#include <stdio.h>

#include <netcdf.h>

#include "read_icon_data.h"

/* ------------------------------------------------ */

static void handle_error(int status) {
  const char *err_string;
  err_string = nc_strerror(status);
  printf ("%s \n", err_string);
  if (status != NC_NOERR) {
    exit(EXIT_FAILURE);
  }
}

/* ------------------------------------------------ */

int open_icon_output ( char * input_file ) {

  int ncid;

  int status = nc_open(input_file, NC_NOWRITE, &ncid);
  if (status != NC_NOERR) handle_error(status);

  return ncid;
}

/* ------------------------------------------------ */

void close_icon_output ( int ncid ) {
    int status = nc_close(ncid);
    if (status != NC_NOERR) handle_error(status);
}

/* ------------------------------------------------ */

void read_icon_dimensions ( int fileId, char * fieldName,
                            struct fieldMetadata * fieldInfo ) {

  int varNdims;
  int varDimids[NC_MAX_VAR_DIMS];   // dimension NetCDF IDs

  int varId;
  nc_type varType;
  int timeId;
  int depthId;
  int cellId;

  int status;

  int i;

  status = nc_inq_varid ( fileId, fieldName, &varId);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_vartype ( fileId, varId, &varType );
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_varndims ( fileId, varId, &varNdims);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_vardimid ( fileId, varId, varDimids);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_dimid ( fileId, "time", &timeId );
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_dimid ( fileId, "depth", &depthId );
  if (status != NC_NOERR) {
     status = nc_inq_dimid ( fileId, "lev", &depthId );
     if (status != NC_NOERR) handle_error(status);
  }

  status = nc_inq_dimid ( fileId, "ncells", &cellId );
  if (status != NC_NOERR) {
     status = nc_inq_dimid ( fileId, "cell", &depthId );
     if (status != NC_NOERR) handle_error(status);
  }

  fieldInfo->start = ( size_t * ) malloc ( varNdims * sizeof (size_t) );
  fieldInfo->count = ( size_t * ) malloc ( varNdims * sizeof (size_t) );

  switch(varType) {
  case NC_FLOAT:
    fieldInfo->varType = FLOAT;
    break;
  case NC_DOUBLE:
    fieldInfo->varType = DOUBLE;
    break;
  default:
    fprintf(stdout, "Datatype not supported %i %i %i \n", varType, NC_FLOAT, NC_DOUBLE);
    exit(EXIT_FAILURE);
  }

  fieldInfo->varId   = varId;

  /* set defaults */

  fieldInfo->nbrLevels    = 1;
  fieldInfo->nbrTimeSteps = 1;
  fieldInfo->nbrHorPoints = 1;
  fieldInfo->timeDim       = -1;
  fieldInfo->levelDim      = -1;
  fieldInfo->horizontalDim = -1;

  for ( i = 0; i < varNdims; i++ ) {

    fieldInfo->start[i] = 0;

    status = nc_inq_dimlen( fileId, varDimids[i], &fieldInfo->count[i]);
    if (status != NC_NOERR) handle_error(status);

  }

  for ( i = 0; i < varNdims; i++ ) {
    if ( varDimids[i] == timeId  ) {
      fieldInfo->nbrTimeSteps = fieldInfo->count[i];
      fieldInfo->timeDim = i;
#ifdef DEBUG
      printf ("number of time steps in file %i \n",  fieldInfo->nbrTimeSteps);
#endif
    }
    if ( varDimids[i] == depthId ) {
      fieldInfo->nbrLevels    = fieldInfo->count[i];
      fieldInfo->levelDim = i;
#ifdef DEBUG
      printf ("number of levels for field   %i \n",  fieldInfo->nbrLevels);
#endif
    }
    if ( varDimids[i] == cellId  ) {
      fieldInfo->nbrHorPoints = fieldInfo->count[i];
      fieldInfo->horizontalDim = i;
#ifdef DEBUG
      printf ("number of horizontal data    %i \n",  fieldInfo->nbrHorPoints);
#endif
    }
  }
}

/* ------------------------------------------------ */

void get_memory ( void ** cell_data, struct fieldMetadata fieldInfo ) {

  switch(fieldInfo.varType) {
  case FLOAT:
    *cell_data = malloc ( fieldInfo.nbrHorPoints * sizeof( float ) );
#ifdef DEBUG
    printf ("Allocated %i elements with datatype NC_FLOAT.\n", fieldInfo.nbrHorPoints);
#endif
    break;
  case DOUBLE:
    *cell_data = malloc ( fieldInfo.nbrHorPoints * sizeof ( double ) );
#ifdef DEBUG
    printf ("Allocated %i elements with datatype NC_DOUBLE.\n", fieldInfo.nbrHorPoints);
#endif
    break;
  default:
    fprintf(stdout, "get_memory: datatype not supported %i %i %i \n", fieldInfo.varType, FLOAT, DOUBLE);
    exit(EXIT_FAILURE);
  }
}

/* ------------------------------------------------ */

void free_memory ( void * data ) {
  free(data);
}

/* ------------------------------------------------ */

void read_icon_timestep_level ( int fileId, void * cell_data, struct fieldMetadata fieldInfo, int timeStep, int level ) {

  /* ... extract one level */

  if ( fieldInfo.levelDim > -1 ) {
    fieldInfo.start[fieldInfo.levelDim] = level-1;
    fieldInfo.count[fieldInfo.levelDim] = 1;
#ifdef DEBUG
    printf ("depth start and count %i %i \n", (int) fieldInfo.start[fieldInfo.levelDim], (int) fieldInfo.count[fieldInfo.levelDim]);
#endif
    if ( level > fieldInfo.nbrLevels ) {
      printf ("A maximum of %i levels/layers is available for the selected field! \n",fieldInfo.nbrLevels);
      exit(EXIT_FAILURE);
    }
  }

  /* ... extract one timestep */

  if ( fieldInfo.timeDim > -1 ) {
    fieldInfo.start[fieldInfo.timeDim] = timeStep;
    fieldInfo.count[fieldInfo.timeDim] = 1;
#ifdef DEBUG
    printf ("time  start and count %i %i \n", (int) fieldInfo.start[fieldInfo.timeDim], (int) fieldInfo.count[fieldInfo.timeDim]);
#endif
  }

  /* ... extract one horizontal data set */

  if ( fieldInfo.horizontalDim > -1 ) {
    fieldInfo.start[fieldInfo.horizontalDim] = 0 ;
    fieldInfo.count[fieldInfo.horizontalDim] = fieldInfo.nbrHorPoints ;
#ifdef DEBUG
    printf ("cell  start and count %i %i \n", (int) fieldInfo.start[fieldInfo.horizontalDim], (int) fieldInfo.count[fieldInfo.horizontalDim]);
#endif
  }
#ifdef DEBUG
  printf (" -------------------- \n ");
#endif
  /* ... get data from netcdf file */

  int status = nc_get_vara ( fileId, fieldInfo.varId, fieldInfo.start, fieldInfo.count, cell_data);
  if (status != NC_NOERR) handle_error(status);

}

