#include "yac_config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <math.h>

#include <netcdf.h>

#include "grid.h"
#include "grid_unstruct.h"
#include "dep_list.h"
#include "utils.h"
#include "read_icon_grid.h"

static void open_icon_file    ( int *ncid, const char *filename);

static void close_icon_file   ( int ncid );

static void get_icon_vertices ( int ncid,
                                int *vertex_type,  // data type for vertex coordinates
                                void **vertex_lon, // longitude of vertex coordinates
                                void **vertex_lat, // latitude of vertex coordinates
                                size_t *nbr_vertices );

static void get_icon_cell_center ( int ncid,
                                   int *cell_type,  // data type for cell coordinates
                                   void **cell_lon, // longitude of cell coordinates
                                   void **cell_lat, // latitude of cell coordinates
                                   size_t *nbr_cells );

static void get_icon_cell_mask ( int ncid,
                                 int *cell_type,  // data type for cell mask
                                 int **cell_mask, // integer cell mask
                                 size_t *nbr_cells );

static void get_icon_connect  ( int ncid,
                                int **vertex_of_cell,
                                size_t **dimlen );

static void handle_error(int status);

void read_part_icon_grid_information(const char * filename, int * nbr_vertices,
                                     int * nbr_cells, int ** num_vertices_per_cell,
                                     int ** cell_to_vertex, double ** x_vertices,
                                     double ** y_vertices, double ** x_cells,
                                     double ** y_cells, int ** global_cell_id,
                                     int ** cell_mask,
                                     int ** global_cell_id_rank, int ** global_corner_id,
                                     int ** global_corner_id_rank, int rank, int size) {

 // read the global grid
  read_icon_grid_information(filename, nbr_vertices, nbr_cells,
                             num_vertices_per_cell, cell_to_vertex,
                             x_vertices, y_vertices, x_cells, y_cells, cell_mask);

  // determine local range
  int local_start = ((unsigned long)(*nbr_cells) * (unsigned long)rank) /
                    (unsigned long)size;
  int num_local_cells = ((unsigned long)(*nbr_cells) * (unsigned long)(rank+1)) /
                        (unsigned long)size - (unsigned long)local_start;

  // mask for required vertices and cells
  int * required_vertices = calloc(*nbr_vertices, sizeof(*required_vertices));
  int * required_cells = calloc(*nbr_cells, sizeof(*required_cells));

  int offset = 0;
  for (int i = 0; i < local_start; ++i)
    offset += (*num_vertices_per_cell)[i];

  // mark all local cells and their vertices as required
  for (int i = local_start; i < local_start + num_local_cells; ++i) {

    required_cells[i] = 1;
    for (int j = 0; j < (*num_vertices_per_cell)[i]; ++j)
      required_vertices[(*cell_to_vertex)[offset+j]] = 1;

    offset += (*num_vertices_per_cell)[i];
  }

  // mask for halo cells
  int * cell_is_on_rank = malloc(*nbr_cells * sizeof(*cell_is_on_rank));

  // mark all halo cells as required and generate cell_is_on_rank
  offset = 0;
  for (int i = 0; i < *nbr_cells; ++i) {

    if (!required_cells[i]) {

      for (int j = 0; j < (*num_vertices_per_cell)[i]; ++j) {

        if (required_vertices[(*cell_to_vertex)[offset+j]]) {

          required_cells[i] = 1;
          cell_is_on_rank[i] = ((unsigned long)i * (unsigned long)size +
                                (unsigned long)size - 1) /
                               (unsigned long)(*nbr_cells);
          break;
        }
      }

    } else {
      cell_is_on_rank[i] = -1;
    }

    offset += (*num_vertices_per_cell)[i];
  }

  // mask for halo vertices
  int * vertex_is_on_rank = malloc(*nbr_vertices * sizeof(*vertex_is_on_rank));

  // mark all halo vertices as required and generate vertex_is_on_rank
  offset = 0;
  for (int i = 0; i < *nbr_vertices; ++i)
    if (required_vertices[i])
      vertex_is_on_rank[i] = -1;
  for (int i = 0; i < *nbr_cells; ++i) {

    if (required_cells[i] && cell_is_on_rank[i] != -1) {

      for (int j = 0; j < (*num_vertices_per_cell)[i]; ++j) {

        if (!required_vertices[(*cell_to_vertex)[offset+j]]) {

          required_vertices[(*cell_to_vertex)[offset+j]] = 1;
          vertex_is_on_rank[(*cell_to_vertex)[offset+j]] = cell_is_on_rank[i];
        }
      }
    }

    offset += (*num_vertices_per_cell)[i];
  }

  // count the number of cells and vertices
  int part_num_vertices = 0;
  int part_num_cells = 0;
  for (int i = 0; i < *nbr_vertices; ++i)
    if (required_vertices[i])
      part_num_vertices++;
  for (int i = 0; i < *nbr_cells; ++i)
    if(required_cells[i])
      part_num_cells++;

  *global_cell_id = malloc(part_num_cells * sizeof(**global_cell_id));
  *global_cell_id_rank = malloc(part_num_cells * sizeof(**global_cell_id_rank));
  *global_corner_id = malloc(part_num_vertices * sizeof(**global_corner_id));
  *global_corner_id_rank = malloc(part_num_vertices * sizeof(**global_corner_id_rank));

  // generate final vertex data
  part_num_vertices = 0;
  int * global_to_local_vertex = malloc(*nbr_vertices * sizeof(*global_to_local_vertex));
  for (int i = 0; i < *nbr_vertices; ++i) {

    if (required_vertices[i]) {

      (*global_corner_id)[part_num_vertices] = i;
      (*global_corner_id_rank)[part_num_vertices] = vertex_is_on_rank[i];
      (*x_vertices)[part_num_vertices] = (*x_vertices)[i];
      (*y_vertices)[part_num_vertices] = (*y_vertices)[i];
      global_to_local_vertex[i] = part_num_vertices;
      part_num_vertices++;
    }
  }

  *x_vertices = realloc(*x_vertices, part_num_vertices * sizeof(**x_vertices));
  *y_vertices = realloc(*y_vertices, part_num_vertices * sizeof(**y_vertices));
  free(vertex_is_on_rank);
  *nbr_vertices = part_num_vertices;
  free(required_vertices);

  // generate final cell data
  int num_cell_vertex_dependencies = 0;
  part_num_cells = 0;
  offset = 0;
  for (int i = 0; i < *nbr_cells; ++i) {

    if (required_cells[i]) {

      (*global_cell_id)[part_num_cells] = i;
      (*global_cell_id_rank)[part_num_cells] = cell_is_on_rank[i];
      (*x_cells)[part_num_cells] = (*x_cells)[i];
      (*y_cells)[part_num_cells] = (*y_cells)[i];
      (*cell_mask)[part_num_cells] = (*cell_mask)[i];

      for (int j = 0; j < (*num_vertices_per_cell)[i]; ++j)
        (*cell_to_vertex)[num_cell_vertex_dependencies++] =
          global_to_local_vertex[(*cell_to_vertex)[offset+j]];

      (*num_vertices_per_cell)[part_num_cells] = (*num_vertices_per_cell)[i];

      part_num_cells++;
    }

    offset += (*num_vertices_per_cell)[i];
  }

  *x_cells   = realloc(*x_cells, part_num_cells * sizeof(**x_cells));
  *y_cells   = realloc(*y_cells, part_num_cells * sizeof(**y_cells));
  *cell_mask = realloc(*cell_mask, part_num_cells * sizeof(**cell_mask));

  *num_vertices_per_cell = realloc(*num_vertices_per_cell, part_num_cells *
                                   sizeof(**num_vertices_per_cell));
  *cell_to_vertex = realloc(*cell_to_vertex, num_cell_vertex_dependencies *
                            sizeof(**cell_to_vertex));
  free(cell_is_on_rank);
  *nbr_cells = part_num_cells;
  free(required_cells);
  free(global_to_local_vertex);
}

void read_icon_grid_information(const char * filename, int * nbr_vertices,
                                int * nbr_cells, int ** num_vertices_per_cell,
                                int ** cell_to_vertex, double ** x_vertices,
                                double ** y_vertices, double ** x_cells,
                                double ** y_cells, int ** cell_mask) {

   int ncid;

   int * vertex_of_cell;

   /* Open file */

   open_icon_file ( &ncid, filename );

   /* Get vertex longitudes and latitudes of cells */

   {
      void * temp_vertex_lon;
      void * temp_vertex_lat;
      size_t temp_nbr_vertices;
      int vertex_type;

      get_icon_vertices ( ncid, &vertex_type, &temp_vertex_lon, &temp_vertex_lat, &temp_nbr_vertices );

      *nbr_vertices = temp_nbr_vertices;

      if (vertex_type == NC_FLOAT) {

         *x_vertices = malloc(temp_nbr_vertices * sizeof(**x_vertices));
         *y_vertices = malloc(temp_nbr_vertices * sizeof(**y_vertices));

         for (size_t i = 0; i < temp_nbr_vertices; ++i) {

            (*x_vertices)[i] = ((float*)temp_vertex_lon)[i];
            (*y_vertices)[i] = ((float*)temp_vertex_lat)[i];
         }
         free(temp_vertex_lon);
         free(temp_vertex_lat);
      } else {
         *x_vertices = temp_vertex_lon;
         *y_vertices = temp_vertex_lat;
      }
   }

   /* Get cell center longitudes and latitudes of cells */

   {
      void * temp_cell_lon;
      void * temp_cell_lat;
      size_t temp_nbr_cells;
      int cell_type;

      get_icon_cell_center ( ncid, &cell_type, &temp_cell_lon, &temp_cell_lat, &temp_nbr_cells );

      *nbr_cells = temp_nbr_cells;

      if (cell_type == NC_FLOAT) {

         *x_cells = malloc(temp_nbr_cells * sizeof(**x_cells));
         *y_cells = malloc(temp_nbr_cells * sizeof(**y_cells));

         for (size_t i = 0; i < temp_nbr_cells; ++i) {

            (*x_cells)[i] = ((float*)temp_cell_lon)[i];
            (*y_cells)[i] = ((float*)temp_cell_lat)[i];
         }
         free(temp_cell_lon);
         free(temp_cell_lat);
      } else {
         *x_cells = temp_cell_lon;
         *y_cells = temp_cell_lat;
      }
   }

   /* Get mask of cells */

   {
      int * temp_cell_mask;
      size_t temp_nbr_cells;
      int cell_type;

      get_icon_cell_mask ( ncid, &cell_type, &temp_cell_mask, &temp_nbr_cells );

      *nbr_cells = temp_nbr_cells;

      *cell_mask = temp_cell_mask;
   }

   /* Get relations between vertices and cells */
  
   {
      size_t * dimlen;

      get_icon_connect ( ncid, &vertex_of_cell, &dimlen);

      *nbr_cells = dimlen[1];

      free(dimlen);
   }

   close_icon_file ( ncid );

   //-------------------------------------------------------------------------//

   *num_vertices_per_cell = malloc(*nbr_cells * sizeof(**num_vertices_per_cell));
   for (int i = 0; i < *nbr_cells; (*num_vertices_per_cell)[i++] = 3);

   *cell_to_vertex = malloc(*nbr_cells * 3 * sizeof(**cell_to_vertex));

   // Unfortunately the data is only available in Fortran order
   for (int i = 0; i < *nbr_cells; ++i) {

      (*cell_to_vertex)[3*i+0] = vertex_of_cell[i+0*(*nbr_cells)] - 1;
      (*cell_to_vertex)[3*i+1] = vertex_of_cell[i+1*(*nbr_cells)] - 1;
      (*cell_to_vertex)[3*i+2] = vertex_of_cell[i+2*(*nbr_cells)] - 1;
   }

   free(vertex_of_cell);
}

#ifdef USE_MPI

// taken from scales-ppm library
// https://www.dkrz.de/redmine/projects/scales-ppm
static inline int
partition_idx_from_element_idx(unsigned element_idx, unsigned num_elements,
                               int num_partitions) {

  return (int)((((unsigned long)element_idx) * ((unsigned long)num_partitions) +
                (unsigned long)num_partitions - 1) /
               ((unsigned long)num_elements));
}

void read_icon_grid_information_parallel(
  const char * filename, MPI_Comm comm, int * nbr_vertices, int * nbr_cells,
  int ** num_vertices_per_cell, int ** cell_to_vertex, int ** global_cell_ids,
  int ** cell_owner, int ** global_vertex_ids, int ** vertex_owner,
  double ** x_vertices, double ** y_vertices,
  double ** x_cells, double ** y_cells, int ** cell_msk,
  unsigned io_proc_step) {

  int comm_rank, comm_size;

  MPI_Comm_rank(comm, &comm_rank);
  MPI_Comm_size(comm, &comm_size);

  if (io_proc_step >= (unsigned)comm_size)
    io_proc_step = (unsigned)sqrtf((float)comm_size);
  io_proc_step = MAX(1, io_proc_step);

  size_t ncells, nvertices;

  size_t read_local_start_cell = 0;
  size_t read_num_local_cells = 0;
  size_t read_local_start_vertex = 0;
  size_t read_num_local_vertices = 0;
  double * read_cell_lon = NULL;
  double * read_cell_lat = NULL;
  int * read_cell_mask = NULL;
  double * read_vertex_lon = NULL;
  double * read_vertex_lat = NULL;
  int * read_dist_vertex_of_cell = NULL;
  int * read_dist_cells_of_vertex = NULL;

  if (((unsigned long)comm_rank % io_proc_step) == 0) {

    unsigned long io_proc_idx =
      (unsigned long)comm_rank/(unsigned long)io_proc_step;
    unsigned long num_io_procs =
      ((unsigned long)comm_size + (unsigned long)io_proc_step - 1)/
      (unsigned long)io_proc_step;

    // open file
    int ncid;
    open_icon_file ( &ncid, filename );

    // get number of cells and vertices
    int dim_id;
    handle_error(nc_inq_dimid(ncid, "cell", &dim_id));
    handle_error(nc_inq_dimlen(ncid, dim_id, &ncells));
    handle_error(nc_inq_dimid(ncid, "vertex", &dim_id));
    handle_error(nc_inq_dimlen(ncid, dim_id, &nvertices));

    // determine local range for cell and vertex data
    read_local_start_cell =
      ((unsigned long)ncells * io_proc_idx) / num_io_procs;
    read_num_local_cells =
      ((unsigned long)ncells * (io_proc_idx+1)) / num_io_procs -
      (unsigned long)read_local_start_cell;
    read_local_start_vertex =
      ((unsigned long)nvertices * io_proc_idx) / num_io_procs;
    read_num_local_vertices =
      ((unsigned long)nvertices * (io_proc_idx+1)) / num_io_procs -
      (unsigned long)read_local_start_vertex;

    // read basic grid data (each process its individual part)
    read_cell_lon = malloc(read_num_local_cells * sizeof(*read_cell_lon));
    read_cell_lat = malloc(read_num_local_cells * sizeof(*read_cell_lat));
    read_cell_mask = malloc(read_num_local_cells * sizeof(*read_cell_mask));
    read_vertex_lon = malloc(read_num_local_vertices * sizeof(*read_vertex_lon));
    read_vertex_lat = malloc(read_num_local_vertices * sizeof(*read_vertex_lat));
    read_dist_vertex_of_cell =
      malloc(read_num_local_cells * 3 * sizeof(*read_dist_vertex_of_cell));
    read_dist_cells_of_vertex =
      malloc(read_num_local_vertices * 6 * sizeof(*read_dist_cells_of_vertex));
    int varid;
    handle_error(nc_inq_varid(ncid, "clon", &varid));
    handle_error(nc_get_vara_double(ncid, varid, &read_local_start_cell,
                                    &read_num_local_cells, read_cell_lon));
    handle_error(nc_inq_varid(ncid, "clat", &varid));
    handle_error(nc_get_vara_double(ncid, varid, &read_local_start_cell,
                                    &read_num_local_cells, read_cell_lat));
    handle_error(nc_inq_varid(ncid, "cell_sea_land_mask", &varid));
    handle_error(nc_get_vara_int(ncid, varid, &read_local_start_cell,
                                 &read_num_local_cells, read_cell_mask));
    handle_error(nc_inq_varid(ncid, "vlon", &varid));
    handle_error(nc_get_vara_double(ncid, varid, &read_local_start_vertex,
                                    &read_num_local_vertices, read_vertex_lon));
    handle_error(nc_inq_varid(ncid, "vlat", &varid));
    handle_error(nc_get_vara_double(ncid, varid, &read_local_start_vertex,
                                    &read_num_local_vertices, read_vertex_lat));
    {
      int * buffer = malloc(MAX(3*read_num_local_cells,
                                6*read_num_local_vertices) * sizeof(*buffer));
      size_t tmp_start[2], tmp_count[2];
      tmp_start[0] = 0, tmp_start[1] = read_local_start_cell;
      tmp_count[0] = 3, tmp_count[1] = read_num_local_cells;
      handle_error(nc_inq_varid(ncid, "vertex_of_cell", &varid));
      handle_error(nc_get_vara_int(ncid, varid, tmp_start, tmp_count, buffer));
      for (size_t i = 0; i < read_num_local_cells; ++i)
        for (int j = 0; j < 3; ++j)
          read_dist_vertex_of_cell[i * 3 + j] =
            buffer[i + j * read_num_local_cells];
      tmp_start[0] = 0, tmp_start[1] = read_local_start_vertex;
      tmp_count[0] = 6, tmp_count[1] = read_num_local_vertices;
      handle_error(nc_inq_varid(ncid, "cells_of_vertex", &varid));
      handle_error(nc_get_vara_int(ncid, varid, tmp_start, tmp_count, buffer));
      for (size_t i = 0; i < read_num_local_vertices; ++i)
        for (int j = 0; j < 6; ++j)
          read_dist_cells_of_vertex[i * 6 + j] =
            buffer[i + j * read_num_local_vertices];
      free(buffer);

      // adjust for c indices
      for (size_t i = 0; i < read_num_local_cells * 3; ++i)
        if (read_dist_vertex_of_cell[i] > 0) read_dist_vertex_of_cell[i]--;
      for (size_t i = 0; i < read_num_local_vertices * 6; ++i)
        read_dist_cells_of_vertex[i]--;
    }

    close_icon_file(ncid);
  } else {
    read_cell_lon = malloc(1 * sizeof(*read_cell_lon));
    read_cell_lat = malloc(1 * sizeof(*read_cell_lat));
    read_cell_mask = malloc(1 * sizeof(*read_cell_mask));
    read_vertex_lon = malloc(1 * sizeof(*read_vertex_lon));
    read_vertex_lat = malloc(1 * sizeof(*read_vertex_lat));
    read_dist_vertex_of_cell = malloc(1 * sizeof(*read_dist_vertex_of_cell));
    read_dist_cells_of_vertex = malloc(1 * sizeof(*read_dist_cells_of_vertex));
  }

  {
    int tmp;
    if (comm_rank == 0) tmp = (int)ncells;
    MPI_Bcast(&tmp, 1, MPI_INT, 0, comm);
    ncells = (size_t)tmp;
    if (comm_rank == 0) tmp = (int)nvertices;
    MPI_Bcast(&tmp, 1, MPI_INT, 0, comm);
    nvertices = (size_t)tmp;
  }

  // determine local range for cell and vertex data
  size_t local_start_cell =
    ((unsigned long)ncells * (unsigned long)comm_rank) /
    (unsigned long)comm_size;
  size_t num_local_cells =
    ((unsigned long)ncells * (unsigned long)(comm_rank+1)) /
    (unsigned long)comm_size - (unsigned long)local_start_cell;
  size_t local_start_vertex =
    ((unsigned long)nvertices * (unsigned long)comm_rank) /
    (unsigned long)comm_size;
  size_t num_local_vertices =
    ((unsigned long)nvertices * (unsigned long)(comm_rank+1)) /
    (unsigned long)comm_size - (unsigned long)local_start_vertex;

  // redistribute basic cell data (from io decomposition)
  double * cell_lon = malloc(num_local_cells * sizeof(*cell_lon));
  double * cell_lat = malloc(num_local_cells * sizeof(*cell_lat));
  int * cell_mask = malloc(num_local_cells * sizeof(*cell_mask));
  int * dist_vertex_of_cell =
    malloc(num_local_cells * 3 * sizeof(*dist_vertex_of_cell));
  {
    int * send_count = calloc(comm_size, sizeof(*send_count));
    int * recv_count = calloc(comm_size, sizeof(*recv_count));

    for (unsigned i = 0; i < read_num_local_cells; ++i)
      send_count[
        partition_idx_from_element_idx(
          read_local_start_cell + i, ncells, comm_size)]++;

    MPI_Alltoall(send_count, 1, MPI_INT, recv_count, 1, MPI_INT, comm);

    int * send_displ = malloc(comm_size * sizeof(*send_displ));
    int * recv_displ = malloc(comm_size * sizeof(*recv_displ));
    int send_accum = 0, recv_accum = 0;
    for (int i = 0; i < comm_size; ++i) {
      send_displ[i] = send_accum;
      recv_displ[i] = recv_accum;
      send_accum += send_count[i];
      recv_accum += recv_count[i];
    }

    MPI_Alltoallv(read_cell_lon, send_count, send_displ, MPI_DOUBLE,
                  cell_lon, recv_count, recv_displ, MPI_DOUBLE, comm);
    MPI_Alltoallv(read_cell_lat, send_count, send_displ, MPI_DOUBLE,
                  cell_lat, recv_count, recv_displ, MPI_DOUBLE, comm);
    MPI_Alltoallv(read_cell_mask, send_count, send_displ, MPI_INT,
                  cell_mask, recv_count, recv_displ, MPI_INT, comm);

    for (int i = 0; i < comm_size; ++i) {
      send_count[i] *= 3;
      send_displ[i] *= 3;
      recv_count[i] *= 3;
      recv_displ[i] *= 3;
    }

    MPI_Alltoallv(read_dist_vertex_of_cell, send_count, send_displ, MPI_INT,
                  dist_vertex_of_cell, recv_count, recv_displ, MPI_INT, comm);

    free(recv_displ);
    free(send_displ);
    free(recv_count);
    free(send_count);
    free(read_cell_lon);
    free(read_cell_lat);
    free(read_cell_mask);
    free(read_dist_vertex_of_cell);
  }

  // redistribute basic vertex data (from io decomposition)
  double * vertex_lon = malloc(num_local_vertices * sizeof(*vertex_lon));
  double * vertex_lat = malloc(num_local_vertices * sizeof(*vertex_lat));
  int * dist_cells_of_vertex =
    malloc(num_local_vertices * 6 * sizeof(*dist_cells_of_vertex));
  {
    int * send_count = calloc(comm_size, sizeof(*send_count));
    int * recv_count = calloc(comm_size, sizeof(*recv_count));

    for (unsigned i = 0; i < read_num_local_vertices; ++i)
      send_count[
        partition_idx_from_element_idx(
          read_local_start_vertex + i, nvertices, comm_size)]++;

    MPI_Alltoall(send_count, 1, MPI_INT, recv_count, 1, MPI_INT, comm);

    int * send_displ = malloc(comm_size * sizeof(*send_displ));
    int * recv_displ = malloc(comm_size * sizeof(*recv_displ));
    int send_accum = 0, recv_accum = 0;
    for (int i = 0; i < comm_size; ++i) {
      send_displ[i] = send_accum;
      recv_displ[i] = recv_accum;
      send_accum += send_count[i];
      recv_accum += recv_count[i];
    }

    MPI_Alltoallv(read_vertex_lon, send_count, send_displ, MPI_DOUBLE,
                  vertex_lon, recv_count, recv_displ, MPI_DOUBLE, comm);
    MPI_Alltoallv(read_vertex_lat, send_count, send_displ, MPI_DOUBLE,
                  vertex_lat, recv_count, recv_displ, MPI_DOUBLE, comm);

    for (int i = 0; i < comm_size; ++i) {
      send_count[i] *= 6;
      send_displ[i] *= 6;
      recv_count[i] *= 6;
      recv_displ[i] *= 6;
    }

    MPI_Alltoallv(read_dist_cells_of_vertex, send_count, send_displ, MPI_INT,
                  dist_cells_of_vertex, recv_count, recv_displ, MPI_INT, comm);

    free(recv_displ);
    free(send_displ);
    free(recv_count);
    free(send_count);
    free(read_vertex_lon);
    free(read_vertex_lat);
    free(read_dist_cells_of_vertex);
  }

  // determine required vertices for core cells
  unsigned num_core_vertices = num_local_cells * 3;
  int * core_vertices = malloc(num_core_vertices * sizeof(*core_vertices));
  {
    memcpy(core_vertices, dist_vertex_of_cell,
           num_core_vertices * sizeof(*core_vertices));
    yac_quicksort_index(core_vertices, num_core_vertices, NULL);
    yac_remove_duplicates_int(core_vertices, &num_core_vertices);
    core_vertices =
      realloc(core_vertices, num_core_vertices * sizeof(*core_vertices));
  }

  // get cells_of_vertex for core vertices and compute dist_vertex_owner
  int * cells_of_vertex_core =
    malloc(num_core_vertices * 6 * sizeof(*cells_of_vertex_core));
  int * dist_vertex_owner =
    malloc(num_local_vertices * sizeof(*dist_vertex_owner));
  {
    int * send_count = calloc(comm_size, sizeof(*send_count));
    int * recv_count = calloc(comm_size, sizeof(*recv_count));

    for (unsigned i = 0; i < num_core_vertices; ++i)
      send_count[((unsigned long)(core_vertices[i]) *
                  (unsigned long)comm_size + (unsigned long)comm_size - 1) /
                 (unsigned long)nvertices]++;

    MPI_Alltoall(send_count, 1, MPI_INT, recv_count, 1, MPI_INT, comm);

    int * send_displ = malloc(comm_size * sizeof(*send_displ));
    int * recv_displ = malloc(comm_size * sizeof(*recv_displ));
    int send_accum = 0, recv_accum = 0;
    for (int i = 0; i < comm_size; ++i) {
      send_displ[i] = send_accum;
      recv_displ[i] = recv_accum;
      send_accum += send_count[i];
      recv_accum += recv_count[i];
    }

    int num_core_vertices_remote = 0;
    for (int i = 0; i < comm_size; ++i)
      num_core_vertices_remote += recv_count[i];

    int * remote_vertex_buffer =
      malloc(num_core_vertices_remote * sizeof(*remote_vertex_buffer));

    MPI_Alltoallv(core_vertices, send_count, send_displ, MPI_INT,
                  remote_vertex_buffer, recv_count, recv_displ, MPI_INT, comm);

    for (size_t i = 0; i < num_local_vertices; ++i) dist_vertex_owner[i] = -1;

    for (int i = 0, j = 0; i < comm_size; ++i)
      for (int k = 0; k < recv_count[i]; ++k, ++j)
        dist_vertex_owner[remote_vertex_buffer[j] - local_start_vertex] = i;

    int * send_cell_of_vertex =
      malloc(num_core_vertices_remote * 6 * sizeof(*send_cell_of_vertex));

    for (int i = 0, l = 0, m = 0; i < comm_size; ++i) {
      for (int j = 0; j < recv_count[i]; ++j, ++l) {
        int idx = remote_vertex_buffer[l] - local_start_vertex;
        for (int k = 0; k < 6; ++k, ++m)
          send_cell_of_vertex[m] = dist_cells_of_vertex[idx * 6 + k];
      }
    }
    free(dist_cells_of_vertex);

    for (int i = 0; i < comm_size; ++i) {
      send_count[i] *= 6;
      send_displ[i] *= 6;
      recv_count[i] *= 6;
      recv_displ[i] *= 6;
    }

    MPI_Alltoallv(send_cell_of_vertex, recv_count, recv_displ, MPI_INT,
                  cells_of_vertex_core, send_count, send_displ, MPI_INT, comm);

    free(send_cell_of_vertex);
    free(remote_vertex_buffer);
    free(recv_displ);
    free(send_displ);
    free(recv_count);
    free(send_count);
  }

  // determine halo cells
  int * halo_cells;
  unsigned num_halo_cells = 0;
  {
    int * tmp_cells = cells_of_vertex_core;
    unsigned num_tmp_cells = num_core_vertices * 6;

    yac_quicksort_index(tmp_cells, num_tmp_cells, NULL);
    yac_remove_duplicates_int(tmp_cells, &num_tmp_cells);

    if ((num_tmp_cells > 0) && (tmp_cells[0] == -1)) {
      num_tmp_cells--;
      tmp_cells++;
    }

    halo_cells = malloc((num_tmp_cells - num_local_cells) * sizeof(*halo_cells));

    unsigned i = 0;
    for (; i < num_tmp_cells && tmp_cells[i] < (int)local_start_cell; ++i)
      halo_cells[num_halo_cells++] = tmp_cells[i];
    i += num_local_cells;
    for (; i < num_tmp_cells; ++i) halo_cells[num_halo_cells++] = tmp_cells[i];

    assert(num_halo_cells == num_tmp_cells - num_local_cells);

    free(cells_of_vertex_core);
  }

  // determine all vertices and get coordinates of halo cells
  unsigned num_all_local_vertices = num_halo_cells * 3 + num_core_vertices;
  int * all_cell_to_vertex;
  int * all_local_vertices = realloc(core_vertices, num_all_local_vertices *
                                     sizeof(*all_local_vertices));
  cell_lon = realloc(cell_lon, (num_local_cells + num_halo_cells) *
                     sizeof(*cell_lon));
  cell_lat = realloc(cell_lat, (num_local_cells + num_halo_cells) *
                     sizeof(*cell_lat));
  cell_mask = realloc(cell_mask, (num_local_cells + num_halo_cells) *
                      sizeof(*cell_mask));
  {
    int * send_count = calloc(comm_size, sizeof(*send_count));
    int * recv_count = calloc(comm_size, sizeof(*recv_count));

    for (unsigned i = 0; i < num_halo_cells; ++i)
      send_count[((unsigned long)(halo_cells[i]) *
                  (unsigned long)comm_size + (unsigned long)comm_size - 1) /
                 (unsigned long)ncells]++;

    MPI_Alltoall(send_count, 1, MPI_INT, recv_count, 1, MPI_INT, comm);

    int * send_displ = malloc(comm_size * sizeof(*send_displ));
    int * recv_displ = malloc(comm_size * sizeof(*recv_displ));
    int send_accum = 0, recv_accum = 0;
    for (int i = 0; i < comm_size; ++i) {
      send_displ[i] = send_accum;
      recv_displ[i] = recv_accum;
      send_accum += send_count[i];
      recv_accum += recv_count[i];
    }

    int num_halo_cells_remote = 0;
    for (int i = 0; i < comm_size; ++i)
      num_halo_cells_remote += recv_count[i];

    int * remote_halo_cell_buffer =
      malloc(num_halo_cells_remote * sizeof(*remote_halo_cell_buffer));

    MPI_Alltoallv(halo_cells, send_count, send_displ, MPI_INT,
                  remote_halo_cell_buffer, recv_count, recv_displ, MPI_INT,
                  comm);

    int * send_halo_cell_vertices =
      malloc(num_halo_cells_remote * 3 * sizeof(*send_halo_cell_vertices));
    double * send_cell_lon =
      malloc(num_halo_cells_remote * sizeof(*send_cell_lon));
    double * send_cell_lat =
      malloc(num_halo_cells_remote * sizeof(*send_cell_lat));
    int * send_cell_mask =
      malloc(num_halo_cells_remote * sizeof(*send_cell_mask));

    for (int i = 0, l = 0, m = 0; i < comm_size; ++i) {
      for (int j = 0; j < recv_count[i]; ++j, ++l) {
        int idx = remote_halo_cell_buffer[l] - local_start_cell;
        for (int k = 0; k < 3; ++k, ++m)
          send_halo_cell_vertices[m] = dist_vertex_of_cell[idx * 3 + k];
        send_cell_lon[l] = cell_lon[idx];
        send_cell_lat[l] = cell_lat[idx];
        send_cell_mask[l] = cell_mask[idx];
      }
    }

    MPI_Alltoallv(send_cell_lon, recv_count, recv_displ, MPI_DOUBLE,
                  cell_lon + num_local_cells, send_count, send_displ,
                  MPI_DOUBLE, comm);
    MPI_Alltoallv(send_cell_lat, recv_count, recv_displ, MPI_DOUBLE,
                  cell_lat + num_local_cells, send_count, send_displ,
                  MPI_DOUBLE, comm);
    MPI_Alltoallv(send_cell_mask, recv_count, recv_displ, MPI_INT,
                  cell_mask + num_local_cells, send_count, send_displ,
                  MPI_INT, comm);

    for (int i = 0; i < comm_size; ++i) {
      send_count[i] *= 3;
      send_displ[i] *= 3;
      recv_count[i] *= 3;
      recv_displ[i] *= 3;
    }

    all_cell_to_vertex =
      realloc(dist_vertex_of_cell, (num_local_cells + num_halo_cells) * 3 *
              sizeof(*all_cell_to_vertex));

    MPI_Alltoallv(send_halo_cell_vertices, recv_count, recv_displ, MPI_INT,
                  all_cell_to_vertex + num_local_cells * 3, send_count,
                  send_displ, MPI_INT, comm);

    memcpy(all_local_vertices + num_core_vertices,
           all_cell_to_vertex + num_local_cells * 3,
           num_halo_cells * 3 * sizeof(*all_local_vertices));

    free(send_cell_mask);
    free(send_cell_lat);
    free(send_cell_lon);
    free(send_halo_cell_vertices);
    free(remote_halo_cell_buffer);
    free(recv_displ);
    free(send_displ);
    free(recv_count);
    free(send_count);

    yac_quicksort_index(all_local_vertices, num_all_local_vertices, NULL);
    yac_remove_duplicates_int(all_local_vertices, &num_all_local_vertices);
  }

  // determine owner and coordinates for all vertices
  double * all_vertex_lon =
    malloc(num_all_local_vertices * sizeof(*all_vertex_lon));
  double * all_vertex_lat =
    malloc(num_all_local_vertices * sizeof(*all_vertex_lat));
  int * all_local_vertices_owner =
    malloc(num_all_local_vertices * sizeof(*all_local_vertices_owner));
  {
    int * send_count = calloc(comm_size, sizeof(*send_count));
    int * recv_count = calloc(comm_size, sizeof(*recv_count));

    for (unsigned i = 0; i < num_all_local_vertices; ++i)
      send_count[((unsigned long)(all_local_vertices[i]) *
                  (unsigned long)comm_size + (unsigned long)comm_size - 1) /
                 (unsigned long)nvertices]++;

    MPI_Alltoall(send_count, 1, MPI_INT, recv_count, 1, MPI_INT, comm);

    int * send_displ = malloc(comm_size * sizeof(*send_displ));
    int * recv_displ = malloc(comm_size * sizeof(*recv_displ));
    int send_accum = 0, recv_accum = 0;
    for (int i = 0; i < comm_size; ++i) {
      send_displ[i] = send_accum;
      recv_displ[i] = recv_accum;
      send_accum += send_count[i];
      recv_accum += recv_count[i];
    }

    int num_all_local_vertices_remote = 0;
    for (int i = 0; i < comm_size; ++i)
      num_all_local_vertices_remote += recv_count[i];

    int * remote_vertex_buffer =
      malloc(num_all_local_vertices_remote * sizeof(*remote_vertex_buffer));

    MPI_Alltoallv(all_local_vertices, send_count, send_displ, MPI_INT,
                  remote_vertex_buffer, recv_count, recv_displ, MPI_INT, comm);

    int * send_vertex_owner = remote_vertex_buffer;
    double * send_vertex_lon =
      malloc(num_all_local_vertices_remote * sizeof(*send_vertex_lon));
    double * send_vertex_lat =
      malloc(num_all_local_vertices_remote * sizeof(*send_vertex_lat));

    for (int i = 0, l = 0; i < comm_size; ++i) {
      for (int j = 0; j < recv_count[i]; ++j, ++l) {
        int idx = remote_vertex_buffer[l] - local_start_vertex;
        send_vertex_owner[l] = dist_vertex_owner[idx];
        send_vertex_lon[l] = vertex_lon[idx];
        send_vertex_lat[l] = vertex_lat[idx];
      }
    }

    free(dist_vertex_owner);
    free(vertex_lon);
    free(vertex_lat);

    MPI_Alltoallv(send_vertex_owner, recv_count, recv_displ, MPI_INT,
                  all_local_vertices_owner, send_count, send_displ, MPI_INT,
                  comm);
    MPI_Alltoallv(send_vertex_lon, recv_count, recv_displ, MPI_DOUBLE,
                  all_vertex_lon, send_count, send_displ, MPI_DOUBLE, comm);
    MPI_Alltoallv(send_vertex_lat, recv_count, recv_displ, MPI_DOUBLE,
                  all_vertex_lat, send_count, send_displ, MPI_DOUBLE, comm);

    free(send_vertex_lat);
    free(send_vertex_lon);
    free(send_vertex_owner);
    free(recv_displ);
    free(send_displ);
    free(recv_count);
    free(send_count);
  }

  // convert global ids within all_cell_to_vertex into local ids
  {
    int * permutation =
      malloc(3 * (num_local_cells + num_halo_cells) * sizeof(*permutation));
    for (unsigned i = 0; i < 3 * (num_local_cells + num_halo_cells); ++i)
      permutation[i] = i;

    yac_quicksort_index(
      all_cell_to_vertex, (num_local_cells + num_halo_cells) * 3, permutation);

    for (unsigned i = 0, j = 0; i < (num_local_cells + num_halo_cells) * 3; ++i) {
      while (all_local_vertices[j] != all_cell_to_vertex[i]) ++j;
      all_cell_to_vertex[i] = j;
    }

    yac_quicksort_index(permutation, (num_local_cells + num_halo_cells) * 3,
                        all_cell_to_vertex);

    free(permutation);
  }

  *nbr_vertices = num_all_local_vertices;
  *nbr_cells = num_local_cells + num_halo_cells;

  *num_vertices_per_cell =
    malloc((num_local_cells + num_halo_cells) * sizeof(**num_vertices_per_cell));
  for (unsigned i = 0; i < num_local_cells + num_halo_cells; ++i)
    (*num_vertices_per_cell)[i] = 3;

  *cell_to_vertex = all_cell_to_vertex;

  *global_cell_ids =
    malloc((num_local_cells + num_halo_cells) * sizeof(**global_cell_ids));
  for (unsigned i = 0; i < num_local_cells; ++i)
    (*global_cell_ids)[i] = local_start_cell + i;
  memcpy(*global_cell_ids + num_local_cells, halo_cells,
         num_halo_cells * sizeof(*halo_cells));
  *global_vertex_ids = realloc(all_local_vertices, num_all_local_vertices *
                               sizeof(*all_local_vertices));

  *cell_owner =
    malloc((num_local_cells + num_halo_cells) * sizeof(**cell_owner));
  for (unsigned i = 0; i < num_local_cells; ++i)
    (*cell_owner)[i] = -1;
  for (unsigned i = 0; i < num_halo_cells; ++i)
    (*cell_owner)[num_local_cells + i] =
      ((unsigned long)(halo_cells[i]) * (unsigned long)comm_size +
      (unsigned long)comm_size - 1) / (unsigned long)ncells;
  free(halo_cells);

  for (unsigned i = 0; i < num_all_local_vertices; ++i)
    if (all_local_vertices_owner[i] == comm_rank)
      all_local_vertices_owner[i] = -1;
  *vertex_owner = all_local_vertices_owner;

  *x_vertices = all_vertex_lon;
  *y_vertices = all_vertex_lat;
  *x_cells = cell_lon;
  *y_cells = cell_lat;
  *cell_msk = cell_mask;
}
#endif // USE_MPI

struct grid * read_icon_grid (char * filename) {

   int nbr_vertices;
   int nbr_cells;
   int * num_vertices_per_cell = NULL;
   int * cell_to_vertex = NULL;
   int * cell_mask = NULL;

   double * x_vertices = NULL;
   double * y_vertices = NULL;
   double * x_cells = NULL;
   double * y_cells = NULL;

   read_icon_grid_information(filename, &nbr_vertices, &nbr_cells,
                              &num_vertices_per_cell, &cell_to_vertex, 
                              &x_vertices, &y_vertices,
                              &x_cells, &y_cells,
                              &cell_mask);

   free(x_cells);
   free(y_cells);
   free(cell_mask);

   size_t * num_vertices_per_cell_ =
      malloc(nbr_cells * sizeof(*num_vertices_per_cell_));
   for (int i = 0; i < nbr_cells; ++i)
      num_vertices_per_cell_[i] = (size_t)num_vertices_per_cell[i];
   free(num_vertices_per_cell);

   struct dep_list * cell_to_vertex_dep_list =
      yac_dep_list_new((size_t)nbr_cells, num_vertices_per_cell_,
                       (unsigned*)cell_to_vertex);

   struct grid * tmp_grid =
      yac_unstruct_grid_new(
         x_vertices, y_vertices, nbr_vertices, cell_to_vertex_dep_list);
   struct grid * grid_copy = yac_copy_grid(tmp_grid);
   yac_delete_grid(tmp_grid);
   yac_dep_list_delete(cell_to_vertex_dep_list);
   free(x_vertices);
   free(y_vertices);

   return grid_copy;
}

static void open_icon_file (int *ncid, const char *filename) {

  int id, status;

  /* open netCDF dataset */

  status = nc_open(filename, NC_NOWRITE, &id);
  if (status != NC_NOERR) handle_error(status);

  *ncid = id;

}

/* ---------------------------------------------------------------- */

static void close_icon_file (int ncid) {

   int status;

  /* close netCDF dataset */

  status = nc_close(ncid);
  if (status != NC_NOERR) handle_error(status);
}

/* ---------------------------------------------------------------- */

static void get_icon_vertices ( int ncid,
                                int  *vertex_type,
                                void **vertex_lon,
                                void **vertex_lat,
                                size_t *nbr_vertices ) {

  int status;

  int glat_id;                        // Various NetCDF IDs
  int glon_id;

  int glon_ndims;                     // number of dimensions in file
  int glat_ndims;

  int glon_dimids[NC_MAX_VAR_DIMS];   // dimension NetCDF IDs
  int glat_dimids[NC_MAX_VAR_DIMS];

  nc_type glon_type;                  // variable type
  nc_type glat_type;

  int nbr_atts;                     // number of attributes

  status = nc_inq_varid (ncid, "vlon", &glon_id);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_varid (ncid, "vlat", &glat_id);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_var (ncid, glon_id, 0, &glon_type, &glon_ndims, glon_dimids, &nbr_atts);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_var (ncid, glat_id, 0, &glat_type, &glat_ndims, glat_dimids, &nbr_atts);
  if (status != NC_NOERR) handle_error(status);

  /* Allocate memory to read in coordinates */

  if ( glon_type != glat_type ) {
    exit(EXIT_FAILURE);
  }

  if (glon_ndims != 1 || glat_ndims != 1) {
    exit(EXIT_FAILURE);
  }

  /* get size of arrays */

  size_t dimlen_lat; // dimension size for latitude
  size_t dimlen_lon; // dimension size for longitude

  status = nc_inq_dimlen(ncid, *glon_dimids, &dimlen_lon);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_dimlen(ncid, *glat_dimids, &dimlen_lat);
  if (status != NC_NOERR) handle_error(status);

  if (dimlen_lon != dimlen_lat)
    exit(EXIT_FAILURE);

  *nbr_vertices = dimlen_lon;

  /* read center coordinates and convert radians into degrees */

  size_t start = 0;

  switch(glon_type) {

  case NC_FLOAT:

    *vertex_lon = (float * ) malloc ( *nbr_vertices * sizeof ( float ) );
    *vertex_lat = (float * ) malloc ( *nbr_vertices * sizeof ( float ) );

    status = nc_get_vara (ncid, glon_id, &start, nbr_vertices, (*vertex_lon));
    if (status != NC_NOERR) handle_error(status);

    status = nc_get_vara (ncid, glat_id, &start, nbr_vertices, (*vertex_lat));
    if (status != NC_NOERR) handle_error(status);

    break;

  case NC_DOUBLE:

    *vertex_lon = (double * ) malloc ( *nbr_vertices * sizeof ( double ) );
    *vertex_lat = (double * ) malloc ( *nbr_vertices * sizeof ( double ) );

    status = nc_get_vara_double (ncid, glon_id, &start, nbr_vertices, (*vertex_lon));
    if (status != NC_NOERR) handle_error(status);

    status = nc_get_vara_double (ncid, glat_id, &start, nbr_vertices, (*vertex_lat));
    if (status != NC_NOERR) handle_error(status);

    break;

  default:

    exit(EXIT_FAILURE);

  }

  *vertex_type = glon_type;
}

/* ---------------------------------------------------------------- */

static void get_icon_cell_center ( int ncid,
                                   int  *cell_type,
                                   void **cell_lon,
                                   void **cell_lat,
                                   size_t *nbr_cells ) {

  int status;

  int glat_id;                        // Various NetCDF IDs
  int glon_id;

  int glon_ndims;                     // number of dimensions in file
  int glat_ndims;

  int glon_dimids[NC_MAX_VAR_DIMS];   // dimension NetCDF IDs
  int glat_dimids[NC_MAX_VAR_DIMS];

  nc_type glon_type;                  // variable type
  nc_type glat_type;

  int nbr_atts;                       // number of attributes

  status = nc_inq_varid (ncid, "clon", &glon_id);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_varid (ncid, "clat", &glat_id);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_var (ncid, glon_id, 0, &glon_type, &glon_ndims, glon_dimids, &nbr_atts);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_var (ncid, glat_id, 0, &glat_type, &glat_ndims, glat_dimids, &nbr_atts);
  if (status != NC_NOERR) handle_error(status);

  /* Allocate memory to read in coordinates */

  if ( glon_type != glat_type ) {
    exit(EXIT_FAILURE);
  }

  if (glon_ndims != 1 || glat_ndims != 1) {
    exit(EXIT_FAILURE);
  }

  /* get size of arrays */

  size_t dimlen_lat; // dimension size for latitude
  size_t dimlen_lon; // dimension size for longitude

  status = nc_inq_dimlen(ncid, *glon_dimids, &dimlen_lon);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_dimlen(ncid, *glat_dimids, &dimlen_lat);
  if (status != NC_NOERR) handle_error(status);

  if (dimlen_lon != dimlen_lat)
    exit(EXIT_FAILURE);

  *nbr_cells = dimlen_lon;

  /* read center coordinates and convert radians into degrees */

  size_t start = 0;

  switch(glon_type) {

  case NC_FLOAT:

    *cell_lon = (float * ) malloc ( *nbr_cells * sizeof ( float ) );
    *cell_lat = (float * ) malloc ( *nbr_cells * sizeof ( float ) );

    status = nc_get_vara (ncid, glon_id, &start, nbr_cells, (*cell_lon));
    if (status != NC_NOERR) handle_error(status);

    status = nc_get_vara (ncid, glat_id, &start, nbr_cells, (*cell_lat));
    if (status != NC_NOERR) handle_error(status);

    break;

  case NC_DOUBLE:

    *cell_lon = (double * ) malloc ( *nbr_cells * sizeof ( double ) );
    *cell_lat = (double * ) malloc ( *nbr_cells * sizeof ( double ) );

    status = nc_get_vara_double (ncid, glon_id, &start, nbr_cells, (*cell_lon));
    if (status != NC_NOERR) handle_error(status);

    status = nc_get_vara_double (ncid, glat_id, &start, nbr_cells, (*cell_lat));
    if (status != NC_NOERR) handle_error(status);

    break;

  default:

    exit(EXIT_FAILURE);

  }

  *cell_type = glon_type;
}

/* ---------------------------------------------------------------- */

static void get_icon_cell_mask ( int ncid,
                                 int  *cell_type,
                                 int **cell_mask,
                                 size_t *nbr_cells ) {

  int status;

  int mask_id;                        // Various NetCDF IDs
  int mask_ndims;                     // number of dimensions in file
  int mask_dimids[NC_MAX_VAR_DIMS];   // dimension NetCDF IDs

  nc_type mask_type;                  // variable type

  int nbr_atts;                       // number of attributes

  status = nc_inq_varid (ncid, "cell_sea_land_mask", &mask_id);

if (status != NC_NOERR) handle_error(status);

  status = nc_inq_var (ncid, mask_id, 0, &mask_type, &mask_ndims, mask_dimids, &nbr_atts);
  if (status != NC_NOERR) handle_error(status);

  /* Allocate memory to read in coordinates */

  if (mask_ndims != 1) exit(EXIT_FAILURE);

  /* get size of arrays */

  size_t dimlen_mask; // dimension size for latitude

  status = nc_inq_dimlen(ncid, *mask_dimids, &dimlen_mask);
  if (status != NC_NOERR) handle_error(status);

  *nbr_cells = dimlen_mask;

  /* read cell mask */

  size_t start = 0;

  switch(mask_type) {

  case NC_INT:

    *cell_mask = (int * ) malloc ( *nbr_cells * sizeof ( int ) );

    status = nc_get_vara_int (ncid, mask_id, &start, nbr_cells, (*cell_mask));
    if (status != NC_NOERR) handle_error(status);

    break;

  default:

    exit(EXIT_FAILURE);

  }

  *cell_type = mask_type;
}

/* ---------------------------------------------------------------- */

static void get_icon_connect ( int ncid, int **vertex_of_cell, size_t **dimlen ) {

  int status;
  
  int conn_id;
  int conn_ndims;
  int conn_dimids[NC_MAX_VAR_DIMS];
  nc_type conn_type;

  int nbr_atts;                     // number of attributes

  status = nc_inq_varid (ncid, "vertex_of_cell", &conn_id);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_var (ncid, conn_id, 0, &conn_type, &conn_ndims, conn_dimids, &nbr_atts);
  if (status != NC_NOERR) handle_error(status);

  /* get size of arrays */

  size_t *start;
  size_t *count;

  start = ( size_t * ) malloc ( conn_ndims * sizeof (size_t) );
  count = ( size_t * ) malloc ( conn_ndims * sizeof (size_t) );

  *dimlen = ( size_t * ) malloc ( conn_ndims * sizeof (size_t) );

  int length = 1;

  int i;

  for ( i = 0; i < conn_ndims; i++ ) {

    start[i] = 0;

    status = nc_inq_dimlen(ncid, conn_dimids[i], &count[i]);
    if (status != NC_NOERR) handle_error(status);

    length = length * count[i];
    (*dimlen)[i] = count[i];

  }

  /* read connectivity */

  *vertex_of_cell = ( int *) malloc ( length * sizeof ( int ) );

  status = nc_get_vara_int(ncid, conn_id, start, count, (*vertex_of_cell));
  if (status != NC_NOERR) handle_error(status);

  free (start);
  free (count);

}

void delete_icon_grid_data( int ** cell_mask,
		                    int ** global_cell_id,
                            int ** global_cell_id_rank,
                            int ** num_vertices_per_cell,
                            int ** global_corner_id,
                            int ** global_corner_id_rank,
                            int ** cell_to_vertex,
                            double ** x_cells,
                            double ** y_cells,
                            double ** x_vertices,
                            double ** y_vertices) {

	free (*cell_mask);
    free (*global_cell_id);
    free (*global_cell_id_rank);
    free (*num_vertices_per_cell);
    free (*global_corner_id);
    free (*global_corner_id_rank);
    free (*cell_to_vertex);
    free (*x_cells);
    free (*y_cells);
    free (*x_vertices);
    free (*y_vertices);
}

/* ---------------------------------------------------------------- */

static void handle_error(int status)
{
  if (status == NC_NOERR) return;
  const char *err_string;
  err_string = nc_strerror(status);
  printf ( "%s \n", err_string );
  exit(EXIT_FAILURE);
}

