//#define VERBOSE
#include <stdio.h>
#include <netcdf.h>

#include "read_woa_data.h"

/* ------------------------------------------------ */

static void handle_error(int status) {
  const char *err_string;
  err_string = nc_strerror(status);
  printf ("%s \n", err_string);
  if (status != NC_NOERR) {
    exit(EXIT_FAILURE);
  }
}

/* ------------------------------------------------ */

int open_woa_output ( char * input_file ) {

  int ncid;

  int status = nc_open(input_file, NC_NOWRITE, &ncid);
  if (status != NC_NOERR) handle_error(status);

  return ncid;
}

/* ------------------------------------------------ */

void close_woa_output ( int ncid ) {
    int status = nc_close(ncid);
    if (status != NC_NOERR) handle_error(status);
}

/* ------------------------------------------------ */

void read_woa_dimensions ( int fileId, char * fieldName,
                            struct fieldMetadata * fieldInfo ) {

  int varNdims;
  int varDimids[NC_MAX_VAR_DIMS];   // dimension NetCDF IDs

  int varId;
  nc_type varType;
  int timeId;
  int depthId;
  int latId, lonId;

  int status;

  int i;

  status = nc_inq_varid ( fileId, fieldName, &varId);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_vartype ( fileId, varId, &varType );
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_varndims ( fileId, varId, &varNdims);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_vardimid ( fileId, varId, varDimids);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_dimid ( fileId, "time", &timeId );
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_dimid ( fileId, "depth", &depthId );
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_dimid ( fileId, "lon", &lonId );
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_dimid ( fileId, "lat", &latId );
  if (status != NC_NOERR) handle_error(status);

  fieldInfo->start = ( size_t * ) malloc ( varNdims * sizeof (size_t) );
  fieldInfo->count = ( size_t * ) malloc ( varNdims * sizeof (size_t) );

  switch(varType) {
  case NC_FLOAT:
    fieldInfo->varType = FLOAT;
    break;
  case NC_DOUBLE:
    fieldInfo->varType = DOUBLE;
    break;
  default:
    fprintf(stdout, "Datatype not supported %i %i %i \n", varType, NC_FLOAT, NC_DOUBLE);
    exit(EXIT_FAILURE);
  }

  fieldInfo->varId   = varId;

  /* set defaults */

  fieldInfo->nbrLevels    = 1;
  fieldInfo->nbrTimeSteps = 1;
  fieldInfo->nbrLatPoints = 1;
  fieldInfo->nbrLonPoints = 1;
  fieldInfo->timeDim      = -1;
  fieldInfo->latDim       = -1;
  fieldInfo->lonDim       = -1;
  fieldInfo->levelDim     = -1;

  for ( i = 0; i < varNdims; i++ ) {

    fieldInfo->start[i] = 0;

    status = nc_inq_dimlen( fileId, varDimids[i], &fieldInfo->count[i]);
    if (status != NC_NOERR) handle_error(status);

  }

  for ( i = 0; i < varNdims; i++ ) {
    if ( varDimids[i] == timeId  ) {
      fieldInfo->nbrTimeSteps = fieldInfo->count[i];
      fieldInfo->timeDim = i;
#ifdef VERBOSE
      printf ("number of time steps in file  %i \n",  fieldInfo->nbrTimeSteps);
#endif
    }
    if ( varDimids[i] == depthId ) {
      fieldInfo->nbrLevels    = fieldInfo->count[i];
      fieldInfo->levelDim = i;
#ifdef VERBOSE
      printf ("number of levels for field    %i \n",  fieldInfo->nbrLevels);
#endif
    }
    if ( varDimids[i] == latId  ) {
      fieldInfo->nbrLatPoints = fieldInfo->count[i];
      fieldInfo->latDim = i;
#ifdef VERBOSE
      printf ("number of horizontal lon data  %i \n",  fieldInfo->nbrLatPoints);
#endif
    }
    if ( varDimids[i] == lonId  ) {
      fieldInfo->nbrLonPoints = fieldInfo->count[i];
      fieldInfo->lonDim = i;
#ifdef VERBOSE
      printf ("number of horizontal data lat  %i \n",  fieldInfo->nbrLonPoints);
#endif
    }
  }
}

/* ------------------------------------------------ */

void get_woa_memory ( void ** cell_data, struct fieldMetadata fieldInfo ) {

  switch(fieldInfo.varType) {
  case FLOAT:
    *cell_data = malloc ( fieldInfo.nbrLatPoints * fieldInfo.nbrLonPoints * sizeof( float ) );
#ifdef VERBOSE
    printf ("Allocated %i elements with datatype NC_FLOAT.\n", fieldInfo.nbrLatPoints * fieldInfo.nbrLonPoints);
#endif
    break;
  case DOUBLE:
    *cell_data = malloc ( fieldInfo.nbrLatPoints * fieldInfo.nbrLonPoints * sizeof ( double ) );
#ifdef VERBOSE
    printf ("Allocated %i elements with datatype NC_DOUBLE.\n", fieldInfo.nbrLatPoints * fieldInfo.nbrLonPoints);
#endif
    break;
  default:
    fprintf(stdout, "get_memory: datatype not supported %i %i %i \n", fieldInfo.varType, FLOAT, DOUBLE);
    exit(EXIT_FAILURE);
  }
}

/* ------------------------------------------------ */

void free_woa_memory ( void * data ) {
  free(data);
}

/* ------------------------------------------------ */

void read_woa_timestep_level ( int fileId, void * cell_data, struct fieldMetadata fieldInfo, int timeStep, int level ) {

  /* ... extract one level */

  if ( fieldInfo.levelDim > -1 ) {
    fieldInfo.start[fieldInfo.levelDim] = level-1;
    fieldInfo.count[fieldInfo.levelDim] = 1;
#ifdef VERBOSE
    printf ("depth start and count %i %i \n", (int) fieldInfo.start[fieldInfo.levelDim], (int) fieldInfo.count[fieldInfo.levelDim]);
#endif
    if ( level > fieldInfo.nbrLevels ) {
      printf ("A maximum of %i levels/layers is available for the selected field! \n",fieldInfo.nbrLevels);
      exit(EXIT_FAILURE);
    }
  }

  /* ... extract one timestep */

  if ( fieldInfo.timeDim > -1 ) {
    fieldInfo.start[fieldInfo.timeDim] = timeStep-1;
    fieldInfo.count[fieldInfo.timeDim] = 1;
#ifdef VERBOSE
    printf ("time start and count %i %i \n", (int) fieldInfo.start[fieldInfo.timeDim], (int) fieldInfo.count[fieldInfo.timeDim]);
#endif
  }

  /* ... extract one horizontal data set */

  if ( fieldInfo.latDim > -1 ) {
    fieldInfo.start[fieldInfo.latDim] = 0 ;
    fieldInfo.count[fieldInfo.latDim] = fieldInfo.nbrLatPoints ;
#ifdef VERBOSE
    printf ("lat  start and count %i %i \n", (int) fieldInfo.start[fieldInfo.latDim], (int) fieldInfo.count[fieldInfo.latDim]);
#endif
  }
  if ( fieldInfo.lonDim > -1 ) {
    fieldInfo.start[fieldInfo.lonDim] = 0 ;
    fieldInfo.count[fieldInfo.lonDim] = fieldInfo.nbrLonPoints ;
  #ifdef VERBOSE
    printf ("lon  start and count %i %i \n", (int) fieldInfo.start[fieldInfo.lonDim], (int) fieldInfo.count[fieldInfo.lonDim]);
  #endif
  }
#ifdef VERBOSE
  printf (" -------------------- \n ");
#endif
  /* ... get data from netcdf file */

  int status = nc_get_vara ( fileId, fieldInfo.varId, fieldInfo.start, fieldInfo.count, cell_data);
  if (status != NC_NOERR) handle_error(status);

}

