// #define VERBOSE

/**
 * @file perf_toy_icon.c
 *
 * @copyright Copyright  (C)  2015 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include "yac_config.h"

#if defined USE_MPI && defined HAVE_C_NETCDF
#include <mpi.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "xml/config_xml.h"
#include "utils.h"
#include "component.h"
#include "fields.h"
#include "yac_interface.h"
#include "geometry.h"
#include "read_icon_grid.h"
#include "test_function.h"

/* ------------------------------------------------- */

/* For simplicity we define the same 8 fields that are in the xml file */

const char * fieldName[] = {"icon_out", "cube_out"};

int main () {

#ifdef USE_XML2

  // Initialisation of MPI

  MPI_Init (0, NULL);

  /* The initialisation phase includes the reading of the XML file */
#ifdef VERBOSE
  printf (". main: calling yac_cinit\n");
#endif

  double tic, toc, time;

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  yac_cinit ("perf_toy.xml", "coupling.xsd");

  /* The usual component definition, here for two sequential components on the same process */

#ifdef VERBOSE
  printf (". main: calling yac_cdef_comp\n");
#endif

  int comp_id;
  char * comp_name = "ICON";

  yac_cdef_comp ( comp_name, &comp_id );
#ifdef VERBOSE
  printf ( ". main: defined %s with local comp ID %i \n", "ICON", comp_id );
#endif

  MPI_Comm local_comm;

  yac_cget_localcomm(&local_comm, comp_id);

  int rank, size;

  MPI_Comm_rank(local_comm,&rank);
  MPI_Comm_size(local_comm,&size);

  int ierror = 0;

  int num_fields = 2;
  int global_comp_id;
  int cell_point_id;
  int mask_id;

  int field_ids[2];
  int domain_id;
  int subdomain_id;

  struct component * component;

  /* Grid definition for the first component (ICON) */

  char * filename = "icon_grid_R02B06.nc";
  //char * filename = "icon_grid_R02B07.nc";
  int num_vertices;
  int num_cells;
  int * num_vertices_per_cell;
  int * cell_to_vertex;
  double * x_vertices;
  double * y_vertices;
  double * x_cells;
  double * y_cells;

  int * cell_mask;
  int * global_cell_id;
  int * global_cell_id_rank;
  int * global_corner_id;
  int * global_corner_id_rank;

  read_part_icon_grid_information(filename, &num_vertices, &num_cells,
                                  &num_vertices_per_cell, &cell_to_vertex,
                                  &x_vertices, &y_vertices, &x_cells,
                                  &y_cells, &global_cell_id,
                                  &cell_mask,
                                  &global_cell_id_rank, &global_corner_id,
                                  &global_corner_id_rank, rank, size);

  double * x_points, * y_points;

  x_points = x_vertices;
  y_points = y_vertices;

  for (int i = 0; i < num_vertices; ++i) {
    x_points[i] /= YAC_RAD;
    y_points[i] /= YAC_RAD;
  }

  double * x_center, * y_center;

  x_center = x_cells;
  y_center = y_cells;

  for (int i = 0; i < num_cells; ++i) {
    x_center[i] /= YAC_RAD;
    y_center[i] /= YAC_RAD;
  }

  yac_cdef_subdomain ( comp_id, "icon_grid", &subdomain_id );

  yac_cdef_elements_unstruct(subdomain_id, num_vertices, num_cells,
                             num_vertices_per_cell, x_vertices, y_vertices,
                             cell_to_vertex);
  yac_cdef_points_unstruct(subdomain_id, num_cells, CELL, x_center, y_center,
                           &cell_point_id );
  yac_cdef_index_location(subdomain_id, num_cells, CELL, global_cell_id,
                          global_cell_id_rank);
  yac_cconnect_subdomains(comp_id, 1, &subdomain_id, &domain_id);
  yac_cdef_mask(cell_mask, cell_point_id, &mask_id);

  /* Field definition for the first component (ICON-atmosphere) */

  component = yac_unique_id_to_pointer(comp_id);
  global_comp_id = yac_component_get_global_id(component);
 
  num_fields = yac_xml_get_number_of_transient_grid_refs_by_component_id(global_comp_id);
  
#ifdef VERBOSE
  printf (". main: number of fields in component %i is %i\n", global_comp_id, num_fields);
#endif

  yac_cdef_field(fieldName[0], comp_id, domain_id, &cell_point_id, &mask_id, 1,
                 &field_ids[0]);
  yac_cdef_field(fieldName[1], comp_id, domain_id, &cell_point_id, &mask_id, 1,
                 &field_ids[1]);

  toc=MPI_Wtime();
  time = toc-tic;
  printf ("ICON: Time for initialisation %f\n", time ); 

  /* Search. */

#ifdef VERBOSE
  printf (". main: calling yac_csearch\n");
#endif

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  yac_csearch ( 1, &comp_id, num_fields, field_ids, &ierror );

  toc=MPI_Wtime();
  time = toc-tic;
  printf ("ICON: Time for search %f\n", time ); 

  double * conserv_in = malloc(num_cells * sizeof(*conserv_in));

  for (int i = 0; i < num_cells; ++i)
    conserv_in[i] = -10;

  double * cell_out = malloc(num_cells * sizeof(*cell_out));

  int err, info;

  int cell_to_vertex_offset = 0;

  for (int i = 0; i < num_cells; ++i) {

    double middle_point[3] = {0, 0, 0};

    for (int j = 0; j < num_vertices_per_cell[i]; ++j) {

      double curr_point[3];

      LLtoXYZ_deg(x_vertices[cell_to_vertex[cell_to_vertex_offset + j]],
                  y_vertices[cell_to_vertex[cell_to_vertex_offset + j]],
                  curr_point);

      middle_point[0] += curr_point[0];
      middle_point[1] += curr_point[1];
      middle_point[2] += curr_point[2];
    }

    double scale = 1.0 / sqrt(middle_point[0] * middle_point[0] + 
                              middle_point[1] * middle_point[1] + 
                              middle_point[2] * middle_point[2]);

    middle_point[0] *= scale;
    middle_point[1] *= scale;
    middle_point[2] *= scale;

    struct point point;

    XYZtoLL(middle_point, &point.lon, &point.lat);

    cell_to_vertex_offset += num_vertices_per_cell[i];

    cell_out[i] = test_func_point(point);
  }

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  {
    double *point_set_data[1];
    double **domain_data[1] = {point_set_data};
    double ***collection_data[1] = {domain_data};

    point_set_data[0] = cell_out;
    yac_cput(field_ids[0], 1, collection_data, &info, &err);
  }

  {
    double *domain_data[1];
    double **collection_data[1] = {domain_data};

    domain_data[0] = conserv_in;
    yac_cget(field_ids[1], 1, collection_data, &info, &err);
  }

  toc=MPI_Wtime();
  time = toc-tic;
  printf ("ICON: Time for ping-pong %f\n", time );

#ifdef VTK_OUTPUT
  //----------------------------------------------------------
  // write field to vtk output file
  //----------------------------------------------------------

  char vtk_filename[32];

  sprintf(vtk_filename, "perf_toy_icon_%d.vtk", rank);

  double point_data[num_vertices][3];
  for (int i = 0; i < num_vertices; ++i) {
   LLtoXYZ_deg(x_vertices[i], y_vertices[i], point_data[i]);
  }

  VTK_FILE *vtk_file = vtk_open(vtk_filename, "unstruct_out");
  vtk_write_point_data(vtk_file, (double *)point_data, num_vertices);
  vtk_write_cell_data(vtk_file, (unsigned *)cell_to_vertex,
                      (unsigned*)num_vertices_per_cell, num_cells);
  vtk_write_point_scalars_int(vtk_file, global_corner_id_rank, num_vertices,
                              "vertex_is_on_rank");
  vtk_write_point_scalars_int(vtk_file, global_corner_id, num_vertices,
                              "global_corner_id");
  vtk_write_cell_scalars_int(vtk_file, global_cell_id_rank, num_cells,
                             "cell_is_on_rank");
  vtk_write_cell_scalars_int(vtk_file, global_cell_id, num_cells,
                             "global_cell_id");

  vtk_write_cell_scalars_double(vtk_file, conserv_in, num_cells, "conserv_in");

  vtk_write_cell_scalars_double(vtk_file, cell_out, num_cells, "cell_out");

  vtk_close(vtk_file);

#endif // VTK_OUTPUT

  yac_cfinalize();

  MPI_Finalize();

  free(num_vertices_per_cell);
  free(cell_to_vertex);
  free(x_cells);
  free(y_cells);
  free(x_vertices);
  free(y_vertices);
  free(cell_mask);
  free(global_cell_id);
  free(global_cell_id_rank);
  free(global_corner_id);
  free(global_corner_id_rank);
  free(conserv_in);
  free(cell_out);

  return EXIT_SUCCESS;

#else

  fprintf(stderr, "ERROR: perf_toy_icon requires XML2\n");

  return EXIT_FAILURE;

#endif // USE_XML2
}

#else
#include <stdlib.h>
#include <stdio.h>
int main () {
  printf ("Examples requires compiling with MPI and NetCDF.\n");
  return EXIT_FAILURE;
}
#endif

