// #define VERBOSE

/**
 * @file toy_multi_unstruct.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include "yac_config.h"

#if defined USE_MPI && defined HAVE_C_NETCDF
#include <mpi.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "xml/config_xml.h"
#include "utils.h"
#include "component.h"
#include "fields.h"
#include "yac_interface.h"
#include "geometry.h"
#include "vtk_output.h"
#include "read_icon_grid.h"
#include "test_function.h"

/* ------------------------------------------------- */

/* For simplicity we define the same 8 fields that are in the xml file */

const char * fieldName[] = {"reg2d_conserv_field_out",
                            "unstruct_conserv_field_out",
                            "unstruct_2_conserv_field_out",
                            "cube_conserv_field_out",
                            "reg2d_avg_field_out",
                            "unstruct_avg_field_out",
                            "unstruct_2_avg_field_out",
                            "cube_avg_field_out",
                            "reg2d_patch_field_out",
                            "unstruct_patch_field_out",
                            "unstruct_2_patch_field_out",
                            "cube_patch_field_out",
                            "reg2d_hcsbb_field_out",
                            "unstruct_hcsbb_field_out",
                            "unstruct_2_hcsbb_field_out",
                            "cube_hcsbb_field_out",
                            "reg2d_2nd_conserv_field_out",
                            "unstruct_2nd_conserv_field_out",
                            "unstruct_2_2nd_conserv_field_out",
                            "cube_2nd_conserv_field_out",
                            "reg2d_rbf_field_out",
                            "unstruct_rbf_field_out",
                            "unstruct_2_rbf_field_out",
                            "cube_rbf_field_out"};

int main (int argc, char *argv[]) {

#ifdef USE_XML2

  // Initialisation of MPI

  MPI_Init (0, NULL);

  /* The initialisation phase includes the reading of the XML file */
#ifdef VERBOSE
  printf (". main: calling yac_cinit\n");
#endif

  double tic, toc, time;

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  char * configFilename = "coupling_toy_multi.xml";
  if ( argc > 2 ) configFilename = argv[2];
#ifdef VERBOSE
  printf ( "coupling configuration file: %s\n",configFilename);
#endif
  yac_cinit (configFilename, "coupling.xsd");

  /* The usual component definition, here for two sequential components on the same process */

#ifdef VERBOSE
  printf (". main: calling yac_cdef_comp\n");
#endif

  int comp_id;
  char * comp_name = "ICON";

  yac_cdef_comp ( comp_name, &comp_id );
#ifdef VERBOSE
  printf ( ". main: defined %s with local comp ID %i \n", "ICON", comp_id );
#endif

  MPI_Comm local_comm;

  yac_cget_localcomm(&local_comm, comp_id);

  int rank, size;

  MPI_Comm_rank(local_comm,&rank);
  MPI_Comm_size(local_comm,&size);

  int ierror = 0;

  int nbr_fields = 1;
  int global_comp_id;
  int cell_point_id;
  int corner_point_id;
  int mask_id[3];

  int field_ids[20];
  int domain_id;
  int subdomain_id;

  struct component * component;

  /* Grid definition for the first component*/

  char * filename = "icon_grid_R02B06.nc";
  if ( argc > 1 ) filename = argv[1];
  printf ( "toy_multi_unstruct: input grid file: %s\n",filename );

  int nbr_vertices;
  int nbr_cells;
  int * num_vertices_per_cell;
  int * cell_to_vertex;
  double * x_vertices;
  double * y_vertices;
  double * x_cells;
  double * y_cells;

  int * cell_mask;
  int * global_cell_id;
  int * global_cell_id_rank;
  int * global_corner_id;
  int * global_corner_id_rank;

  read_part_icon_grid_information(filename, &nbr_vertices, &nbr_cells,
                                  &num_vertices_per_cell, &cell_to_vertex,
                                  &x_vertices, &y_vertices, &x_cells,
                                  &y_cells, &global_cell_id,
                                  &cell_mask,
                                  &global_cell_id_rank, &global_corner_id,
                                  &global_corner_id_rank, rank, size);

  double * x_points, * y_points;

  x_points = x_vertices;
  y_points = y_vertices;

  for (int i = 0; i < nbr_vertices; ++i) {
    x_points[i] /= YAC_RAD;
    y_points[i] /= YAC_RAD;
  }

  double * x_center, * y_center;

  x_center = x_cells;
  y_center = y_cells;

  for (int i = 0; i < nbr_cells; ++i) {
    x_center[i] /= YAC_RAD;
    y_center[i] /= YAC_RAD;
  }

  yac_cdef_subdomain ( comp_id, "icon_grid", &subdomain_id );

  yac_cdef_elements_unstruct ( subdomain_id, nbr_vertices, nbr_cells, num_vertices_per_cell,
                               x_vertices, y_vertices, cell_to_vertex );

  yac_cdef_points_unstruct ( subdomain_id, nbr_vertices, CORNER, x_points, y_points, &corner_point_id );
  yac_cdef_points_unstruct ( subdomain_id, nbr_cells, CELL, x_center, y_center, &cell_point_id );

  yac_cdef_index_location ( subdomain_id, nbr_cells, CELL, global_cell_id, global_cell_id_rank );
  yac_cdef_index_location ( subdomain_id, nbr_vertices, CORNER, global_corner_id, global_corner_id_rank );

  yac_cconnect_subdomains ( comp_id, 1, &subdomain_id, &domain_id );

  yac_cdef_mask ( cell_mask, cell_point_id, &mask_id[0] );

  /* Field definition for the first component*/

  component = yac_unique_id_to_pointer(comp_id);
  global_comp_id = yac_component_get_global_id(component);
 
  nbr_fields = yac_xml_get_number_of_transient_grid_refs_by_component_id(global_comp_id);
  
#ifdef VERBOSE
  printf (". main: number of fields in component %i is %i\n", global_comp_id, nbr_fields);
#endif

  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(fieldName[i+0], comp_id, domain_id, &cell_point_id, &mask_id[0], 1, &field_ids[i+0]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(fieldName[i+4], comp_id, domain_id, &corner_point_id, NULL, 1, &field_ids[i+4]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(fieldName[i+8], comp_id, domain_id, &cell_point_id, &mask_id[0], 1, &field_ids[i+8]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(fieldName[i+12], comp_id, domain_id, &corner_point_id, NULL, 1, &field_ids[i+12]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(fieldName[i+16], comp_id, domain_id, &cell_point_id, NULL, 1, &field_ids[i+16]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(fieldName[i+20], comp_id, domain_id, &cell_point_id, NULL, 1, &field_ids[i+20]);

  toc=MPI_Wtime();
  time = toc-tic;
  printf ("toy_multi_unstruct: Time for initialisation %f\n", time );

  /* Search. */

#ifdef VERBOSE
  printf (". main: calling yac_csearch\n");
#endif

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  yac_csearch ( 1, &comp_id, nbr_fields, field_ids, &ierror );

  toc=MPI_Wtime();
  time = toc-tic;
  printf ("toy_multi_unstruct: Time for search %f\n", time );

  double * conserv_in_reg2d_point_data = malloc(nbr_cells * sizeof(*conserv_in_reg2d_point_data));
  double * conserv_in_unstruct_2_point_data = malloc(nbr_cells * sizeof(*conserv_in_unstruct_2_point_data));
  double * conserv_in_cube_point_data = malloc(nbr_cells * sizeof(*conserv_in_cube_point_data));

  double * avg_in_reg2d_point_data = malloc(nbr_vertices * sizeof(*avg_in_reg2d_point_data));
  double * avg_in_unstruct_2_point_data = malloc(nbr_vertices * sizeof(*avg_in_unstruct_2_point_data));
  double * avg_in_cube_point_data = malloc(nbr_vertices * sizeof(*avg_in_cube_point_data));

  double * patch_in_reg2d_point_data = malloc(nbr_cells * sizeof(*patch_in_reg2d_point_data));
  double * patch_in_unstruct_2_point_data = malloc(nbr_cells * sizeof(*patch_in_unstruct_2_point_data));
  double * patch_in_cube_point_data = malloc(nbr_cells * sizeof(*patch_in_cube_point_data));

  double * hcsbb_in_reg2d_point_data = malloc(nbr_vertices * sizeof(*hcsbb_in_reg2d_point_data));
  double * hcsbb_in_unstruct_2_point_data = malloc(nbr_vertices * sizeof(*hcsbb_in_unstruct_2_point_data));
  double * hcsbb_in_cube_point_data = malloc(nbr_vertices * sizeof(*hcsbb_in_cube_point_data));

  double * conserv_2nd_in_reg2d_point_data = malloc(nbr_cells * sizeof(*conserv_2nd_in_reg2d_point_data));
  double * conserv_2nd_in_unstruct_2_point_data = malloc(nbr_cells * sizeof(*conserv_2nd_in_unstruct_2_point_data));
  double * conserv_2nd_in_cube_point_data = malloc(nbr_cells * sizeof(*conserv_2nd_in_cube_point_data));

  double * rbf_in_reg2d_point_data = malloc(nbr_cells * sizeof(*rbf_in_reg2d_point_data));
  double * rbf_in_unstruct_2_point_data = malloc(nbr_cells * sizeof(*rbf_in_unstruct_2_point_data));
  double * rbf_in_cube_point_data = malloc(nbr_cells * sizeof(*rbf_in_cube_point_data));

  for (int i = 0; i < nbr_cells; ++i) {
    conserv_in_reg2d_point_data[i] = -10;
    conserv_in_unstruct_2_point_data[i] = -10;
    conserv_in_cube_point_data[i] = -10;

    patch_in_reg2d_point_data[i] = -10;
    patch_in_unstruct_2_point_data[i] = -10;
    patch_in_cube_point_data[i] = -10;

    conserv_2nd_in_reg2d_point_data[i] = -10;
    conserv_2nd_in_unstruct_2_point_data[i] = -10;
    conserv_2nd_in_cube_point_data[i] = -10;

    rbf_in_reg2d_point_data[i] = -10;
    rbf_in_unstruct_2_point_data[i] = -10;
    rbf_in_cube_point_data[i] = -10;
  }

  for (int i = 0; i < nbr_vertices; ++i) {
    avg_in_reg2d_point_data[i] = -10;
    avg_in_unstruct_2_point_data[i] = -10;
    avg_in_cube_point_data[i] = -10;

    hcsbb_in_reg2d_point_data[i] = -10;
    hcsbb_in_unstruct_2_point_data[i] = -10;
    hcsbb_in_cube_point_data[i] = -10;
  }

  double * cell_out_point_data = malloc(nbr_cells * sizeof(*cell_out_point_data));
  double * corner_out_point_data = malloc(nbr_vertices * sizeof(*corner_out_point_data));

  int err;
  int info;

  int cell_to_vertex_offset = 0;

  for (int i = 0; i < nbr_cells; ++i) {

    double middle_point[3] = {0, 0, 0};

    for (int j = 0; j < num_vertices_per_cell[i]; ++j) {

      double curr_point[3];

      LLtoXYZ_deg(x_vertices[cell_to_vertex[cell_to_vertex_offset + j]],
                  y_vertices[cell_to_vertex[cell_to_vertex_offset + j]],
                  curr_point);

      middle_point[0] += curr_point[0];
      middle_point[1] += curr_point[1];
      middle_point[2] += curr_point[2];
    }

    double scale = 1.0 / sqrt(middle_point[0] * middle_point[0] +
                              middle_point[1] * middle_point[1] +
                              middle_point[2] * middle_point[2]);

    middle_point[0] *= scale;
    middle_point[1] *= scale;
    middle_point[2] *= scale;

    struct point point;

    XYZtoLL(middle_point, &point.lon, &point.lat);

    cell_to_vertex_offset += num_vertices_per_cell[i];

    cell_out_point_data[i] = test_func_point(point);
  }

  for (int i = 0; i < nbr_vertices; ++i)
    corner_out_point_data[i] = test_func_deg(x_points[i], y_points[i]);

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  {
    double *point_set_data[1];
    double **domain_data[1] = {point_set_data};
    double ***collection_data[1] = {domain_data};

    point_set_data[0] = cell_out_point_data;
    yac_cput(field_ids[1], 1, collection_data, &info, &err);
    point_set_data[0] = corner_out_point_data;
    yac_cput(field_ids[5], 1, collection_data, &info, &err);
    point_set_data[0] = cell_out_point_data;
    yac_cput(field_ids[9], 1, collection_data, &info, &err);
    point_set_data[0] = corner_out_point_data;
    yac_cput(field_ids[13], 1, collection_data, &info, &err);
    point_set_data[0] = cell_out_point_data;
    yac_cput(field_ids[17], 1, collection_data, &info, &err);
    point_set_data[0] = cell_out_point_data;
    yac_cput(field_ids[21], 1, collection_data, &info, &err);
  }

  {
    double *domain_data[1];
    double **collection_data[1] = {domain_data};

    domain_data[0] = conserv_in_reg2d_point_data;
    yac_cget(field_ids[0], 1, collection_data, &info, &err);
    domain_data[0] = conserv_in_unstruct_2_point_data;
    yac_cget(field_ids[2], 1, collection_data, &info, &err);
    domain_data[0] = conserv_in_cube_point_data;
    yac_cget(field_ids[3], 1, collection_data, &info, &err);

    domain_data[0] = avg_in_reg2d_point_data;
    yac_cget(field_ids[4], 1, collection_data, &info, &err);
    domain_data[0] = avg_in_unstruct_2_point_data;
    yac_cget(field_ids[6], 1, collection_data, &info, &err);
    domain_data[0] = avg_in_cube_point_data;
    yac_cget(field_ids[7], 1, collection_data, &info, &err);

    domain_data[0] = patch_in_reg2d_point_data;
    yac_cget(field_ids[8], 1, collection_data, &info, &err);
    domain_data[0] = patch_in_unstruct_2_point_data;
    yac_cget(field_ids[10], 1, collection_data, &info, &err);
    domain_data[0] = patch_in_cube_point_data;
    yac_cget(field_ids[11], 1, collection_data, &info, &err);

    domain_data[0] = hcsbb_in_reg2d_point_data;
    yac_cget(field_ids[12], 1, collection_data, &info, &err);
    domain_data[0] = hcsbb_in_unstruct_2_point_data;
    yac_cget(field_ids[14], 1, collection_data, &info, &err);
    domain_data[0] = hcsbb_in_cube_point_data;
    yac_cget(field_ids[15], 1, collection_data, &info, &err);

    domain_data[0] = conserv_2nd_in_reg2d_point_data;
    yac_cget(field_ids[16], 1, collection_data, &info, &err);
    domain_data[0] = conserv_2nd_in_unstruct_2_point_data;
    yac_cget(field_ids[18], 1, collection_data, &info, &err);
    domain_data[0] = conserv_2nd_in_cube_point_data;
    yac_cget(field_ids[19], 1, collection_data, &info, &err);

    domain_data[0] = rbf_in_reg2d_point_data;
    yac_cget(field_ids[20], 1, collection_data, &info, &err);
    domain_data[0] = rbf_in_unstruct_2_point_data;
    yac_cget(field_ids[22], 1, collection_data, &info, &err);
    domain_data[0] = rbf_in_cube_point_data;
    yac_cget(field_ids[23], 1, collection_data, &info, &err);
  }

  toc=MPI_Wtime();
  time = toc-tic;
  printf ("toy_multi_unstruct: Time for ping-pong %f\n", time );

  //----------------------------------------------------------
  // write field to vtk output file
  //----------------------------------------------------------

  char vtk_filename[32];

  sprintf(vtk_filename, "unstruct_out_%d.vtk", rank);

  double point_data[nbr_vertices][3];
  for (int i = 0; i < nbr_vertices; ++i) {
   LLtoXYZ_deg(x_vertices[i], y_vertices[i], point_data[i]);
  }

  VTK_FILE *vtk_file = vtk_open(vtk_filename, "unstruct_out");
  vtk_write_point_data(vtk_file, (double *)point_data, nbr_vertices);
  vtk_write_cell_data(vtk_file, (unsigned *)cell_to_vertex,
                      (unsigned*)num_vertices_per_cell, nbr_cells);
  vtk_write_point_scalars_int(vtk_file, global_corner_id_rank, nbr_vertices,
                              "vertex_is_on_rank");
  vtk_write_point_scalars_int(vtk_file, global_corner_id, nbr_vertices,
                              "global_corner_id");
  vtk_write_cell_scalars_int(vtk_file, global_cell_id_rank, nbr_cells,
                             "cell_is_on_rank");
  vtk_write_cell_scalars_int(vtk_file, global_cell_id, nbr_cells,
                             "global_cell_id");

  vtk_write_cell_scalars_double(vtk_file, conserv_in_reg2d_point_data, nbr_cells,
                                "conserv_in_reg2d");
  vtk_write_cell_scalars_double(vtk_file, conserv_in_unstruct_2_point_data, nbr_cells,
                                "conserv_in_unstruct_2");
  vtk_write_cell_scalars_double(vtk_file, conserv_in_cube_point_data, nbr_cells,
                                "conserv_in_cube");

  vtk_write_point_scalars_double(vtk_file, avg_in_reg2d_point_data, nbr_vertices,
                                 "avg_in_reg2d");
  vtk_write_point_scalars_double(vtk_file, avg_in_unstruct_2_point_data, nbr_vertices,
                                 "avg_in_unstruct_2");
  vtk_write_point_scalars_double(vtk_file, avg_in_cube_point_data, nbr_vertices,
                                 "avg_in_cube");

  vtk_write_cell_scalars_double(vtk_file, patch_in_reg2d_point_data, nbr_cells,
                                "patch_in_reg2d");
  vtk_write_cell_scalars_double(vtk_file, patch_in_unstruct_2_point_data, nbr_cells,
                                "patch_in_unstruct_2");
  vtk_write_cell_scalars_double(vtk_file, patch_in_cube_point_data, nbr_cells,
                                "patch_in_cube");

  vtk_write_point_scalars_double(vtk_file, hcsbb_in_reg2d_point_data, nbr_vertices,
                                 "hcsbb_in_reg2d");
  vtk_write_point_scalars_double(vtk_file, hcsbb_in_unstruct_2_point_data, nbr_vertices,
                                 "hcsbb_in_unstruct_2");
  vtk_write_point_scalars_double(vtk_file, hcsbb_in_cube_point_data, nbr_vertices,
                                 "hcsbb_in_cube");

  vtk_write_cell_scalars_double(vtk_file, conserv_2nd_in_reg2d_point_data, nbr_cells,
                                "2nd_conserv_in_reg2d");
  vtk_write_cell_scalars_double(vtk_file, conserv_2nd_in_unstruct_2_point_data, nbr_cells,
                                "2nd_conserv_in_unstruct_2");
  vtk_write_cell_scalars_double(vtk_file, conserv_2nd_in_cube_point_data, nbr_cells,
                                "2nd_conserv_in_cube");

  vtk_write_cell_scalars_double(vtk_file, rbf_in_reg2d_point_data, nbr_cells,
                                "rbf_in_reg2d");
  vtk_write_cell_scalars_double(vtk_file, rbf_in_unstruct_2_point_data, nbr_cells,
                                "rbf_in_unstruct_2");
  vtk_write_cell_scalars_double(vtk_file, rbf_in_cube_point_data, nbr_cells,
                                "rbf_in_cube");

  vtk_write_point_scalars_double(vtk_file, corner_out_point_data, nbr_vertices, "corner_out");
  vtk_write_cell_scalars_double(vtk_file, cell_out_point_data, nbr_cells, "cell_out");
  vtk_close(vtk_file);

  delete_icon_grid_data ( &cell_mask,
                          &global_cell_id,
                          &global_cell_id_rank,
                          &num_vertices_per_cell,
                          &global_corner_id,
                          &global_corner_id_rank,
                          &cell_to_vertex,
                          &x_cells,
                          &y_cells,
                          &x_vertices,
                          &y_vertices );

  yac_cfinalize();

  MPI_Finalize();

  free(conserv_in_reg2d_point_data);
  free(conserv_in_unstruct_2_point_data);
  free(conserv_in_cube_point_data);
  free(avg_in_reg2d_point_data);
  free(avg_in_unstruct_2_point_data);
  free(avg_in_cube_point_data);
  free(patch_in_reg2d_point_data);
  free(patch_in_unstruct_2_point_data);
  free(patch_in_cube_point_data);
  free(hcsbb_in_reg2d_point_data);
  free(hcsbb_in_unstruct_2_point_data);
  free(hcsbb_in_cube_point_data);
  free(conserv_2nd_in_reg2d_point_data);
  free(conserv_2nd_in_unstruct_2_point_data);
  free(conserv_2nd_in_cube_point_data);
  free(rbf_in_reg2d_point_data);
  free(rbf_in_unstruct_2_point_data);
  free(rbf_in_cube_point_data);
  free(corner_out_point_data);
  free(cell_out_point_data);

  return EXIT_SUCCESS;

#else

  fprintf(stderr, "ERROR: toy_multi_unstruct requires XML2\n");

  return EXIT_FAILURE;

#endif // USE_XML2
}

#else
#include <stdlib.h>
#include <stdio.h>
int main () {
  printf ("Examples requires compiling with MPI and NetCDF.\n");
  return EXIT_FAILURE;
}
#endif

