package de.mpg.mpimet.yacgui.model;

import java.util.ArrayList;
import java.util.List;

import de.mpg.mpimet.yacgui.model.Model.Align;

/**
 * List wrapper above connections with auxiliary methods
 * 
 * @author Teresa Holfeld
 */
public class Connections {
	
	//TODO: refactoring and rethinking would help to this class
	
	private List<Connection> connections;

	public Connections() {
		connections = new ArrayList<Connection>();
	}

	public boolean isEmpty() {
		return connections.isEmpty();
	}

	public int size() {
		return connections.size();
	}

	/**
	 * Adds given Connection to the collection.
	 * 
	 * <p>
	 * Takes care of preventing duplicate entries and unwanted combinations.
	 * 
	 * @param connection
	 *            to be added
	 * @throws IllegalConnectionException
	 *             if unwanted combination or duplication occurs
	 */
	public void addConnection(Connection connection) throws IllegalConnectionException {

		// unwanted case: connection already exists
		if (contains(connection)) {
			throw new IllegalConnectionException(connection.getSourceAlign(),
					connection.getSourceIndex(), connection.getTargetIndex(),
					"Connection already exists");
		}

		// unwanted case: connection with same target align, index exists
		if (containsTarget(connection)) {
			throw new IllegalConnectionException(connection.getSourceAlign(),
					connection.getSourceIndex(), connection.getTargetIndex(),
					"Connection with same target exists");
		}

		// unwanted case: source align, index == existing target align, index
		if (containsSourceAsTarget(connection)) {
			throw new IllegalConnectionException(connection.getSourceAlign(),
					connection.getSourceIndex(), connection.getTargetIndex(),
					"Connection exists that has the same target as the given connection's source");
		}

		// unwanted case: target align, index == existing source align, index
		if (containsTargetAsSource(connection)) {
			throw new IllegalConnectionException(connection.getSourceAlign(),
					connection.getSourceIndex(), connection.getTargetIndex(),
					"Connection exists that has the same source as the given connection's target");
		}

		// add connection
		connections.add(connection);
	}

	/**
	 * Creates a Connection and adds it to the collection.
	 * 
	 * <p>
	 * Takes care of preventing duplicate entries and unwanted combinations.
	 * 
	 * @param sourceAlign
	 * @param sourceIndex
	 * @param targetIndex
	 * @param hasInterpolation
	 * @return connection
	 * @throws IllegalConnectionException Illegal Connection
	 */
	public Connection addConnection(Align sourceAlign, int sourceIndex, int targetIndex, boolean hasInterpolation)
			throws IllegalConnectionException {
		Connection connection = new Connection(sourceAlign, sourceIndex, targetIndex, hasInterpolation);
		addConnection(connection);
		return connection;
	}

	/**
	 * Removes Connections at given align and given index.
	 * 
	 * @param align
	 *            side where to remove (LEFT or RIGHT)
	 * @param index
	 *            index where to remove
	 */
	public void removeConnection(Align align, int index) {
		List<Connection> toBeRemoved = new ArrayList<Connection>();
		for (Connection connection : connections) {
			if (connection.getSourceAlign() == align && connection.getSourceIndex() == index
					|| connection.getTargetAlign() == align && connection.getTargetIndex() == index) {
				toBeRemoved.add(connection);
			}
		}
		connections.removeAll(toBeRemoved);
	}

	/**
	 * Returns a Connection that have the specified Align and 
	 * index values either at source or at target if present.
	 * 
	 * @param align
	 *            side (LEFT or RIGHT) of source or target
	 * @param index
	 *            index of source or target
	 * @return connection with given attributes, null otherwise
	 */
	public Connection getConnectionAt(Align align, int index) {
		for (Connection connection : connections) {
			if (connection.getSourceAlign() == align && connection.getSourceIndex() == index
					|| connection.getTargetAlign() == align && connection.getTargetIndex() == index) {
				return connection;
			}
		}
		return null;
	}

	/**
	 * Checks if Connection exists that is either source or target at given
	 * align and given index.
	 * 
	 * @param align
	 *            side (LEFT or RIGHT) of source or target
	 * @param index
	 *            index of source or target
	 * @return true if connection exists
	 */
	public boolean hasConnectionAt(Align align, int index) {
		for (Connection connection : connections) {
			if (connection.getSourceAlign() == align && connection.getSourceIndex() == index
					|| connection.getTargetAlign() == align && connection.getTargetIndex() == index) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Checks if collection contains a Connection with given source Align and
	 * index.
	 * 
	 * @param align
	 *            which side (LEFT or RIGHT)
	 * @param index
	 *            index of source
	 * @return true if connection with given source exists
	 */
	public boolean hasSourceAt(Align align, int index) {
		for (Connection connection : connections) {
			if (connection.getSourceAlign() == align && connection.getSourceIndex() == index) {
				return true;
			}
		}
		return false;
	}

	public List<Connection> getConnections() {
		return connections;
	}

	/**
	 * Checks if collection contains given connection.
	 * 
	 * @param connection
	 *            to be compared
	 * @return true if given connection is contained
	 */
	private boolean contains(Connection connection) {
		return connections.contains(connection);
	}

	/**
	 * Checks if collection contains a connection with target Align and index
	 * equal to target Align and index of given connection.
	 * 
	 * @param other
	 *            to be compared
	 * @return true if connection with target Align and index is already
	 *         contained
	 */
	private boolean containsTarget(Connection other) {
		for (Connection c : connections) {
			if (c.getTargetAlign() == other.getTargetAlign()
					&& c.getTargetIndex() == other.getTargetIndex()) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Checks if collections contains a connection with target Align and index 
	 * equal to source Align and index of given connection.
	 * 
	 * @param other	to be compared
	 * @return true if source is contained as target
	 */
	private boolean containsSourceAsTarget(Connection other) {
		for (Connection c : connections) {
			if (c.getTargetAlign() == other.getSourceAlign()
					&& c.getTargetIndex() == other.getSourceIndex()) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Checks if collections contains a connection with source Align and index 
	 * equal to target Align and index of given connection.
	 * 
	 * @param other	to be compared
	 * @return true if target is contained as source
	 */
	private boolean containsTargetAsSource(Connection other) {
		for (Connection c : connections) {
			if (c.getSourceAlign() == other.getTargetAlign()
					&& c.getSourceIndex() == other.getTargetIndex()) {
				return true;
			}
		}
		return false;
	}
}