/**
 * @file event.h
 * @brief Structs and interfaces to handle events
 *
 * Event are defined based on the user input to trigger actions
 * for sending and receiving data (put and get action) during the
 * data exchange of (interpolated) fields.
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#ifndef EVENT_H
#define EVENT_H

#include "mtime_datetime.h"
#include "mtime_timedelta.h"
#include "mtime_eventHandling.h"

/** \example test_events.c
 * This show how to work with events.
 */

enum yac_action_type {
  NONE               = 0,
  COUPLING           = 1,
  RESTART            = 2,
  GET_FOR_RESTART    = 3,
  PUT_FOR_RESTART    = 4,
  GET_FOR_CHECKPOINT = 5,
  PUT_FOR_CHECKPOINT = 6,
  OUT_OF_BOUND       = 7,
};

enum yac_reduction_type {
  INSTANT      = 0,
  ACCUMULATE   = 1,
  TIME_AVERAGE = 2,
};

enum yac_time_unit_type {
   C_MILLISECOND = 0,
   C_SECOND      = 1,
   C_MINUTE      = 2,
   C_HOUR        = 3,
   C_DAY         = 4,
   C_MONTH       = 5,
   C_YEAR        = 6,
   C_ISO_FORMAT  = 7,
};

struct event {
  int is_set;
  struct _timedelta *model_time_step;
  struct _timedelta *coupling_time_step;
  struct _timedelta *dummy_time_step;
  int lag;
  enum yac_reduction_type time_operation;
  enum yac_action_type action;
  enum yac_time_unit_type time_unit;
  struct _datetime *startdate;
  struct _datetime *stopdate;
  struct _datetime *currentdate;
  struct _datetime *offset;
  struct _event* coupling;
};

extern int yac_number_of_events;

/**
 * Initialise an event
 * @param[in]  event_id            event handle
 */
void yac_event_init ( int * event_id );

/**
  * Fill in the information for an event
  * @param[in]  event_id            event handle
  * @param[in]  model_time_step     model time step
  * @param[in]  coupling_time_step  coupling period
  * @param[in]  time_unit           unit of coupling period and model time step
  * @param[in]  lag                 time lag (in model time steps)
  * @param[in]  time_operation      online processing of coupling fields /average/accumulate/instant
  * @param[in]  start_date          start date of the job
  * @param[in]  stop_date           end date of the job
  */
void yac_event_add ( int event_id,
                     void *model_time_step,
                     void *coupling_time_step,
                     enum  yac_time_unit_type time_unit,
                     int lag,
                     enum yac_reduction_type time_operation,
                     const char * start_date,
                     const char * stop_date);

/**
 * Delete an event
 * @param[in]  event_id            event handle
 */
void yac_event_free ( int event_id );

/**
 * Check for an event
 * @param[in]  event_id            event handle
 */
int yac_event_check ( int event_id );

/**
 * Add delta time to an event
 * @param[in]  event_id            event handle
 */
void yac_event_update ( int event_id );

/**
 * Get the time lag of an event
 * @param[in]  event_id            event handle
 */
int yac_get_event_lag ( int event_id );

/**
 * Get the time operation of an event
 * @param[in]  event_id            event handle
 */
int yac_get_event_time_operation ( int event_id );

/**
 * Get the coupling time step for an event
 * @param[in]  event_id            event handle
 * @param[in]  timedelta_str       string memory
 */
char * yac_get_event_coupling_timestep ( int event_id, char * timedelta_str );

/**
 * Get the model time step for an event
 * @param[in]  event_id            event handle
 * @param[in]  timedelta_str       string memory
 */
char * yac_get_event_model_timestep ( int event_id, char * timedelta_str );

#endif

