/**
 * @file field_interpolation_gen_src.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#ifdef HAVE_CONFIG_H
// Get the definition of the 'restrict' keyword.
#include "config.h"
#endif

#ifdef HAVE_C_NETCDF
#include <netcdf.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "field_interpolation_gen.h"
#include "field_interpolation_gen_src.h"
#include "interpolation_method.h"
#include "interpolation_method_file.h"
#include "interpolation_plan.h"
#include "utils.h"
#include "ensure_array_size.h"
#include "global_search.h"

static void
field_interpolation_gen_src_enddef(struct field_interpolation_gen * interpolation_gen);
static struct field_interpolation *
field_interpolation_gen_src_get_field_interpolation
  (struct field_interpolation_gen * interpolation_gen);
static void
field_interpolation_gen_src_add_weight_file_name
  (field_interpolation_gen interpolation_gen, char const * weight_file_name);
static void
field_interpolation_gen_src_delete(struct field_interpolation_gen * interpolation_gen);

static struct field_interpolation_gen_vtable field_interpolation_gen_src_vtable = {
                                       field_interpolation_gen_src_enddef,
                                       field_interpolation_gen_src_get_field_interpolation,
                                       field_interpolation_gen_src_add_weight_file_name,
                                       field_interpolation_gen_src_delete};

struct field_interpolation_gen_src_tags {

   unsigned results_send;
   unsigned point_data_recv;
   unsigned point_data_barrier;
   unsigned point_data_header_recv;
   unsigned halo_data_alltoallv;
   unsigned halo_exchange_send;
   unsigned halo_exchange_recv;
   unsigned tgt_point_location_recv;
   unsigned tgt_point_location_bcast;
   unsigned allreduce_num_weights_per_link;
   unsigned reduce_send_flag;
   unsigned interp_data_msg_count;
   unsigned interp_data_header;
   unsigned interp_data;
};

static void
field_interpolation_src_put_mapping_on_src(
  struct field_interpolation * interpolation, double **** data,
  int collection_size);
static void
field_interpolation_src_put_mapping_on_tgt(
  struct field_interpolation * interpolation, double **** data,
  int collection_size);
static void
field_interpolation_src_get(struct field_interpolation * interpolation,
                            double *** data, int collection_size);
static void
field_interpolation_src_delete(struct field_interpolation * interpolation);

static struct field_interpolation_vtable
  field_interpolation_src_mapping_on_src_vtable =
    {field_interpolation_src_put_mapping_on_src,
     field_interpolation_src_get,
     field_interpolation_src_delete};

static struct field_interpolation_vtable
  field_interpolation_src_mapping_on_tgt_vtable =
    {field_interpolation_src_put_mapping_on_tgt,
     field_interpolation_src_get,
     field_interpolation_src_delete};

//! \todo check which tags are actually needed
struct field_interpolation_src_tags {

   unsigned halo_exchange_send;
   unsigned halo_exchange_recv;
   unsigned put;
};

struct link_data_tags {
  unsigned alltoallv_link_data;
  unsigned allreduce_num_links;
  unsigned allreduce_meta_data;
};

struct fixed_data_tags {
  unsigned allgather_num_fixed_values;
  unsigned allgatherv_meta_data;
  unsigned alltoallv_fixed_data;
  unsigned allreduce_num_fixed;
};

/**
 * request from target processes for data that needs to be interpolated
 */
struct field_interpolation_gen_request {

   double (*coordinates_xyz)[3];
   int * global_ids;
   unsigned num_points;

   enum yac_location point_location;

   unsigned rank;
};

struct interp_data_send_msg {
   unsigned * send_points;   //!< points from the array,  that is returned by
                             //!< \ref yac_execute_interpolation_plan_element,
                             //!< that need to be sent in the put operation
   unsigned num_send_points; //!< number of points to be sent to the target
   unsigned rank;            //!< target process rank
};

struct field_data_send_msg {
   struct {
      unsigned subdomain_idx;
      unsigned pointset_idx;
      unsigned local_id;
   } * send_points;
   unsigned num_points;
   unsigned rank;
};

struct field_data_recv_msg {
   struct {
      unsigned position;
      unsigned pointset_idx;
   } * recv_points;
   unsigned num_recv_points;
   unsigned rank;
};

struct mapping_data_src {
  // messages send from source to target containing interpolated points
  //
  // contains for each target process, that gets interpolated points from
  // this process, the points from the array, that is returned by
  // \ref yac_execute_interpolation_plan_element, that need to be sent in
  // the put operation
  struct interp_data_send_msg * send_msgs;
  unsigned num_send_msgs;
};

struct mapping_data_tgt {
  struct field_data_send_msg * send_msgs;
  unsigned num_send_msgs;
};

union mapping_data {
  struct mapping_data_src src;
  struct mapping_data_tgt tgt;
};

struct field_interpolation_data_src {

  struct communicator * comm;

  struct interpolation_plan_element ** plans;

  struct communicator * halo_comm;

  unsigned total_num_interpolated_points;
  unsigned num_subdomains;
  unsigned num_pointsets;

  unsigned * num_halo_recv_points_per_points_set;
  struct field_data_send_msg * halo_send_msgs;
  struct field_data_recv_msg * halo_recv_msgs;
  unsigned total_num_halo_send_points;
  unsigned total_num_halo_recv_points;
  unsigned num_halo_send_msgs, num_halo_recv_msgs;

  union mapping_data mapping_on;
  unsigned max_send_buffer_size;

  unsigned reference_counter; //!< number of active references if the data in
                              //!< this field interpolation
};

struct field_interpolation_gen_src {

   struct field_interpolation_gen_vtable * vtable;
   struct global_search * search;
   struct interpolation_method ** methods;

   char ** weight_file_names;
   unsigned weight_file_names_array_size;
   unsigned num_weight_file_names;
   unsigned * num_interpolated_points_per_method;
   int * collected_global_ids;
   enum yac_location * src_locations;
   unsigned num_interpolation_methods;
   enum yac_location tgt_location;
   unsigned io_proc_step;
   int mapping_on_source;

   struct field_interpolation_gen_src_tags tags;
   struct field_interpolation_src_tags put_get_tags;
   struct link_data_tags link_data_tags;
   struct fixed_data_tags fixed_data_tags;

   struct field_interpolation_data_src * interp_data;
};

struct put_buffers {

  double ** buffer;
  struct comm_request ** requests;
  unsigned num_open_requests;
  int collection_size;
};

struct field_interpolation_src {

   struct field_interpolation_vtable * vtable;

   struct field_interpolation_src_tags tags;

   struct field_interpolation_data_src * interp_data;

   // buffers and communication requests used in the put operation
   struct put_buffers * put_buffers;
   unsigned put_buffers_array_size;
   unsigned num_put_buffers;
};

struct temp_halo_search_data {

   enum yac_location * src_locations; //!< point locations (dimension [point set idx])
   struct domain * local_domain;
   struct field_interpolation_gen_src * interp_src;
};

struct temp_point_search_data {

   struct interpolation_method ** methods;
   struct field_interpolation_gen_src * interp_src;
   unsigned num_interpolation_methods;
   unsigned * num_interpolated_points_per_method;

   struct field_interpolation_gen_request * target_requests;
   unsigned num_target_requests;
   unsigned num_open_target_requests;

   int * collected_global_ids;             //!< global ids of all target points
                                           //!< that are to be interpolated
   double (*collected_coordinates_xyz)[3]; //!< cartesian coordinates of all
                                           //!< target points that are to be
                                           //!< interpolated
   unsigned total_num_points;              //!< total number of target points
                                           //!< that are to be interpolated
};

struct temp_recv_init_head_data {

   struct temp_point_search_data * search_data;
   unsigned num_points;
   enum yac_location point_location;
   unsigned contains_point_data_coords;
};

struct temp_tgt_point_location_data {

  func_recv_callback callback;
  struct communicator * comm;
  struct temp_point_search_data * search_data;
  unsigned tag;
};

struct link {

  int src_index;
  int tgt_index;
  unsigned pointset_index;
  double * weights;
};

struct link_ {

  int src_index;
  int tgt_index;
  unsigned pointset_index;
  unsigned rank;
  double * weights;
};

struct temp_link_io_proc_data {

  unsigned io_proc_idx;
  unsigned num_io_procs;
  unsigned num_weights_per_link;
  char const * weight_file_name;
  unsigned * num_links_per_io_proc;
  struct link * links;
  int links_have_been_received;
  double * weights;
  unsigned num_links;
  unsigned num_pointsets;
  struct link_data_tags tags;
};

struct temp_fixed_io_proc_data {

  unsigned io_proc_idx;
  unsigned num_io_procs;
  unsigned * num_tgt_per_io_proc;
  char const * weight_file_name;
  unsigned num_fixed_values;
  int ** global_tgt_indices_per_fixed_value;
  struct fixed_data_tags tags;
};

struct temp_link_data {

  struct link_data_tags tags;

  // local data
  unsigned num_links;
  int * src_indices;
  int * tgt_indices;
  unsigned * pointset_indices;
  double * weights;

  // meta data
  enum yac_location * src_locations;
  unsigned num_pointsets;
  enum yac_location tgt_location;

  // collected data
  unsigned total_num_links;
  unsigned num_weights_per_link;
  int min_tgt_index;
  int max_tgt_index;
  unsigned * num_links_per_pointset;
};

struct fixed_value_meta_data {
  double value;
  unsigned count;
};

struct temp_fixed_data {

  struct fixed_data_tags tags;

  // local data
  int * global_tgt_indices;
  unsigned num_tgt_indices;
  unsigned * num_tgt_per_fixed_value;
  double * fixed_values;
  unsigned num_fixed_values;

  // collected data
  struct fixed_value_meta_data * collected_fixed_values;
  unsigned collected_num_fixed_values;
  int min_tgt_index;
  int max_tgt_index;
};

struct temp_weight_file_data {

  char * weight_file_name;
  char const * src_grid_name;
  char const * tgt_grid_name;

  struct temp_link_data link;
  struct temp_fixed_data fixed;

  unsigned io_proc_step;
};

struct temp_interpolation_plan_data {

  // link data
  unsigned num_tgt_indices_links;
  unsigned * num_links_per_pointset;
  unsigned num_weights_per_link;
  unsigned * num_links_per_tgt;
  int * src_indices;
  unsigned * pointset_indices;
  double * weights;

  // fixed data
  int * global_tgt_indices;
  unsigned num_tgt_indices_fixed;
  unsigned * num_tgt_per_fixed_value;
  double * fixed_values;
  unsigned num_fixed_values;
};

struct temp_halo_requests_target_data {

  struct link_ * links;
  unsigned num_links;
  unsigned num_weights_per_link;
  double * weights; // to be used to free all weights in links

  // fixed data
  int * global_tgt_indices;
  unsigned num_tgt_indices_fixed;
  unsigned * num_tgt_per_fixed_value;
  double * fixed_values;
  unsigned num_fixed_values;

  struct field_interpolation_gen_src * interp_src;
  struct field_interpolation_gen_request * target_requests;
  unsigned num_target_requests;
};

struct temp_send_flag_data {
  unsigned rank;
  unsigned tag;
  struct communicator * comm;
};

static void
write_interpolation_to_file(
  struct field_interpolation_gen_src * interp_src,
  struct temp_interpolation_plan_data * interp_plan_data);

static void generate_tag_string(char * buf, char const * tag_base_str,
                                char const * local_comp_name,
                                char const * remote_comp_name,
                                char const * local_domain_name,
                                char const * remote_domain_name,
                                char const * local_field_name,
                                char const * remote_field_name,
                                char const * tag_str) {

   strcpy(buf, tag_base_str);
   strcat(buf, local_comp_name);
   strcat(buf, remote_comp_name);
   strcat(buf, local_domain_name);
   strcat(buf, remote_domain_name);
   strcat(buf, local_field_name);
   strcat(buf, remote_field_name);
   strcat(buf, tag_str);
}

static void
generate_field_interpolation_gen_tag(struct field_interpolation_gen_src * interp_src,
                                     char const * local_field_name,
                                     char const * other_field_name) {

   char const * str_base_str                       = "field interpolation gen";
   char const * str_generator_string               = "generator string";

   char const * str_point_data_barrier             = "point data barrier";
   char const * str_halo_init_data                 = "halo init data";
   char const * str_halo_init_exchange             = "halo init data exchange";
   char const * str_tgt_point_location_bcast       = "tgt point location bcast";
   char const * str_allreduce_num_weights_per_link = "allreduce num weights per link";
   char const * str_reduce_send_flag               = "reduce send flag";

   char const * str_alltoallv_link_data            = "alltoallv link data";
   char const * str_allreduce_num_links            = "allreduce num links";
   char const * str_allreduce_meta_data            = "allreduce link meta data";

   char const * str_allgather_num_fixed_values     = "allgather num fixed values";
   char const * str_allgatherv_meta_data           = "allgatherv fixed meta data";
   char const * str_alltoallv_fixed_data           = "alltoallv fixed data";
   char const * str_allreduce_num_fixed            = "allreduce num fixed";

   char const * str_halo_exchange                  = "halo exchange data";

   struct global_search * search = interp_src->search;

   struct communicator * halo_comm;

   halo_comm = interp_src->interp_data->halo_comm;

   char const * local_comp_name = yac_get_local_component_name(search);
   char const * remote_comp_name = yac_get_remote_component_name(search);

   char const * local_domain_name = yac_get_local_domain_name(search);
   char const * remote_domain_name = yac_get_remote_domain_name(search);

   char const * remote_field_name;

   remote_field_name = other_field_name;

   unsigned max_tag_strlen;

   max_tag_strlen = strlen(str_allreduce_num_weights_per_link);

   unsigned max_str_buf_size = max_tag_strlen;
   max_str_buf_size += strlen(str_base_str);
   max_str_buf_size += strlen(local_comp_name);
   max_str_buf_size += strlen(remote_comp_name);
   max_str_buf_size += strlen(local_domain_name);
   max_str_buf_size += strlen(remote_domain_name);
   max_str_buf_size += strlen(local_field_name);
   max_str_buf_size += strlen(remote_field_name) + 1;

   char * unique_generator_string = malloc(max_str_buf_size);
   generate_tag_string(unique_generator_string, str_base_str, local_comp_name,
                       remote_comp_name, local_domain_name, remote_domain_name,
                       local_field_name, remote_field_name,
                       str_generator_string);

   char * halo_tag_strings[15];
   halo_tag_strings[0] = malloc(max_str_buf_size);
   generate_tag_string(halo_tag_strings[0], str_base_str, remote_comp_name,
                       local_comp_name, remote_domain_name, local_domain_name,
                       remote_field_name, local_field_name,
                       str_point_data_barrier);

   halo_tag_strings[1] = malloc(max_str_buf_size);
   generate_tag_string(halo_tag_strings[1], str_base_str, remote_comp_name,
                       local_comp_name, remote_domain_name, local_domain_name,
                       remote_field_name, local_field_name, str_halo_init_data);

   halo_tag_strings[2] = malloc(max_str_buf_size);
   generate_tag_string(halo_tag_strings[2], str_base_str, remote_comp_name,
                       local_comp_name, remote_domain_name, local_domain_name,
                       remote_field_name, local_field_name,
                       str_halo_init_exchange);

   halo_tag_strings[3] = malloc(max_str_buf_size);
   generate_tag_string(halo_tag_strings[3], str_base_str, remote_comp_name,
                       local_comp_name, remote_domain_name, local_domain_name,
                       remote_field_name, local_field_name,
                       str_tgt_point_location_bcast);

   halo_tag_strings[4] = malloc(max_str_buf_size);
   generate_tag_string(halo_tag_strings[4], str_base_str, local_comp_name,
                       remote_comp_name, local_domain_name, remote_domain_name,
                       local_field_name, remote_field_name,
                       str_allreduce_num_weights_per_link);

   halo_tag_strings[5] = malloc(max_str_buf_size);
   generate_tag_string(halo_tag_strings[5], str_base_str, local_comp_name,
                       remote_comp_name, local_domain_name, remote_domain_name,
                       local_field_name, remote_field_name,
                       str_reduce_send_flag);

   halo_tag_strings[6] = malloc(max_str_buf_size);
   generate_tag_string(halo_tag_strings[6], str_base_str, local_comp_name,
                       remote_comp_name, local_domain_name, remote_domain_name,
                       local_field_name, remote_field_name,
                       str_alltoallv_link_data);

   halo_tag_strings[7] = malloc(max_str_buf_size);
   generate_tag_string(halo_tag_strings[7], str_base_str, local_comp_name,
                       remote_comp_name, local_domain_name, remote_domain_name,
                       local_field_name, remote_field_name,
                       str_allreduce_num_links);

   halo_tag_strings[8] = malloc(max_str_buf_size);
   generate_tag_string(halo_tag_strings[8], str_base_str, local_comp_name,
                       remote_comp_name, local_domain_name, remote_domain_name,
                       local_field_name, remote_field_name,
                       str_allreduce_meta_data);

   halo_tag_strings[9] = malloc(max_str_buf_size);
   generate_tag_string(halo_tag_strings[9], str_base_str, local_comp_name,
                       remote_comp_name, local_domain_name, remote_domain_name,
                       local_field_name, remote_field_name,
                       str_allgather_num_fixed_values);

   halo_tag_strings[10] = malloc(max_str_buf_size);
   generate_tag_string(halo_tag_strings[10], str_base_str, local_comp_name,
                       remote_comp_name, local_domain_name, remote_domain_name,
                       local_field_name, remote_field_name,
                       str_allgatherv_meta_data);

   halo_tag_strings[11] = malloc(max_str_buf_size);
   generate_tag_string(halo_tag_strings[11], str_base_str, local_comp_name,
                       remote_comp_name, local_domain_name, remote_domain_name,
                       local_field_name, remote_field_name,
                       str_alltoallv_fixed_data);

   halo_tag_strings[12] = malloc(max_str_buf_size);
   generate_tag_string(halo_tag_strings[12], str_base_str, local_comp_name,
                       remote_comp_name, local_domain_name, remote_domain_name,
                       local_field_name, remote_field_name,
                       str_allreduce_num_fixed);

   halo_tag_strings[13] = malloc(max_str_buf_size);
   generate_tag_string(halo_tag_strings[13], str_base_str, remote_comp_name,
                       local_comp_name, remote_domain_name, local_domain_name,
                       remote_field_name, local_field_name, str_halo_exchange);
   halo_tag_strings[14] = malloc(max_str_buf_size);
   generate_tag_string(halo_tag_strings[14], str_base_str, remote_comp_name,
                       local_comp_name, remote_domain_name, local_domain_name,
                       remote_field_name, local_field_name, str_halo_exchange);

   unsigned halo_tags[15];

   yac_get_unique_comm_tags(
      halo_comm, unique_generator_string, (char const **)halo_tag_strings, 15,
      halo_tags);

   free(unique_generator_string);

   interp_src->tags.point_data_barrier = halo_tags[0];
   interp_src->tags.halo_data_alltoallv = halo_tags[1];
   interp_src->tags.halo_exchange_send = halo_tags[2];
   interp_src->tags.halo_exchange_recv = halo_tags[2];
   interp_src->tags.tgt_point_location_bcast = halo_tags[3];
   interp_src->tags.allreduce_num_weights_per_link = halo_tags[4];
   interp_src->tags.reduce_send_flag = halo_tags[5];
   interp_src->link_data_tags.alltoallv_link_data = halo_tags[6];
   interp_src->link_data_tags.allreduce_num_links = halo_tags[7];
   interp_src->link_data_tags.allreduce_meta_data = halo_tags[8];
   interp_src->fixed_data_tags.allgather_num_fixed_values = halo_tags[9];
   interp_src->fixed_data_tags.allgatherv_meta_data = halo_tags[10];
   interp_src->fixed_data_tags.alltoallv_fixed_data = halo_tags[11];
   interp_src->fixed_data_tags.allreduce_num_fixed = halo_tags[12];
   interp_src->put_get_tags.halo_exchange_send = halo_tags[13];
   interp_src->put_get_tags.halo_exchange_recv = halo_tags[14];

   for (unsigned i = 0; i < 15; ++i) free(halo_tag_strings[i]);
}

static int compare_int(const void * a, const void * b) {

  return ((*(int*)a) > (*(int*)b)) - (*((int*)a) < (*(int*)b));
}

/**
 * merges all collected target requests (coordinates and global ids) and
 * applies the interpolation method stack to this information
 */
static void
collect_target_requests_barrier_callback(struct communicator * comm,
                                         struct comm_message msg,
                                         void * user_data) {

   UNUSED(comm);
   UNUSED(msg);

   struct temp_point_search_data * search_data =
      (struct temp_point_search_data *)user_data;

   unsigned total_num_points = 0;
   int point_coordinates_are_available = 1;

   struct field_interpolation_gen_request * target_requests =
      search_data->target_requests;
   unsigned num_target_requests = search_data->num_target_requests;

   for (unsigned i = 0; i < num_target_requests; ++i) {
      total_num_points += target_requests[i].num_points;
      if (target_requests[i].num_points > 0)
      point_coordinates_are_available &=
         target_requests[i].coordinates_xyz != NULL;
   }

   int * collected_global_ids =
      malloc(total_num_points * sizeof(*collected_global_ids));
   double (*collected_coordinates_xyz)[3] = NULL;

   if (point_coordinates_are_available) {
      collected_coordinates_xyz =
         malloc(total_num_points * sizeof(*collected_coordinates_xyz));
   }

   enum yac_location tgt_point_location =
      target_requests[0].point_location;

   // merge information of all target requests
   for (unsigned i = 0, offset = 0; i < num_target_requests; ++i) {

      memcpy(collected_global_ids + offset, target_requests[i].global_ids,
             target_requests[i].num_points * sizeof(*collected_global_ids));
      if (point_coordinates_are_available)
         memcpy(collected_coordinates_xyz + offset,
                target_requests[i].coordinates_xyz,
                target_requests[i].num_points *
                sizeof(*collected_coordinates_xyz));

      // we do not need the coordinates from the request anymore
      free(target_requests[i].coordinates_xyz);
      target_requests[i].coordinates_xyz = NULL;

      // sort the global ids, makes some lookups easier later on
      yac_mergesort(target_requests[i].global_ids,
                    target_requests[i].num_points,
                    sizeof(*(target_requests[i].global_ids)), compare_int);

      offset += target_requests[i].num_points;

      if (tgt_point_location != target_requests[i].point_location)
        yac_internal_abort_message("ERROR: mismatching target point location"
                                   "(process_collected_target_requests)\n",
                                   __FILE__, __LINE__);
   }

   //remove duplicated target points (occurs when a point is required by more
   // than one process)
   if (point_coordinates_are_available) {

      int * collected_global_ids_idx =
         malloc(total_num_points * sizeof(*collected_global_ids_idx));
      for (unsigned i = 0; i < total_num_points; ++i)
         collected_global_ids_idx[i] = (int)i;

      yac_quicksort_index(
         collected_global_ids, total_num_points, collected_global_ids_idx);
      unsigned new_total_num_points = total_num_points;

      for (unsigned i = 1; i < total_num_points; ++i) {
         if (collected_global_ids[i-1] == collected_global_ids[i]) {
            collected_global_ids_idx[i] = INT_MAX;
            new_total_num_points--;
         }
      }
      yac_quicksort_index(
         collected_global_ids_idx, total_num_points, collected_global_ids);

      // if there are duplicated global ids
      if (new_total_num_points != total_num_points) {

         total_num_points = new_total_num_points;
         collected_global_ids =
            realloc(collected_global_ids,
                    new_total_num_points * sizeof(*collected_global_ids));

         if (point_coordinates_are_available) {

            unsigned i = 0;
            while (collected_global_ids_idx[i] == (int)i) ++i;

            int idx;
            while ((idx = collected_global_ids_idx[i]) != INT_MAX) {

               collected_coordinates_xyz[i][0] =
                  collected_coordinates_xyz[idx][0];
               collected_coordinates_xyz[i][1] =
                  collected_coordinates_xyz[idx][1];
               collected_coordinates_xyz[i][2] =
                  collected_coordinates_xyz[idx][2];
               ++i;
            }

            collected_coordinates_xyz =
               realloc(collected_coordinates_xyz, new_total_num_points *
                       sizeof(*collected_coordinates_xyz));
         }
      }
      free(collected_global_ids_idx);
   }

   search_data->total_num_points = total_num_points;
   search_data->collected_global_ids = collected_global_ids;
   search_data->interp_src->collected_global_ids = collected_global_ids;
   search_data->collected_coordinates_xyz = collected_coordinates_xyz;

   struct field_interpolation_gen_src * interp_src = search_data->interp_src;
   interp_src->tgt_location = tgt_point_location;

   // apply method stack to collected target information
   yac_do_search_interpolation_method(
      search_data->methods, interp_src->interp_data->plans,
      collected_global_ids, collected_coordinates_xyz, tgt_point_location,
      total_num_points, search_data->num_interpolated_points_per_method,
      interp_src->interp_data->num_pointsets);
}

/**
 * receives and collects the points that have to be interpolated
 * by the target processes
 *
 * once all points are collected an interpolation method stack
 * is applied to the collected data
 *
 * the matching comm_recv_callback can be found in the
 * \ref receive_interpolation_init_header_callback
 */
static void
receive_interpolation_point_data_callback(struct communicator * comm,
                                          struct comm_message msg,
                                          void * user_data) {

   struct temp_recv_init_head_data * temp_data =
     (struct temp_recv_init_head_data *)user_data;

   int * global_ids = malloc(temp_data->num_points * sizeof(*global_ids));
   double (*coordinates_xyz)[3] = NULL;

   // receive point coordinates and global ids and unpack them
   if (temp_data-> num_points > 0 ) {

      int position = 0;

      if (temp_data->contains_point_data_coords) {

         coordinates_xyz = malloc(temp_data->num_points * sizeof(*coordinates_xyz));

         for (unsigned i = 0; i < temp_data->num_points; ++i) {

            yac_comm_unpack(msg.buffer, msg.count, &position, global_ids + i,
                            1, COMM_INT, comm);
            yac_comm_unpack(msg.buffer, msg.count, &position,
                            coordinates_xyz[i], 3, COMM_DBLE, comm);
         }

      } else
         yac_comm_unpack(msg.buffer, msg.count, &position, global_ids,
                         temp_data->num_points, COMM_INT, comm);
   }

   // set up target request
   struct field_interpolation_gen_request * request;
   request = temp_data->search_data->target_requests +
      temp_data->search_data->num_target_requests++;
   request->coordinates_xyz = coordinates_xyz;
   request->global_ids = global_ids;
   request->num_points = temp_data->num_points;
   request->point_location = temp_data->point_location;
   request->rank = msg.src_rank;

   // if the last message was received, wait until source processes have
   // received the target points, this ensures reduced idle time
   //! \todo make this operation atomic
   if (!--(temp_data->search_data->num_open_target_requests))
      yac_comm_barrier_callback(
         temp_data->search_data->interp_src->tags.point_data_barrier,
         temp_data->search_data->interp_src->interp_data->halo_comm,
         temp_data->search_data, collect_target_requests_barrier_callback,
         MSG_PRIO_HIGH);

   free(temp_data);
}

/**
 * receives header information about points that have to be interpolated
 * by the target processes
 *
 * sets up receive for the target point data if necessary
 *
 * the matching comm_recv_callback can be found in the
 * field_interpolation_gen_src constructor (\ref yac_new_field_interpolation_gen_src)
 */
static void
receive_interpolation_init_header_callback(struct communicator * comm,
                                           struct comm_message msg,
                                           void * user_data) {

   struct temp_recv_init_head_data * temp_data = malloc(1 * sizeof(*temp_data));

   temp_data->search_data = (struct temp_point_search_data *) user_data;

   temp_data->num_points = ((unsigned*)(msg.buffer))[0];
   temp_data->point_location =
     (enum yac_location)(((unsigned*)(msg.buffer))[1]);
   temp_data->contains_point_data_coords = ((unsigned*)(msg.buffer))[2];

   // receive point coordinates and global ids
   if (temp_data->num_points > 0) {
      yac_comm_recv_callback(
         ((unsigned*)(msg.buffer))[3], COMM_PACKED, msg.src_rank,
         temp_data->search_data->interp_src->tags.point_data_recv, comm,
         temp_data, receive_interpolation_point_data_callback, NULL,
         MSG_PRIO_HIGH);
   } else {
     struct comm_message temp_msg = msg;

     temp_msg.buffer = NULL;
     temp_msg.type = COMM_PACKED;
     temp_msg.count = 0;
     temp_msg.tag = temp_data->search_data->interp_src->tags.point_data_recv;
     receive_interpolation_point_data_callback(comm, temp_msg, temp_data);
   }
}

static int compare_interp_data_send_msgs (const void * a, const void * b) {
  return (((struct interp_data_send_msg*)a)->rank -
          ((struct interp_data_send_msg*)b)->rank);
}

/**
 * returns to the target processes the global ids of the points that are interpolated
 * by the local source process
 *
 * in addition the send_msgs for the put operation are generated
 */
static void
return_interpolation_points(struct field_interpolation_gen_src * interp_src,
                            unsigned num_interpolation_methods,
                            unsigned * num_interpolated_points_per_method,
                            unsigned num_target_requests,
                            struct field_interpolation_gen_request * target_requests,
                            int * collected_global_ids) {

   unsigned total_num_interpolated_points = 0;

   for (unsigned i = 0; i < num_interpolation_methods; ++i)
      total_num_interpolated_points +=
         num_interpolated_points_per_method[i];

   interp_src->interp_data->total_num_interpolated_points =
     total_num_interpolated_points;

   int * global_id_idx = malloc(total_num_interpolated_points *
                                sizeof(*global_id_idx));

   for (unsigned i = 0; i< total_num_interpolated_points; ++i)
      global_id_idx[i] = i;

   yac_quicksort_index(collected_global_ids, total_num_interpolated_points,
                       global_id_idx);

   //---------------------------------------------------------------------
   // return to the target processes the global ids of the points that are
   // interpolated by the local source process and set up send_msgs
   //---------------------------------------------------------------------

   struct mapping_data_src * mapping_on_src =
      &(interp_src->interp_data->mapping_on.src);
   struct interp_data_send_msg * send_msgs =
      malloc(num_target_requests * sizeof(*send_msgs));
   unsigned num_send_msgs = num_target_requests;
   struct communicator * comm = interp_src->interp_data->comm;

   mapping_on_src->num_send_msgs = num_target_requests;
   mapping_on_src->send_msgs = send_msgs;
   interp_src->interp_data->max_send_buffer_size = 0;

   int * interpolated_points; // global ids of interpolated points that were
                              // requested by the current target process
   unsigned * send_points;

   {
      unsigned max_num_points = 0;
      for (unsigned i = 0; i < num_target_requests; ++i)
         if (target_requests[i].num_points > max_num_points)
            max_num_points = target_requests[i].num_points;

      interpolated_points =
         malloc(max_num_points * sizeof(*interpolated_points));
      send_points = malloc(max_num_points * sizeof(*send_points));
   }

   for (unsigned i = 0; i < num_target_requests; ++i) {

      unsigned curr_num_interpolated_points = 0;

      for (unsigned j = 0, k = 0; j < target_requests[i].num_points; ++j) {

         while ((k < total_num_interpolated_points) &&
                (collected_global_ids[k] < target_requests[i].global_ids[j]))
            ++k;

         if ((k < total_num_interpolated_points) &&
             (collected_global_ids[k] == target_requests[i].global_ids[j])) {

            interpolated_points[curr_num_interpolated_points] =
               collected_global_ids[k];
            send_points[curr_num_interpolated_points] = global_id_idx[k];
            curr_num_interpolated_points++;
         }
      }

      send_msgs[i].send_points =
        malloc(curr_num_interpolated_points *
               sizeof(*(send_msgs[i].send_points)));
      memcpy(send_msgs[i].send_points, send_points,
             curr_num_interpolated_points * sizeof(*send_points));
      send_msgs[i].num_send_points = curr_num_interpolated_points;
      send_msgs[i].rank = target_requests[i].rank;
      if (interp_src->interp_data->max_send_buffer_size <
          curr_num_interpolated_points)
         interp_src->interp_data->max_send_buffer_size =
            curr_num_interpolated_points;

      // return global ids of interpolation points to targets
      // the matching receive can be found in the field_interpolation_gen_tgt
      // routine \ref send_interpolation_data
      yac_comm_bsend(interpolated_points, curr_num_interpolated_points, COMM_INT,
                     target_requests[i].rank, interp_src->tags.results_send,
                     comm);
   }

   free(send_points);
   free(interpolated_points);

   // reorder send messages
   unsigned rank = yac_get_comm_rank(comm);
   unsigned size = yac_get_comm_size(comm);
   for (unsigned i = 0; i < num_send_msgs; ++i)
     if (send_msgs[i].rank < rank) send_msgs[i].rank += size;

   // the send order is invers to the receive order
   qsort(send_msgs, num_send_msgs, sizeof(*send_msgs),
         compare_interp_data_send_msgs);
   for (unsigned i = 0; i < num_send_msgs; ++i)
     if (send_msgs[i].rank >= size) send_msgs[i].rank -= size;

   // bring collected_global_ids into original order
   yac_quicksort_index(global_id_idx, total_num_interpolated_points,
                       collected_global_ids);

   free(global_id_idx);
}

static int compare_field_data_send_msgs (const void * a, const void * b) {
  return (((struct field_data_send_msg*)a)->rank -
          ((struct field_data_send_msg*)b)->rank);
}

/**
 * receive the halo requests from other source processes
 *
 * stores these points in interpolation.halo_send_points, which is
 * then used in the halo exchange before the interpolation in a put
 * operation
 *
 * the matching yac_comm_alltoallv_callback can be found in the
 * \ref exchange_halo_requests_source
 */
static void
alltoallv_interpolation_halo_callback(struct communicator * comm,
                                      struct comm_message msg,
                                      void * user_data) {

  struct temp_halo_search_data * halo_search_data =
    (struct temp_halo_search_data *)user_data;
  struct field_interpolation_gen_src * interp_src =
    halo_search_data->interp_src;
  unsigned num_pointsets = interp_src->interp_data->num_pointsets;

  int * buffer = (int*)(msg.buffer);
  unsigned halo_comm_rank = yac_get_comm_rank(comm);
  unsigned halo_comm_size = yac_get_comm_size(comm);
  unsigned num_halo_send_msgs = 0;
  unsigned total_num_halo_send_points = 0;
  // compute the number of processes that require data from the local process
  for (unsigned i = 0; i < halo_comm_size; ++i) {

    int * num_points_per_pointset = buffer;
    int num_halo_points = 0;

    for (unsigned j = 0; j < num_pointsets; ++j)
      num_halo_points += num_points_per_pointset[j];

    num_halo_send_msgs += (num_halo_points > 0);
    total_num_halo_send_points += num_halo_points;
    buffer += num_halo_points + num_pointsets;
  }

  struct field_data_send_msg * halo_send_msgs =
    malloc(num_halo_send_msgs * sizeof(*halo_send_msgs));

  interp_src->interp_data->total_num_halo_send_points =
    total_num_halo_send_points;
  interp_src->interp_data->halo_send_msgs = halo_send_msgs;
  interp_src->interp_data->num_halo_send_msgs = num_halo_send_msgs;

  struct field_data_send_msg * curr_halo_send_msg = halo_send_msgs;
  buffer = (int*)(msg.buffer);
  int * global_ids_sort_idx = NULL;
  size_t global_ids_sort_idx_array_size = 0;
  unsigned * found_flag = NULL;
  size_t found_flag_array_size = 0;
  unsigned * matching_global_ids_idx = NULL;
  size_t matching_global_ids_idx_array_size = 0;
  unsigned * local_ids = NULL;
  size_t local_ids_array_size = 0;

  for (unsigned i = 0; i < halo_comm_size; ++i) {

    int * num_points_per_pointset = buffer;
    buffer += num_pointsets;

    unsigned curr_num_halo_points = 0;
    for (unsigned j = 0; j < num_pointsets; ++j)
      curr_num_halo_points += num_points_per_pointset[j];

    if (curr_num_halo_points == 0) continue;

    curr_halo_send_msg->send_points =
      malloc(curr_num_halo_points * sizeof(*(curr_halo_send_msg->send_points)));
    curr_halo_send_msg->num_points = curr_num_halo_points;
    curr_halo_send_msg->rank = i;

    for (unsigned j = 0, pointset_offset = 0; j < num_pointsets;
         pointset_offset += num_points_per_pointset[j++]) {

      if (num_points_per_pointset[j] == 0) continue;

      ENSURE_ARRAY_SIZE(global_ids_sort_idx, global_ids_sort_idx_array_size,
                        (size_t)(num_points_per_pointset[j]));
      ENSURE_ARRAY_SIZE(found_flag, found_flag_array_size,
                        (size_t)(num_points_per_pointset[j]));
      ENSURE_ARRAY_SIZE(matching_global_ids_idx,
                        matching_global_ids_idx_array_size,
                        (size_t)(num_points_per_pointset[j]));
      ENSURE_ARRAY_SIZE(local_ids, local_ids_array_size,
                        (size_t)(num_points_per_pointset[j]));

      for (int k = 0; k < num_points_per_pointset[j]; ++k) {
        global_ids_sort_idx[k] = k;
        found_flag[k] = 0;
      }

      // sort the global ids (required by matching_global_ids_idx)
      yac_quicksort_index(buffer, num_points_per_pointset[j],
                          global_ids_sort_idx);

      for (unsigned k = 0; k < interp_src->interp_data->num_subdomains; ++k) {

        unsigned num_matching_global_ids;

        yac_match_global_ids(halo_search_data->local_domain, k,
                             halo_search_data->src_locations[j], buffer,
                             num_points_per_pointset[j], NULL,
                             matching_global_ids_idx, &num_matching_global_ids,
                             local_ids, 0);

        for (unsigned l = 0; l < num_matching_global_ids; ++l) {

           unsigned point_idx = pointset_offset +
                                global_ids_sort_idx[matching_global_ids_idx[l]];
           found_flag[global_ids_sort_idx[matching_global_ids_idx[l]]] = 1;
           curr_halo_send_msg->send_points[point_idx].subdomain_idx = k;
           curr_halo_send_msg->send_points[point_idx].pointset_idx = j;
           curr_halo_send_msg->send_points[point_idx].local_id = local_ids[l];
        }
      }

      buffer += num_points_per_pointset[j];

      for (int k = 0; k < num_points_per_pointset[j]; ++k)
        if (!found_flag[k])
          yac_internal_abort_message(
            "ERROR: could not find all received global ids in local data "
            "(alltoallv_interpolation_halo_callback) (Is "
            "halo data defined correctly?)\n", __FILE__, __LINE__);
    }
    curr_halo_send_msg++;
  }

  free(local_ids);
  free(matching_global_ids_idx);
  free(found_flag);
  free(global_ids_sort_idx);
  free(halo_search_data);

  // reorder send messages
  for (unsigned i = 0; i < num_halo_send_msgs; ++i)
    if (halo_send_msgs[i].rank < halo_comm_rank)
      halo_send_msgs[i].rank += halo_comm_size;
  // the send order is invers to the receive order
  qsort(halo_send_msgs, num_halo_send_msgs, sizeof(*halo_send_msgs),
        compare_field_data_send_msgs);
  for (unsigned i = 0; i < num_halo_send_msgs; ++i)
    if (halo_send_msgs[i].rank >= halo_comm_size)
      halo_send_msgs[i].rank -= halo_comm_size;
}

static int compare_links_tgt_rank(const void * a, const void * b) {
  struct link_ * link_a = (struct link_*)a;
  struct link_ * link_b = (struct link_*)b;

  return (*(int*)&(link_a->rank) - *(int*)&(link_b->rank));
}

static int compare_links_global_target_index(const void * a, const void * b) {
  struct link_ * link_a = (struct link_*)a;
  struct link_ * link_b = (struct link_*)b;

  return (link_a->tgt_index - link_b->tgt_index);
}

static int compare_links_pointset (const void * a, const void * b) {
  struct link_ * link_a = (struct link_*)a;
  struct link_ * link_b = (struct link_*)b;

  if (link_a->pointset_index != link_b->pointset_index)
    return (link_a->pointset_index - link_b->pointset_index);
  if (link_a->src_index != link_b->src_index)
    return (link_a->src_index - link_b->src_index);
  return (link_a->tgt_index - link_b->tgt_index);
}

static int compare_links_src_rank (const void * a, const void * b) {
  struct link_ * link_a = (struct link_*)a;
  struct link_ * link_b = (struct link_*)b;

  if (link_a->rank != link_b->rank)
    return (link_a->rank - link_b->rank);
  return (link_a->tgt_index - link_b->tgt_index);
}

/**
 * this is only called by the root of the source processes (in case of
 * mapping_on_source == false)
 * it receives the number of link messages sent from the source to the target
 * processes
 * the source process sends these message numbers to that target root
 */
static void
reduce_send_flags_callback(struct communicator * halo_comm,
                           struct comm_message msg,
                           void * user_data) {

  UNUSED(halo_comm);

  struct temp_send_flag_data * send_flag_data =
    (struct temp_send_flag_data *)user_data;

  yac_comm_bsend(msg.buffer, msg.count, msg.type, send_flag_data->rank,
                 send_flag_data->tag, send_flag_data->comm);

  free(send_flag_data);
}

static struct field_data_send_msg
generate_send_msg(struct link_ * requested_links, unsigned num_requested_links,
                  unsigned tgt_rank, struct domain * domain,
                  enum yac_location * src_locations) {

  struct field_data_send_msg msg;

  unsigned prev_pointset_index = -1;
  int prev_src_index = -1;
  unsigned num_points = 0;
  unsigned max_pointset_index = 0;

  // count the number of points to be sent
  for (unsigned i = 0; i < num_requested_links; ++i) {
    if ((prev_pointset_index != requested_links[i].pointset_index) ||
        (prev_src_index != requested_links[i].src_index)) {
      ++num_points;
      if (requested_links[i].pointset_index > max_pointset_index)
        max_pointset_index = requested_links[i].pointset_index;
      prev_pointset_index = requested_links[i].pointset_index;
      prev_src_index = requested_links[i].src_index;
    }
  }

  // get all global src_indices that need to be sent
  int * global_src_indices = malloc(num_points * sizeof(*global_src_indices));
  unsigned * num_points_per_pointset =
    calloc((max_pointset_index + 1), sizeof(*num_points_per_pointset));
  unsigned * local_ids = malloc(num_points * sizeof(*local_ids));
  unsigned * matching_global_ids_idx =
    malloc(num_points * sizeof(*matching_global_ids_idx));
  prev_pointset_index = -1;
  prev_src_index = -1;
  num_points = 0;
  for (unsigned i = 0; i < num_requested_links; ++i) {
    if ((prev_pointset_index != requested_links[i].pointset_index) ||
        (prev_src_index != requested_links[i].src_index)) {

      global_src_indices[num_points++] = requested_links[i].src_index;
      num_points_per_pointset[requested_links[i].pointset_index]++;

      prev_pointset_index = requested_links[i].pointset_index;
      prev_src_index = requested_links[i].src_index;
    }
  }

  msg.send_points = malloc(num_points * sizeof(*(msg.send_points)));
  msg.num_points = num_points;
  msg.rank = tgt_rank;
  num_points = 0;

  unsigned num_subdomains = yac_get_num_subdomains(domain);

  // get the local ids of the points that need to be sent
  for (unsigned pointset_index = 0, offset = 0;
       pointset_index <= max_pointset_index;
       offset += num_points_per_pointset[pointset_index++]) {

    for (unsigned subdomain_idx = 0; subdomain_idx < num_subdomains;
         ++subdomain_idx) {

      unsigned num_matching_global_ids;

      yac_match_global_ids(domain, subdomain_idx, src_locations[pointset_index],
                           global_src_indices + offset,
                           num_points_per_pointset[pointset_index], NULL,
                           matching_global_ids_idx, &num_matching_global_ids,
                           local_ids, 0);

      for (unsigned i = 0; i < num_matching_global_ids; ++i) {

        msg.send_points[offset + matching_global_ids_idx[i]].subdomain_idx =
          subdomain_idx;
        msg.send_points[offset + matching_global_ids_idx[i]].pointset_idx =
          pointset_index;
        msg.send_points[offset + matching_global_ids_idx[i]].local_id =
          local_ids[i];
      }
    }
  }

  free(matching_global_ids_idx);
  free(local_ids);
  free(num_points_per_pointset);
  free(global_src_indices);

  return msg;
}

/**
 * receive the links from other source processes that are based on data from
 * this process
 *
 * the link data and fixed value is sent to the respective target processes and
 * matching send lists are generated
 *
 * the matching yac_comm_alltoallv_callback can be found in
 * \ref exchange_halo_requests_target
 */
static void
alltoallv_halo_data_callback(struct communicator * halo_comm,
                             struct comm_message msg, void * user_data) {

  struct temp_halo_requests_target_data * temp_data =
    (struct temp_halo_requests_target_data *)user_data;
  struct field_interpolation_gen_src * interp_src = temp_data->interp_src;
  struct communicator * comm = interp_src->interp_data->comm;
  unsigned comm_size = yac_get_comm_size(comm);
  struct link_ * links = temp_data->links;
  unsigned num_links = temp_data->num_links;
  unsigned num_weights_per_link = temp_data->num_weights_per_link;

  int link_pack_size =
    yac_comm_pack_size(2, COMM_INT, halo_comm); // src_index + tgt_index
  link_pack_size +=
    yac_comm_pack_size(2, COMM_UINT, halo_comm); // pointset_index + src_rank
  link_pack_size +=
    yac_comm_pack_size(num_weights_per_link, COMM_DBLE, halo_comm); // weights

  unsigned max_num_remote_links = msg.count / link_pack_size;
  struct link_ * remote_links =
    malloc(max_num_remote_links * sizeof(*remote_links));
  double * remote_weights =
    malloc(max_num_remote_links * num_weights_per_link * sizeof(*remote_weights));
  unsigned num_remote_links = 0;

  unsigned offset = 0;
  for (unsigned i = 0; offset < msg.count;) {

    int pos = 0;
    char * curr_packing_unit = (char*)msg.buffer + (size_t)offset;
    int max_packing_unit_size = (int)(msg.count - offset);

    unsigned curr_num_remote_links;
    yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &pos,
                    &curr_num_remote_links, 1, COMM_UINT, comm);

    num_remote_links += curr_num_remote_links;

    for (unsigned j = 0; j < curr_num_remote_links; ++j, ++i) {
      yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &pos,
                      &(remote_links[i].src_index), 1, COMM_INT, comm);
      yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &pos,
                      &(remote_links[i].tgt_index), 1, COMM_INT, comm);
      yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &pos,
                      &(remote_links[i].pointset_index), 1, COMM_UINT, comm);
      yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &pos,
                      &(remote_links[i].rank), 1, COMM_UINT, comm);
      remote_links[i].weights = remote_weights + i * num_weights_per_link;
      yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &pos,
                      remote_links[i].weights, num_weights_per_link, COMM_DBLE,
                      comm);
    }

    offset += (unsigned)pos;
  }

  // sort remote links by rank (tgt)
  qsort(remote_links, num_remote_links, sizeof(*remote_links),
        compare_links_tgt_rank);
  // sort remote links by global target index
  qsort(links, num_links, sizeof(*links), compare_links_global_target_index);

  struct link_ * requested_links = NULL;
  size_t requested_links_array_size = 0;

  char * send_buffer = NULL;
  unsigned send_buffer_size = 0;

  unsigned * send_flags = calloc(comm_size, sizeof(*send_flags));

  // for ranks in comm
  for (unsigned tgt_rank = 0; tgt_rank < comm_size; ++tgt_rank) {

    unsigned num_fixed_values = 0;
    unsigned num_tgt_indices_fixed = 0;

    struct field_interpolation_gen_request * target_request = NULL;
    for (unsigned i = 0; i < temp_data->num_target_requests; ++i) {
      if (temp_data->target_requests[i].rank == tgt_rank) {
        target_request = temp_data->target_requests + i;
        break;
      }
    }

    { // count number of links and fixed values that have to be sent to the
      // current source process

      unsigned link_count = 0;

      for (unsigned j = 0; j < num_remote_links; ++j)
        if (remote_links[j].rank == tgt_rank) link_count++;

      if (target_request != NULL) {
        for (unsigned j = 0, k = 0;
             (j < target_request->num_points) && (k < num_links);
             ++j) {
          int tgt_index = target_request->global_ids[j];
          while ((k < num_links) && (links[k].tgt_index < tgt_index)) ++k;
          while ((k < num_links) && (links[k].tgt_index == tgt_index)) {
            ++k;
            ++link_count;
          }
        }

        for (unsigned i = 0, k = 0, l = 0; i < temp_data->num_fixed_values; ++i) {
          unsigned flag = 0;
          for (unsigned j = 0; j < temp_data->num_tgt_per_fixed_value[i]; ++j, ++k) {

            while ((l < target_request->num_points) &&
                   (target_request->global_ids[l] <
                    temp_data->global_tgt_indices[k])) ++l;

            if ((l < target_request->num_points) &&
                (target_request->global_ids[l] ==
                 temp_data->global_tgt_indices[k])) {

              flag = 1;
              num_tgt_indices_fixed++;
            }
          }
          num_fixed_values += flag;
        }
      } // if (target_request != NULL)

      if ((link_count == 0) && (num_fixed_values == 0)) continue;

      size_t min_send_buffer_size = link_count * link_pack_size;
      min_send_buffer_size +=
        yac_comm_pack_size(4, COMM_UINT, comm); // num_weights_per_link +
                                                // num_requested_links +
                                                // num_packed_fixed +
                                                // num_tgt_indices_fixed
      min_send_buffer_size +=
        yac_comm_pack_size(
          num_fixed_values, COMM_UINT, comm); // num_points_per_fixed_value
      min_send_buffer_size +=
        yac_comm_pack_size(
          num_tgt_indices_fixed, COMM_INT, comm); // num_tgt_indices_fixed
      min_send_buffer_size +=
        yac_comm_pack_size(num_fixed_values, COMM_DBLE, comm); // fixed values

      ENSURE_ARRAY_SIZE(requested_links, requested_links_array_size, link_count);
      ENSURE_ARRAY_SIZE(send_buffer, send_buffer_size, min_send_buffer_size);
    }

    send_flags[tgt_rank] = 1;

    unsigned num_requested_links = 0;

    // for all matching remote links
    for (unsigned j = 0; j < num_remote_links; ++j)
      if (remote_links[j].rank == tgt_rank)
        requested_links[num_requested_links++] = remote_links[j];

    if (target_request != NULL) {

      // for all requested global target ids
      for (unsigned j = 0, k = 0;
           (j < target_request->num_points) && (k < num_links); ++j) {

        // find matching local links
        int tgt_index = target_request->global_ids[j];
        while ((k < num_links) && (links[k].tgt_index < tgt_index)) ++k;
        while ((k < num_links) && (links[k].tgt_index == tgt_index))
          requested_links[num_requested_links++] = links[k++];
      }
    }

    // sort requested link by pointset_index, src_index, tgt_index
    qsort(requested_links, num_requested_links, sizeof(*requested_links),
          compare_links_pointset);

    int pos = 0;

    // pack number of weights per links
    yac_comm_pack(&num_weights_per_link, 1, COMM_UINT, send_buffer,
                  send_buffer_size, &pos, comm);

    // pack number of links
    yac_comm_pack(&num_requested_links, 1, COMM_UINT, send_buffer,
                  send_buffer_size, &pos, comm);

    // pack links
    for (unsigned j = 0; j < num_requested_links; ++j) {
      yac_comm_pack(&(requested_links[j].src_index), 1, COMM_INT, send_buffer,
                    send_buffer_size, &pos, comm);
      yac_comm_pack(&(requested_links[j].tgt_index), 1, COMM_INT, send_buffer,
                    send_buffer_size, &pos, comm);
      yac_comm_pack(&(requested_links[j].pointset_index), 1, COMM_UINT,
                    send_buffer, send_buffer_size, &pos, comm);
      yac_comm_pack(requested_links[j].weights, num_weights_per_link,
                    COMM_DBLE, send_buffer, send_buffer_size, &pos, comm);
    }

    // pack number of fixed values
    yac_comm_pack(&num_fixed_values, 1, COMM_UINT, send_buffer,
                  send_buffer_size, &pos, comm);

    // pack total number of fixed target points
    yac_comm_pack(&num_tgt_indices_fixed, 1, COMM_UINT, send_buffer,
                  send_buffer_size, &pos, comm);

    int * temp_fixed_global_tgt_ids =
      malloc(temp_data->num_tgt_indices_fixed *
             sizeof(*temp_fixed_global_tgt_ids));

    // for each fixed value
    for (unsigned j = 0; j < temp_data->num_fixed_values; ++j) {

      if (target_request == NULL) continue;

      unsigned curr_num_fixed_global_tgt_ids = 0;

      for (unsigned i = 0, k = 0, l = 0; i < temp_data->num_fixed_values; ++i) {
        for (unsigned h = 0; h < temp_data->num_tgt_per_fixed_value[i]; ++h, ++k) {

          while ((l < target_request->num_points) &&
                 (target_request->global_ids[l] <
                  temp_data->global_tgt_indices[k])) ++l;

          if ((l < target_request->num_points) &&
              (target_request->global_ids[l] ==
               temp_data->global_tgt_indices[k]))
            temp_fixed_global_tgt_ids[curr_num_fixed_global_tgt_ids++] =
              temp_data->global_tgt_indices[k];
        }
      }

      if (curr_num_fixed_global_tgt_ids == 0) continue;

      // pack current number of point for current fixed value
      yac_comm_pack(&curr_num_fixed_global_tgt_ids, 1, COMM_UINT, send_buffer,
                    send_buffer_size, &pos, comm);

      // pack target indices for current fixed value
      yac_comm_pack(temp_fixed_global_tgt_ids, curr_num_fixed_global_tgt_ids,
                    COMM_INT, send_buffer, send_buffer_size, &pos, comm);

      // pack fixed value
      yac_comm_pack(&(temp_data->fixed_values[j]), 1, COMM_DBLE, send_buffer,
                    send_buffer_size, &pos, comm);
    }

    free(temp_fixed_global_tgt_ids);

    // send header containing size of the send_buffer
    yac_comm_bsend((void const *)&pos, 1, COMM_UINT, tgt_rank,
                   interp_src->tags.interp_data_header, comm);

    // send links to target process
    yac_comm_bsend(send_buffer, pos, COMM_PACKED, tgt_rank,
                   interp_src->tags.interp_data, comm);

    // generate sent list (ignore duplicated global source points)
    struct mapping_data_tgt * mapping_on_tgt =
      &(interp_src->interp_data->mapping_on.tgt);

    mapping_on_tgt->send_msgs =
      realloc(mapping_on_tgt->send_msgs,
              (mapping_on_tgt->num_send_msgs + 1) *
              sizeof(*(mapping_on_tgt->send_msgs)));
    struct field_data_send_msg * curr_send_msg =
      mapping_on_tgt->send_msgs + mapping_on_tgt->num_send_msgs;
    mapping_on_tgt->num_send_msgs++;

    *curr_send_msg =
      generate_send_msg(requested_links, num_requested_links, tgt_rank,
                        yac_get_local_domain(interp_src->search),
                        interp_src->src_locations);
    if (interp_src->interp_data->max_send_buffer_size <
        curr_send_msg->num_points)
      interp_src->interp_data->max_send_buffer_size =
        curr_send_msg->num_points;
  }

  struct temp_send_flag_data * send_flag_data = NULL;
  if (yac_get_comm_rank(halo_comm) == 0) {
    send_flag_data = malloc(1 * sizeof(*send_flag_data));
    send_flag_data->rank =
      yac_get_lowest_remote_domain_rank(interp_src->search);
    send_flag_data->tag = interp_src->tags.interp_data_msg_count;
    send_flag_data->comm = interp_src->interp_data->comm;
  }

  // gather send flags on the root of the source component
  // the source root then sends this to the target root, which in turn scatters
  // the data among the target processes. The target processes use this
  // information to determine when they have received all data
  yac_comm_reduce_callback(send_flags, comm_size, COMM_UINT, COMM_OP_SUM, 0,
                           interp_src->tags.reduce_send_flag,
                           halo_comm, send_flag_data,
                           (send_flag_data == NULL)?NULL:
                             reduce_send_flags_callback, MSG_PRIO_LOW);

  free(send_flags);
  free(send_buffer);
  free(requested_links);
  free(remote_weights);
  free(remote_links);
  for (unsigned i = 0; i < temp_data->num_target_requests; ++i)
    free(temp_data->target_requests[i].global_ids);
  free(temp_data->weights);
  free(temp_data->links);
  free(temp_data->target_requests);
  free(temp_data->global_tgt_indices);
  free(temp_data->num_tgt_per_fixed_value);
  free(temp_data->fixed_values);
  free(temp_data);
}

static int compare_field_data_recv_msgs (const void * a, const void * b) {
  return (((struct field_data_recv_msg*)b)->rank -
          ((struct field_data_recv_msg*)a)->rank);
}

static void num_weights_per_link_allreduce_op(void * invec, void * inoutvec,
                                              unsigned count,
                                              enum yac_comm_data_type type) {

  if (count != 1)
    yac_internal_abort_message(
      "ERROR(num_weights_per_link_allreduce_op): invalid count (should be 1)",
      __FILE__, __LINE__);

  if (type != COMM_UINT)
    yac_internal_abort_message(
      "ERROR(num_weights_per_link_allreduce_op): invalid type should be "
      "COMM_UINT", __FILE__, __LINE__);

  // num_weights_per_link
  if (*((unsigned*)invec) != *((unsigned*)inoutvec)) {
    if ((*((unsigned*)invec) == 0) || (*((unsigned*)inoutvec) == 0))
      *((unsigned*)inoutvec) =
      MAX(*((unsigned*)invec), *((unsigned*)inoutvec));
    else
      yac_internal_abort_message(
        "ERROR(num_weights_per_link_allreduce_op): "
        "num_weights_per_link do not match", __FILE__, __LINE__);
  }
}

/**
 * in case the mapping is to be executed at the target side this routine informs
 * other source processes about halo points required by the targets for the
 * mapping
 */
static void
exchange_halo_requests_target(
  struct temp_point_search_data * search_data,
  struct interpolation_method_data_request * data_request,
  struct temp_interpolation_plan_data * interp_plan_data) {

  // check that all source processes have the same number of weights per link
  // everything else is a but and cause an error
  yac_comm_allreduce_callback(
    &interp_plan_data->num_weights_per_link, 1, COMM_UINT,
    num_weights_per_link_allreduce_op,
    search_data->interp_src->tags.allreduce_num_weights_per_link,
    search_data->interp_src->interp_data->halo_comm, NULL, NULL, MSG_PRIO_LOW);

  // copy link data into an array of "struct link_"
  // this makes working with the links easier
  unsigned num_links = 0;
  for (unsigned i = 0;
       i < search_data->interp_src->interp_data->num_pointsets; ++i)
    num_links += interp_plan_data->num_links_per_pointset[i];

  struct link_ * links = malloc(num_links * sizeof(*links));
  double * weights = malloc(num_links * interp_plan_data->num_weights_per_link *
                            sizeof(*weights));
  memcpy(weights, interp_plan_data->weights,
         num_links * interp_plan_data->num_weights_per_link * sizeof(*weights));

  for (unsigned i = 0; i < num_links; ++i) {
    links[i].src_index = interp_plan_data->src_indices[i];
    links[i].rank = -1;
    links[i].weights = weights + i * interp_plan_data->num_weights_per_link;
  }

  for (unsigned i = 0, k = 0; i < interp_plan_data->num_tgt_indices_links; ++i)
    for (unsigned j = 0; j < interp_plan_data->num_links_per_tgt[i]; ++j, ++k)
      links[k].tgt_index =
        search_data->interp_src->collected_global_ids[i];

  for (unsigned i = 0, k = 0;
       i < search_data->interp_src->interp_data->num_pointsets; ++i)
    for (unsigned j = 0; j < interp_plan_data->num_links_per_pointset[i]; ++j, ++k)
      links[k].pointset_index = i;

  // sort links by first by pointset_index and second by src_index
  qsort(links, num_links, sizeof(*links), compare_links_pointset);

  // use data_request to set links[k].rank for all halo points
  for (unsigned i = 0, j = 0;
       i < search_data->interp_src->interp_data->num_pointsets; ++i) {

    int * global_src_indices = data_request[i].global_ids;
    int * src_ranks = data_request[i].ranks;
    unsigned num_global_src_indices = data_request[i].num_values;

    // sort request points by global indices, which makes lookup easier
    yac_quicksort_index(global_src_indices, num_global_src_indices, src_ranks);

    for (unsigned k = 0; k < num_global_src_indices; ++k) {

      if ((k > 0) && (global_src_indices[k] == global_src_indices[k-1]))
        continue;

      while ((j < num_links) && (links[j].src_index != global_src_indices[k]))
        ++j;

      if ((j == num_links) || (links[j].pointset_index != i))
        yac_internal_abort_message(
          "ERROR(exchange_halo_requests_target): data_request and "
          "interpolation plan link data does not match", __FILE__, __LINE__);

      while ((j < num_links) && (links[j].src_index == global_src_indices[k]))
        links[j++].rank = src_ranks[k];
    }
  }

  // sort links first by source rank then by global target index
  qsort(links, num_links, sizeof(*links), compare_links_src_rank);

  struct communicator * halo_comm =
    search_data->interp_src->interp_data->halo_comm;
  unsigned halo_comm_size = yac_get_comm_size(halo_comm);

  int link_pack_size =
    yac_comm_pack_size(2, COMM_INT, halo_comm); // src_index + tgt_index
  link_pack_size +=
    yac_comm_pack_size(2, COMM_UINT, halo_comm); // pointset_index + src_rank
  link_pack_size +=
    yac_comm_pack_size(
      interp_plan_data->num_weights_per_link, COMM_DBLE, halo_comm); // weights
  int packing_unit_header_size =
    yac_comm_pack_size(1, COMM_UINT, halo_comm);

  unsigned k = 0;
  while ((k < num_links) && (links[k].rank == -1)) ++k;
  unsigned new_num_links = k;

  // pack links of halo points for other target processes
  char * send_buffer = NULL;
  size_t send_buffer_size = 0;
  unsigned * send_size_per_rank =
    calloc(halo_comm_size, sizeof(*send_size_per_rank));
  int offset = 0;

  for (unsigned rank = 0; rank < halo_comm_size; ++rank) {

    while ((k < num_links) && (links[k].rank < rank)) ++k;

    if (k == num_links) break;

    if (links[k].rank == rank) {

      struct link_ * curr_links = links + k;
      unsigned curr_num_links = 0;

      while ((k < num_links) && (links[k].rank == rank)) {
        ++k;
        ++curr_num_links;
      }

      for (unsigned i = 0; i < search_data->num_target_requests; ++i) {

        unsigned num_matches = 0;

        // count the number of links required for current target request
        for (unsigned j = 0, l = 0;
             j < search_data->target_requests[i].num_points; ++j) {

          while ((l < curr_num_links) &&
                 (curr_links[l].tgt_index <
                  search_data->target_requests[i].global_ids[j])) ++l;
          while ((l < curr_num_links) &&
                 (curr_links[l].tgt_index ==
                  search_data->target_requests[i].global_ids[j])) {

            ++l;
            ++num_matches;
          }
        }

        // ensure that pack buffer is big enough
        ENSURE_ARRAY_SIZE(send_buffer, send_buffer_size,
                          offset + packing_unit_header_size +
                          num_matches * link_pack_size);

        int pos = 0;
        int curr_send_buffer_size = (int)send_buffer_size - offset;
        char * curr_send_buffer = send_buffer + offset;

        // pack num_matches
        yac_comm_pack(&num_matches, 1, COMM_UINT, curr_send_buffer,
                      curr_send_buffer_size, &pos, halo_comm);

        // pack links
        for (unsigned j = 0, l = 0;
             j < search_data->target_requests[i].num_points; ++j) {

          while ((l < curr_num_links) &&
                 (curr_links[l].tgt_index <
                  search_data->target_requests[i].global_ids[j])) ++l;
          while ((l < curr_num_links) &&
                 (curr_links[l].tgt_index ==
                  search_data->target_requests[i].global_ids[j])) {

            // pack
            yac_comm_pack(&(curr_links[l].src_index), 1, COMM_INT, curr_send_buffer,
                          curr_send_buffer_size, &pos, halo_comm);
            yac_comm_pack(&(curr_links[l].tgt_index), 1, COMM_INT, curr_send_buffer,
                          curr_send_buffer_size, &pos, halo_comm);
            yac_comm_pack(&(curr_links[l].pointset_index), 1, COMM_UINT,
                          curr_send_buffer, curr_send_buffer_size, &pos, halo_comm);
            yac_comm_pack(&(search_data->target_requests[i].rank), 1, COMM_UINT,
                          curr_send_buffer, curr_send_buffer_size, &pos, halo_comm);
            yac_comm_pack(curr_links[l].weights,
                          interp_plan_data->num_weights_per_link, COMM_DBLE,
                          curr_send_buffer, curr_send_buffer_size, &pos, halo_comm);

            ++l;
          }
        }
        offset += pos;
        send_size_per_rank[rank] += pos;
      }
    }
  }

  struct temp_halo_requests_target_data * temp_data =
    malloc(1 * sizeof(*temp_data));

  temp_data->links = links;
  temp_data->num_links = new_num_links;
  temp_data->num_weights_per_link = interp_plan_data->num_weights_per_link;
  temp_data->weights = weights;
  temp_data->global_tgt_indices =
    malloc(interp_plan_data->num_tgt_indices_fixed *
           sizeof(*(temp_data->global_tgt_indices)));
  memcpy(temp_data->global_tgt_indices,
         interp_plan_data->global_tgt_indices,
         interp_plan_data->num_tgt_indices_fixed *
         sizeof(*(temp_data->global_tgt_indices)));
  temp_data->num_tgt_indices_fixed = interp_plan_data->num_tgt_indices_fixed;
  temp_data->num_tgt_per_fixed_value =
    malloc(interp_plan_data->num_fixed_values *
           sizeof(*(temp_data->num_tgt_per_fixed_value)));
  memcpy(temp_data->num_tgt_per_fixed_value,
         interp_plan_data->num_tgt_per_fixed_value,
         interp_plan_data->num_fixed_values *
         sizeof(*(temp_data->num_tgt_per_fixed_value)));
  temp_data->fixed_values =
    malloc(interp_plan_data->num_fixed_values *
           sizeof(*(temp_data->fixed_values)));
  memcpy(temp_data->fixed_values, interp_plan_data->fixed_values,
         interp_plan_data->num_fixed_values *
         sizeof(*(temp_data->fixed_values)));
  temp_data->num_fixed_values = interp_plan_data->num_fixed_values;
  temp_data->interp_src = search_data->interp_src;
  temp_data->target_requests = search_data->target_requests;
  temp_data->num_target_requests = search_data->num_target_requests;

  search_data->target_requests = NULL;
  search_data->num_target_requests = 0;

  // exchange links containing halo source points
  yac_comm_alltoallv_callback(send_buffer, send_size_per_rank, COMM_PACKED,
                              search_data->interp_src->tags.halo_data_alltoallv,
                              halo_comm, temp_data,
                              alltoallv_halo_data_callback, MSG_PRIO_MIDDLE);
  free(send_buffer);
  free(send_size_per_rank);
}

/**
 * exchanges halo requests with other source processes and set up local halo
 * receive list (mapping_on_source == true)
 */
static void
exchange_halo_requests_source(
  struct field_interpolation_gen_src * interp_src,
  struct interpolation_method_data_request * data_request) {

   unsigned total_num_halo_points = 0;

   unsigned halo_comm_size =
      yac_get_comm_size(interp_src->interp_data->halo_comm);
   unsigned halo_comm_rank =
      yac_get_comm_rank(interp_src->interp_data->halo_comm);

   unsigned * num_points_per_rank =
      calloc(halo_comm_size, sizeof(*num_points_per_rank));

   // get the total number of requested points from all interpolation methods on the
   // interpolation method stack
   for (unsigned i = 0; i < interp_src->interp_data->num_pointsets; ++i) {
      total_num_halo_points += data_request[i].num_values;
      interp_src->interp_data->num_halo_recv_points_per_points_set[i] =
        data_request[i].num_values;
      for (unsigned j = 0; j < data_request[i].num_values; ++j)
        num_points_per_rank[data_request[i].ranks[j]]++;
   }

   // check for data requests from the local process (not allowed)
   if (num_points_per_rank[halo_comm_rank] != 0)
      yac_internal_abort_message(
         "internal error (data request from local process)", __FILE__,__LINE__);

   unsigned num_halo_recv_msgs = 0;
   unsigned * rank_to_halo_msg =
      malloc(halo_comm_size * sizeof(*rank_to_halo_msg));
   for (unsigned i = 0; i < halo_comm_size; ++i) {
     rank_to_halo_msg[i] = num_halo_recv_msgs;
     if (num_points_per_rank[i] > 0) num_halo_recv_msgs++;
   }

   struct field_data_recv_msg * halo_recv_msgs =
      malloc(num_halo_recv_msgs *
             sizeof(*(interp_src->interp_data->halo_recv_msgs)));
   interp_src->interp_data->halo_recv_msgs = halo_recv_msgs;
   interp_src->interp_data->num_halo_recv_msgs = num_halo_recv_msgs;
   interp_src->interp_data->total_num_halo_recv_points = total_num_halo_points;

   num_halo_recv_msgs = 0;
   for (unsigned i = 0; i < halo_comm_size; ++i) {
     if (num_points_per_rank[i] > 0) {
       halo_recv_msgs[num_halo_recv_msgs].recv_points =
          malloc(num_points_per_rank[i] *
                 sizeof(*(halo_recv_msgs[num_halo_recv_msgs].recv_points)));
       halo_recv_msgs[num_halo_recv_msgs].num_recv_points = 0;
       halo_recv_msgs[num_halo_recv_msgs].rank = i;
       num_halo_recv_msgs++;
     }
   }

   int * send_buffer = malloc((total_num_halo_points + halo_comm_size *
                              interp_src->interp_data->num_pointsets) *
                              sizeof(*send_buffer));
   int ** send_buffer_num_points_per_point_set_per_rank =
      malloc(halo_comm_size *
             sizeof(*send_buffer_num_points_per_point_set_per_rank));
   int ** send_buffer_global_ids_per_rank =
      malloc(halo_comm_size * sizeof(*send_buffer_global_ids_per_rank));
   for (unsigned i = 0, offset = 0; i < halo_comm_size; ++i) {
      send_buffer_num_points_per_point_set_per_rank[i] = send_buffer + offset;
      send_buffer_global_ids_per_rank[i] =
         send_buffer + offset + interp_src->interp_data->num_pointsets;
      offset += num_points_per_rank[i] + interp_src->interp_data->num_pointsets;
      for (unsigned j = 0; j < interp_src->interp_data->num_pointsets; ++j)
         send_buffer_num_points_per_point_set_per_rank[i][j] = 0;
   }

   // extract data from data_requests
   // -> generate halo_recv_msgs
   // -> generate message for other source processes that contains information
   //    about halo points required by the local process
   for (unsigned i = 0; i < interp_src->interp_data->num_pointsets; ++i) {
      for (unsigned j = 0; j < data_request[i].num_values; ++j) {

         unsigned rank = data_request[i].ranks[j];
         struct field_data_recv_msg * curr_halo_recv_msg =
            halo_recv_msgs + rank_to_halo_msg[rank];
         unsigned recv_point_idx = curr_halo_recv_msg->num_recv_points++;
         curr_halo_recv_msg->recv_points[recv_point_idx].position = j;
         curr_halo_recv_msg->recv_points[recv_point_idx].pointset_idx = i;

         send_buffer_num_points_per_point_set_per_rank[rank][i]++;
         *(send_buffer_global_ids_per_rank[rank]) =
         data_request[i].global_ids[j];
         send_buffer_global_ids_per_rank[rank]++;
      }
   }

   free(send_buffer_global_ids_per_rank);
   free(send_buffer_num_points_per_point_set_per_rank);
   free(rank_to_halo_msg);

   struct temp_halo_search_data * halo_search_data =
      malloc(1 * sizeof(*halo_search_data));
   halo_search_data->src_locations = interp_src->src_locations;
   halo_search_data->local_domain = yac_get_local_domain(interp_src->search);
   halo_search_data->interp_src = interp_src;

   unsigned * send_count_per_proc = num_points_per_rank;
   for (unsigned i = 0; i < halo_comm_size; ++i)
      send_count_per_proc[i] += interp_src->interp_data->num_pointsets;

   // exchange the required halo points
   yac_comm_alltoallv_callback(send_buffer, send_count_per_proc, COMM_INT,
                               interp_src->tags.halo_data_alltoallv,
                               interp_src->interp_data->halo_comm,
                               halo_search_data,
                               alltoallv_interpolation_halo_callback,
                               MSG_PRIO_MIDDLE);

   free(num_points_per_rank);
   free(send_buffer);

   // reorder halo receive messages
   for (unsigned i = 0; i < num_halo_recv_msgs; ++i)
     if (halo_recv_msgs[i].rank <= halo_comm_rank)
        halo_recv_msgs[i].rank += halo_comm_size;
   // the receive order is invers to the send order
   qsort(halo_recv_msgs, num_halo_recv_msgs, sizeof(*halo_recv_msgs),
         compare_field_data_recv_msgs);
   for (unsigned i = 0; i < num_halo_recv_msgs; ++i)
     if (halo_recv_msgs[i].rank >= halo_comm_size)
        halo_recv_msgs[i].rank -= halo_comm_size;
}

static void
cleanup_search_data(struct temp_point_search_data * search_data) {

   // free search_data
   for (unsigned i = 0; i < search_data->num_target_requests; ++i)
      free(search_data->target_requests[i].global_ids);
   free(search_data->target_requests);
   free(search_data->collected_coordinates_xyz);
   free(search_data);
}

#ifdef HAVE_C_NETCDF
static void handle_error(int status) {
  const char *err_string;
  err_string = nc_strerror(status);
  printf ("%s \n", err_string);
  if (status != NC_NOERR) {
    exit(EXIT_FAILURE);
  }
}

static void
write_fixed_to_file(struct temp_fixed_io_proc_data * temp_data) {

  if (temp_data->num_fixed_values == 0) {

    free(temp_data->global_tgt_indices_per_fixed_value);
    free(temp_data->num_tgt_per_io_proc);
    free(temp_data);
    return;
  }

  int ncid, status;

  // open file

  status = nc_open(temp_data->weight_file_name, NC_WRITE | NC_SHARE, &ncid);
  if (status != NC_NOERR) handle_error(status);

  // inquire variable ids

  int var_dst_add_fixed_id;
  status = nc_inq_varid(ncid, "dst_address_fixed", &var_dst_add_fixed_id);
  if (status != NC_NOERR) handle_error(status);

  // write fixed data to file

  unsigned file_tgt_offset = 0;
  unsigned * num_tgt_per_fixed_value = temp_data->num_tgt_per_io_proc +
                                       temp_data->num_fixed_values *
                                       temp_data->io_proc_idx;

  for (unsigned i = 0; i < temp_data->num_fixed_values; ++i) {

    for (unsigned j = 0; j < temp_data->io_proc_idx; ++j)
      file_tgt_offset +=
        temp_data->num_tgt_per_io_proc[j * temp_data->num_fixed_values + i];

    for (unsigned j = 0; j < num_tgt_per_fixed_value[i]; ++j)
      temp_data->global_tgt_indices_per_fixed_value[i][j]++;

    size_t start[1] = {file_tgt_offset};
    size_t count[1] = {num_tgt_per_fixed_value[i]};
    status = nc_put_vara_int(ncid, var_dst_add_fixed_id, start, count,
                             temp_data->global_tgt_indices_per_fixed_value[i]);
    if (status != NC_NOERR) handle_error(status);
    free(temp_data->global_tgt_indices_per_fixed_value[i]);

    for (unsigned j = temp_data->io_proc_idx; j < temp_data->num_io_procs; ++j)
      file_tgt_offset +=
        temp_data->num_tgt_per_io_proc[j * temp_data->num_fixed_values + i];
  }

  // clean up

  free(temp_data->global_tgt_indices_per_fixed_value);
  free(temp_data->num_tgt_per_io_proc);
  free(temp_data);
}

/**
 * write link data to file
 */
static void
write_links_to_file(struct temp_link_io_proc_data * temp_data) {

  if (temp_data->num_links == 0) {

    free(temp_data->links);
    free(temp_data->weights);
    free(temp_data->num_links_per_io_proc);
    free(temp_data);
    return;
  }

  int ncid, status;

  // open file

  status = nc_open(temp_data->weight_file_name, NC_WRITE | NC_SHARE, &ncid);
  if (status != NC_NOERR) handle_error(status);

  // inquire variable ids

  int var_src_add_id, var_dst_add_id, var_weight_id;
  status = nc_inq_varid(ncid, "src_address", &var_src_add_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_inq_varid(ncid, "dst_address", &var_dst_add_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_inq_varid(ncid, "remap_matrix", &var_weight_id);
  if (status != NC_NOERR) handle_error(status);

  // write mapping data

  unsigned num_links = temp_data->num_links;
  unsigned num_weights_per_link = temp_data->num_weights_per_link;
  struct link * links = temp_data->links;
  int * src_address = malloc(num_links * sizeof(*src_address));
  int * tgt_address = malloc(num_links * sizeof(*tgt_address));
  double * weights =
    malloc(num_links * num_weights_per_link * sizeof(*weights));
  unsigned * num_links_per_pointset = temp_data->num_links_per_io_proc +
                                      temp_data->num_pointsets *
                                      temp_data->io_proc_idx;

  for (unsigned i = 0; i < num_links; i++) {
    src_address[i] = links[i].src_index + 1;
    tgt_address[i] = links[i].tgt_index + 1;
    memcpy(weights + i * num_weights_per_link, links[i].weights,
           num_weights_per_link * sizeof(*weights));
  }

  unsigned file_link_offset = 0, link_offset = 0;

  for (unsigned i = 0; i < temp_data->num_pointsets; ++i) {

    for (unsigned j = 0; j < temp_data->io_proc_idx; ++j)
      file_link_offset +=
        temp_data->num_links_per_io_proc[j * temp_data->num_pointsets + i];

    if (num_links_per_pointset[i] == 0) continue;

    size_t start[2] = {file_link_offset, 0};
    size_t count[2] = {num_links_per_pointset[i], num_weights_per_link};
    status = nc_put_vara_int(ncid, var_src_add_id, start, count,
                             src_address + link_offset);
    if (status != NC_NOERR) handle_error(status);
    status = nc_put_vara_int(ncid, var_dst_add_id, start, count,
                             tgt_address + link_offset);
    if (status != NC_NOERR) handle_error(status);
    status = nc_put_vara_double(ncid, var_weight_id, start, count,
                                weights + link_offset);
    if (status != NC_NOERR) handle_error(status);

    link_offset += num_links_per_pointset[i];

    for (unsigned j = temp_data->io_proc_idx; j < temp_data->num_io_procs; ++j)
      file_link_offset +=
        temp_data->num_links_per_io_proc[j * temp_data->num_pointsets + i];
  }

  // close file
  status = nc_close(ncid);
  if (status != NC_NOERR) handle_error(status);

  free(weights);
  free(tgt_address);
  free(src_address);
  free(temp_data->links);
  free(temp_data->weights);
  free(temp_data->num_links_per_io_proc);
  free(temp_data);
}
#else
static void
write_fixed_to_file(struct temp_fixed_io_proc_data * temp_data) {

  UNUSED(temp_data);
  yac_internal_abort_message("ERROR(write_fixed_to_file): NETCDF not available",
                             __FILE__, __LINE__);
}

static void
write_links_to_file(struct temp_link_io_proc_data * temp_data) {

  UNUSED(temp_data);
  yac_internal_abort_message("ERROR(write_links_to_file): NETCDF not available",
                             __FILE__, __LINE__);
}
#endif // HAVE_C_NETCDF

#ifdef HAVE_C_NETCDF
/**
 * create a weight file with basic meta data
 */
static void
create_weight_file(struct temp_weight_file_data * weight_file_data) {

  int ncid, status;
  unsigned num_fixed_values =
    weight_file_data->fixed.collected_num_fixed_values;
  unsigned num_links = weight_file_data->link.total_num_links;

  // create file
  status = nc_create(weight_file_data->weight_file_name, NC_CLOBBER, &ncid);
  if (status != NC_NOERR) handle_error(status);

  int dim_weight_id[6];

  // define dimensions
  if (num_links > 0) {
    status = nc_def_dim(ncid, "num_links", num_links, &dim_weight_id[0]);
    if (status != NC_NOERR) handle_error(status);
    status = nc_def_dim(ncid, "num_wgts",
                        weight_file_data->link.num_weights_per_link,
                        &dim_weight_id[1]);
    if (status != NC_NOERR) handle_error(status);
  }
  status = nc_def_dim(ncid, "num_src_fields",
                      weight_file_data->link.num_pointsets, &dim_weight_id[2]);
  if (status != NC_NOERR) handle_error(status);
  status = nc_def_dim(ncid, "max_loc_str_len", YAC_MAX_LOC_STR_LEN, &dim_weight_id[3]);
  if (status != NC_NOERR) handle_error(status);

  if (num_fixed_values > 0) {
    status = nc_def_dim(ncid, "num_fixed_values", num_fixed_values,
                        &dim_weight_id[4]);
    if (status != NC_NOERR) handle_error(status);
    unsigned num_fixed_dst = 0;
    for (unsigned i = 0; i < num_fixed_values; ++i)
      num_fixed_dst += weight_file_data->fixed.collected_fixed_values[i].count;
    status = nc_def_dim(ncid, "num_fixed_dst", num_fixed_dst, &dim_weight_id[5]);
    if (status != NC_NOERR) handle_error(status);
  }

  int var_src_add_id, var_dst_add_id, var_weight_id, var_num_links_id,
      src_var_locs_id, tgt_var_loc_id, var_fixed_values_id,
      var_num_dst_per_fixed_value_id, var_dst_add_fixed_id;

  // define variables
  if (num_links > 0) {
    status = nc_def_var(ncid, "src_address", NC_INT, 1, dim_weight_id, &var_src_add_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_def_var(ncid, "dst_address", NC_INT, 1, dim_weight_id, &var_dst_add_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_def_var(ncid, "remap_matrix", NC_DOUBLE, 2, dim_weight_id, &var_weight_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_def_var(ncid, "num_links_per_src_field", NC_INT, 1,
                        &dim_weight_id[2], &var_num_links_id);
    if (status != NC_NOERR) handle_error(status);
  }
  status = nc_def_var(ncid, "src_locations", NC_CHAR, 2, &dim_weight_id[2],
                      &src_var_locs_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_def_var(ncid, "dst_location", NC_CHAR, 1, &dim_weight_id[3],
                      &tgt_var_loc_id);
  if (status != NC_NOERR) handle_error(status);
  if (num_fixed_values > 0) {
    status = nc_def_var(ncid, "fixed_values", NC_DOUBLE, 1, &dim_weight_id[4],
                        &var_fixed_values_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_def_var(ncid, "num_dst_per_fixed_value", NC_INT, 1, &dim_weight_id[4],
                        &var_num_dst_per_fixed_value_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_def_var(ncid, "dst_address_fixed", NC_INT, 1, &dim_weight_id[5],
                        &var_dst_add_fixed_id);
    if (status != NC_NOERR) handle_error(status);
  }

  // put attributes
  status = nc_put_att_text(ncid, NC_GLOBAL, "version",
                           strlen(YAC_WEIGHT_FILE_VERSION_STRING),
                           YAC_WEIGHT_FILE_VERSION_STRING);
  if (status != NC_NOERR) handle_error(status);
  status = nc_put_att_text(ncid, NC_GLOBAL, "src_grid_name",
                           strlen(weight_file_data->src_grid_name),
                           weight_file_data->src_grid_name);
  if (status != NC_NOERR) handle_error(status);
  status = nc_put_att_text(ncid, NC_GLOBAL, "dst_grid_name",
                           strlen(weight_file_data->tgt_grid_name),
                           weight_file_data->tgt_grid_name);
  if (status != NC_NOERR) handle_error(status);
  {
    char const * str_logical[2] = {"FALSE", "TRUE"};
    status = nc_put_att_text(ncid, NC_GLOBAL, "contains_links",
                             strlen(str_logical[num_links > 0]),
                             str_logical[num_links > 0]);
    if (status != NC_NOERR) handle_error(status);
    status = nc_put_att_text(ncid, NC_GLOBAL, "contains_fixed_dst",
                             strlen(str_logical[num_fixed_values > 0]),
                             str_logical[num_fixed_values > 0]);
    if (status != NC_NOERR) handle_error(status);
  }

  // end definition
  status = nc_enddef(ncid);
  if (status != NC_NOERR) handle_error(status);

  // write some basic data

  if (num_links > 0) {
    status =
       nc_put_var_int(ncid, var_num_links_id,
                      (int*)(weight_file_data->link.num_links_per_pointset));
    if (status != NC_NOERR) handle_error(status);
  }

  for (unsigned i = 0; i < weight_file_data->link.num_pointsets; ++i) {
    char const * loc_str = yac_loc2str(weight_file_data->link.src_locations[i]);
    size_t str_start[2] = {i, 0};
    size_t str_count[2] = {1, strlen(loc_str)};
    status = nc_put_vara_text(ncid, src_var_locs_id, str_start, str_count,
                              loc_str);
    if (status != NC_NOERR) handle_error(status);
  }

  {
    char const * loc_str = yac_loc2str(weight_file_data->link.tgt_location);
    size_t str_start[1] = {0};
    size_t str_count[1] = {strlen(loc_str)};
    status = nc_put_vara_text(ncid, tgt_var_loc_id, str_start, str_count,
                              loc_str);
  }
  if (num_fixed_values > 0) {

    double * fixed_values = malloc(num_fixed_values * sizeof(*fixed_values));
    int * num_dst_per_fixed_value =
      malloc(num_fixed_values * sizeof(*num_dst_per_fixed_value));
    for (unsigned i = 0; i < num_fixed_values; ++i) {
      fixed_values[i] = weight_file_data->fixed.collected_fixed_values[i].value;
      num_dst_per_fixed_value[i] = weight_file_data->fixed.collected_fixed_values[i].count;
    }
    status = nc_put_var_double(ncid, var_fixed_values_id, fixed_values);
    if (status != NC_NOERR) handle_error(status);
    status = nc_put_var_int(ncid, var_num_dst_per_fixed_value_id,
                            num_dst_per_fixed_value);
    if (status != NC_NOERR) handle_error(status);
    free(num_dst_per_fixed_value);
    free(fixed_values);
  }

  // close file
  status = nc_close(ncid);
  if (status != NC_NOERR) handle_error(status);
}
#else

static void
create_weight_file(struct temp_weight_file_data * weight_file_data) {

  UNUSED(weight_file_data);

  yac_internal_abort_message("ERROR(create_weight_file): NETCDF not available",
                             __FILE__, __LINE__);
}

#endif // HAVE_C_NETCDF

/**
 * exchanges required halo data between source processes in the put operation
 */
static void
exchange_halo_data(struct field_interpolation_data_src * interp_data,
                   unsigned halo_send_tag, unsigned halo_recv_tag,
                   void **** restrict extended_data, int collection_size,
                   enum yac_comm_data_type data_type) {

   if ((data_type != COMM_DBLE) && (data_type != COMM_INT))
      yac_internal_abort_message("ERROR(exchange_halo_data): invalid data type",
                                 __FILE__, __LINE__);

   unsigned num_requests =
      interp_data->num_halo_send_msgs + interp_data->num_halo_recv_msgs;
   struct comm_request ** requests = malloc(num_requests * sizeof(*requests));

   // recv halo data
   struct field_data_recv_msg *
      curr_halo_recv_msg = interp_data->halo_recv_msgs;
   void * recv_buffer = NULL;
   if (data_type == COMM_DBLE) {
      double * restrict recv_buffer_dble;
      recv_buffer = malloc(interp_data->total_num_halo_recv_points *
                           collection_size * sizeof(*recv_buffer_dble));
      recv_buffer_dble = (double *)recv_buffer;
      for (unsigned i = 0; i < interp_data->num_halo_recv_msgs;
           ++i, ++curr_halo_recv_msg) {

         yac_comm_irecv(recv_buffer_dble, curr_halo_recv_msg->num_recv_points *
                        collection_size, COMM_DBLE, curr_halo_recv_msg->rank,
                        halo_recv_tag, interp_data->halo_comm, &(requests[i]));
         recv_buffer_dble +=
            curr_halo_recv_msg->num_recv_points * collection_size;
      }
   } else {
      int * restrict recv_buffer_int;
      recv_buffer = malloc(interp_data->total_num_halo_recv_points *
                           collection_size * sizeof(*recv_buffer_int));
      recv_buffer_int = (int *)recv_buffer;
      for (unsigned i = 0; i < interp_data->num_halo_recv_msgs;
           ++i, ++curr_halo_recv_msg) {

         yac_comm_irecv(recv_buffer_int, curr_halo_recv_msg->num_recv_points *
                        collection_size, COMM_INT, curr_halo_recv_msg->rank,
                        halo_recv_tag, interp_data->halo_comm, &(requests[i]));
         recv_buffer_int +=
            curr_halo_recv_msg->num_recv_points * collection_size;
      }
   }

   // pack and send halo data
   struct field_data_send_msg *
      curr_halo_send_msg = interp_data->halo_send_msgs;
   char * send_buffer = NULL;
   if (data_type == COMM_DBLE) {
      double * restrict send_buffer_dble =
         malloc(interp_data->total_num_halo_send_points * collection_size *
                sizeof(*send_buffer_dble));
      send_buffer = (void*)send_buffer_dble;
      for (unsigned i = 0; i < interp_data->num_halo_send_msgs;
          ++i, ++curr_halo_send_msg) {

         for (int h = 0; h < collection_size; ++h) {
            double * curr_send_buffer_dble =
               send_buffer_dble + h * curr_halo_send_msg->num_points;
            double *** curr_extended_data = (double***)(extended_data[h]);
            for (unsigned j = 0; j < curr_halo_send_msg->num_points; ++j)
               curr_send_buffer_dble[j] =
                 curr_extended_data
                   [1 + curr_halo_send_msg->send_points[j].subdomain_idx]
                   [curr_halo_send_msg->send_points[j].pointset_idx]
                   [curr_halo_send_msg->send_points[j].local_id];
         }

         yac_comm_isend(send_buffer_dble, curr_halo_send_msg->num_points *
                        collection_size, COMM_DBLE, curr_halo_send_msg->rank,
                        halo_send_tag, interp_data->halo_comm,
                        &(requests[i+interp_data->num_halo_recv_msgs]));
         send_buffer_dble += curr_halo_send_msg->num_points * collection_size;
      }
   } else {
      int * restrict send_buffer_int =
         malloc(interp_data->total_num_halo_send_points * collection_size *
                sizeof(*send_buffer_int));
      send_buffer = (void*)send_buffer_int;
      for(unsigned i = 0; i < interp_data->num_halo_send_msgs;
          ++i, ++curr_halo_send_msg) {

         for (int h = 0; h < collection_size; ++h) {
            int * curr_send_buffer_int =
               send_buffer_int + h * curr_halo_send_msg->num_points;
            int *** curr_extended_data = (int***)(extended_data[h]);
            for (unsigned j = 0; j < curr_halo_send_msg->num_points; ++j)
               curr_send_buffer_int[j] =
                 curr_extended_data
                   [1 + curr_halo_send_msg->send_points[j].subdomain_idx]
                   [curr_halo_send_msg->send_points[j].pointset_idx]
                   [curr_halo_send_msg->send_points[j].local_id];
         }

         yac_comm_isend(send_buffer_int, curr_halo_send_msg->num_points *
                        collection_size, COMM_INT, curr_halo_send_msg->rank,
                        halo_send_tag, interp_data->halo_comm,
                        &(requests[i+interp_data->num_halo_recv_msgs]));
         send_buffer_int += curr_halo_send_msg->num_points * collection_size;
      }
   }

   yac_comm_waitall(num_requests, requests);

   // unpack halo data
   curr_halo_recv_msg = interp_data->halo_recv_msgs;
   if (data_type == COMM_DBLE) {
      double * restrict recv_buffer_dble = (double*)recv_buffer;
      for (unsigned i = 0; i < interp_data->num_halo_recv_msgs;
           ++i, ++curr_halo_recv_msg) {
         for (int h = 0; h < collection_size; ++h) {
            double * curr_recv_buffer_dble =
               recv_buffer_dble + h * curr_halo_recv_msg->num_recv_points;
            double ** curr_extended_data = (double**)(extended_data[h][0]);
            for (unsigned j = 0; j < curr_halo_recv_msg->num_recv_points; ++j)
               curr_extended_data
                  [curr_halo_recv_msg->recv_points[j].pointset_idx]
                  [curr_halo_recv_msg->recv_points[j].position] =
                     curr_recv_buffer_dble[j];
         }
         recv_buffer_dble +=
            curr_halo_recv_msg->num_recv_points * collection_size;
      }
   } else {
      int * restrict recv_buffer_int = (int*)recv_buffer;
      for (unsigned i = 0; i < interp_data->num_halo_recv_msgs;
           ++i, ++curr_halo_recv_msg) {
         for (int h = 0; h < collection_size; ++h) {
            int * curr_recv_buffer_int =
               recv_buffer_int + h * curr_halo_recv_msg->num_recv_points;
            int ** curr_extended_data = (int**)(extended_data[h][0]);
            for (unsigned j = 0; j < curr_halo_recv_msg->num_recv_points; ++j)
               curr_extended_data
                  [curr_halo_recv_msg->recv_points[j].pointset_idx]
                  [curr_halo_recv_msg->recv_points[j].position] =
                     curr_recv_buffer_int[j];
         }
         recv_buffer_int +=
            curr_halo_recv_msg->num_recv_points * collection_size;
      }
   }

   free(send_buffer);
   free(recv_buffer);
   free(requests);
}

static void allreduce_num_links_callback(struct communicator * comm,
                                         struct comm_message msg,
                                         void * user_data) {

  UNUSED(comm);

  struct temp_link_io_proc_data * temp_data =
    (struct temp_link_io_proc_data *)user_data;

  unsigned buffer_size = temp_data->num_io_procs * temp_data->num_pointsets *
       sizeof(unsigned);
  unsigned * num_links_per_io_proc = malloc(buffer_size);
  memcpy(num_links_per_io_proc, msg.buffer, buffer_size);
  temp_data->num_links_per_io_proc = num_links_per_io_proc;

  // if the links have already been received
  if (temp_data->links_have_been_received)
    write_links_to_file(temp_data);
}

static void allreduce_num_fixed_callback(struct communicator * comm,
                                         struct comm_message msg,
                                         void * user_data) {

  UNUSED(comm);

  struct temp_fixed_io_proc_data * temp_data =
    (struct temp_fixed_io_proc_data *)user_data;

  unsigned buffer_size = msg.count * sizeof(*temp_data->num_tgt_per_io_proc);
  unsigned * num_tgt_per_io_proc = malloc(buffer_size);
  memcpy(num_tgt_per_io_proc, msg.buffer, buffer_size);

  temp_data->num_tgt_per_io_proc = num_tgt_per_io_proc;

  // if the links have already been received
  if (temp_data->global_tgt_indices_per_fixed_value != NULL)
    write_fixed_to_file(temp_data);
}

static int compare_links (const void * a, const void * b) {
  struct link * link_a = (struct link*)a;
  struct link * link_b = (struct link*)b;

  if (link_a->pointset_index != link_b->pointset_index)
    return ( link_a->pointset_index - link_b->pointset_index );
  if (link_a->tgt_index != link_b->tgt_index)
    return ( link_a->tgt_index - link_b->tgt_index );
  else
    return ( link_a->src_index - link_b->src_index );
}

static void alltoallv_link_data_callback(struct communicator * comm,
                                         struct comm_message msg,
                                         void * user_data) {

  struct temp_link_io_proc_data * temp_data =
    (struct temp_link_io_proc_data *)user_data;

  unsigned num_weights_per_link = temp_data->num_weights_per_link;
  unsigned buffer_size_per_link = yac_comm_pack_size(2, COMM_INT, comm);
  buffer_size_per_link += yac_comm_pack_size(1, COMM_UINT, comm);
  buffer_size_per_link +=
    yac_comm_pack_size(num_weights_per_link, COMM_DBLE, comm);
  unsigned max_num_links = msg.count / buffer_size_per_link;

  struct link * links = malloc(max_num_links * sizeof(*links));
  double * weights =
    malloc(max_num_links * num_weights_per_link * sizeof(*weights));

  unsigned num_links = 0;
  unsigned offset = 0;
  unsigned i = 0;

  while (offset < msg.count) {

    int pos = 0;
    unsigned curr_num_links;
    char * curr_packing_unit = (char*)msg.buffer + offset;
    int max_packing_unit_size = (int)(msg.count - offset);

    yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &pos,
                    &curr_num_links, 1, COMM_UINT, comm);

    for (unsigned j = 0; j < curr_num_links; ++i, ++j) {

      yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &pos,
                      &(links[i].src_index), 1, COMM_INT, comm);
      yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &pos,
                      &(links[i].tgt_index), 1, COMM_INT, comm);
      yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &pos,
                      &(links[i].pointset_index), 1, COMM_UINT, comm);
      links[i].weights = weights + i * num_weights_per_link;
      yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &pos,
                      links[i].weights, num_weights_per_link, COMM_DBLE, comm);
    }

    num_links += curr_num_links;
    offset += (unsigned)pos;
  }

  weights = realloc(weights, num_links * num_weights_per_link *
                                        sizeof(*(temp_data->weights)));
  for (unsigned j = 0; j < num_links; ++j)
    links[j].weights = weights + j * num_weights_per_link;

  qsort(links, num_links, sizeof(*links), compare_links);

  // remark: the assignment of the following points needs to be done after the
  //         unpacking is completed, otherwise a race condition with
  //         allreduce_num_links_callback might occure
  temp_data->num_links = num_links;
  temp_data->links = realloc(links, num_links * sizeof(*(temp_data->links)));
  temp_data->weights = weights;
  temp_data->links_have_been_received = 1;

  // if the number of links per io proc is known
  if (temp_data->num_links_per_io_proc != NULL)
    write_links_to_file(temp_data);
}

static void alltoallv_fixed_data_callback(struct communicator * comm,
                                         struct comm_message msg,
                                         void * user_data) {

  struct temp_fixed_io_proc_data * temp_data =
    (struct temp_fixed_io_proc_data *)user_data;

  unsigned comm_size = yac_get_comm_size(comm);

  unsigned num_fixed_values = temp_data->num_fixed_values;

  int ** global_tgt_indices_per_fixed_value =
    calloc(num_fixed_values, sizeof(*global_tgt_indices_per_fixed_value));
  unsigned * array_sizes = calloc(num_fixed_values, sizeof(*array_sizes));
  unsigned * num_tgt_per_fixed_value =
    calloc(num_fixed_values, sizeof(*num_tgt_per_fixed_value));
  unsigned total_num_global_tgt_indices = 0;

  char * curr_packing_unit = (char*)msg.buffer;
  unsigned max_packing_unit_size = msg.count;

  for (unsigned i = 0; i < comm_size; ++i) {

    int position = 0;
    for (unsigned j = 0; j < num_fixed_values; ++j) {

      unsigned count;
      yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &position,
                      &count, 1, COMM_UINT, comm);

      if (count == 0) continue;

      ENSURE_ARRAY_SIZE(global_tgt_indices_per_fixed_value[j],
                        array_sizes[j], num_tgt_per_fixed_value[j] + count);
      yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &position,
                      global_tgt_indices_per_fixed_value[j] +
                      num_tgt_per_fixed_value[j], count,
                      COMM_INT, comm);
      total_num_global_tgt_indices += count;
      num_tgt_per_fixed_value[j] += count;
    }
    curr_packing_unit += position;
    max_packing_unit_size -= position;
  }

  free(array_sizes);

  if (max_packing_unit_size != 0)
    yac_internal_abort_message("ERROR(alltoallv_fixed_data_callback): unpacking"
                               " failed\n", __FILE__, __LINE__);

  for (unsigned i = 0; i < num_fixed_values; ++i)
    qsort(global_tgt_indices_per_fixed_value[i], num_tgt_per_fixed_value[i],
          sizeof(**global_tgt_indices_per_fixed_value), compare_int);

  free(num_tgt_per_fixed_value);

  // remark: the assignment of the following array needs to be done after the
  //         unpacking is completed, otherwise a race condition with
  //         allreduce_num_fixed_callback might occure
  temp_data->global_tgt_indices_per_fixed_value =
    global_tgt_indices_per_fixed_value;

  // if the number of links per io proc is known
  if (temp_data->num_tgt_per_io_proc != NULL)
    write_fixed_to_file(temp_data);
}

static void
redistribute_link_data(struct temp_weight_file_data * weight_file_data,
                       struct communicator * comm) {

  unsigned comm_size = yac_get_comm_size(comm);
  unsigned comm_rank = yac_get_comm_rank(comm);

  // redistribute link data among the io procs

  unsigned io_proc_step = weight_file_data->io_proc_step;
  unsigned num_io_procs = (comm_size + io_proc_step - 1) / io_proc_step;
  unsigned is_io_proc = !(comm_rank % io_proc_step);

  unsigned num_link_step =
    (weight_file_data->link.max_tgt_index -
     weight_file_data->link.min_tgt_index) / num_io_procs + 1;

  unsigned * num_links_per_io_proc =
    calloc(num_io_procs * weight_file_data->link.num_pointsets,
           sizeof(*num_links_per_io_proc));
  unsigned * send_buffer_size_per_proc =
    calloc(comm_size, sizeof(*send_buffer_size_per_proc));
  int * positions = calloc(num_io_procs, sizeof(*positions));
  int * offsets = malloc(num_io_procs * sizeof(*positions));

  unsigned buffer_size_per_link = yac_comm_pack_size(2, COMM_INT, comm);
  buffer_size_per_link += yac_comm_pack_size(1, COMM_UINT, comm);
  buffer_size_per_link +=
    yac_comm_pack_size(
      weight_file_data->link.num_weights_per_link, COMM_DBLE, comm);
  unsigned packing_unit_header_size = yac_comm_pack_size(1, COMM_UINT, comm);

  for (unsigned i = 0; i < weight_file_data->link.num_links; ++i)
    ++num_links_per_io_proc[
      weight_file_data->link.num_pointsets *
      ((weight_file_data->link.tgt_indices[i] -
        weight_file_data->link.min_tgt_index)/num_link_step) +
      weight_file_data->link.pointset_indices[i]];
  for (unsigned i = 0; i < num_io_procs; ++i) {
    unsigned curr_send_buffer_size_per_proc = 0;
    for (unsigned j = 0; j < weight_file_data->link.num_pointsets; ++j)
      curr_send_buffer_size_per_proc +=
        num_links_per_io_proc[i * weight_file_data->link.num_pointsets + j];
    curr_send_buffer_size_per_proc *= buffer_size_per_link;
    curr_send_buffer_size_per_proc += packing_unit_header_size;

    send_buffer_size_per_proc[i * io_proc_step] =
      curr_send_buffer_size_per_proc;
  }

  int accum = 0;
  for (unsigned i = 0; i < num_io_procs; ++i) {
    offsets[i] = accum;
    accum += send_buffer_size_per_proc[i * io_proc_step];
  }

  char * send_buffer = malloc((unsigned)accum);

  for (unsigned i = 0; i < num_io_procs; ++i) {

    char * curr_send_buffer = send_buffer + offsets[i];
    unsigned curr_send_buffer_size =
      send_buffer_size_per_proc[i * io_proc_step];

    unsigned curr_num_links = 0;
    for (unsigned j = 0; j < weight_file_data->link.num_pointsets; ++j)
      curr_num_links +=
        num_links_per_io_proc[i * weight_file_data->link.num_pointsets + j];

    yac_comm_pack(&curr_num_links, 1, COMM_UINT, curr_send_buffer,
                  curr_send_buffer_size, positions + i, comm);
  }

  for(unsigned i = 0; i < weight_file_data->link.num_links; ++i) {

    unsigned io_proc_idx =
      (weight_file_data->link.tgt_indices[i] -
       weight_file_data->link.min_tgt_index) / num_link_step;

    char * curr_send_buffer = send_buffer + offsets[io_proc_idx];
    unsigned curr_send_buffer_size =
      send_buffer_size_per_proc[io_proc_idx * io_proc_step];

    yac_comm_pack(weight_file_data->link.src_indices + i, 1, COMM_INT,
                  curr_send_buffer, curr_send_buffer_size,
                  positions + io_proc_idx, comm);
    yac_comm_pack(weight_file_data->link.tgt_indices + i, 1, COMM_INT,
                  curr_send_buffer, curr_send_buffer_size,
                  positions + io_proc_idx, comm);
    yac_comm_pack(weight_file_data->link.pointset_indices + i, 1, COMM_UINT,
                  curr_send_buffer, curr_send_buffer_size,
                  positions + io_proc_idx, comm);
    yac_comm_pack(weight_file_data->link.weights +
                  i * weight_file_data->link.num_weights_per_link,
                  weight_file_data->link.num_weights_per_link, COMM_DBLE,
                  curr_send_buffer, curr_send_buffer_size,
                  positions + io_proc_idx, comm);
  }

  // the following is just to ensure that the packing does not leave any holes
  // in the send buffer (should theoretically never happen)
  char * to_buffer = send_buffer;
  for (unsigned i = 0; i < num_io_procs; ++i) {
    char * from_buffer = send_buffer + offsets[i];
    unsigned packed_size = positions[i];
    send_buffer_size_per_proc[i * io_proc_step] = packed_size;
    // should never happen
    if (from_buffer != to_buffer) memmove(to_buffer, from_buffer, packed_size);
    to_buffer += packed_size;
  }

  {
    struct temp_link_io_proc_data * temp_data;

    if (is_io_proc) {
      temp_data = malloc(1 * sizeof(*temp_data));
      temp_data->io_proc_idx = comm_rank / io_proc_step;
      temp_data->num_io_procs = num_io_procs;
      temp_data->num_weights_per_link =
        weight_file_data->link.num_weights_per_link;
      temp_data->weight_file_name = weight_file_data->weight_file_name;
      temp_data->num_links_per_io_proc = NULL;
      temp_data->num_pointsets = weight_file_data->link.num_pointsets;
      temp_data->links = NULL;
      temp_data->links_have_been_received = 0;
      temp_data->weights = NULL;
      temp_data->tags = weight_file_data->link.tags;
    } else {
      temp_data = NULL;
    }

    // gather link data
    yac_comm_alltoallv_callback(send_buffer, send_buffer_size_per_proc,
                                COMM_PACKED,
                                weight_file_data->link.tags.alltoallv_link_data,
                                comm, temp_data,
                                (is_io_proc)?alltoallv_link_data_callback:NULL,
                                MSG_PRIO_MIDDLE);

    // determine number of links per io process
    yac_comm_allreduce_callback(num_links_per_io_proc,
                                num_io_procs *
                                weight_file_data->link.num_pointsets, COMM_UINT,
                                COMM_OP_SUM,
                                weight_file_data->link.tags.allreduce_num_links,
                                comm, temp_data,
                                (is_io_proc)?allreduce_num_links_callback:NULL,
                                MSG_PRIO_MIDDLE);
  }

  free(offsets);
  free(send_buffer);
  free(positions);
  free(send_buffer_size_per_proc);
  free(num_links_per_io_proc);
}

static void
free_temp_weight_file_data(struct temp_weight_file_data * weight_file_data){


  free(weight_file_data->link.src_indices);
  free(weight_file_data->link.tgt_indices);
  free(weight_file_data->link.pointset_indices);
  free(weight_file_data->link.weights);
  free(weight_file_data->link.num_links_per_pointset);
  free(weight_file_data->fixed.global_tgt_indices);
  free(weight_file_data->fixed.num_tgt_per_fixed_value);
  free(weight_file_data->fixed.fixed_values);
  free(weight_file_data->fixed.collected_fixed_values);
  free(weight_file_data);
}

static void
redistribute_fixed_data(struct temp_weight_file_data * weight_file_data,
                        struct communicator * comm) {

  unsigned comm_size = yac_get_comm_size(comm);
  unsigned comm_rank = yac_get_comm_rank(comm);

  // redistribute fixed data among the io procs

  unsigned io_proc_step = weight_file_data->io_proc_step;
  unsigned num_io_procs = (comm_size + io_proc_step - 1) / io_proc_step;
  unsigned is_io_proc = !(comm_rank % io_proc_step);

  unsigned fixed_step =
    (weight_file_data->fixed.max_tgt_index -
     weight_file_data->fixed.min_tgt_index) / num_io_procs + 1;

  unsigned * num_tgt_per_io_proc =
    calloc(num_io_procs * weight_file_data->fixed.collected_num_fixed_values,
           sizeof(*num_tgt_per_io_proc));
  unsigned * send_buffer_size_per_proc =
    calloc(comm_size, sizeof(*send_buffer_size_per_proc));
  int * positions = calloc(num_io_procs, sizeof(*positions));
  int * offsets = malloc(num_io_procs * sizeof(*offsets));

  unsigned buffer_size_header = yac_comm_pack_size(1, COMM_UINT, comm);
  unsigned buffer_size_tgt = yac_comm_pack_size(1, COMM_INT, comm);

  // for all local fixed values
  for (unsigned i = 0, offset = 0; i < weight_file_data->fixed.num_fixed_values;
       ++i) {

    unsigned fixed_value_idx = 0;
    for (; fabs(weight_file_data->fixed.fixed_values[i] -
                weight_file_data->fixed.collected_fixed_values[
                  fixed_value_idx].value) > 0.0;
         ++fixed_value_idx);

    for (unsigned j = 0; j < weight_file_data->fixed.num_tgt_per_fixed_value[i];
         ++j, ++offset) {

      ++num_tgt_per_io_proc[
        weight_file_data->fixed.collected_num_fixed_values *
        ((weight_file_data->fixed.global_tgt_indices[offset] -
          weight_file_data->fixed.min_tgt_index)/fixed_step) + fixed_value_idx];
    }
  }

  unsigned send_buffer_size = 0;

  for (unsigned i = 0; i < num_io_procs; ++i) {
    send_buffer_size_per_proc[i * io_proc_step] =
      buffer_size_header * weight_file_data->fixed.collected_num_fixed_values;
    for (unsigned j = 0; j < weight_file_data->fixed.collected_num_fixed_values;
         ++j) {
      send_buffer_size_per_proc[i * io_proc_step] +=
        num_tgt_per_io_proc[
          i * weight_file_data->fixed.collected_num_fixed_values + j] *
        buffer_size_tgt;
    }
    send_buffer_size += send_buffer_size_per_proc[i * io_proc_step];
  }

  char * send_buffer = malloc(send_buffer_size);

  int accum = 0;
  for (unsigned i = 0; i < num_io_procs; ++i) {
    offsets[i] = accum;
    accum += send_buffer_size_per_proc[i * io_proc_step];
  }

  for (unsigned i = 0;
       i < weight_file_data->fixed.collected_num_fixed_values; ++i) {

    for (unsigned j = 0; j < num_io_procs; ++j)
      yac_comm_pack(
        num_tgt_per_io_proc +
        j * weight_file_data->fixed.collected_num_fixed_values + i, 1,
        COMM_UINT, send_buffer + offsets[j], send_buffer_size - offsets[j],
        positions + j, comm);

    unsigned offset = 0;
    unsigned fixed_value_idx = 0;
    for (; (fixed_value_idx < weight_file_data->fixed.num_fixed_values) &&
         fabs(weight_file_data->fixed.fixed_values[fixed_value_idx] -
              weight_file_data->fixed.collected_fixed_values[i].value) > 0.0;
         offset +=
          weight_file_data->fixed.num_tgt_per_fixed_value[fixed_value_idx++]);

    if (fixed_value_idx >= weight_file_data->fixed.num_fixed_values) continue;

    for (unsigned j = 0;
         j < weight_file_data->fixed.num_tgt_per_fixed_value[fixed_value_idx];
         ++j, ++offset) {

      unsigned io_proc_idx =
        (weight_file_data->fixed.global_tgt_indices[offset] -
         weight_file_data->fixed.min_tgt_index) / fixed_step;

      yac_comm_pack(weight_file_data->fixed.global_tgt_indices + offset, 1,
                    COMM_UINT, send_buffer + offsets[io_proc_idx],
                    send_buffer_size - offsets[io_proc_idx],
                    positions + io_proc_idx, comm);
    }
  }

  // the following is just to ensure that the packing does not leave any holes
  // in the send buffer (should theoretically never happen)
  char * to_buffer = send_buffer;
  for (unsigned i = 0; i < num_io_procs; ++i) {
    char * from_buffer = send_buffer + offsets[i];
    unsigned packed_size = positions[i];
    send_buffer_size_per_proc[i * io_proc_step] = packed_size;
    // should never happen
    if (from_buffer != to_buffer) memmove(to_buffer, from_buffer, packed_size);
    to_buffer += packed_size;
  }

  free(offsets);
  free(positions);

  {
    struct temp_fixed_io_proc_data * temp_data;

    if (is_io_proc) {
      temp_data = malloc(1 * sizeof(*temp_data));
      temp_data->io_proc_idx = comm_rank / io_proc_step;
      temp_data->num_io_procs = num_io_procs;
      temp_data->num_tgt_per_io_proc = NULL;
      temp_data->weight_file_name = weight_file_data->weight_file_name;
      temp_data->num_fixed_values =
        weight_file_data->fixed.collected_num_fixed_values;
      temp_data->global_tgt_indices_per_fixed_value = NULL;
      temp_data->tags = weight_file_data->fixed.tags;
    } else {
      temp_data = NULL;
    }

    // gather fixed data
    yac_comm_alltoallv_callback(send_buffer, send_buffer_size_per_proc,
                                COMM_PACKED,
                                weight_file_data->fixed.tags.alltoallv_fixed_data,
                                comm, temp_data,
                                (is_io_proc)?alltoallv_fixed_data_callback:NULL,
                                MSG_PRIO_MIDDLE);

    // determine number of fixed tgt indices per io process
    yac_comm_allreduce_callback(num_tgt_per_io_proc,
                                num_io_procs *
                                weight_file_data->fixed.collected_num_fixed_values,
                                COMM_UINT, COMM_OP_SUM,
                                weight_file_data->fixed.tags.allreduce_num_fixed,
                                comm, temp_data,
                                (is_io_proc)?allreduce_num_fixed_callback:NULL,
                                MSG_PRIO_MIDDLE);

    free(send_buffer);
    free(send_buffer_size_per_proc);
    free(num_tgt_per_io_proc);
  }
}

static void redistribute_weight_file_data(
  struct temp_weight_file_data * weight_file_data,
  struct communicator * comm) {

  // if this is root
  if (yac_get_comm_rank(comm) == 0)
    // create weight file and write fixed and link meta data
    create_weight_file(weight_file_data);

  // distribute fixed and link data among I/O processes and write it to file
  redistribute_link_data(weight_file_data, comm);
  if (weight_file_data->fixed.collected_num_fixed_values > 0)
    redistribute_fixed_data(weight_file_data, comm);
  free_temp_weight_file_data(weight_file_data);
}

static void allreduce_weight_file_meta_data_callback(struct communicator * comm,
                                                     struct comm_message msg,
                                                     void * user_data) {

  struct temp_weight_file_data * weight_file_data =
    (struct temp_weight_file_data *)user_data;

  weight_file_data->link.total_num_links = ((unsigned*)msg.buffer)[0];
  weight_file_data->link.num_weights_per_link = ((unsigned*)msg.buffer)[1];
  weight_file_data->link.min_tgt_index = ((int*)msg.buffer)[3];
  weight_file_data->link.max_tgt_index = ((int*)msg.buffer)[4];
  weight_file_data->link.num_links_per_pointset =
    malloc(weight_file_data->link.num_pointsets *
           sizeof(*weight_file_data->link.num_links_per_pointset));
  memcpy(weight_file_data->link.num_links_per_pointset,
         ((unsigned*)msg.buffer) + 5,
         weight_file_data->link.num_pointsets *
         sizeof(*weight_file_data->link.num_links_per_pointset));

  // if the link data meta data has already been distributed
  if (weight_file_data->fixed.collected_num_fixed_values != -1)
    redistribute_weight_file_data(weight_file_data, comm);
}

static void weight_file_meta_data_allreduce_op(void * invec, void * inoutvec,
                                               unsigned count,
                                               enum yac_comm_data_type type) {

  if (count < 5)
    yac_internal_abort_message(
      "ERROR(weight_file_meta_data_allreduce_op): count too small",
      __FILE__, __LINE__);

  if (type != COMM_UINT)
    yac_internal_abort_message(
      "ERROR(weight_file_meta_data_allreduce_op): invalid type",
      __FILE__, __LINE__);

  // num links
  unsigned * new_num_links = (unsigned*)inoutvec;
  *new_num_links = ((unsigned*)invec)[0] + ((unsigned*)inoutvec)[0];
  // num_weights_per_link
  if (((unsigned*)invec)[1] != ((unsigned*)inoutvec)[1]) {
    if ((((unsigned*)invec)[1] == 0) || (((unsigned*)inoutvec)[1] == 0))
      ((unsigned*)inoutvec)[1] =
      MAX(((unsigned*)invec)[1], ((unsigned*)inoutvec)[1]);
    else
      yac_internal_abort_message(
        "ERROR(weight_file_meta_data_allreduce_op): "
        "num_weights_per_link do not match", __FILE__, __LINE__);
  }
  // num_pointsets
  if (((unsigned*)invec)[2] != ((unsigned*)inoutvec)[2])
    yac_internal_abort_message(
      "ERROR(weight_file_meta_data_allreduce_op): "
      "num_pointsets do not match", __FILE__, __LINE__);
  unsigned num_pointsets = ((unsigned*)invec)[2];
  // min_tgt_index
  int * new_min_tgt_index = (int*)inoutvec + 3;
  *new_min_tgt_index = MIN(((int*)invec)[3], ((int*)inoutvec)[3]);
  // max_tgt_index
  int * new_max_tgt_index = (int*)inoutvec + 4;
  *new_max_tgt_index = MAX(((int*)invec)[4], ((int*)inoutvec)[4]);

  if ((count - 5) != num_pointsets)
    yac_internal_abort_message(
      "ERROR(weight_file_meta_data_allreduce_op): invalid count",
      __FILE__, __LINE__);

  // num links per pointset
  unsigned * new_num_links_per_pointset = (unsigned*)inoutvec + 5;
  for (unsigned i = 0; i < num_pointsets; ++i)
    new_num_links_per_pointset[i] = ((unsigned*)invec)[i+5] +
                                    new_num_links_per_pointset[i];
}

static void
unpack_fixed_meta_data(void * buffer, unsigned count,
                       struct communicator * comm,
                       struct fixed_value_meta_data ** fixed_values,
                       unsigned * num_fixed_values,
                       int * min_tgt_index, int * max_tgt_index) {

  unsigned single_pack_size = yac_comm_pack_size(1, COMM_DBLE, comm);
  single_pack_size += yac_comm_pack_size(2, COMM_INT, comm);
  single_pack_size += yac_comm_pack_size(1, COMM_UINT, comm);
  unsigned packing_unit_header_size = yac_comm_pack_size(1, COMM_UINT, comm);

  unsigned comm_size = yac_get_comm_size(comm);

  unsigned tmp_num_fixed_values =
    (count - comm_size * packing_unit_header_size) / single_pack_size;

  if (tmp_num_fixed_values == 0) {

    *fixed_values = NULL;
    *num_fixed_values = 0;
    *min_tgt_index = -1;
    *max_tgt_index = -1;
    return;
  }

  struct fixed_value_meta_data * tmp_fixed_values =
    malloc(tmp_num_fixed_values * sizeof(*tmp_fixed_values));
  int * min_max_tgt_index_per_fixed_value =
    malloc(2 * tmp_num_fixed_values *
           sizeof(*min_max_tgt_index_per_fixed_value));

  char * curr_packing_unit = (char*)buffer;
  unsigned max_packing_unit_size = count;

  for (unsigned i = 0; i < comm_size; ++i) {

    int position = 0;
    unsigned curr_num_fixed_values;

    yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &position,
                    &curr_num_fixed_values, 1, COMM_UINT, comm);

    for (unsigned j = 0; j < curr_num_fixed_values; ++j) {

      yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &position,
                      &tmp_fixed_values[j].value, 1, COMM_DBLE, comm);
      yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &position,
                      min_max_tgt_index_per_fixed_value + 2 * j + 0,
                      1, COMM_INT, comm);
      yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &position,
                      min_max_tgt_index_per_fixed_value + 2 * j + 1,
                      1, COMM_INT, comm);
      yac_comm_unpack(curr_packing_unit, max_packing_unit_size, &position,
                      &tmp_fixed_values[j].count, 1, COMM_UINT, comm);
    }

    unsigned curr_packing_unit_size =
      packing_unit_header_size + curr_num_fixed_values * single_pack_size;
    curr_packing_unit += curr_packing_unit_size;
    max_packing_unit_size -= curr_packing_unit_size;
  }

  int tmp_min_tgt_index = min_max_tgt_index_per_fixed_value[0];
  int tmp_max_tgt_index = min_max_tgt_index_per_fixed_value[1];

  for (unsigned i = 1; i < tmp_num_fixed_values; ++i) {
    tmp_min_tgt_index = MIN(tmp_min_tgt_index,
                            min_max_tgt_index_per_fixed_value[2 * i + 0]);
    tmp_max_tgt_index = MAX(tmp_max_tgt_index,
                            min_max_tgt_index_per_fixed_value[2 * i + 1]);
  }

  free(min_max_tgt_index_per_fixed_value);

  *fixed_values = tmp_fixed_values;
  *num_fixed_values = tmp_num_fixed_values;
  *min_tgt_index = tmp_min_tgt_index;
  *max_tgt_index = tmp_max_tgt_index;
}

static int compare_fixed_values(const void * a, const void * b) {

  return (((struct fixed_value_meta_data*)a)->value >
          ((struct fixed_value_meta_data*)b)->value) -
         (((struct fixed_value_meta_data*)a)->value <
          ((struct fixed_value_meta_data*)b)->value);
}

static void
determine_unique_fixed_values(struct fixed_value_meta_data * fixed_values,
                              unsigned * num_fixed_values) {

  if (*num_fixed_values == 0) return;

  unsigned old_num_fixed_values, new_num_fixed_values;

  old_num_fixed_values = *num_fixed_values;

  qsort(fixed_values, old_num_fixed_values, sizeof(*fixed_values),
        compare_fixed_values);

  new_num_fixed_values = 1;
  double prev_fixed_value = fixed_values[0].value;
  for (unsigned i = 1; i < old_num_fixed_values; ++i) {
    if (fabs(prev_fixed_value - fixed_values[i].value) > 0.0) {
      ++new_num_fixed_values;
    } else {
      prev_fixed_value = fixed_values[i].value;
      fixed_values[new_num_fixed_values-1].count += fixed_values[i].count;
    }
  }

  *num_fixed_values = new_num_fixed_values;
}

static void
allgatherv_fixed_meta_data_callback(struct communicator * comm,
                                    struct comm_message msg,
                                    void * user_data) {

  struct temp_weight_file_data * weight_file_data =
    (struct temp_weight_file_data *)user_data;

  struct fixed_value_meta_data * fixed_values;
  unsigned num_fixed_values;

  unpack_fixed_meta_data(msg.buffer, msg.count, comm, &fixed_values,
                         &num_fixed_values,
                         &weight_file_data->fixed.min_tgt_index,
                         &weight_file_data->fixed.max_tgt_index);

  determine_unique_fixed_values(fixed_values, &num_fixed_values);

  weight_file_data->fixed.collected_fixed_values = fixed_values;
  weight_file_data->fixed.collected_num_fixed_values = num_fixed_values;

  // if the link data meta data has already been distributed
  if (weight_file_data->link.total_num_links != -1)
    redistribute_weight_file_data(weight_file_data, comm);
}

static void
allgather_num_fixed_values_callback(struct communicator * comm,
                                    struct comm_message msg,
                                    void * user_data) {

  struct temp_weight_file_data * weight_file_data =
    (struct temp_weight_file_data *)user_data;

  //---------------------------------------------------------------
  // determin minimum and maximum target index for each fixed value
  //---------------------------------------------------------------

  int * min_max_tgt_index_per_fixed_value =
    malloc(2 * weight_file_data->fixed.num_fixed_values *
           sizeof(*min_max_tgt_index_per_fixed_value));

  for (unsigned i = 0, offset = 0;
       i < weight_file_data->fixed.num_fixed_values; ++i) {

    min_max_tgt_index_per_fixed_value[2 * i + 0] =
      weight_file_data->fixed.global_tgt_indices[offset];
    min_max_tgt_index_per_fixed_value[2 * i + 1] =
      weight_file_data->fixed.global_tgt_indices[offset];

    for (unsigned j = 0;
         j < weight_file_data->fixed.num_tgt_per_fixed_value[i];
         ++j, ++offset) {

      min_max_tgt_index_per_fixed_value[2 * i + 0] =
        MIN(min_max_tgt_index_per_fixed_value[2 * i + 0],
            weight_file_data->fixed.global_tgt_indices[offset]);
      min_max_tgt_index_per_fixed_value[2 * i + 1] =
        MAX(min_max_tgt_index_per_fixed_value[2 * i + 1],
            weight_file_data->fixed.global_tgt_indices[offset]);
    }
  }

  //---------------------
  // pack local meta data
  //---------------------

  unsigned single_pack_size = yac_comm_pack_size(1, COMM_DBLE, comm);
  single_pack_size += yac_comm_pack_size(2, COMM_INT, comm);
  single_pack_size += yac_comm_pack_size(1, COMM_UINT, comm);
  unsigned packing_unit_header_size = yac_comm_pack_size(1, COMM_UINT, comm);

  unsigned send_buffer_size =
    weight_file_data->fixed.num_fixed_values * single_pack_size +
    packing_unit_header_size;

  char * send_buffer = malloc(send_buffer_size);

  int position = 0;
  yac_comm_pack(&(weight_file_data->fixed.num_fixed_values),
                1, COMM_UINT, send_buffer, send_buffer_size, &position,
                comm);

  for (unsigned i = 0; i < weight_file_data->fixed.num_fixed_values; ++i) {

    yac_comm_pack(weight_file_data->fixed.fixed_values + i,
                  1, COMM_DBLE, send_buffer, send_buffer_size, &position,
                  comm);
    yac_comm_pack(min_max_tgt_index_per_fixed_value + 2 * i,
                  2, COMM_INT, send_buffer, send_buffer_size, &position,
                  comm);
    yac_comm_pack(weight_file_data->fixed.num_tgt_per_fixed_value + i,
                  1, COMM_UINT, send_buffer, send_buffer_size, &position,
                  comm);
  }

  free(min_max_tgt_index_per_fixed_value);

  //--------------------------------------------
  // gather remaining meta data on all processes
  //--------------------------------------------

  unsigned comm_size = yac_get_comm_size(comm);
  unsigned * num_fixed_values_per_rank = (unsigned *)(msg.buffer);

  unsigned * counts = malloc(comm_size * sizeof(*counts));

  for (unsigned i = 0; i < comm_size; ++i)
    counts[i] = packing_unit_header_size +
                single_pack_size * num_fixed_values_per_rank[i];

  yac_comm_allgatherv_callback(
    send_buffer, counts, COMM_PACKED,
    weight_file_data->fixed.tags.allgatherv_meta_data, comm,
    weight_file_data, allgatherv_fixed_meta_data_callback, MSG_PRIO_MIDDLE);

  free(counts);
  free(send_buffer);
}

static void handle_interp_stack_fixed_data(
  struct field_interpolation_gen_src * interp_src,
  struct temp_interpolation_plan_data * interp_plan_data,
  struct temp_weight_file_data * weight_file_data) {

  weight_file_data->fixed.tags = interp_src->fixed_data_tags;

  //-----------------------------------------------------------------
  // get fixed target point information from interpolation plan stack
  //-----------------------------------------------------------------

  weight_file_data->fixed.global_tgt_indices =
    malloc(interp_plan_data->num_tgt_indices_fixed *
           sizeof(*(weight_file_data->fixed.global_tgt_indices)));
  memcpy(weight_file_data->fixed.global_tgt_indices,
         interp_plan_data->global_tgt_indices,
         interp_plan_data->num_tgt_indices_fixed *
         sizeof(*(weight_file_data->fixed.global_tgt_indices)));
  weight_file_data->fixed.num_tgt_indices =
    interp_plan_data->num_tgt_indices_fixed;
  weight_file_data->fixed.num_tgt_per_fixed_value =
    malloc(interp_plan_data->num_fixed_values *
           sizeof(*(weight_file_data->fixed.num_tgt_per_fixed_value)));
  memcpy(weight_file_data->fixed.num_tgt_per_fixed_value,
         interp_plan_data->num_tgt_per_fixed_value,
         interp_plan_data->num_fixed_values *
         sizeof(*(weight_file_data->fixed.num_tgt_per_fixed_value)));
  weight_file_data->fixed.fixed_values =
    malloc(interp_plan_data->num_fixed_values *
           sizeof(*(weight_file_data->fixed.fixed_values)));
  memcpy(weight_file_data->fixed.fixed_values, interp_plan_data->fixed_values,
         interp_plan_data->num_fixed_values *
         sizeof(*(weight_file_data->fixed.fixed_values)));
  weight_file_data->fixed.num_fixed_values = interp_plan_data->num_fixed_values;

  //-----------------------------------------------
  // gather number of fixed values on all processes
  //-----------------------------------------------

  yac_comm_allgather_callback(
    &weight_file_data->fixed.num_fixed_values, 1, COMM_UINT,
    weight_file_data->fixed.tags.allgather_num_fixed_values,
    interp_src->interp_data->halo_comm, weight_file_data,
    allgather_num_fixed_values_callback, MSG_PRIO_MIDDLE);
}

static void handle_interp_stack_link_data(
  struct field_interpolation_gen_src * interp_src,
  struct temp_interpolation_plan_data * interp_plan_data,
  struct temp_weight_file_data * weight_file_data) {

  unsigned num_links = 0;
  for (unsigned i = 0; i < interp_src->interp_data->num_pointsets; ++i)
    num_links += interp_plan_data->num_links_per_pointset[i];

  weight_file_data->link.src_indices =
    malloc(num_links * sizeof(*(weight_file_data->link.src_indices)));
  memcpy(weight_file_data->link.src_indices, interp_plan_data->src_indices,
         num_links * sizeof(*(weight_file_data->link.src_indices)));
  weight_file_data->link.pointset_indices =
    malloc(num_links * sizeof(*(weight_file_data->link.pointset_indices)));
  memcpy(weight_file_data->link.pointset_indices,
         interp_plan_data->pointset_indices,
         num_links * sizeof(*(weight_file_data->link.pointset_indices)));
  weight_file_data->link.weights =
    malloc(num_links * sizeof(*(weight_file_data->link.weights)));
  memcpy(weight_file_data->link.weights, interp_plan_data->weights,
         num_links * sizeof(*(weight_file_data->link.weights)));

  //-----------------------
  // gather local link data
  //-----------------------

  int * tgt_indices = malloc(num_links * sizeof(*tgt_indices));

  for (unsigned num_remaining_links = num_links, tgt_idx = 0, offset = 0;
       num_remaining_links > 0;
       num_remaining_links -= interp_plan_data->num_links_per_tgt[tgt_idx++]) {

    unsigned tgt_id = interp_src->collected_global_ids[tgt_idx];

    for (unsigned i = 0; i < interp_plan_data->num_links_per_tgt[tgt_idx];
         ++i, ++offset)
      tgt_indices[offset] = tgt_id;
  }

  //--------------------------------------------------------------------------
  // allgather min/max tgt index, max num weights per file and num links,
  // afterwards redistribute link data among io procs
  //--------------------------------------------------------------------------

  weight_file_data->link.tags = interp_src->link_data_tags;
  weight_file_data->link.num_links = num_links;
  weight_file_data->link.tgt_indices = tgt_indices;
  weight_file_data->link.num_pointsets = interp_src->interp_data->num_pointsets;
  weight_file_data->link.src_locations = interp_src->src_locations;
  weight_file_data->link.tgt_location = interp_src->tgt_location;

  int min_tgt_index = 0;
  int max_tgt_index = 0;
  if (num_links > 0) {
    min_tgt_index = tgt_indices[0];
    max_tgt_index = tgt_indices[0];
    for (unsigned i = 0; i < num_links; ++i) {
      if (tgt_indices[i] < min_tgt_index) min_tgt_index = tgt_indices[i];
      else if (tgt_indices[i] > max_tgt_index) max_tgt_index = tgt_indices[i];
    }
  }

  unsigned * weight_file_meta_data =
    malloc((5 + weight_file_data->link.num_pointsets) *
           sizeof(*weight_file_meta_data));
  weight_file_meta_data[0] = num_links;
  weight_file_meta_data[1] = interp_plan_data->num_weights_per_link;
  weight_file_meta_data[2] = weight_file_data->link.num_pointsets;
  ((int*)weight_file_meta_data)[3] = min_tgt_index;
  ((int*)weight_file_meta_data)[4] = max_tgt_index;
  for (unsigned i = 0; i < weight_file_data->link.num_pointsets; ++i)
    weight_file_meta_data[5+i] = interp_plan_data->num_links_per_pointset[i];

  // reduce meta data
  yac_comm_allreduce_callback(weight_file_meta_data,
                              5 + weight_file_data->link.num_pointsets,
                              COMM_UINT, weight_file_meta_data_allreduce_op,
                              weight_file_data->link.tags.allreduce_meta_data,
                              interp_src->interp_data->halo_comm,
                              weight_file_data,
                              allreduce_weight_file_meta_data_callback,
                              MSG_PRIO_MIDDLE);

  free(weight_file_meta_data);
}

static void
write_interpolation_to_file(
  struct field_interpolation_gen_src * interp_src,
  struct temp_interpolation_plan_data * interp_plan_data) {

  struct temp_weight_file_data * weight_file_data =
    malloc(1 * sizeof(*weight_file_data));

  weight_file_data->link.total_num_links = -1;
  weight_file_data->fixed.collected_num_fixed_values = -1;
  weight_file_data->weight_file_name = interp_src->weight_file_names[0];
  weight_file_data->src_grid_name =
    yac_get_local_domain_name(interp_src->search);
  weight_file_data->tgt_grid_name =
    yac_get_remote_domain_name(interp_src->search);
  weight_file_data->io_proc_step = interp_src->io_proc_step;

  //-------------------------------------------
  // handle link data from interpolation stack
  //-------------------------------------------

  handle_interp_stack_link_data(interp_src, interp_plan_data, weight_file_data);

  //------------------------------------------------
  // handle fixed tgt data from interpolation stack
  //------------------------------------------------

  handle_interp_stack_fixed_data(interp_src, interp_plan_data,
                                 weight_file_data);
}

static struct temp_interpolation_plan_data *
get_interpolation_plan_data(
  struct field_interpolation_gen_src * interp_src,
  struct interpolation_method_data_request * data_request) {

  struct temp_interpolation_plan_data * interp_plan_data =
    malloc(1 * sizeof(*interp_plan_data));

  //--------------------------------------------------------------------------
  // create input data for interpolation plan containing global source indices
  //--------------------------------------------------------------------------

  int *** global_indices =
    malloc((interp_src->interp_data->num_subdomains + 1) *
           sizeof(*global_indices));
  global_indices[0] =
    malloc(interp_src->interp_data->num_pointsets *
           sizeof(**global_indices));
  for (unsigned i = 0; i < interp_src->interp_data->num_pointsets; ++i)
    global_indices[0][i] = data_request[i].global_ids;

  struct domain * local_domain = yac_get_local_domain(interp_src->search);

  for (unsigned i = 0; i < interp_src->interp_data->num_subdomains; ++i) {

     global_indices[i+1] = malloc(interp_src->interp_data->num_pointsets *
                                  sizeof(**global_indices));

     for (unsigned j = 0; j < interp_src->interp_data->num_pointsets; ++j)
        global_indices[i+1][j] =
           (int*)yac_get_global_ids(local_domain, i,
                                    interp_src->src_locations[j]);
  }

  // get link data
  yac_get_link_data(interp_src->interp_data->plans, global_indices,
                    interp_src->interp_data->num_pointsets,
                    &interp_plan_data->num_tgt_indices_links,
                    &interp_plan_data->num_links_per_pointset,
                    &interp_plan_data->num_weights_per_link,
                    &interp_plan_data->num_links_per_tgt,
                    &interp_plan_data->src_indices,
                    &interp_plan_data->pointset_indices,
                    &interp_plan_data->weights);

  // global indices are not needed anymore
  for (unsigned i = 0; i < interp_src->interp_data->num_subdomains + 1; ++i)
    free(global_indices[i]);
  free(global_indices);

  // get fixed data
  yac_get_fixed_data(interp_src->interp_data->plans,
                     interp_src->collected_global_ids,
                     &interp_plan_data->global_tgt_indices,
                     &interp_plan_data->num_tgt_indices_fixed,
                     &interp_plan_data->num_tgt_per_fixed_value,
                     &interp_plan_data->fixed_values,
                     &interp_plan_data->num_fixed_values);

  // sort the global target indices for each fixed value
  for (unsigned i = 0, offset = 0; i < interp_plan_data->num_fixed_values;
       ++i) {

    qsort(interp_plan_data->global_tgt_indices + offset,
          interp_plan_data->num_tgt_per_fixed_value[i],
          sizeof(*(interp_plan_data->global_tgt_indices)),
          compare_int);
    offset += interp_plan_data->num_tgt_per_fixed_value[i];
  }

  return interp_plan_data;
}

static void
free_interpolation_plan_data(
  struct temp_interpolation_plan_data * interp_plan_data) {

  if (interp_plan_data == NULL) return;

  free(interp_plan_data->num_links_per_pointset);
  free(interp_plan_data->num_links_per_tgt);
  free(interp_plan_data->src_indices);
  free(interp_plan_data->pointset_indices);
  free(interp_plan_data->weights);
  free(interp_plan_data->global_tgt_indices);
  free(interp_plan_data->num_tgt_per_fixed_value);
  free(interp_plan_data->fixed_values);
  free(interp_plan_data);
}

/**
 * routine is called automatically once the processing of the
 * interpolation method stack is finished
 *
 * (this routine is called on the source side)
 *
 * the callback is set in the field_interpolation_gen_src constructor
 *(\ref yac_new_field_interpolation_gen_src)via the interpolation method
 * routine \ref yac_set_finish_callback_interpolation_method
 */
static void method_finish_callback(void * data) {

   struct temp_point_search_data * search_data;

   search_data = (struct temp_point_search_data *) data;

   // if the mapping is to be executed on the source processes
   if (search_data->interp_src->mapping_on_source)
      // return to the target processes the points that can be interpolated by
      // the local process
      // in addition the send_list for the put operation is generated
      return_interpolation_points(
         search_data->interp_src, search_data->num_interpolation_methods,
         search_data->num_interpolated_points_per_method,
         search_data->num_target_requests, search_data->target_requests,
         search_data->collected_global_ids);

   // get the data_requests of the interpolation methods
   struct interpolation_method_data_request * data_request =
      malloc(search_data->interp_src->interp_data->num_pointsets *
             sizeof(*data_request));
   yac_get_source_exchange_data_interpolation_method(
      search_data->methods, &data_request,
      search_data->interp_src->interp_data->num_pointsets);

   struct temp_interpolation_plan_data * interp_plan_data = NULL;
   if ((search_data->interp_src->num_weight_file_names > 0) ||
       !(search_data->interp_src->mapping_on_source))
      interp_plan_data =
        get_interpolation_plan_data(search_data->interp_src, data_request);

   // write the weight file if required
   if (search_data->interp_src->num_weight_file_names > 0)
      write_interpolation_to_file(search_data->interp_src, interp_plan_data);


   // if the mapping is to be executed on the source processes
   if (search_data->interp_src->mapping_on_source)
      // exchange halo requests with other source processes and set up local halo
      // receive list
      exchange_halo_requests_source(search_data->interp_src,
                                    data_request);
   else
      // exchange halo requests with other source processes and inform target
      // processes about which processes will be provided by which source
      // process
      exchange_halo_requests_target(
        search_data, data_request, interp_plan_data);

   free_interpolation_plan_data(interp_plan_data);

   // free the data request
   yac_free_interpolation_method_data_request(
      data_request, search_data->interp_src->interp_data->num_pointsets);

   cleanup_search_data(search_data);
}

static void tgt_point_location_recv_callback(struct communicator * comm,
                                             struct comm_message msg,
                                             void * user_data) {

  UNUSED(comm);

  struct temp_tgt_point_location_data * temp_data =
    (struct temp_tgt_point_location_data *)user_data;

  unsigned root = yac_get_comm_rank(temp_data->comm);

  yac_comm_bcast_callback(msg.buffer, 1, COMM_INT, root, temp_data->tag,
                          temp_data->comm, temp_data->search_data,
                          temp_data->callback, MSG_PRIO_HIGH);

  free(user_data);
}

// in case the source process has no overlap with any target process, it
// gets it from the local root. This is required for calling the
// interpolation method stack.
// (the matching yac_comm_bcast_callback call for this routine can be found in
//  \ref yac_new_field_interpolation_gen_src)
static void tgt_point_location_bcast_callback(struct communicator * comm,
                                              struct comm_message msg,
                                              void * user_data) {

   UNUSED(comm);

   struct temp_point_search_data * search_data =
      (struct temp_point_search_data *)user_data;

   enum yac_location target_point_location =
      (enum yac_location)(*(int *)msg.buffer);

   // call interpolation method stack
   yac_do_search_interpolation_method(
     search_data->methods, search_data->interp_src->interp_data->plans,
     NULL, NULL, target_point_location, 0,
     search_data->num_interpolated_points_per_method,
     search_data->interp_src->interp_data->num_pointsets);
}

struct field_interpolation_gen *
yac_new_field_interpolation_gen_src_(struct points *** points,
                                     unsigned num_pointsets,
                                     char const * field_config_name,
                                     struct global_search * search,
                                     const char * other_field_config_name,
                                     struct interpolation_method ** methods,
                                     char const * weight_file_name,
                                     unsigned io_proc_step,
                                     struct field_interpolation_tags tags,
                                     int mapping_on_source) {

   if (points == NULL)
      yac_internal_abort_message("ERROR: no point set defined", __FILE__, __LINE__);

   unsigned num_interpolation_methods;

   for (num_interpolation_methods = 0;
        methods[num_interpolation_methods] != NULL;
        num_interpolation_methods++);

   struct field_interpolation_gen_src * interp_src;

   interp_src = malloc(1 * sizeof(*interp_src));
   interp_src->interp_data = malloc(1 * sizeof(*(interp_src->interp_data)));

   interp_src->vtable = &field_interpolation_gen_src_vtable;
   interp_src->search = search;
   interp_src->interp_data->comm = yac_get_communicator(search);
   interp_src->interp_data->num_subdomains =
     yac_get_num_subdomains(yac_get_local_domain(search));
   interp_src->interp_data->num_pointsets = num_pointsets;
   interp_src->interp_data->halo_comm =
     yac_get_domain_comm(yac_get_local_domain(search));
   interp_src->methods = methods;

   generate_field_interpolation_gen_tag(interp_src, field_config_name,
                                        other_field_config_name);
   interp_src->tags.results_send = tags.results_exchange;
   interp_src->tags.point_data_recv = tags.point_data_exchange;
   interp_src->tags.point_data_header_recv = tags.point_data_header_exchange;
   interp_src->tags.tgt_point_location_recv = tags.tgt_point_location_exchange;
   interp_src->tags.interp_data_msg_count = tags.interp_data_msg_count;
   interp_src->tags.interp_data_header = tags.interp_data_header;
   interp_src->tags.interp_data = tags.interp_data;
   interp_src->put_get_tags.put = tags.exchange;

   interp_src->interp_data->total_num_interpolated_points = 0;
   if (mapping_on_source) {
      interp_src->interp_data->mapping_on.src.send_msgs = NULL;
      interp_src->interp_data->mapping_on.src.num_send_msgs = 0;
   } else {
      interp_src->interp_data->mapping_on.tgt.send_msgs = NULL;
      interp_src->interp_data->mapping_on.tgt.num_send_msgs = 0;
   }
   interp_src->interp_data->max_send_buffer_size = 0;
   interp_src->interp_data->reference_counter = 1;
   interp_src->num_interpolated_points_per_method =
      calloc(num_interpolation_methods,
             sizeof(*(interp_src->num_interpolated_points_per_method)));
   interp_src->collected_global_ids = NULL;
   interp_src->src_locations =
      malloc(num_pointsets * sizeof(*interp_src->src_locations));
   interp_src->num_interpolation_methods = num_interpolation_methods;
   for (unsigned i = 0; i < num_pointsets; ++i)
      interp_src->src_locations[i] = points[0][i]->location;
   interp_src->weight_file_names = NULL;
   interp_src->weight_file_names_array_size = 0;
   interp_src->num_weight_file_names = 0;
   field_interpolation_gen_src_add_weight_file_name(
     (field_interpolation_gen)interp_src, weight_file_name);
   interp_src->interp_data->plans = malloc((num_interpolation_methods + 1)*
                                           sizeof(*(interp_src->interp_data->plans)));
   for (unsigned i = 0; i < num_interpolation_methods+1; ++i)
      interp_src->interp_data->plans[i] = NULL;
   interp_src->io_proc_step = io_proc_step;
   interp_src->mapping_on_source = mapping_on_source;

   unsigned comm_size;

   comm_size = yac_get_comm_size(interp_src->interp_data->comm);

   unsigned num_open_target_requests = 0;
   for (unsigned i = 0; i < comm_size; ++i)
      if (yac_source_process_has_overlap_with(interp_src->search, i))
         num_open_target_requests++;

   struct temp_point_search_data * search_data;
   search_data = malloc(1 * sizeof(*search_data));
   search_data->methods = methods;
   search_data->interp_src = interp_src;
   search_data->num_interpolation_methods = num_interpolation_methods;
   search_data->num_interpolated_points_per_method =
      interp_src->num_interpolated_points_per_method;

   yac_set_finish_callback_interpolation_method(methods, method_finish_callback,
                                                search_data);

   search_data->num_open_target_requests = num_open_target_requests;
   search_data->num_target_requests = 0;
   search_data->target_requests = malloc(search_data->num_open_target_requests *
                                         sizeof(*(search_data->target_requests)));
   search_data->collected_global_ids = NULL;
   search_data->collected_coordinates_xyz = NULL;

   interp_src->interp_data->halo_send_msgs = NULL;
   interp_src->interp_data->halo_recv_msgs = NULL;
   interp_src->interp_data->num_halo_send_msgs = 0;
   interp_src->interp_data->num_halo_recv_msgs = 0;
   interp_src->interp_data->total_num_halo_send_points = 0;
   interp_src->interp_data->total_num_halo_recv_points = 0;
   interp_src->interp_data->num_halo_recv_points_per_points_set = NULL;

   interp_src->interp_data->num_halo_recv_points_per_points_set =
     calloc(interp_src->interp_data->num_pointsets,
       sizeof(*(interp_src->interp_data->num_halo_recv_points_per_points_set)));

   func_recv_callback p_tgt_point_location_bcast_callback;

   if (num_open_target_requests > 0) {

      p_tgt_point_location_bcast_callback = NULL;

      for (unsigned i = 0; i < comm_size; ++i)
         if (yac_source_process_has_overlap_with(interp_src->search, i))
            // the matching send can be found in field_interpolation_gen_tgt
            // routine send_interpolation_data
            yac_comm_recv_callback(4, COMM_UINT, i,
                                   interp_src->tags.point_data_header_recv,
                                   interp_src->interp_data->comm, search_data,
                                   receive_interpolation_init_header_callback,
                                   NULL, MSG_PRIO_HIGH);
   } else {

      // for source processes that overlap with any target process the following
      // barrier is called after the points that have to be interpolated have
      // been received
      yac_comm_barrier_callback(interp_src->tags.point_data_barrier,
                                interp_src->interp_data->halo_comm, NULL, NULL,
                                MSG_PRIO_HIGH);

      p_tgt_point_location_bcast_callback =
         tgt_point_location_bcast_callback;
   }

   unsigned halo_comm_local_root_rank;
   unsigned local_root_rank =
      yac_get_lowest_local_domain_rank(interp_src->search,
                                       &halo_comm_local_root_rank);

   if (local_root_rank == yac_get_comm_rank(interp_src->interp_data->comm)) {

      struct temp_tgt_point_location_data * temp_data =
         malloc(1 * sizeof(*temp_data));

      temp_data->callback = p_tgt_point_location_bcast_callback;
      temp_data->comm = interp_src->interp_data->halo_comm;
      temp_data->search_data = search_data;
      temp_data->tag = interp_src->tags.tgt_point_location_bcast;

      unsigned target_rank =
         yac_get_lowest_remote_domain_rank(interp_src->search);

      // receive the target point location and broadcast it among all source
      // processes
      yac_comm_recv_callback(1, COMM_INT, target_rank,
                             interp_src->tags.tgt_point_location_recv,
                             interp_src->interp_data->comm, temp_data,
                             tgt_point_location_recv_callback, NULL,
                             MSG_PRIO_HIGH);

   } else {

      // get the target point location
      yac_comm_bcast_callback(NULL, 1, COMM_INT, halo_comm_local_root_rank,
                              interp_src->tags.tgt_point_location_bcast,
                              interp_src->interp_data->halo_comm, search_data,
                              p_tgt_point_location_bcast_callback, MSG_PRIO_HIGH);
   }

   return (struct field_interpolation_gen*)interp_src;
}

struct field_interpolation_gen *
yac_new_field_interpolation_gen_src(struct points *** points,
                                    unsigned num_pointsets,
                                    char const * field_config_name,
                                    struct global_search * search,
                                    const char * other_field_config_name,
                                    struct interpolation_method ** methods,
                                    char const * weight_file_name,
                                    unsigned io_proc_step,
                                    int mapping_on_source) {

   if (points == NULL)
      yac_internal_abort_message(
         "ERROR: no point set defined", __FILE__, __LINE__);

   struct communicator * comm = yac_get_communicator(search);
   char const * local_comp_name = yac_get_local_component_name(search);
   char const * remote_comp_name = yac_get_remote_component_name(search);
   char const * local_domain_name = yac_get_local_domain_name(search);
   char const * remote_domain_name = yac_get_remote_domain_name(search);

   struct field_interpolation_tags tags =
      yac_field_interpolation_generate_tag(
         comm, local_comp_name, remote_comp_name, local_domain_name,
         remote_domain_name, field_config_name, other_field_config_name);

   return
      (struct field_interpolation_gen*)yac_new_field_interpolation_gen_src_(
         points, num_pointsets, field_config_name, search,
         other_field_config_name, methods, weight_file_name, io_proc_step,
         tags, mapping_on_source);
}

static void
field_interpolation_gen_src_enddef(
   struct field_interpolation_gen * interpolation_gen) {

   struct field_interpolation_gen_src * interp_src;

   interp_src = (struct field_interpolation_gen_src *)interpolation_gen;

   // the order in which the while's and the yac_enddef_interpolation_method
   // is called is significant
   while (yac_comm_wait(interp_src->interp_data->comm,
                       interp_src->tags.point_data_header_recv) ||
          yac_comm_wait(interp_src->interp_data->comm,
                         interp_src->tags.tgt_point_location_recv) ||
          yac_comm_wait(interp_src->interp_data->comm,
                         interp_src->tags.point_data_recv) ||
          yac_comm_wait(interp_src->interp_data->halo_comm,
                         interp_src->tags.point_data_barrier) ||
          yac_comm_wait(interp_src->interp_data->halo_comm,
                         interp_src->tags.tgt_point_location_bcast));
   yac_enddef_interpolation_method(interp_src->methods);
   while(yac_comm_wait(interp_src->interp_data->halo_comm,
                       interp_src->tags.halo_data_alltoallv) ||
         yac_comm_wait(interp_src->interp_data->halo_comm,
                       interp_src->tags.results_send) ||
         yac_comm_wait(interp_src->interp_data->halo_comm,
                       interp_src->tags.allreduce_num_weights_per_link) ||
         yac_comm_wait(interp_src->interp_data->halo_comm,
                       interp_src->tags.reduce_send_flag) ||
         yac_comm_wait(interp_src->interp_data->comm,
                       interp_src->tags.interp_data_msg_count) ||
         yac_comm_wait(interp_src->interp_data->comm,
                       interp_src->tags.interp_data_header) ||
         yac_comm_wait(interp_src->interp_data->comm,
                       interp_src->tags.interp_data));

   interp_src->methods = NULL;

   if (interp_src->num_weight_file_names > 0) {

      while (yac_comm_wait(interp_src->interp_data->halo_comm,
                           interp_src->link_data_tags.alltoallv_link_data) ||
             yac_comm_wait(interp_src->interp_data->halo_comm,
                           interp_src->link_data_tags.allreduce_num_links) ||
             yac_comm_wait(interp_src->interp_data->halo_comm,
                           interp_src->link_data_tags.allreduce_meta_data) ||
             yac_comm_wait(interp_src->interp_data->halo_comm,
                           interp_src->fixed_data_tags.allgather_num_fixed_values) ||
             yac_comm_wait(interp_src->interp_data->halo_comm,
                           interp_src->fixed_data_tags.allgatherv_meta_data) ||
             yac_comm_wait(interp_src->interp_data->halo_comm,
                           interp_src->fixed_data_tags.alltoallv_fixed_data) ||
             yac_comm_wait(interp_src->interp_data->halo_comm,
                           interp_src->fixed_data_tags.allreduce_num_fixed));

      // if the weight file is to be stored under multiple file names
      if ((interp_src->num_weight_file_names > 1) &&
          (yac_get_comm_rank(interp_src->interp_data->halo_comm) == 0)) {

        char * command = NULL;
        unsigned command_size = 0;
        unsigned from_size = strlen(interp_src->weight_file_names[0]);

        for (unsigned i = 1; i < interp_src->num_weight_file_names; ++i) {

          unsigned to_size = strlen(interp_src->weight_file_names[1]);
          ENSURE_ARRAY_SIZE(command, command_size, 5 + from_size + to_size);
          sprintf(command, "cp %s %s", interp_src->weight_file_names[0],
                  interp_src->weight_file_names[1]);
          system(command);
        }

        free(command);
      }
   }
}

static struct field_interpolation *
field_interpolation_gen_src_get_field_interpolation
  (struct field_interpolation_gen * interpolation_gen) {

  //! \todo check that field_interpolation_gen_src_enddef has been called
  //!       before

  struct field_interpolation_gen_src * interp_gen_src;
  struct field_interpolation_src * interp_src;

  interp_gen_src = (struct field_interpolation_gen_src *)interpolation_gen;
  interp_src = malloc(1 * sizeof(*interp_src));

  if (interp_gen_src->mapping_on_source)
    interp_src->vtable = &field_interpolation_src_mapping_on_src_vtable;
  else
    interp_src->vtable = &field_interpolation_src_mapping_on_tgt_vtable;
  interp_src->interp_data = interp_gen_src->interp_data;
  interp_src->interp_data->reference_counter++;
  interp_src->put_buffers = NULL;
  interp_src->put_buffers_array_size = 0;
  interp_src->num_put_buffers = 0;

  interp_src->tags = interp_gen_src->put_get_tags;

  return (struct field_interpolation *)interp_src;
}

static struct put_buffers *
get_put_buffers(struct field_interpolation_src * interp_src,
                int collection_size) {

  for (unsigned i = 0; i < interp_src->num_put_buffers; ++i) {

    if (interp_src->put_buffers[i].collection_size == collection_size) {

      // wait until potential previous put operation is finished
      yac_comm_waitall(interp_src->put_buffers[i].num_open_requests,
                       interp_src->put_buffers[i].requests);

      interp_src->put_buffers[i].num_open_requests = 0;

      return interp_src->put_buffers + i;
    }
  }

  // no matching put buffers have been found

  ENSURE_ARRAY_SIZE(interp_src->put_buffers, interp_src->put_buffers_array_size,
                    interp_src->num_put_buffers + 1);

  struct put_buffers * put_buffers =
    interp_src->put_buffers + interp_src->num_put_buffers;
  interp_src->num_put_buffers++;

  put_buffers->buffer =
    malloc(interp_src->interp_data->mapping_on.src.num_send_msgs *
           sizeof(*(put_buffers->buffer)));
  put_buffers->requests =
    malloc(interp_src->interp_data->mapping_on.src.num_send_msgs *
           sizeof(*(put_buffers->requests)));
  put_buffers->num_open_requests = 0;
  put_buffers->collection_size = collection_size;

  for (unsigned i = 0;
       i < interp_src->interp_data->mapping_on.src.num_send_msgs; ++i)
    put_buffers->buffer[i] =
      malloc(collection_size *
             interp_src->interp_data->mapping_on.src.send_msgs[i].
             num_send_points * sizeof(*(put_buffers->buffer[i])));

  return put_buffers;
}

static void
field_interpolation_src_put_mapping_on_src(
  struct field_interpolation * interpolation, double **** data,
  int collection_size) {

   struct field_interpolation_src * interp_src =
      (struct field_interpolation_src *)interpolation;

   double **** extended_data; // data array extended by additional data from
                              // other source processes
                              // dimensions: [collection_idx]
                              //             [subdomain_idx]
                              //             [pointset_idx]
                              //             [local_idx]
   double *** extended_data_single;

   if (collection_size == 1) extended_data = &extended_data_single;
   else extended_data = malloc(collection_size * sizeof(*extended_data));

   extended_data[0] = malloc((interp_src->interp_data->num_subdomains + 1) *
                             collection_size * sizeof(**extended_data));
   extended_data[0][0] = malloc(interp_src->interp_data->num_pointsets *
                                collection_size * sizeof(***extended_data));
   extended_data[0][0][0] =
      malloc(interp_src->interp_data->total_num_halo_recv_points *
             collection_size * sizeof(****extended_data));
   {
      size_t offset = 0;
      for (unsigned j = 0; j < interp_src->interp_data->num_pointsets; ++j) {
         extended_data[0][0][j] = extended_data[0][0][0] + offset;
         offset +=
            interp_src->interp_data->num_halo_recv_points_per_points_set[j];
      }
   }

   for (int i = 1; i < collection_size; ++i) {
      extended_data[i] =
         extended_data[0] + i * (interp_src->interp_data->num_subdomains + 1);
      extended_data[i][0] =
         extended_data[0][0] + i * interp_src->interp_data->num_pointsets;
      size_t offset = i * interp_src->interp_data->total_num_halo_recv_points;
      for (unsigned j = 0; j < interp_src->interp_data->num_pointsets; ++j) {
         extended_data[i][0][j] = extended_data[0][0][0] + offset;
         offset +=
            interp_src->interp_data->num_halo_recv_points_per_points_set[j];
      }
   }

   for (int h = 0; h < collection_size; ++h)
       for (unsigned i = 0; i < interp_src->interp_data->num_subdomains; ++i)
           extended_data[h][i+1] = data[h][i];

   // do pre interpolation communication
   if ((interp_src->interp_data->num_halo_send_msgs > 0) ||
       (interp_src->interp_data->num_halo_recv_msgs > 0))
      exchange_halo_data(interp_src->interp_data,
                         interp_src->tags.halo_exchange_send,
                         interp_src->tags.halo_exchange_recv,
                         (void****)extended_data, collection_size,
                         COMM_DBLE);

   struct put_buffers * put_buffers =
      get_put_buffers(interp_src, collection_size);
   double ** restrict send_buffers = put_buffers->buffer;

   double * interpolated_data =
    malloc(interp_src->interp_data->total_num_interpolated_points *
           sizeof(*interpolated_data));

   for (int h = 0; h < collection_size; ++h) {

       // execute interpolation plan
       yac_execute_interpolation_plan_element(
          interp_src->interp_data->plans, extended_data[h], interpolated_data);

       //----------------
       // collect results
       //----------------
       for (unsigned i = 0;
            i < interp_src->interp_data->mapping_on.src.num_send_msgs; ++i) {

          if (interp_src->interp_data->mapping_on.src.send_msgs[i].
              num_send_points == 0)
             continue;

          unsigned const * send_points =
             interp_src->interp_data->mapping_on.src.send_msgs[i].send_points;
          unsigned num_points_to_send =
             interp_src->interp_data->mapping_on.src.send_msgs[i].num_send_points;

          for (unsigned j = 0; j < num_points_to_send; ++j)
             send_buffers[i][h*num_points_to_send+j] =
                interpolated_data[send_points[j]];
       }
   }
   free(interpolated_data);

   //-------------------
   // distribute results
   //-------------------

   unsigned num_open_requests = put_buffers->num_open_requests;
   struct comm_request ** requests = put_buffers->requests;

   for (unsigned i = 0;
        i < interp_src->interp_data->mapping_on.src.num_send_msgs; ++i) {

       if (interp_src->interp_data->mapping_on.src.send_msgs[i].
             num_send_points > 0) {

          if (num_open_requests > 64)
            num_open_requests =
              yac_comm_testsome(num_open_requests, requests);

          yac_comm_isend(
            send_buffers[i], collection_size *
            interp_src->interp_data->mapping_on.src.send_msgs[i].
              num_send_points, COMM_DBLE,
            interp_src->interp_data->mapping_on.src.send_msgs[i].rank,
            interp_src->tags.put, interp_src->interp_data->comm,
            requests + num_open_requests++);
       }
   }

   put_buffers->num_open_requests = num_open_requests;

   free(extended_data[0][0][0]);
   free(extended_data[0][0]);
   free(extended_data[0]);
   if (collection_size > 1) free(extended_data);
}

static void
field_interpolation_src_put_mapping_on_tgt(
  struct field_interpolation * interpolation, double **** data,
  int collection_size) {

  struct field_interpolation_src * interp_src =
    (struct field_interpolation_src *)interpolation;

  struct put_buffers * put_buffers =
    get_put_buffers(interp_src, collection_size);

  double ** restrict send_buffers = put_buffers->buffer;
  unsigned num_open_requests = put_buffers->num_open_requests;
  struct comm_request ** requests = put_buffers->requests;

  // send field data to target processes for the mapping
  for (unsigned i = 0;
       i < interp_src->interp_data->mapping_on.tgt.num_send_msgs; ++i) {

    if (interp_src->interp_data->mapping_on.tgt.send_msgs[i].num_points > 0) {

      for (unsigned j = 0, l = 0; j < (unsigned)collection_size; ++j)
        for (unsigned k = 0;
             k < interp_src->interp_data->mapping_on.tgt.send_msgs[i].num_points;
             ++k, ++l)
          send_buffers[i][l] =
            data[j]
                [interp_src->interp_data->mapping_on.tgt.send_msgs[i].send_points[k].subdomain_idx]
                [interp_src->interp_data->mapping_on.tgt.send_msgs[i].send_points[k].pointset_idx]
                [interp_src->interp_data->mapping_on.tgt.send_msgs[i].send_points[k].local_id];

      if (num_open_requests > 64)
        num_open_requests =
          yac_comm_testsome(num_open_requests, requests);

      yac_comm_isend(
        send_buffers[i], collection_size *
        interp_src->interp_data->mapping_on.tgt.send_msgs[i].num_points,
        COMM_DBLE, interp_src->interp_data->mapping_on.tgt.send_msgs[i].rank,
        interp_src->tags.put, interp_src->interp_data->comm,
        requests + num_open_requests++);
    }
  }
   put_buffers->num_open_requests = num_open_requests;
}

static void
field_interpolation_src_get(struct field_interpolation * interpolation,
                            double *** data, int collection_size) {

  UNUSED(interpolation);
  UNUSED(data);
  UNUSED(collection_size);

  yac_internal_abort_message(
    "ERROR: get operation is not defined for a source field interpolation\n",
    __FILE__, __LINE__);
}


static void
field_interpolation_gen_src_add_weight_file_name
  (field_interpolation_gen interpolation_gen, char const * weight_file_name) {

  struct field_interpolation_gen_src * interp_src =
    (struct field_interpolation_gen_src *)interpolation_gen;

  if (weight_file_name == NULL) return;

  unsigned idx = interp_src->num_weight_file_names;

  interp_src->num_weight_file_names++;

  ENSURE_ARRAY_SIZE(interp_src->weight_file_names,
                    interp_src->weight_file_names_array_size,
                    interp_src->num_weight_file_names);
  size_t str_size = strlen(weight_file_name)+1;
  interp_src->weight_file_names[idx] = malloc(str_size);
  memcpy(interp_src->weight_file_names[idx], weight_file_name, str_size);
}

static void
field_interpolation_data_src_delete(struct field_interpolation_data_src * data) {

  data->reference_counter--;

  if (!data->reference_counter) {

    struct interpolation_plan_element ** plan;
    plan = data->plans;
    while(*plan != NULL) {
      yac_free_interpolation_plan_element(*plan);
      free(*(plan++));
    }
    free(data->plans);

    for (unsigned i = 0; i < data->mapping_on.src.num_send_msgs; ++i)
       free(data->mapping_on.src.send_msgs[i].send_points);
    free(data->mapping_on.src.send_msgs);

    for (unsigned i = 0; i < data->num_halo_send_msgs; ++i)
       free(data->halo_send_msgs[i].send_points);
    free(data->halo_send_msgs);
    for (unsigned i = 0; i < data->num_halo_recv_msgs; ++i)
       free(data->halo_recv_msgs[i].recv_points);
    free(data->halo_recv_msgs);
    free(data->num_halo_recv_points_per_points_set);

    free(data);
  }
}

static void
field_interpolation_gen_src_delete
  (struct field_interpolation_gen * interpolation_gen) {

   struct field_interpolation_gen_src * interp_src;

   interp_src = (struct field_interpolation_gen_src *)interpolation_gen;

   field_interpolation_data_src_delete(interp_src->interp_data);
   for (unsigned i = 0; i < interp_src->num_weight_file_names; ++i)
     free(interp_src->weight_file_names[i]);
   free(interp_src->weight_file_names);
   free(interp_src->num_interpolated_points_per_method);
   free(interp_src->src_locations);
   free(interp_src->collected_global_ids);

   free(interpolation_gen);
}

static void
field_interpolation_src_delete(struct field_interpolation * interpolation) {

   struct field_interpolation_src * interp_src;

   interp_src = (struct field_interpolation_src *)interpolation;

   for (unsigned i = 0; i < interp_src->num_put_buffers; ++i) {

      yac_comm_waitall(interp_src->put_buffers[i].num_open_requests,
                       interp_src->put_buffers[i].requests);

      for (unsigned j = 0;
           j < interp_src->interp_data->mapping_on.src.num_send_msgs; ++j)
        free(interp_src->put_buffers[i].buffer[j]);
      free(interp_src->put_buffers[i].buffer);
      free(interp_src->put_buffers[i].requests);
   }
   free(interp_src->put_buffers);

   field_interpolation_data_src_delete(interp_src->interp_data);

   free(interpolation);
}

