/**
 * @file field_interpolation_gen_src_tgt.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdlib.h>
#include <string.h>

#include "field_interpolation_gen.h"
#include "field_interpolation_gen_src_tgt.h"
#include "field_interpolation_gen_src.h"
#include "field_interpolation_gen_tgt.h"

static void
field_interpolation_gen_src_tgt_enddef(struct field_interpolation_gen * interpolation_gen);
static struct field_interpolation *
field_interpolation_gen_src_tgt_get_field_interpolation
  (struct field_interpolation_gen * interpolation_gen);
static void 
field_interpolation_gen_src_tgt_add_weight_file_name
  (field_interpolation_gen interpolation_gen, char const * weight_file_name);
static void
field_interpolation_gen_src_tgt_delete(struct field_interpolation_gen * interpolation_gen);

static struct field_interpolation_gen_vtable field_interpolation_gen_src_tgt_vtable = {
                                       field_interpolation_gen_src_tgt_enddef,
                                       field_interpolation_gen_src_tgt_get_field_interpolation,
                                       field_interpolation_gen_src_tgt_add_weight_file_name,
                                       field_interpolation_gen_src_tgt_delete};

static void
field_interpolation_src_tgt_put(struct field_interpolation * interpolation_gen,
                                double **** data, int collection_size);
static void
field_interpolation_src_tgt_get(struct field_interpolation * interpolation_gen,
                                double *** data, int collection_size);
static void
field_interpolation_src_tgt_delete(struct field_interpolation * interpolation);

static struct field_interpolation_vtable field_interpolation_src_tgt_vtable = {
                                       field_interpolation_src_tgt_put,
                                       field_interpolation_src_tgt_get,
                                       field_interpolation_src_tgt_delete};

struct field_interpolation_gen_src_tgt {

   // basic field interpolation data
   struct field_interpolation_gen_vtable * vtable;

   // source+target field interpolation data
   struct field_interpolation_gen * inter_src, * inter_tgt;
};

struct field_interpolation_src_tgt {

   // basic field interpolation data
   struct field_interpolation_vtable * vtable;

   // source+target field interpolation data
   struct field_interpolation * inter_src, * inter_tgt;
};

struct field_interpolation_gen *
yac_new_field_interpolation_gen_src_tgt_mask(struct points *** points,
                                             struct mask *** masks,
                                             unsigned num_pointsets,
                                             char const * field_config_name,
                                             struct global_search * search,
                                             const char * other_field_config_name,
                                             struct interpolation_method ** methods,
                                             unsigned points_idx,
                                             enum yac_field_interpolation_tgt_mask_handling
                                             mask_handling,
                                             char const * weight_file_name,
                                             unsigned io_proc_step,
                                             int mapping_on_source) {

   char const * local_component_name = yac_get_local_component_name(search);
   char const * remote_component_name = yac_get_remote_component_name(search);
   char const * local_domain_name = yac_get_local_domain_name(search);
   char const * remote_domain_name = yac_get_remote_domain_name(search);
   int cmp_comp_names = strcmp(local_component_name, remote_component_name);
   int cmp_domain_names = strcmp(local_domain_name, remote_domain_name);
   int cmp_field_names = strcmp(field_config_name, other_field_config_name);

   int order;

   if (cmp_comp_names) order = cmp_comp_names > 0;
   else if (cmp_domain_names) order = cmp_domain_names > 0;
   else order = cmp_field_names > 0;

   struct field_interpolation_gen_src_tgt * inter_src_tgt;

   inter_src_tgt = malloc(1 * sizeof(*inter_src_tgt));
   inter_src_tgt->vtable = &field_interpolation_gen_src_tgt_vtable;

   for (int i = 0; i < 2; ++i) {
      if (i ^ order)
         inter_src_tgt->inter_src =
            yac_new_field_interpolation_gen_src(
               points, num_pointsets, field_config_name, search,
               other_field_config_name, methods, weight_file_name,
               io_proc_step, mapping_on_source);
      else
         inter_src_tgt->inter_tgt =
            yac_new_field_interpolation_gen_tgt_mask(
               points, masks, num_pointsets, field_config_name, search,
               other_field_config_name, points_idx, mask_handling,
               mapping_on_source);
   }

   return (struct field_interpolation_gen*)inter_src_tgt;
}

void yac_new_field_interpolation_gen_src_tgt_mask_2(
  struct points *** points_src, unsigned num_pointsets_src,
  char const * field_config_name_src, struct global_search * search_src,
  struct interpolation_method ** methods_src, char const * weight_file_name_src,
  unsigned io_proc_step_src, struct points *** points_tgt,
  struct mask *** masks_tgt, unsigned num_pointsets_tgt,
  char const * field_config_name_tgt, struct global_search * search_tgt,
  unsigned points_idx_tgt,
  enum yac_field_interpolation_tgt_mask_handling mask_handling_tgt,
  struct field_interpolation_gen ** interp_gen_src,
  struct field_interpolation_gen ** interp_gen_tgt, int mapping_on_source) {

  struct communicator * comm = yac_get_communicator(search_src);
  char const * local_component_name = yac_get_local_component_name(search_src);
  char const * remote_component_name = yac_get_remote_component_name(search_src);
  char const * local_domain_name = yac_get_local_domain_name(search_src);
  char const * remote_domain_name = yac_get_remote_domain_name(search_src);
  struct field_interpolation_tags tags =
     yac_field_interpolation_generate_tag(
        comm, local_component_name, remote_component_name,
        local_domain_name, remote_domain_name,
        field_config_name_src, field_config_name_tgt);

  *interp_gen_src =
     yac_new_field_interpolation_gen_src_(
        points_src, num_pointsets_src, field_config_name_src, search_src,
        field_config_name_tgt, methods_src, weight_file_name_src,
        io_proc_step_src, tags, mapping_on_source);
  *interp_gen_tgt =
     yac_new_field_interpolation_gen_tgt_mask_(
        points_tgt, masks_tgt, num_pointsets_tgt, field_config_name_tgt,
        search_tgt, field_config_name_src, points_idx_tgt,
        mask_handling_tgt, tags, mapping_on_source);
}

struct field_interpolation_gen *
yac_new_field_interpolation_gen_src_tgt(struct points *** points,
                                        unsigned num_pointsets,
                                        char const * field_config_name,
                                        struct global_search * search,
                                        const char * other_field_config_name,
                                        struct interpolation_method ** methods,
                                        unsigned points_idx,
                                        char const * weight_file_name,
                                        unsigned io_proc_step,
                                        int mapping_on_source) {

   char const * local_component_name = yac_get_local_component_name(search);
   char const * remote_component_name = yac_get_remote_component_name(search);
   char const * local_domain_name = yac_get_local_domain_name(search);
   char const * remote_domain_name = yac_get_remote_domain_name(search);
   int cmp_comp_names = strcmp(local_component_name, remote_component_name);
   int cmp_domain_names = strcmp(local_domain_name, remote_domain_name);
   int cmp_field_names = strcmp(field_config_name, other_field_config_name);

   int order;

   if (cmp_comp_names) order = cmp_comp_names > 0;
   else if (cmp_domain_names) order = cmp_domain_names > 0;
   else order = cmp_field_names > 0;

   struct field_interpolation_gen_src_tgt * inter_src_tgt =
      malloc(1 * sizeof(*inter_src_tgt));

   inter_src_tgt->vtable = &field_interpolation_gen_src_tgt_vtable;

   for (int i = 0; i < 2; ++i) {
      if (i ^ order)
         inter_src_tgt->inter_src =
            yac_new_field_interpolation_gen_src(
               points, num_pointsets, field_config_name, search,
               other_field_config_name, methods, weight_file_name,
               io_proc_step, mapping_on_source);
      else
         inter_src_tgt->inter_tgt =
            yac_new_field_interpolation_gen_tgt(
               points, num_pointsets, field_config_name, search,
               other_field_config_name, points_idx, mapping_on_source);
   }

   return (struct field_interpolation_gen*)inter_src_tgt;
}

void yac_new_field_interpolation_gen_src_tgt_2(
  struct points *** points_src, unsigned num_pointsets_src,
  char const * field_config_name_src, struct global_search * search_src,
  struct interpolation_method ** methods_src, char const * weight_file_name_src,
  unsigned io_proc_step_src, struct points *** points_tgt,
  unsigned num_pointsets_tgt, char const * field_config_name_tgt,
  struct global_search * search_tgt, unsigned points_idx_tgt,
  struct field_interpolation_gen ** interp_gen_src,
  struct field_interpolation_gen ** interp_gen_tgt, int mapping_on_source) {

  yac_new_field_interpolation_gen_src_tgt_mask_2(
    points_src, num_pointsets_src, field_config_name_src, search_src,
    methods_src, weight_file_name_src, io_proc_step_src, points_tgt, NULL,
    num_pointsets_tgt, field_config_name_tgt, search_tgt, points_idx_tgt,
    TGT_IGNORE_MASK, interp_gen_src, interp_gen_tgt, mapping_on_source);
}

static struct field_interpolation *
field_interpolation_gen_src_tgt_get_field_interpolation
  (struct field_interpolation_gen * interpolation_gen) {

   struct field_interpolation_gen_src_tgt * interp_gen_src_tgt;
   struct field_interpolation_src_tgt * interp_src_tgt;

   interp_gen_src_tgt = (struct field_interpolation_gen_src_tgt *)interpolation_gen;
   interp_src_tgt = malloc(1 * sizeof(*interp_src_tgt));

   interp_src_tgt->vtable = &field_interpolation_src_tgt_vtable;
   interp_src_tgt->inter_src =
     yac_get_field_interpolation(interp_gen_src_tgt->inter_src);
   interp_src_tgt->inter_tgt =
     yac_get_field_interpolation(interp_gen_src_tgt->inter_tgt);

   return (struct field_interpolation*)interp_src_tgt;
}

static void 
field_interpolation_gen_src_tgt_add_weight_file_name
  (field_interpolation_gen interpolation_gen, char const * weight_file_name) {

  struct field_interpolation_gen_src_tgt * interp_gen_src_tgt =
    (struct field_interpolation_gen_src_tgt *)interpolation_gen;

  yac_field_interpolation_gen_add_weight_file_name(
    interp_gen_src_tgt->inter_src, weight_file_name);
}

static void
field_interpolation_gen_src_tgt_enddef(struct field_interpolation_gen * interpolation_gen) {

   struct field_interpolation_gen_src_tgt * inter_src_tgt;

   inter_src_tgt = (struct field_interpolation_gen_src_tgt *)interpolation_gen;

   inter_src_tgt->inter_src->vtable->enddef(inter_src_tgt->inter_src);
   inter_src_tgt->inter_tgt->vtable->enddef(inter_src_tgt->inter_tgt);
}

static void
field_interpolation_src_tgt_put(struct field_interpolation * interpolation,
                                double **** data, int collection_size) {

   struct field_interpolation_src_tgt * inter_src_tgt;

   inter_src_tgt = (struct field_interpolation_src_tgt *)interpolation;

   inter_src_tgt->inter_src->vtable->put(inter_src_tgt->inter_src, data, collection_size);
}

static void
field_interpolation_src_tgt_get(struct field_interpolation * interpolation,
                                double *** data, int collection_size) {

   struct field_interpolation_src_tgt * inter_src_tgt;

   inter_src_tgt = (struct field_interpolation_src_tgt *)interpolation;

   inter_src_tgt->inter_tgt->vtable->get(inter_src_tgt->inter_tgt, data, collection_size);
}

static void
field_interpolation_gen_src_tgt_delete(struct field_interpolation_gen * interpolation_gen) {

   struct field_interpolation_gen_src_tgt * inter_src_tgt;

   inter_src_tgt = (struct field_interpolation_gen_src_tgt *)interpolation_gen;

   yac_delete_field_interpolation_gen(inter_src_tgt->inter_src);
   yac_delete_field_interpolation_gen(inter_src_tgt->inter_tgt);
   free(interpolation_gen);
}

static void
field_interpolation_src_tgt_delete(struct field_interpolation * interpolation) {

   struct field_interpolation_src_tgt * inter_src_tgt;

   inter_src_tgt = (struct field_interpolation_src_tgt *)interpolation;

   yac_delete_field_interpolation(inter_src_tgt->inter_src);
   yac_delete_field_interpolation(inter_src_tgt->inter_tgt);
   free(interpolation);
}

