/**
 * @file global_search.h
 * @brief Structs and interfaces for the global search between two global domains
 *
 * The global search takes into acount information of adjacent domains which
 * reside on a different process. In this sense it considers the information
 * of the global grid (in other words: the complete compute domain) rather
 * than just the information located on the local process.  
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#ifndef GLOBAL_SEARCH_H
#define GLOBAL_SEARCH_H

/** \example test_global_search.c
 * This contains examples on how to use global search.
 */

#include "domain.h"
#include "communicator.h"

//! search results for a target cell
struct source_search_results {

  int global_id;              //!< global id of the target cell
  int tgt_rank;               //!< rank of process, which is the owner of the target
                              //!< cell (there may be other target processes that
                              //!< have the same cell but are not mentioned here)
  struct domain * tgt_domain; //!< local copy of the remote target domain
  unsigned local_tgt_cell_id; //!< local cell id of the target cell in the local copy
                              //!< of the remote target domain
  unsigned tgt_subdomain_idx; //!< subdomain id in the local copy of the remote
                              //!< target domain

  unsigned ** overlapping_src_cells;                  //!< local ids of overlapping
                                                      //!< source cells
                                                      //!< (dimensions [subdomain id][src cell])
  unsigned * num_overlapping_src_cells_per_subdomain; //!< number of cell of each local
                                                      //!< source subdomain that overlap
                                                      //!< with the target cell(dimensions
                                                      //!< [subdomain id])

  unsigned * other_src_ranks;   //!< ranks of other source processes that
                                //!< have an overlap with this target cell
  unsigned num_other_src_ranks; //!< number of other source processes that
                                //!< have an overlap with this target cell
};

/**
 * generates and starts the initialisation of a global search object for two global
 * domains \n the initialisation is asynchronous, a call to \ref yac_enddef_global_search
 * is necessary to ensure that the global search is finished
 * @param[in] local_domain %domain of the local process
 * @param[in] remote_domain_name name of the remote %domain
 * @param[in] comm %communicator containing the all processes of both global domains
 * @param[in] local_domain_is_source specifies whether the local %domain acts as a source
 * @param[in] local_domain_is_target specifies whether the local %domain acts as a target
 * @param[in] local_component_name name of the local component
 * @param[in] remote_component_name name of the remote component
 * @return global search object
 * @see enddef_global_search
 *
 * \remark it is valid to specify that the local %domain is both source an target
 * \remark the values for local_domain_is_source and local_domain_is_target must match with the values provided on the remote %domain
 */
struct global_search * yac_global_search_new(struct domain * local_domain,
                                             char const * remote_domain_name,
                                             struct communicator * comm,
                                             unsigned local_domain_is_source,
                                             unsigned local_domain_is_target,
                                             char const * local_component_name,
                                             char const * remote_component_name);

/**
 * generates and starts the initialisation of global search objects for two
 * locally available global domains \n
 * the initialisation is asynchronous, a call to \ref yac_enddef_global_search
 * is necessary to ensure that the global search is finished
 * @param[in]  domains          pointers to both domains
 * @param[in]  comm             %communicator containing the all processes of
 *                              both global domains
 * @param[in]  domain_is_source specifies whether the domains acts as a sources
 * @param[in]  component_names  names of the components of both domains
 * @param[out] searches         global search objects
 * @see enddef_global_search
 *
 * \remark it is valid to specify that both domains are source
 * \remark the values for domain_is_source must match with the values provided
 *         on the other processes
 */
void yac_global_search_all_local_new(struct domain * domains[2],
                                     struct communicator * comm,
                                     unsigned domain_is_source[2],
                                     char const * component_names[2],
                                     struct global_search * searches[2]);

/**
 * blocks until the global search is finished
 * @param[in] search
 */
void yac_enddef_global_search(struct global_search * search);

/**
 * returns the grid search object for a given %subdomain (only available on domains
 * that are sources)
 * @param[in] search
 * @param[in] subdomain_idx index of the respective %subdomain
 * @return grid search objext for the grid of the specified %subdomain
 */
struct grid_search * yac_get_subdomain_grid_search(struct global_search * search, unsigned subdomain_idx);

/**
 * determines whether the local source process has an overlap with a potential
 * target process
 * @param[in] search
 * @param[in] rank   rank of the potential target process
 * @return 0 if the other process is no target or has no overlap with the
 *         local process
 */
int yac_source_process_has_overlap_with(struct global_search * search,
                                        int rank);

/**
 * determines whether the local target process has an overlap with a potential
 * source process
 * @param[in] search
 * @param[in] rank   rank of the potential source process
 * @return 0 if the other process is no source or has no overlap with the
 *         local process
 */
int yac_target_process_has_overlap_with(struct global_search * search,
                                        int rank);

/**
 * returns a dependency list that contains for each local subdomain the local cell
 * ids that overlap with a source domain on the given rank (this routine is only
 * valid for target domains)
 * @param[in] search
 * @param[in] rank   rank of a process
 * @return           dependency list containing the overlapping local cell
 *                   ids of each subdomain \n
 *                   NULL if there is no overlap with the given process
 */
struct dep_list * yac_get_matching_local_cell_ids(
   struct global_search * search, unsigned rank);

/**
 * returns the search results for a given target cell
 * @param[in] search
 * @param[in] global_cell_id global id of target cell
 * @returns search results for the given target cell\n
 *          NULL if there matching local source cells were found
 */
struct source_search_results const * yac_get_search_results_for_tgt_cell (
   struct global_search * search, int global_cell_id);

/**
 * returns the search results for a given target cell
 * @param[in] search
 * @param[in] global_cell_ids global ids of target cells
 * @param[in] num_global_cell_ids number of ids in global_cell_ids
 * @param[in,out] results search results for the given target cells\n
 *                NULL for target cells, for which no matching local
 *                source cells were found
 * @remark entries in global_cell_ids have to be sorted in ascending order
 * @remark the user needs to provide for results an array that is big enough to
 *         hold at least num_global_cell_ids entries
 */
void yac_get_search_results_for_tgt_cells (
   struct global_search * search, int * global_cell_ids,
   unsigned num_global_cell_ids,
   struct source_search_results const ** results);

/**
 * gets the search %communicator
 * @param[in] search
 * @return %communicator associated to search
 */
struct communicator * yac_get_communicator(struct global_search * search);

/**
 * gets the lowest rank of a process in the search %communicator that is part of
 * the same global %domain as the local process
 * @param[in]  search
 * @param[out] local_comm_rank matching rank in the domain comm of the rank
 *                                      returned by this routine
 * @return lowest rank of a process in the search %communicator that is part of
 *         the same global %domain as the local process
 * @remark local_comm_rank can be NULL
 */
unsigned yac_get_lowest_local_domain_rank(struct global_search * search,
                                          unsigned * local_comm_rank);

/**
 * gets the lowest rank of a process in the search %communicator that is part of
 * the remote global %domain
 * @param[in] search
 * @return lowest rank of a process in the search %communicator that is part of
 *         the remote global %domain
 */
unsigned yac_get_lowest_remote_domain_rank(struct global_search * search);

/**
 * gets number of processes in the remote %domain within the search
 * %communicator
 * @param[in] search
 * @return number of processes in the remote domain
 */
unsigned yac_get_num_remote_domain_ranks(struct global_search * search);

/**
 * gets ranks of processes of the remote %domain within the search
 * %communicator
 * @param[in] search
 * @return ranks of the remote domain processes
 */
unsigned const * yac_get_remote_domain_ranks(struct global_search * search);

/**
 * get the name of the local domain
 * @param[in] search
 * @return name of the local domain
 */
char const * yac_get_local_domain_name(struct global_search * search);

/**
 * returns the local domain
 * @param[in] search
 * @return local domain
 */
struct domain * yac_get_local_domain(struct global_search * search);

/**
 * get the name of the remote domain
 * @param[in] search
 * @return name of the remote domain
 */
char const * yac_get_remote_domain_name(struct global_search * search);

/**
 * after enddef_global_search was called for the given global search object,
 * this routine returns a remote domain that for a given rank.
 * @param[in] search
 * @param[in] remote_domain_rank rank of the remote process
 * @return copy of domain from remote process\n
 *         NULL if the remote domain does not overlap with the local one
 * @remarks the returned domain may only contain the cells of the remote domain
 *          that overlap with the local data
 */
struct domain * yac_get_remote_domain(struct global_search * search,
                                      unsigned remote_domain_rank);

/**
 * get the name of the local component
 * @param[in] search
 * @return name of the local component
 */
char const * yac_get_local_component_name(struct global_search * search);

/**
 * get the name of the remote component
 * @param[in] search
 * @return name of the remote component
 */
char const * yac_get_remote_component_name(struct global_search * search);

/**
 * frees all memory associated to the global search object
 * @param[in] search
 */
void yac_global_search_delete(struct global_search * search);

#endif // GLOBAL_SEARCH_H

