/**
 * @file interpolation_method_check.c
 *
 * @copyright Copyright  (C)  2014 Moritz Hanke <hanke@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "interpolation_method.h"
#include "interpolation_method_check.h"
#include "interpolation_plan.h"
#include "communicator.h"
#include "utils.h"
#include "ensure_array_size.h"

static void
do_search_interpolation_method_check(struct interpolation_method ** method,
                                     struct interpolation_plan_element ** plan_stack,
                                     int * global_ids, double (*coordinates_xyz)[3],
                                     enum yac_location tgt_location,
                                     unsigned num_points,
                                     unsigned * num_interpolated_points,
                                     struct interpolation_method_data_request ** data_request);
static void 
enddef_interpolation_method_check(struct interpolation_method ** method);
static void 
get_source_exchange_data_interpolation_method_check(
   struct interpolation_method ** method,
   struct interpolation_method_data_request *** data_request);
static void delete_interpolation_method_check(struct interpolation_method * method);
static struct global_search *
get_global_search(struct interpolation_method * method);
static void
set_finish_callback(struct interpolation_method * method,
                    void (*callback)(void * user_data),
                    void * user_data);

static struct interpolation_method_vtable
   interpolation_method_check_vtable = {do_search_interpolation_method_check,
                                        enddef_interpolation_method_check,
                                        get_source_exchange_data_interpolation_method_check,
                                        delete_interpolation_method_check,
                                        get_global_search,
                                        set_finish_callback};

struct interpolation_method_check {

   struct interpolation_method_vtable * vtable;
   func_do_search do_search_callback;
   void * do_search_user_data;
   struct global_search * search;
   void (*finish_callback)(void * user_data);
   void * finish_user_data;
   struct interpolation_method_data_request ** data_request;
};

struct interpolation_method *
yac_new_interpolation_method_check(func_constructor constructor_callback,
                               void * constructor_user_data,
                               func_do_search do_search_callback,
                               void * do_search_user_data,
                               struct global_search * search) {

   struct interpolation_method_check * method;

   method = malloc(1 * sizeof(*method));

   method->vtable = &interpolation_method_check_vtable;
   method->do_search_callback = do_search_callback;
   method->do_search_user_data = do_search_user_data;
   method->search = search;
   method->finish_callback = NULL;
   method->finish_user_data = NULL;
   method->data_request = NULL;

   if (constructor_callback != NULL)
      constructor_callback(constructor_user_data);

   yac_init_interpolation_method((struct interpolation_method *)method, search);

   return (struct interpolation_method *)method;
}

static void
do_search_interpolation_method_check(struct interpolation_method ** method,
                                     struct interpolation_plan_element ** plan_stack,
                                     int * global_ids, double (*coordinates_xyz)[3],
                                     enum yac_location tgt_location,
                                     unsigned num_points,
                                     unsigned * num_interpolated_points,
                                     struct interpolation_method_data_request ** data_request) {

   struct interpolation_method_check * method_check;

   method_check = (struct interpolation_method_check *)(method[0]);
   method_check->data_request = data_request;

   if (method_check->do_search_callback != NULL)
      method_check->do_search_callback(
         global_ids, (const double (*)[3])coordinates_xyz, num_points,
         method_check->do_search_user_data);

   // we can process all points
   *num_interpolated_points = 0;

   if (method[1] != NULL)
      method[1]->vtable->do_search(method+1, plan_stack, global_ids,
                                   coordinates_xyz, tgt_location, num_points,
                                   num_interpolated_points+1, data_request+1);

   if (method_check->finish_callback != NULL)
      method_check->finish_callback(method_check->finish_user_data);

   return;
}

static void
enddef_interpolation_method_check(struct interpolation_method ** method) {

   // struct interpolation_method_check * method_check;

   // method_check = (struct interpolation_method_check *)(method[0]);

   if (method[1] != NULL)
      method[1]->vtable->enddef(method + 1);
}

static void
get_source_exchange_data_interpolation_method_check(
   struct interpolation_method ** method,
   struct interpolation_method_data_request *** data_request) {

   struct interpolation_method_check * method_check;

   method_check = (struct interpolation_method_check *)(method[0]);

   *data_request = method_check->data_request;
   method_check->data_request = NULL;
}

void delete_interpolation_method_check(struct interpolation_method * method) {

   free(method);
}

static struct global_search *
get_global_search(struct interpolation_method * method) {

   return ((struct interpolation_method_check*)(method))->search;
}

static void
set_finish_callback(struct interpolation_method * method,
                     void (*callback)(void * user_data),
                     void * user_data) {

   struct interpolation_method_check * method_check;

   method_check = (struct interpolation_method_check*)method;

   method_check->finish_callback = callback;
   method_check->finish_user_data = user_data;
}

enum callback_type {
      CONSTRUCTOR,
      DO_SEARCH,
};
typedef void (*func_dummy)(void);
static struct {
  struct {
    func_dummy callback;
    void * user_data;
  } value;
  enum callback_type type;
  char * key;
} * callback_lookup_table = NULL;
static unsigned callback_lookup_table_array_size = 0;
static unsigned callback_lookup_table_size = 0;

static void interpolation_method_check_add_callback(
  func_dummy callback, enum callback_type type, void * user_data,
  char const * key) {

  for (unsigned i = 0; i < callback_lookup_table_size; ++i) {
    if ((!strcmp(callback_lookup_table[i].key, key)) &&
        (callback_lookup_table[i].type == type)) {
      if ((callback_lookup_table[i].value.callback != callback) ||
          (callback_lookup_table[i].value.user_data != user_data))
        yac_internal_abort_message("ERROR(interpolation_method_check_add_do_search_callback)"
                                   "identical key has been set before with different "
                                   "value(s)\n", __FILE__, __LINE__);
      return;
    }
  }

  ENSURE_ARRAY_SIZE(callback_lookup_table, callback_lookup_table_array_size,
                    callback_lookup_table_size + 1);

  callback_lookup_table[callback_lookup_table_size].key =
    malloc((strlen(key)+1) * sizeof(*key));
  strcpy(callback_lookup_table[callback_lookup_table_size].key, key);
  callback_lookup_table[callback_lookup_table_size].type = type;
  callback_lookup_table[callback_lookup_table_size].value.callback = callback;
  callback_lookup_table[callback_lookup_table_size].value.user_data = user_data;
  callback_lookup_table_size++;
}

static void interpolation_method_get_callback(char const * key,
                                              func_dummy * callback,
                                              enum callback_type type, 
                                              void ** user_data) {

  for (unsigned i = 0; i < callback_lookup_table_size; ++i) {
    if ((!strcmp(callback_lookup_table[i].key, key)) &&
        (callback_lookup_table[i].type == type)) {
      *callback = callback_lookup_table[i].value.callback;
      *user_data = callback_lookup_table[i].value.user_data;
      return;
    }
  }
  *callback = NULL;
  *user_data = NULL;
}

void yac_interpolation_method_check_add_constructor_callback(
  func_constructor constructor_callback, void * user_data, char const * key) {

  interpolation_method_check_add_callback((func_dummy)constructor_callback,
                                          CONSTRUCTOR, user_data, key);
}

void yac_interpolation_method_get_constructor_callback(
  char const * key, func_constructor * constructor_callback, void ** user_data) {

  interpolation_method_get_callback(key, (func_dummy*)constructor_callback,
                                    CONSTRUCTOR, user_data);
}

void yac_interpolation_method_check_add_do_search_callback(
  func_do_search do_search_callback, void * user_data, char const * key) {

  interpolation_method_check_add_callback((func_dummy)do_search_callback,
                                          DO_SEARCH, user_data, key);
}

void yac_interpolation_method_get_do_search_callback(
  char const * key, func_do_search * do_search_callback, void ** user_data) {

  interpolation_method_get_callback(key, (func_dummy*)do_search_callback,
                                    DO_SEARCH, user_data);
}

