/**
 * @file interpolation_method_file.c
 *
 * @copyright Copyright  (C)  2015 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_C_NETCDF
#include <netcdf.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "interpolation_method.h"
#include "interpolation_method_file.h"
#include "interpolation_plan.h"
#include "grid_search.h"
#include "global_search.h"
#include "ensure_array_size.h"
#include "utils.h"

static void
do_search_interpolation_method_file(struct interpolation_method ** method,
                                    struct interpolation_plan_element ** plan_stack,
                                    int * global_ids, double (*coordinates_xyz)[3],
                                    enum yac_location tgt_location,
                                    unsigned num_points,
                                    unsigned * num_interpolated_points,
                                    struct interpolation_method_data_request ** data_request);
static void
enddef_interpolation_method_file(struct interpolation_method ** method);
static void
get_source_exchange_data_interpolation_method_file(
   struct interpolation_method ** method,
   struct interpolation_method_data_request *** data_request);
static void
delete_interpolation_method_file(struct interpolation_method * method);
static struct global_search *
get_global_search(struct interpolation_method * method);
static void
set_finish_callback(struct interpolation_method * method,
                    void (*callback)(void * user_data),
                    void * user_data);

static struct interpolation_method_vtable
   interpolation_method_file_vtable = {do_search_interpolation_method_file,
                                       enddef_interpolation_method_file,
                                       get_source_exchange_data_interpolation_method_file,
                                       delete_interpolation_method_file,
                                       get_global_search,
                                       set_finish_callback};

struct interpolation_method_file_tags {

  unsigned tgt_indices_alltoall;
  unsigned src_indices_alltoall_cell;
  unsigned src_indices_alltoall_corner;
  unsigned src_indices_alltoall_edge;
  unsigned tgt_location_bcast;
  unsigned link_data_alltoall_1;
  unsigned link_data_alltoall_2;
  unsigned link_results_alltoall;
  unsigned fixed_data_alltoall;
  unsigned fixed_data_fixed_value_bcast;
  unsigned fixed_data_num_fixed_bcast;
  unsigned fixed_results_alltoall;
};

struct link_data {
  int src_idx, tgt_idx;
  unsigned rank, points_idx;
  double weight;
};

struct link_data_reorder {
  int src_idx;
  unsigned points_idx;
  unsigned orig_pos;
};

struct dist_link_data {
  struct link_data * links;
  unsigned count;
  enum yac_location tgt_location;
};

struct dist_fixed_data {
  unsigned num_fixed_values;
  double * fixed_values;
  unsigned * num_tgt_indices_per_fixed_value;
  int ** global_tgt_indices_per_fixed_value;
};

struct index_data {
  int index;
  unsigned rank;
};

struct dist_index_data {
  struct index_data * data;
  unsigned count;
  enum yac_location location;
};

struct intermediate_method_file_data_link {

  struct interpolation_method_file * method_file;
  struct communicator * comm;
  struct dist_link_data dist_link_data;
  struct dist_index_data * dist_src_indices;
  unsigned num_locations;
};

struct intermediate_method_file_data_fixed {

  struct interpolation_method_file * method_file;
  struct communicator * comm;
  unsigned num_fixed_values;
  double * fixed_values;
};

struct intermediate_results {

  struct interpolation_method_file * method_file;

  struct link_data * links;
  unsigned num_links;
  unsigned * link_flags;

  int * global_tgt_ids;
  unsigned num_fixed_tgt;
  unsigned * fixed_flags;
  double * fixed_values;
  unsigned num_fixed_values;
};

struct interpolation_method_file {

  struct interpolation_method_vtable * vtable;
  void (*finish_callback)(void * user_data);
  void * finish_user_data;
  struct global_search * search;
  struct domain * domain;
  enum yac_location * src_locations;
  unsigned num_pointsets;
  struct interpolation_method_file_tags tags;

  struct interpolation_method ** method_stack;
  struct interpolation_plan_element ** plan_stack;
  int * global_tgt_ids;
  double (*coordinates_xyz)[3];
  unsigned num_global_tgt_ids;
  unsigned * num_interpolated_points;
  struct interpolation_method_data_request ** data_request;

  struct dist_index_data dist_tgt_indices;
  struct dist_link_data dist_link_data;
  struct dist_fixed_data dist_fixed_data;

  unsigned io_proc_step;
};

struct tgt_reorder_data {
  unsigned rank;
  int index;
  double coord[3];
};

static void generate_tag_string(char * buf, size_t buffer_size,
                                char const * tag_base_str,
                                char const * source_comp_name,
                                char const * target_comp_name,
                                char const * source_domain_name,
                                char const * target_domain_name,
                                char const * file_name,
                                char const * stack_name,
                                int unique_interpolation_method_id,
                                char const * tag_str) {

   int ret_val;
   ret_val = snprintf(buf, buffer_size, "%s%s%s%s%s%s%s%d%s", tag_base_str,
                      source_comp_name, target_comp_name, source_domain_name,
                      target_domain_name, file_name, stack_name,
                      unique_interpolation_method_id, tag_str);

   if (ret_val < 0 || (size_t)ret_val >= buffer_size)
      yac_internal_abort_message("ERROR(generate_tag_string): truncation",
                                 __FILE__, __LINE__);
}

static void
generate_interpolation_method_file_tag(
  struct interpolation_method_file * method, char const * stack_name,
  char const * file_name, struct interpolation_method_file_tags * tags) {

   char const * str_base_str                     = "interpolation method file";
   char const * str_generator_string             = "generator string";
   char const * str_tgt_indices_alltoall         = "target indices alltoall";
   char const * str_src_indices_alltoall_cell    = "source indices alltoall cell";
   char const * str_src_indices_alltoall_corner  = "source indices alltoall corner";
   char const * str_src_indices_alltoall_edge    = "source indices alltoall edge";
   char const * str_tgt_location_bcast           = "target location bcast";
   char const * str_link_data_alltoall_1         = "link data alltoall 1";
   char const * str_link_data_alltoall_2         = "link_data alltoall 2";
   char const * str_link_results_alltoall        = "link results alltoall";
   char const * str_fixed_data_alltoall          = "fixed data alltoall";
   char const * str_fixed_data_fixed_value_bcast = "fixed values bcase";
   char const * str_fixed_data_num_fixed_bcast   = "num fixed bcast";
   char const * str_fixed_results_alltoall       = "fixed results alltoall";

   struct global_search * search;

   search = get_global_search((struct interpolation_method *)method);

   struct communicator * comm;

   comm = yac_get_domain_comm(yac_get_local_domain(search));

   char const * source_comp_name = yac_get_local_component_name(search);
   char const * target_comp_name = yac_get_remote_component_name(search);

   char const * source_domain_name = yac_get_local_domain_name(search);
   char const * target_domain_name = yac_get_remote_domain_name(search);

   unsigned max_tag_strlen = strlen(str_src_indices_alltoall_corner);

   unsigned max_str_buf_size = max_tag_strlen;

   max_str_buf_size += strlen(str_base_str);
   max_str_buf_size += strlen(source_comp_name);
   max_str_buf_size += strlen(target_comp_name);
   max_str_buf_size += strlen(source_domain_name);
   max_str_buf_size += strlen(target_domain_name);
   max_str_buf_size += strlen(file_name);
   max_str_buf_size += strlen(stack_name) + 10 + 10 + 1; // (+ 10) for points_idx
                                                         // (+ 10) for unique_method_id
                                                         // (+ 1)  for '\0'

   char * unique_generator_string = malloc(max_str_buf_size);
   generate_tag_string(unique_generator_string, max_str_buf_size, str_base_str,
                       source_comp_name, target_comp_name, source_domain_name,
                       target_domain_name, file_name, stack_name,
                       yac_get_unique_id_interpolation_method(
                        (struct interpolation_method *)method),
                       str_generator_string);

   char * tag_strings[12];
   tag_strings[0] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[0], max_str_buf_size, str_base_str,
                       source_comp_name, target_comp_name, source_domain_name,
                       target_domain_name, file_name, stack_name,
                       yac_get_unique_id_interpolation_method(
                        (struct interpolation_method *)method),
                       str_tgt_indices_alltoall);

   tag_strings[1] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[1], max_str_buf_size, str_base_str,
                       source_comp_name, target_comp_name, source_domain_name,
                       target_domain_name, file_name, stack_name,
                       yac_get_unique_id_interpolation_method(
                        (struct interpolation_method *)method),
                       str_src_indices_alltoall_cell);

   tag_strings[2] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[2], max_str_buf_size, str_base_str,
                       source_comp_name, target_comp_name, source_domain_name,
                       target_domain_name, file_name, stack_name,
                       yac_get_unique_id_interpolation_method(
                        (struct interpolation_method *)method),
                       str_src_indices_alltoall_corner);

   tag_strings[3] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[3], max_str_buf_size, str_base_str,
                       source_comp_name, target_comp_name, source_domain_name,
                       target_domain_name, file_name, stack_name,
                       yac_get_unique_id_interpolation_method(
                        (struct interpolation_method *)method),
                       str_src_indices_alltoall_edge);

   tag_strings[4] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[4], max_str_buf_size, str_base_str,
                       source_comp_name, target_comp_name, source_domain_name,
                       target_domain_name, file_name, stack_name,
                       yac_get_unique_id_interpolation_method(
                        (struct interpolation_method *)method),
                       str_tgt_location_bcast);

   tag_strings[5] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[5], max_str_buf_size, str_base_str,
                       source_comp_name, target_comp_name, source_domain_name,
                       target_domain_name, file_name, stack_name,
                       yac_get_unique_id_interpolation_method(
                        (struct interpolation_method *)method),
                       str_link_data_alltoall_1);

   tag_strings[6] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[6], max_str_buf_size, str_base_str,
                       source_comp_name, target_comp_name, source_domain_name,
                       target_domain_name, file_name, stack_name,
                       yac_get_unique_id_interpolation_method(
                        (struct interpolation_method *)method),
                       str_link_data_alltoall_2);

   tag_strings[7] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[7], max_str_buf_size, str_base_str,
                       source_comp_name, target_comp_name, source_domain_name,
                       target_domain_name, file_name, stack_name,
                       yac_get_unique_id_interpolation_method(
                        (struct interpolation_method *)method),
                       str_link_results_alltoall);

   tag_strings[8] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[8], max_str_buf_size, str_base_str,
                       source_comp_name, target_comp_name, source_domain_name,
                       target_domain_name, file_name, stack_name,
                       yac_get_unique_id_interpolation_method(
                        (struct interpolation_method *)method),
                       str_fixed_data_alltoall);

   tag_strings[9] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[9], max_str_buf_size, str_base_str,
                       source_comp_name, target_comp_name, source_domain_name,
                       target_domain_name, file_name, stack_name,
                       yac_get_unique_id_interpolation_method(
                        (struct interpolation_method *)method),
                       str_fixed_data_fixed_value_bcast);

   tag_strings[10] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[10], max_str_buf_size, str_base_str,
                       source_comp_name, target_comp_name, source_domain_name,
                       target_domain_name, file_name, stack_name,
                       yac_get_unique_id_interpolation_method(
                        (struct interpolation_method *)method),
                       str_fixed_data_num_fixed_bcast);

   tag_strings[11] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[11], max_str_buf_size, str_base_str,
                       source_comp_name, target_comp_name, source_domain_name,
                       target_domain_name, file_name, stack_name,
                       yac_get_unique_id_interpolation_method(
                        (struct interpolation_method *)method),
                       str_fixed_results_alltoall);

   unsigned tag_values[12];

   yac_get_unique_comm_tags(
      comm, unique_generator_string, (char const **)tag_strings, 12, tag_values);

   free(unique_generator_string);

   tags->tgt_indices_alltoall = tag_values[0];
   tags->src_indices_alltoall_cell = tag_values[1];
   tags->src_indices_alltoall_corner = tag_values[2];
   tags->src_indices_alltoall_edge = tag_values[3];
   tags->tgt_location_bcast = tag_values[4];
   tags->link_data_alltoall_1 = tag_values[5];
   tags->link_data_alltoall_2 = tag_values[6];
   tags->link_results_alltoall = tag_values[7];
   tags->fixed_data_alltoall = tag_values[8];
   tags->fixed_data_fixed_value_bcast = tag_values[9];
   tags->fixed_data_num_fixed_bcast = tag_values[10];
   tags->fixed_results_alltoall = tag_values[11];

   for (unsigned i = 0; i < 12; ++i) free(tag_strings[i]);
}

static inline unsigned
compute_bucket(int value, unsigned num_buckets) {
  return (value / 128) % num_buckets;
}

static int compare_link_data_tgt_idx (const void* a, const void* b) {
  
  if (((struct link_data*)a)->tgt_idx != ((struct link_data*)b)->tgt_idx)
    return (((struct link_data*)a)->tgt_idx - ((struct link_data*)b)->tgt_idx);
  else if ((((struct link_data*)a)->src_idx != ((struct link_data*)b)->src_idx))
    return (((struct link_data*)a)->src_idx - ((struct link_data*)b)->src_idx);
  else
    return (((struct link_data*)a)->weight > ((struct link_data*)b)->weight) -
           (((struct link_data*)a)->weight < ((struct link_data*)b)->weight);
}

static int compare_link_data_src_idx (const void* a, const void* b) {

  if (((struct link_data*)a)->points_idx != ((struct link_data*)b)->points_idx)
    return (((struct link_data*)a)->points_idx -
            ((struct link_data*)b)->points_idx);
  else
    return (((struct link_data*)a)->src_idx - ((struct link_data*)b)->src_idx);
}

#ifdef HAVE_C_NETCDF
static void handle_error(int status) {
  const char *err_string;
  err_string = nc_strerror(status);
  printf ("%s \n", err_string);
  if (status != NC_NOERR) {
    exit(EXIT_FAILURE);
  }
}
#endif

static void
get_io_proc_setup(unsigned comm_size, unsigned comm_rank,
                  unsigned io_proc_step, unsigned * num_io_procs,
                  unsigned * is_io_proc) {

  // compute which procs do io
  *num_io_procs = (comm_size + io_proc_step - 1) / io_proc_step;
  *is_io_proc = !(comm_rank % io_proc_step);
}

static void
read_link_data(struct link_data ** links, unsigned * num_links,
               enum yac_location * src_locations,
               enum yac_location * tgt_location, unsigned num_pointsets,
               char const * src_grid_name, char const * tgt_grid_name,
               int ncid, struct communicator * comm, unsigned io_proc_step) {

  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned comm_size = yac_get_comm_size(comm);

  unsigned num_io_procs, is_io_proc;

  get_io_proc_setup(comm_size, comm_rank, io_proc_step, &num_io_procs,
                    &is_io_proc);

  *links = NULL;
  *num_links = 0;
  *tgt_location = LOC_INVALID;

  if (!is_io_proc) return;

#ifdef HAVE_C_NETCDF

  //\todo use parallel netcdf if available

  unsigned io_idx = comm_rank / io_proc_step;

  int status, dimid, var_id;

  // global attributes
  size_t version_len;
  char * version;
  status = nc_inq_attlen(ncid, NC_GLOBAL, "version", &version_len);
  if (status != NC_NOERR) handle_error(status);
  version = malloc(version_len + 1);
  version[version_len] = '\0';
  status = nc_get_att_text(ncid, NC_GLOBAL, "version", version);
  if (status != NC_NOERR) handle_error(status);
  if ((strlen(YAC_WEIGHT_FILE_VERSION_STRING) != version_len) ||
      strncmp(version, YAC_WEIGHT_FILE_VERSION_STRING, version_len)) {
    if (io_idx == 0)
      fprintf(stderr, "version: %s (weight file) %s (interface)\n",
              version, YAC_WEIGHT_FILE_VERSION_STRING);
    yac_internal_abort_message(
      "ERROR(read_link_data): version string from weight file does not match "
      "with YAC version\n", __FILE__, __LINE__);
  }
  free(version);

  size_t grid_name_len;
  char * grid_name;
  status = nc_inq_attlen(ncid, NC_GLOBAL, "src_grid_name", &grid_name_len);
  if (status != NC_NOERR) handle_error(status);
  grid_name = malloc(grid_name_len + 1);
  grid_name[grid_name_len] = '\0';
  status = nc_get_att_text(ncid, NC_GLOBAL, "src_grid_name", grid_name);
  if (status != NC_NOERR) handle_error(status);
  if ((strlen(src_grid_name) != grid_name_len) ||
      strncmp(src_grid_name, grid_name, grid_name_len)) {
    if (io_idx == 0)
      fprintf(stderr, "grid name: %s (weight file) %s (interface)\n",
              grid_name, src_grid_name);
    yac_internal_abort_message(
      "ERROR(read_link_data): source grid name from weight file does not match "
      "with the one provided through the interface\n", __FILE__, __LINE__);
  }

  free(grid_name);

  status = nc_inq_attlen(ncid, NC_GLOBAL, "dst_grid_name", &grid_name_len);
  if (status != NC_NOERR) handle_error(status);
  grid_name = malloc(grid_name_len + 1);
  grid_name[grid_name_len] = '\0';
  status = nc_get_att_text(ncid, NC_GLOBAL, "dst_grid_name", grid_name);
  if (status != NC_NOERR) handle_error(status);
  if ((strlen(tgt_grid_name) != grid_name_len) ||
      strncmp(tgt_grid_name, grid_name, grid_name_len)) {
    if (io_idx == 0)
      fprintf(stderr, "grid name: %s (weight file) %s (interface)\n",
              grid_name, tgt_grid_name);
    yac_internal_abort_message(
      "ERROR(read_link_data): target grid name from weight file does not match "
      "with the one provided through the interface\n", __FILE__, __LINE__);
  }

  free(grid_name);

  size_t str_link_len;
  char * str_link;
  var_id = NC_GLOBAL;
  status = nc_inq_attlen(ncid, var_id, "contains_links", &str_link_len);
  if (status != NC_NOERR) handle_error(status);
  str_link = malloc(str_link_len + 1);
  str_link[str_link_len] = '\0';
  status = nc_get_att_text(ncid, var_id, "contains_links", str_link);
  if (status != NC_NOERR) handle_error(status);

  unsigned contains_links = 0;
  if ((strlen("TRUE") == str_link_len) &&
      !strncmp("TRUE", str_link, str_link_len)) {

    contains_links = 1;

  } else if ((strlen("FALSE") != str_link_len) ||
             strncmp("FALSE", str_link, str_link_len))
    yac_internal_abort_message("ERROR(read_link_data): invalid global "
                               "attribute contains_links", __FILE__,
                               __LINE__);
  free(str_link);

  size_t num_wgts = 0;
  if (contains_links) {
    status = nc_inq_dimid(ncid, "num_wgts", &dimid);
    if (status != NC_NOERR) handle_error(status);
    status = nc_inq_dimlen(ncid, dimid, &num_wgts);
    if (status != NC_NOERR) handle_error(status);
    if (num_wgts != 1)
        yac_internal_abort_message("ERROR(read_link_data): YAC only supports "
                                   "num_wgts == 1\n", __FILE__, __LINE__);
  }

  // get number of links from file

  size_t num_links_in_file = 0;
  if (contains_links) {
    status = nc_inq_dimid(ncid, "num_links", &dimid);
    if (status != NC_NOERR) handle_error(status);
    status = nc_inq_dimlen(ncid, dimid, &num_links_in_file);
    if (status != NC_NOERR) handle_error(status);
    if (num_links_in_file == 0)
        yac_internal_abort_message("ERROR(read_link_data): no links defined",
                                   __FILE__, __LINE__);
  }

  // get number of pointsets
  size_t tmp_num_pointsets = 0;
  if (contains_links) {
    status = nc_inq_dimid(ncid, "num_src_fields", &dimid);
    if (status != NC_NOERR) handle_error(status);
    status = nc_inq_dimlen(ncid, dimid, &tmp_num_pointsets);
    if (status != NC_NOERR) handle_error(status);
    if (tmp_num_pointsets == 0)
        yac_internal_abort_message("ERROR(read_link_data): no point sets in file\n",
                                   __FILE__, __LINE__);
    if (tmp_num_pointsets != num_pointsets) {
      if (io_idx == 0) fprintf(stderr, "number of source fields from file: %zu "
                               "interface: %u\n", tmp_num_pointsets, num_pointsets);
      yac_internal_abort_message("ERROR(read_link_data): number of source "
                                 "fields in file do not match with the number "
                                 "provided through the interface\n",
                                 __FILE__, __LINE__);
    }
  }

  // get max location string length from file
  size_t max_loc_str_len;
  status = nc_inq_dimid(ncid, "max_loc_str_len", &dimid);
  if (status != NC_NOERR) handle_error(status);
  status = nc_inq_dimlen(ncid, dimid, &max_loc_str_len);
  if (status != NC_NOERR) handle_error(status);
  if (max_loc_str_len != YAC_MAX_LOC_STR_LEN)
      yac_internal_abort_message("ERROR(read_link_data): wrong max location "
                                 "string length in weight file", __FILE__,
                                 __LINE__);

  // get source locations
  enum yac_location * tmp_src_locations =
    malloc(num_pointsets * sizeof(*tmp_src_locations));
  status = nc_inq_varid(ncid, "src_locations", &var_id);
  if (status != NC_NOERR) handle_error(status);

  for (unsigned i = 0; i < num_pointsets; ++i) {

    char loc_str[YAC_MAX_LOC_STR_LEN];

    size_t str_start[2] = {i, 0};
    size_t str_count[2] = {1, YAC_MAX_LOC_STR_LEN};

    status = nc_get_vara_text(ncid, var_id, str_start, str_count, loc_str);
    if (status != NC_NOERR) handle_error(status);

    tmp_src_locations[i] = yac_str2loc(loc_str);
  }

  for (unsigned i = 0; i < num_pointsets; ++i) {

    if (tmp_src_locations[i] != src_locations[i]) {

      int collumn_width[2] = {MAX(strlen("weight file"), YAC_MAX_LOC_STR_LEN),
                              MAX(strlen("interface"), YAC_MAX_LOC_STR_LEN)};
      if (io_idx == 0) {
        fprintf(stderr, "source locations:\n"
                        "%*s : %*s\n",
                        collumn_width[0], "weight file",
                        collumn_width[1], "interface");
        for (unsigned j = 0; j < num_pointsets; ++j) {

          char const * str_location_wgt = yac_loc2str(tmp_src_locations[j]);
          char const * str_location_tgt = yac_loc2str(src_locations[j]);
          fprintf(stderr, "%*s : %*s\n", collumn_width[0], str_location_wgt,
                  collumn_width[1], str_location_tgt);
        }
      }

      yac_internal_abort_message("ERROR(read_link_data): source locations in "
                                 "file do not match with the locations "
                                 "provided through the interface\n",
                                 __FILE__, __LINE__);
    }
  }

  free(tmp_src_locations);

  // get target location
  status = nc_inq_varid(ncid, "dst_location", &var_id);
  {
    char loc_str[YAC_MAX_LOC_STR_LEN];
    status = nc_get_var_text(ncid, var_id, loc_str);
    if (status != NC_NOERR) handle_error(status);
    *tgt_location = yac_str2loc(loc_str);
  }

  if (!contains_links) return;

  // get number of links per pointset
  unsigned * num_links_per_pointset =
    malloc(num_pointsets * sizeof(*num_links_per_pointset));
  status = nc_inq_varid(ncid, "num_links_per_src_field", &var_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_get_var_uint(ncid, var_id, num_links_per_pointset);
  if (status != NC_NOERR) handle_error(status);

  unsigned local_size = (num_links_in_file + num_io_procs - 1) / num_io_procs;

  size_t offset = MIN(local_size * io_idx, num_links_in_file);
  size_t count = MAX(MIN(local_size, num_links_in_file - local_size * io_idx), 0);

  *num_links = count;

  *links = malloc(count * sizeof(**links));

  if (count > 0) {
    unsigned points_idx = 0, points_idx_offset = 0;
    while ((points_idx < num_pointsets) &&
           (points_idx_offset + num_links_per_pointset[points_idx] < offset)) {

      points_idx_offset += num_links_per_pointset[points_idx];
      ++points_idx;
    };
    if (points_idx >= num_pointsets)
      yac_internal_abort_message(
        "ERROR(read_link_data): problem in num_links_per_src_field", __FILE__,
        __LINE__);
    {
      unsigned k = 0;
      unsigned pointset_off = offset - points_idx_offset;
      for (; (points_idx < num_pointsets) && (k < count);
           pointset_off = 0, ++points_idx) {

        for (; (pointset_off < num_links_per_pointset[points_idx]) && (k < count);
             ++pointset_off, ++k) {
          (*links)[k].points_idx = points_idx;
        }
      }
    }
  }
  free(num_links_per_pointset);

  // get links
  int * address = malloc(count * sizeof(*address));
  status = nc_inq_varid(ncid, "src_address", &var_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_get_vara_int(ncid, var_id, &offset, &count, address);
  if (status != NC_NOERR) handle_error(status);
  for (unsigned i = 0; i < count; ++i)
    (*links)[i].src_idx = address[i] - 1;

  status = nc_inq_varid(ncid, "dst_address", &var_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_get_vara_int(ncid, var_id, &offset, &count, address);
  if (status != NC_NOERR) handle_error(status);
  for (unsigned i = 0; i < count; ++i)
    (*links)[i].tgt_idx = address[i] - 1;
  free(address);

  double * weights = malloc(count * sizeof(*weights));

  status = nc_inq_varid(ncid, "remap_matrix", &var_id);
  if (status != NC_NOERR) handle_error(status);
  size_t offsets[2] = {offset, 0};
  size_t counts[2] = {count, 1};
  status = nc_get_vara_double(ncid, var_id, offsets, counts, weights);
  if (status != NC_NOERR) handle_error(status);
  for (size_t i = 0; i < count; ++i)
    (*links)[i].weight = weights[i];

  free(weights);
#else
  UNUSED(src_locations);
  UNUSED(num_pointsets);
  UNUSED(src_grid_name);
  UNUSED(tgt_grid_name);
  UNUSED(ncid);
#endif
}

static void
reorder_target_points(struct interpolation_method_file * method_file,
                      unsigned * flags, unsigned num_fixed_values,
                      unsigned * num_interpolated_points,
                      unsigned * num_interpolated_points_link,
                      unsigned ** num_interpolated_points_fixed,
                      unsigned * num_remaining_points) {

  unsigned * tmp_num_interpolated_points =
    calloc(num_fixed_values + 2, sizeof(*tmp_num_interpolated_points));

  for (unsigned i = 0; i < method_file->num_global_tgt_ids; ++i)
    if (flags[i] != (unsigned)-1)
      tmp_num_interpolated_points[flags[i]]++;

  unsigned * offsets = malloc((num_fixed_values + 2) * sizeof(*offsets));
  unsigned tmp_num_remaining_points = tmp_num_interpolated_points[0];
  unsigned total_num_interpolated_points = tmp_num_interpolated_points[1];

  offsets[1] = 0;
  for (unsigned i = 0; i < num_fixed_values; ++i) {
    total_num_interpolated_points += tmp_num_interpolated_points[i + 2];
    offsets[i + 2] = offsets[i + 1] + tmp_num_interpolated_points[i + 1];
    tmp_num_interpolated_points[i] = tmp_num_interpolated_points[i + 2];
  }
  offsets[0] = total_num_interpolated_points;

  *num_interpolated_points_link = tmp_num_interpolated_points[1];
  *num_interpolated_points = total_num_interpolated_points;
  *num_remaining_points = tmp_num_remaining_points;
  *num_interpolated_points_fixed = tmp_num_interpolated_points;

  unsigned new_num_global_tgt_ids =
    tmp_num_remaining_points + total_num_interpolated_points;
  int * new_global_tgt_ids =
    malloc(new_num_global_tgt_ids * sizeof(*new_global_tgt_ids));
  unsigned have_coordinates = (method_file->coordinates_xyz != NULL);
  double (*new_coordinates_xyz)[3] =
    (have_coordinates)?
    (malloc(new_num_global_tgt_ids * sizeof(*new_coordinates_xyz))):(NULL);

  for (unsigned i = 0; i < method_file->num_global_tgt_ids; ++i) {

    unsigned curr_flag = flags[i];

    switch (curr_flag) {
      case ((unsigned)-1):
        break;
      default:
        new_global_tgt_ids[offsets[curr_flag]] = method_file->global_tgt_ids[i];
        if (have_coordinates) {
          new_coordinates_xyz[offsets[curr_flag]][0] =
            method_file->coordinates_xyz[i][0];
          new_coordinates_xyz[offsets[curr_flag]][1] =
            method_file->coordinates_xyz[i][1];
          new_coordinates_xyz[offsets[curr_flag]][2] =
            method_file->coordinates_xyz[i][2];
        }
        ++offsets[curr_flag];
    };
  }
  free(offsets);

  method_file->num_global_tgt_ids = new_num_global_tgt_ids;
  memcpy(method_file->global_tgt_ids, new_global_tgt_ids,
         new_num_global_tgt_ids * sizeof(*method_file->global_tgt_ids));
  if (have_coordinates) {
    memcpy(method_file->coordinates_xyz, new_coordinates_xyz,
           new_num_global_tgt_ids * sizeof(*method_file->coordinates_xyz));
  }

  free(new_global_tgt_ids);
  free(new_coordinates_xyz);
}

static int compare_link_data_reorder (const void* a, const void* b) {

  if (((struct link_data_reorder*)a)->points_idx !=
      ((struct link_data_reorder*)b)->points_idx)
    return (((struct link_data_reorder*)a)->points_idx -
            ((struct link_data_reorder*)b)->points_idx);
  return ((struct link_data_reorder*)a)->src_idx - ((struct link_data_reorder*)b)->src_idx;
}

/**
 * determines for all given local source indices the matching subdomain id and
 * local id
 * for all local source indices the respective rank is set to -1
 * the subdomain_ids of remote source indices is set to -1
 */
static void
lookup_local_global_src_ids(struct link_data * links, unsigned num_links,
                            unsigned rank, struct domain * domain,
                            enum yac_location * locations, unsigned num_pointsets,
                            unsigned * subdomain_ids,
                            unsigned * local_src_ids) {

  if (num_links == 0) return;

  unsigned num_local_src_indices = 0;

  struct link_data_reorder * links_reorder =
    malloc(num_links * sizeof(*links_reorder));

  for (unsigned i = 0; i < num_links; ++i) {

    if (links[i].rank == rank) {

      links_reorder[num_local_src_indices].src_idx = links[i].src_idx;
      links_reorder[num_local_src_indices].points_idx = links[i].points_idx;
      links_reorder[num_local_src_indices].orig_pos = i;
      num_local_src_indices++;
      links[i].rank = -1;
    } else
      subdomain_ids[i] = -1;
  }

  if (num_local_src_indices == 0) {
    free(links_reorder);
    return;
  }

  qsort(links_reorder, num_local_src_indices, sizeof(*links_reorder),
        compare_link_data_reorder);

  unsigned * num_local_points_per_pointset =
    calloc(num_pointsets, sizeof(*num_local_points_per_pointset));
  int * src_indices = malloc(num_local_src_indices * sizeof(*src_indices));

  for (unsigned i = 0; i < num_local_src_indices; ++i) {

    src_indices[i] = links_reorder[i].src_idx;
    num_local_points_per_pointset[links_reorder[i].points_idx]++;
  }

  unsigned num_subdomains = yac_get_num_subdomains(domain);

  unsigned * matching_ids_idx = malloc(num_local_src_indices *
                                       sizeof(*matching_ids_idx));
  unsigned * temp_local_src_ids = malloc(num_local_src_indices *
                                         sizeof(*temp_local_src_ids));
  unsigned num_matching_ids;

  for (unsigned i = 0, offset = 0; i < num_pointsets;
       offset += num_local_points_per_pointset[i++]) {

    for (unsigned j = 0; j < num_subdomains; ++j) {

      yac_match_global_ids(domain, j, locations[i], src_indices + offset,
                           num_local_points_per_pointset[i], NULL,
                           matching_ids_idx, &num_matching_ids,
                           temp_local_src_ids, 0);

      for (unsigned k = 0; k < num_matching_ids; ++k) {
        unsigned orig_pos =
          links_reorder[offset + matching_ids_idx[k]].orig_pos;
        subdomain_ids[orig_pos] = j;
        local_src_ids[orig_pos] = temp_local_src_ids[k];
      }
    }
  }

  free(temp_local_src_ids);
  free(matching_ids_idx);
  free(src_indices);
  free(num_local_points_per_pointset);
  free(links_reorder);
}

static void
generate_data_request(struct interpolation_method_file * method_file,
                      struct link_data * links, unsigned num_links,
                      unsigned * local_src_ids,
                      unsigned * num_remote_points_per_pointset) {

  // initialise data request
  method_file->data_request[0] =
    malloc(method_file->num_pointsets * sizeof(*(method_file->data_request[0])));
  struct interpolation_method_data_request * data_request =
    method_file->data_request[0];

  for (unsigned i = 0; i < method_file->num_pointsets; ++i) {

    data_request[i].num_values = 0;
    data_request[i].global_ids = NULL;
    data_request[i].ranks = NULL;
    num_remote_points_per_pointset[i] = 0;
  }

  unsigned total_remote_src_indices = 0;

  for (unsigned i = 0; i < num_links; ++i) {

    if (links[i].points_idx >= method_file->num_pointsets)
      yac_internal_abort_message(
        "ERROR(generate_data_request): points_idx to big", __FILE__, __LINE__);

    if (links[i].rank != -1) ++total_remote_src_indices;
  }

  if (total_remote_src_indices == 0) return;

  int * remote_src_indices = malloc(total_remote_src_indices *
                                    sizeof(*remote_src_indices));
  unsigned * remote_src_indices_orig_pos =
    malloc(total_remote_src_indices * sizeof(*remote_src_indices_orig_pos));

  for (unsigned i = 0, j = 0; i < num_links; ++i) {
    if (links[i].rank != -1) {
      remote_src_indices[j] = links[i].src_idx;
      remote_src_indices_orig_pos[j] = i;
      ++j;
    }
  }

  yac_quicksort_index(remote_src_indices, total_remote_src_indices,
                      (int*)remote_src_indices_orig_pos);

  unsigned * ranks = malloc(total_remote_src_indices * sizeof(*ranks));
  unsigned * points_idx = malloc(total_remote_src_indices * sizeof(*points_idx)); 

  unsigned num_unique_src_indices = 1;
  local_src_ids[remote_src_indices_orig_pos[0]] = 0;
  ranks[0] = links[remote_src_indices_orig_pos[0]].rank;
  points_idx[0] = links[remote_src_indices_orig_pos[0]].points_idx;
  num_remote_points_per_pointset[
    links[remote_src_indices_orig_pos[0]].points_idx]++;

  for (unsigned i = 1; i < total_remote_src_indices; ++i) {

    if (remote_src_indices[i] != remote_src_indices[i-1]) {

      remote_src_indices[num_unique_src_indices] =  remote_src_indices[i];
      ranks[num_unique_src_indices] =
        links[remote_src_indices_orig_pos[i]].rank;
      points_idx[num_unique_src_indices] =
        links[remote_src_indices_orig_pos[i]].points_idx;
      num_remote_points_per_pointset[
        links[remote_src_indices_orig_pos[i]].points_idx]++;

      local_src_ids[remote_src_indices_orig_pos[i]] = num_unique_src_indices++;
    } else {
      local_src_ids[remote_src_indices_orig_pos[i]] =
        num_unique_src_indices - 1;
    }
  }

  free(remote_src_indices_orig_pos);

  for (unsigned i = 0; i < method_file->num_pointsets; ++i) {

    unsigned num_values = num_remote_points_per_pointset[i];

    if (num_values > 0) {

      data_request[i].num_values = 0;
      data_request[i].global_ids =
        malloc(num_values * sizeof(*(data_request[i].global_ids)));
      data_request[i].ranks =
        malloc(num_values * sizeof(*(data_request[i].ranks)));
    }
  }

  for (unsigned i = 0; i < num_unique_src_indices; ++i) {

    unsigned curr_points_idx = points_idx[i];
    data_request[curr_points_idx].global_ids[
      data_request[curr_points_idx].num_values] = remote_src_indices[i];
    data_request[curr_points_idx].ranks[
      data_request[curr_points_idx].num_values] = ranks[i];
    data_request[curr_points_idx].num_values++;
  }

  free(remote_src_indices);
  free(ranks);
  free(points_idx);
}

static struct interpolation_plan_element *
generate_interpolation_plan_wpointsum(
  unsigned num_pointsets, struct link_data * links, unsigned num_links,
  unsigned * num_src_per_tgt, unsigned num_interpolated_points,
  unsigned * subdomain_ids, unsigned * local_src_ids,
  unsigned * num_remote_points_per_pointset) {

  if (num_links == 0) return NULL;

  unsigned * src_indices = malloc(3 * num_links * sizeof(*src_indices));

  for (unsigned i = 0; i < num_links; ++i) {

    // for remote points subdomain_ids is -1 and local_src_ids contains
    // an entry matching the data_request
    src_indices[3 * i + 0] = subdomain_ids[i] + 1;
    src_indices[3 * i + 1] = links[i].points_idx;
    src_indices[3 * i + 2] = local_src_ids[i];
  }

  struct interpolation_plan_element * plan = malloc(1 * sizeof(*plan));
  double * weights = malloc(num_links * sizeof(*weights));

  for (unsigned i = 0; i < num_links; ++i) weights[i] = links[i].weight;

  yac_init_interpolation_plan_element_wpointsum(
    plan, num_interpolated_points, src_indices, num_src_per_tgt,
    weights, num_remote_points_per_pointset, num_pointsets);

  free(weights);
  free(src_indices);

  return plan;
}

/**
 * generates the interpolation plan based on the results \n
 */
static struct interpolation_plan_element *
generate_interpolation_plan(struct interpolation_method_file * method_file,
                            struct intermediate_results * results,
                            unsigned num_interpolated_points_link,
                            unsigned * num_interpolated_points_fixed,
                            unsigned * subdomain_ids, unsigned * local_src_ids,
                            unsigned * num_remote_points_per_pointset) {

  struct interpolation_plan_element * plan_wpointsum =
    generate_interpolation_plan_wpointsum(
      method_file->num_pointsets, results->links, results->num_links,
      results->link_flags, num_interpolated_points_link, subdomain_ids,
      local_src_ids, num_remote_points_per_pointset);

  unsigned has_fixed = 0;
  for (unsigned i = 0; i < results->num_fixed_tgt; ++i) {
    if (results->fixed_flags[i] != (unsigned)-1) {
      has_fixed = 1;
      break;
    }
  }
    
  if (!has_fixed) return plan_wpointsum;

  unsigned has_links = results->num_links > 0;
  unsigned num_used_fixed_values = 0;

  for (unsigned i = 0; i < results->num_fixed_values; ++i)
    num_used_fixed_values += num_interpolated_points_fixed[i] > 0;

  if ((!has_links) && (num_used_fixed_values == 1)) {
    for (unsigned i = 0; i < results->num_fixed_values; ++i) {
      if (num_interpolated_points_fixed[i] > 0) {
        struct interpolation_plan_element * plan_fixed =
          malloc(1 * sizeof(*plan_fixed));
        yac_init_interpolation_plan_element_fixed(
          plan_fixed, num_interpolated_points_fixed[i],
          results->fixed_values[i]);
        return plan_fixed;
      }
    }
  }

  struct interpolation_plan_element ** plans =
    malloc((has_links + num_used_fixed_values) * sizeof(*plans));

  if (has_links) plans[0] = plan_wpointsum;

  for (unsigned i = 0, offset = has_links; i < results->num_fixed_values; ++i) {
    if (num_interpolated_points_fixed[i] != 0) {
      plans[offset] = malloc(1 * sizeof(**plans));
      yac_init_interpolation_plan_element_fixed(
        plans[offset], num_interpolated_points_fixed[i],
        results->fixed_values[i]);
      ++offset;
    }
  }

  struct interpolation_plan_element * plan_coll =
    malloc(1 * sizeof(*plan_coll));
  yac_init_interpolation_plan_element_collection(
    plan_coll, has_links + num_used_fixed_values, plans);

  for (unsigned i = 0; i < has_links + num_used_fixed_values; ++i)
    free(plans[i]);
  free(plans);
  return plan_coll;
}

static void
reorder_link_flags(unsigned * link_flags, unsigned num_global_tgt_ids) {

  for (unsigned i = 0, offset = 0; i < num_global_tgt_ids; ++i) {
    if ((link_flags[i] != 0) && (link_flags[i] != (unsigned)-1)) {
      if (i != offset) link_flags[offset] = link_flags[i];
      ++offset;
    }
  }
}

static unsigned *
merge_flags(struct intermediate_results * results) {

  struct interpolation_method_file * method_file = results->method_file;

  unsigned * merged_flags =
    malloc(method_file->num_global_tgt_ids * sizeof(*merged_flags));

  for (unsigned i = 0; i < method_file->num_global_tgt_ids; ++i) {
    switch(results->link_flags[i]) {
      case (0):
      case ((unsigned)-1):
        merged_flags[i] = results->link_flags[i];
        break;
      default:
        merged_flags[i] = 1;
    };
  }

  for (unsigned i = 0, idx = 0; i < results->num_fixed_tgt; ++i) {

    for (; method_file->global_tgt_ids[idx] != results->global_tgt_ids[i];
         ++idx);

    if (merged_flags[idx] != 0) {
      fprintf(stderr, "Weight file contains for a target (global id %d) links "
              "and a fixed value, which is not allowed.\n",
              results->global_tgt_ids[i]);
      yac_internal_abort_message("ERROR(merge_flags): tgt has link and fixed "
                                 "value\n", __FILE__, __LINE__);
    }

    switch (results->fixed_flags[i]) {
      case ((unsigned)-1):
        merged_flags[idx] = (unsigned)-1;
        break;
      default:
        merged_flags[idx] = results->fixed_flags[i] + 2;
    };
  }

  return merged_flags;
}

static void
finalise_results(struct intermediate_results * results,
                 struct communicator * comm) {

  struct interpolation_method_file * method_file = results->method_file;

  //--------------------------------------------------------------------
  // merge the link and fixed flags
  //--------------------------------------------------------------------
  unsigned * merged_flags = merge_flags(results);

  //--------------------------------------------------------------------
  // put valid link flags at the beginning of the array
  //--------------------------------------------------------------------
  reorder_link_flags(results->link_flags, method_file->num_global_tgt_ids);
  //--------------------------------------------------------------------
  // reorder global target indices and their coordinates (if available)
  // target points that can be interpolated using links first, then points
  // that receive fixed values, and at the end points that cannot be
  // interpolated. Points that are interpolated by other source processes
  // are ignored.
  //--------------------------------------------------------------------
  unsigned num_remaining_points;
  unsigned num_interpolated_points_link;
  unsigned * num_interpolated_points_fixed;
  reorder_target_points(method_file, merged_flags,
                        results->num_fixed_values,
                        method_file->num_interpolated_points,
                        &num_interpolated_points_link,
                        &num_interpolated_points_fixed,
                        &num_remaining_points);
  free(merged_flags);

  // subdomain id for each source point required for the interpolation
  unsigned * subdomain_ids =
    malloc(results->num_links * sizeof(*subdomain_ids));
  // local index for each source point required for the interpolation
  unsigned * local_src_ids =
    malloc(results->num_links * sizeof(*local_src_ids));

  //--------------------------------------------------------------------
  // the link data contains only global ids
  // we now determine the local indices and subdomain ids of all
  // source points that are locally available
  //--------------------------------------------------------------------
  lookup_local_global_src_ids(results->links, results->num_links,
                              yac_get_comm_rank(comm), method_file->domain,
                              method_file->src_locations,
                              method_file->num_pointsets, subdomain_ids,
                              local_src_ids);

  //--------------------------------------------------------------------
  // generate data request and prepare local indices
  //--------------------------------------------------------------------
  unsigned * num_remote_points_per_pointset =
    malloc(method_file->num_pointsets * sizeof(*num_remote_points_per_pointset));
  generate_data_request(method_file, results->links, results->num_links,
                        local_src_ids, num_remote_points_per_pointset);
  //--------------------------------------------------------------------
  // generate interpolation plan
  //--------------------------------------------------------------------
  method_file->plan_stack[0] =
    generate_interpolation_plan(method_file, results,
                                num_interpolated_points_link,
                                num_interpolated_points_fixed, subdomain_ids,
                                local_src_ids, num_remote_points_per_pointset);
  if (method_file->plan_stack[0] != NULL) method_file->plan_stack++;
  free(num_interpolated_points_fixed);
  free(num_remote_points_per_pointset);
  free(local_src_ids);
  free(subdomain_ids);
  free(results->links);
  free(results->link_flags);
  free(results->global_tgt_ids);
  free(results->fixed_flags);

  //--------------------------------------------------------------------
  // call next interpolation in interpolation stack
  //--------------------------------------------------------------------
  struct interpolation_method ** method_stack = method_file->method_stack+1;
  struct interpolation_plan_element ** plan_stack = method_file->plan_stack;
  unsigned * num_interpolated_points = method_file->num_interpolated_points+1;
  double (*coordinates_xyz)[3] = method_file->coordinates_xyz;
  int * global_ids = method_file->global_tgt_ids +
                     *(method_file->num_interpolated_points);

  if (coordinates_xyz != NULL)
     coordinates_xyz += *(method_file->num_interpolated_points);

  // call next interpolation method on stack
  if (*method_stack != NULL)
     (*method_stack)->vtable->do_search(
        method_stack, plan_stack, global_ids, coordinates_xyz,
        method_file->dist_tgt_indices.location, num_remaining_points,
        num_interpolated_points, method_file->data_request+1);

  free(results->fixed_values);
  free(results);

  if (method_file->finish_callback != NULL)
    method_file->finish_callback(method_file->finish_user_data);
}

static void
link_results_alltoall_callback(struct communicator * comm,
                               struct comm_message msg,
                               void * user_data) {

  struct intermediate_results * results =
    (struct intermediate_results *)user_data;

  struct interpolation_method_file * method_file = results->method_file;

  unsigned * flags = malloc(method_file->num_global_tgt_ids * sizeof(*flags));
  struct link_data * links = NULL;
  size_t links_array_size = 0;

  int offset = 0;
  unsigned link_idx = 0;

  // unpack results
  for (unsigned i = 0; i < method_file->num_global_tgt_ids; ++i) {

    char * curr_packing_unit = (char*)msg.buffer + offset;
    unsigned packing_unit_size = msg.count - (unsigned)offset;
    int position = 0;

    yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                    flags + i, 1, COMM_UINT, comm);

    if (flags[i] == (unsigned)-1) {
      offset += position;
      continue;
    }

    ENSURE_ARRAY_SIZE(links, links_array_size, link_idx + flags[i]);

    for (unsigned j = 0; j < flags[i]; ++j, ++link_idx) {

      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      &(links[link_idx].src_idx), 1, COMM_INT, comm);
      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      &(links[link_idx].tgt_idx), 1, COMM_INT, comm);
      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      &(links[link_idx].rank), 1, COMM_UINT, comm);
      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      &(links[link_idx].points_idx), 1, COMM_UINT, comm);
      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      &(links[link_idx].weight), 1, COMM_DBLE, comm);
    }
    offset += position;
  }

  results->links = realloc(links, link_idx * sizeof(*links));
  results->num_links = link_idx;
  results->link_flags = flags;

  // if the link results have already been received
  if (results->num_fixed_tgt != (unsigned)-1) finalise_results(results, comm);
}

static void
process_tgt_requests_links(struct intermediate_results * results) {

  struct interpolation_method_file * method_file = results->method_file;

  struct communicator * comm = yac_get_domain_comm(method_file->domain);
  unsigned comm_size = yac_get_comm_size(comm);

  unsigned req_idx, link_idx;
  unsigned num_links = method_file->dist_link_data.count;
  unsigned num_req = method_file->dist_tgt_indices.count;
  struct index_data * tgt_req = method_file->dist_tgt_indices.data;
  struct link_data * links = method_file->dist_link_data.links;
  unsigned * send_size_per_rank =
    calloc(comm_size, sizeof(*send_size_per_rank));
  unsigned * flags = calloc(num_req, sizeof(*flags));

  unsigned link_pack_size = yac_comm_pack_size(2, COMM_INT, comm);
  link_pack_size += yac_comm_pack_size(2, COMM_UINT, comm);
  link_pack_size += yac_comm_pack_size(1, COMM_DBLE, comm);
  unsigned uint_pack_size = yac_comm_pack_size(1, COMM_UINT, comm);

  // determine for target indices that are requested by multiple source
  // processes the source process that will process the target point
  req_idx = 0;
  link_idx = 0;
  while (req_idx < num_req) {

    unsigned start_req_idx = req_idx++;
    int tgt_idx = tgt_req[start_req_idx].index;
    while ((req_idx < num_req) && (tgt_req[req_idx].index == tgt_idx))
      ++req_idx;

    // if only one process requests the current target index
    if (req_idx - start_req_idx == 1) continue;

    // find the first link that matches the target index
    while ((link_idx < num_links) &&
           (links[link_idx].tgt_idx < tgt_idx)) ++link_idx;

    // if there is no link for the current target index
    if ((link_idx >= num_links) || (links[link_idx].tgt_idx != tgt_idx))
      continue;

    // find the first link that matches the target index and has a source points
    // that is located on one of the target processes requesting the target
    // index
    unsigned i = req_idx;
    while ((link_idx < num_links) && (links[link_idx].tgt_idx == tgt_idx) &&
           (i == req_idx)) {
      for (i = start_req_idx; i < req_idx; ++i)
        if (tgt_req[i].rank == links[link_idx].rank) break;
      ++link_idx;
    }

    if (i == req_idx) i = start_req_idx;
    for (unsigned j = start_req_idx; j < req_idx; ++j)
      flags[j] = (unsigned)-1;
    flags[i] = 0;
  }

  // count the number of links per requested target index
  // if a target index is requested by more than one process, count the links
  // only for one process, set the link count to -1 for the other processes
  for (req_idx = 0, link_idx = 0; (req_idx < num_req) && (link_idx < num_links);
       ++req_idx) {

    int tgt_idx = tgt_req[req_idx].index;
    unsigned rank = tgt_req[req_idx].rank;

    send_size_per_rank[rank] += uint_pack_size;

    if (flags[req_idx] == (unsigned)-1) continue;

    while ((link_idx < num_links) &&
           (links[link_idx].tgt_idx < tgt_idx)) ++link_idx;

    while ((link_idx < num_links) &&
           (links[link_idx].tgt_idx == tgt_idx)) {

      flags[req_idx]++;
      ++link_idx;
    }

    send_size_per_rank[rank] += flags[req_idx] * link_pack_size;
  }
  for (; req_idx < num_req; ++req_idx)
    send_size_per_rank[tgt_req[req_idx].rank] += uint_pack_size;

  unsigned * offsets = malloc(comm_size * sizeof(*offsets));
  int * pack_sizes = calloc(comm_size, sizeof(*pack_sizes));
  unsigned send_buffer_size = 0;

  for (unsigned i = 0; i < comm_size; ++i) {
    offsets[i] = send_buffer_size;
    send_buffer_size += send_size_per_rank[i];
  }

  char * send_buffer = malloc(send_buffer_size);

  // pack answers
  for (req_idx = 0, link_idx = 0; req_idx < num_req; ++req_idx) {

    unsigned rank = tgt_req[req_idx].rank;
    int tgt_idx = tgt_req[req_idx].index;

    char * curr_packing_unit = send_buffer + offsets[rank] + pack_sizes[rank];
    unsigned packing_unit_size = send_size_per_rank[rank] - pack_sizes[rank];
    int position = 0;

    yac_comm_pack(flags + req_idx, 1, COMM_UINT, curr_packing_unit,
                  packing_unit_size, &position, comm);

    if (flags[req_idx] == (unsigned)-1) {
      pack_sizes[rank] += position;
      continue;
    }

    while ((link_idx < num_links) &&
           (tgt_idx != tgt_req[req_idx].index)) ++link_idx;

    for (unsigned i = 0; i < flags[req_idx]; ++i, ++link_idx) {

      yac_comm_pack(&(links[link_idx].src_idx), 1, COMM_INT,
                    curr_packing_unit, packing_unit_size, &position, comm);
      yac_comm_pack(&(links[link_idx].tgt_idx), 1, COMM_INT,
                    curr_packing_unit, packing_unit_size, &position, comm);
      yac_comm_pack(&(links[link_idx].rank), 1, COMM_UINT,
                    curr_packing_unit, packing_unit_size, &position, comm);
      yac_comm_pack(&(links[link_idx].points_idx), 1, COMM_UINT,
                    curr_packing_unit, packing_unit_size, &position, comm);
      yac_comm_pack(&(links[link_idx].weight), 1, COMM_DBLE,
                    curr_packing_unit, packing_unit_size, &position, comm);
    }
    pack_sizes[rank] += position;
  }

  // the following is just to ensure that the packing does not leave any holes
  // in the send buffer (should theoretically never happen)
  char * to_buffer = send_buffer;
  for (unsigned i = 0; i < comm_size; ++i) {
    char * from_buffer = send_buffer + offsets[i];
    unsigned packed_size = pack_sizes[i];
    send_size_per_rank[i] = packed_size;
    // should never happen
    if (from_buffer != to_buffer) memmove(to_buffer, from_buffer, packed_size);
    to_buffer += packed_size;
  }

  free(pack_sizes);
  free(offsets);
  free(flags);

  // distribute link results
  yac_comm_alltoallv_callback(send_buffer, send_size_per_rank, COMM_PACKED,
                              method_file->tags.link_results_alltoall, comm,
                              results, link_results_alltoall_callback,
                              MSG_PRIO_LOW);

  free(send_buffer);
  free(send_size_per_rank);
  free(links);
  free(tgt_req);
}

static void
fixed_results_alltoall_callback(struct communicator * comm,
                                struct comm_message msg,
                                void * user_data) {

  struct intermediate_results * results =
    (struct intermediate_results *)user_data;

  int * global_tgt_ids = NULL;
  unsigned * flags = NULL;
  size_t global_tgt_ids_size = 0;
  size_t flags_size = 0;
  unsigned num_fixed_tgt = 0;

  unsigned offset = 0;

  while (offset < msg.count) {

    char * curr_packing_unit = (char*)msg.buffer + offset;
    unsigned packing_unit_size = msg.count - offset;
    int position = 0;

    unsigned curr_num_fixed_tgt;
    yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                    &curr_num_fixed_tgt, 1, COMM_UINT, comm);

    ENSURE_ARRAY_SIZE(global_tgt_ids, global_tgt_ids_size,
                      num_fixed_tgt + curr_num_fixed_tgt);
    ENSURE_ARRAY_SIZE(flags, flags_size, num_fixed_tgt + curr_num_fixed_tgt);

    for (unsigned i = 0; i < curr_num_fixed_tgt; ++i) {

      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      global_tgt_ids + i, 1, COMM_INT, comm);
      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      flags + i, 1, COMM_UINT, comm);
    }
    offset += position;

    num_fixed_tgt += curr_num_fixed_tgt;
  }

  results->global_tgt_ids = realloc(global_tgt_ids,
                                    num_fixed_tgt * sizeof(*global_tgt_ids));
  results->fixed_flags = realloc(flags, num_fixed_tgt * sizeof(*flags));
  results->num_fixed_tgt = num_fixed_tgt;

  // if the link results have already been received
  if (results->num_links != (unsigned)-1) finalise_results(results, comm);
}

static void
process_tgt_requests_fixed(struct intermediate_results * results) {

  struct interpolation_method_file * method_file = results->method_file;

  struct communicator * comm = yac_get_domain_comm(method_file->domain);
  unsigned comm_size = yac_get_comm_size(comm);

  unsigned num_req = method_file->dist_tgt_indices.count;
  unsigned num_fixed_values = method_file->dist_fixed_data.num_fixed_values;

  struct index_data * tgt_req = method_file->dist_tgt_indices.data;

  unsigned * flags = calloc(num_req, sizeof(*flags));

  // for all fixed values
  for (unsigned fixed_value_idx = 0; fixed_value_idx < num_fixed_values;
       ++fixed_value_idx) {

    if (method_file->dist_fixed_data.num_tgt_indices_per_fixed_value[
          fixed_value_idx] == 0) continue;

    unsigned req_idx = 0, fixed_tgt_idx = 0;
    unsigned prev_tgt_idx = (unsigned)-1;
    int * fixed_tgt_indices =
      method_file->dist_fixed_data.global_tgt_indices_per_fixed_value[
        fixed_value_idx];

    // while there are requests and tgt indices left
    while ((req_idx < num_req) &&
           (fixed_tgt_idx <
            method_file->dist_fixed_data.num_tgt_indices_per_fixed_value[
              fixed_value_idx])) {

      // if current requested tgt index is equal to current tgt index of current
      // fixed value
      if (tgt_req[req_idx].index == fixed_tgt_indices[fixed_tgt_idx]) {

        // if another fixed values has already been assigned to the target
        if (flags[req_idx] != 0) {
          fprintf(stderr, "A weight file containes for target point %d "
                  "more than one fixed value.\n", tgt_req[req_idx].index);
          yac_internal_abort_message(
            "ERROR(process_tgt_requests_fixed): more than on fixed value per "
            "target point", __FILE__, __LINE__);
        }

        // if current requested tgt index is identical to the previously found
        // one
        if (prev_tgt_idx == tgt_req[req_idx].index) {

          // mark flag with -1 (to be removed on the respective source process)
          flags[req_idx] = (unsigned)-1;

        } else {

          // set flag to fixed value index + 1
          flags[req_idx] = fixed_value_idx + 1;
          prev_tgt_idx = tgt_req[req_idx].index;
        }

        ++req_idx;

      // if current requested tgt index is smaller than current tgt index of
      // current fixed value
      } else if (tgt_req[req_idx].index < fixed_tgt_indices[fixed_tgt_idx]) {

        // increase request index
        ++req_idx;

      } else {

        // increase tgt index of current fixed value
        ++fixed_tgt_idx;
      }
    }
  }

  unsigned fixed_data_pack_size = yac_comm_pack_size(1, COMM_INT, comm);
  fixed_data_pack_size += yac_comm_pack_size(1, COMM_UINT, comm);
  unsigned packing_unit_header_size = yac_comm_pack_size(1, COMM_UINT, comm);

  unsigned * send_size_per_rank =
    calloc(comm_size, sizeof(*send_size_per_rank));

  for (unsigned i = 0; i < num_req; ++i)
    if (flags[i] != 0)
      send_size_per_rank[tgt_req[i].rank] += fixed_data_pack_size;

  int * offsets = malloc(comm_size * sizeof(*offsets));
  int * positions = calloc(comm_size, sizeof(*positions));
  unsigned send_buffer_size = 0;
  for (unsigned i = 0; i < comm_size; ++i) {
    offsets[i] = send_buffer_size;
    if (send_size_per_rank[i] > 0)
      send_size_per_rank[i] += packing_unit_header_size;
    send_buffer_size += send_size_per_rank[i];
  }

  char * send_buffer = malloc(send_buffer_size);

  for (unsigned i = 0; i < comm_size; ++i) {
    if (send_size_per_rank[i] == 0) continue;
    unsigned packing_unit_header =
      (send_size_per_rank[i] - packing_unit_header_size) / fixed_data_pack_size;
    yac_comm_pack(&packing_unit_header, 1, COMM_UINT, send_buffer + offsets[i],
                  send_size_per_rank[i], positions + i, comm);
  }

  for (unsigned i = 0; i < num_req; ++i) {

    if (flags[i] == 0) continue;
    if (flags[i] != (unsigned)-1) flags[i]--;

    unsigned rank = tgt_req[i].rank;
    char * curr_packing_unit = send_buffer + offsets[rank];
    unsigned packing_unit_size = send_size_per_rank[rank];
    int * position = positions + rank;

    yac_comm_pack(&(tgt_req[i].index), 1, COMM_INT, curr_packing_unit,
                  packing_unit_size, position, comm);
    yac_comm_pack(flags + i, 1, COMM_UINT, curr_packing_unit,
                  packing_unit_size, position, comm);
  }

  // the following is just to ensure that the packing does not leave any holes
  // in the send buffer (should theoretically never happen)
  char * to_buffer = send_buffer;
  for (unsigned i = 0; i < comm_size; ++i) {
    char * from_buffer = send_buffer + offsets[i];
    unsigned packed_size = positions[i];
    send_size_per_rank[i] = packed_size;
    // should never happen
    if (from_buffer != to_buffer) memmove(to_buffer, from_buffer, packed_size);
    to_buffer += packed_size;
  }

  free(positions);
  free(offsets);
  free(flags);

  // distribute fixed results
  yac_comm_alltoallv_callback(send_buffer, send_size_per_rank, COMM_PACKED,
                              method_file->tags.fixed_results_alltoall, comm,
                              results, fixed_results_alltoall_callback,
                              MSG_PRIO_LOW);

  free(send_buffer);
  free(send_size_per_rank);

  free(method_file->dist_fixed_data.num_tgt_indices_per_fixed_value);
  if (method_file->dist_fixed_data.global_tgt_indices_per_fixed_value != NULL)
    for (unsigned i = 0; i < method_file->dist_fixed_data.num_fixed_values; ++i)
      free(method_file->dist_fixed_data.global_tgt_indices_per_fixed_value[i]);
  free(method_file->dist_fixed_data.global_tgt_indices_per_fixed_value);
}

static void
process_tgt_requests(struct interpolation_method_file * method_file) {

  // mark target location invalid to avoid this routine being called more than
  // once
  enum yac_location tgt_location = method_file->dist_link_data.tgt_location;
  method_file->dist_link_data.tgt_location = LOC_INVALID;

  struct communicator * comm = yac_get_domain_comm(method_file->domain);

  if (tgt_location != method_file->dist_tgt_indices.location) {

    if (yac_get_comm_rank(comm) == 0) {
        char const * str_tgt_location_wgt = yac_loc2str(tgt_location);
        char const * str_tgt_location_tgt =
          yac_loc2str(method_file->dist_tgt_indices.location);
        fprintf(stderr, "target location: %s (weight file) %s (interface)\n",
                str_tgt_location_wgt,   str_tgt_location_tgt);
    }
    yac_internal_abort_message("ERROR(process_tgt_requests): target location "
                               "in file does not match with the location "
                               "provided through the interface\n",
                               __FILE__, __LINE__);
  }

  struct intermediate_results * results = malloc(1 * sizeof(*results));

  results->method_file = method_file;
  results->links = NULL;
  results->num_links = (unsigned)-1;
  results->link_flags = NULL;
  results->num_fixed_tgt = (unsigned)-1;
  results->num_fixed_values = method_file->dist_fixed_data.num_fixed_values;
  results->fixed_values = method_file->dist_fixed_data.fixed_values;

  process_tgt_requests_fixed(results);
  process_tgt_requests_links(results);

  method_file->dist_link_data.count = (unsigned)-1;
  method_file->dist_fixed_data.num_fixed_values = (unsigned)-1;
  method_file->dist_tgt_indices.count = (unsigned)-1;
}

static void
remove_invalid_links(struct link_data * link_data, unsigned * num_links) {

  unsigned old_num_links = *num_links;
  unsigned flag = 0;

  // check for links for which not all requires source points are available
  // -> mark them
  for (unsigned i = 0; i < old_num_links; ++i) {

    // if the source point for the link is not available on any source process
    if (link_data[i].rank == (unsigned)-1) {

      flag = 1;

      int tgt_idx = link_data[i].tgt_idx;

      // mark all links with the same target index as: "to be removed"
      for (unsigned j = i - 1;
           (j < old_num_links) && (link_data[j].tgt_idx == tgt_idx); --j)
        link_data[j].rank = -1;
      for (++i;(i < old_num_links) && (link_data[i].tgt_idx == tgt_idx); ++i)
        link_data[i].rank = -1;
      --i;
    }
  }

  // if there are invalid links
  if (flag) {
    unsigned new_num_links = 0;

    for (unsigned i = 0; i < old_num_links; ++i) {

      if (link_data[i].rank == (unsigned)-1) continue;

      if (new_num_links != i) link_data[new_num_links] = link_data[i];
      ++new_num_links;
    }

    *num_links = new_num_links;
  }
}

static void
link_data_alltoall_2_callback(struct communicator * comm,
                              struct comm_message msg, void * user_data) {

  struct interpolation_method_file * method_file = 
    (struct interpolation_method_file *)user_data;

  struct link_data * links = NULL;
  size_t links_array_size = 0;
  unsigned num_links = 0;

  unsigned offset = 0;

  while (offset < msg.count) {

    ENSURE_ARRAY_SIZE(links, links_array_size, num_links + 128);

    for (unsigned i = 0; (i < 128) && (offset < msg.count); ++i, ++num_links) {

      char * curr_packing_unit = (char*)msg.buffer + offset;
      unsigned packing_unit_size = msg.count - offset;
      int position = 0;

      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      &(links[num_links].src_idx), 1, COMM_INT, comm);
      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      &(links[num_links].tgt_idx), 1, COMM_INT, comm);
      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      &(links[num_links].rank), 1, COMM_UINT, comm);
      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      &(links[num_links].points_idx), 1, COMM_UINT, comm);
      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      &(links[num_links].weight), 1, COMM_DBLE, comm);

      offset += (unsigned)position;
    }
  }

  // sort links based on target indices
  qsort(links, num_links, sizeof(*links), compare_link_data_tgt_idx);

  // remove links for target points for which at least one source point is not
  // available
  remove_invalid_links(links, &num_links);

  method_file->dist_link_data.links =
    realloc(links, num_links * sizeof(*links));
  method_file->dist_link_data.count = num_links;

  // if the target requests, target location (from file) and fixed target points
  // have already been distributed
  if ((method_file->dist_tgt_indices.count != (unsigned)-1) &&
      (method_file->dist_link_data.tgt_location != LOC_INVALID) &&
      (method_file->dist_fixed_data.num_fixed_values != (unsigned)-1))
    process_tgt_requests(method_file);
}

static void
finalise_dist_link_data(struct intermediate_method_file_data_link * temp_data) {

  unsigned link_idx = 0;
  unsigned num_links = temp_data->dist_link_data.count;
  struct link_data * links = temp_data->dist_link_data.links;

  // sort links based on location and source indices
  qsort(temp_data->dist_link_data.links, num_links,
        sizeof(*temp_data->dist_link_data.links), compare_link_data_src_idx);

  // determine owners of the source indices of all links
  while (link_idx < num_links) {

    unsigned curr_points_idx = links[link_idx].points_idx;
    unsigned location_idx;
    for (location_idx = 0; location_idx < temp_data->num_locations;
         ++location_idx)
      if (temp_data->dist_src_indices[location_idx].location ==
          temp_data->method_file->src_locations[curr_points_idx]) break;

    if (temp_data->dist_src_indices[location_idx].location !=
        temp_data->method_file->src_locations[curr_points_idx])
      yac_internal_abort_message(
        "ERROR(finalise_dist_link_data): invalid location", 
        __FILE__, __LINE__);

    unsigned src_idx = 0;
    unsigned num_src_indices = temp_data->dist_src_indices[location_idx].count;
    struct index_data * src_indices =
      temp_data->dist_src_indices[location_idx].data;

    while ((src_idx < num_src_indices) && (link_idx < num_links) &&
           (curr_points_idx == links[link_idx].points_idx)) {

      if (src_indices[src_idx].index < links[link_idx].src_idx) {
        src_idx++;
      } else if (src_indices[src_idx].index == links[link_idx].src_idx) {
        links[link_idx].rank = src_indices[src_idx].rank;
        link_idx++;
      } else {
        links[link_idx].rank = -1; // no process has the respective source index
        link_idx++;
      }
    }

    if ((link_idx < num_links) &&
        (curr_points_idx == links[link_idx].points_idx)) break;
  }

  for (; link_idx < num_links; ++link_idx)
    links[link_idx].rank = -1; // no process has the respective source index

  struct communicator * comm =
    yac_get_domain_comm(temp_data->method_file->domain);
  unsigned comm_size = yac_get_comm_size(comm);

  unsigned link_pack_size = yac_comm_pack_size(2, COMM_INT, comm);
  link_pack_size += yac_comm_pack_size(2, COMM_UINT, comm);
  link_pack_size += yac_comm_pack_size(1, COMM_DBLE, comm);

  unsigned * send_size_per_rank =
    calloc(comm_size, sizeof(*send_size_per_rank));

  for (unsigned i = 0; i < num_links; ++i) {
    unsigned rank = compute_bucket(links[i].tgt_idx, comm_size);
    send_size_per_rank[rank] += link_pack_size;
  }

  unsigned * offsets = malloc(comm_size * sizeof(*offsets));
  int * pack_sizes = calloc(comm_size, sizeof(*pack_sizes));

  unsigned accum = 0;
  for (unsigned i = 0; i < comm_size; ++i) {
    offsets[i] = accum;
    accum += send_size_per_rank[i];
  }

  unsigned send_buffer_size = num_links * link_pack_size;
  char * send_buffer = malloc(send_buffer_size);

  // pack links into the send buffer
  for (unsigned i = 0; i < num_links; ++i) {

    unsigned rank = compute_bucket(links[i].tgt_idx, comm_size);

    char * curr_packing_unit = send_buffer + offsets[rank] + pack_sizes[rank];
    unsigned packing_unit_size = send_size_per_rank[rank] - pack_sizes[rank];
    int position = 0;

    yac_comm_pack(&(links[i].src_idx), 1, COMM_INT, curr_packing_unit,
                  packing_unit_size, &position, comm);
    yac_comm_pack(&(links[i].tgt_idx), 1, COMM_INT, curr_packing_unit,
                  packing_unit_size, &position, comm);
    yac_comm_pack(&(links[i].rank), 1, COMM_UINT, curr_packing_unit,
                  packing_unit_size, &position, comm);
    yac_comm_pack(&(links[i].points_idx), 1, COMM_UINT, curr_packing_unit,
                  packing_unit_size, &position, comm);
    yac_comm_pack(&(links[i].weight), 1, COMM_DBLE, curr_packing_unit,
                  packing_unit_size, &position, comm);

    pack_sizes[rank] += position;
  }

  // the following is just to ensure that the packing does not leave any holes
  // in the send buffer (should theoretically never happen)
  char * to_buffer = send_buffer;
  for (unsigned i = 0; i < comm_size; ++i) {
    char * from_buffer = send_buffer + offsets[i];
    unsigned packed_size = pack_sizes[i];
    send_size_per_rank[i] = packed_size;
    // should never happen
    if (from_buffer != to_buffer) memmove(to_buffer, from_buffer, packed_size);
    to_buffer += packed_size;
  }

  // distribute the link data 
  yac_comm_alltoallv_callback(send_buffer, send_size_per_rank, COMM_PACKED,
                              temp_data->method_file->tags.link_data_alltoall_2,
                              comm, temp_data->method_file,
                              link_data_alltoall_2_callback, MSG_PRIO_HIGH);

  free(pack_sizes);
  free(offsets);
  free(send_size_per_rank);
  free(send_buffer);
  free(links);
  for (unsigned i = 0; i < temp_data->num_locations; ++i)
    free(temp_data->dist_src_indices[i].data);
  free(temp_data->dist_src_indices);
  free(temp_data);
}

static void
link_data_alltoall_1_callback(struct communicator * comm,
                              struct comm_message msg, void * user_data) {

  struct intermediate_method_file_data_link * temp_data = 
    (struct intermediate_method_file_data_link *)user_data;

  struct link_data * links = NULL;
  size_t links_array_size = 0;
  unsigned num_links = 0;

  unsigned offset = 0;
  while (offset < msg.count) {

    ENSURE_ARRAY_SIZE(links, links_array_size, num_links + 128);

    for (unsigned i = 0; (i < 128) && (offset < msg.count); ++i, ++num_links) {

      char * curr_packing_unit = (char*)msg.buffer + offset;
      unsigned packing_unit_size = msg.count - offset;
      int position = 0;

      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      &(links[num_links].src_idx), 1, COMM_INT, comm);
      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      &(links[num_links].tgt_idx), 1, COMM_INT, comm);
      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      &(links[num_links].points_idx), 1, COMM_UINT, comm);
      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      &(links[num_links].weight), 1, COMM_DBLE, comm);

      offset += (unsigned)position;
    }
  }

  temp_data->dist_link_data.links = realloc(links, num_links * sizeof(*links));
  temp_data->dist_link_data.count = num_links;

  // if the source indices have already been distributed
  for (unsigned i = 0; i < temp_data->num_locations; ++i)
    if (temp_data->dist_src_indices[i].count == (unsigned)-1) return;

  finalise_dist_link_data(temp_data);
}

static void tgt_location_bcast_callback(
  struct communicator * comm, struct comm_message msg, void * user_data) {

  UNUSED(comm);

  struct interpolation_method_file * method_file =
    (struct interpolation_method_file *)user_data;

  method_file->dist_link_data.tgt_location =
    yac_get_location(*(int*)(msg.buffer));

  // if the links, fixed target points and target requests have already
  // been distributed
  if ((method_file->dist_link_data.count != (unsigned)-1) &&
      (method_file->dist_tgt_indices.count != (unsigned)-1) &&
      (method_file->dist_fixed_data.num_fixed_values != (unsigned)-1))
    process_tgt_requests(method_file);
}

static void
setup_dist_link_data(struct intermediate_method_file_data_link * temp_data,
                     int ncid) {

  struct communicator * comm = temp_data->comm;
  unsigned comm_size = yac_get_comm_size(comm);

  // read link data (each IO process has only a section of the global data)
  unsigned num_links;
  struct link_data * links;
  enum yac_location tgt_location;
  char const * local_domain_name =
    yac_get_local_domain_name(temp_data->method_file->search);
  char const * remote_domain_name =
    yac_get_remote_domain_name(temp_data->method_file->search);
  read_link_data(&links, &num_links, temp_data->method_file->src_locations,
                 &tgt_location, temp_data->method_file->num_pointsets,
                 local_domain_name, remote_domain_name,
                 ncid, comm, temp_data->method_file->io_proc_step);

  // broadcast target point location from file
  int tmp_tgt_location = (int)tgt_location;
  unsigned root = 0;
  yac_comm_bcast_callback(&tmp_tgt_location, 1, COMM_INT, root,
                          temp_data->method_file->tags.tgt_location_bcast,
                          comm, temp_data->method_file,
                          tgt_location_bcast_callback, MSG_PRIO_HIGH);

  unsigned link_pack_size = yac_comm_pack_size(2, COMM_INT, comm);
  link_pack_size += yac_comm_pack_size(1, COMM_UINT, comm);
  link_pack_size += yac_comm_pack_size(1, COMM_DBLE, comm);

  unsigned * send_size_per_rank =
    calloc(comm_size, sizeof(*send_size_per_rank));
  for (unsigned i = 0; i < num_links; ++i) {

    links[i].rank = compute_bucket(links[i].src_idx, comm_size);
    send_size_per_rank[links[i].rank] += link_pack_size;
  }

  unsigned * offsets = malloc(comm_size * sizeof(*offsets));
  int * pack_sizes = calloc(comm_size, sizeof(*pack_sizes));
  unsigned accum = 0;
  for (unsigned i = 0; i < comm_size; ++i) {
    offsets[i] = accum;
    accum += send_size_per_rank[i];
  }

  unsigned send_buffer_size = link_pack_size * num_links;
  char * send_buffer = malloc(send_buffer_size);

  // pack links into the send buffer
  for (unsigned i = 0; i < num_links; ++i) {

    unsigned rank = links[i].rank;
    char * curr_packing_unit = send_buffer + offsets[rank] + pack_sizes[rank];
    unsigned packing_unit_size = send_size_per_rank[rank] - pack_sizes[rank];

    int position = 0;

    yac_comm_pack(&(links[i].src_idx), 1, COMM_INT, curr_packing_unit,
                  packing_unit_size, &position, comm);
    yac_comm_pack(&(links[i].tgt_idx), 1, COMM_INT, curr_packing_unit,
                  packing_unit_size, &position, comm);
    yac_comm_pack(&(links[i].points_idx), 1, COMM_UINT, curr_packing_unit,
                  packing_unit_size, &position, comm);
    yac_comm_pack(&(links[i].weight), 1, COMM_DBLE, curr_packing_unit,
                  packing_unit_size, &position, comm);

    pack_sizes[rank] += position;
  }

  // the following is just to ensure that the packing does not leave any holes
  // in the send buffer (should theoretically never happen)
  char * to_buffer = send_buffer;
  for (unsigned i = 0; i < comm_size; ++i) {
    char * from_buffer = send_buffer + offsets[i];
    unsigned packed_size = pack_sizes[i];
    send_size_per_rank[i] = packed_size;
    // should never happen
    if (from_buffer != to_buffer) memmove(to_buffer, from_buffer, packed_size);
    to_buffer += packed_size;
  }

  free(pack_sizes);
  free(offsets);
  free(links);

  // distribute the link data 
  yac_comm_alltoallv_callback(send_buffer, send_size_per_rank, COMM_PACKED,
                              temp_data->method_file->tags.link_data_alltoall_1,
                              comm, temp_data, link_data_alltoall_1_callback,
                              MSG_PRIO_HIGH);

  free(send_buffer);
  free(send_size_per_rank);
}

static void
fixed_data_alltoall_callback(struct communicator * comm,
                             struct comm_message msg,
                             void * user_data) {

  struct intermediate_method_file_data_fixed * temp_data =
    (struct intermediate_method_file_data_fixed *)user_data;

  int ** global_tgt_indices_per_fixed_value =
    calloc(temp_data->num_fixed_values,
           sizeof(*global_tgt_indices_per_fixed_value));
  unsigned * global_tgt_indices_per_fixed_value_array_size =
    calloc(temp_data->num_fixed_values,
           sizeof(*global_tgt_indices_per_fixed_value_array_size));
  unsigned * num_tgt_indices_per_fixed_value =
    calloc(temp_data->num_fixed_values,
           sizeof(*num_tgt_indices_per_fixed_value));

  unsigned offset = 0;

  while (offset < msg.count) {

    for (unsigned i = 0; i < temp_data->num_fixed_values; ++i) {

      int position = 0;
      char * curr_packing_unit = (char*)msg.buffer + offset;
      unsigned packing_unit_size = msg.count - offset;

      unsigned num_tgt_indices;
      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      &num_tgt_indices, 1, COMM_INT, comm);

      ENSURE_ARRAY_SIZE(global_tgt_indices_per_fixed_value[i],
                        global_tgt_indices_per_fixed_value_array_size[i],
                        num_tgt_indices_per_fixed_value[i] + num_tgt_indices);

      yac_comm_unpack(curr_packing_unit, packing_unit_size, &position,
                      global_tgt_indices_per_fixed_value[i] +
                      num_tgt_indices_per_fixed_value[i], num_tgt_indices,
                      COMM_UINT, comm);
      num_tgt_indices_per_fixed_value[i] += num_tgt_indices;

      offset += (int)position;
    }
  }

  free(global_tgt_indices_per_fixed_value_array_size);

  struct interpolation_method_file * method_file = temp_data->method_file;

  method_file->dist_fixed_data.num_fixed_values = temp_data->num_fixed_values;
  method_file->dist_fixed_data.fixed_values = temp_data->fixed_values;
  method_file->dist_fixed_data.num_tgt_indices_per_fixed_value =
    num_tgt_indices_per_fixed_value;
  method_file->dist_fixed_data.global_tgt_indices_per_fixed_value =
    global_tgt_indices_per_fixed_value;

  free(temp_data);

  // if the target requests, the links and target location (from file) have
  // already been distributed
  if ((method_file->dist_tgt_indices.count != (unsigned)-1) &&
      (method_file->dist_link_data.tgt_location != LOC_INVALID) &&
      (method_file->dist_link_data.count != (unsigned)-1))
    process_tgt_requests(method_file);
}

static void
fixed_data_fixed_value_bcast_callback(struct communicator * comm,
                                      struct comm_message msg,
                                      void * user_data) {

  struct intermediate_method_file_data_fixed * temp_data =
    (struct intermediate_method_file_data_fixed *)user_data;

  unsigned buffer_size =
    temp_data->num_fixed_values * sizeof(*temp_data->fixed_values);
  temp_data->fixed_values = malloc(buffer_size);
  memcpy(temp_data->fixed_values, msg.buffer, buffer_size);

  unsigned comm_size = yac_get_comm_size(comm);

  unsigned * send_size_per_rank =
    calloc(comm_size, sizeof(*send_size_per_rank));

  // distribute the fixed data 
  yac_comm_alltoallv_callback(NULL, send_size_per_rank, COMM_PACKED,
                              temp_data->method_file->tags.fixed_data_alltoall,
                              comm, temp_data, fixed_data_alltoall_callback,
                              MSG_PRIO_HIGH);

  free(send_size_per_rank);
}

static void
fixed_data_num_fixed_bcast_callback(struct communicator * comm,
                                    struct comm_message msg, void * user_data) {

  struct intermediate_method_file_data_fixed * temp_data =
    (struct intermediate_method_file_data_fixed *)user_data;

  temp_data->num_fixed_values = *(unsigned*)msg.buffer;

  unsigned root = 0;

  yac_comm_bcast_callback( NULL, temp_data->num_fixed_values, COMM_DBLE, root,
    temp_data->method_file->tags.fixed_data_fixed_value_bcast, comm, temp_data,
    fixed_data_fixed_value_bcast_callback, MSG_PRIO_HIGH);
}

static void
read_fixed_data(int ** global_tgt_indices, unsigned * num_tgt_indices,
                unsigned ** num_tgt_indices_per_fixed_value,
                double ** fixed_values, unsigned * num_fixed_values,
                int ncid, struct communicator * comm, unsigned io_proc_step) {

  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned comm_size = yac_get_comm_size(comm);

  unsigned num_io_procs, is_io_proc;

  get_io_proc_setup(comm_size, comm_rank, io_proc_step, &num_io_procs,
                    &is_io_proc);

  *global_tgt_indices = NULL;
  *num_tgt_indices = 0;
  *num_tgt_indices_per_fixed_value = NULL;
  *fixed_values = NULL;
  *num_fixed_values = 0;

  if (!is_io_proc) return;

#ifdef HAVE_C_NETCDF

  //\todo use parallel netcdf if available

  int status, var_id, dimid;

  // global attributes
  size_t str_fixed_len;
  char * str_fixed;
  var_id = NC_GLOBAL;
  status = nc_inq_attlen(ncid, var_id, "contains_fixed_dst", &str_fixed_len);
  if (status != NC_NOERR) handle_error(status);
  str_fixed = malloc(str_fixed_len + 1);
  str_fixed[str_fixed_len] = '\0';
  status = nc_get_att_text(ncid, var_id, "contains_fixed_dst", str_fixed);
  if (status != NC_NOERR) handle_error(status);

  if ((strlen("FALSE") == str_fixed_len) &&
      !strncmp("FALSE", str_fixed, str_fixed_len)) {

    free(str_fixed);

    return;
  } else if ((strlen("TRUE") != str_fixed_len) ||
             strncmp("TRUE", str_fixed, str_fixed_len))
    yac_internal_abort_message("ERROR(read_fixed_data): invalid global "
                               "attribute contains_fixed_dst", __FILE__,
                               __LINE__);
  free(str_fixed);

  // get number of fixed values
  size_t num_fixed_values_;
  status = nc_inq_dimid(ncid, "num_fixed_values", &dimid);
  if (status != NC_NOERR) handle_error(status);
  status = nc_inq_dimlen(ncid, dimid, &num_fixed_values_);
  if (status != NC_NOERR) handle_error(status);

  // get number for fixed target points
  size_t num_fixed_tgt;
  status = nc_inq_dimid(ncid, "num_fixed_dst", &dimid);
  if (status != NC_NOERR) handle_error(status);
  status = nc_inq_dimlen(ncid, dimid, &num_fixed_tgt);
  if (status != NC_NOERR) handle_error(status);

  *fixed_values = malloc(num_fixed_values_ * sizeof(*fixed_values));
  *num_tgt_indices_per_fixed_value =
    malloc(num_fixed_values_ * sizeof(*num_tgt_indices_per_fixed_value));
  *global_tgt_indices = malloc(num_fixed_tgt * sizeof(*global_tgt_indices));

  // read data
  status = nc_inq_varid(ncid, "fixed_values", &var_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_get_var_double(ncid, var_id, *fixed_values);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_varid(ncid, "num_dst_per_fixed_value", &var_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_get_var_int(ncid, var_id,
                          (int*)(*num_tgt_indices_per_fixed_value));
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_varid(ncid, "dst_address_fixed", &var_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_get_var_int(ncid, var_id, *global_tgt_indices);
  if (status != NC_NOERR) handle_error(status);

  for (unsigned i = 0; i < num_fixed_tgt; ++i)
    (*global_tgt_indices)[i]--;

  *num_tgt_indices = num_fixed_tgt;
  *num_fixed_values = num_fixed_values_;
#else
  UNUSED(ncid);
#endif
}

static void
setup_dist_fixed_data(struct intermediate_method_file_data_fixed * temp_data,
                      int ncid) {

  struct communicator * comm = temp_data->comm;
  unsigned comm_size = yac_get_comm_size(comm);
  unsigned comm_rank = yac_get_comm_rank(comm);

  unsigned num_io_procs, is_io_proc;
  get_io_proc_setup(comm_size, comm_rank, temp_data->method_file->io_proc_step,
                    &num_io_procs, &is_io_proc);

  // read link data (each IO process has only a section of the global data)
  int * global_tgt_indices;
  unsigned num_tgt_indices;
  unsigned * num_tgt_indices_per_fixed_value;
  double * fixed_values;
  unsigned num_fixed_values;
  read_fixed_data(&global_tgt_indices, &num_tgt_indices,
                  &num_tgt_indices_per_fixed_value, &fixed_values,
                  &num_fixed_values, ncid, comm,
                  temp_data->method_file->io_proc_step);

  // broadcast number of fixed values and the fixed values themselves
  unsigned root = 0;
  yac_comm_bcast_callback(
    &num_fixed_values, 1, COMM_UINT, root,
    temp_data->method_file->tags.fixed_data_num_fixed_bcast, comm, temp_data,
    (is_io_proc)?NULL:fixed_data_num_fixed_bcast_callback, MSG_PRIO_HIGH);
  if (!is_io_proc) return;
  yac_comm_bcast_callback(
    fixed_values, num_fixed_values, COMM_DBLE, root,
    temp_data->method_file->tags.fixed_data_fixed_value_bcast, comm, temp_data,
    NULL, MSG_PRIO_HIGH);

  temp_data->num_fixed_values = num_fixed_values;
  temp_data->fixed_values = fixed_values;

  unsigned * num_tgt_per_rank =
    calloc(comm_size * num_fixed_values, sizeof(*num_tgt_per_rank));
  unsigned uint_pack_size = yac_comm_pack_size(1, COMM_UINT, comm);
  unsigned int_pack_size = yac_comm_pack_size(1, COMM_INT, comm);
  unsigned * send_size_per_rank =
    calloc(comm_size, sizeof(*send_size_per_rank));
  unsigned send_buffer_size = 0;

  for (unsigned i = 0, k = 0; i < num_fixed_values; ++i) {
    for (unsigned j = 0; j < num_tgt_indices_per_fixed_value[i]; ++j, ++k) {

      unsigned rank = compute_bucket(global_tgt_indices[k], comm_size);
      send_size_per_rank[rank] += int_pack_size;
      num_tgt_per_rank[rank * num_fixed_values + i]++;
    }
  }
  for (unsigned i = 0; i < comm_size; ++i) {
    if (send_size_per_rank[i] > 0)
      send_size_per_rank[i] += uint_pack_size * num_fixed_values;
    send_buffer_size += send_size_per_rank[i];
  }

  int * positions = calloc(comm_size, sizeof(*positions));
  unsigned * offsets = malloc(comm_size * sizeof(*offsets));
  int * pack_sizes = calloc(comm_size, sizeof(*pack_sizes));

  int accum = 0;
  for (unsigned i = 0; i < comm_size; ++i) {
    offsets[i] = accum;
    accum += send_size_per_rank[i];
  }

  char * send_buffer = malloc(send_buffer_size);

  // pack global target indices
  for (unsigned i = 0, k = 0; i < num_fixed_values; ++i) {

    for (unsigned j = 0; j < comm_size; ++j) {
      positions[j] = 0;
      if (send_size_per_rank[j] > 0) {
        yac_comm_pack(num_tgt_per_rank + j * num_fixed_values + i, 1, COMM_UINT,
                      send_buffer + offsets[j] + pack_sizes[j],
                      send_size_per_rank[j] - pack_sizes[j], positions + j,
                      comm);
      }
    }

    for (unsigned j = 0; j < num_tgt_indices_per_fixed_value[i]; ++j, ++k) {

      unsigned rank = compute_bucket(global_tgt_indices[k], comm_size);

      yac_comm_pack(global_tgt_indices + k, 1, COMM_INT,
                    send_buffer + offsets[rank] + pack_sizes[rank],
                    send_buffer_size - pack_sizes[rank], positions + rank,
                    comm);
    }

    for (unsigned j = 0; j < comm_size; ++j) pack_sizes[j] += positions[j];
  }

  free(global_tgt_indices);
  free(num_tgt_indices_per_fixed_value);
  free(num_tgt_per_rank);

  // the following is just to ensure that the packing does not leave any holes
  // in the send buffer (should theoretically never happen)
  char * to_buffer = send_buffer;
  for (unsigned i = 0; i < comm_size; ++i) {
    char * from_buffer = send_buffer + offsets[i];
    unsigned packed_size = pack_sizes[i];
    send_size_per_rank[i] = packed_size;
    // should never happen
    if (from_buffer != to_buffer) memmove(to_buffer, from_buffer, packed_size);
    to_buffer += packed_size;
  }

  // distribute the fixed data 
  yac_comm_alltoallv_callback(send_buffer, send_size_per_rank, COMM_PACKED,
                              temp_data->method_file->tags.fixed_data_alltoall,
                              comm, temp_data, fixed_data_alltoall_callback,
                              MSG_PRIO_HIGH);

  free(send_buffer);
  free(pack_sizes);
  free(offsets);
  free(positions);
  free(send_size_per_rank);
}

static int compare_index_data_index (const void* a, const void* b) {
  
  if (((struct index_data*)a)->index != ((struct index_data*)b)->index)
    return (((struct index_data*)a)->index - ((struct index_data*)b)->index);
  else
    return ((((struct index_data*)b)->rank) - ((struct index_data*)a)->rank);
}

static void src_indices_alltoall_callback(
  struct communicator * comm, struct comm_message msg, void * user_data) {

  UNUSED(comm);

  struct intermediate_method_file_data_link * temp_data =
    (struct intermediate_method_file_data_link *)user_data;

  enum yac_location location;

  if (temp_data->method_file->tags.src_indices_alltoall_cell == msg.tag)
    location = CELL;
  else if (temp_data->method_file->tags.src_indices_alltoall_corner == msg.tag)
    location = CORNER;
  else if (temp_data->method_file->tags.src_indices_alltoall_edge == msg.tag)
    location = EDGE;
  else
    yac_internal_abort_message("ERROR(src_indices_alltoall_callback): "
                               "invalid location", __FILE__, __LINE__);

  unsigned dist_src_indices_idx;

  for (dist_src_indices_idx = 0;
       dist_src_indices_idx < temp_data->num_locations; ++dist_src_indices_idx)
    if (temp_data->dist_src_indices[dist_src_indices_idx].location ==
        location) break;

  if (dist_src_indices_idx == temp_data->num_locations)
    yac_internal_abort_message(
      "ERROR(src_indices_alltoall_callback): location not found",
      __FILE__, __LINE__);

  unsigned num_src_indices = msg.count / 2;

  temp_data->dist_src_indices[dist_src_indices_idx].data =
    malloc(num_src_indices *
           sizeof(*(temp_data->dist_src_indices[dist_src_indices_idx].data)));

  int * recv_buffer = msg.buffer;

  for (unsigned i = 0; i < num_src_indices; ++i) {
    temp_data->dist_src_indices[dist_src_indices_idx].data[i].index =
      recv_buffer[2 * i + 0];
    temp_data->dist_src_indices[dist_src_indices_idx].data[i].rank =
      recv_buffer[2 * i + 1];
  }

  // sort source indices
  qsort(temp_data->dist_src_indices[dist_src_indices_idx].data, num_src_indices,
        sizeof(*temp_data->dist_src_indices[dist_src_indices_idx].data),
        compare_index_data_index);

  // this indicates that the source indices have been distributed
  temp_data->dist_src_indices[dist_src_indices_idx].count = num_src_indices;

  // if the links have already been distributed
  if (temp_data->dist_link_data.count == (unsigned)-1) return;
  for (unsigned i = 0; i < temp_data->num_locations; ++i)
    if (temp_data->dist_src_indices[i].count == (unsigned)-1) return;

  finalise_dist_link_data(temp_data);
}

static void
setup_dist_src_indices(struct intermediate_method_file_data_link * temp_data,
                       enum yac_location location) {

  struct communicator * comm = temp_data->comm;
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned comm_size = yac_get_comm_size(comm);
  struct domain * domain = temp_data->method_file->domain;

  // distribute all non-halo global source indices among the source processes
  unsigned num_subdomains = yac_get_num_subdomains(domain);
  unsigned total_num_src_indices = 0;

  for (unsigned i = 0; i < num_subdomains; ++i) {
    total_num_src_indices += yac_get_subdomain_data_size(domain, i, location);
    total_num_src_indices -= yac_get_num_total_halo_points(domain, i, location);
  }

  int * src_indices = malloc(total_num_src_indices * sizeof(*src_indices));
  int * ranks = malloc(total_num_src_indices * sizeof(*ranks));
  unsigned * num_points_per_rank =
    calloc(comm_size, sizeof(*num_points_per_rank));

  // gather source indices
  for (unsigned i = 0, offset = 0; i < num_subdomains; ++i) {

    yac_get_non_halo_points(domain, i, location, src_indices + offset, NULL);
    offset += yac_get_subdomain_data_size(domain, i, location);
    offset -= yac_get_num_total_halo_points(domain, i, location);
  }

  for (unsigned i = 0; i < total_num_src_indices; ++i) {

    ranks[i] = compute_bucket(src_indices[i], comm_size);
    num_points_per_rank[ranks[i]] += 2;
  }

  // sort source indices based on the ranks they are assigned to
  yac_quicksort_index(ranks, total_num_src_indices, src_indices);

  int * send_buffer = malloc(2 * total_num_src_indices * sizeof(*send_buffer));

  for (unsigned i = 0; i < total_num_src_indices; ++i) {
    send_buffer[2 * i + 0] = src_indices[i];
    send_buffer[2 * i + 1] = comm_rank;
  }

  free(ranks);
  free(src_indices);

  unsigned tag;

  switch (location) {
    case (CELL):
      tag = temp_data->method_file->tags.src_indices_alltoall_cell;
      break;
    case (CORNER):
      tag = temp_data->method_file->tags.src_indices_alltoall_corner;
      break;
    case (EDGE):
      tag = temp_data->method_file->tags.src_indices_alltoall_edge;
      break;
    default:
      yac_internal_abort_message(
        "ERROR(setup_dist_src_indices): invalid location", __FILE__, __LINE__);
  };

  // distribute source indices among all processes
  yac_comm_alltoallv_callback(send_buffer, num_points_per_rank, COMM_INT,
                              tag, comm, temp_data,
                              src_indices_alltoall_callback, MSG_PRIO_HIGH);

  free(send_buffer);
  free(num_points_per_rank);
}

/**
 * constructor for interpolation method file
 */
struct interpolation_method *
yac_new_interpolation_method_file(struct global_search * search,
                                  struct domain * domain,
                                  char const * stack_name,
                                  enum yac_location * src_locations,
                                  unsigned num_pointsets,
                                  char const * file_name,
                                  unsigned io_proc_step) {

  struct interpolation_method_file * method_file =
    malloc(1 * sizeof(*method_file));

  method_file->vtable = &interpolation_method_file_vtable;
  method_file->finish_callback = NULL;
  method_file->finish_user_data = NULL;
  method_file->search = search;
  method_file->domain = domain;
  method_file->src_locations = malloc(num_pointsets * sizeof(*(src_locations)));
  memcpy(method_file->src_locations, src_locations,
         num_pointsets * sizeof(*(src_locations)));
  method_file->num_pointsets = num_pointsets;
  method_file->dist_tgt_indices.count = (unsigned)-1;
  method_file->dist_tgt_indices.location = LOC_INVALID;
  method_file->dist_link_data.count = (unsigned)-1;
  method_file->dist_link_data.tgt_location = LOC_INVALID;
  method_file->dist_fixed_data.num_fixed_values = (unsigned)-1;
  method_file->io_proc_step = io_proc_step;

  yac_init_interpolation_method((struct interpolation_method *)method_file,
                                search);

  generate_interpolation_method_file_tag(method_file, stack_name, file_name,
                                         &(method_file->tags));

  struct communicator * comm = yac_get_domain_comm(method_file->domain);

  struct intermediate_method_file_data_link * temp_data_link =
    malloc(1 * sizeof(*temp_data_link));

  temp_data_link->method_file = method_file;
  temp_data_link->comm = comm;
  temp_data_link->dist_link_data.count = (unsigned)-1;

  {
    unsigned cell_flag = 0, corner_flag = 0, edge_flag = 0;
    for (unsigned i = 0; i < num_pointsets; ++i) {
      switch (src_locations[i]) {
        case (CELL):   cell_flag = 1;   break;
        case (CORNER): corner_flag = 1; break;
        case (EDGE):   edge_flag = 1;   break;
        default: yac_internal_abort_message("ERROR: invalid src_location", __FILE__, __LINE__);
      };
    }
    temp_data_link->num_locations = cell_flag + corner_flag + edge_flag;
    temp_data_link->dist_src_indices =
      malloc(temp_data_link->num_locations *
             sizeof(*(temp_data_link->dist_src_indices)));
    if (cell_flag)
      temp_data_link->dist_src_indices[0].location = CELL;
    if (corner_flag)
      temp_data_link->dist_src_indices[cell_flag].location = CORNER;
    if (edge_flag)
      temp_data_link->dist_src_indices[cell_flag+corner_flag].location = CORNER;
  }

  // set up a distributed directory for all global indices and their owners in
  // the source domain
  for (unsigned i = 0; i < temp_data_link->num_locations; ++i)
    temp_data_link->dist_src_indices[i].count = (unsigned)-1;
  for (unsigned i = 0; i < temp_data_link->num_locations; ++i)
    setup_dist_src_indices(temp_data_link, src_locations[i]);

  unsigned num_io_procs, is_io_proc;
  unsigned comm_size = yac_get_comm_size(comm);
  unsigned comm_rank = yac_get_comm_rank(comm);
  get_io_proc_setup(
    comm_size, comm_rank, io_proc_step, &num_io_procs, &is_io_proc);

  int ncid;
#ifndef HAVE_C_NETCDF
  ncid = -1;
  yac_internal_abort_message("ERROR(yac_new_interpolation_method_file): "
                             "HAVE_C_NETCDF not defined", __FILE__, __LINE__);
#else
  // open weight file on io processes
  if (is_io_proc) {
    int status = nc_open(file_name, NC_NOWRITE, &ncid);
    if (status != NC_NOERR) handle_error(status);
  }
#endif

  // set up a distributed directory for all links from the weight files
  setup_dist_link_data(temp_data_link, ncid);

  struct intermediate_method_file_data_fixed * temp_data_fixed =
    malloc(1 * sizeof(*temp_data_fixed));

  temp_data_fixed->method_file = method_file;
  temp_data_fixed->comm = comm;

  // set up a distributed directory for all fixed target points from the weight
  // file
  setup_dist_fixed_data(temp_data_fixed, ncid);

#ifdef HAVE_C_NETCDF
  // close weight file on io processes
  if (is_io_proc) {
    int status = nc_close(ncid);
    if (status != NC_NOERR) handle_error(status);
  }
#endif

  return (struct interpolation_method *)method_file;
}

static void tgt_indices_alltoall_callback(
  struct communicator * comm, struct comm_message msg, void * user_data) {

  UNUSED(comm);

  struct interpolation_method_file * method_file =
    (struct interpolation_method_file *)user_data;

  unsigned num_tgt_indices = msg.count / 2;

  method_file->dist_tgt_indices.data =
    malloc(num_tgt_indices *
           sizeof(*(method_file->dist_tgt_indices.data)));

  int * recv_buffer = msg.buffer;

  for (unsigned i = 0; i < num_tgt_indices; ++i) {
    method_file->dist_tgt_indices.data[i].index = recv_buffer[2 * i + 0];
    method_file->dist_tgt_indices.data[i].rank = recv_buffer[2 * i + 1];
  }

  // sort source indices
  qsort(method_file->dist_tgt_indices.data, num_tgt_indices,
        sizeof(*method_file->dist_tgt_indices.data), compare_index_data_index);

  // this indicates that the source indices have been distributed
  method_file->dist_tgt_indices.count = num_tgt_indices;

  // if the links, fixed target points and the target location(from file) have
  // already been distributed
  if ((method_file->dist_link_data.count != (unsigned)-1) &&
      (method_file->dist_link_data.tgt_location != LOC_INVALID) &&
      (method_file->dist_fixed_data.num_fixed_values != (unsigned)-1))
    process_tgt_requests(method_file);
}

static int compare_tgt_reorder_data(const void * a, const void * b) {

  if (((struct tgt_reorder_data*)a)->rank !=
      ((struct tgt_reorder_data*)b)->rank)
    return (((struct tgt_reorder_data*)a)->rank -
            ((struct tgt_reorder_data*)b)->rank);
  else
    return (((struct tgt_reorder_data*)a)->index -
            ((struct tgt_reorder_data*)b)->index);
}

static void
distribute_tgt_indices(struct interpolation_method_file * method_file,
                       unsigned num_global_tgt_ids) {

  struct communicator * comm = yac_get_domain_comm(method_file->domain);
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned comm_size = yac_get_comm_size(comm);

  struct tgt_reorder_data * tgt_data =
    malloc(num_global_tgt_ids * sizeof(*tgt_data));
  unsigned * num_points_per_rank =
    calloc(comm_size, sizeof(*num_points_per_rank));

  unsigned have_coordinates = (method_file->coordinates_xyz != NULL);

  for (unsigned i = 0; i < num_global_tgt_ids; ++i) {

    tgt_data[i].rank =
      compute_bucket(method_file->global_tgt_ids[i], comm_size);
    tgt_data[i].index = method_file->global_tgt_ids[i];
    if (have_coordinates) {
      tgt_data[i].coord[0] = method_file->coordinates_xyz[i][0];
      tgt_data[i].coord[1] = method_file->coordinates_xyz[i][1];
      tgt_data[i].coord[2] = method_file->coordinates_xyz[i][2];
    }
    num_points_per_rank[tgt_data[i].rank] += 2;
  }

  // sort the target data first based on rank and second on index
  qsort(tgt_data, num_global_tgt_ids, sizeof(*tgt_data),
        compare_tgt_reorder_data);

  int * send_buffer = malloc(2 * num_global_tgt_ids * sizeof(*send_buffer));

  for (unsigned i = 0; i < num_global_tgt_ids; ++i) {

    method_file->global_tgt_ids[i] = tgt_data[i].index;
    if (have_coordinates) {
      method_file->coordinates_xyz[i][0] = tgt_data[i].coord[0];
      method_file->coordinates_xyz[i][1] = tgt_data[i].coord[1];
      method_file->coordinates_xyz[i][2] = tgt_data[i].coord[2];
    }
    send_buffer[2 * i + 0] = tgt_data[i].index;
    send_buffer[2 * i + 1] = comm_rank;
  }

  free(tgt_data);

  // distribute source indices among all processes
  yac_comm_alltoallv_callback(send_buffer, num_points_per_rank, COMM_INT,
                              method_file->tags.tgt_indices_alltoall,
                              comm, method_file, tgt_indices_alltoall_callback,
                              MSG_PRIO_HIGH);

  free(send_buffer);
  free(num_points_per_rank);
}

static void
do_search_interpolation_method_file(struct interpolation_method ** method,
                                    struct interpolation_plan_element ** plan_stack,
                                    int * global_ids, double (*coordinates_xyz)[3],
                                    enum yac_location tgt_location,
                                    unsigned num_points,
                                    unsigned * num_interpolated_points,
                                    struct interpolation_method_data_request ** data_request) {

  struct interpolation_method_file * method_file =
    (struct interpolation_method_file*)(method[0]);

  method_file->method_stack = method;
  method_file->plan_stack = plan_stack;
  method_file->global_tgt_ids = global_ids;
  method_file->coordinates_xyz = coordinates_xyz;
  method_file->num_global_tgt_ids = num_points;
  method_file->num_interpolated_points = num_interpolated_points;
  method_file->data_request = data_request;
  method_file->dist_tgt_indices.location = tgt_location;

  //---------------------------------------------------------------------------
  // lookup target points in distributed link data and initiate lookup of links
  //---------------------------------------------------------------------------
  distribute_tgt_indices(method_file, num_points);
}

static void
enddef_interpolation_method_file(struct interpolation_method ** method) {

  struct interpolation_method_file * method_file =
    (struct interpolation_method_file*)(method[0]);
  struct communicator * comm = yac_get_domain_comm(method_file->domain);

  while(yac_comm_wait(comm, method_file->tags.tgt_indices_alltoall) ||
        yac_comm_wait(comm, method_file->tags.src_indices_alltoall_cell) ||
        yac_comm_wait(comm, method_file->tags.src_indices_alltoall_corner) ||
        yac_comm_wait(comm, method_file->tags.src_indices_alltoall_edge) ||
        yac_comm_wait(comm, method_file->tags.link_data_alltoall_1) ||
        yac_comm_wait(comm, method_file->tags.link_data_alltoall_2) ||
        yac_comm_wait(comm, method_file->tags.link_results_alltoall) ||
        yac_comm_wait(comm, method_file->tags.fixed_data_alltoall) ||
        yac_comm_wait(comm, method_file->tags.fixed_data_fixed_value_bcast) ||
        yac_comm_wait(comm, method_file->tags.fixed_data_num_fixed_bcast) ||
        yac_comm_wait(comm, method_file->tags.fixed_results_alltoall));

  if (method[1] != NULL)
    method[1]->vtable->enddef(method + 1);
}

static void
get_source_exchange_data_interpolation_method_file(
   struct interpolation_method ** method,
   struct interpolation_method_data_request *** data_request) {

   struct interpolation_method_file * method_file;

   method_file = (struct interpolation_method_file*)(method[0]);

   *data_request = method_file->data_request;
   method_file->data_request = NULL;
}

static
void delete_interpolation_method_file(struct interpolation_method * method) {

  struct interpolation_method_file * method_file =
    (struct interpolation_method_file*)method;

  free(method_file->src_locations);
  free(method);
}

static struct global_search *
get_global_search(struct interpolation_method * method) {

  return ((struct interpolation_method_file*)(method))->search;
}

static void
set_finish_callback(struct interpolation_method * method,
                    void (*callback)(void * user_data),
                    void * user_data) {

  struct interpolation_method_file * method_file;

  method_file = (struct interpolation_method_file*)method;

  method_file->finish_callback = callback;
  method_file->finish_user_data = user_data;
}

