/**
 * @file config_xml.h
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *                                 Teresa Holfeld <teresa.holfeld@zmaw.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 *         Teresa Holfeld <teresa.holfeld@zmaw.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 *             Teresa Holfeld <teresa.holfeld@zmaw.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#ifndef CONFIG_XML_H
#define CONFIG_XML_H

#include "yac_config.h"

#ifdef USE_XML2
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xmlschemas.h>
#endif
#include "interpolation_enum.h"



/** @file config_xml.h
 *
 * Interface for retrieving values given in coupling configuration XML file.
 * Parses XML file and provides interface function for accessing all values and
 * attributes.
 *
 * XML file must validate against schema coupling.xsd.
 *
 * For some examples on how to use the routines listed hear see:
 * \ref config_xml_example_doc
 */

/** \example test_config_xml.c
 * This example shows how to use the XML interface routines
 */

//* typedefs, enums ***********************************************************/


/// type for component, transient, and yac_xml_grid
typedef struct Yac_xml_element_info {
   char * name;
   int global_id;
} yac_xml_component, yac_xml_transient, yac_xml_grid;

/// enum for time unit
typedef enum Yac_xml_time_unit {
  T_MILLISECOND, T_SECOND, T_MINUTE, T_HOUR, T_DAY, T_MONTH, T_YEAR, T_ISO_FORMAT, T_NO_FORMAT
} yac_xml_timestep_unit;

/// enum for time yac_xml_operation in coupling period
typedef enum Yac_xml_time_operation {
   T_ACCUMULATE, T_AVERAGE, T_NONE
} yac_xml_operation;

/// enum for calendar
typedef enum Yac_xml_calendar {
	C_PROLEPTIC_GREGORIAN, C_REGULAR_360D, C_REGULAR_365D
} yac_xml_calendar;



//* init and destroy **********************************************************/


/**
 * Parses the given xml file and initializes the document tree.
 *
 * @throws error if xml or schema file does not exist
 * @throws error if xml file could not be parsed, or parsing failed
 * @throws error if xml file does not validate against schema
 * @param xml_filename     filename of xml instance with relative path
 * @param schema_filename  filename of xml schema with relative path
 */
void yac_xml_init_config_xml(const char * xml_filename, const char * schema_filename);

/**
 * Frees corresponding xmlChar in xml tree.
 *
 * @param string  pointer to char in xml tree
 */
void yac_xml_free_xml_string(const char * string);

/**
 * Frees corresponding xmlChar in xml tree.
 *
 * @param component  array of yac_xml_component containing pointer to char in xml tree
 */
void yac_xml_free_xml_component(yac_xml_component * component);

/**
 * Frees corresponding xmlChar in xml tree.
 *
 * @param components  array of yac_xml_component containing pointers to char in xml tree
 */
void yac_xml_free_xml_components(yac_xml_component * components);

/**
 * Frees corresponding xmlChar in xml tree.
 *
 * @param transient  array of yac_xml_transient containing pointer to char in xml tree
 */
void yac_xml_free_xml_transient(yac_xml_transient * transient);

/**
 * Frees corresponding xmlChar in xml tree.
 *
 * @param transients  array of yac_xml_transient containing pointers to char in xml tree
 * @param n           size of array
 */
void yac_xml_free_xml_transients(yac_xml_transient * transients, const int n);

/**
 * Frees corresponding xmlChar in xml tree.
 *
 * @param grid  array of yac_xml_grid containing pointer to char in xml tree
 */
void yac_xml_free_xml_grid(yac_xml_grid * grid);

/**
 * Frees corresponding xmlChar in xml tree.
 *
 * @param grids  array of yac_xml_grid containing pointers to char in xml tree
 * @param n      size of array
 */
void yac_xml_free_xml_grids(yac_xml_grid * grids, const int n);

/**
 * Frees the document tree and sets the pointers of document and coupling to \c
 * NULL.
 */
void yac_xml_destroy_config_xml();



//* interface functions for \<components\> ************************************/

/**
 * @returns 1 if redirect_stdout is existing and set to true, 0 otherwise.
 */
int yac_xml_redirect_stdout();

/**
 * @returns number of \<component\> elements
 */
int yac_xml_get_number_of_components();

/**
 * Checks if \<component\> element with given \<name\> value exists in
 * coupling xml.
 *
 * @param component_name   name of component as specified in element \<component\>\<name\>
 * @returns 1 if true, 0 if false
 */
int yac_xml_component_name_exists_in_xml(const char * component_name);

/**
 * Checks if \<component\> element with given ID exists in coupling xml.
 *
 * @param component_id  ID specifying a \<component\> element
 * @returns 1 if true, 0 if false
 */
int yac_xml_component_id_exists_in_xml(const int component_id);

/**
 * Gets ID of given component.
 *
 * Make sure that component_index is in range of [0,n) where n is the number of
 * components, e.g. retrieved by \ref yac_xml_get_number_of_components. If the
 * value exceeds this range, an error will be thrown.
 *
 * @param component_index  index specifying a \<component\> element, [0,n)
 * @returns value of id attribute of \<component\>
 */
int yac_xml_get_component_id(const int component_index);

/**
 * Gets ID of component with given name.
 *
 * @throws error if component_name is NULL or ""
 * @param component_name   name of component as specified in element \<component\>\<name\>
 * @returns value of id attribute of \<component\>
 * @returns -1 if there is no such \<component\> with given name
 */
int yac_xml_get_component_id_by_component_name(const char * component_name);

/**
 * Gets ID and name of given component.
 * Sets attributes of given yac_xml_component type.
 *
 * Make sure that component_index is in range of [0,n) where n is the number of
 * components, e.g. retrieved by \ref yac_xml_get_number_of_components. If the
 * value exceeds this range, an error will be thrown.
 *
 * After using it, \ref yac_xml_free_xml_component has to be called on
 * component.
 *
 * @param component_index  index specifying a \<component\> element, [0,n)
 * @param component        yac_xml_component to set attributes
 */
void yac_xml_get_component(const int component_index,
      yac_xml_component * component);

/**
 * Gets ID and name of component with given ID.
 * Sets attributes of given yac_xml_component type.
 *
 * Value of component_id can be retrieved by e.g. \ref yac_xml_get_component_id.
 *
 * After using it, \ref yac_xml_free_xml_component has to be called on
 * component.
 *
 * @param component_id  ID specifying a \<component\> element
 * @param component     yac_xml_component to set attributes
 */
void yac_xml_get_component_by_component_id(const int component_id,
      yac_xml_component * component);

/**
 * Gets ID of component with given name.
 * Sets attributes of given yac_xml_component type.
 *
 * After using it, \ref yac_xml_free_xml_component has to be called on
 * component.
 *
 * @throws error if component_name is NULL or ""
 * @throws error if there is no such \<component\> with given name
 * @param component_name   name of component as specified in element \<component\>\<name\>
 * @param component        yac_xml_component to set attributes
 */
void yac_xml_get_component_by_component_name(const char * component_name,
      yac_xml_component * component);

/**
 * Gets name of given component.
 *
 * Make sure that component_index is in range of [0,n) where n is the number of
 * components, e.g. retrieved by \ref yac_xml_get_number_of_components. If the
 * value exceeds this range, an error will be thrown.
 *
 * @param component_index  index specifying a \<component\> element, [0,n)
 * @returns content of \<name\> element in \<component\>
 */
const char * yac_xml_get_component_name(const int component_index);

/**
 * Gets name of component with given ID.
 *
 * Value of component_id can be retrieved by e.g. \ref yac_xml_get_component_id.
 *
 * @param component_id  ID specifying a \<component\> element
 * @returns content of \<name\> element in \<component\>
 */
const char * yac_xml_get_component_name_by_component_id(const int component_id);

/**
 * Gets model name of given component.
 *
 * Make sure that component_index is in range of [0,n) where n is the number of
 * components, e.g. retrieved by \ref yac_xml_get_number_of_components. If the
 * value exceeds this range, an error will be thrown.
 *
 * @param component_index  index specifying a \<component\> element, [0,n)
 * @returns content of \<model\> element in \<component\>
 */
const char * yac_xml_get_component_model(const int component_index);

/**
 * Gets model name of component with given ID.
 *
 * Value of component_id can be retrieved by e.g. \ref yac_xml_get_component_id.
 *
 * @param component_id  ID specifying a \<component\> element
 * @returns content of \<model\> element in \<component\>
 */
const char * yac_xml_get_component_model_by_component_id(
      const int component_id);

/**
 * Gets information of what is simulated in given component.
 *
 * Make sure that component_index is in range of [0,n) where n is the number of
 * components, e.g. retrieved by \ref yac_xml_get_number_of_components. If the
 * value exceeds this range, an error will be thrown.
 *
 * @param component_index  index specifying a \<component\> element, [0,n)
 * @returns content of \<simulated\> element in \<component\>
 */
const char * yac_xml_get_component_simulated(const int component_index);

/**
 * Gets information of what is simulated in component with given ID.
 *
 * Value of component_id can be retrieved by e.g. \ref yac_xml_get_component_id.
 *
 * @param component_id  ID specifying a \<component\> element
 * @returns content of \<simulated\> element in \<component\>
 */
const char * yac_xml_get_component_simulated_by_component_id(
      const int component_id);

/**
 * Gets number of transient_grid_ref elements in given component.
 *
 * Make sure that component_index is in range of [0,n) where n is the number of
 * components, e.g. retrieved by \ref yac_xml_get_number_of_components. If the
 * value exceeds this range, an error will be thrown.
 *
 * @param component_index  index specifying a \<component\> element, [0,n)
 * @returns number of \<transient_grid_ref\> child elements of \<component\>
 */
int yac_xml_get_number_of_transient_grid_refs(const int component_index);

/**
 * Gets number of transient_grid_ref elements in component with given ID.
 *
 * Value of component_id can be retrieved by e.g. \ref yac_xml_get_component_id.
 *
 * @param component_id  ID specifying a \<component\> element
 * @returns number of \<transient_grid_ref\> child elements of \<component\>
 */
int yac_xml_get_number_of_transient_grid_refs_by_component_id(
      const int component_id);

/**
 * Gets the ID of transient_grid_ref element at given transient_grid_ref_index
 * in component at given component_index.
 *
 * Make sure that component_index is in range of [0,n) where n is the number of
 * components, e.g. retrieved by \ref yac_xml_get_number_of_components. If the
 * value exceeds this range, an error will be thrown.
 *
 * Make sure that transient_grid_ref_index is in range of [0,m) where m is the
 * number of transient_grid_ref elements in the given component, e.g. retrieved
 * by \ref yac_xml_get_number_of_transient_grid_refs. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param component_index           index specifying a \<component\> element, [0,n)
 * @param transient_grid_ref_index  index specifying a \<transient_grid_ref\> element, [0,m)
 * @returns value of parameter "id" in \<transient_grid_ref\> element
 */
int yac_xml_get_transient_grid_ref_id(const int component_index,
      const int transient_grid_ref_index);

/**
 * Gets the ID of transient_grid_ref element at given transient_grid_ref_index
 * in component with given ID.
 *
 * Value of component_id can be retrieved by e.g. \ref yac_xml_get_component_id.
 *
 * Make sure that transient_grid_ref_index is in range of [0,m) where m is the
 * number of transient_grid_ref elements in the given component, e.g. retrieved
 * by \ref yac_xml_get_number_of_transient_grid_refs. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param component_id  ID specifying a \<component\> element
 * @param transient_grid_ref_index  index specifying a \<transient_grid_ref\> element, [0,m)
 * @returns value of parameter "id" in \<transient_grid_ref\> element
 */
int yac_xml_get_transient_grid_ref_id_by_component_id(const int component_id,
      const int transient_grid_ref_index);

/**
 * Gets the ID of transient referenced by transient_grid_ref with given ID in
 * given component.
 *
 * Make sure that component_index is in range of [0,n) where n is the number of
 * components, e.g. retrieved by \ref yac_xml_get_number_of_components. If the
 * value exceeds this range, an error will be thrown.
 *
 * Value of transient_grid_ref_id can be retrieved by e.g.
 * \ref yac_xml_get_transient_grid_ref_id.
 *
 * @param component_index        index specifying a \<component\> element, [0,n)
 * @param transient_grid_ref_id  ID specifying a \<transient_grid_ref\> element
 * @returns value of parameter "transient_ref" in \<transient_grid_ref\> element
 */
int yac_xml_get_transient_id_by_transient_grid_ref_id(const int component_index,
      const int transient_grid_ref_id);

/**
 * Gets the ID of transient referenced by transient_grid_ref with given ID in
 * component with given ID.
 *
 * Value of component_id can be retrieved by e.g. \ref yac_xml_get_component_id.
 *
 * Value of transient_grid_ref_id can be retrieved by e.g.
 * \ref yac_xml_get_transient_grid_ref_id.
 *
 * @param component_id           ID specifying a \<component\> element
 * @param transient_grid_ref_id  ID specifying a \<transient_grid_ref\> element
 * @returns value of parameter "transient_ref" in \<transient_grid_ref\> element
 */
int yac_xml_get_transient_id_by_transient_grid_ref_id_by_component_id(
      const int component_id, const int transient_grid_ref_id);

/**
 * Gets the ID of grid referenced by transient_grid_ref with given ID in given
 * component.
 *
 * Make sure that component_index is in range of [0,n) where n is the number of
 * components, e.g. retrieved by \ref yac_xml_get_number_of_components. If the
 * value exceeds this range, an error will be thrown.
 *
 * Value of transient_grid_ref_id can be retrieved by e.g.
 * \ref yac_xml_get_transient_grid_ref_id.
 *
 * @param component_index        index specifying a \<component\> element, [0,n)
 * @param transient_grid_ref_id  ID specifying a \<transient_grid_ref\> element
 * @returns value of parameter "grid_ref" in \<transient_grid_ref\> element
 */
int yac_xml_get_grid_id_by_transient_grid_ref_id(const int component_index,
      const int transient_grid_ref_id);

/**
 * Gets the ID of yac_xml_grid referenced by transient_grid_ref with given ID in
 * component with given ID.
 *
 * Value of component_id can be retrieved by e.g. \ref yac_xml_get_component_id.
 *
 * Value of transient_grid_ref_id can be retrieved by e.g.
 * \ref yac_xml_get_transient_grid_ref_id.
 *
 * @param component_id           ID specifying a \<component\> element
 * @param transient_grid_ref_id  ID specifying a \<transient_grid_ref\> element
 * @returns value of parameter "grid_ref" in \<transient_grid_ref\> element
 */
int yac_xml_get_grid_id_by_transient_grid_ref_id_by_component_id(
      const int component_id, const int transient_grid_ref_id);



//* interface functions for \<transients\> ************************************/


/**
 * @returns number of \<transient\> elements
 */
int yac_xml_get_number_of_transients();

/**
 * Gets ID of transient with given transient standard name.
 *
 * @param transient_standard_name   as specified in attribute transient_standard_name of \<transient\> element
 * @returns ID specifying \<transient\> element, or -1 if element does not exist
 */
int yac_xml_get_transient_id_by_transient_standard_name(
      const char * transient_standard_name);

/**
 * Checks if \<transient\> element with given ID exists in coupling xml.
 *
 * @param transient_id  ID specifying a \<transient\> element
 * @returns 1 if true, 0 if false
 */
int yac_xml_transient_id_exists_in_xml(const int transient_id);

/**
 * Gets transient standard name of transient with given ID.
 *
 * After using it, \ref yac_xml_free_xml_string has to be called on returned
 * char *.
 *
 * @param transient_id  ID specifying a \<transient\> element
 * @returns value of attribute "transient_standard_name" of element
 * \<transient\>
 */
const char * yac_xml_get_transient_standard_name_by_transient_id(
      const int transient_id);

/**
 * Checks if \<transient\> element with given transient standard name exists in
 * coupling xml.
 *
 * @param transient_standard_name   as specified in attribute
 * transient_standard_name of \<transient\> element
 * @returns 1 if true, 0 if false
 */
int yac_xml_transient_standard_name_exists_in_xml(
      const char * transient_standard_name);

/**
 * Gets ID and transient standard name of given transient.
 * Sets attributes of given yac_xml_transient type.
 *
 * Make sure that transient_index is in range of [0,n) where n is the number of
 * transients, e.g. retrieved by \ref yac_xml_get_number_of_transients. If the
 * value exceeds this range, an error will be thrown.
 *
 * After using it, \ref yac_xml_free_xml_transient has to be called on
 * transient.
 *
 * @param transient_index  index specifying a \<transient\> element, [0,n)
 * @param transient        yac_xml_transient to set attributes
 */
void yac_xml_get_transient(const int transient_index,
      yac_xml_transient * transient);

/**
 * Gets ID and name of transient with given ID.
 * Sets attributes of given yac_xml_transient type.
 *
 * Value of transient_id can be retrieved by e.g.
 * \ref yac_xml_get_transient_id_by_transient_standard_name.
 *
 * After using it, \ref yac_xml_free_xml_transient has to be called on
 * transient.
 *
 * @param transient_id  ID specifying a \<transient\> element
 * @param transient     yac_xml_transient to set attributes
 */
void yac_xml_get_transient_by_id(const int transient_id,
      yac_xml_transient * transient);

/**
 * Gets IDs and transient standard names of all given transients in coupling
 * xml.
 * Sets attributes of given array of yac_xml_transient types.
 *
 * Make sure to allocate memory for as many transient types as given by
 * @ref yac_xml_get_number_of_transients<br>
 * Example: \code yac_xml_transient * transients = malloc(number_of_transients * sizeof(yac_xml_transient)); \endcode
 *
 * After using it, \ref yac_xml_free_xml_transients has to be called on
 * transients[].
 *
 * @param transients[]     yac_xml_transient[] to set attributes
 */
void yac_xml_get_all_transients(yac_xml_transient transients[]);



//* interface functions for \<grids\> *****************************************/


/**
 * @returns number of \<grid\> elements
 */
int yac_xml_get_number_of_grids();

/**
 * Checks if \<grid\> element with given ID exists in coupling xml.
 *
 * @param grid_id  ID specifying a \<grid\> element
 * @returns 1 if true, 0 if false
 */
int yac_xml_grid_id_exists_in_xml(const int grid_id);

/**
 * Gets alias name of grid with given ID.
 *
 * Value of grid_id can be retrieved by e.g. \ref yac_xml_get_grid.
 *
 * After using it, \ref yac_xml_free_xml_string has to be called on returned
 * char *.
 *
 * @param grid_id ID specifying a \<grid\> element
 */
const char * yac_xml_get_grid_alias_name_by_grid_id(const int grid_id);

/**
 * Gets ID and alias name of given grid.
 * Sets attributes of given yac_xml_grid type.
 *
 * Make sure that grid_index is in range of [0,n) where n is the number of
 * grids, e.g. retrieved by \ref yac_xml_get_number_of_grids.
 * If the value exceeds this range, an error will be thrown.
 *
 * After using it, \ref yac_xml_free_xml_grid has to be called on grid.
 *
 * @param grid_index  index specifying a \<grid\> element, [0,n)
 * @param grid        yac_xml_grid to set attributes
 */
void yac_xml_get_grid(const int grid_index, yac_xml_grid * grid);

/**
 * Gets ID and alias name of grid with given ID.
 * Sets attributes of given yac_xml_grid type.
 *
 * After using it, \ref yac_xml_free_xml_grid has to be called on grid.
 *
 * @param grid_id    ID specifying a \<grid\> element
 * @param grid       yac_xml_grid to set attributes
 */
void yac_xml_get_grid_by_id(const int grid_id, yac_xml_grid * grid);

/**
 * Gets IDs and alias names of all given grids in coupling xml.
 * Sets attributes of array given yac_xml_grid type.
 *
 * Make sure to allocate memory for as many transient types as given by
 * @ref yac_xml_get_number_of_grids<br>
 * Example: \code yac_xml_grid * grids = malloc(number_of_grids * sizeof(yac_xml_grid)); \endcode
 *
 * After using it, \ref yac_xml_free_xml_grids has to be called on grid.
 *
 * @param grids[]     yac_xml_grid[] to set attributes
 */
void yac_xml_get_all_grids(yac_xml_grid grids[]);



//* interface functions for \<dates\> and \<timestep_unit\> *******************/


/**
 * @returns value of \<start_date\> element in \<dates\>
 */
const char * yac_xml_get_start_date();

/**
 * @returns value of \<end_date\> element in \<dates\>
 */
const char * yac_xml_get_end_date();

/**
 * @returns value of \<calendar\> element in \<dates\>
 */
yac_xml_calendar yac_xml_get_calendar();

/**
 * @returns value of \<timestep_unit\> element
 */
yac_xml_timestep_unit yac_xml_get_timestep_unit();



//* interface functions for \<couples\> ***************************************/


/**
 * @returns number of \<component\> elements
 */
int yac_xml_get_number_of_couples();

/**
 * Gets global ID of first component of given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * @param couple_index  index specifying a \<couple\> element, [0,n)
 * @returns ID of referenced \<component\> element
 */
int yac_xml_get_first_component_id(const int couple_index);

/**
 * Gets global ID of second component of given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * @param couple_index  index specifying a \<couple\> element, [0,n)
 * @returns ID of referenced \<component\> element
 */
int yac_xml_get_second_component_id(const int couple_index);

/**
 * Gets number of transient couples of given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * @param couple_index  index specifying a \<couple\> element, [0,n)
 * @returns number of \<transient_couple\> elements in \<couple\>
 */
int yac_xml_get_number_of_transient_couples(const int couple_index);

/**
 * Gets ID of transient of given transient couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns ID of referenced \<transient\> element
 */
int yac_xml_get_transient_couple_transient_id(const int couple_index,
      const int transient_couple_index);

/**
 * Gets ID and transient standard name of transient of given transient couple.
 * Sets attributes of given yac_xml_transient type.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * After using it, \ref yac_xml_free_xml_transient has to be called on
 * transient.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @param transient              yac_xml_transient to set attributes
 * @returns ID of referenced \<transient\> element
 */
void yac_xml_get_transient_couple_transient(const int couple_index,
      const int transient_couple_index, yac_xml_transient * transient);

/**
 * Gets global ID of source component of given transient couple in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns ID of referenced \<component\> element listed as source
 */
int yac_xml_get_source_component_id(const int couple_index,
      const int transient_couple_index);

/**
 * Gets global ID of target component of given transient couple in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns ID of referenced \<component\> element listed as target
 */
int yac_xml_get_target_component_id(const int couple_index,
      const int transient_couple_index);

/**
 * Gets ID of referenced transient yac_xml_grid ref of source in given transient couple
 * in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns value of attribute 'transient_grid_ref' in \<source\> element
 */
int yac_xml_get_source_transient_grid_ref_id(const int couple_index,
      const int transient_couple_index);

/**
 * Gets ID of referenced transient_grid_ref of target in given transient couple
 * in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns value of attribute 'transient_grid_ref' in \<target\> element
 */
int yac_xml_get_target_transient_grid_ref_id(const int couple_index,
      const int transient_couple_index);

/**
 * Gets the source timestep of given transient couple in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns value of \<source\> element in \<timestep\> element
 */
int yac_xml_get_source_timestep(const int couple_index,
      const int transient_couple_index);

/**
 * Gets the target timestep of given transient couple in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns value of \<target\> element in \<timestep\> element
 */
int yac_xml_get_target_timestep(const int couple_index,
      const int transient_couple_index);

/**
 * Gets coupling period of given transient couple in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns content of \<coupling_period\> element in \<timestep\> element
 */
int yac_xml_get_coupling_period(const int couple_index,
      const int transient_couple_index);

/**
 * Gets the source timestep given in ISO Format of given transient couple in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns value of \<source\> element in \<timestep\> element
 */
char * yac_xml_get_source_isotimestep(const int couple_index,
      const int transient_couple_index);

/**
 * Gets the target timestep given in ISO Format of given transient couple in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns value of \<target\> element in \<timestep\> element
 */
char * yac_xml_get_target_isotimestep(const int couple_index,
      const int transient_couple_index);

/**
 * Gets coupling period given in ISO Format of given transient couple in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns content of \<coupling_period\> element in \<timestep\> element
 */
char * yac_xml_get_isocoupling_period(const int couple_index,
      const int transient_couple_index);

/**
 * Gets unit of coupling period of given transient couple in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * After using it, \ref yac_xml_free_xml_string has to be called on returned
 * char *.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns content of yac_xml_operation attribute of \<coupling_period\> element in \<timestep\> element
 */
yac_xml_operation yac_xml_get_coupling_period_operation(const int couple_index,
      const int transient_couple_index);

/**
 * Gets coupling time lag for a source of given transient couple in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns content of \<source_timelag\> element in \<timestep\> element
 */
int yac_xml_get_source_timelag(const int couple_index,
      const int transient_couple_index);

/**
 * Gets coupling time lag for a target of given transient couple in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns content of \<target_timelag\> element in \<timestep\> element
 */
int yac_xml_get_target_timelag(const int couple_index,
      const int transient_couple_index);

/**
 * Gets global ID of component that does the interpolation.
 * Refers to either source or target component in given transient couple in
 * given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns ID of \<component\> element referenced by 'interpolation_done_by'
 * attribute in \<interpolation_requirements\> element
 */
int yac_xml_get_component_id_of_interpolation_done_by(const int couple_index,
      const int transient_couple_index);

/**
 * Gets number of interpolation of given transient couple in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns number of \<interpolation\> elements in
 * \<couple\>\<transient_couple\>\<interpolation_requirements\>
 */
int yac_xml_get_number_of_interpolations(const int couple_index,
      const int transient_couple_index);

/**
 * Compares child elements of 2 \<interpolation_requirement\> elements.
 *
 * If the lists specify the same sequence of interpolation methods, the method
 * will return 1. This will be the case even if the parameters of same methods
 * specify different values.
 *
 * Make sure that couple_index_1 and couple_index_2 are in range of [0,n) where
 * n is the number of couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_couples.
 * If the value exceeds this range, an error will be thrown. couple_index_1 and
 * couple_index_2 can have the same value, ergo specifying the same couple
 * element.
 *
 * Make sure that transient_couple_index_1 and transient_couple_index_2 are in
 * range of [0,m) where m is the number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index_1            index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index_1  index specifying a \<transient_couple\> element, [0,m)
 * @param couple_index_2            index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index_2  index specifying a \<transient_couple\> element, [0,m)
 * @returns 1 if interpolation lists are equal, 0 otherwise
 */
int yac_xml_compare_interpolation_lists(const int couple_index_1,
      const int transient_couple_index_1, const int couple_index_2,
      const int transient_couple_index_2);

/**
 * Gets interpolation method of given interpolation in given transient couple
 * in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * Make sure that interpolation_index is in range of [0,k) where k is the
 * number of interpolations, e.g. retrieved by
 * \ref yac_xml_get_number_of_interpolations. If the value exceeds this range,
 * an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @param interpolation_index    index specifying a \<interpolation\> element [0,k]
 * @returns enum value of method attribute of \<interpolation\> element
 */
enum yac_interpolation_list yac_xml_get_interpolation_method(const int couple_index,
      const int transient_couple_index, const int interpolation_index);

/**
 * Checks if given interpolation in given transient couple in given couple has
 * parameter specified.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * Make sure that interpolation_index is in range of [0,k) where k is the
 * number of interpolations, e.g. retrieved by
 * \ref yac_xml_get_number_of_interpolations. If the value exceeds this range,
 * an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @param interpolation_index    index specifying a \<interpolation\> element [0,k]
 * @returns 1 if parameter was specified, 0 otherwise
 */
int yac_xml_interpolation_has_parameters(const int couple_index,
      const int transient_couple_index, const int interpolation_index);

/**
 * Checks if given interpolation in given transient couple in given couple has
 * a given parameter specified.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * Make sure that interpolation_index is in range of [0,k) where k is the
 * number of interpolations, e.g. retrieved by
 * \ref yac_xml_get_number_of_interpolations. If the value exceeds this range,
 * an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @param interpolation_index    index specifying a \<interpolation\> element [0,k]
 * @param parameter              name of attribute specifying the parameter
 * @returns 1 if parameter was specified, 0 otherwise
 */
int yac_xml_interpolation_has_parameter(const int couple_index,
      const int transient_couple_index, const int interpolation_index,
      const char * parameter);

/**
 * Gets the number of parameters of given interpolation in given transient
 * couple in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * Make sure that interpolation_index is in range of [0,k) where k is the
 * number of interpolations, e.g. retrieved by
 * \ref yac_xml_get_number_of_interpolations. If the value exceeds this range,
 * an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @param interpolation_index    index specifying a \<interpolation\> element [0,k]
 * @returns number of parameters
 */
int yac_xml_get_interpolation_number_of_parameters(const int couple_index,
      const int transient_couple_index, const int interpolation_index);

/**
 * Gets the name of the given parameter in given interpolation in given
 * transient couple in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * Make sure that interpolation_index is in range of [0,k) where k is the
 * number of interpolations, e.g. retrieved by
 * \ref yac_xml_get_number_of_interpolations. If the value exceeds this range,
 * an error will be thrown.
 *
 * Make sure that parameter_index is in range of [0,l) where l is the number of
 * parameters, e.g. retrieved by
 * \ref yac_xml_get_interpolation_number_of_parameters.
 * If the value exceeds this range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @param interpolation_index    index specifying a \<interpolation\> element, [0,k]
 * @param parameter_index        index specifying the parameter, [0,l)
 * @returns name of attribute specified by parameter in \<interpolation\> element
 */
const char * yac_xml_get_interpolation_parameter_name(const int couple_index,
      const int transient_couple_index, const int interpolation_index,
      const int parameter_index);

/**
 * Gets the value of the given parameter in given interpolation in given
 * transient couple in given couple.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * Make sure that interpolation_index is in range of [0,k) where k is the
 * number of interpolations, e.g. retrieved by
 * \ref yac_xml_get_number_of_interpolations. If the value exceeds this range,
 * an error will be thrown.
 *
 * Make sure specified parameter exists in interpolation, otherwise an error
 * will be thrown. Can be checked by \ref yac_xml_interpolation_has_parameter.
 *
 * After using it, \ref yac_xml_free_xml_string has to be called on returned
 * char *.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @param interpolation_index    index specifying a \<interpolation\> element [0,k]
 * @param parameter              name of attribute specifying the parameter
 * @returns value of attribute specified by parameter in \<interpolation\> element
 */
const char * yac_xml_get_interpolation_parameter_value_string(
      const int couple_index, const int transient_couple_index,
      const int interpolation_index, const char * parameter);

/**
 * Gets the integer value of the given parameter in given interpolation in given
 * transient couple in given couple.
 *
 * Only use this function if the expected type of the given parameter is of type
 * int. If it is not, an error will be thrown.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * Make sure that interpolation_index is in range of [0,k) where k is the
 * number of interpolations, e.g. retrieved by
 * \ref yac_xml_get_number_of_interpolations. If the value exceeds this range,
 * an error will be thrown.
 *
 * Make sure specified parameter exists in interpolation, otherwise an error
 * will be thrown. Can be checked by \ref yac_xml_interpolation_has_parameter
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @param interpolation_index    index specifying a \<interpolation\> element [0,k]
 * @param parameter              name of attribute specifying the parameter
 * @returns int value of attribute specified by parameter in \<interpolation\> element
 */
int yac_xml_get_interpolation_parameter_value_int(const int couple_index,
      const int transient_couple_index, const int interpolation_index,
      const char * parameter);

/**
 * Gets the double value of the given parameter in given interpolation in given
 * transient couple in given couple.
 *
 * Only use this function if the expected type of the given parameter is of type
 * double. If it is not, an error will be thrown.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * Make sure that interpolation_index is in range of [0,k) where k is the
 * number of interpolations, e.g. retrieved by
 * \ref yac_xml_get_number_of_interpolations. If the value exceeds this range,
 * an error will be thrown.
 *
 * Make sure specified parameter exists in interpolation, otherwise an error
 * will be thrown. Can be checked by \ref yac_xml_interpolation_has_parameter
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @param interpolation_index    index specifying a \<interpolation\> element [0,k]
 * @param parameter              name of attribute specifying the parameter
 * @returns double value of attribute specified by parameter in \<interpolation\> element
 */
double yac_xml_get_interpolation_parameter_value_double(
      const int couple_index, const int transient_couple_index,
      const int interpolation_index, const char * parameter);

/**
 * Checks if for the given transient couple in given couple use_source_mask is
 * set
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns 1 if \<use_source_mask\> element exists\n
 *          0 otherwise
 */
int yac_xml_use_source_mask_is_set(const int couple_index,
                                   const int transient_couple_index);

/**
 * Determines whether the source mask is to be applied to transient couple in
 * given couple.
 *
 * Make sure that element \<use_source_mask\> exists. This can be done with
 * \ref yac_xml_use_source_mask_is_set.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns whether the source mask is to be applied
 */
int yac_xml_get_use_source_mask(const int couple_index,
                                const int transient_couple_index);

/**
 * Checks if for the given transient couple in given couple use_target_mask is
 * set
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns 1 if \<use_target_mask\> element exists\n
 *          0 otherwise
 */
int yac_xml_use_target_mask_is_set(const int couple_index,
                                   const int transient_couple_index);

/**
 * Determines whether the target mask is to be applied to transient couple in
 * given couple.
 *
 * Make sure that element \<use_target_mask\> exists. This can be done with
 * \ref yac_xml_use_target_mask_is_set.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns whether the target mask is to be applied
 */
int yac_xml_get_use_target_mask(const int couple_index,
                                const int transient_couple_index);

/**
 * Checks if mapping is to be done at the source component or the target.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns 1 if content \<mapping_on_source\> is "true" or nothing is
            specified, 0 otherwise
 */
int yac_xml_mapping_on_source(const int couple_index,
      const int transient_couple_index);

/**
 * Checks if debug output is done at source before interpolation.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns 1 if value of attribute 'at_source_before_interpolation' of element
 * \<debug_mode\> is 'true', 0 otherwise
 */
int yac_xml_debug_mode_at_source_before_interpolation(const int couple_index,
      const int transient_couple_index);

/**
 * Checks if debug output is done at source after interpolation.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns 1 if value of attribute 'at_source_after_interpolation' of element
 * \<debug_mode\> is 'true', 0 otherwise
 */
int yac_xml_debug_mode_at_source_after_interpolation(const int couple_index,
      const int transient_couple_index);

/**
 * Checks if debug output is done at target.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns 1 if value of attribute 'at_target' of element \<debug_mode\> is
 * 'true', 0 otherwise
 */
int yac_xml_debug_mode_at_target(const int couple_index,
      const int transient_couple_index);

/**
 * Checks if transient couple enforces write restart.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns 1 if content \<enforce_write_restart\> is "true", 0 otherwise
 */
int yac_xml_enforce_write_restart(const int couple_index,
      const int transient_couple_index);

/**
 * Checks if transient couple enforces writing of weight files.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns 1 if content \<enforce_write_weight_file\> is "true", 0 otherwise
 */
int yac_xml_enforce_write_weight_file(const int couple_index,
      const int transient_couple_index);

/**
 * Gets the filename that is to be used to write the weight file.
 *
 * Make sure that couple_index is in range of [0,n) where n is the number of
 * couples, e.g. retrieved by \ref yac_xml_get_number_of_couples. If the value
 * exceeds this range, an error will be thrown.
 *
 * Make sure that transient_couple_index is in range of [0,m) where m is the
 * number of transient_couples, e.g. retrieved by
 * \ref yac_xml_get_number_of_transient_couples. If the value exceeds this
 * range, an error will be thrown.
 *
 * Make sure the filename is specified, otherwise an error will be thrown.
 *
 * After using it, \ref yac_xml_free_xml_string has to be called on the
 * filename.
 *
 * @param couple_index           index specifying a \<couple\> element, [0,n)
 * @param transient_couple_index index specifying a \<transient_couple\> element, [0,m)
 * @returns filename
 */
const char *  yac_xml_get_weight_file_name(const int couple_index,
      const int transient_couple_index);

/**
 * Gets the collection size of specific transient_grid_ref referenced by transient_grid_ref_id in
 * component with given ID.
 *
 * Value of component_id can be retrieved by e.g. \ref yac_xml_get_component_id.
 *
 * Value of transient_grid_ref_id can be retrieved by e.g.
 * \ref yac_xml_get_transient_grid_ref_id.
 *
 * @param component_id           ID specifying a \<component\> element
 * @param transient_grid_ref_id  ID specifying a \<transient_grid_ref\> element
 * @returns value of parameter "collection_size" in \<transient_grid_ref\> element
 */
int yac_xml_get_transient_collection_size(const int component_id, const int transient_grid_ref_id);


#endif /* CONFIG_XML_H */

