/**
 * @file yac_lapack_interface.h
 * @brief lapack includes and alternative interfaces
 *
 * This include file assembles the necessary include files to access the
 * lapack routines required by YAC. In YAC itself only calls to the
 * LAPACKE variant are used. If LAPACKE is not available the required interfaces
 * are provided here to give access to ATLAS clapack, Netlib clapack and the
 * original LAPACK Fortran functions via cfortran.h.
 *
 * - lapack/dgels.f
 * - lapack/dgesv.f
 * - lapack/dgetrf.f
 * - lapack/dgetri.f
 * - lapack/dsytrf.f
 * - lapack/dsytri.f
 *
 * @copyright Copyright  (C)  2017 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */
#ifndef YAC_LAPACK_INTERFACE_H
#define YAC_LAPACK_INTERFACE_H

#ifdef HAVE_CONFIG_H
// Get the definition of the lapack interface macros.
#include "config.h"
#else
// Use Fortran lapack if there is no config.h.
#define HAVE_FORTRAN_LAPACK 1
#endif

#if defined HAVE_MKL_LAPACKE
#  include <mkl_lapacke.h>
#elif defined HAVE_LAPACKE
#  include <lapacke.h>
#elif HAVE_ATLAS_CLAPACK
#  include <assert.h>
#  include <clapack.h>
#  define LAPACK_COL_MAJOR CblasColMajor
// see e.g. http://www.icl.utk.edu/files/publications/2017/icl-utk-950-2017.pdf ; July 2, 2017
//
// Note: CBLAS_ORDER was renamed CBLAS_LAYOUT around LAPACK 3.6.0.
// typedef  enum {CblasRowMajor =101,  CblasColMajor =102}  CBLAS_LAYOUT;
// typedef  enum {CblasNoTrans =111,  CblasTrans =112,  CblasConjTrans =113}  CBLAS_TRANSPOSE;
// typedef  enum {CblasUpper =121,  CblasLower =122}  CBLAS_UPLO;
// typedef  enum {CblasNonUnit =131,  CblasUnit =132}  CBLAS_DIAG;
// typedef  enum {CblasLeft =141,  CblasRight =142}  CBLAS_SIDE;
#  ifdef YAC_CBLAS_ENUM_TYPEDEF_MISSING
#    define CBLAS_LAYOUT enum CBLAS_ORDER
#    define CBLAS_TRANSPOSE enum CBLAS_ ## TRANSPOSE
#    define CBLAS_UPLO enum CBLAS_ ## UPLO
#  endif
static int
LAPACKE_dgels_work(int matrix_order, char trans, int M, int N,
                   int NRHS, double *A, int LDA, double *B, int LDB,
                   double *work, int lwork)
{
  assert(matrix_order == LAPACK_COL_MAJOR);
  CBLAS_TRANSPOSE transpose = trans == 'N' ? CblasNoTrans : CblasTrans;
  return clapack_dgels((CBLAS_LAYOUT)matrix_order, transpose,
                       M, N, NRHS, A, LDA, B, LDB);
}

static int
LAPACKE_dgesv(int matrix_order, int N, int NRHS,
               double *A, int LDA, int *IPIV,
               double *B, int LDB)
{
  assert(matrix_order == LAPACK_COL_MAJOR);
  return clapack_dgesv((CBLAS_LAYOUT)matrix_order,
                       N, NRHS, A, LDA, IPIV, B, LDB);
}

static int
LAPACKE_dgetrf(int matrix_order, int M, int N,
               double *A, int LDA, int *IPIV)
{
  assert(matrix_order == LAPACK_COL_MAJOR);
  return clapack_dgetrf((CBLAS_LAYOUT)matrix_order,
                        M, N, A, LDA, IPIV);
}

static int
LAPACKE_dgetri_work(int matrix_order, int N, double *A,
                    int LDA, const int *IPIV,
                    double *WORK, int LWORK)
{
  assert(matrix_order == LAPACK_COL_MAJOR);
  return clapack_dgetri((CBLAS_LAYOUT)matrix_order,
                        N, A, LDA, IPIV);
}

#  if 0
/* FIXME: doesn't work for ATLAS, work-around needed (if this is fixed, please
          adjust routine inverse in interpolation_method_hcsbb.c accordingly)*/
static int
LAPACKE_dsytrf_work( int matrix_order, char uplo, int N,
                     double* A, int LDA, int* IPIV,
                     double* WORK, int LWORK )
{
  assert(matrix_order == LAPACK_COL_MAJOR);
  return clapack_dsytrf((CBLAS_LAYOUT)matrix_order,
                        (CBLAS_UPLO)uplo,
                         N, A, LDA, IPIV);
}

/* FIXME: doesn't work for ATLAS, work-around needed (if this is fixed, please
          adjust routine inverse in interpolation_method_hcsbb.c accordingly)*/
static int
LAPACKE_dsytri_work(int matrix_order, char uplo, int N,
               double* A, int LDA, int* IPIV, double * WORK)
{
  assert(matrix_order == LAPACK_COL_MAJOR);
  return clapack_dsytri((CBLAS_LAYOUT)matrix_order,
                        (CBLAS_UPLO)uplo,
                        N, A, LDA, IPIV);
}
#  endif
#elif HAVE_FORTRAN_LAPACK
#  include <assert.h>
#  include "cfortran.h"
#  define LAPACK_COL_MAJOR 102
static int
LAPACKE_dgels_work(int matrix_order, char TRANS, int M, int N,
                   int NRHS, double *A, int LDA, double *B, int LDB,
                   double *WORK, int LWORK)
{ 
  assert(matrix_order == LAPACK_COL_MAJOR);
  PROTOCCALLSFSUB11(DGELS,dgels,STRING,INT,INT,INT,DOUBLEV,
                    INT,DOUBLEV,INT,DOUBLEV,INT,
                    PINT);
  
  int INFO;
  char TRANS_[2] = { TRANS, '\0' };
  
  CCALLSFSUB11(DGELS,dgels,STRING,INT,INT,INT,DOUBLEV,
               INT,DOUBLEV,INT,DOUBLEV,INT,PINT,
               TRANS_,M,N,NRHS,A,LDA,B,LDB,WORK,LWORK,INFO);

  return INFO;
}

static int
LAPACKE_dgesv(int matrix_order, int N, int NRHS,
               double *A, int LDA, int *IPIV,
               double *B, int LDB)
{
  assert(matrix_order == LAPACK_COL_MAJOR);
  PROTOCCALLSFSUB8(DGESV,dgesv,INT,INT,DOUBLEV,
                   INT,INTV,DOUBLEV,INT,PINT);

  int INFO;

  CCALLSFSUB8(DGESV,dgesv,INT,INT,DOUBLEV,INT,INTV,
              DOUBLEV,INT,PINT,
              N,NRHS,A,LDA,IPIV,B,LDB,INFO);

  return INFO;
}

static int
LAPACKE_dgetrf(int matrix_order, int M, int N,
               double *A, int LDA, int *IPIV)
{
  assert(matrix_order == LAPACK_COL_MAJOR);
  PROTOCCALLSFSUB6(DGETRF,dgetrf,INT,INT,DOUBLEV,INT,INTV,PINT);

  int INFO;

  CCALLSFSUB6(DGETRF,dgetrf,INT,INT,DOUBLEV,INT,INTV,PINT,
               M,N,A,LDA,IPIV,INFO);

  return INFO;
}

static int
LAPACKE_dgetri_work(int matrix_order, int N, double *A,
                    int LDA, const int *IPIV,
                    double *WORK, int LWORK)
{
  assert(matrix_order == LAPACK_COL_MAJOR);
  PROTOCCALLSFSUB7(DGETRI,dgetri,INT,DOUBLEV,INT,INTV,DOUBLEV,INT,PINT);

  int INFO;

  CCALLSFSUB7(DGETRI,dgetri,INT,DOUBLEV,INT,INTV,DOUBLEV,INT,PINT,
              N,A,LDA,(int *)IPIV,WORK,LWORK,INFO);

  return INFO;
}

static int
LAPACKE_dsytrf_work(int matrix_order, char UPLO, int N,
                    double* A, int LDA, int* IPIV,
                    double* WORK, int LWORK )
{
  assert(matrix_order == LAPACK_COL_MAJOR);

  int INFO;
  char UPLO_[2] = { UPLO, '\0' };

  CCALLSFSUB8(DSYTRF,dsytrf,STRING,INT,DOUBLEV,INT,INTV,DOUBLEV,INT,PINT,
              UPLO_,N,A,LDA,(int *)IPIV,WORK,LWORK,INFO);

  return INFO;
}

static int
LAPACKE_dsytri_work(int matrix_order, char UPLO, int N,
               double* A, int LDA, int* IPIV, double * WORK)
{
  assert(matrix_order == LAPACK_COL_MAJOR);

  int INFO;
  char UPLO_[2] = { UPLO, '\0' };

  CCALLSFSUB7(DSYTRI,dsytri,STRING,INT,DOUBLEV,INT,INTV,DOUBLEV,PINT,
              UPLO_,N,A,LDA,(int *)IPIV,WORK,INFO);

  return INFO;
}
#elif HAVE_CLAPACK
#  include <assert.h>
#  include <f2c.h>
#  include <clapack.h>
#  define LAPACK_COL_MAJOR 102
static int
LAPACKE_dgels_work(int matrix_order, char TRANS, int M, int N,
                   int NRHS, double *A, int LDA, double *B, int LDB,
                   double *WORK, int LWORK)
{
  assert(matrix_order == LAPACK_COL_MAJOR);

  integer M_ = M, N_ = N, NRHS_ = NRHS,
          LDA_ = LDA, LDB_ = LDB,
          LWORK_ = LWORK, INFO_;

  dgels_(&TRANS,&M_,&N_,&NRHS_,A,&LDA_,B,&LDB_,WORK,&LWORK_,&INFO_);

  return (int)INFO_;
}

static int
LAPACKE_dgesv(int matrix_order, int N, int NRHS,
               double *A, int LDA, int *IPIV,
               double *B, int LDB)
{
  assert(matrix_order == LAPACK_COL_MAJOR);

  int i;
  integer N_ = N, NRHS_ = NRHS,
          LDA_ = LDA, LDB_ = LDB,
          IPIV_[N], INFO_;

  dgesv_(&N_,&NRHS_,A,&LDA_,IPIV_,B,&LDB_,&INFO_);
  for (i = 0; i != N; ++i) { IPIV[i] = IPIV_[i]; }

  return (int)INFO_;
}

static int
LAPACKE_dgetrf(int matrix_order, int M, int N,
               double *A, int LDA, int *IPIV)
{
  assert(matrix_order == LAPACK_COL_MAJOR);

  int IPIV_size = min(N, M), i;

  integer M_ = M, N_ = N, LDA_ = LDA,
          IPIV_[IPIV_size], INFO_;

  dgetrf_(&M_,&N_,A,&LDA_,IPIV_,&INFO_);
  for (i = 0; i != IPIV_size; ++i) { IPIV[i] = IPIV_[i]; }

  return (int)INFO_;
}

static int
LAPACKE_dgetri_work(int matrix_order, int N, double *A,
                    int LDA, const int *IPIV,
                    double *WORK, int LWORK)
{
  assert(matrix_order == LAPACK_COL_MAJOR);

  int i;
  integer N_ = N, LDA_ = LDA,
          IPIV_[N], LWORK_ = LWORK,
          INFO_;

  for (i = 0; i != N; ++i) { IPIV_[i] = IPIV[i]; }
  dgetri_(&N_,A,&LDA_,IPIV_,WORK,&LWORK_,&INFO_);

  return (int)INFO_;
}

static int
LAPACKE_dsytrf_work(int matrix_order, char UPLO, int N,
                    double* A, int LDA, int* IPIV,
                    double* WORK, int LWORK )
{
  assert(matrix_order == LAPACK_COL_MAJOR);

  int i;
  integer N_ = N, LDA_ = LDA,
          IPIV_[N], LWORK_ = LWORK,
          INFO_;

  dsytrf_(&UPLO,&N_,A,&LDA_,IPIV_,WORK,&LWORK_,&INFO_);
  for (i = 0; i != N; ++i) { IPIV[i] = IPIV_[i]; }

  return (int)INFO_;
}

static int
LAPACKE_dsytri_work(int matrix_order, char UPLO, int N,
               double* A, int LDA, int* IPIV, double * WORK)
{
  assert(matrix_order == LAPACK_COL_MAJOR);

  int i;
  integer N_ = N, LDA_ = LDA,
          IPIV_[N], INFO_;

  for (i = 0; i != N; ++i) { IPIV_[i] = IPIV[i]; }
  dsytri_(&UPLO,&N_,A,&LDA_,IPIV_,WORK,&INFO_);

  return (int)INFO_;
}
#else

#  error None of the supported LAPACK interfaces is available

#endif

#endif // YAC_LAPACK_INTERFACE_H

