/**
 * @file test_couple_config.c
 *
 * @copyright Copyright  (C)  2015 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "tests.h"

#include "yacpar.h"
#include "xml/config_xml.h"
#include "communicator.h"
#include "communicator_mpi.h"
#include "grid_reg2d.h"
#include "domain.h"
#include "couple_config.h"
#include "utils.h"
#include "yac_interface.h"

int main (void) {

  yac_world_comm = yac_mpi_comm_new();

  if (yac_get_comm_size(yac_world_comm) != 3) {

    PUT_ERR("ERROR: wrong number of processes\n");
    return TEST_EXIT_CODE;
  }

  { // tests with couple_config_test_1.xml
    yac_xml_init_config_xml("couple_config_test_1.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // no process defines a components
      yac_setup_couple_config(0, NULL, 0, NULL);
      yac_enddef_couple_config();
      yac_free_couple_config();
    }

    { // first two process define component comp_1 and last defines comp_2

      int rank = yac_get_comm_rank(yac_world_comm);

      struct component * components[1];
      char * component_names[2] = {"comp_1", "comp_2"};
      unsigned component_id;
      struct communicator * component_comm;

      components[0] = yac_component_new(component_names[rank >> 1]);
      component_id = yac_pointer_to_unique_id(components[0]);
      component_comm = yac_component_get_communicator(components[0]);

      struct grid * grid;
      unsigned num_cells[2] = {1,1};
      unsigned cyclic[2] = {0,0};
      double coordinates_x[3][2] = {{0,1},{1,2},{0.5,1.5}};
      double coordinates_y[3][2] = {{0,1},{0,1},{0,1}};

      grid = yac_reg2d_grid_new(coordinates_x[rank], coordinates_y[rank],
                                num_cells, cyclic);

      struct subdomain subdomain;
      int global_cell_ids [3][1] = {{0},{1},{0}};
      unsigned cell_is_on_rank = -1;
      int global_corner_ids[3][4] = {{0,1,3,4},{1,2,4,5},{0,1,2,3}};
      unsigned corner_is_on_rank[3][4] = {{-1,-1,-1,1},{0,-1,-1,-1},{-1,-1,-1,-1}};

      yac_init_subdomain_unstruct(&subdomain, grid, component_comm,
                                  global_corner_ids[rank], corner_is_on_rank[rank],
                                  global_cell_ids[rank], &cell_is_on_rank,
                                  NULL, NULL);

      struct domain domain;
      struct subdomain * subdomains[1] = {&subdomain};
      char * domain_name[2] = {"grid1", "grid2"};

      yac_init_domain(&domain, component_comm, subdomains, 1,
                      domain_name[rank>>1]);
      yac_enddef_domain(&domain);

      struct points points;
      struct points * points_per_subdomain[1] = {&points};
      struct points ** point_sets[1] = {points_per_subdomain};

      yac_init_points(&points, grid, CORNER, coordinates_x[rank],
                  coordinates_y[rank]);

      struct coupling_field * coupling_fields[1];
      char * field_name = "transient_1";

      coupling_fields[0] = yac_coupling_field_new(field_name, component_id,
                                                  &domain, point_sets, NULL, 1);

      yac_setup_couple_config(1, components, 1, coupling_fields);
      yac_enddef_couple_config();

      yac_coupling_field_delete(coupling_fields[0]);
      yac_free_points(&points);
      yac_free_domain(&domain);
      yac_delete_grid(grid);
      yac_component_delete(components[0]);

      yac_free_couple_config();

    }

    yac_components_new_first_call_reset();
    yac_xml_destroy_config_xml();
  }

  { // tests with couple_config_test_2.xml
    yac_xml_init_config_xml("couple_config_test_2.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // each process has its own component and data exchange occurs in a
      // round robin fashion

      int rank = yac_get_comm_rank(yac_world_comm);

      struct component * components[1];
      char * component_names[3] = {"comp_1", "comp_2", "comp_3"};
      unsigned component_id;
      struct communicator * component_comm;

      components[0] = yac_component_new(component_names[rank]);
      component_id = yac_pointer_to_unique_id(components[0]);
      component_comm = yac_component_get_communicator(components[0]);

      struct grid * grid;
      unsigned num_cells[2] = {1,1};
      unsigned cyclic[2] = {0,0};
      double coordinates_x[2] = {0,1};
      double coordinates_y[2] = {0,1};

      grid = yac_reg2d_grid_new(coordinates_x, coordinates_y, num_cells, cyclic);

      struct subdomain subdomain;
      int global_cell_ids [1] = {0};
      unsigned cell_is_on_rank[1] = {-1};
      int global_corner_ids[4] = {0,1,2,3};
      unsigned corner_is_on_rank[4] = {-1,-1,-1,-1};

      yac_init_subdomain_unstruct(&subdomain, grid, component_comm,
                                  global_corner_ids, corner_is_on_rank,
                                  global_cell_ids, cell_is_on_rank,
                                  NULL, NULL);

      struct domain domain;
      struct subdomain * subdomains[1] = {&subdomain};
      char * domain_name[3] = {"grid1", "grid2", "grid3"};

      yac_init_domain(&domain, component_comm, subdomains, 1,
                      domain_name[rank]);
      yac_enddef_domain(&domain);

      struct points points;
      struct points * points_per_subdomain[1] = {&points};
      struct points ** point_sets[1] = {points_per_subdomain};

      yac_init_points(&points, grid, CORNER, coordinates_x, coordinates_y);

      struct coupling_field * coupling_fields[2];
      char * field_name[3] = {"transient_1", "transient_2", "transient_3"};

      coupling_fields[0] = yac_coupling_field_new(field_name[rank], component_id,
                                                  &domain, point_sets, NULL, 1);
      coupling_fields[1] = yac_coupling_field_new(field_name[(rank + 2) % 3], component_id,
                                                  &domain, point_sets, NULL, 1);

      yac_setup_couple_config(1, components, 2, coupling_fields);
      yac_enddef_couple_config();

      yac_coupling_field_delete(coupling_fields[0]);
      yac_coupling_field_delete(coupling_fields[1]);
      yac_free_points(&points);
      yac_free_domain(&domain);
      yac_delete_grid(grid);
      yac_component_delete(components[0]);

      yac_free_couple_config();
    }

    yac_components_new_first_call_reset();
    yac_xml_destroy_config_xml();
  }

  { // tests with couple_config_test_3.xml
    yac_xml_init_config_xml("couple_config_test_3.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // each process has its own component and data exchange occurs in a
      // round robin fashion

      int rank = yac_get_comm_rank(yac_world_comm);

      struct component * components[1];
      char * component_names[3] = {"comp_1", "comp_2", "comp_3"};
      unsigned component_id;
      struct communicator * component_comm;

      components[0] = yac_component_new(component_names[rank]);
      component_id = yac_pointer_to_unique_id(components[0]);
      component_comm = yac_component_get_communicator(components[0]);

      struct grid * grid;
      unsigned num_cells[2] = {1,1};
      unsigned cyclic[2] = {0,0};
      double coordinates_x[2] = {0,1};
      double coordinates_y[2] = {0,1};

      grid = yac_reg2d_grid_new(coordinates_x, coordinates_y, num_cells, cyclic);

      struct subdomain subdomain;
      int global_cell_ids [1] = {0};
      unsigned cell_is_on_rank[1] = {-1};
      int global_corner_ids[4] = {0,1,2,3};
      unsigned corner_is_on_rank[4] = {-1,-1,-1,-1};

      yac_init_subdomain_unstruct(&subdomain, grid, component_comm,
                                  global_corner_ids, corner_is_on_rank,
                                  global_cell_ids, cell_is_on_rank,
                                  NULL, NULL);

      struct domain domain;
      struct subdomain * subdomains[1] = {&subdomain};
      char * domain_name[3] = {"grid1", "grid2", "grid3"};

      yac_init_domain(&domain, component_comm, subdomains, 1,
                      domain_name[rank]);
      yac_enddef_domain(&domain);

      struct points points;
      struct points * points_per_subdomain[1] = {&points};
      struct points ** point_sets[1] = {points_per_subdomain};

      yac_init_points(&points, grid, CORNER, coordinates_x, coordinates_y);

      struct coupling_field * coupling_fields[4];
      char * field_name[6] = {"transient_1", "transient_2", "transient_3",
                              "transient_4", "transient_5", "transient_6"};

      coupling_fields[0] = yac_coupling_field_new(field_name[rank], component_id,
                                                  &domain, point_sets, NULL, 1);
      coupling_fields[1] = yac_coupling_field_new(field_name[(rank + 2) % 3],
                                                  component_id, &domain, point_sets,
                                                  NULL, 1);
      coupling_fields[2] = yac_coupling_field_new(field_name[rank + 3],
                                                  component_id, &domain, point_sets,
                                                  NULL, 1);
      coupling_fields[3] = yac_coupling_field_new(field_name[((rank + 2) % 3) + 3],
                                                  component_id, &domain, point_sets,
                                                  NULL, 1);

      yac_setup_couple_config(1, components, 4, coupling_fields);
      yac_enddef_couple_config();

      yac_coupling_field_delete(coupling_fields[0]);
      yac_coupling_field_delete(coupling_fields[1]);
      yac_coupling_field_delete(coupling_fields[2]);
      yac_coupling_field_delete(coupling_fields[3]);
      yac_free_points(&points);
      yac_free_domain(&domain);
      yac_delete_grid(grid);
      yac_component_delete(components[0]);

      yac_free_couple_config();
    }

    yac_components_new_first_call_reset();
    yac_xml_destroy_config_xml();
  }

  { // tests with couple_config_test_4.xml
    yac_xml_init_config_xml("couple_config_test_4.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // each process has its own component
      // process one has one field that is to be sent to both other
      // processes in the same put

      int rank = yac_get_comm_rank(yac_world_comm);

      struct component * components[1];
      char * component_names[3] = {"comp_1", "comp_2", "comp_3"};
      unsigned component_id;
      struct communicator * component_comm;

      components[0] = yac_component_new(component_names[rank]);
      component_id = yac_pointer_to_unique_id(components[0]);
      component_comm = yac_component_get_communicator(components[0]);

      struct grid * grid;
      unsigned num_cells[2] = {1,1};
      unsigned cyclic[2] = {0,0};
      double coordinates_x[2] = {0,1};
      double coordinates_y[2] = {0,1};

      grid = yac_reg2d_grid_new(coordinates_x, coordinates_y, num_cells, cyclic);

      struct subdomain subdomain;
      int global_cell_ids [1] = {0};
      unsigned cell_is_on_rank[1] = {-1};
      int global_corner_ids[4] = {0,1,2,3};
      unsigned corner_is_on_rank[4] = {-1,-1,-1,-1};

      yac_init_subdomain_unstruct(&subdomain, grid, component_comm,
                                  global_corner_ids, corner_is_on_rank,
                                  global_cell_ids, cell_is_on_rank,
                                  NULL, NULL);

      struct domain domain;
      struct subdomain * subdomains[1] = {&subdomain};
      char * domain_name[3] = {"grid1", "grid2", "grid3"};

      yac_init_domain(&domain, component_comm, subdomains, 1,
                      domain_name[rank]);
      yac_enddef_domain(&domain);

      struct points points;
      struct points * points_per_subdomain[1] = {&points};
      struct points ** point_sets[1] = {points_per_subdomain};

      yac_init_points(&points, grid, CORNER, coordinates_x, coordinates_y);

      struct coupling_field * coupling_fields[1];
      char * field_name = "transient_1";

      coupling_fields[0] = yac_coupling_field_new(field_name, component_id,
                                                  &domain, point_sets, NULL, 1);

      yac_setup_couple_config(1, components, 1, coupling_fields);
      yac_enddef_couple_config();

      yac_coupling_field_delete(coupling_fields[0]);
      yac_free_points(&points);
      yac_free_domain(&domain);
      yac_delete_grid(grid);
      yac_component_delete(components[0]);

      yac_free_couple_config();
    }

    yac_components_new_first_call_reset();
    yac_xml_destroy_config_xml();
  }

  { // tests with couple_config_test_5.xml
    yac_xml_init_config_xml("couple_config_test_5.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // each process has its own component
      // process one has one field that is to be sent to both other
      // processes in the same put

      int rank = yac_get_comm_rank(yac_world_comm);

      struct component * components[1];
      char * component_names[3] = {"comp_1", "comp_2", "comp_3"};
      unsigned component_id;
      struct communicator * component_comm;

      components[0] = yac_component_new(component_names[rank]);
      component_id = yac_pointer_to_unique_id(components[0]);
      component_comm = yac_component_get_communicator(components[0]);

      struct grid * grid;
      unsigned num_cells[2] = {1,1};
      unsigned cyclic[2] = {0,0};
      double coordinates_x[2] = {0,1};
      double coordinates_y[2] = {0,1};

      grid = yac_reg2d_grid_new(coordinates_x, coordinates_y, num_cells, cyclic);

      struct subdomain subdomain;
      int global_cell_ids [1] = {0};
      unsigned cell_is_on_rank[1] = {-1};
      int global_corner_ids[4] = {0,1,2,3};
      unsigned corner_is_on_rank[4] = {-1,-1,-1,-1};

      yac_init_subdomain_unstruct(&subdomain, grid, component_comm,
                                  global_corner_ids, corner_is_on_rank,
                                  global_cell_ids, cell_is_on_rank,
                                  NULL, NULL);

      struct domain domain;
      struct subdomain * subdomains[1] = {&subdomain};
      char * domain_name[3] = {"grid1", "grid2", "grid3"};

      yac_init_domain(&domain, component_comm, subdomains, 1,
                      domain_name[rank]);
      yac_enddef_domain(&domain);

      struct points points;
      struct points * points_per_subdomain[1] = {&points};
      struct points ** point_sets[1] = {points_per_subdomain};

      yac_init_points(&points, grid, CELL, coordinates_x, coordinates_y);

      struct coupling_field * coupling_fields[1];
      char * field_name = "transient_1";

      coupling_fields[0] = yac_coupling_field_new(field_name, component_id,
                                                  &domain, point_sets, NULL, 1);

      yac_setup_couple_config(1, components, 1, coupling_fields);
      yac_enddef_couple_config();

      yac_coupling_field_delete(coupling_fields[0]);
      yac_free_points(&points);
      yac_free_domain(&domain);
      yac_delete_grid(grid);
      yac_component_delete(components[0]);

      yac_free_couple_config();
    }

    yac_components_new_first_call_reset();
    yac_xml_destroy_config_xml();
  }

#ifdef HAVE_C_NETCDF
  { // tests with couple_config_test_6.xml
    yac_xml_init_config_xml("couple_config_test_6.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // first two process define component comp_1 and last defines comp_2

      int rank = yac_get_comm_rank(yac_world_comm);

      struct component * components[1];
      char * component_names[2] = {"comp_1", "comp_2"};
      unsigned component_id;
      struct communicator * component_comm;

      components[0] = yac_component_new(component_names[rank >> 1]);
      component_id = yac_pointer_to_unique_id(components[0]);
      component_comm = yac_component_get_communicator(components[0]);

      struct grid * grid;
      unsigned num_cells[2] = {1,1};
      unsigned cyclic[2] = {0,0};
      double coordinates_x[3][2] = {{0,1},{1,2},{0.5,1.5}};
      double coordinates_y[3][2] = {{0,1},{0,1},{0,1}};

      grid = yac_reg2d_grid_new(coordinates_x[rank], coordinates_y[rank],
                                num_cells, cyclic);

      struct subdomain subdomain;
      int global_cell_ids [3][1] = {{0},{1},{0}};
      unsigned cell_is_on_rank = -1;
      int global_corner_ids[3][4] = {{0,1,3,4},{1,2,4,5},{0,1,2,3}};
      unsigned corner_is_on_rank[3][4] = {{-1,-1,-1,1},{0,-1,-1,-1},{-1,-1,-1,-1}};

      yac_init_subdomain_unstruct(&subdomain, grid, component_comm,
                                  global_corner_ids[rank], corner_is_on_rank[rank],
                                  global_cell_ids[rank], &cell_is_on_rank,
                                  NULL, NULL);

      struct domain domain;
      struct subdomain * subdomains[1] = {&subdomain};
      char * domain_name[2] = {"grid1", "grid2"};

      yac_init_domain(&domain, component_comm, subdomains, 1,
                      domain_name[rank>>1]);
      yac_enddef_domain(&domain);

      struct points points;
      struct points * points_per_subdomain[1] = {&points};
      struct points ** point_sets[1] = {points_per_subdomain};

      yac_init_points(&points, grid, CORNER, coordinates_x[rank],
                  coordinates_y[rank]);

      struct coupling_field * coupling_fields[1];
      char * field_name = "transient_1";

      char * weight_file_name = "weight_file.nc";

      // delete weight file if it extists
      if (rank == 0) unlink(weight_file_name);

      coupling_fields[0] = yac_coupling_field_new(field_name, component_id,
                                                  &domain, point_sets, NULL, 1);

      yac_setup_couple_config(1, components, 1, coupling_fields);
      yac_enddef_couple_config();

      // ensure that the weight file has been written
      MPI_Barrier(MPI_COMM_WORLD);

      // check whether weight file exists
      if (rank == 0) {
        if (access(weight_file_name, F_OK ) == -1)
          PUT_ERR("weight file is missing\n");
        // delete weight file if it extists
        unlink(weight_file_name);
      }

      yac_coupling_field_delete(coupling_fields[0]);
      yac_free_points(&points);
      yac_free_domain(&domain);
      yac_delete_grid(grid);
      yac_component_delete(components[0]);

      yac_free_couple_config();
    }

    yac_components_new_first_call_reset();
    yac_xml_destroy_config_xml();
  }
#endif

#ifdef HAVE_C_NETCDF
  { // tests with couple_config_test_7.xml
    yac_xml_init_config_xml("couple_config_test_7.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // first two process define component comp_1 and last defines comp_2

      int rank = yac_get_comm_rank(yac_world_comm);

      struct component * components[1];
      char * component_names[2] = {"comp_1", "comp_2"};
      unsigned component_id;
      struct communicator * component_comm;

      components[0] = yac_component_new(component_names[rank >> 1]);
      component_id = yac_pointer_to_unique_id(components[0]);
      component_comm = yac_component_get_communicator(components[0]);

      struct grid * grid;
      unsigned num_cells[2] = {1,1};
      unsigned cyclic[2] = {0,0};
      double coordinates_x[3][2] = {{0,1},{1,2},{0.5,1.5}};
      double coordinates_y[3][2] = {{0,1},{0,1},{0,1}};

      grid = yac_reg2d_grid_new(coordinates_x[rank], coordinates_y[rank],
                                num_cells, cyclic);

      struct subdomain subdomain;
      int global_cell_ids [3][1] = {{0},{1},{0}};
      unsigned cell_is_on_rank = -1;
      int global_corner_ids[3][4] = {{0,1,3,4},{1,2,4,5},{0,1,2,3}};
      unsigned corner_is_on_rank[3][4] = {{-1,-1,-1,1},{0,-1,-1,-1},{-1,-1,-1,-1}};

      yac_init_subdomain_unstruct(&subdomain, grid, component_comm,
                                  global_corner_ids[rank], corner_is_on_rank[rank],
                                  global_cell_ids[rank], &cell_is_on_rank,
                                  NULL, NULL);

      struct domain domain;
      struct subdomain * subdomains[1] = {&subdomain};
      char * domain_name[2] = {"grid1", "grid2"};

      yac_init_domain(&domain, component_comm, subdomains, 1,
                      domain_name[rank>>1]);
      yac_enddef_domain(&domain);

      struct points points;
      struct points * points_per_subdomain[1] = {&points};
      struct points ** point_sets[1] = {points_per_subdomain};

      yac_init_points(&points, grid, CORNER, coordinates_x[rank],
                  coordinates_y[rank]);

      struct coupling_field * coupling_fields[3];
      char * field_name[3] = {"transient_1", "transient_2", "transient_3"};

      char * weight_file_name[3] =
        {NULL, "weight_file_1.nc", "weight_file_2.nc"};

      // delete weight file if it extists
      for (unsigned i = 0; i < 3; ++i) {

        if (rank == 0) unlink(weight_file_name[i]);

        coupling_fields[i] =
          yac_coupling_field_new(field_name[i], component_id, &domain,
                                 point_sets, NULL, 1);
      }

      yac_setup_couple_config(1, components, 3, coupling_fields);
      yac_enddef_couple_config();

      // ensure that the weight file has been written
      MPI_Barrier(MPI_COMM_WORLD);

      // check whether weight file exists
      if (rank == 0) {
        for (unsigned i = 0; i < 3; ++i) {
          if (weight_file_name[i] == NULL) continue;
          if (access(weight_file_name[i], F_OK ) == -1)
            PUT_ERR("weight file is missing\n");
          // delete weight file if it extists
          unlink(weight_file_name[i]);
        }
      }

      for (unsigned i = 0; i < 3; ++i)
        yac_coupling_field_delete(coupling_fields[i]);
      yac_free_points(&points);
      yac_free_domain(&domain);
      yac_delete_grid(grid);
      yac_component_delete(components[0]);

      yac_free_couple_config();
    }

    yac_components_new_first_call_reset();
    yac_xml_destroy_config_xml();
  }
#endif


  { // tests with couple_config_test_8.xml
    yac_xml_init_config_xml("couple_config_test_8.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // all processes define comp_1, but the xml lists comp_1 and comp_2

      int rank = yac_get_comm_rank(yac_world_comm);

      struct component * component = yac_component_new("comp_1");
      unsigned component_id = yac_pointer_to_unique_id(component);
      struct communicator * component_comm =
        yac_component_get_communicator(component);

      struct grid * grid;
      unsigned num_cells[2] = {1,1};
      unsigned cyclic[2] = {0,0};
      double coordinates_x[3][2] = {{0,1},{1,2},{2,3}};
      double coordinates_y[3][2] = {{0,1},{0,1},{0,1}};

      grid = yac_reg2d_grid_new(coordinates_x[rank], coordinates_y[rank],
                                num_cells, cyclic);

      struct subdomain subdomain;
      int global_cell_ids [3][1] = {{0},{1},{2}};
      unsigned cell_is_on_rank = -1;
      int global_corner_ids[3][4] = {{0,1,4,5},{1,2,5,6},{2,3,6,7}};
      unsigned corner_is_on_rank[3][4] = {{-1,-1,-1,1},{0,-1,-1,2},{1,-1,-1,-1}};

      yac_init_subdomain_unstruct(&subdomain, grid, component_comm,
                                  global_corner_ids[rank], corner_is_on_rank[rank],
                                  global_cell_ids[rank], &cell_is_on_rank,
                                  NULL, NULL);

      struct domain domain;
      struct subdomain * subdomains[1] = {&subdomain};

      yac_init_domain(&domain, component_comm, subdomains, 1, "grid1");
      yac_enddef_domain(&domain);

      struct points points;
      struct points * points_per_subdomain[1] = {&points};
      struct points ** point_sets[1] = {points_per_subdomain};

      yac_init_points(&points, grid, CORNER, coordinates_x[rank],
                  coordinates_y[rank]);

      struct coupling_field * coupling_fields[3];
      char * field_name[3] = {"transient_1", "transient_2", "transient_3"};

      // delete weight file if it extists
      for (unsigned i = 0; i < 3; ++i) {

        coupling_fields[i] =
          yac_coupling_field_new(field_name[i], component_id, &domain,
                                 point_sets, NULL, 1);
      }

      yac_setup_couple_config(1, &component, 3, coupling_fields);
      yac_enddef_couple_config();

      yac_coupling_field_delete(coupling_fields[0]);
      yac_coupling_field_delete(coupling_fields[1]);
      yac_coupling_field_delete(coupling_fields[2]);
      yac_free_points(&points);
      yac_free_domain(&domain);
      yac_delete_grid(grid);
      yac_component_delete(component);

      yac_free_couple_config();
    }

    yac_components_new_first_call_reset();
    yac_xml_destroy_config_xml();
  }

  yac_free_communicator(yac_world_comm);

  yac_comm_finalize();

  return TEST_EXIT_CODE;
}

