/**
 * @file test_dep_list.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <string.h>
#include "tests.h"
#include "dep_list.h"
#include "test_common.h"
#include "communicator_local.h"

static void compare_dep_list(struct dep_list * a, struct dep_list * b);

int main (void) {

   struct communicator * comm = yac_local_comm_new();

   //---------------
   // setup
   //---------------

   size_t num_elements = 4;
   size_t num_deps_per_element[4] = {1,2,2,4};
   size_t * p_num_deps_per_element;
   unsigned dependencies[9] = {0, 0,1, 0,2, 0,1,2,3};
   unsigned * p_dependencies;

   // test initialisation of dependency list
   p_num_deps_per_element = TO_POINTER(num_deps_per_element);
   p_dependencies = TO_POINTER(dependencies);
   struct dep_list * list =
      yac_dep_list_new(num_elements, p_num_deps_per_element, p_dependencies);

   //---------------
   // testing
   //---------------

   // test inversion of dependency list
   {
      struct dep_list * inv_list = yac_invert_dep_list(list);

      // reference data
      size_t ref_num_elements = 4;
      size_t ref_num_deps_per_element[4] = {4,2,2,1};
      unsigned ref_dependencies[9] = {0,1,2,3, 1,3, 2,3, 3};
      size_t * p_ref_num_deps_per_element = TO_POINTER(ref_num_deps_per_element);
      unsigned * p_ref_dependencies = TO_POINTER(ref_dependencies);
      struct dep_list * ref_dep_list =
         yac_dep_list_new(
            ref_num_elements, p_ref_num_deps_per_element, p_ref_dependencies);

      compare_dep_list(inv_list, ref_dep_list);

      yac_dep_list_delete(ref_dep_list);
      yac_dep_list_delete(inv_list);
   }

   // test dependency index <--> index + element index conversion
   {

      unsigned ref_idx_element_pair[9][2] = {{0,0}, {1,0},{1,1}, {2,0},{2,2},
                                             {3,0},{3,1},{3,2},{3,3}};

      for (size_t i = 0; i < 9; ++i) {

         if (yac_get_dependency_index(
                list, (size_t)ref_idx_element_pair[i][0],
                ref_idx_element_pair[i][1]) != i)
            PUT_ERR("error get_dependency_index\n")

         size_t index;
         unsigned element;
         yac_get_dependency (list, i, &index, &element);

         if (index != (size_t)ref_idx_element_pair[i][0] ||
             element != ref_idx_element_pair[i][1])
            PUT_ERR("error get_dependency\n")
      }
   }

   { // test remove_dependencies_of_elements
      struct dep_list * temp_list = yac_copy_dep_list(list);
      size_t element_to_remove = 2;

      yac_remove_dependencies_of_elements(temp_list, &element_to_remove, 1);

      size_t ref_num_elements = 4;
      size_t ref_num_deps_per_element[4] = {1,2,0,4};
      unsigned ref_dependencies[7] = {0, 0,1, 0,1,2,3};
      size_t * p_ref_num_deps_per_element = TO_POINTER(ref_num_deps_per_element);
      unsigned * p_ref_dependencies = TO_POINTER(ref_dependencies);
      struct dep_list * ref_dep_list =
         yac_dep_list_new(
            ref_num_elements, p_ref_num_deps_per_element, p_ref_dependencies);

      compare_dep_list(temp_list, ref_dep_list);

      yac_dep_list_delete(ref_dep_list);
      yac_dep_list_delete(temp_list);
   }

   { // test remove_dependencies
      struct dep_list * temp_list = yac_copy_dep_list(list);
      unsigned index_to_remove = 0;

      yac_remove_dependencies(temp_list, &index_to_remove, 1);

      size_t ref_num_elements = 4;
      size_t ref_num_deps_per_element[4] = {0,1,1,3};
      unsigned ref_dependencies[5] = {1, 2, 1,2,3};
      size_t * p_ref_num_deps_per_element = TO_POINTER(ref_num_deps_per_element);
      unsigned * p_ref_dependencies = TO_POINTER(ref_dependencies);
      struct dep_list * ref_dep_list =
         yac_dep_list_new(
            ref_num_elements, p_ref_num_deps_per_element, p_ref_dependencies);

      compare_dep_list(temp_list, ref_dep_list);

      yac_dep_list_delete(ref_dep_list);
      yac_dep_list_delete(temp_list);
   }

   { // test add_dependencies
      struct dep_list * temp_list = yac_copy_dep_list(list);

      unsigned new_dependencies[2] = {2,3};

      yac_add_dependencies(temp_list, 1, 2, new_dependencies);

      size_t ref_num_elements = 4;
      size_t ref_num_deps_per_element[4] = {1,4,2,4};
      unsigned ref_dependencies[11] = {0, 0,1,2,3, 0,2, 0,1,2,3};
      size_t * p_ref_num_deps_per_element = TO_POINTER(ref_num_deps_per_element);
      unsigned * p_ref_dependencies = TO_POINTER(ref_dependencies);
      struct dep_list * ref_dep_list =
         yac_dep_list_new(
            ref_num_elements, p_ref_num_deps_per_element, p_ref_dependencies);

      compare_dep_list(temp_list, ref_dep_list);

      yac_dep_list_delete(ref_dep_list);
      yac_dep_list_delete(temp_list);
   }

   { // test add_dependencies
      struct dep_list * temp_list = yac_copy_dep_list(list);

      unsigned new_dependencies[2][2] = {{1,2},{4,5}};

      yac_add_dependencies(temp_list, 0, 2, new_dependencies[0]);
      yac_add_dependencies(temp_list, 3, 2, new_dependencies[1]);

      size_t ref_num_elements = 4;
      size_t ref_num_deps_per_element[4] = {3,2,2,6};
      unsigned ref_dependencies[13] = {0,1,2, 0,1, 0,2, 0,1,2,3,4,5};
      size_t * p_ref_num_deps_per_element = TO_POINTER(ref_num_deps_per_element);
      unsigned * p_ref_dependencies = TO_POINTER(ref_dependencies);
      struct dep_list * ref_dep_list =
         yac_dep_list_new(
            ref_num_elements, p_ref_num_deps_per_element, p_ref_dependencies);

      compare_dep_list(temp_list, ref_dep_list);

      yac_dep_list_delete(ref_dep_list);
      yac_dep_list_delete(temp_list);
   }

   { // test packing and unpacking of dep_list
      struct dep_list * unpacked_list[2];

      int pack_buffer_size = 0;
      for (size_t i = 0; i < 2; ++i)
        pack_buffer_size += yac_dep_list_pack_size(list, comm);

      void * pack_buffer = malloc(pack_buffer_size);

      int position = 0;
      for (size_t i = 0; i < 2; ++i)
         yac_pack_dep_list(
            list, pack_buffer, pack_buffer_size, &position, comm);

      position = 0;

      for (size_t i = 0; i < 2; ++i) {

         unpacked_list[i] =
            yac_unpack_dep_list(pack_buffer, pack_buffer_size, &position, comm);
         compare_dep_list(unpacked_list[i], list);
      }

      //---------------
      // cleanup
      //---------------

      yac_dep_list_delete(unpacked_list[0]);
      yac_dep_list_delete(unpacked_list[1]);
      free(pack_buffer);
   }

   { // test packing and unpacking of an empty dep_list

      struct dep_list * empty_list = yac_dep_list_new(0, NULL, NULL);

      struct dep_list * unpacked_list[2];

      int pack_buffer_size = 0;
      for (size_t i = 0; i < 2; ++i)
        pack_buffer_size += yac_dep_list_pack_size(empty_list, comm);

      void * pack_buffer = malloc(pack_buffer_size);

      int position = 0;
      for (size_t i = 0; i < 2; ++i)
         yac_pack_dep_list(
            empty_list, pack_buffer, pack_buffer_size, &position, comm);

      position = 0;

      for (size_t i = 0; i < 2; ++i) {

         unpacked_list[i] =
            yac_unpack_dep_list(pack_buffer, pack_buffer_size, &position, comm);

         if (unpacked_list[i] == NULL) PUT_ERR("wrong unpacking\n")

         if (yac_get_num_elements(unpacked_list[i]) != 0)
            PUT_ERR("wrong num_elements after unpacking\n")
      }

      //---------------
      // cleanup
      //---------------

      yac_dep_list_delete(unpacked_list[0]);
      yac_dep_list_delete(unpacked_list[1]);
      free(pack_buffer);
   }

   yac_dep_list_delete(list);

   yac_free_communicator(comm);
   yac_comm_finalize();

   return TEST_EXIT_CODE;
}

static void compare_dep_list(struct dep_list * a, struct dep_list * b) {

  if ((a == NULL) ^ (b == NULL)) {
    PUT_ERR("wrong (a == NULL) ^ (b == NULL)\n");
    return;
  }

  if (a == NULL) return;

  size_t num_elements_a = yac_get_num_elements(a);
  size_t num_elements_b = yac_get_num_elements(b);

  if (num_elements_a != num_elements_b) {
    PUT_ERR("wrong num_elements\n");
    return;
  }

  for (size_t i = 0; i < num_elements_a; ++i) {

    size_t total_num_dependencies_a = yac_get_total_num_dependencies(a);
    size_t total_num_dependencies_b = yac_get_total_num_dependencies(b);

    if (total_num_dependencies_a != total_num_dependencies_b) {
      PUT_ERR("wrong total_num_dependencies\n");
      return;
    }

    size_t curr_num_deps_a = yac_get_num_dependencies_of_element(a, i);
    size_t curr_num_deps_b = yac_get_num_dependencies_of_element(b, i);
    unsigned const * curr_deps_a = yac_get_dependencies_of_element(a, i);
    unsigned const * curr_deps_b = yac_get_dependencies_of_element(b, i);

    if (curr_num_deps_a != curr_num_deps_b) {
       PUT_ERR("wrong num_deps_per_element\n");
       return;
    }

    for (size_t j = 0; j < curr_num_deps_a; ++j) {

      if (curr_deps_a[j] != curr_deps_b[j]) {
        PUT_ERR("wrong dependencies\n");
        return;
      }
    }
  }
}

