/**
 * @file test_field_interpolation.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <unistd.h>

#include "tests.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_local.h"
#include "domain.h"
#include "global_search.h"
#include "fields.h"
#include "interpolation_method_avg.h"
#include "interpolation_method_fixed.h"
#include "interpolation_method_file.h"
#include "interpolation_method_nnn.h"
#include "field_interpolation_gen_src.h"
#include "field_interpolation_gen_tgt.h"
#include "utils.h"
#include "weight_file_common.h"
#include "test_common.h"

int main(void) {

   int component_id = yac_pointer_to_unique_id(NULL);

   // simple fixed value "interpolation" on a single process
   for (unsigned mapping_on_source = 0; mapping_on_source < 2;
        ++mapping_on_source) {

      //--------------------
      // set up global search
      //--------------------

         // one directional global search with two domains on two processes with one subdomain each

         // comm worlds on both processes
         struct communicator * comm_p1, * comm_p2;

         comm_p1 = yac_local_comm_new();
         comm_p2 = yac_local_comm_new();

         // named comms for the domains on both processes
         struct communicator * comm_named_p1, * comm_named_p2;

         comm_named_p1 = yac_named_local_comm_new(comm_p1, "P1");
         comm_named_p2 = yac_named_local_comm_new(comm_p2, "P2");

         // define a domain on P1

         struct grid * grid_p1;
         unsigned num_cells[2] = {2,2};
         unsigned cyclic[2] = {0,0};
         double coordinates_x[2][3] = {{-1,0,1}, {-1.5,-0.5,0.5}};
         double coordinates_y[2][3] = {{-1,0,1}, {-1.5,-0.5,0.5}};

         for ( unsigned i = 0; i < 2; ++i ) {
            for ( unsigned j = 0; j < 3; ++j ) {
               coordinates_x[i][j] *= YAC_RAD;
               coordinates_y[i][j] *= YAC_RAD;
            }
         }

         grid_p1 = yac_reg2d_grid_new(coordinates_x[0], coordinates_y[0],
                                      num_cells, cyclic);

         struct subdomain subdomain_p1;

         int global_corner_ids_p1[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p1[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p1[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p1[4] = {-1,-1,-1,-1};
         int global_edge_ids_p1[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p1[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p1, grid_p1, comm_named_p1,
                                     global_corner_ids_p1, corner_is_on_rank_p1,
                                     global_cell_ids_p1, cell_is_on_rank_p1,
                                     global_edge_ids_p1, edge_is_on_rank_p1);

         struct domain l_domain_p1;

         struct subdomain * blocks_p1[1] = {&subdomain_p1};

         yac_init_domain(&l_domain_p1, comm_named_p1, blocks_p1, 1, "domain_p1");

         // define a domain on P2, which matches the one on P1

         struct grid * grid_p2;

         grid_p2 = yac_reg2d_grid_new(coordinates_x[1], coordinates_y[1],
                                      num_cells, cyclic);

         struct subdomain subdomain_p2;

         int global_corner_ids_p2[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p2[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p2[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p2[4] = {-1,-1,-1,-1};
         int global_edge_ids_p2[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p2[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p2, grid_p2, comm_named_p2,
                                     global_corner_ids_p2, corner_is_on_rank_p2,
                                     global_cell_ids_p2, cell_is_on_rank_p2,
                                     global_edge_ids_p2, edge_is_on_rank_p2);

         struct domain l_domain_p2;

         struct subdomain * blocks_p2[1] = {&subdomain_p2};

         yac_init_domain(&l_domain_p2, comm_named_p2, blocks_p2, 1, "domain_p2");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1);
         yac_enddef_domain(&l_domain_p2);

         // init global search on both processes

         struct global_search * search[2];

         search[0] = yac_global_search_new(&l_domain_p1, "domain_p2", comm_p1,
                                           1,0, "component_p1", "component_p2");
         search[1] = yac_global_search_new(&l_domain_p2, "domain_p1", comm_p2,
                                           0,1, "component_p2", "component_p1");

         // finish initialisation of global search

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);

         // define points

         struct points points[2];

         yac_init_points(points+0,grid_p1, CORNER,
                         coordinates_x[0], coordinates_y[0]);
         yac_init_points(points+1,grid_p2, CORNER,
                         coordinates_x[1], coordinates_y[1]);

         // set up coupling field

         struct coupling_field * field[2];
         struct points * points_per_subdomain[2][1] = {{points+0}, {points+1}};
         struct points ** points_set[2] = {points_per_subdomain[0],
                                           points_per_subdomain[1]};

         field[0] = yac_coupling_field_new("source field", component_id,
                                           &l_domain_p1, points_set + 0,
                                           NULL, 1);
         field[1] = yac_coupling_field_new("target field", component_id,
                                           &l_domain_p2, points_set + 1,
                                           NULL, 1);

         // set up fixed interpolation method

         struct interpolation_method * method_fixed;
         double fixed_value = 1337;

         method_fixed = yac_new_interpolation_method_fixed(fixed_value, search[0]);

      //---------------------------
      // set up field interpolation
      //---------------------------

      struct field_interpolation_gen * interpolation_gen[2];
      struct interpolation_method * interpolation_method_stack[] = {method_fixed,
                                                                    NULL};

      interpolation_gen[0] =
        yac_new_field_interpolation_gen(points_set + 0, 1, "source field", search[0],
                                        "target field", 1, 0,
                                        interpolation_method_stack, 0, NULL, 1,
                                        mapping_on_source);
      interpolation_gen[1] =
        yac_new_field_interpolation_gen(points_set + 1, 1, "target field", search[1],
                                        "source field", 0, 1, NULL, 0, NULL, 1,
                                        mapping_on_source);
      yac_enddef_field_interpolation_gen(interpolation_gen[0]);
      yac_enddef_field_interpolation_gen(interpolation_gen[1]);

      struct field_interpolation * interpolation[2];

      interpolation[0] = yac_get_field_interpolation(interpolation_gen[0]);
      interpolation[1] = yac_get_field_interpolation(interpolation_gen[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double source_field[9] = {1,1,1, 1,1,1, 1,1,1};
      double * source_field_[1] = {source_field};
      double ** source_field__[1] = {source_field_};
      double target_field[9] = {-1,-1,-1, -1,-1,-1, -1,-1,-1};
      double * target_field_[1] = {target_field};

      double *** source_data[1] = {source_field__};
      double ** target_data[1] = {target_field_};

      yac_put_field(interpolation[0], source_data, 1);
      yac_get_field(interpolation[1], target_data, 1);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      for (unsigned i = 0; i < 9; ++i)
         if (double_are_unequal(target_field[i], fixed_value))
            PUT_ERR("error in results\n");

      //---------------
      // cleanup
      //---------------

      yac_delete_field_interpolation_gen(interpolation_gen[0]);
      yac_delete_field_interpolation_gen(interpolation_gen[1]);

      yac_delete_field_interpolation(interpolation[0]);
      yac_delete_field_interpolation(interpolation[1]);

      yac_delete_interpolation_method(method_fixed);

      yac_coupling_field_delete(field[0]);
      yac_coupling_field_delete(field[1]);

      yac_free_points(points+0);
      yac_free_points(points+1);

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);

      yac_free_domain(&l_domain_p1);
      yac_free_domain(&l_domain_p2);

      yac_delete_grid(grid_p1);
      yac_delete_grid(grid_p2);

      yac_free_communicator(comm_p1);
      yac_free_communicator(comm_p2);

      yac_free_communicator(comm_named_p1);
      yac_free_communicator(comm_named_p2);

      yac_reset_interpolation_method();
   }

   // NNN interpolation without any result using a single process
   for (unsigned mapping_on_source = 0; mapping_on_source < 2;
        ++mapping_on_source) {

      //--------------------
      // set up global search
      //--------------------

         // one directional global search with two domains on two processes with one subdomain each

         // comm worlds on both processes
         struct communicator * comm_p1, * comm_p2;

         comm_p1 = yac_local_comm_new();
         comm_p2 = yac_local_comm_new();

         // named comms for the domains on both processes
         struct communicator * comm_named_p1, * comm_named_p2;

         comm_named_p1 = yac_named_local_comm_new(comm_p1, "P1");
         comm_named_p2 = yac_named_local_comm_new(comm_p2, "P2");

         // define a domain on P1

         struct grid * grid_p1;
         unsigned num_cells[2] = {2,2};
         unsigned cyclic[2] = {0,0};
         double coordinates_x[2][3] = {{-1,0,1}, {-1.5,-0.5,0.5}};
         double coordinates_y[2][3] = {{-1,0,1}, {-1.5,-0.5,0.5}};

         for ( unsigned i = 0; i < 2; ++i ) {
            for ( unsigned j = 0; j < 3; ++j ) {
               coordinates_x[i][j] *= YAC_RAD;
               coordinates_y[i][j] *= YAC_RAD;
            }
         }

         grid_p1 = yac_reg2d_grid_new(coordinates_x[0], coordinates_y[0],
                                      num_cells, cyclic);

         struct subdomain subdomain_p1;

         int global_corner_ids_p1[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p1[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p1[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p1[4] = {-1,-1,-1,-1};
         int global_edge_ids_p1[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p1[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p1, grid_p1, comm_named_p1,
                                     global_corner_ids_p1, corner_is_on_rank_p1,
                                     global_cell_ids_p1, cell_is_on_rank_p1,
                                     global_edge_ids_p1, edge_is_on_rank_p1);

         struct domain l_domain_p1;

         struct subdomain * blocks_p1[1] = {&subdomain_p1};

         yac_init_domain(&l_domain_p1, comm_named_p1, blocks_p1, 1, "domain_p1");

         // define a domain on P2, which matches the one on P1

         struct grid * grid_p2;

         grid_p2 = yac_reg2d_grid_new(coordinates_x[1], coordinates_y[1],
                                      num_cells, cyclic);

         struct subdomain subdomain_p2;

         int global_corner_ids_p2[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p2[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p2[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p2[4] = {-1,-1,-1,-1};
         int global_edge_ids_p2[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p2[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p2, grid_p2, comm_named_p2,
                                     global_corner_ids_p2, corner_is_on_rank_p2,
                                     global_cell_ids_p2, cell_is_on_rank_p2,
                                     global_edge_ids_p2, edge_is_on_rank_p2);

         struct domain l_domain_p2;

         struct subdomain * blocks_p2[1] = {&subdomain_p2};

         yac_init_domain(&l_domain_p2, comm_named_p2, blocks_p2, 1, "domain_p2");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1);
         yac_enddef_domain(&l_domain_p2);

         // init global search on both processes

         struct global_search * search[2];

         search[0] = yac_global_search_new(&l_domain_p1, "domain_p2", comm_p1,
                                           1,0, "component_p1", "component_p2");
         search[1] = yac_global_search_new(&l_domain_p2, "domain_p1", comm_p2,
                                           0,1, "component_p2", "component_p1");

         // finish initialisation of global search

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);

         // define points

         struct points points[2];

         yac_init_points(points+0,grid_p1, CORNER,
                         coordinates_x[0], coordinates_y[0]);
         yac_init_points(points+1,grid_p2, CORNER,
                         coordinates_x[1], coordinates_y[1]);

         // set up coupling field

         struct coupling_field * field[2];
         struct points * points_per_subdomain[2][1] = {{points+0}, {points+1}};
         struct points ** points_set[2] = {points_per_subdomain[0],
                                           points_per_subdomain[1]};

         field[0] = yac_coupling_field_new("source field", component_id, &l_domain_p1, points_set + 0, NULL, 1);
         field[1] = yac_coupling_field_new("target field", component_id, &l_domain_p2, points_set + 1, NULL, 1);

         // set up fixed interpolation method

         struct interpolation_method * method_nnn;
         unsigned num_pointsets = 1;
         enum yac_nnn_reduction_type reduction_type = NNN_DISTANCE_WEIGHTED;
         unsigned n = 13; // 13 the source grid only has 12 corners, so the
                          // result will be empty
         unsigned points_idx = 0;

         method_nnn =
          yac_new_interpolation_method_nnn(
            search[0], &l_domain_p1, "nnn stack", points_set + 0, num_pointsets,
            reduction_type, n, YAC_DEFAULT_RBF_SCALE, YAC_DEFAULT_GAUSS_SCALE, points_idx);

      //---------------------------
      // set up field interpolation
      //---------------------------

      char const filename_out[] = "weight_file_out.nc";

      struct field_interpolation_gen * interpolation_gen[2];
      struct interpolation_method * interpolation_method_stack[] = {method_nnn, NULL};

      interpolation_gen[0] =
        yac_new_field_interpolation_gen(points_set + 0, 1, "source field", search[0],
                                        "target field", 1, 0,
                                        interpolation_method_stack, 0,
                                        filename_out, 1, mapping_on_source);
      interpolation_gen[1] =
        yac_new_field_interpolation_gen(points_set + 1, 1, "target field", search[1],
                                        "source field", 0, 1, NULL, 0, NULL, 1,
                                        mapping_on_source);
      yac_enddef_field_interpolation_gen(interpolation_gen[0]);
      yac_enddef_field_interpolation_gen(interpolation_gen[1]);

      struct field_interpolation * interpolation[2];

      interpolation[0] = yac_get_field_interpolation(interpolation_gen[0]);
      interpolation[1] = yac_get_field_interpolation(interpolation_gen[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double source_field[9] = {1,1,1, 1,1,1, 1,1,1};
      double * source_field_[1] = {source_field};
      double ** source_field__[1] = {source_field_};
      double target_field[9] = {-1,-1,-1, -1,-1,-1, -1,-1,-1};
      double * target_field_[1] = {target_field};

      double *** source_data[1] = {source_field__};
      double ** target_data[1] = {target_field_};

      yac_put_field(interpolation[0], source_data, 1);
      yac_get_field(interpolation[1], target_data, 1);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      for (unsigned i = 0; i < 9; ++i)
         if (double_are_unequal(target_field[i], -1.0))
            PUT_ERR("error in results\n");

      //---------------
      // cleanup
      //---------------

      yac_delete_field_interpolation_gen(interpolation_gen[0]);
      yac_delete_field_interpolation_gen(interpolation_gen[1]);

      yac_delete_field_interpolation(interpolation[0]);
      yac_delete_field_interpolation(interpolation[1]);

      yac_coupling_field_delete(field[0]);
      yac_coupling_field_delete(field[1]);

      yac_delete_interpolation_method(method_nnn);

      yac_free_points(points+0);
      yac_free_points(points+1);

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);

      yac_free_domain(&l_domain_p1);
      yac_free_domain(&l_domain_p2);

      yac_delete_grid(grid_p1);
      yac_delete_grid(grid_p2);

      yac_free_communicator(comm_p1);
      yac_free_communicator(comm_p2);

      yac_free_communicator(comm_named_p1);
      yac_free_communicator(comm_named_p2);

      unlink(filename_out);

      yac_reset_interpolation_method();
   }

   // interpolation using a single source process that has two subdomains which
   // share some points
   for (unsigned mapping_on_source = 0; mapping_on_source < 2;
        ++mapping_on_source) {

      //--------------------
      // set up global search
      //--------------------

         // one directional global search with two domains on two processes with one subdomain each

         // comm worlds on both processes
         struct communicator * comm_p1, * comm_p2;

         comm_p1 = yac_local_comm_new();
         comm_p2 = yac_local_comm_new();

         // named comms for the domains on both processes
         struct communicator * comm_named_p1, * comm_named_p2;

         comm_named_p1 = yac_named_local_comm_new(comm_p1, "P1");
         comm_named_p2 = yac_named_local_comm_new(comm_p2, "P2");

         // define a domain on P1

         struct grid * grid_p1[2];
         unsigned num_cells_p1[2] = {2,2};
         unsigned cyclic[2] = {0,0};
         double coordinates_x_p1[2][3] = {{0,1,2}, {0,1,2}};
         double coordinates_y_p1[2][3] = {{0,1,2}, {2,3,4}};

         for ( unsigned i = 0; i < 2; ++i ) {
            for ( unsigned j = 0; j < 3; ++j ) {
               coordinates_x_p1[i][j] *= YAC_RAD;
               coordinates_y_p1[i][j] *= YAC_RAD;
            }
         }

         for (unsigned i = 0; i < 2; ++i)
            grid_p1[i] =
               yac_reg2d_grid_new(
                  coordinates_x_p1[i], coordinates_y_p1[i], num_cells_p1, cyclic);

         struct subdomain subdomain_p1[2];

         int global_corner_ids_p1[2][9] =
            {{0,1,2,3,4,5,6,7,8}, {6,7,8,9,10,11,12,13,14}};
         unsigned corner_is_on_rank_p1[2][9] =
            {{-1,-1,-1,-1,-1,-1,-1,-1,0},{0,-1,-1,-1,-1,-1,-1,-1,-1}};
         int global_cell_ids_p1[2][4] =
            {{0,1,2,3}, {4,5,6,7}};
         unsigned cell_is_on_rank_p1[2][4] =
            {{-1,-1,-1,-1}, {-1,-1,-1,-1}};

         for (unsigned i = 0; i < 2; ++i)
            yac_init_subdomain_unstruct(
               &(subdomain_p1[i]), grid_p1[i], comm_named_p1,
               global_corner_ids_p1[i], corner_is_on_rank_p1[i],
               global_cell_ids_p1[i], cell_is_on_rank_p1[i], NULL, NULL);

         struct domain l_domain_p1;

         struct subdomain * blocks_p1[2] =
            {&(subdomain_p1[0]), &(subdomain_p1[1])};

         yac_init_domain(&l_domain_p1, comm_named_p1, blocks_p1, 2, "domain_p1");

         // define a domain on P2, which matches the one on P1

         struct grid * grid_p2;
         unsigned num_cells_p2[2] = {2,4};
         double coordinates_x_p2[3] = {0,1,2};
         double coordinates_y_p2[5] = {0,1,2,3,4};

         for (unsigned i = 0; i < 3; ++i) coordinates_x_p2[i] *= YAC_RAD;
         for (unsigned i = 0; i < 5; ++i) coordinates_y_p2[i] *= YAC_RAD;

         grid_p2 = yac_reg2d_grid_new(coordinates_x_p2, coordinates_y_p2,
                                      num_cells_p2, cyclic);

         struct subdomain subdomain_p2;

         int global_corner_ids_p2[15] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14};
         unsigned corner_is_on_rank_p2[15] =
            {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p2[8] = {0,1,2,3,4,5,6,7};
         unsigned cell_is_on_rank_p2[8] = {-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p2, grid_p2, comm_named_p2,
                                     global_corner_ids_p2, corner_is_on_rank_p2,
                                     global_cell_ids_p2, cell_is_on_rank_p2,
                                     NULL, NULL);

         struct domain l_domain_p2;

         struct subdomain * blocks_p2[1] = {&subdomain_p2};

         yac_init_domain(&l_domain_p2, comm_named_p2, blocks_p2, 1, "domain_p2");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1);
         yac_enddef_domain(&l_domain_p2);

         // init global search on both processes

         struct global_search * search[2];

         search[0] = yac_global_search_new(&l_domain_p1, "domain_p2", comm_p1,
                                           0,1, "component_p1", "component_p2");
         search[1] = yac_global_search_new(&l_domain_p2, "domain_p1", comm_p2,
                                           1,0, "component_p2", "component_p1");

         // finish initialisation of global search

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);

         // define points

         struct points points[3];

         for (unsigned i = 0; i < 2; ++i)
            yac_init_points(&(points[i]), grid_p1[i], CORNER,
                            coordinates_x_p1[i], coordinates_y_p1[i]);
         yac_init_points(&(points[2]), grid_p2, CORNER,
                         coordinates_x_p2, coordinates_y_p2);

         // set up coupling field

         struct coupling_field * field[2];
         struct points * points_per_subdomain[3][1] = {{points+0}, {points+1}, {points+2}};
         struct points ** points_set[3] = {points_per_subdomain[0],
                                           points_per_subdomain[1],
                                           points_per_subdomain[2]};

         field[0] = yac_coupling_field_new("target field", component_id, &l_domain_p1, points_set + 0, NULL, 1);
         field[1] = yac_coupling_field_new("source field", component_id, &l_domain_p2, points_set + 2, NULL, 1);

         // set up nnn interpolation method

         struct interpolation_method * method_nnn;
         unsigned num_pointsets = 1;
         enum yac_nnn_reduction_type reduction_type = NNN_DISTANCE_WEIGHTED;
         unsigned n = 1;
         unsigned points_idx = 0;

         method_nnn =
          yac_new_interpolation_method_nnn(
            search[1], &l_domain_p2, "nnn stack", points_set + 2,
            num_pointsets, reduction_type, n, YAC_DEFAULT_RBF_SCALE, YAC_DEFAULT_GAUSS_SCALE, points_idx);

      //---------------------------‚
      // set up field interpolation
      //---------------------------

      struct field_interpolation_gen * interpolation_gen[2];
      struct interpolation_method * interpolation_method_stack[] = {method_nnn, NULL};

      interpolation_gen[0] =
        yac_new_field_interpolation_gen(points_set + 0, 1, "target field", search[0],
                                        "source field", 0, 1,
                                        NULL, 0, NULL, 1, mapping_on_source);
      interpolation_gen[1] =
        yac_new_field_interpolation_gen(points_set + 2, 1, "source field", search[1],
                                        "target field", 1, 0,
                                        interpolation_method_stack, 0, NULL, 1,
                                        mapping_on_source);
      yac_enddef_field_interpolation_gen(interpolation_gen[0]);
      yac_enddef_field_interpolation_gen(interpolation_gen[1]);

      struct field_interpolation * interpolation[2];

      interpolation[0] = yac_get_field_interpolation(interpolation_gen[0]);
      interpolation[1] = yac_get_field_interpolation(interpolation_gen[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double source_field[15] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14};
      double * source_field_[1] = {source_field};
      double ** source_field__[1] = {source_field_};
      double target_field[2][9] =
         {{-1,-1,-1, -1,-1,-1, -1,-1,-1}, {-1,-1,-1, -1,-1,-1, -1,-1,-1}};
      double * target_field_[2] =
         {&(target_field[0][0]), &(target_field[1][0])};

      double *** source_data[1] = {source_field__};
      double ** target_data[1] = {target_field_};

      yac_put_field(interpolation[1], source_data, 1);
      yac_get_field(interpolation[0], target_data, 1);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      for (unsigned i = 0; i < 2; ++i) {
         for (unsigned j = 0; j < 9; ++j) {
            int target_field_is_not_global_corner_id =
              fabs(target_field[i][j] - global_corner_ids_p1[i][j]) > 0.0;
            int target_field_is_set = fabs(target_field[i][j] + 1.0) > 0.0;
            if (((corner_is_on_rank_p1[i][j] == -1) &&
                 target_field_is_not_global_corner_id) ||
                ((corner_is_on_rank_p1[i][j] != -1) && target_field_is_set))
               PUT_ERR("error in results\n");
         }
      }

      //---------------
      // cleanup
      //---------------

      yac_delete_field_interpolation_gen(interpolation_gen[0]);
      yac_delete_field_interpolation_gen(interpolation_gen[1]);

      yac_delete_field_interpolation(interpolation[0]);
      yac_delete_field_interpolation(interpolation[1]);

      yac_coupling_field_delete(field[0]);
      yac_coupling_field_delete(field[1]);

      yac_delete_interpolation_method(method_nnn);

      yac_free_points(points+0);
      yac_free_points(points+1);
      yac_free_points(points+2);

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);

      yac_free_domain(&l_domain_p1);
      yac_free_domain(&l_domain_p2);

      yac_delete_grid(grid_p1[0]);
      yac_delete_grid(grid_p1[1]);
      yac_delete_grid(grid_p2);

      yac_free_communicator(comm_p1);
      yac_free_communicator(comm_p2);

      yac_free_communicator(comm_named_p1);
      yac_free_communicator(comm_named_p2);

      yac_reset_interpolation_method();
   }

   // simple fixed value "interpolation" on a single process
   // for two domains that do not overlap -> points are not interpolated
   for (unsigned mapping_on_source = 0; mapping_on_source < 2;
        ++mapping_on_source) {

      //--------------------
      // set up global search
      //--------------------

         // one directional global search with two domains on two processes with one subdomain each

         // comm worlds on both processes
         struct communicator * comm_p1, * comm_p2;

         comm_p1 = yac_local_comm_new();
         comm_p2 = yac_local_comm_new();

         // named comms for the domains on both processes
         struct communicator * comm_named_p1, * comm_named_p2;

         comm_named_p1 = yac_named_local_comm_new(comm_p1, "P1");
         comm_named_p2 = yac_named_local_comm_new(comm_p2, "P2");

         // define a domain on P1

         struct grid * grid_p1;
         unsigned num_cells[2] = {2,2};
         unsigned cyclic[2] = {0,0};
         double coordinates_x[2][3] = {{-1,0,1}, {-181.5,-180.5,180.5}};
         double coordinates_y[2][3] = {{-1,0,1}, {-1,0,1}};

         for ( unsigned i = 0; i < 2; ++i ) {
            for ( unsigned j = 0; j < 3; ++j ) {
               coordinates_x[i][j] *= YAC_RAD;
               coordinates_y[i][j] *= YAC_RAD;
            }
         }

         grid_p1 = yac_reg2d_grid_new(coordinates_x[0], coordinates_y[0],
                                      num_cells, cyclic);

         struct subdomain subdomain_p1;

         int global_corner_ids_p1[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p1[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p1[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p1[4] = {-1,-1,-1,-1};
         int global_edge_ids_p1[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p1[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p1, grid_p1, comm_named_p1,
                                     global_corner_ids_p1, corner_is_on_rank_p1,
                                     global_cell_ids_p1, cell_is_on_rank_p1,
                                     global_edge_ids_p1, edge_is_on_rank_p1);

         struct domain l_domain_p1;

         struct subdomain * blocks_p1[1] = {&subdomain_p1};

         yac_init_domain(&l_domain_p1, comm_named_p1, blocks_p1, 1, "domain_p1");

         // define a domain on P2, which matches the one on P1

         struct grid * grid_p2;

         grid_p2 = yac_reg2d_grid_new(coordinates_x[1], coordinates_y[1],
                                      num_cells, cyclic);

         struct subdomain subdomain_p2;

         int global_corner_ids_p2[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p2[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p2[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p2[4] = {-1,-1,-1,-1};
         int global_edge_ids_p2[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p2[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p2, grid_p2, comm_named_p2,
                                     global_corner_ids_p2, corner_is_on_rank_p2,
                                     global_cell_ids_p2, cell_is_on_rank_p2,
                                     global_edge_ids_p2, edge_is_on_rank_p2);

         struct domain l_domain_p2;

         struct subdomain * blocks_p2[1] = {&subdomain_p2};

         yac_init_domain(&l_domain_p2, comm_named_p2, blocks_p2, 1, "domain_p2");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1);
         yac_enddef_domain(&l_domain_p2);

         // init global search on both processes

         struct global_search * search[2];

         search[0] = yac_global_search_new(&l_domain_p1, "domain_p2", comm_p1,
                                           1,0, "component_p1", "component_p2");
         search[1] = yac_global_search_new(&l_domain_p2, "domain_p1", comm_p2,
                                           0,1, "component_p2", "component_p1");

         // finish initialisation of global search

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);

         // define points

         struct points points[2];

         yac_init_points(points+0, grid_p1, CORNER,
                         coordinates_x[0], coordinates_y[0]);
         yac_init_points(points+1, grid_p2, CORNER,
                         coordinates_x[1], coordinates_y[1]);

         // set up coupling field

         struct coupling_field * field[2];
         struct points * points_per_subdomain[2][1] = {{points+0}, {points+1}};
         struct points ** points_set[2] = {points_per_subdomain[0],
                                           points_per_subdomain[1]};

         field[0] = yac_coupling_field_new("source field", component_id, &l_domain_p1, points_set + 0, NULL, 1);
         field[1] = yac_coupling_field_new("target field", component_id, &l_domain_p2, points_set + 1, NULL, 1);

         // set up fixed interpolation method

         struct interpolation_method * method_fixed;
         double fixed_value = 1337;

         method_fixed = yac_new_interpolation_method_fixed(fixed_value, search[0]);

      //---------------------------
      // set up field interpolation
      //---------------------------

      struct field_interpolation_gen * interpolation_gen[2];
      struct interpolation_method * interpolation_method_stack[] = {method_fixed,
                                                                    NULL};

      interpolation_gen[0] =
        yac_new_field_interpolation_gen(points_set + 1, 1, "source field", search[0],
                                        "target field", 1, 0,
                                        interpolation_method_stack, 0, NULL, 1,
                                        mapping_on_source);
      interpolation_gen[1] =
        yac_new_field_interpolation_gen(points_set + 1, 1, "target field", search[1],
                                        "source field", 0, 1, NULL, 0, NULL, 1,
                                        mapping_on_source);
      yac_enddef_field_interpolation_gen(interpolation_gen[0]);
      yac_enddef_field_interpolation_gen(interpolation_gen[1]);

      struct field_interpolation * interpolation[2];

      interpolation[0] = yac_get_field_interpolation(interpolation_gen[0]);
      interpolation[1] = yac_get_field_interpolation(interpolation_gen[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double source_field[9] = {1,1,1, 1,1,1, 1,1,1};
      double * source_field_[1] = {source_field};
      double ** source_field__[1] = {source_field_};

      double target_field[9] = {-1,-1,-1, -1,-1,-1, -1,-1,-1};
      double *target_field_[1] = {target_field};

      double *** source_data[1] = {source_field__};
      double ** target_data[1] = {target_field_};

      yac_put_field(interpolation[0], source_data, 1);
      yac_get_field(interpolation[1], target_data, 1);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      for (unsigned i = 0; i < 9; ++i)
         if (double_are_unequal(target_data[0][0][i], -1.0))
            PUT_ERR("error in results\n");

      //---------------
      // cleanup
      //---------------

      yac_delete_field_interpolation(interpolation[0]);
      yac_delete_field_interpolation(interpolation[1]);

      yac_delete_field_interpolation_gen(interpolation_gen[0]);
      yac_delete_field_interpolation_gen(interpolation_gen[1]);

      yac_delete_interpolation_method(method_fixed);

      yac_coupling_field_delete(field[0]);
      yac_coupling_field_delete(field[1]);

      yac_free_points(points+0);
      yac_free_points(points+1);

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);

      yac_free_domain(&l_domain_p1);
      yac_free_domain(&l_domain_p2);

      yac_delete_grid(grid_p1);
      yac_delete_grid(grid_p2);

      yac_free_communicator(comm_p1);
      yac_free_communicator(comm_p2);

      yac_free_communicator(comm_named_p1);
      yac_free_communicator(comm_named_p2);

      yac_reset_interpolation_method();
   }

   yac_comm_finalize();
   yac_free_pointer_unique_lookup();

   return TEST_EXIT_CODE;
}

