/**
 * @file test_field_interpolation_parallel7.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <mpi.h>

#include "tests.h"
#include "test_common.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_mpi.h"
#include "domain.h"
#include "global_search.h"
#include "interpolation_method_fixed.h"
#include "interpolation_method_avg.h"
#include "fields.h"
#include "field_interpolation_gen.h"
#include "field_interpolation_gen_src.h"
#include "field_interpolation_gen_tgt.h"
#include "field_interpolation_gen_src_tgt.h"

int mapping_on_source;

int main(int argc, char *argv[]) {

   if (argc != 2) {
      PUT_ERR("wrong number of arguments\n");
      return TEST_EXIT_CODE;
   }

   mapping_on_source = strcmp(argv[1], "src") == 0;
   if (!mapping_on_source && strcmp(argv[1], "tgt")) {
      PUT_ERR("invalid argument (has to be either \"src\" or \"tgt\")\n");
      return TEST_EXIT_CODE;
   }

   struct communicator * global_comm;

   global_comm = yac_mpi_comm_new();

   if (yac_get_comm_size(global_comm) != 3) {

      PUT_ERR("wrong number of processes");
      yac_free_communicator(global_comm);

      MPI_Finalize();

      return TEST_EXIT_CODE;
   }

   int my_global_rank = yac_get_comm_rank(global_comm);

   unsigned is_source = my_global_rank > 0;
   unsigned is_target = my_global_rank < 2;

   struct communicator * target_comm = yac_comm_split(global_comm, is_target, 0);
   struct communicator * source_comm = yac_comm_split(global_comm, is_source, 0);

   /*
    * The source domain is distributed among 2 processes, of which one process has
    * one and the other two subdomains.
    *
    * The global source domain is a 5x4 grid:
    *
    *    24--44--25--45--26--46--27--47--28--48--29
    *    |       |       |       |       |       |
    *    34  15  36  16  38  17  40  18  42  19  43
    *    |       |       |       |       |       |
    *    18--33--19--35--20--37--21--39--22--41--23
    *    |       |       |       |       |       |
    *    23  10  25  11  27  12  29  13  31  14  32
    *    |       |       |       |       |       |
    *    12--22--13--24--14--26--15--28--16--30--17
    *    |       |       |       |       |       |
    *    12  05  14  06  16  07  18  08  20  09  21
    *    |       |       |       |       |       |
    *    06--11--07--13--08--15--09--17--10--19--11
    *    |       |       |       |       |       |
    *    01  00  03  01  05  02  07  03  09  04  10
    *    |       |       |       |       |       |
    *    00--01--01--02--02--04--03--06--04--08--05
    *
    * Subdomains are distributed as follows (xy => x = domain, y = subdomain)
    *
    *    01--01--01--01--01--01--01--01--01--01--01
    *    |       |       |       |       |       |
    *    01  01  01  01  01  01  01  01  01  01  01
    *    |       |       |       |       |       |
    *    01--01--01--01--01--01--01--01--01--01--10
    *    |       |       |       |       |       |
    *    01  01  01  01  01  10  10  10  10  10  10
    *    |       |       |       |       |       |
    *    01--01--00--01--01--10--10--10--10--10--10
    *    |       |       |       |       |       |
    *    00  00  00  00  00  10  10  10  10  10  10
    *    |       |       |       |       |       |
    *    00--00--00--00--00--10--10--10--10--10--10
    *    |       |       |       |       |       |
    *    00  00  00  00  10  10  10  10  10  10  10
    *    |       |       |       |       |       |
    *    00--00--00--00--10--10--10--10--10--10--10
    */

   unsigned num_cells_p0_src[2][2] = {{2,2},{5,2}},
            num_cells_p1_src[2] =  {3,3};
   unsigned cyclic_src[2] = {0,0};
   double coordinates_x_p0_src[2][6] = {{0*YAC_RAD,1*YAC_RAD,2*YAC_RAD},
                                        {0*YAC_RAD,1*YAC_RAD,2*YAC_RAD,
                                         3*YAC_RAD,4*YAC_RAD,5*YAC_RAD}},
          coordinates_x_p1_src[4] = {2*YAC_RAD,3*YAC_RAD,4*YAC_RAD,5*YAC_RAD};
   double coordinates_y_p0_src[2][3] = {{0*YAC_RAD,1*YAC_RAD,2*YAC_RAD},
                                        {2*YAC_RAD,3*YAC_RAD,4*YAC_RAD}};
   double coordinates_y_p1_src[4] = {0,1*YAC_RAD,2*YAC_RAD,3*YAC_RAD};

   int global_corner_ids_p0_src[2][18] =
      {{0,1,2,6,7,8,12,13,14},
       {12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29}};
   unsigned corner_is_on_rank_p0_src[2][18] =
      {{-1,-1,1,-1,-1,-1,0,-1,0},
       {-1, 0,-1, 1, 1, 1,-1,-1,-1,-1,-1, 1,-1,-1,-1,-1,-1,-1}};
   int global_cell_ids_p0_src[2][10] =
      {{0,1,2,5,6,7},
       {10,11,12,13,14,15,16,17,18,19}};
   unsigned cell_is_on_rank_p0_src[2][10] =
      {{-1,-1,-1,-1,-1,-1},
       {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}};

   int global_corner_ids_p1_src[16] =
      { 2, 3, 4, 5,  8, 9,10,11, 14,15,16,17, 20,21,22,23};
   unsigned corner_is_on_rank_p1_src[16] =
      {-1,-1,-1,-1,  0,-1,-1,-1,  0,-1,-1,-1,  0, 0, 0,-1};
   int global_cell_ids_p1_src[9] =
      {2,3,4,7,8,9,12,13,14};
   unsigned cell_is_on_rank_p1_src[9] =
      {-1,-1,-1,-1,-1,-1,-1,-1,-1};

   int my_rank_src = 0;
   unsigned num_subdomains_src = 0;
   struct grid * grid_src[2];
   struct subdomain subdomain_src[2];
   struct subdomain * subdomains_src[2];
   struct points points_src[2];
   struct points * points_per_subdomain_src[2][1];
   struct points ** points_set_src[2];
   struct domain domain_src;
   struct coupling_field * field_src;

   if (is_source) {

      my_rank_src = yac_get_comm_rank(source_comm);

      num_subdomains_src = (my_rank_src == 0)?2:1;

      if (my_rank_src == 0)
         for (unsigned i = 0; i < 2; ++i)
            grid_src[i] =
               yac_reg2d_grid_new(coordinates_x_p0_src[i], coordinates_y_p0_src[i],
                                  num_cells_p0_src[i], cyclic_src);
      else
         grid_src[0] =
            yac_reg2d_grid_new(coordinates_x_p1_src, coordinates_y_p1_src,
                               num_cells_p1_src, cyclic_src);

      if (my_rank_src == 0)
         for (unsigned i = 0; i < 2; ++i)
            yac_init_subdomain_unstruct(
               subdomain_src+i, grid_src[i], source_comm,
               global_corner_ids_p0_src[i], corner_is_on_rank_p0_src[i],
               global_cell_ids_p0_src[i], cell_is_on_rank_p0_src[i], NULL, NULL);
      else
         yac_init_subdomain_unstruct(
            subdomain_src+0, grid_src[0], source_comm,
            global_corner_ids_p1_src, corner_is_on_rank_p1_src,
            global_cell_ids_p1_src, cell_is_on_rank_p1_src, NULL, NULL);

      subdomains_src[0] = subdomain_src+0;
      subdomains_src[1] = subdomain_src+1;

      yac_init_domain(&domain_src, source_comm, subdomains_src,
                      num_subdomains_src, "source_domain");
      yac_enddef_domain(&domain_src);

      if (my_rank_src == 0)
         for (unsigned i = 0; i < 2; ++i)
            yac_init_points(points_src+i, grid_src[i], CORNER,
                            coordinates_x_p0_src[i], coordinates_y_p0_src[i]);
      else
         yac_init_points(points_src+0, grid_src[0], CORNER,
                         coordinates_x_p1_src, coordinates_y_p1_src);

      points_per_subdomain_src[0][0] = points_src+0;
      points_per_subdomain_src[1][0] = points_src+1;
      points_set_src[0] = points_per_subdomain_src[0];
      points_set_src[1] = points_per_subdomain_src[1];

      if (my_rank_src == 0)
         field_src = yac_coupling_field_new("source field", 0, &domain_src,
                                            points_set_src, NULL, 1);
      else
         field_src = yac_coupling_field_new("source field", 0, &domain_src,
                                            points_set_src, NULL, 1);
   }

   /*
    * The target domain is distributed among 2 processes with one subdomain each
    *
    * The global target domain is a 5x3 grid:
    *
    *    21--39--22--40--23--41--24--42--25--43--26--44--27
    *    |       |       |       |       |       |       |
    *    27  12  29  13  31  14  33  15  35  16  37  17  38
    *    |       |       |       |       |       |       |
    *    14--26--15--28--16--30--17--32--18--34--19--36--20
    *    |       |       |       |       |       |       |
    *    14  06  16  07  18  08  20  09  22  10  24  11  25
    *    |       |       |       |       |       |       |
    *    07--13--08--15--09--17--10--19--11--21--12--23--13
    *    |       |       |       |       |       |       |
    *    01  00  03  01  05  02  07  03  09  04  11  05  12
    *    |       |       |       |       |       |       |
    *    00--00--01--02--02--04--03--06--04--08--05--10--06
    *
    * Subdomains are distributed as follows (xy => x = domain, y = subdomain)
    *
    *    00--00--00--00--00--00--00--10--10--10--10--10--10
    *    |       |       |       |       |       |       |
    *    00  00  00  00  00  00  00  10  10  10  10  10  10
    *    |       |       |       |       |       |       |
    *    00--00--00--00--00--00--00--10--10--10--10--10--10
    *    |       |       |       |       |       |       |
    *    00  00  00  00  00  00  00  10  10  10  10  10  10
    *    |       |       |       |       |       |       |
    *    00--00--00--00--00--00--00--10--10--10--10--10--10
    *    |       |       |       |       |       |       |
    *    00  00  00  00  00  00  00  10  10  10  10  10  10
    *    |       |       |       |       |       |       |
    *    00--00--00--00--00--00--00--10--10--10--10--10--10
    */

   unsigned num_cells_tgt[2] = {3,3};
   unsigned cyclic_tgt[2] = {0,0};
   double coordinates_x_tgt[2][4] = {{0.5,1.5,2.5,3.5},{3.5,4.5,5.5,6.5}};
   double coordinates_y_tgt[4] = {0.5,1.5,2.5,3.5};

   int global_corner_ids_tgt[2][16] =
      {{0,1,2,3, 7,8,9,10, 14,15,16,17, 21,22,23,24},
       {3,4,5,6, 10,11,12,13, 17,18,19,20, 24,25,26,27}};
   unsigned corner_is_on_rank_tgt[2][16] =
      {{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
       {0,-1,-1,-1,0,-1,-1,-1,0,-1,-1,-1,0,-1,-1,-1}};
   int global_cell_ids_tgt[2][9] = {{0,1,2,6,7,8,12,13,14}, {3,4,5,9,10,11,15,16,17}};
   unsigned cell_is_on_rank_tgt[2][9] = {{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                     {-1,-1,-1,-1,-1,-1,-1,-1,-1}};

   int my_rank_tgt;
   struct grid * grid_tgt;
   struct subdomain subdomain_tgt;
   struct subdomain * subdomains_tgt[1];
   struct points points_tgt;
   struct points * points_per_subdomain_tgt[1];
   struct points ** points_set_tgt[1];
   struct domain domain_tgt;
   struct coupling_field * field_tgt;

   if (is_target) {
      my_rank_tgt = yac_get_comm_rank(target_comm);

      for ( unsigned j = 0; j < 4; ++j ) {
         coordinates_y_tgt[j] *= YAC_RAD;
         for ( unsigned i = 0; i < 2; ++i ) {
            coordinates_x_tgt[i][j] *= YAC_RAD;
         }
      }

      grid_tgt =
         yac_reg2d_grid_new(coordinates_x_tgt[my_rank_tgt], coordinates_y_tgt,
                            num_cells_tgt, cyclic_tgt);

      yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, target_comm,
                                  global_corner_ids_tgt[my_rank_tgt],
                                  corner_is_on_rank_tgt[my_rank_tgt],
                                  global_cell_ids_tgt[my_rank_tgt],
                                  cell_is_on_rank_tgt[my_rank_tgt], NULL, NULL);

      subdomains_tgt[0] = &subdomain_tgt;

      yac_init_domain(&domain_tgt, target_comm, subdomains_tgt, 1, "target_domain");
      yac_enddef_domain(&domain_tgt);

      yac_init_points(&points_tgt, grid_tgt, CORNER,
                      coordinates_x_tgt[my_rank_tgt], coordinates_y_tgt);

      points_per_subdomain_tgt[0] = &points_tgt;
      points_set_tgt[0] = points_per_subdomain_tgt;

      field_tgt =
         yac_coupling_field_new("target field", 1, &domain_tgt, points_set_tgt,
                                NULL, 1);
   }

   //---------------------------------------------
   // generate global search for source and target
   //---------------------------------------------

   struct global_search * search_src, * search_tgt;

   if (is_source && is_target) {
      struct domain * domains[2] = {&domain_src, &domain_tgt};
      unsigned domain_is_source[2] = {1, 0};
      char const * component_names[2] =
         {"source_component", "target_component"};
      struct global_search * searches[2];
      yac_global_search_all_local_new(domains, global_comm, domain_is_source,
                                      component_names, searches);
      search_src = searches[0];
      search_tgt = searches[1];
      yac_enddef_global_search(search_src);
      yac_enddef_global_search(search_tgt);
   } else if (is_source) {
      search_src =
         yac_global_search_new(&domain_src, "target_domain", global_comm, 1, 0,
                               "source_component", "target_component");
      yac_enddef_global_search(search_src);
   } else {
      search_tgt =
         yac_global_search_new(&domain_tgt, "source_domain", global_comm, 0, 1,
                               "target_component", "source_component");
      yac_enddef_global_search(search_tgt);
   }

   struct interpolation_method * method_fixed, * method_avg;
   struct interpolation_method * interpolation_method_stack[3];

   if (is_source) {

      double fixed_value = 1337;

      method_fixed = yac_new_interpolation_method_fixed(fixed_value, search_src);
      method_avg =
         yac_new_interpolation_method_avg(search_src, &domain_src, "avg stack",
                                          points_set_src, NULL, 1, 0, 0, 0);

      interpolation_method_stack[0] = method_avg;
      interpolation_method_stack[1] = method_fixed;
      interpolation_method_stack[2] = NULL;
   }

   //---------------------------
   // set up field interpolation
   //---------------------------

   char const filename_out[] = "weight_file_out.nc";

   struct field_interpolation_gen * interpolation_gen_src,
                                  * interpolation_gen_tgt;

   if (is_source && is_target) {
      yac_new_field_interpolation_gen_src_tgt_2(
        points_set_src, 1, "source field", search_src,
        interpolation_method_stack, filename_out, 1, points_set_tgt, 1,
        "target field", search_tgt, 0, &interpolation_gen_src,
        &interpolation_gen_tgt, mapping_on_source);
   } else if (is_source) {
     interpolation_gen_src =
        yac_new_field_interpolation_gen_src(
           points_set_src, 1, "source field", search_src, "target field",
           interpolation_method_stack, filename_out, 1, mapping_on_source);
      interpolation_gen_tgt = NULL;
   } else {
      interpolation_gen_src = NULL;
      interpolation_gen_tgt =
        yac_new_field_interpolation_gen_tgt(
           points_set_tgt, 1, "target field", search_tgt, "source field", 0,
           mapping_on_source);
   }

   if (is_source) yac_enddef_field_interpolation_gen(interpolation_gen_src);
   if (is_target) yac_enddef_field_interpolation_gen(interpolation_gen_tgt);

   struct field_interpolation * interpolation_src, * interpolation_tgt;

   if (is_source)
      interpolation_src = yac_get_field_interpolation(interpolation_gen_src);
   if (is_target)
      interpolation_tgt = yac_get_field_interpolation(interpolation_gen_tgt);

   if (is_source) {

      //---------------------
      // do the interpolation
      //---------------------

      double *** source_data[1];

      double source_field_p0_subdomain0[] = { 0, 1,-1,
                                              6, 7, 8,
                                             -1,13,-1};
      double * source_field_p0_subdomain0_[1] = {source_field_p0_subdomain0};

      double source_field_p0_subdomain1[] = {12,-1,14,-1,-1,-1,
                                             18,19,20,21,22,-1,
                                             24,25,26,27,28,29};

      double * source_field_p0_subdomain1_[1] = {source_field_p0_subdomain1};

      double source_field_p1[] = { 2, 3, 4, 5,
                                  -1, 9,10,11,
                                  -1,15,16,17,
                                  -1,-1,-1,23};
      double * source_field_p1_[1] = {source_field_p1};

      double ** source_field_p0_subdomain0__[2];

      if (my_rank_src == 0) {
         source_field_p0_subdomain0__[0] = source_field_p0_subdomain0_;
         source_field_p0_subdomain0__[1] = source_field_p0_subdomain1_;
      } else {
         source_field_p0_subdomain0__[0] = source_field_p1_;
         source_field_p0_subdomain0__[1] = NULL;
      }

      source_data[0] = source_field_p0_subdomain0__;

      yac_put_field(interpolation_src, source_data, 1);
   }

   if (is_target) {

      //---------------------
      // do the interpolation
      //---------------------

      double target_field[16] = {-1,-1,-1,-1,
                                 -1,-1,-1,-1,
                                 -1,-1,-1,-1,
                                 -1,-1,-1,-1};
      double * target_field_[1] = {target_field};
      double ** target_data[1] = {target_field_};

      yac_get_field(interpolation_tgt, target_data, 1);

      //----------------------------
      // check interpolation results
      //----------------------------

      double ref_target_data[2][16] = {{3.5,4.5,5.5,6.5,
                                        9.5,10.5,11.5,12.5,
                                        15.5,16.5,17.5,18.5,
                                        21.5,22.5,23.5,24.5},
                                       {-1,7.5,1337,-1,
                                        -1,13.5,1337,-1,
                                        -1,19.5,1337,-1,
                                        -1,25.5,1337,-1}};
      unsigned field_size[2] = {16,16};

      for (unsigned i = 0; i < field_size[my_rank_tgt]; ++i)
         if (double_are_unequal(
              target_data[0][0][i], ref_target_data[my_rank_tgt][i]))
            PUT_ERR("error in interpolated data on target side\n");
   }

   //---------
   // clean up
   //---------

   if (is_source) {
      yac_delete_field_interpolation_gen(interpolation_gen_src);
      yac_delete_field_interpolation(interpolation_src);
      yac_delete_interpolation_method(method_avg);
      yac_delete_interpolation_method(method_fixed);
      yac_global_search_delete(search_src);
      yac_coupling_field_delete(field_src);
      for (unsigned i = 0; i < num_subdomains_src; ++i)
         yac_free_points(&(points_src[i]));
      yac_free_domain(&domain_src);
      for (unsigned i = 0; i < num_subdomains_src; ++i)
         yac_delete_grid(grid_src[i]);
   }
   if (is_target) {
      yac_delete_field_interpolation_gen(interpolation_gen_tgt);
      yac_delete_field_interpolation(interpolation_tgt);
      yac_global_search_delete(search_tgt);
      yac_coupling_field_delete(field_tgt);
      yac_free_points(&points_tgt);
      yac_free_domain(&domain_tgt);
      yac_delete_grid(grid_tgt);
   }

   yac_free_communicator(source_comm);
   yac_free_communicator(target_comm);
   yac_free_communicator(global_comm);

   MPI_Finalize();

   // delete weight file
   if (my_global_rank == 0)
     unlink(filename_out);

   return TEST_EXIT_CODE;
}

