/**
 * @file test_grid.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "tests.h"
#include "grid.h"
#include "grid_reg2d.h"
#include "grid_curve2d.h"
#include "grid_unstruct.h"
#include "geometry.h"
#include "ensure_array_size.h"
#include "area.h"
#include "test_common.h"
#include "utils.h"
#include "communicator_local.h"

enum grid_type {

   REG2D,
   UNSTRUCT,
   CURVE2D,
};

static double const tol = 1.0e-12;

static int test_grid_cell_corner(
   struct grid_cell cell_a, size_t corner_index_a,
   struct grid_cell cell_b, size_t corner_index_b) {

   double temp;

   temp = cell_a.coordinates_x[corner_index_a];

   while (cell_b.coordinates_x[corner_index_b] - temp > M_PI)
      temp += 2*M_PI;
   while (temp - cell_b.coordinates_x[corner_index_b] > M_PI)
      temp -= 2*M_PI;

   return ((fabs(temp - cell_b.coordinates_x[corner_index_b]) < tol) &&
           (fabs(cell_a.coordinates_y[corner_index_a] -
                 cell_b.coordinates_y[corner_index_b]) < tol) &&
           (get_vector_angle(cell_a.coordinates_xyz[corner_index_a],
                             cell_b.coordinates_xyz[corner_index_b]) < tol));
}

static int test_grid_cell (struct grid_cell cell_a, struct grid_cell cell_b) {

   size_t i, j;

   if (cell_a.num_corners != cell_b.num_corners) return 0;

   // find the first matching corner (order can be reversed or start at different corners)
   // this assumes that all corners are different
   for (i = 0; i < cell_a.num_corners; ++i) {

      if (test_grid_cell_corner(cell_a, i, cell_b, 0)) break;
   }

   if (i == cell_a.num_corners) return 0;

   if (test_grid_cell_corner(cell_a, i+1, cell_b, 1)) {

      for (j = 2; j < cell_a.num_corners; ++j) {

         if (!test_grid_cell_corner(cell_a, (i+j)%cell_a.num_corners,
                                    cell_b, j)) return 0;
      }

   } else if (test_grid_cell_corner(cell_a, i+1, cell_b, cell_a.num_corners - 1)) {

      for (j = 2; j < cell_a.num_corners; ++j) {

         if (!test_grid_cell_corner(cell_a, (i+j)%cell_a.num_corners,
                                    cell_b, cell_a.num_corners - j)) return 0;
      }
   } else {
      return 0;
   }

   return 1;
}

static int test_corner_indices(
   unsigned const * indices_a, unsigned const * indices_b,
   unsigned num_indices) {

   unsigned i, j;

   for (i = 0; i < num_indices; ++i) {

      if (indices_a[i] == indices_b[0]) break;
   }

   if (i == num_indices) return 0;

   if (indices_a[(i+1)%num_indices] == indices_b[1]) {

      for (j = 2; j < num_indices; ++j) {

         if (indices_a[(i+j)%num_indices] !=indices_b[j]) return 0;
      }

   } else if (indices_a[(i+1)%num_indices] == indices_b[num_indices - 1]) {

      for (j = 2; j < num_indices; ++j) {

         if (indices_a[(i+j)%num_indices] != indices_b[num_indices - j]) return 0;
      }
   } else {
      return 0;
   }

   return 1;
}

static int test_indices(unsigned const * indices_a, unsigned const * indices_b,
                        unsigned num_indices) {

   unsigned i, j;

   // check that both arrays only contain unique values
   for (i = 0; i < num_indices; ++i) {
      for (j = 0; j < num_indices; ++j) {

         if (i != j && (indices_a[i] == indices_a[j] || indices_b[i] == indices_b[j]))
            return 0;
      }
   }

   unsigned count = 0;

   for (i = 0; i < num_indices; ++i)
      for (j = 0; j < num_indices; ++j)
         if (indices_a[i] == indices_b[j]) count++;

   return count == num_indices;
}

static void test_grid_1x1(struct grid * grid, char * grid_name) {

  struct grid_cell cell, ref_cell;

  double ref_cell_coords_x[4] = {0.5,0.5,1.5,1.5};
  double ref_cell_coords_y[4] = {0.5,1.5,1.5,0.5};
  double ref_cell_coords_xyz[4][3];

  for (unsigned i = 0; i < 4; ++i) {
    ref_cell_coords_x[i] *= YAC_RAD;
    ref_cell_coords_y[i] *= YAC_RAD;
    LLtoXYZ(ref_cell_coords_x[i], ref_cell_coords_y[i], ref_cell_coords_xyz[i]);
  }

  ref_cell.num_corners = 4;
  ref_cell.coordinates_x = ref_cell_coords_x;
  ref_cell.coordinates_y = ref_cell_coords_y;
  ref_cell.coordinates_xyz = ref_cell_coords_xyz;

#ifdef VERBOSE
  printf("testing grid: %s\n", grid_name);
#else
   UNUSED(grid_name);
#endif

  if (yac_get_num_grid_corners(grid) != 4) PUT_ERR("error in get_num_grid_corners\n")
  if (yac_get_num_grid_cells(grid) != 1) PUT_ERR("error in get_num_grid_cells\n")
  if (yac_get_num_grid_edges(grid) != 4) PUT_ERR("error in get_num_grid_edges\n")

  yac_init_grid_cell(&cell);
  yac_get_grid_cell(grid, 0, &cell);
  if (!test_grid_cell(ref_cell, cell))
    PUT_ERR("error in get_grid_cell \n")
  yac_free_grid_cell(&cell);

  for (unsigned i = 0; i < 4; ++i)
    if (yac_get_edge_type(grid, i) != GREAT_CIRCLE)
      PUT_ERR("error in get_edge_type\n")
}

// this test assumes that the underlying grid data is based on a regular
// 2d grid with x coords == {0,1,2} and y coord == {0,1,2}
static void test_grid_2x2(
   struct grid * grid, char * grid_name, enum grid_type type) {

   unsigned i, j, k;

   struct grid_cell cell, ref_cell;

   double ref_cell_coords_x[4][4] = {{0,1,1,0},{1,2,2,1},{0,1,1,0},{1,2,2,1}};
   double ref_cell_coords_y[4][4] = {{0,0,1,1},{0,0,1,1},{1,1,2,2},{1,1,2,2}};
   double ref_cell_coords_xyz[4][4][3];

   for ( i = 0; i < 4; ++i ) {
      for ( j = 0; j < 4; ++j ) {
         ref_cell_coords_x[i][j] *= YAC_RAD;
         ref_cell_coords_y[i][j] *= YAC_RAD;
         LLtoXYZ(ref_cell_coords_x[i][j], ref_cell_coords_y[i][j],
                 ref_cell_coords_xyz[i][j]);
      }
   }

   double ref_grid_extent[2][2] = {{0,2*YAC_RAD},{0,2*YAC_RAD}};
   double grid_extent[2][2];

   unsigned ref_cell_corner_indices[4][4] = {{0,1,4,3}, {1,2,5,4}, {3,4,7,6}, {4,5,8,7}};
   unsigned const * cell_corner_indices;

   unsigned ref_bnd_corners[8] = {0,1,2,3,5,6,7,8};
   unsigned bnd_corners[9], num_bnd_corners;
   unsigned num_matching_bnd_corners;

   unsigned ref_num_edges[9] = {2,3,2,3,4,3,2,3,2};
   unsigned ref_edge_corners[9][4] = {{1,3,-1,-1},{0,2,4,-1},{1,5,-1,-1},
                                      {0,4,6,-1},{1,3,5,7},{2,4,8,-1},
                                      {3,7,-1,-1},{4,6,8,-1},{5,7,-1,-1}};
   unsigned const * curr_edge_corners;
   unsigned matching_edges;

   unsigned const * curr_cell_edges;
   unsigned ref_cell_edges [4][4] = {{0,3,5,1},{2,4,7,3},{5,8,10,6},{7,9,11,8}};

   unsigned ref_num_corner_cells[9] = {1,2,1,2,4,2,1,2,1};
   unsigned ref_corner_cells[9][4] = {{0,-1,-1,-1},{0,1,-1,-1},{1,-1,-1,-1},
                                      {0,2,-1,-1},{0,1,2,3},{1,3,-1,-1},
                                      {2,-1,-1,-1},{2,3,-1,-1},{3,-1,-1,-1}};
   unsigned ref_aux_grid_cell_indices[9][4] = {{-1},{-1},{-1}, {-1},{0,1,3,2},{-1}, {-1},{-1},{-1}};

#ifdef VERBOSE
   printf("testing grid: %s\n", grid_name);
#else
   UNUSED(grid_name);
#endif

   if (yac_get_num_grid_corners(grid) != 9) PUT_ERR("error in get_num_grid_corners\n")
   if (yac_get_num_grid_cells(grid) != 4) PUT_ERR("error in get_num_grid_cells\n")
   if (yac_get_num_grid_edges(grid) != 12) PUT_ERR("error in get_num_grid_edges\n")

   yac_get_2d_grid_extent(grid, grid_extent);

   if (ref_grid_extent[0][0] < grid_extent[0][0] ||
       ref_grid_extent[0][1] > grid_extent[0][1] ||
       ref_grid_extent[1][0] < grid_extent[1][0] ||
       ref_grid_extent[1][1] > grid_extent[1][1]) {
      PUT_ERR("error in get_2d_grid_extent\n")
   }
   
   yac_init_grid_cell(&cell);
   ref_cell.num_corners = 4;
   for (i = 0; i < 4; ++i) {

      ref_cell.coordinates_x = ref_cell_coords_x[i];
      ref_cell.coordinates_y = ref_cell_coords_y[i];
      ref_cell.coordinates_xyz = ref_cell_coords_xyz[i];

      yac_get_grid_cell(grid, i, &cell);

      if (!test_grid_cell (ref_cell, cell))
         PUT_ERR("error in get_grid_cell \n")
   }

   for (i = 0; i < 4; ++i) {

      cell_corner_indices = yac_get_cell_corner_indices(grid, i);

      if (!test_corner_indices(cell_corner_indices, ref_cell_corner_indices[i],4))
         PUT_ERR("error in get_cell_corner_indices\n")
   }

   num_matching_bnd_corners = 0;

   yac_get_boundary_corners (grid, bnd_corners, &num_bnd_corners);

   if (num_bnd_corners != 8)
      PUT_ERR("error in get_boundary_corners (num_bnd_corners)\n")

   for (i = 0; i < 8; ++i) {
   
      for (j = 0; j < 8; ++j) {

         if (bnd_corners[i] == ref_bnd_corners[j]) ++num_matching_bnd_corners;
      }
   }

   if (num_matching_bnd_corners != num_bnd_corners)
      PUT_ERR("error in get_boundary_corners (bnd_corners)\n")

   for (i = 0; i < 9; ++i) {

      if (yac_get_num_corner_edges(grid, i) != ref_num_edges[i])
         PUT_ERR("error in get_num_corner_edges\n")

      curr_edge_corners = yac_get_corner_edges(grid, i);

      matching_edges = 0;

      for (j = 0; j < ref_num_edges[i]; ++j) {
         for (k = 0; k < ref_num_edges[i]; ++k) {

            if (curr_edge_corners[j] == ref_edge_corners[i][k])
               ++matching_edges;
         }
      }

      if (matching_edges != ref_num_edges[i])
         PUT_ERR("error in get_corner_edges\n")
   }

   for (i = 0; i < 4; ++i) {

      if (yac_get_num_cell_edges(grid,i) != 4)
         PUT_ERR("error in get_num_cell_edges\n")

      curr_cell_edges = yac_get_cell_edge_indices(grid, i);

      matching_edges = 0;

      for (j = 0; j < 4; ++j) {
         for (k = 0; k < 4; ++k) {

            if (curr_cell_edges[j] == ref_cell_edges[i][k])
               ++matching_edges;
         }
      }

      if (matching_edges != 4)
         PUT_ERR("error in get_cell_edge_indices\n")
   }

   yac_free_grid_cell(&cell);

   if (type == REG2D) {

      enum yac_edge_type ref_edge_type[12] = {LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE, LON_CIRCLE,
                                              LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE, LON_CIRCLE,
                                              LAT_CIRCLE, LAT_CIRCLE};

      for (i = 0; i < 12; ++i)
         if (yac_get_edge_type(grid, i) != ref_edge_type[i])
            PUT_ERR("error in get_edge_type\n")

   } else if (type == CURVE2D || type == UNSTRUCT) {

      for (i = 0; i < 12; ++i)
         if (yac_get_edge_type(grid, i) != GREAT_CIRCLE)
            PUT_ERR("error in get_edge_type\n")
   }

   for (i = 0; i < 9; ++i)
      if (yac_get_num_corner_cells(grid, i) != ref_num_corner_cells[i])
         PUT_ERR("error in get_num_corner_cells\n");

   for (i = 0; i < 9; ++i) {

      unsigned const * corner_cells = yac_get_corner_cell_indices(grid, i);

      for (j = 0; j < ref_num_corner_cells[i]; ++j)
         if (corner_cells[j] != ref_corner_cells[i][j])
            PUT_ERR("error in get_corner_cell_indices\n");
   }

   for (i = 0; i < 9; ++i) {

      unsigned cell_indices[4];
      enum yac_edge_type edge_type[4];
      int ret_val;

      ret_val = yac_get_aux_grid_cell(grid, i, cell_indices, edge_type);

      if ((ret_val && (ref_num_corner_cells[i] != 4)) ||
          (!ret_val && (ref_num_corner_cells[i] == 4))) {
         PUT_ERR("error in get_aux_grid_cell\n");
      } else if (ret_val) {
         if (!test_corner_indices(ref_aux_grid_cell_indices[i], cell_indices, 4))
            PUT_ERR("error in get_aux_grid_cell\n");
      }
   }
}

// this test assumes that the underlying grid data is based on a regular
// 2d grid with x coords == {0,1,2} and y coord == {0,1,2,3}
static void test_grid_2x3(
   struct grid * grid, char * grid_name, enum grid_type type) {

   unsigned i, j, k;

   struct grid_cell cell, ref_cell;

   double ref_cell_coords_x[6][4] = {{0,1,1,0},{1,2,2,1},{0,1,1,0},{1,2,2,1},{0,1,1,0},{1,2,2,1}};
   double ref_cell_coords_y[6][4] = {{0,0,1,1},{0,0,1,1},{1,1,2,2},{1,1,2,2},{2,2,3,3},{2,2,3,3}};
   double ref_cell_coords_xyz[6][4][3];

   for ( i = 0; i < 6; ++i ) {
      for ( j = 0; j < 4; ++j ) {
         ref_cell_coords_x[i][j] *= YAC_RAD;
         ref_cell_coords_y[i][j] *= YAC_RAD;
         LLtoXYZ(ref_cell_coords_x[i][j], ref_cell_coords_y[i][j],
                 ref_cell_coords_xyz[i][j]);
      }
   }

   double ref_grid_extent[2][2] = {{0,2*YAC_RAD},{0,3*YAC_RAD}};
   double grid_extent[2][2];

   unsigned ref_cell_corner_indices[6][4] = {{0,1,4,3}, {1,2,5,4}, {3,4,7,6},
                                             {4,5,8,7}, {6,7,10,9}, {7,8,11,10}};
   unsigned const * cell_corner_indices;

   unsigned ref_bnd_corners[10] = {0,1,2,3,5,6,8,9,10,11};
   unsigned bnd_corners[12], num_bnd_corners;
   unsigned num_matching_bnd_corners;

   unsigned ref_num_edges[12] = {2,3,2,3,4,3,3,4,3,2,3,2};
   unsigned ref_edge_corners[12][4] = {{1,3,-1,-1},{0,2,4,-1},{1,5,-1,-1},
                                       {0,4,6,-1},{1,3,5,7},{2,4,8,-1},
                                       {3,7,9,-1},{4,6,8,10},{5,7,11,-1},
                                       {6,10,-1,-1},{7,9,11,-1},{8,10,-1,-1}};
   unsigned const * curr_edge_corners;
   unsigned matching_edges;

   unsigned const * curr_cell_edges;
   unsigned ref_cell_edges [6][4] = {{0,3,5,1},{2,4,7,3},{5,8,10,6},
                                     {7,9,12,8},{10,13,15,11},{12,14,16,13}};

   unsigned ref_num_corner_cells[12] = {1,2,1, 2,4,2, 2,4,2, 1,2,1};
   unsigned ref_corner_cells[12][4] = {{0},{0,1},{1},
                                       {0,2},{0,1,2,3},{1,3},
                                       {2,4},{2,3,4,5},{3,5},
                                       {4},{4,5},{5}};
   unsigned ref_aux_grid_cell_indices[12][4] = {{-1},{-1},{-1}, {-1},{0,1,3,2},{-1},
                                                {-1},{2,3,5,4},{-1}, {-1},{-1},{-1}};

#ifdef VERBOSE
   printf("testing grid: %s\n", grid_name);
#else
   UNUSED(grid_name);
#endif

   if (yac_get_num_grid_corners(grid) != 12) PUT_ERR("error in get_num_grid_corners\n")
   if (yac_get_num_grid_cells(grid) != 6) PUT_ERR("error in get_num_grid_cells\n")
   if (yac_get_num_grid_edges(grid) != 17) PUT_ERR("error in get_num_grid_edges\n")

   yac_get_2d_grid_extent(grid, grid_extent);
   if (ref_grid_extent[0][0] < grid_extent[0][0] ||
       ref_grid_extent[0][1] > grid_extent[0][1] ||
       ref_grid_extent[1][0] < grid_extent[1][0] ||
       ref_grid_extent[1][1] > grid_extent[1][1]) {
      PUT_ERR("error in get_2d_grid_extent\n")
   }
   
   yac_init_grid_cell(&cell);
   ref_cell.num_corners = 4;
   for (i = 0; i < 6; ++i) {

      ref_cell.coordinates_x = ref_cell_coords_x[i];
      ref_cell.coordinates_y = ref_cell_coords_y[i];
      ref_cell.coordinates_xyz = ref_cell_coords_xyz[i];

      yac_get_grid_cell(grid, i, &cell);

      if (!test_grid_cell (ref_cell, cell))
         PUT_ERR("error in get_grid_cell \n")
   }

   for (i = 0; i < 6; ++i) {

      cell_corner_indices = yac_get_cell_corner_indices(grid, i);
   
      if (!test_corner_indices(cell_corner_indices, ref_cell_corner_indices[i],4))
         PUT_ERR("error in get_cell_corner_indices\n")
   }

   num_matching_bnd_corners = 0;

   yac_get_boundary_corners (grid, bnd_corners, &num_bnd_corners);

   if (num_bnd_corners != 10)
      PUT_ERR("error in get_boundary_corners (num_bnd_corners)\n")

   for (i = 0; i < 10; ++i) {
   
      for (j = 0; j < 10; ++j) {

         if (bnd_corners[i] == ref_bnd_corners[j]) ++num_matching_bnd_corners;
      }
   }

   if (num_matching_bnd_corners != num_bnd_corners)
      PUT_ERR("error in get_boundary_corners (bnd_corners)\n")

   for (i = 0; i < 11; ++i) {

      if (yac_get_num_corner_edges(grid, i) != ref_num_edges[i])
         PUT_ERR("error in get_num_corner_edges\n")

      curr_edge_corners = yac_get_corner_edges(grid, i);

      matching_edges = 0;

      for (j = 0; j < ref_num_edges[i]; ++j) {
         for (k = 0; k < ref_num_edges[i]; ++k) {

            if (curr_edge_corners[j] == ref_edge_corners[i][k])
               ++matching_edges;
         }
      }

      if (matching_edges != ref_num_edges[i])
         PUT_ERR("error in get_corner_edges\n")
   }

   for (i = 0; i < 6; ++i) {

      if (yac_get_num_cell_edges(grid,i) != 4)
         PUT_ERR("error in get_num_cell_edges\n")

      curr_cell_edges = yac_get_cell_edge_indices(grid, i);

      matching_edges = 0;

      for (j = 0; j < 4; ++j) {
         for (k = 0; k < 4; ++k) {

            if (curr_cell_edges[j] == ref_cell_edges[i][k])
               ++matching_edges;
         }
      }

      if (matching_edges != 4)
         PUT_ERR("error in get_cell_edge_indices\n")
   }

   yac_free_grid_cell(&cell);

   if (type == REG2D) {

      enum yac_edge_type ref_edge_type[17] = {LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE, LON_CIRCLE,
                                              LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE, LON_CIRCLE,
                                              LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE, LON_CIRCLE,
                                              LAT_CIRCLE, LAT_CIRCLE};

      for (i = 0; i < 17; ++i)
         if (yac_get_edge_type(grid, i) != ref_edge_type[i])
            PUT_ERR("error in get_edge_type\n")

   } else if (type == CURVE2D || type == UNSTRUCT) {

      for (i = 0; i < 17; ++i)
         if (yac_get_edge_type(grid, i) != GREAT_CIRCLE)
            PUT_ERR("error in get_edge_type\n")
   }

   for (i = 0; i < 12; ++i)
      if (yac_get_num_corner_cells(grid, i) != ref_num_corner_cells[i])
         PUT_ERR("error in get_num_corner_cells\n");

   for (i = 0; i < 12; ++i) {

      unsigned const * corner_cells = yac_get_corner_cell_indices(grid, i);

      for (j = 0; j < ref_num_corner_cells[i]; ++j)
         if (corner_cells[j] != ref_corner_cells[i][j])
            PUT_ERR("error in get_corner_cell_indices\n");
   }

   for (i = 0; i < 12; ++i) {

      unsigned cell_indices[4];
      enum yac_edge_type edge_type[4];
      int ret_val;

      ret_val = yac_get_aux_grid_cell(grid, i, cell_indices, edge_type);

      if ((ret_val && (ref_num_corner_cells[i] != 4)) ||
          (!ret_val && (ref_num_corner_cells[i] == 4))) {
         PUT_ERR("error in get_aux_grid_cell\n");
      } else if (ret_val) {
         if (!test_corner_indices(ref_aux_grid_cell_indices[i], cell_indices, 4))
            PUT_ERR("error in get_aux_grid_cell\n");
      }
   }
}

// this test assumes that the underlying grid data is based on a regular
// 2d grid with x coords == {60,180,300,420} and y coord == {-60,0,60}
static void test_grid_3x2_1x0
   (struct grid * grid, char * grid_name, enum grid_type type) {

   unsigned i, j, k, l;

   struct grid_cell cell, ref_cell;

   double ref_cell_coords_x[6][4] = {{60,180,180,60},{180,300,300,180},{300,60,60,300},
                                     {60,180,180,60},{180,300,300,180},{300,60,60,300}};
   double ref_cell_coords_y[6][4] = {{-60,-60,0,0},{-60,-60,0,0},{-60,-60,0,0},
                                     {0,0,60,60},{0,0,60,60},{0,0,60,60}};
   double ref_cell_coords_xyz[6][4][3];

   for ( i = 0; i < 6; ++i ) {
      for ( j = 0; j < 4; ++j ) {
         ref_cell_coords_x[i][j] *= YAC_RAD;
         ref_cell_coords_y[i][j] *= YAC_RAD;
         LLtoXYZ(ref_cell_coords_x[i][j], ref_cell_coords_y[i][j],
                 ref_cell_coords_xyz[i][j]);
      }
   }

   double grid_extent[2][2];

   unsigned num_matching_cells;

   unsigned ref_cell_corner_indices[6][4] = {{0,1,4,3},{1,2,5,4},{2,0,3,5},
                                             {3,4,7,6},{4,5,8,7},{5,3,6,8}};
   unsigned const * cell_corner_indices;

   unsigned ref_bnd_corners[6] = {0,1,2,6,7,8};
   unsigned bnd_corners[9], num_bnd_corners;
   unsigned num_matching_bnd_corners;

   unsigned ref_num_edges[9] = {3,3,3,4,4,4,3,3,3};
   unsigned ref_edge_corners[9][4] = {{1,2,3,-1},{0,2,4,-1},{0,1,5,-1},
                                      {0,4,5,6},{1,3,5,7},{2,3,4,8},
                                      {3,7,8,-1},{4,6,8,-1},{5,6,7,-1}};
   unsigned const * curr_edge_corners;
   unsigned matching_edges;

   unsigned const * curr_cell_edges;
   unsigned ref_cell_edges [6][4] = {{0,4,6,2},{3,5,9,4},{1,2,7,5},
                                     {6,10,12,8},{9,11,14,10},{7,8,13,11}};

   unsigned ref_num_corner_cells[9] = {2,2,2, 4,4,4, 2,2,2};
   unsigned ref_corner_cells[9][4] = {{2,0},{0,1},{1,2},
                                      {2,0,5,3},{0,1,3,4},{1,2,4,5},
                                      {5,3},{3,4},{4,5}};
   unsigned ref_aux_grid_cell_indices[9][4] = {{-1},{-1},{-1}, {2,0,3,5},{0,1,4,3},{1,2,5,4}, {-1},{-1},{-1}};
   // alternative configuration might happen if grid is automatically generated
   // cell grid of a 3x3 grid
   unsigned ref_corner_cells_[9][4] = {{0,1},{1,2},{2,0},
                                       {0,1,3,4},{1,2,4,5},{2,0,5,3},
                                       {3,4},{4,5},{5,3}};
   unsigned ref_aux_grid_cell_indices_[9][4] = {{-1},{-1},{-1}, {0,1,4,3},{1,2,5,4},{2,0,3,5}, {-1},{-1},{-1}};
#ifdef VERBOSE
   printf("testing grid: %s\n", grid_name);
#else
   UNUSED(grid_name);
#endif

   if (yac_get_num_grid_corners(grid) != 9) PUT_ERR("error in get_num_grid_corners\n")
   if (yac_get_num_grid_cells(grid) != 6) PUT_ERR("error in get_num_grid_cells\n")
   if (yac_get_num_grid_edges(grid) != 15) PUT_ERR("error in get_num_grid_edges\n")

   yac_get_2d_grid_extent(grid, grid_extent);
   if (((grid_extent[0][1] - grid_extent[0][0]) < 360*YAC_RAD-tol) ||
        (grid_extent[1][0] > -60*YAC_RAD) || (grid_extent[1][1] < 60*YAC_RAD))
      PUT_ERR("error in get_2d_grid_extent\n")

   num_matching_cells = 0;
   yac_init_grid_cell(&cell);
   ref_cell.num_corners = 4;
   for (i = 0; i < 6; ++i) {

      yac_get_grid_cell(grid, i, &cell);

      for (j = 0; j < 6; ++j) {

         ref_cell.coordinates_x = ref_cell_coords_x[j];
         ref_cell.coordinates_y = ref_cell_coords_y[j];
         ref_cell.coordinates_xyz = ref_cell_coords_xyz[j];

         if (test_grid_cell (ref_cell, cell)) {
            num_matching_cells++;
            break;
         }
      }
   }

   if (num_matching_cells != 6)
      PUT_ERR("error in get_grid_cell \n")

   num_matching_cells = 0;
   for (i = 0; i < 6; ++i) {

      cell_corner_indices = yac_get_cell_corner_indices(grid, i);

      for (j = 0; j < 6; ++j) {

         if (test_corner_indices(cell_corner_indices, ref_cell_corner_indices[j],4))
            num_matching_cells++;
      }
   }

   if (num_matching_cells != 6)
      PUT_ERR("error in get_cell_corner_indices\n")

   num_matching_bnd_corners = 0;

   yac_get_boundary_corners (grid, bnd_corners, &num_bnd_corners);

   if (num_bnd_corners != 6)
      PUT_ERR("error in get_boundary_corners (num_bnd_corners)\n")

   for (i = 0; i < 6; ++i) {
   
      for (j = 0; j < 6; ++j) {

         if (bnd_corners[i] == ref_bnd_corners[j]) ++num_matching_bnd_corners;
      }
   }

   if (num_matching_bnd_corners != num_bnd_corners)
      PUT_ERR("error in get_boundary_corners (bnd_corners)\n")

   for (i = 0; i < 9; ++i) {

      if (yac_get_num_corner_edges(grid, i) != ref_num_edges[i])
         PUT_ERR("error in get_num_corner_edges\n")

      curr_edge_corners = yac_get_corner_edges(grid, i);

      matching_edges = 0;

      for (j = 0; j < ref_num_edges[i]; ++j) {
         for (k = 0; k < ref_num_edges[i]; ++k) {

            if (curr_edge_corners[j] == ref_edge_corners[i][k])
               ++matching_edges;
         }
      }

      if (matching_edges != ref_num_edges[i])
         PUT_ERR("error in get_corner_edges\n")
   }

   num_matching_cells = 0;
   for (i = 0; i < 6; ++i) {

      if (yac_get_num_cell_edges(grid,i) != 4)
         PUT_ERR("error in get_num_cell_edges\n")

      curr_cell_edges = yac_get_cell_edge_indices(grid, i);

      for (j = 0; j < 6; ++j) {

         matching_edges = 0;

         for (k = 0; k < 4; ++k) {
            for (l = 0; l < 4; ++l) {

               if (curr_cell_edges[k] == ref_cell_edges[j][l])
                  ++matching_edges;
            }
         }

         if (matching_edges == 4) {
            ++num_matching_cells;
            break;
         }
      }
   }

   if (num_matching_cells != 6)
      PUT_ERR("error in get_cell_edge_indices\n")

   yac_free_grid_cell(&cell);

   if (type == REG2D) {

      enum yac_edge_type ref_edge_type[15] = {LAT_CIRCLE, LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE, LON_CIRCLE,
                                              LAT_CIRCLE, LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE, LON_CIRCLE,
                                              LAT_CIRCLE, LAT_CIRCLE, LAT_CIRCLE};

      for (i = 0; i < 15; ++i)
         if (yac_get_edge_type(grid, i) != ref_edge_type[i])
            PUT_ERR("error in get_edge_type\n")

   } else if (type == CURVE2D || type == UNSTRUCT) {

      for (i = 0; i < 15; ++i)
         if (yac_get_edge_type(grid, i) != GREAT_CIRCLE)
            PUT_ERR("error in get_edge_type\n")
   }

   for (i = 0; i < 9; ++i)
      if (yac_get_num_corner_cells(grid, i) != ref_num_corner_cells[i])
         PUT_ERR("error in get_num_corner_cells\n");

   {
      int error = 0;

      for (i = 0; i < 9; ++i) {

         unsigned const * corner_cells = yac_get_corner_cell_indices(grid, i);

         if (!test_indices(corner_cells, ref_corner_cells[i], ref_num_corner_cells[i]))
            error++;
      }

      // check alternative configuration if error occured
      if (error) {
         for (i = 0; i < 9; ++i) {

            unsigned const * corner_cells = yac_get_corner_cell_indices(grid, i);

            if (!test_indices(corner_cells, ref_corner_cells_[i], ref_num_corner_cells[i]))
               PUT_ERR("error in get_corner_cell_indices\n");
         }
      }

      error = 0;

      for (i = 0; i < 9; ++i) {

         unsigned cell_indices[4];
         enum yac_edge_type edge_type[4];
         int ret_val;

         ret_val = yac_get_aux_grid_cell(grid, i, cell_indices, edge_type);

         if ((ret_val && (ref_num_corner_cells[i] != 4)) ||
             (!ret_val && (ref_num_corner_cells[i] == 4))) {
            error++;
         } else if (ret_val) {
            if (!test_corner_indices(ref_aux_grid_cell_indices[i], cell_indices, 4))
               error++;
         }
      }

      // check alternative configuration if error occured
      if (error) {
         for (i = 0; i < 9; ++i) {

            unsigned cell_indices[4];
            enum yac_edge_type edge_type[4];
            int ret_val;

            ret_val = yac_get_aux_grid_cell(grid, i, cell_indices, edge_type);

            if ((ret_val && (ref_num_corner_cells[i] != 4)) ||
                (!ret_val && (ref_num_corner_cells[i] == 4))) {
               PUT_ERR("error in get_aux_grid_cell\n");
            } else if (ret_val) {
               if (!test_corner_indices(ref_aux_grid_cell_indices_[i], cell_indices, 4))
                  PUT_ERR("error in get_aux_grid_cell\n");
            }
         }
      }
   }
}

// this test assumes that the underlying grid data is based on a regular
// 2d grid with x coords == {0,120,240,360} and y coord == {-90,-30,30,90}
static void test_grid_3x3_1x0(
   struct grid * grid, char * grid_name, enum grid_type type) {

   unsigned i, j, k, l;

   struct grid_cell cell, ref_cell;

   double ref_cell_coords_x[9][4] = {{0,120,120,0},{120,240,240,120},{240,360,360,240},
                                     {0,120,120,0},{120,240,240,120},{240,360,360,240},
                                     {0,120,120,0},{120,240,240,120},{240,360,360,240}};
   double ref_cell_coords_y[9][4] = {{-90,-90,-30,-30},{-90,-90,-30,-30},{-90,-90,-30,-30},
                                     {-30,-30,30,30},{-30,-30,30,30},{-30,-30,30,30},
                                     {30,30,90,90},{30,30,90,90},{30,30,90,90}};
   double ref_cell_coords_xyz[9][4][3];

   for ( i = 0; i < 9; ++i ) {
      for ( j = 0; j < 4; ++j ) {
         ref_cell_coords_x[i][j] *= YAC_RAD;
         ref_cell_coords_y[i][j] *= YAC_RAD;
         LLtoXYZ(ref_cell_coords_x[i][j], ref_cell_coords_y[i][j],
                 ref_cell_coords_xyz[i][j]);
      }
   }

   double grid_extent[2][2];

   unsigned num_matching_cells;

   unsigned ref_cell_corner_indices[9][4] = {{0,1,4,3},{1,2,5,4},{2,0,3,5},
                                             {3,4,7,6},{4,5,8,7},{5,3,6,8},
                                             {6,7,10,9},{7,8,11,10},{8,6,9,11}};
   unsigned const * cell_corner_indices;

   unsigned ref_bnd_corners[6] = {0,1,2,9,10,11};
   unsigned bnd_corners[12], num_bnd_corners;
   unsigned num_matching_bnd_corners;

   unsigned ref_num_edges[12] = {3,3,3,4,4,4,4,4,4,3,3,3};
   unsigned ref_edge_corners[12][4] = {{1,2,3,-1},{0,2,4,-1},{0,1,5,-1},
                                       {0,4,5,6},{1,3,5,7},{2,3,4,8},
                                       {3,7,8,9},{4,6,8,10},{5,6,7,11},
                                       {6,10,11,-1},{7,9,11,-1},{8,9,10,-1}};
   unsigned const * curr_edge_corners;
   unsigned matching_edges;

   unsigned const * curr_cell_edges;
   unsigned ref_cell_edges [9][4] = {{0,4,6,2},{3,5,9,4},{1,2,7,5},
                                     {6,10,12,8},{9,11,15,10},{7,8,13,11},
                                     {12,16,18,14},{15,17,20,16},{13,14,19,17}};

   unsigned ref_num_corner_cells[12] = {2,2,2, 4,4,4, 4,4,4, 2,2,2};
   unsigned ref_corner_cells[12][4] = {{2,0},{0,1},{1,2},
                                       {2,0,5,3},{0,1,3,4},{1,2,4,5},
                                       {5,3,8,6},{3,4,6,7},{4,5,7,8},
                                       {8,6},{6,7},{7,8}};
   unsigned ref_aux_grid_cell_indices[12][4] = {{-1},{-1},{-1}, {2,0,3,5},{0,1,4,3},{1,2,5,4},
                                                {5,3,6,8},{3,4,7,6},{4,5,8,7}, {-1},{-1},{-1},};

#ifdef VERBOSE
   printf("testing grid: %s\n", grid_name);
#else
   UNUSED(grid_name);
#endif

   if (yac_get_num_grid_corners(grid) != 12) PUT_ERR("error in get_num_grid_corners\n")
   if (yac_get_num_grid_cells(grid) != 9) PUT_ERR("error in get_num_grid_cells\n")
   if (yac_get_num_grid_edges(grid) != 21) PUT_ERR("error in get_num_grid_edges\n")

   yac_get_2d_grid_extent(grid, grid_extent);
   if (((grid_extent[0][1] - grid_extent[0][0]) < 360*YAC_RAD-tol) ||
        double_are_unequal(grid_extent[1][0], -90*YAC_RAD) ||
        double_are_unequal(grid_extent[1][1], 90*YAC_RAD))
      PUT_ERR("error in get_2d_grid_extent\n")

   num_matching_cells = 0;
   yac_init_grid_cell(&cell);
   ref_cell.num_corners = 4;
   for (i = 0; i < 9; ++i) {

      yac_get_grid_cell(grid, i, &cell);

      for (j = 0; j < 9; ++j) {

         ref_cell.coordinates_x = ref_cell_coords_x[j];
         ref_cell.coordinates_y = ref_cell_coords_y[j];
         ref_cell.coordinates_xyz = ref_cell_coords_xyz[j];

         if (test_grid_cell (ref_cell, cell)) {
            num_matching_cells++;
            break;
         }
      }
   }

   if (num_matching_cells != 9)
      PUT_ERR("error in get_grid_cell \n")

   num_matching_cells = 0;
   for (i = 0; i < 9; ++i) {

      cell_corner_indices = yac_get_cell_corner_indices(grid, i);

      for (j = 0; j < 9; ++j) {

         if (test_corner_indices(cell_corner_indices, ref_cell_corner_indices[j],4))
            num_matching_cells++;
      }
   }

   if (num_matching_cells != 9)
      PUT_ERR("error in get_cell_corner_indices\n")

   num_matching_bnd_corners = 0;

   yac_get_boundary_corners (grid, bnd_corners, &num_bnd_corners);

   if (num_bnd_corners != 6)
      PUT_ERR("error in get_boundary_corners (num_bnd_corners)\n")

   for (i = 0; i < 6; ++i) {
   
      for (j = 0; j < 6; ++j) {

         if (bnd_corners[i] == ref_bnd_corners[j]) ++num_matching_bnd_corners;
      }
   }

   if (num_matching_bnd_corners != num_bnd_corners)
      PUT_ERR("error in get_boundary_corners (bnd_corners)\n")

   for (i = 0; i < 12; ++i) {

      if (yac_get_num_corner_edges(grid, i) != ref_num_edges[i])
         PUT_ERR("error in get_num_corner_edges\n")

      curr_edge_corners = yac_get_corner_edges(grid, i);

      matching_edges = 0;

      for (j = 0; j < ref_num_edges[i]; ++j) {
         for (k = 0; k < ref_num_edges[i]; ++k) {

            if (curr_edge_corners[j] == ref_edge_corners[i][k])
               ++matching_edges;
         }
      }

      if (matching_edges != ref_num_edges[i])
         PUT_ERR("error in get_corner_edges\n")
   }

   num_matching_cells = 0;
   for (i = 0; i < 9; ++i) {

      if (yac_get_num_cell_edges(grid,i) != 4)
         PUT_ERR("error in get_num_cell_edges\n")

      curr_cell_edges = yac_get_cell_edge_indices(grid, i);

      for (j = 0; j < 9; ++j) {

         matching_edges = 0;

         for (k = 0; k < 4; ++k) {
            for (l = 0; l < 4; ++l) {

               if (curr_cell_edges[k] == ref_cell_edges[j][l])
                  ++matching_edges;
            }
         }

         if (matching_edges == 4) {
            ++num_matching_cells;
            break;
         }
      }
   }

   if (num_matching_cells != 9)
      PUT_ERR("error in get_cell_edge_indices\n")

   yac_free_grid_cell(&cell);

   if (type == REG2D) {

      enum yac_edge_type ref_edge_type[21] = {LAT_CIRCLE, LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE, LON_CIRCLE,
                                              LAT_CIRCLE, LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE, LON_CIRCLE,
                                              LAT_CIRCLE, LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE, LON_CIRCLE,
                                              LAT_CIRCLE, LAT_CIRCLE, LAT_CIRCLE};

      for (i = 0; i < 21; ++i)
         if (yac_get_edge_type(grid, i) != ref_edge_type[i])
            PUT_ERR("error in get_edge_type\n")

   } else if (type == CURVE2D || type == UNSTRUCT) {

      for (i = 0; i < 21; ++i)
         if (yac_get_edge_type(grid, i) != GREAT_CIRCLE)
            PUT_ERR("error in get_edge_type\n")
   }

   for (i = 0; i < 12; ++i)
      if (yac_get_num_corner_cells(grid, i) != ref_num_corner_cells[i])
         PUT_ERR("error in get_num_corner_cells\n");

   for (i = 0; i < 12; ++i) {

      unsigned const * corner_cells = yac_get_corner_cell_indices(grid, i);

      if (!test_indices(corner_cells, ref_corner_cells[i], ref_num_corner_cells[i]))
         PUT_ERR("error in get_corner_cell_indices\n");
   }

   for (i = 0; i < 12; ++i) {

      unsigned cell_indices[4];
      enum yac_edge_type edge_type[4];
      int ret_val;

      ret_val = yac_get_aux_grid_cell(grid, i, cell_indices, edge_type);

      if ((ret_val && (ref_num_corner_cells[i] != 4)) ||
          (!ret_val && (ref_num_corner_cells[i] == 4))) {
         PUT_ERR("error in get_aux_grid_cell\n");
      } else if (ret_val) {
         if (!test_corner_indices(ref_aux_grid_cell_indices[i], cell_indices, 4))
            PUT_ERR("error in get_aux_grid_cell\n");
      }
   }
}

// this test assumes that the underlying grid data is based on a regular
// 2d grid with x coords == {0,1,2,3} and y coord == {0,1,2,3} and cyclic == {0,0}
// cell indices:
//  6  7  8
//  3  4  5
//  0  1  2
static void test_grid_cell_neigh_dep_3x3_0x0(
   struct grid * grid, char * grid_name) {

   unsigned cell_neigh_dependencies[9][8] = {{1,4,3,-1,-1,-1,-1,-1}, // neighs of cell 0
                                             {2,5,4,3,0,-1,-1,-1}, // neighs of cell 1
                                             {5,4,1,-1,-1,-1,-1,-1}, // ...
                                             {0,1,4,7,6,-1,-1,-1},
                                             {1,2,5,8,7,6,3,0},
                                             {2,8,7,4,1,-1,-1,-1},
                                             {3,4,7,-1,-1,-1,-1,-1},
                                             {4,5,8,6,3,-1,-1,-1},
                                             {5,7,4,-1,-1,-1,-1,-1}};
   size_t num_neigh[9] = {3,5,3,5,8,5,3,5,3};

   struct dep_list * cell_neigh_dep = yac_get_cell_neigh_dep_list(grid);

#ifdef VERBOSE
   printf("testing grid: %s\n", grid_name);
#else
   UNUSED(grid_name);
#endif

   size_t ref_num_deps = 40;

   if (yac_get_total_num_dependencies(cell_neigh_dep) != ref_num_deps)
      PUT_ERR("error in get_total_num_dependencies\n")

   if (yac_get_num_elements(cell_neigh_dep) != 9)
      PUT_ERR("wrong number of elements in cell_neigh_dep\n")

   for (size_t i = 0; i < yac_get_num_elements(cell_neigh_dep); ++i) {

      unsigned const * curr_neighs =
         yac_get_dependencies_of_element(cell_neigh_dep,i);
      size_t curr_num_neighs =
         yac_get_num_dependencies_of_element(cell_neigh_dep, i);

      if (curr_num_neighs != num_neigh[i])
         PUT_ERR("wrong number neighbours in cell_neigh_dep.num_deps_per_element\n")

      size_t num_matching_neighs = 0;
      for (size_t j = 0; j < curr_num_neighs; ++j)
         for (size_t k = 0; k < curr_num_neighs; ++k)
            if (cell_neigh_dependencies[i][j] == curr_neighs[k])
               ++num_matching_neighs;

      if (num_neigh[i] != num_matching_neighs) {
         PUT_ERR("wrong dependency in cell_neigh_dep.dependencies\n")
#ifdef VERBOSE
         printf("cell index: %zu\n", i);
         puts("expected neighs :\n");
         for (size_t j = 0; j < num_neigh[i]; ++j)
            printf("%u ", cell_neigh_dependencies[i][j]);
         puts("\nneighs in cell_neigh_dep:\n");
         for (size_t j = 0; j < curr_num_neighs; ++j)
            printf("%u ", curr_neighs[j]);
         puts("\n\n");
#endif
      }
   }
}

// this test assumes that the underlying grid data is based on a regular
// 2d grid with x coords == {0,1,2,3} and y coord == {0,1,2} and cyclic == {0,0}
// cell indices:
//  3  4  5
//  0  1  2
static void test_grid_cell_neigh_dep_3x2_0x0(
   struct grid * grid, char * grid_name) {

   unsigned cell_neigh_dependencies[6][5] = {{1,4,3,-1,-1}, // neighs of cell 0
                                             {2,5,4,3,0}, // neighs of cell 1
                                             {5,4,1,-1,-1}, // ...
                                             {0,1,4,-1,-1},
                                             {1,2,5,3,0},
                                             {2,4,1,-1,-1}};
   size_t num_neigh[6] = {3,5,3,3,5,3};

   struct dep_list * cell_neigh_dep = yac_get_cell_neigh_dep_list(grid);

#ifdef VERBOSE
   printf("testing grid: %s\n", grid_name);
#else
   UNUSED(grid_name);
#endif

   size_t ref_num_deps = 22;

   if (yac_get_total_num_dependencies(cell_neigh_dep) != ref_num_deps)
      PUT_ERR("error in get_total_num_dependencies\n")

   if (yac_get_num_elements(cell_neigh_dep) != 6)
      PUT_ERR("wrong number of elements in cell_neigh_dep\n")

   for (size_t i = 0; i < yac_get_num_elements(cell_neigh_dep); ++i) {

      unsigned const * curr_neighs =
         yac_get_dependencies_of_element(cell_neigh_dep,i);
      size_t curr_num_neighs =
         yac_get_num_dependencies_of_element(cell_neigh_dep, i);

      if (curr_num_neighs != num_neigh[i])
         PUT_ERR("wrong number neighbours in cell_neigh_dep.num_deps_per_element\n")

      size_t num_matching_neighs = 0;
      for (size_t j = 0; j < curr_num_neighs; ++j)
         for (size_t k = 0; k < curr_num_neighs; ++k)
            if (cell_neigh_dependencies[i][j] == curr_neighs[k])
               ++num_matching_neighs;

      if (num_neigh[i] != num_matching_neighs) {
         PUT_ERR("wrong dependency in cell_neigh_dep.dependencies\n")
#ifdef VERBOSE
         printf("cell index: %zu\n", i);
         puts("expected neighs :\n");
         for (size_t j = 0; j < num_neigh[i]; ++j)
            printf("%u ", cell_neigh_dependencies[i][j]);
         puts("\nneighs in cell_neigh_dep:\n");
         for (size_t j = 0; j < curr_num_neighs; ++j)
            printf("%u ", curr_neighs[j]);
         puts("\n\n");
#endif
      }
   }
}

// this test assumes that the underlying grid data is based on a regular
// 2d grid with 3x3 cells and cyclic == {1,0}
// cell indices:
//  6  7  8
//  3  4  5
//  0  1  2
static void test_grid_cell_neigh_dep_3x3_1x0(
   struct grid * grid, char * grid_name) {

   unsigned cell_neigh_dependencies[9][8] = {{1,4,3,5,2,-1,-1,-1}, // neighs of cell 0
                                             {2,5,4,3,0,-1,-1,-1}, // neighs of cell 1
                                             {5,4,1,0,3,-1,-1,-1}, // ...
                                             {0,1,4,7,6,8,5,2},
                                             {1,2,5,8,7,6,3,0},
                                             {2,8,7,4,1,6,3,0},
                                             {3,4,7,8,5,-1,-1,-1},
                                             {4,5,8,6,3,-1,-1,-1},
                                             {5,7,4,3,6,-1,-1,-1}};
   size_t num_neigh[9] = {5,5,5,8,8,8,5,5,5};

   struct dep_list * cell_neigh_dep = yac_get_cell_neigh_dep_list(grid);

#ifdef VERBOSE
   printf("testing grid: %s\n", grid_name);
#else
   UNUSED(grid_name);
#endif

   size_t ref_num_deps = 54;

   if (yac_get_total_num_dependencies(cell_neigh_dep) != ref_num_deps)
      PUT_ERR("error in get_total_num_dependencies\n")

   if (yac_get_num_elements(cell_neigh_dep) != 9)
      PUT_ERR("wrong number of elements in cell_neigh_dep\n")

   for (size_t i = 0; i < yac_get_num_elements(cell_neigh_dep); ++i) {

      unsigned const * curr_neighs =
         yac_get_dependencies_of_element(cell_neigh_dep,i);
      size_t curr_num_neighs =
         yac_get_num_dependencies_of_element(cell_neigh_dep, i);

      if (curr_num_neighs != num_neigh[i])
         PUT_ERR("wrong number neighbours in cell_neigh_dep.num_deps_per_element\n")

      size_t num_matching_neighs = 0;
      for (size_t j = 0; j < curr_num_neighs; ++j)
         for (size_t k = 0; k < curr_num_neighs; ++k)
            if (cell_neigh_dependencies[i][j] == curr_neighs[k])
               ++num_matching_neighs;

      if (num_neigh[i] != num_matching_neighs) {

         PUT_ERR("wrong dependency in cell_neigh_dep.dependencies\n")
#ifdef VERBOSE
         printf("cell index: %zu\n", i);
         puts("expected neighs :\n");
         for (size_t j = 0; j < num_neigh[i]; ++j)
            printf("%u ", cell_neigh_dependencies[i][j]);
         puts("\nneighs in cell_neigh_dep:\n");
         for (size_t j = 0; j < curr_num_neighs; ++j)
            printf("%u ", curr_neighs[j]);
         puts("\n\n");
#endif
      }
   }
}

// this test assumes that the underlying grid data is based on a regular
// 2d grid with 3x2 cells and cyclic == {0,0}
// cell indices:
//  3  4  5
//  0  1  2
static void test_grid_cell_neigh_dep_3x2_1x0(
   struct grid * grid, char * grid_name) {

   unsigned cell_neigh_dependencies[6][5] = {{  1,2,3,4,5}, // neighs of cell 0
                                             {0,  2,3,4,5}, // neighs of cell 1
                                             {0,1,  3,4,5}, // ...
                                             {0,1,2,  4,5},
                                             {0,1,2,3,  5},
                                             {0,1,2,3,4, }};
   size_t num_neigh[6] = {5,5,5,5,5,5};

   struct dep_list * cell_neigh_dep = yac_get_cell_neigh_dep_list(grid);

#ifdef VERBOSE
   printf("testing grid: %s\n", grid_name);
#else
   UNUSED(grid_name);
#endif

   size_t ref_num_deps = 30;

   if (yac_get_total_num_dependencies(cell_neigh_dep) != ref_num_deps)
      PUT_ERR("error in get_total_num_dependencies\n")

   if (yac_get_num_elements(cell_neigh_dep) != 6)
      PUT_ERR("wrong number of elements in cell_neigh_dep\n")

   for (size_t i = 0; i < yac_get_num_elements(cell_neigh_dep); ++i) {

      unsigned const * curr_neighs =
         yac_get_dependencies_of_element(cell_neigh_dep,i);
      size_t curr_num_neighs =
         yac_get_num_dependencies_of_element(cell_neigh_dep, i);

      if (curr_num_neighs != num_neigh[i])
         PUT_ERR("wrong number neighbours in cell_neigh_dep.num_deps_per_element\n")

      size_t num_matching_neighs = 0;
      for (size_t j = 0; j < curr_num_neighs; ++j)
         for (size_t k = 0; k < curr_num_neighs; ++k)
            if (cell_neigh_dependencies[i][j] == curr_neighs[k])
               ++num_matching_neighs;

      if (num_neigh[i] != num_matching_neighs) {

         PUT_ERR("wrong dependency in cell_neigh_dep.dependencies\n")
#ifdef VERBOSE
         printf("cell index: %zu\n", i);
         puts("expected neighs :\n");
         for (size_t j = 0; j < num_neigh[i]; ++j)
            printf("%u ", cell_neigh_dependencies[i][j]);
         puts("\nneighs in cell_neigh_dep:\n");
         for (size_t j = 0; j < curr_num_neighs; ++j)
            printf("%u ", curr_neighs[j]);
         puts("\n\n");
#endif
      }
   }
}

// this test assumes that the underlying grid data is based on a regular
// 2d grid with 4x1 cells and cyclic == {0,0}
// cell indices:
//  0  1  2  3
static void test_grid_cell_neigh_dep_4x1_0x0(
   struct grid * grid, char * grid_name) {

   unsigned cell_neigh_dependencies[4][2] = {{   1}, // neighs of cell 0
                                             {0, 2}, // neighs of cell 1
                                             {1, 3}, // ...
                                             {2   }};
   size_t num_neigh[4] = {1,2,2,1};

   struct dep_list * cell_neigh_dep = yac_get_cell_neigh_dep_list(grid);

#ifdef VERBOSE
   printf("testing grid: %s\n", grid_name);
#else
   UNUSED(grid_name);
#endif

   size_t ref_num_deps = 6;

   if (yac_get_total_num_dependencies(cell_neigh_dep) != ref_num_deps)
      PUT_ERR("error in get_total_num_dependencies\n")

   if (yac_get_num_elements(cell_neigh_dep) != 4)
      PUT_ERR("wrong number of elements in cell_neigh_dep\n")
      
   for (size_t i = 0; i < yac_get_num_elements(cell_neigh_dep); ++i) {

   }

   for (size_t i = 0; i < yac_get_num_elements(cell_neigh_dep); ++i) {

      unsigned const * curr_neighs =
         yac_get_dependencies_of_element(cell_neigh_dep,i);
      size_t curr_num_neighs =
         yac_get_num_dependencies_of_element(cell_neigh_dep, i);

      if (curr_num_neighs != num_neigh[i])
         PUT_ERR("wrong number neighbours in cell_neigh_dep.num_deps_per_element\n")

      size_t num_matching_neighs = 0;
      for (size_t j = 0; j < curr_num_neighs; ++j)
         for (size_t k = 0; k < curr_num_neighs; ++k)
            if (cell_neigh_dependencies[i][j] == curr_neighs[k])
               ++num_matching_neighs;

      if (num_neigh[i] != num_matching_neighs) {

         PUT_ERR("wrong dependency in cell_neigh_dep.dependencies\n")
#ifdef VERBOSE
         printf("cell index: %zu\n", i);
         puts("expected neighs :\n");
         for (size_t j = 0; j < num_neigh[i]; ++j)
            printf("%u ", cell_neigh_dependencies[i][j]);
         puts("\nneighs in cell_neigh_dep:\n");
         for (size_t j = 0; j < curr_num_neighs; ++j)
            printf("%u ", curr_neighs[j]);
         puts("\n\n");
#endif
      }
   }
}

// this test assumes that the underlying grid data is based on a regular
// 2d grid with 1x4 cells and cyclic == {0,0}
// cell indices:
//  3
//  2
//  1
//  0
static void test_grid_cell_neigh_dep_1x4_0x0 (struct grid * grid, char * grid_name) {

   test_grid_cell_neigh_dep_4x1_0x0(grid, grid_name);
}

static int compare_cells(struct grid_cell a, struct grid_cell b) {

  if (a.num_corners != b.num_corners) return 1;

  if (a.num_corners == 0) return 0;

  for (int order = -1; order <= 1; order += 2) {
    for (int start = 0; start < (int)(a.num_corners); ++start) {

      int differences = 0;

      for (int i = 0; i < (int)(a.num_corners); ++i) {

        int j =
          (((int)(a.num_corners)) + start + order * i) % ((int)(a.num_corners));

        if (get_vector_angle(a.coordinates_xyz[i],
                             b.coordinates_xyz[j]) > tol)
          ++differences;

        if (fabs(a.coordinates_y[i] - b.coordinates_y[j]) > tol)
          ++differences;
        else if ((fabs(fabs(a.coordinates_y[i]) - 90.0*YAC_RAD) > tol) &&
                 (get_angle(a.coordinates_x[i], b.coordinates_x[j]) > tol))
          ++differences;
        else if (a.edge_type[i] !=
                 b.edge_type[(((int)(a.num_corners))+j-(order<0))%
                             ((int)(a.num_corners))])
          ++differences;
      }

      if (!differences) return 0;
    }
  }

  return 1;
}

static void compare_grid(struct grid * grid_a, struct grid * grid_b) {

  unsigned num_grid_cells_a = yac_get_num_grid_cells(grid_a);
  unsigned num_grid_cells_b = yac_get_num_grid_cells(grid_b);
  unsigned num_grid_corners_a = yac_get_num_grid_corners(grid_a);
  unsigned num_grid_corners_b = yac_get_num_grid_corners(grid_b);
  unsigned num_grid_edges_a = yac_get_num_grid_edges(grid_a);
  unsigned num_grid_edges_b = yac_get_num_grid_edges(grid_b);

  if (num_grid_cells_a != num_grid_cells_b) {
    PUT_ERR("ERROR(compare_grid): yac_get_num_grid_cells\n");
  } else if (num_grid_corners_a != num_grid_corners_b) {
    PUT_ERR("ERROR(compare_grid): yac_get_num_grid_corners\n");
  } else if (num_grid_edges_a != num_grid_edges_b) {
    PUT_ERR("ERROR(compare_grid): yac_get_num_grid_edges\n");
  } else {
    unsigned num_cells = yac_get_num_grid_cells(grid_a);

    struct grid_cell * grid_cells_a = malloc(num_cells * sizeof(*grid_cells_a));
    struct grid_cell * grid_cells_b = malloc(num_cells * sizeof(*grid_cells_b));

    unsigned * flag_a = calloc(num_cells, sizeof(*flag_a));
    unsigned * flag_b = calloc(num_cells, sizeof(*flag_b));

    for (unsigned i = 0; i < num_cells; ++i) {

      yac_init_grid_cell(grid_cells_a + i);
      yac_get_grid_cell(grid_a, i, grid_cells_a + i);
      yac_init_grid_cell(grid_cells_b + i);
      yac_get_grid_cell(grid_b, i, grid_cells_b + i);
    }

    for (unsigned i = 0; i < num_cells; ++i) {
      for (unsigned j = 0; j < num_cells; ++j) {
        if (!compare_cells(grid_cells_a[i], grid_cells_b[j])) {
          flag_a[i] += 1;
          flag_b[j] += 1;
        }
      }
    }

    for (unsigned i = 0; i < num_cells; ++i)
      if ((flag_a[i] != 1) || (flag_b[i] != 1))
        PUT_ERR("ERROR(compare_grid): cells do not match\n");

    free(flag_a);
    free(flag_b);

    for (unsigned i = 0; i < num_cells; ++i) {

      yac_free_grid_cell(grid_cells_a + i);
      yac_free_grid_cell(grid_cells_b + i);
    }
    free(grid_cells_a);
    free(grid_cells_b);
  }
}

static void test_edge_grid(
  struct grid * edge_grid, struct grid * orig_grid,
  double * edge_point_coord_x, double * edge_point_coord_y) {

  // generate reference grid
  struct grid * ref_edge_grid;
  struct dep_list * edge_grid_cell_to_vertex;
  {
    size_t num_edge_grid_cells = 0;
    unsigned count = yac_get_num_grid_cells(orig_grid);
    count += yac_get_num_grid_corners(orig_grid);
    size_t * num_corners_per_edge_grid_cell =
      malloc(count * sizeof(*num_corners_per_edge_grid_cell));
    unsigned * corners_of_edge_grid_cells = NULL;
    size_t corners_of_edge_grid_cells_array_size = 0;
    size_t total_num_corners_per_edge_grid_cell = 0;

    unsigned num_orig_grid_cells = yac_get_num_grid_cells(orig_grid);

    // for each orig_grid cell
    for (unsigned i = 0; i < num_orig_grid_cells; ++i) {

      // the edges of the current cell build a cell of the edge_grid
      unsigned const * cell_edges =
        yac_get_cell_edge_indices(orig_grid, i);
      size_t num_cell_edges = (size_t)yac_get_num_cell_edges(orig_grid, i);

      ENSURE_ARRAY_SIZE(
        corners_of_edge_grid_cells, corners_of_edge_grid_cells_array_size,
        total_num_corners_per_edge_grid_cell + num_cell_edges);
      num_corners_per_edge_grid_cell[num_edge_grid_cells++] = num_cell_edges;

      for (size_t j = 0; j < num_cell_edges; ++j)
        corners_of_edge_grid_cells[total_num_corners_per_edge_grid_cell++] =
          cell_edges[j];
    }

    unsigned num_orig_grid_corners = yac_get_num_grid_corners(orig_grid);

    // for each orig_grid corner
    for (unsigned i = 0; i < num_orig_grid_corners; ++i) {

      // get number of neighbour corners
      unsigned num_neigh_corners = yac_get_num_corner_edges(orig_grid, i);

      // get number of neighbour cells
      size_t num_neigh_cells = (size_t)yac_get_num_corner_cells(orig_grid, i);

      // if number of neighbour corners equals the number of neighbour cells,
      // the current corner is an inner corner
      // -> edges around this corner build a cell of the edge_grid
      if (num_neigh_corners == num_neigh_cells) {

        // get neighbour cells
        unsigned const * cell_indices =
          yac_get_corner_cell_indices(orig_grid, i);

        unsigned num_cell_edges[num_neigh_cells];
        unsigned max_num_edges_per_cell = 0;
        for (size_t j = 0; j < num_neigh_cells; ++j)
          max_num_edges_per_cell =
            MAX(max_num_edges_per_cell,
                num_cell_edges[j] =
                  yac_get_num_cell_edges(orig_grid, cell_indices[j]));

        unsigned cell_edges[num_neigh_cells][max_num_edges_per_cell];

        for (size_t j = 0; j < num_neigh_cells; ++j) {

          unsigned const * curr_cell_edges =
            yac_get_cell_edge_indices(orig_grid, cell_indices[j]);
          for (unsigned k = 0; k < num_cell_edges[j]; ++k)
            cell_edges[j][k] = curr_cell_edges[k];
        }

        ENSURE_ARRAY_SIZE(
          corners_of_edge_grid_cells, corners_of_edge_grid_cells_array_size,
          total_num_corners_per_edge_grid_cell + num_neigh_cells);
        num_corners_per_edge_grid_cell[num_edge_grid_cells++] =
          num_neigh_cells;

        unsigned curr_edge = -1;
        size_t curr_cell = 0;
        for (size_t j = 0; j < num_neigh_cells; ++j) {

          unsigned flag = 0;
          for (unsigned k = 0; k < num_cell_edges[curr_cell]; ++k) {

            if (cell_edges[curr_cell][k] == curr_edge) continue;

            for (size_t l = 0; l < num_neigh_cells; ++l) {

              if (l == curr_cell) continue;

              for (unsigned m = 0; m < num_cell_edges[l]; ++m) {

                if (cell_edges[curr_cell][k] == cell_edges[l][m]) {

                  curr_edge = cell_edges[curr_cell][k];
                  curr_cell = l;
                  flag = 1;
                  break;
                }
              }
              if (flag) break;
            }
            if (flag) break;
          }

          corners_of_edge_grid_cells[total_num_corners_per_edge_grid_cell++] =
            curr_edge;
        }
      }
    }

    edge_grid_cell_to_vertex =
      yac_dep_list_new(num_edge_grid_cells, num_corners_per_edge_grid_cell,
                       corners_of_edge_grid_cells);

    unsigned num_edge_grid_vertices = yac_get_num_grid_edges(orig_grid);
    ref_edge_grid =
      yac_unstruct_grid_new(
        edge_point_coord_x, edge_point_coord_y, num_edge_grid_vertices,
        edge_grid_cell_to_vertex);

  }

  compare_grid(edge_grid, ref_edge_grid);

  yac_delete_grid(ref_edge_grid);
  yac_dep_list_delete(edge_grid_cell_to_vertex);
}

static void check_cell_triangulation(
  double * coordinates_x, double * coordinates_y, size_t num_corners) {

  double tmp_coordinates_x[2][num_corners];
  double tmp_coordinates_y[2][num_corners];
  double tmp_coordinates_xyz[2][num_corners][3];
  enum yac_edge_type tmp_edge_type[num_corners];
  for (size_t i = 0; i < num_corners; ++i) {
    tmp_coordinates_x[0][i] = YAC_RAD * coordinates_x[i];
    tmp_coordinates_y[0][i] = YAC_RAD * coordinates_y[i];
    tmp_edge_type[i] = GREAT_CIRCLE;
  }

  double cell_area;
  {
    struct grid_cell cell = {.coordinates_x = tmp_coordinates_x[0],
                             .coordinates_y = tmp_coordinates_y[0],
                             .coordinates_xyz = tmp_coordinates_xyz[0],
                             .edge_type = tmp_edge_type,
                             .num_corners = num_corners,
                             .array_size = num_corners};
    generate_XYZ(cell);

    cell_area = yac_huiliers_area(cell);
  }

  struct grid_cell triangles[num_corners-2];
  for (size_t j = 0; j < num_corners-2; ++j)
    yac_init_grid_cell(&(triangles[j]));

  struct grid_cell tmp_cell =
    {.coordinates_x = tmp_coordinates_x[1],
     .coordinates_y = tmp_coordinates_y[1],
     .coordinates_xyz = tmp_coordinates_xyz[1],
     .edge_type = tmp_edge_type,
     .num_corners = num_corners,
     .array_size = num_corners};

  // check both directions
  for (int order = -1; order < 2; order += 2) {

    // for all start corners
    for (int j = 0; j < (int)num_corners; ++j) {

      for (int k = 0; k < (int)num_corners; ++k) {
        int idx = (order*k+j+((int)num_corners))%((int)num_corners);
        tmp_cell.coordinates_x[k] = tmp_coordinates_x[0][idx];
        tmp_cell.coordinates_y[k] = tmp_coordinates_y[0][idx];
        tmp_cell.coordinates_xyz[k][0] = tmp_coordinates_xyz[0][idx][0];
        tmp_cell.coordinates_xyz[k][1] = tmp_coordinates_xyz[0][idx][1];
        tmp_cell.coordinates_xyz[k][2] = tmp_coordinates_xyz[0][idx][2];
      }

      // for all start corners of the triangulation
      for (size_t k = 0; k < num_corners; ++k) {

        // compute triangulation
        yac_triangulate_cell(tmp_cell, k, triangles);

        double total_area = 0.0;

        // for all triangles
        for (size_t l = 0; l < num_corners - 2; ++l) {

          if (triangles[l].num_corners != 3)
            PUT_ERR("wrong number of corners in triangle\n");

          total_area += yac_huiliers_area(triangles[l]);
        }

        if (fabs(total_area - cell_area) > 1e-9)
          PUT_ERR("error in cell triangulation\n");
      }
    }
  }

  for (size_t j = 0; j < num_corners-2; ++j)
    yac_free_grid_cell(&(triangles[j]));
}

int main (void) {

   struct communicator * comm = yac_local_comm_new();

   struct grid * grid_copy;

   struct grid * grid_unpacked[2];

   { // test setting up of 2d regular 2x2 grids

      double coord_x[3] = {0,1,2};
      double coord_y[3] = {0,1,2};

      for ( unsigned i = 0; i < 3 ; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      unsigned num_cells[2] = {2,2};
      unsigned cyclic[2] = {0,0};
      struct grid * reg_grid;

      reg_grid = yac_reg2d_grid_new(coord_x, coord_y, num_cells, cyclic);

      test_grid_2x2(reg_grid, "regular 2d grid", REG2D);

      if (yac_get_size_cell_grid_x_coords(reg_grid) != 2)
         PUT_ERR("error in get_size_cell_grid_x_coords\n")
      if (yac_get_size_cell_grid_y_coords(reg_grid) != 2)
         PUT_ERR("error in get_size_cell_grid_y_coords\n")

      grid_copy = yac_copy_grid(reg_grid);

      test_grid_2x2(grid_copy, "regular 2d grid", REG2D);

      yac_delete_grid(grid_copy);
      yac_delete_grid(reg_grid);
   }

   { // test setting up of unstructured 2x2 grids

      double coord_x[9] = {0,1,2,0,1,2,0,1,2};
      double coord_y[9] = {0,0,0,1,1,1,2,2,2};

      for ( unsigned i = 0; i < 9 ; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      unsigned num_corners = 9;
      size_t num_cells = 4;
      size_t num_corners_per_cell[4] = {4,4,4,4};
      unsigned cell_to_corner_dep[16] = {0,1,4,3, 1,2,5,4, 3,4,7,6, 4,5,8,7};
      size_t * p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
      unsigned * p_cell_to_corner_dep = TO_POINTER(cell_to_corner_dep);
      struct grid * unstruct_grid;

      struct dep_list * cell_to_vertex =
        yac_dep_list_new(
          num_cells, p_num_corners_per_cell, p_cell_to_corner_dep);

      unstruct_grid = yac_unstruct_grid_new(coord_x, coord_y, num_corners, cell_to_vertex);

      test_grid_2x2(unstruct_grid, "unstructured grid", UNSTRUCT);

      if (yac_get_size_cell_grid_x_coords(unstruct_grid) != 4)
         PUT_ERR("error in get_size_cell_grid_x_coords\n")
      if (yac_get_size_cell_grid_y_coords(unstruct_grid) != 4)
         PUT_ERR("error in get_size_cell_grid_y_coords\n")

      grid_copy = yac_copy_grid(unstruct_grid);

      test_grid_2x2(grid_copy, "unstructured grid", UNSTRUCT);

      yac_delete_grid(unstruct_grid);
      yac_delete_grid(grid_copy);
      yac_dep_list_delete(cell_to_vertex);
   }

   { // test setting up of curvilinear 2x2 grids

      double coord_x[9] = {0,1,2,0,1,2,0,1,2};
      double coord_y[9] = {0,0,0,1,1,1,2,2,2};

      for ( unsigned i = 0; i < 9; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      unsigned num_cells[2] = {2,2};
      unsigned cyclic[2] = {0,0};
      struct grid * curvi_grid;

      curvi_grid = yac_curve2d_grid_new(coord_x, coord_y, num_cells, cyclic);

      test_grid_2x2(curvi_grid, "curvilinear grid", CURVE2D);

      if (yac_get_size_cell_grid_x_coords(curvi_grid) != 4)
         PUT_ERR("error in get_size_cell_grid_x_coords\n")
      if (yac_get_size_cell_grid_y_coords(curvi_grid) != 4)
         PUT_ERR("error in get_size_cell_grid_y_coords\n")

      grid_copy = yac_copy_grid(curvi_grid);

      test_grid_2x2(grid_copy, "curvilinear grid", CURVE2D);

      yac_delete_grid(curvi_grid);
      yac_delete_grid(grid_copy);
   }

   { // test setting up of 2d regular 2x2 grids

      double coord_x[3] = {0,1*YAC_RAD,2*YAC_RAD};
      double coord_y[4] = {0,1*YAC_RAD,2*YAC_RAD,3*YAC_RAD};
      unsigned num_cells[2] = {2,3};
      unsigned cyclic[2] = {0,0};
      struct grid * reg_grid;

      reg_grid = yac_reg2d_grid_new(coord_x, coord_y, num_cells, cyclic);

      test_grid_2x3(reg_grid, "regular 2d grid", REG2D);

      if (yac_get_size_cell_grid_x_coords(reg_grid) != 2)
         PUT_ERR("error in get_size_cell_grid_x_coords\n")
      if (yac_get_size_cell_grid_y_coords(reg_grid) != 3)
         PUT_ERR("error in get_size_cell_grid_y_coords\n")

      grid_copy = yac_copy_grid(reg_grid);

      test_grid_2x3(grid_copy, "regular 2d grid", REG2D);

      yac_delete_grid(reg_grid);
      yac_delete_grid(grid_copy);
   }

   { // test setting up of unstructured 2x2 grids

      double coord_x[12] = {0,1,2,0,1,2,0,1,2,0,1,2};
      double coord_y[12] = {0,0,0,1,1,1,2,2,2,3,3,3};

      for ( unsigned i = 0; i < 12; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      unsigned num_corners = 12;
      size_t num_cells = 6;
      size_t num_corners_per_cell[6] = {4,4,4,4,4,4};
      unsigned cell_to_corner_dep[24] = {0,1,4,3, 1,2,5,4, 3,4,7,6, 4,5,8,7, 6,7,10,9, 7,8,11,10};
      size_t * p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
      unsigned * p_cell_to_corner_dep = TO_POINTER(cell_to_corner_dep);
      struct grid * unstruct_grid;

      struct dep_list * cell_to_vertex =
        yac_dep_list_new(
          num_cells, p_num_corners_per_cell, p_cell_to_corner_dep);

      unstruct_grid = yac_unstruct_grid_new(coord_x, coord_y, num_corners, cell_to_vertex);

      test_grid_2x3(unstruct_grid, "unstructured grid", UNSTRUCT);

      if (yac_get_size_cell_grid_x_coords(unstruct_grid) != 6)
         PUT_ERR("error in get_size_cell_grid_x_coords\n")
      if (yac_get_size_cell_grid_y_coords(unstruct_grid) != 6)
         PUT_ERR("error in get_size_cell_grid_y_coords\n")

      grid_copy = yac_copy_grid(unstruct_grid);

      test_grid_2x3(grid_copy, "unstructured grid", UNSTRUCT);

      yac_delete_grid(unstruct_grid);
      yac_delete_grid(grid_copy);
      yac_dep_list_delete(cell_to_vertex);
   }

   { // test setting up of curvilinear 2x2 grids

      double coord_x[12] = {0,1,2,0,1,2,0,1,2,0,1,2};
      double coord_y[12] = {0,0,0,1,1,1,2,2,2,3,3,3};

      for ( unsigned i = 0; i < 12; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      unsigned num_cells[2] = {2,3};
      unsigned cyclic[2] = {0,0};
      struct grid * curvi_grid;

      curvi_grid = yac_curve2d_grid_new(coord_x, coord_y, num_cells, cyclic);

      test_grid_2x3(curvi_grid, "curvilinear grid", CURVE2D);

      if (yac_get_size_cell_grid_x_coords(curvi_grid) != 6)
         PUT_ERR("error in get_size_cell_grid_x_coords\n")
      if (yac_get_size_cell_grid_y_coords(curvi_grid) != 6)
         PUT_ERR("error in get_size_cell_grid_y_coords\n")

      grid_copy = yac_copy_grid(curvi_grid);

      test_grid_2x3(grid_copy, "curvilinear grid", CURVE2D);

      yac_delete_grid(curvi_grid);
      yac_delete_grid(grid_copy);
   }

   { // test neighbour dependencies of 2d regular grids

      double coord_x[] = {0,1,2,3};
      double coord_y[] = {0,1,2,3};

      for ( unsigned i = 0; i < 4; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      unsigned num_cells[] = {3,3};
      unsigned cyclic[] = {0,0};
      struct grid * reg_grid;

      reg_grid = yac_reg2d_grid_new(coord_x, coord_y, num_cells, cyclic);

      test_grid_cell_neigh_dep_3x3_0x0(reg_grid, "regular 2d grid");

      grid_copy = yac_copy_grid(reg_grid);

      test_grid_cell_neigh_dep_3x3_0x0(grid_copy, "regular 2d grid");

      yac_delete_grid(reg_grid);
      yac_delete_grid(grid_copy);
   }

   { // test neighbour dependencies of unstructured grids

      double coord_x[16] = {0,1,2,3,0,1,2,3,0,1,2,3,0,1,2,3};
      double coord_y[16] = {0,0,0,0,1,1,1,1,2,2,2,2,3,3,3,3};

      for ( unsigned i = 0; i < 16; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      unsigned num_corners = 16;
      size_t num_cells = 9;
      size_t num_corners_per_cell[9] = {4,4,4,4,4,4,4,4,4};
      unsigned cell_to_corner_dep[36] = {0,1,5,4, 1,2,6,5, 2,3,7,6,
                                         4,5,9,8, 5,6,10,9, 6,7,11,10,
                                         8,9,13,12, 9,10,14,13, 10,11,15,14};
      size_t * p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
      unsigned * p_cell_to_corner_dep = TO_POINTER(cell_to_corner_dep);
      struct grid * unstruct_grid;

      struct dep_list * cell_to_vertex =
        yac_dep_list_new(
          num_cells, p_num_corners_per_cell, p_cell_to_corner_dep);

      unstruct_grid = yac_unstruct_grid_new(coord_x, coord_y, num_corners, cell_to_vertex);

      test_grid_cell_neigh_dep_3x3_0x0(unstruct_grid, "unstructured grid");

      grid_copy = yac_copy_grid(unstruct_grid);

      test_grid_cell_neigh_dep_3x3_0x0(grid_copy, "unstructured grid");

      yac_delete_grid(unstruct_grid);
      yac_delete_grid(grid_copy);
      yac_dep_list_delete(cell_to_vertex);
   }

   { // test neighbour dependencies of curvilinear grids

      double coord_x[16] = {0,1,2,3,0,1,2,3,0,1,2,3,0,1,2,3};
      double coord_y[16] = {0,0,0,0,1,1,1,1,2,2,2,2,3,3,3,3};

      for ( unsigned i = 0; i < 16; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      unsigned num_cells[] = {3,3};
      unsigned cyclic[] = {0,0};
      struct grid * curvi_grid;

      curvi_grid = yac_curve2d_grid_new(coord_x, coord_y, num_cells, cyclic);

      test_grid_cell_neigh_dep_3x3_0x0(curvi_grid, "curvilinear grid");

      grid_copy = yac_copy_grid(curvi_grid);

      test_grid_cell_neigh_dep_3x3_0x0(grid_copy, "curvilinear grid");

      yac_delete_grid(curvi_grid);
      yac_delete_grid(grid_copy);
   }

   { // test neighbour dependencies of 2d regular grids

      double coord_x[] = {0,1*YAC_RAD,2*YAC_RAD,3*YAC_RAD};
      double coord_y[] = {0,1*YAC_RAD,2*YAC_RAD};
      unsigned num_cells[] = {3,2};
      unsigned cyclic[] = {0,0};
      struct grid * reg_grid;

      reg_grid = yac_reg2d_grid_new(coord_x, coord_y, num_cells, cyclic);

      test_grid_cell_neigh_dep_3x2_0x0(reg_grid, "regular 2d grid");

      grid_copy = yac_copy_grid(reg_grid);

      test_grid_cell_neigh_dep_3x2_0x0(grid_copy, "regular 2d grid");

      yac_delete_grid(reg_grid);
      yac_delete_grid(grid_copy);
   }

   { // test neighbour dependencies of unstructured grids

      double coord_x[12] = {0,1,2,3,0,1,2,3,0,1,2,3};
      double coord_y[12] = {0,0,0,0,1,1,1,1,2,2,2,2};

      for ( unsigned i = 0; i < 12; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      unsigned num_corners = 12;
      size_t num_cells = 6;
      size_t num_corners_per_cell[6] = {4,4,4,4,4,4};
      unsigned cell_to_corner_dep[36] = {0,1,5,4, 1,2,6,5, 2,3,7,6,
                                         4,5,9,8, 5,6,10,9, 6,7,11,10};
      size_t * p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
      unsigned * p_cell_to_corner_dep = TO_POINTER(cell_to_corner_dep);
      struct grid * unstruct_grid;

      struct dep_list * cell_to_vertex =
        yac_dep_list_new(
          num_cells, p_num_corners_per_cell, p_cell_to_corner_dep);

      unstruct_grid = yac_unstruct_grid_new(coord_x, coord_y, num_corners, cell_to_vertex);

      test_grid_cell_neigh_dep_3x2_0x0(unstruct_grid, "unstructured grid");

      grid_copy = yac_copy_grid(unstruct_grid);

      test_grid_cell_neigh_dep_3x2_0x0(grid_copy, "unstructured grid");

      yac_delete_grid(unstruct_grid);
      yac_delete_grid(grid_copy);
      yac_dep_list_delete(cell_to_vertex);
   }

   { // test neighbour dependencies of curvilinear grids

      double coord_x[12] = {0,1,2,3,0,1,2,3,0,1,2,3};
      double coord_y[12] = {0,0,0,0,1,1,1,1,2,2,2,2};

      for ( unsigned i = 0; i < 12; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      unsigned num_cells[] = {3,2};
      unsigned cyclic[] = {0,0};
      struct grid * curvi_grid;

      curvi_grid = yac_curve2d_grid_new(coord_x, coord_y, num_cells, cyclic);

      test_grid_cell_neigh_dep_3x2_0x0(curvi_grid, "curvilinear grid");

      grid_copy = yac_copy_grid(curvi_grid);

      test_grid_cell_neigh_dep_3x2_0x0(grid_copy, "curvilinear grid");

      yac_delete_grid(curvi_grid);
      yac_delete_grid(grid_copy);
   }

   { // test neighbour dependencies of 2d regular grids

      double coord_x[] = {0,1*YAC_RAD,2*YAC_RAD,3*YAC_RAD,4*YAC_RAD};
      double coord_y[] = {0,1*YAC_RAD};
      unsigned num_cells[] = {4,1};
      unsigned cyclic[] = {0,0};
      struct grid * reg_grid;

      reg_grid = yac_reg2d_grid_new(coord_x, coord_y, num_cells, cyclic);

      test_grid_cell_neigh_dep_4x1_0x0(reg_grid, "regular 2d grid");

      grid_copy = yac_copy_grid(reg_grid);

      test_grid_cell_neigh_dep_4x1_0x0(grid_copy, "regular 2d grid");

      yac_delete_grid(reg_grid);
      yac_delete_grid(grid_copy);
   }

   { // test neighbour dependencies of unstructured grids

      double coord_x[12] = {0,1,2,3,4,0,1,2,3,4};
      double coord_y[12] = {0,0,0,0,0,1,1,1,1,1};

      for ( unsigned i = 0; i < 12; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      unsigned num_corners = 10;
      size_t num_cells = 4;
      size_t num_corners_per_cell[4] = {4,4,4,4};
      unsigned cell_to_corner_dep[4*4] = {0,1,6,5, 1,2,7,6, 2,3,8,7, 3,4,9,8};
      size_t * p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
      unsigned * p_cell_to_corner_dep = TO_POINTER(cell_to_corner_dep);
      struct grid * unstruct_grid;

      struct dep_list * cell_to_vertex =
        yac_dep_list_new(
          num_cells, p_num_corners_per_cell, p_cell_to_corner_dep);

      unstruct_grid = yac_unstruct_grid_new(coord_x, coord_y, num_corners, cell_to_vertex);

      test_grid_cell_neigh_dep_4x1_0x0(unstruct_grid, "unstructured grid");

      grid_copy = yac_copy_grid(unstruct_grid);

      test_grid_cell_neigh_dep_4x1_0x0(grid_copy, "unstructured grid");

      yac_delete_grid(unstruct_grid);
      yac_delete_grid(grid_copy);
      yac_dep_list_delete(cell_to_vertex);
   }

   { // test neighbour dependencies of curvilinear grids

      double coord_x[12] = {0,1,2,3,4,0,1,2,3,4};
      double coord_y[12] = {0,0,0,0,0,1,1,1,1,1};

      for ( unsigned i = 0; i < 12; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      unsigned num_cells[] = {4,1};
      unsigned cyclic[] = {0,0};
      struct grid * curvi_grid;

      curvi_grid = yac_curve2d_grid_new(coord_x, coord_y, num_cells, cyclic);

      test_grid_cell_neigh_dep_4x1_0x0(curvi_grid, "curvilinear grid");

      grid_copy = yac_copy_grid(curvi_grid);

      test_grid_cell_neigh_dep_4x1_0x0(grid_copy, "curvilinear grid");

      yac_delete_grid(curvi_grid);
      yac_delete_grid(grid_copy);
   }

   { // test neighbour dependencies of 2d regular grids

      double coord_x[] = {0,1*YAC_RAD};
      double coord_y[] = {0,1*YAC_RAD,2*YAC_RAD,3*YAC_RAD,4*YAC_RAD};
      unsigned num_cells[] = {1,4};
      unsigned cyclic[] = {0,0};
      struct grid * reg_grid;

      reg_grid = yac_reg2d_grid_new(coord_x, coord_y, num_cells, cyclic);

      test_grid_cell_neigh_dep_1x4_0x0(reg_grid, "regular 2d grid");

      grid_copy = yac_copy_grid(reg_grid);

      test_grid_cell_neigh_dep_1x4_0x0(grid_copy, "regular 2d grid");

      yac_delete_grid(reg_grid);
      yac_delete_grid(grid_copy);
   }

   { // test neighbour dependencies of unstructured grids

      double coord_x[12] = {0,1,0,1,0,1,0,1,0,1};
      double coord_y[12] = {0,0,1,1,2,2,3,3,4,4};

      for ( unsigned i = 0; i < 12; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      unsigned num_corners = 10;
      size_t num_cells = 4;
      size_t num_corners_per_cell[4] = {4,4,4,4};
      unsigned cell_to_corner_dep[4*4] = {0,1,3,2, 2,3,5,4, 4,5,7,6, 6,7,9,8};
      size_t * p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
      unsigned * p_cell_to_corner_dep = TO_POINTER(cell_to_corner_dep);
      struct grid * unstruct_grid;

      struct dep_list * cell_to_vertex =
        yac_dep_list_new(
          num_cells, p_num_corners_per_cell, p_cell_to_corner_dep);

      unstruct_grid = yac_unstruct_grid_new(coord_x, coord_y, num_corners, cell_to_vertex);

      test_grid_cell_neigh_dep_1x4_0x0(unstruct_grid, "unstructured grid");

      grid_copy = yac_copy_grid(unstruct_grid);

      test_grid_cell_neigh_dep_1x4_0x0(grid_copy, "unstructured grid");

      yac_delete_grid(unstruct_grid);
      yac_delete_grid(grid_copy);
      yac_dep_list_delete(cell_to_vertex);
   }

   { // test neighbour dependencies of curvilinear grids

      double coord_x[12] = {0,1,0,1,0,1,0,1,0,1};
      double coord_y[12] = {0,0,1,1,2,2,3,3,4,4};

      for ( unsigned i = 0; i < 12; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      unsigned num_cells[] = {1,4};
      unsigned cyclic[] = {0,0};
      struct grid * curvi_grid;

      curvi_grid = yac_curve2d_grid_new(coord_x, coord_y, num_cells, cyclic);

      test_grid_cell_neigh_dep_1x4_0x0(curvi_grid, "curvilinear grid");

      grid_copy = yac_copy_grid(curvi_grid);

      test_grid_cell_neigh_dep_1x4_0x0(grid_copy, "curvilinear grid");

      yac_delete_grid(curvi_grid);
      yac_delete_grid(grid_copy);
   }

   { // test cell grid generation for 2d regular grids

      double coord_x[] = {0,1,2,3};
      double coord_y[] = {0,1,2,3};

      for ( unsigned i = 0; i < 4; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      double cell_point_coord_x[] = {0,1,2};
      double cell_point_coord_y[] = {0,1,2};

      for ( unsigned i = 0; i < 3; ++i ) {
         cell_point_coord_x[i] *= YAC_RAD;
         cell_point_coord_y[i] *= YAC_RAD;
      }

      double edge_point_coord_x[] = {0.5,0,1.5,1,2.5,2,3,
                                     0.5,0,1.5,1,2.5,2,3,
                                     0.5,0,1.5,1,2.5,2,3,
                                     0.5,1.5,2.5};
      double edge_point_coord_y[] = {0,0.5,0,0.5,0,0.5,0.5,
                                     1,1.5,1,1.5,1,1.5,1.5,
                                     2,2.5,2,2.5,2,2.5,2.5,
                                     3,3,3};
      for ( unsigned i = 0;
            i < sizeof(edge_point_coord_x)/sizeof(edge_point_coord_x[0]);
            ++i ) {
         edge_point_coord_x[i] *= YAC_RAD;
         edge_point_coord_y[i] *= YAC_RAD;
      }

      unsigned num_cells[] = {3,3};
      unsigned cyclic[] = {0,0};
      struct grid * reg_grid, * cell_grid, * edge_grid;

      reg_grid = yac_reg2d_grid_new(coord_x, coord_y, num_cells, cyclic);
      cell_grid = yac_generate_cell_grid(reg_grid, cell_point_coord_x, cell_point_coord_y);
      edge_grid = yac_generate_edge_grid(reg_grid, edge_point_coord_x, edge_point_coord_y);

      test_grid_2x2(cell_grid, "regular 2d grid", REG2D);
      test_edge_grid(edge_grid, reg_grid, edge_point_coord_x, edge_point_coord_y);

      yac_delete_grid(reg_grid);
      yac_delete_grid(cell_grid);
      yac_delete_grid(edge_grid);
   }

   { // test cell grid generation for unstructured grids

      double coord_x[16] = {0,1,2,3,0,1,2,3,0,1,2,3,0,1,2,3};
      double coord_y[16] = {0,0,0,0,1,1,1,1,2,2,2,2,3,3,3,3};

      for ( unsigned i = 0; i < 16; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      double cell_point_coord_x[] = {0,1,2,0,1,2,0,1,2};
      double cell_point_coord_y[] = {0,0,0,1,1,1,2,2,2};

      for ( unsigned i = 0; i < 9; ++i ) {
         cell_point_coord_x[i] *= YAC_RAD;
         cell_point_coord_y[i] *= YAC_RAD;
      }

      double edge_point_coord_x[] = {0.5,0,1.5,1,2.5,2,3,
                                     0.5,0,1.5,1,2.5,2,3,
                                     0.5,0,1.5,1,2.5,2,3,
                                     0.5,1.5,2.5};
      double edge_point_coord_y[] = {0,0.5,0,0.5,0,0.5,0.5,
                                     1,1.5,1,1.5,1,1.5,1.5,
                                     2,2.5,2,2.5,2,2.5,2.5,
                                     3,3,3};

      for ( unsigned i = 0;
            i < sizeof(edge_point_coord_x)/sizeof(edge_point_coord_x[0]);
            ++i ) {
         edge_point_coord_x[i] *= YAC_RAD;
         edge_point_coord_y[i] *= YAC_RAD;
      }

      unsigned num_corners = 16;
      size_t num_cells = 9;
      size_t num_corners_per_cell[9] = {4,4,4,4,4,4,4,4,4};
      unsigned cell_to_corner_dep[36] = {0,1,5,4, 1,2,6,5, 2,3,7,6,
                                         4,5,9,8, 5,6,10,9, 6,7,11,10,
                                         8,9,13,12, 9,10,14,13, 10,11,15,14};
      size_t * p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
      unsigned * p_cell_to_corner_dep = TO_POINTER(cell_to_corner_dep);
      struct grid * unstruct_grid, * cell_grid, * edge_grid;

      struct dep_list * cell_to_vertex =
        yac_dep_list_new(
          num_cells, p_num_corners_per_cell, p_cell_to_corner_dep);

      unstruct_grid = yac_unstruct_grid_new(coord_x, coord_y, num_corners, cell_to_vertex);
      cell_grid = yac_generate_cell_grid(unstruct_grid, cell_point_coord_x, cell_point_coord_y);
      edge_grid = yac_generate_edge_grid(unstruct_grid, edge_point_coord_x, edge_point_coord_y);

      test_grid_2x2(cell_grid, "unstructured grid", UNSTRUCT);
      test_edge_grid(edge_grid, unstruct_grid, edge_point_coord_x, edge_point_coord_y);

      yac_delete_grid(unstruct_grid);
      yac_delete_grid(cell_grid);
      yac_delete_grid(edge_grid);
      yac_dep_list_delete(cell_to_vertex);
   }

   { // test cell grid generation for unstruct_grid grids
     // 6--10--7--11--8
     // |      |      |
     // 9   2  8   1  6
     // |      |      |
     // 5-- 7--4-- 5--3
     // |      |      |
     // 1   0  3   3  4
     // |      |      |
     // 0-- 0--1-- 2--2

      double coord_x[] = {0,1,2,2,1,0,0,1,2};
      double coord_y[] = {0,0,0,1,1,1,2,2,2};

      for ( unsigned i = 0;
            i < sizeof(coord_x)/sizeof(coord_x[0]);
            ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      double cell_point_coord_x[] = {0.5,1.5,0.5,1.5};
      double cell_point_coord_y[] = {0.5,1.5,1.5,0.5};

      for ( unsigned i = 0;
            i < sizeof(cell_point_coord_x)/sizeof(cell_point_coord_x[0]);
            ++i ) {
         cell_point_coord_x[i] *= YAC_RAD;
         cell_point_coord_y[i] *= YAC_RAD;
      }

      double edge_point_coord_x[] =
        {0.5,0,1.5,1,2, 1.5,2,0.5,1,0, 0.5,1.5};
      double edge_point_coord_y[] =
        {0,0.5,0,0.5,0.5, 1,1.5,1,1.5,1.5, 3,3};

      for ( unsigned i = 0;
            i < sizeof(edge_point_coord_x)/sizeof(edge_point_coord_x[0]);
            ++i ) {
         edge_point_coord_x[i] *= YAC_RAD;
         edge_point_coord_y[i] *= YAC_RAD;
      }

      unsigned num_corners = 9;
      size_t num_cells = 4;
      size_t num_corners_per_cell[4] = {4,4,4,4};
      unsigned cell_to_corner_dep[16] = {0,1,4,5, 4,3,8,7, 4,5,6,7, 1,4,3,2};
      size_t * p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
      unsigned * p_cell_to_corner_dep = TO_POINTER(cell_to_corner_dep);
      struct grid * unstruct_grid, * cell_grid, * edge_grid;

      struct dep_list * cell_to_vertex =
        yac_dep_list_new(
          num_cells, p_num_corners_per_cell, p_cell_to_corner_dep);

      unstruct_grid = yac_unstruct_grid_new(coord_x, coord_y, num_corners, cell_to_vertex);
      cell_grid = yac_generate_cell_grid(unstruct_grid, cell_point_coord_x, cell_point_coord_y);
      edge_grid = yac_generate_edge_grid(unstruct_grid, edge_point_coord_x, edge_point_coord_y);

      test_grid_1x1(cell_grid, "unstructured grid");
      test_edge_grid(edge_grid, unstruct_grid, edge_point_coord_x, edge_point_coord_y);

      yac_delete_grid(unstruct_grid);
      yac_delete_grid(cell_grid);
      yac_delete_grid(edge_grid);
      yac_dep_list_delete(cell_to_vertex);
   }

   { // test cell grid generation for curvilinear grids

      double coord_x[16] = {0,1,2,3,0,1,2,3,0,1,2,3,0,1,2,3};
      double coord_y[16] = {0,0,0,0,1,1,1,1,2,2,2,2,3,3,3,3};

      for ( unsigned i = 0; i < 16; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      double cell_point_coord_x[] = {0,1,2,0,1,2,0,1,2};
      double cell_point_coord_y[] = {0,0,0,1,1,1,2,2,2};

      for ( unsigned i = 0; i < 9; ++i ) {
         cell_point_coord_x[i] *= YAC_RAD;
         cell_point_coord_y[i] *= YAC_RAD;
      }

      double edge_point_coord_x[] = {0.5,0,1.5,1,2.5,2,3,
                                     0.5,0,1.5,1,2.5,2,3,
                                     0.5,0,1.5,1,2.5,2,3,
                                     0.5,1.5,2.5};
      double edge_point_coord_y[] = {0,0.5,0,0.5,0,0.5,0.5,
                                     1,1.5,1,1.5,1,1.5,1.5,
                                     2,2.5,2,2.5,2,2.5,2.5,
                                     3,3,3};

      for ( unsigned i = 0;
            i < sizeof(edge_point_coord_x)/sizeof(edge_point_coord_x[0]);
            ++i ) {
         edge_point_coord_x[i] *= YAC_RAD;
         edge_point_coord_y[i] *= YAC_RAD;
      }

      unsigned num_cells[] = {3,3};
      unsigned cyclic[] = {0,0};
      struct grid * curvi_grid, * cell_grid, * edge_grid;

      curvi_grid = yac_curve2d_grid_new(coord_x, coord_y, num_cells, cyclic);
      cell_grid = yac_generate_cell_grid(curvi_grid, cell_point_coord_x, cell_point_coord_y);
      edge_grid = yac_generate_edge_grid(curvi_grid, edge_point_coord_x, edge_point_coord_y);

      test_grid_2x2(cell_grid, "curvilinear grid", CURVE2D);
      test_edge_grid(edge_grid, curvi_grid, edge_point_coord_x, edge_point_coord_y);

      yac_delete_grid(curvi_grid);
      yac_delete_grid(cell_grid);
      yac_delete_grid(edge_grid);
   }

   { // test a cyclic behaviour for 2d regular grids and packing feature

      double coord_x[] = {0,120*YAC_RAD,240*YAC_RAD};
      double coord_y[] = {-90*YAC_RAD,-30*YAC_RAD,30*YAC_RAD,90*YAC_RAD};
      double cell_point_coord_x[] = {60*YAC_RAD,180*YAC_RAD,300*YAC_RAD};
      double cell_point_coord_y[] = {-60*YAC_RAD,0,60*YAC_RAD};
      double edge_point_coord_x[] = { 60*YAC_RAD,300*YAC_RAD,  0*YAC_RAD,
                                     180*YAC_RAD,120*YAC_RAD,240*YAC_RAD,
                                      60*YAC_RAD,300*YAC_RAD,  0*YAC_RAD,
                                     180*YAC_RAD,120*YAC_RAD,240*YAC_RAD,
                                      60*YAC_RAD,300*YAC_RAD,  0*YAC_RAD,
                                     180*YAC_RAD,120*YAC_RAD,240*YAC_RAD,
                                      60*YAC_RAD,300*YAC_RAD,180*YAC_RAD};
      double edge_point_coord_y[] = {  0*YAC_RAD,  0*YAC_RAD,-60*YAC_RAD,
                                       0*YAC_RAD,-60*YAC_RAD,-60*YAC_RAD,
                                     -30*YAC_RAD,-30*YAC_RAD,  0*YAC_RAD,
                                     -30*YAC_RAD,  0*YAC_RAD,  0*YAC_RAD,
                                      30*YAC_RAD, 30*YAC_RAD, 60*YAC_RAD,
                                      30*YAC_RAD, 60*YAC_RAD, 60*YAC_RAD,
                                      90*YAC_RAD, 90*YAC_RAD, 90*YAC_RAD};
      unsigned num_cells[] = {3,3};
      unsigned cyclic[] = {1,0};
      struct grid * reg_grid, * cell_grid, * edge_grid;

      reg_grid = yac_reg2d_grid_new(coord_x, coord_y, num_cells, cyclic);

      test_grid_3x3_1x0(reg_grid, "regular 2d grid", REG2D);
      test_grid_cell_neigh_dep_3x3_1x0(reg_grid, "regular 2d grid");

      cell_grid = yac_generate_cell_grid(reg_grid, cell_point_coord_x, cell_point_coord_y);
      edge_grid = yac_generate_edge_grid(reg_grid, edge_point_coord_x, edge_point_coord_y);

      test_grid_3x2_1x0(cell_grid, "regular 2d grid", REG2D);
      test_grid_cell_neigh_dep_3x2_1x0(cell_grid, "regular 2d grid");

      int pack_buffer_size = yac_grid_pack_size(reg_grid, comm);
      pack_buffer_size    += yac_grid_pack_size(cell_grid, comm);
      void * pack_buffer = malloc(pack_buffer_size);
      int position = 0;

      yac_pack_grid(reg_grid, pack_buffer, pack_buffer_size, &position, comm);
      yac_pack_grid(cell_grid, pack_buffer, pack_buffer_size, &position, comm);

      position = 0;

      grid_unpacked[0] = yac_unpack_grid(pack_buffer, pack_buffer_size, &position, comm);
      grid_unpacked[1] = yac_unpack_grid(pack_buffer, pack_buffer_size, &position, comm);

      test_grid_3x3_1x0(grid_unpacked[0], "regular 2d grid", REG2D);
      test_grid_3x2_1x0(grid_unpacked[1], "regular 2d grid", REG2D);
      test_edge_grid(edge_grid, reg_grid, edge_point_coord_x, edge_point_coord_y);

      yac_delete_grid(grid_unpacked[0]);
      yac_delete_grid(grid_unpacked[1]);
      free(pack_buffer);
      yac_delete_grid(edge_grid);
      yac_delete_grid(cell_grid);
      yac_delete_grid(reg_grid);
   }

   { // test a cyclic behaviour for unstructured grids and packing feature

      double coord_x[12] = {0,120,240,0,120,240,0,120,240,0,120,240};
      double coord_y[12] = {-90,-90,-90,-30,-30,-30,30,30,30,90,90,90};

      for ( unsigned i = 0; i < 12; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      double cell_point_coord_x[] = {60,180,300,60,180,300,60,180,300};
      double cell_point_coord_y[] = {-60,-60,-60,0,0,0,60,60,60};

      for ( unsigned i = 0; i < 9; ++i ) {
         cell_point_coord_x[i] *= YAC_RAD;
         cell_point_coord_y[i] *= YAC_RAD;
      }

      double edge_point_coord_x[] = {60,300,0,180,120,240,
                                     60,300,0,180,120,240,
                                     60,300,0,180,120,240,
                                     60,300,180};
      double edge_point_coord_y[] = {  0,  0,-60,  0,-60,-60,
                                     -30,-30,  0,-30,  0,  0,
                                      30, 30, 60, 30, 60, 60,
                                      90, 90, 90};

      for ( unsigned i = 0; i < 21; ++i ) {
         edge_point_coord_x[i] *= YAC_RAD;
         edge_point_coord_y[i] *= YAC_RAD;
      }

      unsigned num_corners = 12;
      size_t num_cells = 9;
      size_t num_corners_per_cell[9] = {4,4,4,4,4,4,4,4,4};
      unsigned cell_to_corner_dep[36] = {0,1,4,3, 1,2,5,4, 2,0,3,5,
                                         3,4,7,6, 4,5,8,7, 5,3,6,8,
                                         6,7,10,9, 7,8,11,10, 8,6,9,11};
      size_t * p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
      unsigned * p_cell_to_corner_dep = TO_POINTER(cell_to_corner_dep);
      struct grid * unstruct_grid, * cell_grid, * edge_grid;

      struct dep_list * cell_to_vertex =
        yac_dep_list_new(
          num_cells, p_num_corners_per_cell, p_cell_to_corner_dep);

      unstruct_grid = yac_unstruct_grid_new(coord_x, coord_y, num_corners, cell_to_vertex);

      test_grid_3x3_1x0(unstruct_grid, "unstructured grid", UNSTRUCT);
      test_grid_cell_neigh_dep_3x3_1x0(unstruct_grid, "unstructured grid");

      cell_grid = yac_generate_cell_grid(unstruct_grid, cell_point_coord_x, cell_point_coord_y);
      edge_grid = yac_generate_edge_grid(unstruct_grid, edge_point_coord_x, edge_point_coord_y);

      test_grid_3x2_1x0(cell_grid, "unstructured grid", UNSTRUCT);
      test_grid_cell_neigh_dep_3x2_1x0(cell_grid, "unstructured grid");

      int pack_buffer_size = yac_grid_pack_size(unstruct_grid, comm);
      pack_buffer_size    += yac_grid_pack_size(cell_grid, comm);
      void * pack_buffer = malloc(pack_buffer_size);
      int position = 0;

      yac_pack_grid(unstruct_grid, pack_buffer, pack_buffer_size, &position, comm);
      yac_pack_grid(cell_grid, pack_buffer, pack_buffer_size, &position, comm);

      position = 0;

      grid_unpacked[0] = yac_unpack_grid(pack_buffer, pack_buffer_size, &position, comm);
      grid_unpacked[1] = yac_unpack_grid(pack_buffer, pack_buffer_size, &position, comm);

      test_grid_3x3_1x0(grid_unpacked[0], "unstructured grid", UNSTRUCT);
      test_grid_3x2_1x0(grid_unpacked[1], "unstructured grid", UNSTRUCT);
      test_edge_grid(edge_grid, unstruct_grid, edge_point_coord_x, edge_point_coord_y);

      yac_delete_grid(grid_unpacked[0]);
      yac_delete_grid(grid_unpacked[1]);
      free(pack_buffer);
      yac_delete_grid(unstruct_grid);
      yac_delete_grid(cell_grid);
      yac_dep_list_delete(cell_to_vertex);
      yac_delete_grid(edge_grid);
   }

   { // test a cyclic behaviour for curvilinear grids and packing feature

      double coord_x[16] = {0,120,240,0,120,240,0,120,240,0,120,240};
      double coord_y[16] = {-90,-90,-90,-30,-30,-30,30,30,30,90,90,90};

      for ( unsigned i = 0; i < 16; ++i ) {
         coord_x[i] *= YAC_RAD;
         coord_y[i] *= YAC_RAD;
      }

      double cell_point_coord_x[] = {60,180,300,60,180,300,60,180,300};
      double cell_point_coord_y[] = {-60,-60,-60,0,0,0,60,60,60};

      for ( unsigned i = 0; i < 9; ++i ) {
         cell_point_coord_x[i] *= YAC_RAD;
         cell_point_coord_y[i] *= YAC_RAD;
      }

      double edge_point_coord_x[] = {60,300,0,180,120,240,
                                     60,300,0,180,120,240,
                                     60,300,0,180,120,240,
                                     60,300,180};
      double edge_point_coord_y[] = {  0,  0,-60,  0,-60,-60,
                                     -30,-30,  0,-30,  0,  0,
                                      30, 30, 60, 30, 60, 60,
                                      90, 90, 90};

      for ( unsigned i = 0; i < 21; ++i ) {
         edge_point_coord_x[i] *= YAC_RAD;
         edge_point_coord_y[i] *= YAC_RAD;
      }

      unsigned num_cells[] = {3,3};
      unsigned cyclic[] = {1,0};
      struct grid * curvi_grid, * cell_grid, * edge_grid;

      curvi_grid = yac_curve2d_grid_new(coord_x, coord_y, num_cells, cyclic);

      test_grid_3x3_1x0(curvi_grid, "curvilinear grid", CURVE2D);
      test_grid_cell_neigh_dep_3x3_1x0(curvi_grid, "curvilinear grid");


      cell_grid = yac_generate_cell_grid(curvi_grid, cell_point_coord_x, cell_point_coord_y);
      edge_grid = yac_generate_edge_grid(curvi_grid, edge_point_coord_x, edge_point_coord_y);

      test_grid_3x2_1x0(cell_grid, "curvilinear grid", CURVE2D);
      test_grid_cell_neigh_dep_3x2_1x0(cell_grid, "curvilinear grid");

      int pack_buffer_size = yac_grid_pack_size(curvi_grid, comm);
      pack_buffer_size    += yac_grid_pack_size(cell_grid, comm);
      void * pack_buffer = malloc(pack_buffer_size);
      int position = 0;

      yac_pack_grid(curvi_grid, pack_buffer, pack_buffer_size, &position, comm);
      yac_pack_grid(cell_grid, pack_buffer, pack_buffer_size, &position, comm);

      position = 0;

      grid_unpacked[0] = yac_unpack_grid(pack_buffer, pack_buffer_size, &position, comm);
      grid_unpacked[1] = yac_unpack_grid(pack_buffer, pack_buffer_size, &position, comm);

      test_grid_3x3_1x0(grid_unpacked[0], "unstructured grid", UNSTRUCT);
      test_grid_3x2_1x0(grid_unpacked[1], "unstructured grid", UNSTRUCT);
      test_edge_grid(edge_grid, curvi_grid, edge_point_coord_x, edge_point_coord_y);

      yac_delete_grid(grid_unpacked[0]);
      yac_delete_grid(grid_unpacked[1]);
      free(pack_buffer);
      yac_delete_grid(curvi_grid);
      yac_delete_grid(cell_grid);
      yac_delete_grid(edge_grid);
   }

   { // test get_2d_grid_extent on some mean cases

      {
         double coord_x[3] = {0,120*YAC_RAD,240*YAC_RAD};
         double coord_y[3] = {85*YAC_RAD,85*YAC_RAD,85*YAC_RAD};
         unsigned num_corners = 3;
         size_t num_cells = 1;
         size_t num_corners_per_cell[1] = {3};
         unsigned cell_to_corner_dep[3] = {0,1,2};
         size_t * p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
         unsigned * p_cell_to_corner_dep = TO_POINTER(cell_to_corner_dep);
         struct grid * unstruct_grid;

         struct dep_list * cell_to_vertex =
           yac_dep_list_new(
             num_cells, p_num_corners_per_cell, p_cell_to_corner_dep);

         unstruct_grid = yac_unstruct_grid_new(coord_x, coord_y, num_corners, cell_to_vertex);

         double grid_extent[2][2];

         yac_get_2d_grid_extent(unstruct_grid, grid_extent);

         if (grid_extent[0][1] - grid_extent[0][0] < 360.0*YAC_RAD - tol ||
             grid_extent[1][0] > 85.0*YAC_RAD + tol ||
             double_are_unequal(grid_extent[1][1], 90.0*YAC_RAD))
            PUT_ERR("error in get_2d_grid_extent\n")

         yac_delete_grid(unstruct_grid);
         yac_dep_list_delete(cell_to_vertex);
      }

      {
         double coord_x[4] = {0,0,120,240};
         double coord_y[4] = {85,89,85,85};

         for ( unsigned i = 0; i < 4; ++i ) {
            coord_x[i] *= YAC_RAD;
            coord_y[i] *= YAC_RAD;
         }

         unsigned num_corners = 4;
         size_t num_cells = 3;
         size_t num_corners_per_cell[3] = {3,3,3};
         unsigned cell_to_corner_dep[9] = {0,1,2, 1,2,3, 0,1,3};
         size_t * p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
         unsigned * p_cell_to_corner_dep = TO_POINTER(cell_to_corner_dep);
         struct grid * unstruct_grid;

         struct dep_list * cell_to_vertex =
           yac_dep_list_new(
             num_cells, p_num_corners_per_cell, p_cell_to_corner_dep);

         unstruct_grid = yac_unstruct_grid_new(coord_x, coord_y, num_corners, cell_to_vertex);

         double grid_extent[2][2];

         yac_get_2d_grid_extent(unstruct_grid, grid_extent);

         if (grid_extent[0][1] - grid_extent[0][0] < 360*YAC_RAD - tol ||
             grid_extent[1][0] > 85.0*YAC_RAD + tol ||
             double_are_unequal(grid_extent[1][1], 90*YAC_RAD))
            PUT_ERR("error in get_2d_grid_extent\n")

         yac_delete_grid(unstruct_grid);
         yac_dep_list_delete(cell_to_vertex);
      }
   }

   { // test subgrid generation

      unsigned i;

      unsigned ref_included_cells[] = {7,8,9,10,13,14,15,16};
      unsigned ref_included_corners[] = {7,8,9,10,11,13,14,15,16,17,19,20,21,22,23};
      unsigned ref_included_edges[] = {15,16,17,18,19,20,21,22,23,27,28,29,30,31,32,33,34,35,39,41,43,45};

      unsigned cells_to_include[] = {7,8,9,10,13,14,15,16};

      { // regular 2d grids

         double coord_x[] = {0,60*YAC_RAD,120*YAC_RAD,180*YAC_RAD,240*YAC_RAD,300*YAC_RAD};
         double coord_y[] = {-90*YAC_RAD,-45*YAC_RAD,0,45*YAC_RAD,90*YAC_RAD};
         unsigned num_cells[] = {6,4};
         unsigned cyclic[] = {1,0};
         struct grid * reg_grid;
         struct grid * subgrid;

         reg_grid = yac_reg2d_grid_new(coord_x, coord_y, num_cells, cyclic);

         unsigned *included_cells, *included_corners, *included_edges;

         subgrid = yac_generate_subgrid(reg_grid, cells_to_include, 8, &included_cells,
                                        &included_corners, &included_edges);

         for (i = 0; i < 8; ++i)
            if (included_cells[i] != ref_included_cells[i])
               PUT_ERR("error in generate_subgrid\n")

         for (i = 0; i < 15; ++i)
            if (included_corners[i] != ref_included_corners[i])
               PUT_ERR("error in generate_subgrid\n")

         for (i = 0; i < 22; ++i)
            if (included_edges[i] != ref_included_edges[i])
               PUT_ERR("error in generate_subgrid\n")

         free(included_cells);
         free(included_corners);
         free(included_edges);
         yac_delete_grid(subgrid);
         yac_delete_grid(reg_grid);
      }

      { // unstructured grids

         double coord_x[30] = {0,60,120,180,240,300,0,60,120,180,240,300,0,60,120,180,240,300,0,60,120,180,240,300,0,60,120,180,240,300,};
         double coord_y[30] = {-90,-90,-90,-90,-90,-90,-45,-45,-45,-45,-45,-45,0,0,0,0,0,0,45,45,45,45,45,45,90,90,90,90,90,90};

         for ( i = 0; i < 30; ++i ) {
            coord_x[i] *= YAC_RAD;
            coord_y[i] *= YAC_RAD;
         }

         unsigned num_corners = 30;
         size_t num_cells = 24;
         size_t num_corners_per_cell[30] = {4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4};
         unsigned cell_to_corner_dep[96] = {0,1,7,6, 1,2,8,7, 2,3,9,8, 3,4,10,9, 4,5,11,10, 5,0,6,11,
                                            6,7,13,12, 7,8,14,13, 8,9,15,14, 9,10,16,15, 10,11,17,16, 11,6,12,17,
                                            12,13,19,18, 13,14,20,19, 14,15,21,20, 15,16,22,21, 16,17,23,22, 17,12,18,23,
                                            18,19,25,24, 19,20,26,25, 20,21,27,26, 21,22,28,27, 22,23,29,28, 23,18,24,29};
         size_t * p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
         unsigned * p_cell_to_corner_dep = TO_POINTER(cell_to_corner_dep);
         struct grid * unstruct_grid;
         struct grid * subgrid;

         struct dep_list * cell_to_vertex =
           yac_dep_list_new(
             num_cells, p_num_corners_per_cell, p_cell_to_corner_dep);

         unstruct_grid = yac_unstruct_grid_new(coord_x, coord_y, num_corners, cell_to_vertex);

         unsigned *included_cells, *included_corners, *included_edges;

         subgrid = yac_generate_subgrid(unstruct_grid, cells_to_include, 8, &included_cells,
                                        &included_corners, &included_edges);

         for (i = 0; i < 8; ++i)
            if (included_cells[i] != ref_included_cells[i])
               PUT_ERR("error in generate_subgrid\n")

         for (i = 0; i < 15; ++i)
            if (included_corners[i] != ref_included_corners[i])
               PUT_ERR("error in generate_subgrid\n")

         for (i = 0; i < 22; ++i)
            if (included_edges[i] != ref_included_edges[i])
               PUT_ERR("error in generate_subgrid\n")

         free(included_cells);
         free(included_corners);
         free(included_edges);
         yac_delete_grid(subgrid);
         yac_delete_grid(unstruct_grid);
         yac_dep_list_delete(cell_to_vertex);
      }

      { // curve2d grids

         double coord_x[30] = {0,60,120,180,240,300,0,60,120,180,240,300,0,60,120,180,240,300,0,60,120,180,240,300,0,60,120,180,240,300,};
         double coord_y[30] = {-90,-90,-90,-90,-90,-90,-45,-45,-45,-45,-45,-45,0,0,0,0,0,0,45,45,45,45,45,45,90,90,90,90,90,90};

         for ( i = 0; i < 30; ++i ) {
            coord_x[i] *= YAC_RAD;
            coord_y[i] *= YAC_RAD;
         }

         unsigned num_cells[] = {6,4};
         unsigned cyclic[] = {1,0};
         struct grid * curve2d_grid;
         struct grid * subgrid;

         curve2d_grid = yac_curve2d_grid_new(coord_x, coord_y, num_cells, cyclic);

         unsigned *included_cells, *included_corners, *included_edges;

         subgrid = yac_generate_subgrid(curve2d_grid, cells_to_include, 8, &included_cells,
                                        &included_corners, &included_edges);

         for (i = 0; i < 8; ++i)
            if (included_cells[i] != ref_included_cells[i])
               PUT_ERR("error in generate_subgrid\n")

         for (i = 0; i < 15; ++i)
            if (included_corners[i] != ref_included_corners[i])
               PUT_ERR("error in generate_subgrid\n")

         for (i = 0; i < 22; ++i)
            if (included_edges[i] != ref_included_edges[i])
               PUT_ERR("error in generate_subgrid\n")

         free(included_cells);
         free(included_corners);
         free(included_edges);
         yac_delete_grid(subgrid);
         yac_delete_grid(curve2d_grid);
      }
   }

   { // test triangulation of cells

      check_cell_triangulation((double[]){0,1,0}, (double[]){0,0,1}, 3);
      check_cell_triangulation((double[]){0,1,1,0}, (double[]){0,0,1,1}, 4);
      check_cell_triangulation(
        (double[]){0,1,1,0.5,0}, (double[]){0,0,1,1.5,1}, 5);
      check_cell_triangulation(
        (double[]){0,0.5,1,1,0.5,0}, (double[]){0,-0.5,0,1,1.5,1}, 6);
      check_cell_triangulation(
        (double[]){-0.25,0, 0.5,1,1.25,1,0},
        (double[]){ 0.5 ,0,-0.5,0,0.5 ,1,1}, 7);
      check_cell_triangulation(
        (double[]){-0.25,0, 0.5,1,1.25,1,0.5,0},
        (double[]){ 0.5 ,0,-0.5,0,0.5 ,1,1.5,1}, 8);
   }

   yac_free_communicator(comm);
   yac_comm_finalize();

   return TEST_EXIT_CODE;
}

