/**
 * @file test_nnn_search_parallel.c
 *
 * @copyright Copyright  (C)  2017 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <mpi.h>

#include "tests.h"
#include "test_common.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_mpi.h"
#include "domain.h"
#include "utils.h"
#include "geometry.h"
#include "sphere_part.h"
#include "nnn_search.h"

// corner and cell ids for a 7 x 7 grid (x == request point position)

// 56-----57-----58-----59-----60-----61-----62-----63
//  |     x|     x|     x|     x|     x|     x|     x|
//  |  42  |  43  |  44  |  45  |  46  |  47  |  48  |
//  |      |      |      |      |      |      |      |
// 48-----49-----50-----51-----52-----53-----54-----55
//  |     x|     x|     x|     x|     x|     x|     x|
//  |  35  |  36  |  37  |  38  |  39  |  40  |  41  |
//  |      |      |      |      |      |      |      |
// 40-----41-----42-----43-----44-----45-----46-----47
//  |     x|     x|     x|     x|     x|     x|     x|
//  |  28  |  29  |  30  |  31  |  32  |  33  |  34  |
//  |      |      |      |      |      |      |      |
// 32-----33-----34-----35-----36-----37-----38-----39
//  |     x|     x|     x|     x|     x|     x|     x|
//  |  21  |  22  |  23  |  24  |  25  |  26  |  27  |
//  |      |      |      |      |      |      |      |
// 24-----25-----26-----27-----28-----29-----30-----31
//  |     x|     x|     x|     x|     x|     x|     x|
//  |  14  |  15  |  16  |  17  |  18  |  19  |  20  |
//  |      |      |      |      |      |      |      |
// 16-----17-----18-----19-----20-----21-----22-----23
//  |     x|     x|     x|     x|     x|     x|     x|
//  |  07  |  08  |  09  |  10  |  11  |  12  |  13  |
//  |      |      |      |      |      |      |      |
// 08-----09-----10-----11-----12-----13-----14-----15
//  |     x|     x|     x|     x|     x|     x|     x|
//  |  00  |  01  |  02  |  03  |  04  |  05  |  06  |
//  |      |      |      |      |      |      |      |
// 00-----01-----02-----03-----04-----05-----06-----07
//
// the domain is distributed among the processes as follows:
// (index == process)
//
// 3---3---3---3---3---3---3---3
// | 3 | 3 | 3 | 3 | 3 | 3 | 3 |
// 3---3---3---3---3---3---3---3
// | 3 | 3 | 3 | 3 | 3 | 3 | 3 |
// 3---3---3---1---2---2---3---3
// | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
// 1---1---1---2---2---2---2---2
// | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
// 1---1---1---1---2---2---2---2
// | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
// 1---1---1---0---0---0---2---2
// | 0 | 0 | 0 | 0 | 0 | 0 | 0 |
// 0---0---0---0---0---0---0---0
// | 0 | 0 | 0 | 0 | 0 | 0 | 0 |
// 0---0---0---0---0---0---0---0
//
// the mask looks as follows (# == masked point)
//
// +---+---+---+---+---+---+---#
// |   |   |   |   |   |   |   |
// +---+---+---+---+---+---#---+
// |   |   |   |   |   |   |   |
// +---+---+---+---+---#---+---+
// |   |   |   |   |   |   |   |
// +---+---+---+---#---+---+---+
// |   |   |   |   |   |   |   |
// +---+---+---#---+---+---+---+
// |   |   |   |   |   |   |   |
// +---+---#---+---+---+---+---+
// |   |   |   |   |   |   |   |
// +---#---+---+---+---+---+---+
// |   |   |   |   |   |   |   |
// #---+---+---+---+---+---+---+

#define NUM_PROCS (4)

unsigned num_cells[NUM_PROCS][2] = {{7,3},{4,5},{5,5},{7,3}};
unsigned cyclic[2] = {0,0};
double coordinates_x[NUM_PROCS][8] = {{0,1,2,3,4,5,6,7},{0,1,2,3,4},
                                      {2,3,4,5,6,7},{0,1,2,3,4,5,6,7}};
double coordinates_y[NUM_PROCS][8] = {{0,1,2,3},{1,2,3,4,5,6},
                                      {1,2,3,4,5,6},{4,5,6,7}};

int global_corner_ids[NUM_PROCS][36] = {{ 0, 1, 2, 3, 4, 5, 6, 7,
                                          8, 9,10,11,12,13,14,15,
                                         16,17,18,19,20,21,22,23,
                                         24,25,26,27,28,29,30,31},
                                        { 8, 9,10,11,12,
                                         16,17,18,19,20,
                                         24,25,26,27,28,
                                         32,33,34,35,36,
                                         40,41,42,43,44,
                                         48,49,50,51,52},
                                        {10,11,12,13,14,15,
                                         18,19,20,21,22,23,
                                         26,27,28,29,30,31,
                                         34,35,36,37,38,39,
                                         42,43,44,45,46,47,
                                         50,51,52,53,54,55},
                                        {32,33,34,35,36,37,38,39,
                                         40,41,42,43,44,45,46,47,
                                         48,49,50,51,52,53,54,55,
                                         56,57,58,59,60,61,62,63}};
unsigned corner_is_on_rank[NUM_PROCS][36] = {{-1,-1,-1,-1,-1,-1,-1,-1,
                                              -1,-1,-1,-1,-1,-1,-1,-1,
                                               1, 1, 1,-1,-1,-1, 2, 2,
                                               1, 1, 1, 1, 2, 2, 2, 2},
                                             { 0, 0, 0, 0, 0,
                                              -1,-1,-1, 0, 0,
                                              -1,-1,-1,-1, 2,
                                              -1,-1,-1, 2, 2,
                                               3, 3, 3,-1, 2,
                                               3, 3, 3, 3, 3},
                                             { 0, 0, 0, 0, 0, 0,
                                               1, 0, 0, 0,-1,-1,
                                               1, 1,-1,-1,-1,-1,
                                               1,-1,-1,-1,-1,-1,
                                               3, 1,-1,-1, 3, 3,
                                               3, 3, 3, 3, 3, 3},
                                             { 1, 1, 1, 2, 2, 2, 2, 2,
                                              -1,-1,-1, 1, 2, 2,-1,-1,
                                              -1,-1,-1,-1,-1,-1,-1,-1,
                                              -1,-1,-1,-1,-1,-1,-1,-1}};
int global_cell_ids[NUM_PROCS][25] = {{ 0, 1, 2, 3, 4, 5, 6,
                                        7, 8, 9,10,11,12,13,
                                       14,15,16,17,18,19,20},
                                      { 7, 8, 9,10,
                                       14,15,16,17,
                                       21,22,23,24,
                                       28,29,30,31,
                                       35,36,37,38},
                                      { 9,10,11,12,13,
                                       16,17,18,19,20,
                                       23,24,25,26,27,
                                       30,31,32,33,34,
                                       37,38,39,40,41},
                                      {28,29,30,31,32,33,34,
                                       35,36,37,38,39,40,41,
                                       42,43,44,45,46,47,48}};
unsigned cell_is_on_rank[NUM_PROCS][25] = {{-1,-1,-1,-1,-1,-1,-1,
                                            -1,-1,-1,-1,-1,-1,-1,
                                             1, 1, 1, 2, 2, 2, 2},
                                           { 0, 0, 0, 0,
                                            -1,-1,-1, 2,
                                            -1,-1,-1, 2,
                                            -1,-1,-1, 2,
                                             3, 3, 3, 3},
                                           { 0, 0, 0, 0, 0,
                                             1,-1,-1,-1,-1,
                                             1,-1,-1,-1,-1,
                                             1,-1,-1,-1,-1,
                                             3, 3, 3, 3, 3},
                                           { 1, 1, 1, 2, 2, 2, 2,
                                             3, 3, 3, 3, 3, 3, 3,
                                             3, 3, 3, 3, 3, 3, 3}};


unsigned mask_values[4][36] = {{1,0,0,0,0,0,0,0,
                                0,1,0,0,0,0,0,0,
                                0,0,1,0,0,0,0,0,
                                0,0,0,1,0,0,0,0},
                               {0,1,0,0,0,
                                0,0,1,0,0,
                                0,0,0,1,0,
                                0,0,0,0,1,
                                0,0,0,0,0,
                                0,0,0,0,0},
                               {0,0,0,0,0,0,
                                1,0,0,0,0,0,
                                0,1,0,0,0,0,
                                0,0,1,0,0,0,
                                0,0,0,1,0,0,
                                0,0,0,0,1,0},
                               {0,0,0,0,1,0,0,0,
                                0,0,0,0,0,1,0,0,
                                0,0,0,0,0,0,1,0,
                                0,0,0,0,0,0,0,1}};

static void nnn_finish_callback(
  void * user_data, struct request_point * request_points);

struct nnn_test_user_data {
  size_t num_points;
  struct request_point * ref_request_points;
};

static struct request_point * generate_ref_request_points(
  int * global_ids, double (*coordinates_xyz)[3], size_t num_points, unsigned n,
  unsigned my_rank);

int main (void) {

  struct communicator * global_comm = yac_mpi_comm_new();

  if (yac_get_comm_size(global_comm) != 4) {
    PUT_ERR("wrong number of processes");
    return TEST_EXIT_CODE;
  }

  //--------------------
  // set up global search
  //--------------------

  int my_rank = yac_get_comm_rank(global_comm);

  // define a domain
  for (size_t i = 0; i < 4; ++i) {
    for (size_t j = 0; j < 8; ++j) {
      coordinates_x[i][j] *= YAC_RAD;
      coordinates_y[i][j] *= YAC_RAD;
    }
  }

  struct grid * grid =
    yac_reg2d_grid_new(
      coordinates_x[my_rank], coordinates_y[my_rank], num_cells[my_rank],
      cyclic);

  struct subdomain subdomain;

  yac_init_subdomain_unstruct(
    &subdomain, grid, global_comm, global_corner_ids[my_rank],
    corner_is_on_rank[my_rank], global_cell_ids[my_rank],
    cell_is_on_rank[my_rank], NULL, NULL);

  struct domain domain;
  struct subdomain * blocks[1] = {&subdomain};
  yac_init_domain(&domain, global_comm, blocks, 1, "domain");

  // finish domain definition
  yac_enddef_domain(&domain);

  // points
  struct points points;
  yac_init_points(
    &points, grid, CORNER, coordinates_x[my_rank], coordinates_y[my_rank]);

  // masks
  struct mask mask;
  yac_init_mask(&mask, mask_values[my_rank], &points);

  // point sets
  struct points * points_per_subdomain[1] = {&points};
  struct points ** points_set[1] = {&(points_per_subdomain[0])};
  struct mask * masks_per_subdomain[1] = {&mask};
  struct mask ** masks_set[1] = {&(masks_per_subdomain[0])};

  //---------------------------------
  // setup n-nearest-neighbour search
  //---------------------------------

  struct nnn_search * search =
    yac_nnn_search_new(
      &domain, "unique_string", &(points_set[0]), &(masks_set[0]), 1, 0);

  //------------------------------
  // do 1-nearest-neighbour search
  //------------------------------

  {
    int point_global_ids[NUM_PROCS][49] =
      {{ 0, 1, 2, 3, 4, 5, 6,
         7, 8, 9,10,11,12,13,
        14,15,16,17,18,19,20,
        21,22,23,24,25,26,27,
        28,29,30,31,32,33,34,
        35,36,37,38,39,40,41,
        42,43,44,45,46,47,48},
       { 0, 1, 2, 3, 4, 5, 6,
         7, 8, 9,10,11,12,13,
        14,15,16,17,18,19,20,
        27,26,25,24,23,22,21,
        28,29,30,31,32,33,34,
        35,36,37,38,39,40,41,
        42,43,44,45,46,47,48},
       { 0, 1, 2, 3, 4, 5, 6,
         7, 8, 9,10,11,12,13,
        14,15,16,17,18,19,20,
        28,29,30,31,32,33,34,
        21,22,23,24,25,26,27,
        35,36,37,38,39,40,41,
        42,43,44,45,46,47,48},
       { 0, 1, 2, 3, 4, 5, 6,
         7, 8, 9,10,11,12,13,
        14,15,16,17,18,19,20,
        21,22,23,24,25,26,27,
        28,29,30,31,32,33,34,
        35,36,37,38,39,40,41,
        42,43,44,45,46,47,48},};
    double point_coordinates_x[NUM_PROCS][49] =
      {{0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75},
       {0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        6.75,5.75,4.75,3.75,2.75,1.75,0.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75},
       {0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75},
       {0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75,
        0.75,1.75,2.75,3.75,4.75,5.75,6.75}};
    double point_coordinates_y[NUM_PROCS][49] =
      {{0.75,0.75,0.75,0.75,0.75,0.75,0.75,
        1.75,1.75,1.75,1.75,1.75,1.75,1.75,
        2.75,2.75,2.75,2.75,2.75,2.75,2.75,
        3.75,3.75,3.75,3.75,3.75,3.75,3.75,
        4.75,4.75,4.75,4.75,4.75,4.75,4.75,
        5.75,5.75,5.75,5.75,5.75,5.75,5.75,
        6.75,6.75,6.75,6.75,6.75,6.75,6.75},
       {0.75,0.75,0.75,0.75,0.75,0.75,0.75,
        1.75,1.75,1.75,1.75,1.75,1.75,1.75,
        2.75,2.75,2.75,2.75,2.75,2.75,2.75,
        3.75,3.75,3.75,3.75,3.75,3.75,3.75,
        4.75,4.75,4.75,4.75,4.75,4.75,4.75,
        5.75,5.75,5.75,5.75,5.75,5.75,5.75,
        6.75,6.75,6.75,6.75,6.75,6.75,6.75},
       {0.75,0.75,0.75,0.75,0.75,0.75,0.75,
        1.75,1.75,1.75,1.75,1.75,1.75,1.75,
        2.75,2.75,2.75,2.75,2.75,2.75,2.75,
        4.75,4.75,4.75,4.75,4.75,4.75,4.75,
        3.75,3.75,3.75,3.75,3.75,3.75,3.75,
        5.75,5.75,5.75,5.75,5.75,5.75,5.75,
        6.75,6.75,6.75,6.75,6.75,6.75,6.75},
       {0.75,0.75,0.75,0.75,0.75,0.75,0.75,
        1.75,1.75,1.75,1.75,1.75,1.75,1.75,
        2.75,2.75,2.75,2.75,2.75,2.75,2.75,
        3.75,3.75,3.75,3.75,3.75,3.75,3.75,
        4.75,4.75,4.75,4.75,4.75,4.75,4.75,
        5.75,5.75,5.75,5.75,5.75,5.75,5.75,
        6.75,6.75,6.75,6.75,6.75,6.75,6.75}};
    double point_coordinates_xyz[NUM_PROCS][49][3];

    size_t num_points =
      sizeof(point_global_ids[0]) / sizeof(point_global_ids[0][0]);
    for (size_t rank = 0; rank < NUM_PROCS; ++rank)
      for (size_t i = 0; i < num_points; ++i)
        LLtoXYZ(point_coordinates_x[rank][i]*YAC_RAD,
                point_coordinates_y[rank][i]*YAC_RAD,
                &(point_coordinates_xyz[rank][i][0]));
    unsigned n = 1;
    struct nnn_test_user_data data =
      {.num_points = num_points,
       .ref_request_points =
         generate_ref_request_points(
           &(point_global_ids[my_rank][0]), point_coordinates_xyz[my_rank],
           num_points, n, my_rank)};

    yac_nnn_search_do_search(
      search, n, data.num_points, point_coordinates_xyz[my_rank],
      &(point_global_ids[my_rank][0]), (void*)(&data), nnn_finish_callback);

    yac_nnn_search_wait(search);

    yac_nnn_search_free_request_points(data.ref_request_points, num_points);
    free(data.ref_request_points);

  }

  //---------------
  // cleanup
  //---------------

  yac_nnn_search_delete(search);
  yac_free_domain(&domain);
  yac_free_mask(&mask);
  yac_free_points(&points);
  yac_delete_grid(grid);

  yac_free_communicator(global_comm);
  MPI_Finalize();

  return TEST_EXIT_CODE;
}

static void check_request_points(
  struct request_point * request_points,
  struct request_point * ref_request_points, size_t num_points) {

  for (size_t i = 0; i < num_points; ++i) {

    if (request_points[i].global_id != ref_request_points[i].global_id)
      PUT_ERR("wrong global id");

    if (double_are_unequal(request_points[i].coordinate_xyz[0],
                           ref_request_points[i].coordinate_xyz[0]) ||
        double_are_unequal(request_points[i].coordinate_xyz[1],
                           ref_request_points[i].coordinate_xyz[1]) ||
        double_are_unequal(request_points[i].coordinate_xyz[2],
                           ref_request_points[i].coordinate_xyz[2]))
      PUT_ERR("wrong coordinate_xyz");

    if (fabs(request_points[i].cos_angle - ref_request_points[i].cos_angle) >
        1e-9) PUT_ERR("wrong cos_angle");

    if (request_points[i].num_found_points !=
        ref_request_points[i].num_found_points) {
      PUT_ERR("wrong num_found_points");
    } else {

      for (size_t j = 0; j < ref_request_points[i].num_found_points; ++j) {

        if (fabs(request_points[i].results[j].cos_angle -
                 ref_request_points[i].results[j].cos_angle) > 1e-9)
          PUT_ERR("wrong results[j].cos_angle");

        if (double_are_unequal(
              request_points[i].results[j].coordinate_xyz[0],
              ref_request_points[i].results[j].coordinate_xyz[0]) ||
            double_are_unequal(
              request_points[i].results[j].coordinate_xyz[1],
              ref_request_points[i].results[j].coordinate_xyz[1]) ||
            double_are_unequal(
              request_points[i].results[j].coordinate_xyz[2],
              ref_request_points[i].results[j].coordinate_xyz[2]))
          PUT_ERR("wrong results[j].coordinate_xyz");

        if (request_points[i].results[j].global_id !=
            ref_request_points[i].results[j].global_id)
          PUT_ERR("wrong results[j].global_id");

        if (request_points[i].results[j].is_local !=
            ref_request_points[i].results[j].is_local) {
          PUT_ERR("wrong results[j].is_local");
        } else if (ref_request_points[i].results[j].is_local) {
          if (request_points[i].results[j].data.local.local_id !=
              ref_request_points[i].results[j].data.local.local_id)
            PUT_ERR("wrong results[j].data.local.local_id");
          if (request_points[i].results[j].data.local.subdomain_id !=
              ref_request_points[i].results[j].data.local.subdomain_id)
            PUT_ERR("wrong results[j].data.local.subdomain_id");
        } else {
          if (request_points[i].results[j].data.remote.rank !=
              ref_request_points[i].results[j].data.remote.rank)
            PUT_ERR("wrong results[j].data.remote.rank");
        }
      }
    }
  }
}

static void nnn_finish_callback(
  void * user_data, struct request_point * request_points) {

  struct nnn_test_user_data * data = (struct nnn_test_user_data *)user_data;

  check_request_points(
    request_points, data->ref_request_points, data->num_points);

  yac_nnn_search_free_request_points(request_points, data->num_points);
  free(request_points);
}

struct grid_vertex {
  int global_id;
  unsigned local_id;
  double coordinate_xyz[3];
  unsigned owner;
};

static int compare_results(const void * a, const void * b) {

  int res = (((struct result_point*)b)->cos_angle >
             ((struct result_point*)a)->cos_angle) -
            (((struct result_point*)a)->cos_angle >
             ((struct result_point*)b)->cos_angle);
  if (res) return res;
  res = ((struct result_point*)a)->global_id -
        ((struct result_point*)b)->global_id;
  if (res) return res;
  for (size_t i = 0; i < 3; ++i) {
    res = (((struct result_point*)b)->coordinate_xyz[i] >
           ((struct result_point*)a)->coordinate_xyz[i]) -
          (((struct result_point*)a)->coordinate_xyz[i] >
           ((struct result_point*)b)->coordinate_xyz[i]);
    if (res) return res;
  }
  return 0;
}

struct request_point * generate_ref_request_points (
  int * global_ids, double (*point_coordinates_xyz)[3], size_t num_points,
  unsigned n, unsigned my_rank) {

  struct grid_vertex * vertices;
  {
    size_t max_num_vertices = 0;
    for (size_t rank = 0; rank < NUM_PROCS; ++rank)
      for (size_t x_idx = 0; x_idx <= num_cells[rank][0]; ++x_idx)
        max_num_vertices += (num_cells[rank][0]+1)*(num_cells[rank][1]+1);
    vertices = malloc(max_num_vertices * sizeof(*vertices));
  }

  size_t num_vertices = 0;
  for (size_t rank = 0; rank < NUM_PROCS; ++rank) {

    for (size_t x_idx = 0; x_idx <= num_cells[rank][0]; ++x_idx) {

      for (size_t y_idx = 0; y_idx <= num_cells[rank][1]; ++y_idx) {

        size_t vertex_idx = y_idx * (num_cells[rank][0] + 1) + x_idx;

        if ((corner_is_on_rank[rank][vertex_idx] != -1) ||
            (mask_values[rank][vertex_idx] == 1)) continue;

        vertices[num_vertices].global_id = global_corner_ids[rank][vertex_idx];
        vertices[num_vertices].local_id = (unsigned)vertex_idx;
        LLtoXYZ(coordinates_x[rank][x_idx], coordinates_y[rank][y_idx],
                &(vertices[num_vertices].coordinate_xyz[0]));
        vertices[num_vertices].owner = rank;
        ++num_vertices;
      }
    }
  }

  struct point_sphere_part_search * search;
  {
    double (*coordinates_xyz)[3] =
      malloc(num_vertices * sizeof(*coordinates_xyz));

    for (size_t i = 0; i < num_vertices; ++i) {
      coordinates_xyz[i][0] = vertices[i].coordinate_xyz[0];
      coordinates_xyz[i][1] = vertices[i].coordinate_xyz[1];
      coordinates_xyz[i][2] = vertices[i].coordinate_xyz[2];
    }
    search =
      yac_point_sphere_part_search_new(num_vertices, coordinates_xyz);
    free(coordinates_xyz);
  }

  double * cos_angles = NULL;
  double (*result_coordinates_xyz)[3] = NULL;
  size_t * local_point_ids = NULL;
  size_t * num_local_point_ids =
    malloc(num_points * sizeof(*num_local_point_ids));

  size_t cos_angles_array_size = 0;
  size_t result_coordinates_xyz_array_size = 0;
  size_t local_point_ids_array_size = 0;
  yac_point_sphere_part_search_NNN(
    search, num_points, point_coordinates_xyz, (size_t)n, &cos_angles,
    &cos_angles_array_size, &result_coordinates_xyz,
    &result_coordinates_xyz_array_size, &local_point_ids,
    &local_point_ids_array_size, num_local_point_ids);

  struct request_point * request_points =
    malloc(num_points * sizeof(*request_points));
  for (size_t i = 0, k = 0; i < num_points; ++i) {

    unsigned num_found_points = MIN(n, num_local_point_ids[i]);

    request_points[i].global_id = global_ids[i];
    request_points[i].coordinate_xyz[0] = point_coordinates_xyz[i][0];
    request_points[i].coordinate_xyz[1] = point_coordinates_xyz[i][1];
    request_points[i].coordinate_xyz[2] = point_coordinates_xyz[i][2];
    request_points[i].results =
      malloc(num_local_point_ids[i] * sizeof(*(request_points[i].results)));
    for (size_t j = 0; j < num_local_point_ids[i]; ++j, ++k) {
      unsigned local_id = (unsigned)(local_point_ids[k]);
      request_points[i].results[j].cos_angle = cos_angles[k];
      request_points[i].results[j].coordinate_xyz[0] =
        result_coordinates_xyz[k][0];
      request_points[i].results[j].coordinate_xyz[1] =
        result_coordinates_xyz[k][1];
      request_points[i].results[j].coordinate_xyz[2] =
        result_coordinates_xyz[k][2];
      request_points[i].results[j].global_id = vertices[local_id].global_id;
      request_points[i].results[j].is_local =
        my_rank == vertices[local_id].owner;
      if (request_points[i].results[j].is_local) {
        request_points[i].results[j].data.local.local_id =
          vertices[local_id].local_id;
        request_points[i].results[j].data.local.subdomain_id = 0;
      } else {
        request_points[i].results[j].data.remote.rank =
          vertices[local_id].owner;
      }
    }
    qsort(
      request_points[i].results, num_found_points,
      sizeof(*(request_points[i].results)), compare_results);
    request_points[i].num_found_points = num_found_points;
    request_points[i].cos_angle =
      request_points[i].results[num_found_points-1].cos_angle;
  }

  free(local_point_ids);
  free(result_coordinates_xyz);
  free(cos_angles);
  free(num_local_point_ids);
  free(vertices);
  yac_delete_point_sphere_part_search(search);

  return request_points;
}

