// #define VERBOSE

/**
 * @file test_nnn_search_parallel_2.c
 *
 * @copyright Copyright  (C)  2018 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <mpi.h>

#include "tests.h"
#include "test_common.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_mpi.h"
#include "domain.h"
#include "utils.h"
#include "geometry.h"
#include "sphere_part.h"
#include "nnn_search.h"

/* ------------------------------------------------- */

static void compute_reg2d_decomp(
  unsigned num_points[2], unsigned total_num_procs, unsigned * num_procs);

static void nnn_finish_callback(
  void * user_data, struct request_point * request_points);

static struct communicator * global_comm;
static unsigned my_rank,  my_size;

int main (void) {

  global_comm = yac_mpi_comm_new();
  my_rank = yac_get_comm_rank(global_comm);
  my_size = yac_get_comm_size(global_comm);

  if (my_size != 2) {
    PUT_ERR("wrong number of processes\n");
    return TEST_EXIT_CODE;
  }

  //------------------------------------
  // compute decomposition and grid data
  //------------------------------------

#define NUM_CELLS_X (20)
#define NUM_CELLS_Y (10)

  unsigned total_nbr_cells[2] = {NUM_CELLS_X, NUM_CELLS_Y};
  unsigned num_procs[2];

  compute_reg2d_decomp(total_nbr_cells, my_size, num_procs);

  unsigned block_pos[2];
  unsigned block_size[2];
  unsigned nbr_cells[2];
  unsigned cyclic[2];
  unsigned nbr_points[2];
  block_pos[0] = my_rank%num_procs[0];
  block_pos[1] = my_rank/num_procs[0];
  block_size[0] = (NUM_CELLS_X + num_procs[0] - 1)/num_procs[0];
  block_size[1] = (NUM_CELLS_Y + num_procs[1] - 1)/num_procs[1];
  nbr_cells[0] = MIN(block_size[0], NUM_CELLS_X - block_size[0] * block_pos[0]);
  nbr_cells[1] = MIN(block_size[1], NUM_CELLS_Y - block_size[1] * block_pos[1]);
  cyclic[0] = num_procs[0] == 1;
  cyclic[1] = 0;

  // halo
  if (!cyclic[0])
   nbr_cells[0] += 2;
  if (num_procs[1] > 1) {

    nbr_cells[1] += 1;

    if ((block_pos[1] != 0) && (block_pos[1] != num_procs[1]-1))
      nbr_cells[1] += 1;
  }

  if (cyclic[0])
    nbr_points[0] = nbr_cells[0];
  else
    nbr_points[0] = nbr_cells[0]+1;

  nbr_points[1] = nbr_cells[1] + 1;

  double * global_x_vertices = malloc((NUM_CELLS_X+3) * sizeof(*global_x_vertices));
  double * global_y_vertices = malloc((NUM_CELLS_Y+1) * sizeof(*global_y_vertices));;
  double * x_vertices;
  double * y_vertices;

  for (unsigned i = 0; i < NUM_CELLS_X+3; ++i)
    global_x_vertices[i] =
      (0.0 + (360.0 / ((double)NUM_CELLS_X)) * (double)(i-1)) * YAC_RAD;
  for (unsigned i = 0; i < NUM_CELLS_Y; ++i)
    global_y_vertices[i] =
      (-90.0 + (180.0 / ((double)NUM_CELLS_Y)) * (double)i) * YAC_RAD;
  global_y_vertices[NUM_CELLS_Y] = 90.0 * YAC_RAD;

  if (cyclic[0])
    x_vertices = global_x_vertices + 1;
  else
    x_vertices = global_x_vertices + block_size[0] * block_pos[0];

  y_vertices = global_y_vertices + block_size[1] * block_pos[1];

  if (block_pos[1] != 0) y_vertices -= 1;

  double * x_cell_points = malloc(nbr_cells[0] * sizeof(*x_cell_points));
  double * y_cell_points = malloc(nbr_cells[1] * sizeof(*y_cell_points));

  for (unsigned i = 0; i < nbr_cells[0]; ++i)
    x_cell_points[i] = (x_vertices[i] + x_vertices[i + 1]) * 0.5;
  for (unsigned i = 0; i < nbr_cells[1]; ++i)
    y_cell_points[i] = (y_vertices[i] + y_vertices[i + 1]) * 0.5;

  //--------------------
  // compute domain data
  //--------------------

  int * global_global_cell_id = malloc(NUM_CELLS_Y * (NUM_CELLS_X + 2) * sizeof(*global_global_cell_id));
  unsigned * global_global_cell_id_rank = malloc(NUM_CELLS_Y * (NUM_CELLS_X + 2) * sizeof(*global_global_cell_id_rank));
  int * global_global_corner_id = malloc((NUM_CELLS_Y + 1) * (NUM_CELLS_X + 3) * sizeof(*global_global_corner_id));

  for (unsigned j = 0, id = 0; j < NUM_CELLS_Y; ++j) {

    for (unsigned i = 1; i <= NUM_CELLS_X; ++i) {

      global_global_cell_id[j * (NUM_CELLS_X + 2) + i] = id;
      global_global_cell_id_rank[j * (NUM_CELLS_X + 2) + i] =
         (id%NUM_CELLS_X) / block_size[0] + ((id/NUM_CELLS_X) / block_size[1]) * num_procs[0];
      id++;
    }
  }
  for (unsigned i = 0; i < NUM_CELLS_Y; ++i) {
    global_global_cell_id[i * (NUM_CELLS_X + 2) + 0] = global_global_cell_id[i * (NUM_CELLS_X + 2) + NUM_CELLS_X];
    global_global_cell_id_rank[i * (NUM_CELLS_X + 2) + 0] = global_global_cell_id_rank[i * (NUM_CELLS_X + 2) + NUM_CELLS_X];
    global_global_cell_id[i * (NUM_CELLS_X + 2) + NUM_CELLS_X+1] = global_global_cell_id[i * (NUM_CELLS_X + 2) + 1];
    global_global_cell_id_rank[i * (NUM_CELLS_X + 2) + NUM_CELLS_X+1] = global_global_cell_id_rank[i * (NUM_CELLS_X + 2) + 1];
  }

  if (num_procs[0] == 1) {

    for (unsigned j = 0, id = 0; j < NUM_CELLS_Y + 1; ++j) {

      for (unsigned i = 1; i <= NUM_CELLS_X; ++i)
        global_global_corner_id[j * (NUM_CELLS_X + 3) + i] = id++;
    }

  } else {

    for (unsigned j = 0, id = 0; j < NUM_CELLS_Y + 1; ++j) {

      for (unsigned i = 1; i <= NUM_CELLS_X + 1; ++i)
        global_global_corner_id[j * (NUM_CELLS_X + 3) + i] = id++;
    }

    for (unsigned i = 0; i < NUM_CELLS_Y + 1; ++i) {

      global_global_corner_id[i * (NUM_CELLS_X + 3) + 0] = global_global_corner_id[i * (NUM_CELLS_X + 3) + NUM_CELLS_X + 1];
      global_global_corner_id[i * (NUM_CELLS_X + 3) + NUM_CELLS_X + 2] = global_global_corner_id[i * (NUM_CELLS_X + 3) + 1];
    }
  }

  for (unsigned i = 0; i < (NUM_CELLS_Y)*(NUM_CELLS_X + 2); ++i)
    if(global_global_cell_id_rank[i] == my_rank) global_global_cell_id_rank[i] = -1;

  int * local_global_cell_id = malloc(nbr_cells[1] * nbr_cells[0] * sizeof(*local_global_cell_id));
  unsigned * local_global_cell_id_rank = malloc(nbr_cells[1] * nbr_cells[0] * sizeof(*local_global_cell_id_rank));
  int * local_global_corner_id = malloc(nbr_points[1] * nbr_points[0] * sizeof(*local_global_corner_id));
  unsigned * local_global_corner_id_rank = malloc(nbr_points[1] * nbr_points[0] * sizeof(*local_global_corner_id_rank));

  unsigned offset_x, offset_y;

  if (cyclic[0])
    offset_x = 1;
  else
    offset_x = block_size[0] * block_pos[0];
  offset_y = block_size[1] * block_pos[1] - (block_pos[1] != 0);

  for (unsigned j = 0; j < nbr_cells[1]; ++j) {
    for (unsigned i = 0; i < nbr_cells[0]; ++i) {
      local_global_cell_id[j * nbr_cells[0] + i] = global_global_cell_id[(j+offset_y) * (NUM_CELLS_X + 2) + i+offset_x];
      local_global_cell_id_rank[j * nbr_cells[0] + i] = global_global_cell_id_rank[(j+offset_y) * (NUM_CELLS_X + 2) +i+offset_x];
    }
  }

  free(global_global_cell_id_rank);
  free(global_global_cell_id);

  for (unsigned j = 0; j < nbr_points[1]; ++j)
    for (unsigned i = 0; i < nbr_points[0]; ++i)
      local_global_corner_id[j * nbr_points[0] + i] = global_global_corner_id[(j+offset_y) * (NUM_CELLS_X + 3) +i+offset_x];

   free(global_global_corner_id);

  for (unsigned j = 0; j < nbr_points[1]; ++j)
    for (unsigned i = 0; i < nbr_points[0]; ++i)
      local_global_corner_id_rank[j * nbr_points[0] + i] = -1;

  for (unsigned i = 0; i < nbr_points[0]-2; ++i) {
    local_global_corner_id_rank[0 * nbr_points[0] + i] = local_global_cell_id_rank[0 * nbr_cells[0] + i];
    local_global_corner_id_rank[(nbr_points[1]-1) * nbr_points[0] + i] = local_global_cell_id_rank[(nbr_cells[1]-1) * nbr_cells[0] + i];
  }
  for (unsigned j = 0; j < nbr_points[1]-2; ++j) {
    local_global_corner_id_rank[j * nbr_points[0] + 0] = local_global_cell_id_rank[j * nbr_cells[0] + 0];
    local_global_corner_id_rank[j * nbr_points[0] + nbr_points[0]-1] = local_global_cell_id_rank[j * nbr_cells[0] + nbr_cells[0]-1];
  }
  local_global_corner_id_rank[(nbr_points[1]-1)* nbr_points[0] + nbr_points[0]-1] = local_global_cell_id_rank[(nbr_cells[1]-1) * nbr_cells[0] + nbr_cells[0]-1];
  local_global_corner_id_rank[(nbr_points[1]-2)* nbr_points[0] + 0] = local_global_cell_id_rank[(nbr_cells[1]-2) * nbr_cells[0] + 0];
  local_global_corner_id_rank[0 * nbr_points[0] + nbr_points[0]-2] = local_global_cell_id_rank[0 * nbr_cells[0] + nbr_cells[0]-2];
  local_global_corner_id_rank[(nbr_points[1]-2) * nbr_points[0] + nbr_points[0]-1] = local_global_cell_id_rank[(nbr_cells[1]-2) * nbr_cells[0] + nbr_cells[0]-1];
  local_global_corner_id_rank[(nbr_points[1]-1) * nbr_points[0] + nbr_points[0]-2] = local_global_cell_id_rank[(nbr_cells[1]-1) * nbr_cells[0] + nbr_cells[0]-2];

  //---------------------
  // set up global search
  //---------------------

  struct grid * grid =
    yac_reg2d_grid_new(x_vertices, y_vertices, nbr_cells, cyclic);

    struct subdomain subdomain;

  yac_init_subdomain_unstruct(
    &subdomain, grid, global_comm, local_global_corner_id,
    local_global_corner_id_rank, local_global_cell_id,
    local_global_cell_id_rank, NULL, NULL);

  struct domain domain;
  struct subdomain * blocks[1] = {&subdomain};
  yac_init_domain(&domain, global_comm, blocks, 1, "domain");

  // finish domain definition
  yac_enddef_domain(&domain);

  // points
  struct points cell_points;
  yac_init_points(&cell_points, grid, CELL, x_cell_points, y_cell_points);

  // point sets
  struct points * points_per_subdomain[1] = {&cell_points};
  struct points ** points_set[1] = {&(points_per_subdomain[0])};

  //---------------------------------
  // setup n-nearest-neighbour search
  //---------------------------------

  struct nnn_search * search =
    yac_nnn_search_new(
      &domain, "unique_string", &(points_set[0]), NULL, 1, 0);

  //------------------------------
  // do 1-nearest-neighbour search
  //------------------------------

  // search for cell point with global id 20

  int global_ids[] = {20};
  double dx = 360.0 / (double)(NUM_CELLS_X);
  double dy = 180.0 / (double)(NUM_CELLS_Y);
#define NUM_POINTS (sizeof(global_ids)/sizeof(global_ids[0]))
  double coordinates_x[NUM_POINTS];
  double coordinates_y[NUM_POINTS];
  double coordinates_xyz[NUM_POINTS][3];

  for (size_t i = 0; i < NUM_POINTS; ++i) {

    int cell_pos[2] = {global_ids[i]%NUM_CELLS_X, global_ids[i]/NUM_CELLS_X};
    coordinates_x[i] = ((dx * 0.5) + ((double)(cell_pos[0])) * dx) * YAC_RAD;
    coordinates_y[i] = ((dy * 0.5 - 90.0) + ((double)(cell_pos[1])) * dy) * YAC_RAD;
    LLtoXYZ(coordinates_x[i], coordinates_y[i], coordinates_xyz[i]);
  }

  unsigned n = 12;
  unsigned num_points = NUM_POINTS;

  yac_nnn_search_do_search(
    search, n, num_points, coordinates_xyz,
    &(global_ids[0]), (void*)(&num_points), nnn_finish_callback);

  yac_nnn_search_wait(search);

  //---------------
  // cleanup
  //---------------

  yac_nnn_search_delete(search);
  yac_free_domain(&domain);
  yac_free_points(&cell_points);
  yac_delete_grid(grid);
  free(local_global_corner_id_rank);
  free(local_global_corner_id);
  free(local_global_cell_id_rank);
  free(local_global_cell_id);
  free(x_cell_points);
  free(y_cell_points);
  free(global_x_vertices);
  free(global_y_vertices);

  yac_free_communicator(global_comm);
  MPI_Finalize();

  return TEST_EXIT_CODE;
}

static void compute_reg2d_decomp(
  unsigned num_points[2], unsigned total_num_procs, unsigned * num_procs) {

  int flag = num_points[0] > num_points[1];
  double ratio = (flag)?((double)(num_points[0]) / (double)(num_points[1])) :
                ((double)(num_points[1]) / (double)(num_points[1]));

  unsigned prev_num_procs[2];
  double curr_proc_ratio, prev_proc_ratio;

  prev_num_procs[!flag] = total_num_procs;
  prev_num_procs[flag] = 1;

  prev_proc_ratio = (double)total_num_procs;
  curr_proc_ratio = (double)total_num_procs;

  for (unsigned i = total_num_procs-1; i > 0; --i) {

    if (total_num_procs%i == 0) {

      curr_proc_ratio = (double)(i) / (double)(total_num_procs/i);

      double temp[2] = {fabs(ratio - curr_proc_ratio),
                        fabs(ratio - prev_proc_ratio)};
      if (temp[0] < temp[1]) {
        prev_num_procs[!flag] = i;
        prev_num_procs[flag] = total_num_procs/i;
        prev_proc_ratio = curr_proc_ratio;
      }
    }
  }

  num_procs[0] = prev_num_procs[0];
  num_procs[1] = prev_num_procs[1];
}

static void nnn_finish_callback(
  void * user_data, struct request_point * request_points) {

  unsigned num_points = *(unsigned*)user_data;

  unsigned * all_num_result_points =
    malloc(my_size * sizeof(*all_num_result_points));

  for (size_t i = 0; i < num_points; ++i) {

    unsigned num_result_points = request_points[i].num_found_points;

    yac_comm_allgather(
      &num_result_points, 1, all_num_result_points, 1, COMM_UINT, global_comm);

    int * total_result_global_ids;
    double * total_result_distances;
    int test_flag = 1;

    unsigned total_num_result_points = 0;
    for (unsigned j = 0; j < my_size; ++j) {
      if (num_result_points != all_num_result_points[j]) {
        PUT_ERR("wrong number of result points");
        test_flag = 0;
      }
      total_num_result_points += all_num_result_points[j];
    }
    total_result_global_ids =
      malloc(total_num_result_points * sizeof(*total_result_global_ids));
    total_result_distances =
      malloc(total_num_result_points * sizeof(*total_result_distances));

    int * result_global_ids =
      malloc(num_result_points * sizeof(*result_global_ids));
    double * result_distances =
      malloc(num_result_points * sizeof(*result_distances));

    for (unsigned j = 0; j < num_result_points; ++j) {
      result_global_ids[j] = request_points[i].results[j].global_id;
      result_distances[j] = request_points[i].results[j].cos_angle;
    }

    yac_comm_allgatherv(
      result_global_ids, num_result_points, total_result_global_ids,
      all_num_result_points, COMM_INT, global_comm);
    yac_comm_allgatherv(
      result_distances, num_result_points, total_result_distances,
      all_num_result_points, COMM_DBLE, global_comm);

    if (test_flag) {

      for (unsigned j = 0; j < num_result_points; ++j) {

        for (unsigned k = 1; k < my_size; ++k) {

          if (total_result_global_ids[j] !=
              total_result_global_ids[j + k * num_result_points])
            PUT_ERR("wrong result global id");

          if (double_are_unequal(
                total_result_distances[j],
                total_result_distances[j + k * num_result_points]))
            PUT_ERR("wrong result distance");
        }
      }
    }

    free(result_distances);
    free(result_global_ids);
    free(total_result_distances);
    free(total_result_global_ids);
  }

  free(all_num_result_points);

  yac_nnn_search_free_request_points(request_points, num_points);
  free(request_points);
}

