/**
 * @file test_patch_recovery.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "grid.h"
#include "tests.h"
#include "utils.h"
#include "patch_recovery.h"
#include "geometry.h"

int main (void) {

  unsigned num_points;

  unsigned gauss_order;

  double *abscissa;

  double *test_location;
  double *triangle_location; // location in real triangle
  double *quad_location;     // 2D location in real quadrilateral
  double *quad_xyzlocation;   // 3D location in real quadrilateral

  double reference[3];

  struct xyz_triangle triangle;
  struct grid_cell test_cell;

  double const epsilon = 1.0e-14;

  /* gauss integration points for triangles */

#ifdef VERBOSE
  printf ( "\n + + +  gauss integration points for triangles + + + \n\n");
#endif

  for ( gauss_order = 1; gauss_order < 8; ++gauss_order ) {

    abscissa = malloc ( gauss_order * sizeof(*abscissa) );

    yac_linear_gauss_legendre ( gauss_order, abscissa );

#ifdef VERBOSE
    for (unsigned i = 0; i < gauss_order; ++i )
      printf ( "gauss_order = %i : abscissa[%i] = %f \n", gauss_order, i, abscissa[i] );
    printf ( " ----------- \n");
#endif

    free(abscissa);
  }

  /* gauss integration points for quadrilaterals */

  unsigned const dim2 = 2;
  unsigned const dim3 = 3;

#ifdef VERBOSE
  printf ( "\n + + +  gauss integration points for quadrilaterals + + + \n\n");
#endif

  for ( gauss_order = 1; gauss_order < 8; ++gauss_order ) {

    abscissa = malloc ( dim2 * gauss_order * gauss_order * sizeof(*abscissa) );

    yac_quad_2d_locations ( gauss_order, abscissa );

#ifdef VERBOSE
    for (unsigned i = 0; i < gauss_order; ++i )
      for ( int j = 0; j < gauss_order; ++j )
        printf ( "gauss_order = %i : %i %i abscissa = %f %f\n", gauss_order, i, j, abscissa[(i*gauss_order+j)*dim2], abscissa[(i*gauss_order+j)*dim2+1] );
    printf ( " ----------- \n");
#endif
    free(abscissa);
  }

  /* transformation from reference to real world quadrilateral */

  test_cell.num_corners = 4;

  test_cell.coordinates_x = malloc( test_cell.num_corners * sizeof(*test_cell.coordinates_x) );
  test_cell.coordinates_y = malloc( test_cell.num_corners * sizeof(*test_cell.coordinates_y) );

  test_cell.coordinates_x[0] = 0.0*YAC_RAD; test_cell.coordinates_y[0] = 0.0*YAC_RAD;
  test_cell.coordinates_x[1] = 2.0*YAC_RAD; test_cell.coordinates_y[1] = 0.0*YAC_RAD;
  test_cell.coordinates_x[2] = 2.0*YAC_RAD; test_cell.coordinates_y[2] = 2.0*YAC_RAD;
  test_cell.coordinates_x[3] = 0.0*YAC_RAD; test_cell.coordinates_y[3] = 2.0*YAC_RAD;

  gauss_order = 3;
  num_points = gauss_order * gauss_order;

  quad_xyzlocation = malloc ( dim3 * num_points * sizeof(*quad_xyzlocation) );
  quad_location    = malloc ( dim2 * num_points * sizeof(*quad_location) );
  abscissa         = malloc ( dim2 * num_points * sizeof(*abscissa) );

  /* get the reference locations for a particular order */

  yac_quad_2d_locations ( gauss_order, abscissa );

#ifdef VERBOSE
  printf ( "\n + + + transformation from reference to real world quadrilateral + + + \n\n");

  for (unsigned i = 0; i < num_points; ++i )
    printf ( "reference quadrilateral vertices %f %f \n", abscissa[i*dim2], abscissa[i*dim2+1] );
#endif

  /* transformation of reference locations into the test cell
     on the sphere. resulting points are stored in quad_xyzlocations. */

  yac_reference_to_xyz_quadrilateral ( gauss_order, test_cell,
                                       abscissa, quad_xyzlocation );

#ifdef VERBOSE
  for (unsigned i = 0; i < num_points; ++i )
    printf ( "reference in 3D real world %f %f %f\n", quad_xyzlocation[i*dim3], quad_xyzlocation[i*dim3+1] , quad_xyzlocation[i*dim3+2] );
#endif

  /* transformation of reference locations into the test cell
     in 2d lon-lat space. resulting points are stored in quad_locations. */

  for (unsigned i = 0; i < dim2 * num_points; ++i )
     abscissa[i] = 2.0 * abscissa[i] - 1.0;

  test_cell.coordinates_x[0] /= YAC_RAD; test_cell.coordinates_y[0] /= YAC_RAD;
  test_cell.coordinates_x[1] /= YAC_RAD; test_cell.coordinates_y[1] /= YAC_RAD;
  test_cell.coordinates_x[2] /= YAC_RAD; test_cell.coordinates_y[2] /= YAC_RAD;
  test_cell.coordinates_x[3] /= YAC_RAD; test_cell.coordinates_y[3] /= YAC_RAD;

  yac_reference_to_xy_quadrilateral ( num_points, test_cell, abscissa, quad_location );

#ifdef VERBOSE
  for (unsigned i = 0; i < num_points; ++i )
    printf ( "reference in real world %f %f \n", quad_location[i*dim2], quad_location[i*dim2+1] );
#endif

  free (abscissa);
  free (quad_location);
  free (quad_xyzlocation);

  free (test_cell.coordinates_x);
  free (test_cell.coordinates_y);

  /* some tests for triangle functions */

#ifdef VERBOSE
    printf ( "\n + + + tri + + + tri + + + tri + + + \n\n");
#endif

  for ( gauss_order = 1; gauss_order < 8; ++gauss_order ) {

    num_points = yac_get_number_of_quadrature_points (gauss_order);

    abscissa = malloc ( 2* num_points * sizeof(*abscissa) );

    yac_triangle_gauss_legendre ( gauss_order, abscissa );

#ifdef VERBOSE
    for (unsigned i = 0; i < num_points; ++i )
      printf ( "gauss_order = %i : abscissa %f %f\n", gauss_order,
                 abscissa[2*i], abscissa[2*i+1] );
    printf ( " ----------- \n");
#endif

    free(abscissa);
  }

  /* transformations between reference and real world triangle */

  test_cell.coordinates_x = malloc( test_cell.num_corners * sizeof(*test_cell.coordinates_x) );
  test_cell.coordinates_y = malloc( test_cell.num_corners * sizeof(*test_cell.coordinates_y) );

  test_cell.coordinates_x[0] = 0.0*YAC_RAD; test_cell.coordinates_y[0] = 0.0*YAC_RAD;
  test_cell.coordinates_x[1] = 0.0*YAC_RAD; test_cell.coordinates_y[1] = 2.0*YAC_RAD;
  test_cell.coordinates_x[2] = 3.0*YAC_RAD; test_cell.coordinates_y[2] = 4.0*YAC_RAD;

  LLtoXYZ ( 0.0*YAC_RAD, 0.0*YAC_RAD, triangle.a );
  LLtoXYZ ( 0.0*YAC_RAD, 2.0*YAC_RAD, triangle.b );
  LLtoXYZ ( 3.0*YAC_RAD, 4.0*YAC_RAD, triangle.c );

  gauss_order = 3;

  num_points = yac_get_number_of_quadrature_points (gauss_order);

  triangle_location = malloc ( dim3 * num_points * sizeof(*triangle_location) );
  abscissa          = malloc ( dim3 * num_points * sizeof(*abscissa) );
  test_location     = malloc ( dim3 * num_points * sizeof(*test_location) );

  yac_triangle_gauss_legendre ( gauss_order, abscissa );

  double * abscissa_3d = malloc ( 3 * num_points * sizeof(*abscissa_3d) );

  for (unsigned i = 0; i < num_points; ++ i ) {
     abscissa_3d[i*3+0] = abscissa[i*2+0];
     abscissa_3d[i*3+1] = abscissa[i*2+1];
     abscissa_3d[i*3+2] = 0.0;
  }

#ifdef VERBOSE
  for (unsigned i = 0; i < num_points; ++i )
    printf ( "locations in reference %f %f %f\n", abscissa_3d[i*dim3], abscissa_3d[i*dim3+1], abscissa_3d[i*dim3+2] );
#endif

  yac_reference_to_xyz_triangle ( num_points, triangle, abscissa_3d, triangle_location );

#ifdef VERBOSE
    printf ( " ... should be identical to \n" );
#endif

  yac_xyz_triangle_to_reference ( num_points, triangle, test_location, triangle_location );

#ifdef VERBOSE
  for (unsigned i = 0; i < num_points; ++i )
    printf ( "locations in reference %f %f %f\n", test_location[i*dim3], test_location[i*dim3+1], test_location[i*dim3+2] );
#endif

  for (unsigned i = 0; i < dim3*num_points; ++i )
    if ( fabs(abscissa_3d[i]-test_location[i]) > epsilon )
      PUT_ERR("error in forward and backward transformation \n");

  free (test_location);
  free (abscissa);
  free (abscissa_3d);
  free (triangle_location);
  free (test_cell.coordinates_x);
  free (test_cell.coordinates_y);

  /* patch recovery test
     -------------------

   Tasks for patch recovery

   - get all elements needed for the current patch
   - (optional) get coordinates of Gauss integration points for each element in the patch (-> goes into P)
   - (optional) get the function values at each Gauss point in the current patch (-> part of b)
   - (optional) get coordinates of nodes for each element in the patch (-> goes into P)
   - (optional) get the value at each node in the current patch (-> part of b)
   - determine P by inserting (x,y) into the different contributions for P

   Define a patch of triangles 


     |     Element numbers are given in the middle of the triangles
     |     Node number are given at each node
     | 
     6                     10________11
     |                    /\        /\
     |                   /  \  11  /  \
     |                  /    \    /    \
     |                 /  10  \  /  12  \
     4                7________\8________9
     |               /\        /\        /\
     |              /  \  6   /  \  8   /  \
     |             /    \    /    \    /    \
     |            /  5   \  /  7   \  /  9   \
     2           3________\4________\5________6
     |           \        /\        /\        /
     |            \  0   /  \  2   /  \  4   /
     |             \    /    \    /    \    /
     |              \  /  1   \  /  3   \  /
     0               \0________\1________\2
     |
     |_________________________________________________  
                0         2         4          6

     Element number 7 is the center element around which
     the test patch is constructed.

     Elements 1, 0, 5         with value 2.0
     Elements 2, 3, 6,  7, 10 with value 3.0
     Elements 4, 8, 9, 11, 12 with value 4.0

     Alternative: constant value of 2.0 for each element

*/

#define NUM_CELLS 13

  double fscale = 0.1; // scaling of field values
  double cscale = 1.0; // scaling of coordinate space

  unsigned num_sampling_points;

  double node_coordinates[2*(NUM_CELLS-1)] = { 1.0, 0.0, 3.0, 0.0, 5.0, 0.0,
                                               0.0, 2.0, 2.0, 2.0, 4.0, 2.0, 6.0, 2.0, 
                                               1.0, 4.0, 3.0, 4.0, 5.0, 4.0,
                                               2.0, 6.0, 4.0, 6.0 };

  unsigned connectivity[3*NUM_CELLS] = { 0,  4,  3,
                                         0,  1,  4,
                                         1,  5,  4,
                                         1,  2,  5,
                                         2,  6,  5,
                                         3,  4,  7,
                                         4,  8,  7,
                                         4,  5,  8,
                                         5,  9,  8,
                                         5,  6,  9,
                                         7,  8, 10,
                                         8, 11, 10,
                                         8,  9, 11 };

  /* Define values at the center of the triangles */

  double cell_value[NUM_CELLS] = { 2.0, 2.0, 3.0, 3.0, 4.0,
                                   2.0, 3.0, 3.0, 4.0, 4.0,
                                   3.0, 4.0, 4.0 };

  /* double cell_value[NUM_CELLS] = { 2.0, 2.0, 2.0, 2.0, 2.0,
                                  2.0, 2.0, 2.0, 2.0, 2.0,
                                  2.0, 2.0, 2.0 }; */

  struct xyz_triangle cell[NUM_CELLS];

  for (unsigned i = 0; i < 2*(NUM_CELLS-1); ++i ) node_coordinates[i] *= cscale;
  for (unsigned i = 0; i < NUM_CELLS; ++i ) cell_value[i] *= fscale;

  for (unsigned i = 0; i < NUM_CELLS; ++i) {
     for (unsigned n = 0; n < 3; ++n) {

        int ci = connectivity[i*3+n];

        if ( n == 0 )
           LLtoXYZ ( node_coordinates[ci*2+0]*YAC_RAD, node_coordinates[ci*2+1]*YAC_RAD, cell[i].a );
        if ( n == 1 )
           LLtoXYZ ( node_coordinates[ci*2+0]*YAC_RAD, node_coordinates[ci*2+1]*YAC_RAD, cell[i].b );
        if ( n == 2 )
           LLtoXYZ ( node_coordinates[ci*2+0]*YAC_RAD, node_coordinates[ci*2+1]*YAC_RAD, cell[i].c );
     }
  }

  double * target_location = malloc ( 3 * sizeof(*target_location) );

  target_location[0] = cscale * 3.0;
  target_location[1] = cscale * 2.0;
#ifdef VERBOSE
  unsigned gauss_order_min = 0;
  unsigned gauss_order_max = 0;

  unsigned poly_order_min = 0;
  unsigned poly_order_max = 0;
#endif
  double test;
  double error_min;
  double error_max;
  double error_rel;

  unsigned polynom_order;

  /* Get Gauss integration points on the reference triangle and transform them to the patch elements */


#ifdef VERBOSE
  printf ("\n + + + + + + + Patch Recovery Test in 3D + + + + + + + + \n\n");
#endif

  /* Set the centre of mass of patch centre triangle as target location in 3D Cartesian*/

  double test1[3] = {0.33333, 0.33333, 0.0};
  yac_reference_to_xyz_triangle ( 1, cell[7], test1, reference );

  /* Set location of target in 3D Cartesian */

  LLtoXYZ( target_location[0]*YAC_RAD, target_location[1]*YAC_RAD, target_location );

  // for (unsigned i = 0; i < 3; ++i ) reference[i] = target_location[i];

  error_min =  9999999.9;
  error_max = -9999999.9;

  for ( gauss_order = 1; gauss_order < 8; ++gauss_order ) {
    for ( polynom_order = 0; polynom_order < 4; ++polynom_order ) {

      num_points = yac_get_number_of_quadrature_points (gauss_order);

      if ( gauss_order > 0 ) {

        abscissa = malloc ( 2 * num_points * sizeof(*abscissa) );

        num_sampling_points = NUM_CELLS * num_points;

        yac_triangle_gauss_legendre ( gauss_order, abscissa );
      }

      else

         num_sampling_points = NUM_CELLS-1;

      /* Get the gauss integration points (triangle_location)
         and values (gauss_values) at each integration point
         for each cell in the patch.

         For simplicity we first assume that we have a P0 element,
         i.e. the value is constant of the whole element. */

      double *gauss_values = malloc ( num_sampling_points * sizeof(*gauss_values) );

      triangle_location = malloc ( 3 * num_sampling_points * sizeof(*triangle_location) );

      if ( gauss_order == 0 ) {
         // for more clarity this should better be the cell centers
        for (unsigned i = 0; i < num_sampling_points; ++ i ) {
          LLtoXYZ( node_coordinates[i*2]*YAC_RAD, node_coordinates[i*2+1]*YAC_RAD, &triangle_location[i*3] );
          gauss_values[i] = cell_value[i];
        }
      }

      else {

        abscissa_3d = malloc ( 3 * num_points * sizeof(*abscissa_3d) );

        for (unsigned i = 0; i < num_points; ++ i ) {
           abscissa_3d[i*dim3+0] = abscissa[i*dim2+0];
           abscissa_3d[i*dim3+1] = abscissa[i*dim2+1];
           abscissa_3d[i*dim3+2] = 0.0;
        }

        for (unsigned i = 0; i < NUM_CELLS; ++ i ) {
          yac_reference_to_xyz_triangle ( num_points, cell[i], abscissa_3d, &triangle_location[i*3*num_points] );
          for (unsigned n = 0; n < num_points; ++n)
            gauss_values[i*num_points+n] = cell_value[i];
        }
        free (abscissa);
        free (abscissa_3d);
      }
      /* Get the value for the target points via polynomial fit */

      unsigned num_coeff = yac_get_num_polynomial_coefficients ( polynom_order );

      /* a sufficient number of sampling points is required to get reasonable results */

      if ( num_sampling_points >= 2 * num_coeff ) {

         /* 1st) Determine Polynomial terms P for each integration point */

         double *PT = malloc ( num_sampling_points*num_coeff * sizeof(*PT) );

         yac_set_polynomial_terms ( polynom_order, num_sampling_points, triangle_location, reference, PT );

         /* 2nd) Get the polynomial for the target point */

         double *target_pterm = malloc ( num_coeff * sizeof(*target_pterm) );

         yac_set_polynomial_terms ( polynom_order, 1, target_location, target_location, target_pterm );

         /* 3rd) Get the new value at the target point */

         double *weights = malloc ( NUM_CELLS * sizeof(*weights) );

         unsigned *num_gausspoints_per_cell = malloc ( NUM_CELLS * sizeof(*num_gausspoints_per_cell) );

         for (unsigned n = 0; n < NUM_CELLS; ++n)
            num_gausspoints_per_cell[n] = num_points;

         yac_patch_P0weights_opt ( num_sampling_points, num_coeff, NUM_CELLS, num_gausspoints_per_cell, PT, weights );

         free(num_gausspoints_per_cell);

         double value = yac_patch_P0value ( NUM_CELLS, weights, cell_value );

         error_rel = fabs(value-cell_value[7]) / cell_value[7] * 100.0;
#ifdef VERBOSE
         printf ("Gauss order %i, Polynom order %i : rel. error %f percent\n", gauss_order, polynom_order, error_rel );
#endif

         double value_alt1 = yac_patch_value_alt1 ( num_sampling_points, num_coeff, target_pterm, PT, gauss_values );

         error_rel = fabs(value_alt1-cell_value[7]) / cell_value[7] * 100.0;
#ifdef VERBOSE
         printf ("Gauss order %i, Polynom order %i : rel. error %f percent\n", gauss_order, polynom_order, error_rel );
#endif

         /* Alternative method
            ------------------

            Determine unknown coefficients a of P

            P is fixed over time and can be stored for each patch as
            long as the grid does not change during the time of integration.

         The gauss_values are the current values of the interpolation
         source field at a particular time step and does change with time!

         So do the a's.

         During the interpolation this is what we have to do:

         1st) Get the new a's

         Attention: P and gauss_values get overwritten! */

         double *a = malloc ( num_coeff * sizeof(*a) );

         yac_get_polynomial_coefficients ( num_sampling_points, num_coeff, PT, gauss_values, a );

         double value_alt2 = yac_patch_value_alt2 ( num_coeff, target_pterm, a );

         /* Check the error */

         error_rel = fabs(value_alt2-cell_value[7]) / cell_value[7] * 100.0;
#ifdef VERBOSE
         printf ("Gauss order %i, Polynom order %i : rel. error %f percent\n", gauss_order, polynom_order, error_rel );
#endif
         test = MIN(error_min,error_rel) ;

         if ( test < error_min ) {
            error_min = test;
#ifdef VERBOSE
            gauss_order_min = gauss_order;
            poly_order_min  = polynom_order;
#endif
         }

         test = MAX(error_max,error_rel);

         if ( test > error_max ) {
            error_max = test;
#ifdef VERBOSE
            gauss_order_max = gauss_order;
            poly_order_max  = polynom_order;
#endif
         }

         if ( error_rel > 100.0 ) {
            for (unsigned i = 0; i < num_coeff; ++i )
               printf ( "Polynom coefficients Pa[%u] = %f * %f \n", i, target_pterm[i], a[i] );
            printf ("value at test point %f %f %f is %f\n", target_location[0], target_location[1], target_location[2], value );
            PUT_ERR("error in patch recovery, rel error too large \n");
         }
         free(target_pterm);
         free(weights);
         free(PT);
         free(a);
      }
      free(gauss_values);
      free(triangle_location);
    }
  }

#ifdef VERBOSE
  printf ("maximum rel. error: %f percent with gauss order %i and polynom order %i\n", error_max, gauss_order_max, poly_order_max );
  printf ("minimum rel. error: %f percent with gauss order %i and polynom order %i\n", error_min, gauss_order_min, poly_order_min );
#endif

  free(target_location);

  return TEST_EXIT_CODE;
}

