/**
 * @file test_point_in_cell.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include "tests.h"
#include "search.h"
#include "test_common.h"

int main (void) {

   double point_3d[3];
   struct grid_cell cell;

   { // simple inside test

      double x_coords[] = {-1,1,1,-1};
      double y_coords[] = {-1,-1,1,1};

      for ( unsigned i = 0; i < 4; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[i] *= YAC_RAD;
      }

      enum yac_edge_type edge_type[] = {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE};

      LLtoXYZ(0, 0, point_3d);
      cell.coordinates_x = x_coords;
      cell.coordinates_y = y_coords;
      cell.coordinates_xyz = (double[4][3]){{-1}};
      cell.edge_type = edge_type;
      cell.num_corners = 4;

      generate_XYZ(cell);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (simple inside test)\n")
   }

   { // simple outside test

      double x_coords[] = {-1,1,1,-1};
      double y_coords[] = {-1,-1,1,1};

      for ( unsigned i = 0; i < 4; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[i] *= YAC_RAD;
      }

      enum yac_edge_type edge_type[] = {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE};

      LLtoXYZ(0, 1.5*YAC_RAD, point_3d);
      cell.coordinates_x = x_coords;
      cell.coordinates_y = y_coords;
      cell.coordinates_xyz = (double[4][3]){{-1}};
      cell.edge_type = edge_type;
      cell.num_corners = 4;

      generate_XYZ(cell);

      if (yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (simple outside test)\n")
   }

   { // test point on edge

      double x_coords[] = {-1,1,1,-1};
      double y_coords[] = {-1,-1,1,1};

      for ( unsigned i = 0; i < 4; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[i] *= YAC_RAD;
      }

      enum yac_edge_type edge_type[] = {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE};

      LLtoXYZ(0, 1*YAC_RAD, point_3d);
      cell.coordinates_x = x_coords;
      cell.coordinates_y = y_coords;
      cell.coordinates_xyz = (double[4][3]){{-1}};
      cell.edge_type = edge_type;
      cell.num_corners = 4;

      generate_XYZ(cell);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (test point on edge)\n")
   }

   { // test near the pole

      double x_coords[] = {0,45,45,0};
      double y_coords[] = {85,85,88,88};

      for ( unsigned i = 0; i < 4; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[i] *= YAC_RAD;
      }

      enum yac_edge_type edge_type[] = {LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE};

      LLtoXYZ(22.5*YAC_RAD, 85.1*YAC_RAD, point_3d);
      cell.coordinates_x = x_coords;
      cell.coordinates_y = y_coords;
      cell.coordinates_xyz = (double[4][3]){{-1}};
      cell.edge_type = edge_type;
      cell.num_corners = 4;

      generate_XYZ(cell);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (lon-lat cell near the pole)\n")
   }

   { // test near the pole

      double x_coords[] = {0,45,45,0};
      double y_coords[] = {85,85,88,88};

      for ( unsigned i = 0; i < 4; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[i] *= YAC_RAD;
      }

      enum yac_edge_type edge_type[] = {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE};

      LLtoXYZ(22.5*YAC_RAD, 85.1*YAC_RAD, point_3d);
      cell.coordinates_x = x_coords;
      cell.coordinates_y = y_coords;
      cell.coordinates_xyz = (double[4][3]){{-1}};
      cell.edge_type = edge_type;
      cell.num_corners = 4;

      generate_XYZ(cell);

      if (yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (great circle cell near the pole)\n")
   }

   { // simple test at the pole

      double x_coords[] = {0,90,180,270};
      double y_coords[] = {88,88,88,88};

      for ( unsigned i = 0; i < 4; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[i] *= YAC_RAD;
      }

      enum yac_edge_type edge_type[] = {LAT_CIRCLE, LAT_CIRCLE, LAT_CIRCLE, LAT_CIRCLE};

      LLtoXYZ(45*YAC_RAD, 88.3*YAC_RAD, point_3d);
      cell.coordinates_x = x_coords;
      cell.coordinates_y = y_coords;
      cell.coordinates_xyz = (double[4][3]){{-1}};
      cell.edge_type = edge_type;
      cell.num_corners = 4;

      generate_XYZ(cell);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (lon-lat cell on the pole)\n")
   }

   { // simple test at the pole

      double x_coords[] = {0,90,180,270};
      double y_coords[] = {88,88,88,88};

      for ( unsigned i = 0; i < 4; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[i] *= YAC_RAD;
      }

      enum yac_edge_type edge_type[] = {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE};

      LLtoXYZ(45*YAC_RAD, 88.3*YAC_RAD, point_3d);
      cell.coordinates_x = x_coords;
      cell.coordinates_y = y_coords;
      cell.coordinates_xyz = (double[4][3]){{-1}};
      cell.edge_type = edge_type;
      cell.num_corners = 4;

      generate_XYZ(cell);

      if (yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (great circle cell on the pole)\n")
   }

   { // test point on cell corner and edge

      double x_coords[] = {0,1,1,0};
      double y_coords[] = {0,0,1,1};

      for ( unsigned i = 0; i < 4; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[i] *= YAC_RAD;
      }

      enum yac_edge_type edge_type[] = {LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE};

      LLtoXYZ(1*YAC_RAD, 1*YAC_RAD, point_3d);
      cell.coordinates_x = x_coords;
      cell.coordinates_y = y_coords;
      cell.coordinates_xyz = (double[4][3]){{-1}};
      cell.edge_type = edge_type;
      cell.num_corners = 4;

      generate_XYZ(cell);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (point on corner of lon-lat cell)\n")

      LLtoXYZ(0.5*YAC_RAD, 0, point_3d);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (point on edge of lon-lat cell)\n")
   }

   { // test point on cell corner

      double x_coords[] = {0,1,0,-1};
      double y_coords[] = {0,1,2,1};

      for ( unsigned i = 0; i < 4; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[i] *= YAC_RAD;
      }

      enum yac_edge_type edge_type[] = {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE};

      LLtoXYZ(1*YAC_RAD, 1*YAC_RAD, point_3d);
      cell.coordinates_x = x_coords;
      cell.coordinates_y = y_coords;
      cell.coordinates_xyz = (double[4][3]){{-1}};
      cell.edge_type = edge_type;
      cell.num_corners = 4;

      generate_XYZ(cell);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (point on corner of great circle cell)\n")

      LLtoXYZ(0, 2*YAC_RAD, point_3d);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (point on corner of great circle cell)\n")

      LLtoXYZ(0, 1*YAC_RAD, point_3d);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (point in cell)\n")

      LLtoXYZ(0, 3*YAC_RAD, point_3d);

      if (yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (point in cell)\n")
   }

   { // test point outside of cell

      double x_coords[] = {0,1,1,0};
      double y_coords[] = {0,0,1,1};

      for ( unsigned i = 0; i < 4; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[i] *= YAC_RAD;
      }

      enum yac_edge_type edge_type[] = {LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE};

      LLtoXYZ(1.5*YAC_RAD, 0.5*YAC_RAD, point_3d);
      cell.coordinates_x = x_coords;
      cell.coordinates_y = y_coords;
      cell.coordinates_xyz = (double[4][3]){{-1}};
      cell.edge_type = edge_type;
      cell.num_corners = 4;

      generate_XYZ(cell);

      if (yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell\n")
   }

   { // test point on longitude edge of cell

      double x_coords[] = {0,0,5};
      double y_coords[] = {40,50,45};

      for ( unsigned i = 0; i < 3; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[i] *= YAC_RAD;
      }

      for (unsigned i = 0; i < 2; ++i) {
         for (int j = 0; j < 3; ++j) {
            enum yac_edge_type edge_type[2][3] =
               {{GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE},
                {LON_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE}};

            double point_lat[3] = {40, 45, 50};

            LLtoXYZ(0.0*YAC_RAD, point_lat[j]*YAC_RAD, point_3d);
            cell.coordinates_x = x_coords;
            cell.coordinates_y = y_coords;
            cell.coordinates_xyz = (double[3][3]){{-1}};
            cell.edge_type = edge_type[i];
            cell.num_corners = 3;

            generate_XYZ(cell);

            if (!yac_point_in_cell(point_3d, cell))
               PUT_ERR("error in point_in_cell\n")
         }
      }
   }

   { // test point on latitude edge of cell

      double x_coords[] = {0,10,0};
      double y_coords[2][3] = {{10,10,15}, {10,10,5}};

      for ( unsigned i = 0; i < 3; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[0][i] *= YAC_RAD;
         y_coords[1][i] *= YAC_RAD;
      }

      for (unsigned i = 0; i < 2; ++i) {
         enum yac_edge_type edge_type[3] = {LAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE};

         LLtoXYZ(5.0*YAC_RAD, 10*YAC_RAD, point_3d);
         cell.coordinates_x = x_coords;
         cell.coordinates_y = y_coords[i];
         cell.coordinates_xyz = (double[3][3]){{-1}};
         cell.edge_type = edge_type;
         cell.num_corners = 3;

         generate_XYZ(cell);

         if (!yac_point_in_cell(point_3d, cell))
            PUT_ERR("error in point_in_cell\n")
      }
   }

   { // test point equator edge of cell

      double x_coords[] = {0,10,0};
      double y_coords[2][3] = {{0,0,5}, {0,0,-5}};

      for ( unsigned i = 0; i < 3; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[0][i] *= YAC_RAD;
         y_coords[1][i] *= YAC_RAD;
      }

      for (unsigned i = 0; i < 2; ++i) {
         enum yac_edge_type edge_type[3] =
            {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE};

         LLtoXYZ(5.0*YAC_RAD, 0*YAC_RAD, point_3d);
         cell.coordinates_x = x_coords;
         cell.coordinates_y = y_coords[i];
         cell.coordinates_xyz = (double[3][3]){{-1}};
         cell.edge_type = edge_type;
         cell.num_corners = 3;

         generate_XYZ(cell);

         if (!yac_point_in_cell(point_3d, cell))
            PUT_ERR("error in point_in_cell\n")
      }
   }

   { // test point outside of cell

      double x_coords[] = {0,0,5};
      double y_coords[3] = {5,10,20};

      for ( unsigned i = 0; i < 3; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[i] *= YAC_RAD;
      }

      for (unsigned i = 0; i < 2; ++i) {
         enum yac_edge_type edge_type[2][3] =
            {{LON_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE},
             {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE}};

         LLtoXYZ(0.0*YAC_RAD, 11*YAC_RAD, point_3d);
         cell.coordinates_x = x_coords;
         cell.coordinates_y = y_coords;
         cell.coordinates_xyz = (double[3][3]){{-1}};
         cell.edge_type = edge_type[i];
         cell.num_corners = 3;

         generate_XYZ(cell);

         if (yac_point_in_cell(point_3d, cell))
            PUT_ERR("error in point_in_cell\n")
      }
   }

   { // test point inside of concave cell

      double x_coords[] = {0, 0, -5, 5};
      double y_coords[] = {0, 5, 10, 10};

      for ( unsigned i = 0; i < 4; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[i] *= YAC_RAD;
      }

      for (unsigned i = 0; i < 2; ++i) {
         enum yac_edge_type edge_type[2][4] =
            {{LON_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE},
             {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE}};

         LLtoXYZ(0.0*YAC_RAD, 7.0*YAC_RAD, point_3d);
         cell.coordinates_x = x_coords;
         cell.coordinates_y = y_coords;
         cell.coordinates_xyz = (double[4][3]){{-1}};
         cell.edge_type = edge_type[i];
         cell.num_corners = 4;

         generate_XYZ(cell);

         if (!yac_point_in_cell(point_3d, cell))
            PUT_ERR("error in point_in_cell\n")
      }
   }

   { // test point outside of concave cell

      double x_coords[] = {0, 0, -5, 5};
      double y_coords[] = {0, 5, 10, 10};

      for ( unsigned i = 0; i < 4; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[i] *= YAC_RAD;
      }

      for (unsigned i = 0; i < 2; ++i) {
         enum yac_edge_type edge_type[2][4] =
            {{LON_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE},
             {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE}};

         LLtoXYZ(-2.5*YAC_RAD, 6.0*YAC_RAD, point_3d);
         cell.coordinates_x = x_coords;
         cell.coordinates_y = y_coords;
         cell.coordinates_xyz = (double[4][3]){{-1}};
         cell.edge_type = edge_type[i];
         cell.num_corners = 4;

         generate_XYZ(cell);

         if (yac_point_in_cell(point_3d, cell))
            PUT_ERR("error in point_in_cell\n")
      }
   }

   { // from Renes clipping tests

      double x_coords[] = { 0.0, 0.5, 0.0,-0.5};
      double y_coords[] = {-0.5, 0.0, 0.5, 0.0};

      for ( unsigned i = 0; i < 4; ++i ) {
         x_coords[i] *= YAC_RAD;
         y_coords[i] *= YAC_RAD;
      }

      enum yac_edge_type edge_type[] = {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE};

      cell.coordinates_x = x_coords;
      cell.coordinates_y = y_coords;
      cell.coordinates_xyz = (double[4][3]){{-1}};
      cell.edge_type = edge_type;
      cell.num_corners = 4;

      generate_XYZ(cell);

      for (unsigned i = 0; i < 4; ++i) {
         LLtoXYZ(x_coords[i], y_coords[i], point_3d);
         if (!yac_point_in_cell(point_3d, cell))
            PUT_ERR("error in point_in_cell\n")
      }
   }

   { // example from toy

      double x_coords[] = {-2.4523357834544868,
                           -2.5132741228718340,
                           -2.5742124622891822};
      double y_coords[] = {-0.8955349680874743,
                           -0.8292526951666540,
                           -0.8955349680874748};

      enum yac_edge_type edge_type[] = {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE};

      cell.coordinates_x = x_coords;
      cell.coordinates_y = y_coords;
      cell.coordinates_xyz = (double[3][3]){{-1}};
      cell.edge_type = edge_type;
      cell.num_corners = 3;

      generate_XYZ(cell);

      LLtoXYZ(-2.5132741229357842, -0.8511807007280388, point_3d);
      if (!yac_point_in_cell(point_3d, cell))
            PUT_ERR("error in point_in_cell\n");
   }

   return TEST_EXIT_CODE;
}

