//#define VERBOSE

/**
 * @file test_clipping.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "grid.h"
#include "clipping.h"
#include "geometry.h"
#include "tests.h"
#include "area.h"
#include "test_common.h"

static void test_clipping(struct grid_cell cells[2],
                          struct grid_cell * ref_cells,
                          unsigned num_ref_cells);

static int compare_cells(struct grid_cell a, struct grid_cell b);

int main (void) {

  enum yac_edge_type gc_edges[] = {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE,
                                   GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE};
  enum yac_edge_type latlon_edges[] = {LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE};

  struct grid_cell overlap_cell[2];

  yac_init_grid_cell(overlap_cell);
  yac_init_grid_cell(overlap_cell+1);

#ifdef VERBOSE
  printf ("\n 1.) Simple test case with two quadrangle\n\n");
#endif
  /* 1.) Simple test case with two quadrangle

         source cell test data

              0.0 (lon) [2]
              0.7 (lat)
              / \
             /   \
   -0.5 [1] /     \  0.5  [3]
    0.0     \     /  0.0
             \   /
              \ /
              0.0  [0]
             -0.5         
  */

  /* target cell test data

            0.0 (lon)  [0]
            0.6 (lat)
             / \
            /   \
  -0.6 [3] /     \  0.6  [1]
   0.0     \     /  0.0
            \   /
             \ /
             0.0  [2]
            -0.6  
  */

  {
    struct point intersection;

    if (!yac_intersect((struct edge){.points[0] = {.lon = 0.0*YAC_RAD,
                                                  .lat = 0.6*YAC_RAD},
                                     .points[1] = {.lon = 0.6*YAC_RAD,
                                                   .lat = 0.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 0.0*YAC_RAD,
                                                   .lat = 0.7*YAC_RAD},
                                     .points[1] = {.lon = 0.5*YAC_RAD,
                                                   .lat = 0.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                         &intersection))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                          0.5*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                         -0.5*YAC_RAD},
                             .coordinates_y = (double[]){-0.5*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                          0.7*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                          0.6*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                         -0.6*YAC_RAD},
                             .coordinates_y = (double[]){ 0.6*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                         -0.6*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                        intersection.lon,
                                                         0.5*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                        -0.5*YAC_RAD,
                                                        -intersection.lon},
                            .coordinates_y = (double[]){ 0.6*YAC_RAD,
                                                        intersection.lat,
                                                         0.0*YAC_RAD,
                                                        -0.5*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                        intersection.lat},
                            .coordinates_xyz = (double[6][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 6}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 2.a) Simple test case with one quadrangle and one triangle \n\n");
#endif
  /* 2.a) Simple test case with one quadrangle and one triangle

         source cell test data

              0.0 (lon) [2]
              0.7 (lat)
              / \
             /   \
   -0.5 [1] /     \  0.5  [3]
    0.0     \     /  0.0
             \   /
              \ /
              0.0  [0]
             -0.5         
  */

  /* target cell test data

            0.0 (lon)
            0.6 (lat)
             / \
            /   \
  -0.6     /     \   0.6
   0.0    /_______\  0.0

  */

  {
    struct point intersection;

    if (!yac_intersect((struct edge){.points[0] = {.lon = 0.0*YAC_RAD,
                                                   .lat = 0.6*YAC_RAD},
                                     .points[1] = {.lon = 0.6*YAC_RAD,
                                                   .lat = 0.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 0.0*YAC_RAD,
                                                   .lat = 0.7*YAC_RAD},
                                     .points[1] = {.lon = 0.5*YAC_RAD,
                                                   .lat = 0.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                         &intersection))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                         -0.5*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                          0.5*YAC_RAD},
                             .coordinates_y = (double[]){-0.5*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                          0.7*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){ 0.6*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                         -0.6*YAC_RAD},
                             .coordinates_y = (double[]){ 0.0*YAC_RAD,
                                                          0.6*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                        intersection.lon,
                                                         0.5*YAC_RAD,
                                                        -0.5*YAC_RAD,
                                                        -intersection.lon},
                            .coordinates_y = (double[]){ 0.6*YAC_RAD,
                                                        intersection.lat,
                                                         0.0*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                        intersection.lat},
                            .coordinates_xyz = (double[5][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 5}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 2.b) Simple test case with one quadrangle and one triangle \n\n");
#endif
  /* 2.b) Simple test case with one quadrangle and one triangle

         source cell test data

              0.0 (lon) [2]
              0.7 (lat)
              / \
             /   \
   -0.5 [1] /     \  0.5  [3]
    0.0     \     /  0.0
             \   /
              \ /
              0.0  [0]
             -0.5         
  */

  /* target cell test data

            0.6 (lon)
            0.6 (lat)
             / \
            /   \
   0.0     /     \   1.0
   0.1    /_______\  0.1

  */

  {
    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon = 0.0*YAC_RAD,
                                                   .lat = 0.1*YAC_RAD},
                                     .points[1] = {.lon = 1.0*YAC_RAD,
                                                   .lat = 0.1*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 0.0*YAC_RAD,
                                                   .lat = 0.7*YAC_RAD},
                                     .points[1] = {.lon = 0.5*YAC_RAD,
                                                   .lat = 0.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                         intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = 0.0*YAC_RAD,
                                                   .lat = 0.1*YAC_RAD},
                                     .points[1] = {.lon = 0.6*YAC_RAD,
                                                   .lat = 0.6*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 0.0*YAC_RAD,
                                                   .lat = 0.7*YAC_RAD},
                                     .points[1] = {.lon = 0.5*YAC_RAD,
                                                   .lat = 0.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                         intersections+1))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                         -0.5*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                          0.5*YAC_RAD},
                             .coordinates_y = (double[]){-0.5*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                          0.7*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){1.0*YAC_RAD,
                                                         0.6*YAC_RAD,
                                                         0.0*YAC_RAD},
                             .coordinates_y = (double[]){0.1*YAC_RAD,
                                                         0.6*YAC_RAD,
                                                         0.1*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){0.0*YAC_RAD,
                                                        intersections[0].lon,
                                                        intersections[1].lon},
                            .coordinates_y = (double[]){0.1*YAC_RAD,
                                                        intersections[0].lat,
                                                        intersections[1].lat},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 3}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 2.c) Simple test case with one quadrangle and one triangle \n\n");
#endif
  /* 2.c) Simple test case with one quadrangle and one triangle

         source cell test data

              0.0 (lon) [2]
              0.7 (lat)
              / \
             /   \
   -0.5 [1] /     \  0.5  [3]
    0.0     \     /  0.0
             \   /
              \ /
              0.0  [0]
             -0.5         
  */

  /* target cell test data

            0.0 (lon)
            0.6 (lat)
             / \
            /   \
  -0.5     /     \   0.5
   0.1    /_______\  0.1

  */

  {
    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon = 0.0*YAC_RAD,
                                                   .lat = 0.6*YAC_RAD},
                                     .points[1] = {.lon = 0.5*YAC_RAD,
                                                   .lat = 0.1*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 0.0*YAC_RAD,
                                                   .lat = 0.7*YAC_RAD},
                                     .points[1] = {.lon = 0.5*YAC_RAD,
                                                   .lat = 0.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                     intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = -0.5*YAC_RAD,
                                                   .lat =  0.1*YAC_RAD},
                                     .points[1] = {.lon =  0.5*YAC_RAD,
                                                   .lat =  0.1*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 0.0*YAC_RAD,
                                                   .lat = 0.7*YAC_RAD},
                                     .points[1] = {.lon = 0.5*YAC_RAD,
                                                   .lat = 0.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                         intersections+1))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                         -0.5*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                          0.5*YAC_RAD},
                             .coordinates_y = (double[]){-0.5*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                          0.7*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){ 0.5*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                         -0.5*YAC_RAD},
                             .coordinates_y = (double[]){ 0.1*YAC_RAD,
                                                          0.6*YAC_RAD,
                                                          0.1*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){0.0*YAC_RAD,
                                                        intersections[0].lon,
                                                        intersections[1].lon,
                                                        -intersections[1].lon,
                                                        -intersections[0].lon},
                            .coordinates_y = (double[]){0.6*YAC_RAD,
                                                        intersections[0].lat,
                                                        intersections[1].lat,
                                                        intersections[1].lat,
                                                        intersections[0].lat},
                            .coordinates_xyz = (double[5][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 5}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 3.) Two source cells overlapping one target cell\n\n");
#endif

  /* 3.) Two source cells overlapping one target cell

     target cell test data

            0.0 (lon)
            0.7 (lat)
             / \
            1   0 
  -0.5     /     \  0.5
   0.0     \     /  0.0
            2   3
             \ /
             0.0
            -0.7  

  */

  /* source cell test data

            0.0 (lon)
            0.5 (lat)
             / \
            1   0
  -0.6     /     \   0.6
   0.0    /___2___\  0.0
          _________
          \   0   /
           \     /
            1   2
             \ /
             0.0
            -0.5
  */

  // generate reference intersection cells
  {
    struct point intersection;

    if (!yac_intersect((struct edge){.points[0] = {.lon = 0.0*YAC_RAD,
                                                   .lat = 0.5*YAC_RAD},
                                     .points[1] = {.lon = 0.6*YAC_RAD,
                                                   .lat = 0.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 0.0*YAC_RAD,
                                                   .lat = 0.7*YAC_RAD},
                                     .points[1] = {.lon = 0.5*YAC_RAD,
                                                   .lat = 0.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                     &intersection))
      return EXIT_FAILURE;

    struct grid_cell ref_cells[] = {{.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                                  intersection.lon,
                                                                  0.5*YAC_RAD,
                                                                 -0.5*YAC_RAD,
                                                                 -intersection.lon},
                                     .coordinates_y = (double[]){0.5*YAC_RAD,
                                                                 intersection.lat,
                                                                 0.0*YAC_RAD,
                                                                 0.0*YAC_RAD,
                                                                 intersection.lat},
                                     .coordinates_xyz = (double[5][3]){{-1}},
                                     .edge_type = gc_edges,
                                     .num_corners = 5},
                                    {.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                                 -intersection.lon,
                                                                 -0.5*YAC_RAD,
                                                                  0.5*YAC_RAD,
                                                                  intersection.lon},
                                     .coordinates_y = (double[]){-0.5*YAC_RAD,
                                                                 -intersection.lat,
                                                                  0.0*YAC_RAD,
                                                                  0.0*YAC_RAD,
                                                                 -intersection.lat},
                                     .coordinates_xyz = (double[5][3]){{-1}},
                                     .edge_type = gc_edges,
                                     .num_corners = 5}};

    for (size_t i = 0;
         i < sizeof(ref_cells)/sizeof(ref_cells[0]); ++i)
      generate_XYZ(ref_cells[i]);

    for (int src_order_a = -1; src_order_a <= 1; src_order_a += 2) {
      for (int src_order_b = -1; src_order_b <= 1; src_order_b += 2) {
        for (int tgt_order = -1; tgt_order <= 1; tgt_order += 2) {
          for (int src_start_a = 0; src_start_a < 3; ++src_start_a) {
            for (int src_start_b = 0; src_start_b < 3; ++src_start_b) {
              for (int tgt_start = 0; tgt_start < 4; ++tgt_start) {

                struct grid_cell SourceCells[] =
                  {{.coordinates_x = (double[]){ 0.6*YAC_RAD,
                                                 0.0*YAC_RAD,
                                                -0.6*YAC_RAD},
                    .coordinates_y = (double[]){0.0*YAC_RAD,
                                                0.5*YAC_RAD,
                                                0.0*YAC_RAD},
                    .coordinates_xyz = (double[3][3]){{-1}},
                    .edge_type = gc_edges,
                    .num_corners = 3},
                   {.coordinates_x = (double[]){ 0.6*YAC_RAD,
                                                -0.6*YAC_RAD,
                                                 0.0*YAC_RAD},
                    .coordinates_y = (double[]){ 0.0*YAC_RAD,
                                                 0.0*YAC_RAD,
                                                -0.5*YAC_RAD},
                    .coordinates_xyz = (double[3][3]){{-1}},
                    .edge_type = gc_edges,
                    .num_corners = 3}};
                struct grid_cell TargetCell =
                  {.coordinates_x = (double[]){ 0.5*YAC_RAD,
                                                0.0*YAC_RAD,
                                               -0.5*YAC_RAD,
                                                0.0*YAC_RAD},
                   .coordinates_y = (double[]){ 0.0*YAC_RAD,
                                                0.7*YAC_RAD,
                                                0.0*YAC_RAD,
                                               -0.7*YAC_RAD},
                   .coordinates_xyz = (double[4][3]){{-1}},
                   .edge_type = gc_edges,
                   .num_corners = 4};

                for (size_t i = 0;
                     i < sizeof(SourceCells)/sizeof(SourceCells[0]); ++i)
                  generate_XYZ(SourceCells[i]);

                generate_XYZ(TargetCell);

                yac_cell_clipping ( 2, SourceCells, TargetCell, overlap_cell );
                if (compare_cells(overlap_cell[0], ref_cells[0]) ||
                    compare_cells(overlap_cell[1], ref_cells[1]))
                  PUT_ERR("ERROR: wrong clipping cell\n");
              }
            }
          }
        }
      }
    }
  }

#ifdef VERBOSE
  printf ("\n 4.a.1) Simple cases: identity \n\n");
#endif

  /* 4.a.1) Simple cases: identity

     target cell test data

            0.0 (lon)
            0.5 (lat)
             / \
            1   0 
  -0.6     /     \  0.6
   0.0     \     /  0.0
            2   3
             \ /
             0.0
            -0.5

  */

  /* source cell test data

            0.0 (lon)
            0.5 (lat)
             / \
            1   0
  -0.6     /     \   0.6
   0.0    /___2___\  0.0
          _________
          \   0   /
           \     /
            1   2
             \ /
             0.0
            -0.5
  */

  // generate reference intersection cells
  {
    struct grid_cell ref_cells[] = {{.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                                  0.6*YAC_RAD,
                                                                 -0.6*YAC_RAD},
                                     .coordinates_y = (double[]){0.5*YAC_RAD,
                                                                 0.0*YAC_RAD,
                                                                 0.0*YAC_RAD},
                                     .coordinates_xyz = (double[3][3]){{-1}},
                                     .edge_type = gc_edges,
                                     .num_corners = 3},
                                    {.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                                  0.6*YAC_RAD,
                                                                 -0.6*YAC_RAD},
                                     .coordinates_y = (double[]){-0.5*YAC_RAD,
                                                                  0.0*YAC_RAD,
                                                                  0.0*YAC_RAD},
                                     .coordinates_xyz = (double[3][3]){{-1}},
                                     .edge_type = gc_edges,
                                     .num_corners = 3}};

    for (size_t i = 0; i < sizeof(ref_cells)/sizeof(ref_cells[0]); ++i)
      generate_XYZ(ref_cells[i]);

    for (int src_order_a = -1; src_order_a <= 1; src_order_a += 2) {
      for (int src_order_b = -1; src_order_b <= 1; src_order_b += 2) {
        for (int tgt_order = -1; tgt_order <= 1; tgt_order += 2) {
          for (int src_start_a = 0; src_start_a < 3; ++src_start_a) {
            for (int src_start_b = 0; src_start_b < 3; ++src_start_b) {
              for (int tgt_start = 0; tgt_start < 4; ++tgt_start) {

                struct grid_cell SourceCells[] =
                  {{.coordinates_x = (double[]){ 0.6*YAC_RAD,
                                                 0.0*YAC_RAD,
                                                -0.6*YAC_RAD},
                    .coordinates_y = (double[]){0.0*YAC_RAD,
                                                0.5*YAC_RAD,
                                                0.0*YAC_RAD},
                    .coordinates_xyz = (double[3][3]){{-1}},
                    .edge_type = gc_edges,
                    .num_corners = 3},
                   {.coordinates_x = (double[]){ 0.6*YAC_RAD,
                                                -0.6*YAC_RAD,
                                                 0.0*YAC_RAD},
                    .coordinates_y = (double[]){ 0.0*YAC_RAD,
                                                 0.0*YAC_RAD,
                                                -0.5*YAC_RAD},
                    .coordinates_xyz = (double[3][3]){{-1}},
                    .edge_type = gc_edges,
                    .num_corners = 3}};
                struct grid_cell TargetCell =
                  {.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                               -0.6*YAC_RAD,
                                                0.0*YAC_RAD,
                                                0.6*YAC_RAD},
                   .coordinates_y = (double[]){ 0.5*YAC_RAD,
                                                0.0*YAC_RAD,
                                               -0.5*YAC_RAD,
                                                0.0*YAC_RAD},
                   .coordinates_xyz = (double[4][3]){{-1}},
                   .edge_type = gc_edges,
                   .num_corners = 4};

                for (size_t i = 0;
                     i < sizeof(SourceCells)/sizeof(SourceCells[0]); ++i)
                  generate_XYZ(SourceCells[i]);

                generate_XYZ(TargetCell);

                yac_cell_clipping ( 2, SourceCells, TargetCell, overlap_cell );
                if (compare_cells(overlap_cell[0], ref_cells[0]) ||
                    compare_cells(overlap_cell[1], ref_cells[1]))
                  PUT_ERR("ERROR: wrong clipping cell\n");
                yac_free_grid_cell(overlap_cell);
                yac_free_grid_cell(overlap_cell+1);
              }
            }
          }
        }
      }
    }
  }

#ifdef VERBOSE
  printf ("\n 4.a.2) Simple cases: identity \n\n");
#endif

  /* 4.a.2) Simple cases: identity

     target cell test data

            0.0 (lon)
            0.5 (lat)
             / \
            1   0 
  -0.6     /     \  0.6
   0.0     \     /  0.0
            2   3
             \ /
             0.0
            -0.5

  */

  /* source cell test data

            0.0 (lon)
            0.5 (lat)
             / \
            1   0 
  -0.6     /     \  0.6
   0.0     \     /  0.0
            2   3
             \ /
             0.0
            -0.5
  */

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                         -0.6*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                          0.6*YAC_RAD},
                             .coordinates_y = (double[]){ 0.5*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                         -0.5*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                         -0.6*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                          0.6*YAC_RAD},
                             .coordinates_y = (double[]){ 0.5*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                         -0.5*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                        -0.6*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                         0.6*YAC_RAD},
                            .coordinates_y = (double[]){ 0.5*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                        -0.5*YAC_RAD,
                                                         0.0*YAC_RAD},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 4}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 4.b) Simple cases: inside cell\n\n");
#endif

  /* 4.b) Simple cases: source inside target cell

     target cell test data

            0.0 (lon)
            0.5 (lat)
             / \
            1   0 
  -0.6     /     \  0.6
   0.0     \     /  0.0
            2   3
             \ /
             0.0
            -0.5

  */

  /* source cell test data

            0.0 (lon)
            0.4 (lat)
             / \
            1   0 
  -0.5     /     \  0.5
   0.0     \     /  0.0
            2   3
             \ /
             0.0
            -0.4
  */

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                         -0.5*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                          0.5*YAC_RAD},
                             .coordinates_y = (double[]){ 0.4*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                         -0.4*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                         -0.6*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                          0.6*YAC_RAD},
                             .coordinates_y = (double[]){ 0.5*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                         -0.5*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                        -0.5*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                         0.5*YAC_RAD},
                            .coordinates_y = (double[]){ 0.4*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                        -0.4*YAC_RAD,
                                                         0.0*YAC_RAD},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 4}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 4.c) Simple cases: target inside source cell\n\n");
#endif

  /* 4.c) Simple cases: target inside source cell

     target cell test data

            0.0 (lon)
            0.5 (lat)
             / \
            1   0 
  -0.6     /     \  0.6
   0.0     \     /  0.0
            2   3
             \ /
             0.0
            -0.5

  */

  /* source cell test data

            0.0 (lon)
            0.6 (lat)
             / \
            1   0 
  -0.7     /     \  0.7
   0.0     \     /  0.0
            2   3
             \ /
             0.0
            -0.6
  */

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){  0.0*YAC_RAD,
                                                          -0.7*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                           0.7*YAC_RAD},
                             .coordinates_y = (double[]){  0.6*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                          -0.6*YAC_RAD,
                                                           0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                         -0.6*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                          0.6*YAC_RAD},
                             .coordinates_y = (double[]){ 0.5*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                         -0.5*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                        -0.6*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                         0.6*YAC_RAD},
                            .coordinates_y = (double[]){ 0.5*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                        -0.5*YAC_RAD,
                                                         0.0*YAC_RAD},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 4}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 4.d) Simple cases: degenerated cell \n\n");
#endif

  /* 4.d) Simple cases: degenerated cell 

     target cell test data

            0.0 (lon)
            0.5 (lat)
             / \
            1   0 
  -0.6     /     \  0.6
   0.0     \     /  0.0
            2   3
             \ /
             0.0
            -0.5

  */

  /* source cell test data

       0.0     0.0 (lon)
       0.5     0.6 (lat)
        --------
         \      \
          \      \
           \      \
            --------
           0.6     1.0
           0.0     0.0
  */

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){0.0*YAC_RAD,
                                                         0.6*YAC_RAD,
                                                         1.0*YAC_RAD,
                                                         0.5*YAC_RAD},
                             .coordinates_y = (double[]){0.5*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                         0.5*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                         -0.6*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                          0.6*YAC_RAD},
                             .coordinates_y = (double[]){ 0.5*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                         -0.5*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 5.) Special case from ICON_toy \n\n");
#endif
  /* 5.) source cell test data

     +------+
     |      |
     |      |
     |      |
     +------+
  */

  /* target cell test data

     /|
    / |
   /  |
  /   |
  \   |
   \  |
    \ |
     \|
  */

  {
    struct point intersections[4];

    if (!yac_intersect((struct edge){.points[0] = {.lon = 108.0000*YAC_RAD,
                                                   .lat =  -7.2167*YAC_RAD},
                                     .points[1] = {.lon =  88.4979*YAC_RAD,
                                                   .lat = -19.2921*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  90.0*YAC_RAD,
                                                   .lat =   0.0*YAC_RAD},
                                     .points[1] = {.lon =  90.0*YAC_RAD,
                                                   .lat = -22.5*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  88.4979*YAC_RAD,
                                                   .lat = -19.2921*YAC_RAD},
                                     .points[1] = {.lon = 108.0000*YAC_RAD,
                                                   .lat = -26.5650*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  90.0*YAC_RAD,
                                                   .lat =   0.0*YAC_RAD},
                                     .points[1] = {.lon =  90.0*YAC_RAD,
                                                   .lat = -22.5*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  88.4979*YAC_RAD,
                                                   .lat = -19.2921*YAC_RAD},
                                     .points[1] = {.lon = 108.0000*YAC_RAD,
                                                   .lat = -26.5650*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  90.0*YAC_RAD,
                                                   .lat = -22.5*YAC_RAD},
                                     .points[1] = {.lon = 112.5*YAC_RAD,
                                                   .lat = -22.5*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+2))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = 108.0000*YAC_RAD,
                                                   .lat = -26.5650*YAC_RAD},
                                     .points[1] = {.lon = 108.0000*YAC_RAD,
                                                   .lat =  -7.2167*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  90.0*YAC_RAD,
                                                   .lat = -22.5*YAC_RAD},
                                     .points[1] = {.lon = 112.5*YAC_RAD,
                                                   .lat = -22.5*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+3))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){ 90.0*YAC_RAD,
                                                         112.5*YAC_RAD,
                                                         112.5*YAC_RAD,
                                                          90.0*YAC_RAD},
                             .coordinates_y = (double[]){-22.5*YAC_RAD,
                                                         -22.5*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                           0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){108.0000*YAC_RAD,
                                                          88.4979*YAC_RAD,
                                                         108.0000*YAC_RAD},
                             .coordinates_y = (double[]){ -7.2167*YAC_RAD,
                                                         -19.2921*YAC_RAD,
                                                         -26.5650*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){108.0000*YAC_RAD,
                                                        intersections[0].lon,
                                                        intersections[1].lon,
                                                        intersections[2].lon,
                                                        intersections[3].lon},
                            .coordinates_y = (double[]){-7.2167*YAC_RAD,
                                                        intersections[0].lat,
                                                        intersections[1].lat,
                                                        intersections[2].lat,
                                                        intersections[3].lat},
                            .coordinates_xyz = (double[5][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 5}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 6.) Special case from ICON_toy \n\n");
#endif
  /* 6.) source cell test data

     +------+  0.5
     |      |
     |      |
     |      |
     +------+ -0.5
   -0.5    0.5
  */

  /* target cell test data

    -0.75   0.0
        ______  0.0
        \    |
         \   |
          \  |
           \ |
            \| -0.75
  */

  {
    struct point intersections[4];

    if (!yac_intersect((struct edge){.points[0] = {.lon =  0.0 *YAC_RAD,
                                                   .lat =  0.0 *YAC_RAD},
                                     .points[1] = {.lon = -0.75*YAC_RAD,
                                                   .lat =  0.0 *YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -0.5*YAC_RAD,
                                                   .lat =  0.5*YAC_RAD},
                                     .points[1] = {.lon = -0.5*YAC_RAD,
                                                   .lat = -0.5*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = -0.75*YAC_RAD,
                                                   .lat =  0.0 *YAC_RAD},
                                     .points[1] = {.lon =  0.0 *YAC_RAD,
                                                   .lat = -0.75*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -0.5*YAC_RAD,
                                                   .lat =  0.5*YAC_RAD},
                                     .points[1] = {.lon = -0.5*YAC_RAD,
                                                   .lat = -0.5*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = -0.75*YAC_RAD,
                                                   .lat =  0.0 *YAC_RAD},
                                     .points[1] = {.lon =  0.0 *YAC_RAD,
                                                   .lat = -0.75*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -0.5*YAC_RAD,
                                                   .lat = -0.5*YAC_RAD},
                                     .points[1] = {.lon =  0.5*YAC_RAD,
                                                   .lat = -0.5*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+2))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  0.0 *YAC_RAD,
                                                   .lat = -0.75*YAC_RAD},
                                     .points[1] = {.lon =  0.0 *YAC_RAD,
                                                   .lat =  0.0 *YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -0.5*YAC_RAD,
                                                   .lat = -0.5*YAC_RAD},
                                     .points[1] = {.lon =  0.5*YAC_RAD,
                                                   .lat = -0.5*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+3))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-0.5*YAC_RAD,
                                                          0.5*YAC_RAD,
                                                          0.5*YAC_RAD,
                                                         -0.5*YAC_RAD},
                             .coordinates_y = (double[]){-0.5*YAC_RAD,
                                                         -0.5*YAC_RAD,
                                                          0.5*YAC_RAD,
                                                          0.5*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){ 0.0 *YAC_RAD,
                                                         -0.75*YAC_RAD,
                                                          0.0 *YAC_RAD},
                             .coordinates_y = (double[]){ 0.0 *YAC_RAD,
                                                          0.0 *YAC_RAD,
                                                         -0.75*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){0.0*YAC_RAD,
                                                        intersections[0].lon,
                                                        intersections[1].lon,
                                                        intersections[2].lon,
                                                        intersections[3].lon},
                            .coordinates_y = (double[]){0.0*YAC_RAD,
                                                        intersections[0].lat,
                                                        intersections[1].lat,
                                                        intersections[2].lat,
                                                        intersections[3].lat},
                            .coordinates_xyz = (double[5][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 5}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 7.) Special case in which no corner on any square is within the other \n\n");
#endif
  /* 7.) source cell test data

     +------+  0.5
     |      |
     |      |
     |      |
     +------+ -0.5
   -0.5    0.5
  */

  /* target cell test data

   -0.7  0.0  0.7
          /\     0.7
         /  \
        /    \
       /      \  0.0
       \      /
        \    /
         \  /
          \/    -0.7
  */

  {
    struct edge edges[2];
    struct point intersection;
    double coord_mem[2][8];
    enum yac_edge_type edge_mem[8];
    double src_data[2][4] = {{-0.5*YAC_RAD,-0.5*YAC_RAD, 0.5*YAC_RAD, 0.5*YAC_RAD},
                             { 0.5*YAC_RAD,-0.5*YAC_RAD,-0.5*YAC_RAD, 0.5*YAC_RAD}};
    double tgt_data[2][4] = {{ 0.0*YAC_RAD,-0.7*YAC_RAD, 0.0*YAC_RAD, 0.7*YAC_RAD},
                             { 0.7*YAC_RAD, 0.0*YAC_RAD,-0.7*YAC_RAD, 0.0*YAC_RAD}};

    edges[0].edge_type = GREAT_CIRCLE;
    edges[1].edge_type = GREAT_CIRCLE;

    for (int i = 0; i < 8; ++i) {
      edges[0].points[0].lon = src_data[0][i/2];
      edges[0].points[0].lat = src_data[1][i/2];
      edges[0].points[1].lon = src_data[0][(i/2+1)%4];
      edges[0].points[1].lat = src_data[1][(i/2+1)%4];
      edges[1].points[0].lon = tgt_data[0][((i+1)/2)%4];
      edges[1].points[0].lat = tgt_data[1][((i+1)/2)%4];
      edges[1].points[1].lon = tgt_data[0][((i+1)/2+1)%4];
      edges[1].points[1].lat = tgt_data[1][((i+1)/2+1)%4];
      if (!yac_intersect(edges[0], edges[1], &intersection))
        return EXIT_FAILURE;
      coord_mem[0][i] = intersection.lon;
      coord_mem[1][i] = intersection.lat;
      edge_mem[i] = GREAT_CIRCLE;
    }

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-0.5*YAC_RAD,
                                                          0.5*YAC_RAD,
                                                          0.5*YAC_RAD,
                                                         -0.5*YAC_RAD},
                             .coordinates_y = (double[]){-0.5*YAC_RAD,
                                                         -0.5*YAC_RAD,
                                                          0.5*YAC_RAD,
                                                          0.5*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                          0.7*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                         -0.7*YAC_RAD},
                             .coordinates_y = (double[]){-0.7*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                          0.7*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = coord_mem[0],
                            .coordinates_y = coord_mem[1],
                            .coordinates_xyz = (double[8][3]){{-1}},
                            .edge_type = edge_mem,
                            .num_corners = 8}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 8.) Test non intersecting cells \n\n");
#endif
  /* 8.) source cell test data

     +------+ 1
     |      |
     |      |
     |      |
     +------+ 0
     0      1
  */

  /* target cell test data

     +------+ 1
     |      |
     |      |
     |      |
     +------+ 0
     2      3
  */

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){0.0*YAC_RAD,
                                                         1.0*YAC_RAD,
                                                         1.0*YAC_RAD,
                                                         0.0*YAC_RAD},
                             .coordinates_y = (double[]){0.0*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                         1.0*YAC_RAD,
                                                         1.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){2.0*YAC_RAD,
                                                         3.0*YAC_RAD,
                                                         3.0*YAC_RAD,
                                                         2.0*YAC_RAD},
                             .coordinates_y = (double[]){0.0*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                         1.0*YAC_RAD,
                                                         1.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 9.) triangle-square intersection \n\n");
#endif
  /* 9.) source cell test data

    -1.5     1.5
      --------  0.5
      \      /
       \    /
        \  /
         \/    -0.5
  */

  /* target cell test data

     +------+  1
     |      |
     |      |
     |      |
     +------+ -1
    -1      1
  */

  {
    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon =  1.0*YAC_RAD,
                                                   .lat = -1.0*YAC_RAD},
                                     .points[1] = {.lon =  1.0*YAC_RAD,
                                                   .lat =  1.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  0.0*YAC_RAD,
                                                   .lat = -0.5*YAC_RAD},
                                     .points[1] = {.lon =  1.5*YAC_RAD,
                                                   .lat =  0.5*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  1.0*YAC_RAD,
                                                   .lat = -1.0*YAC_RAD},
                                     .points[1] = {.lon =  1.0*YAC_RAD,
                                                   .lat =  1.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  1.5*YAC_RAD,
                                                   .lat =  0.5*YAC_RAD},
                                     .points[1] = {.lon = -1.5*YAC_RAD,
                                                   .lat =  0.5*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){ 1.5*YAC_RAD,
                                                         -1.5*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_y = (double[]){ 0.5*YAC_RAD,
                                                          0.5*YAC_RAD,
                                                         -0.5*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){-1.0*YAC_RAD,
                                                          1.0*YAC_RAD,
                                                          1.0*YAC_RAD,
                                                         -1.0*YAC_RAD},
                             .coordinates_y = (double[]){-1.0*YAC_RAD,
                                                         -1.0*YAC_RAD,
                                                          1.0*YAC_RAD,
                                                          1.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){0.0*YAC_RAD,
                                                        intersections[0].lon,
                                                        intersections[1].lon,
                                                        -intersections[1].lon,
                                                        -intersections[0].lon},
                            .coordinates_y = (double[]){-0.5*YAC_RAD,
                                                        intersections[0].lat,
                                                        intersections[1].lat,
                                                        intersections[1].lat,
                                                        intersections[0].lat},
                            .coordinates_xyz = (double[5][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 5}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 10.1) touching edges of two triangles \n\n");
#endif
  /* 10.1) source cell test data

       /\     1
      /  \
     /    \
    /      \
    --------  0
    0      2
  */
  /* target cell test data

      --------  0.0
      \      /
       \    /
        \  /
         \/    -0.5
     0  0.5  1
  */

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){0.0*YAC_RAD,
                                                         2.0*YAC_RAD,
                                                         1.0*YAC_RAD},
                             .coordinates_y = (double[]){0.0*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                         1.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                          1.0*YAC_RAD,
                                                          0.5*YAC_RAD},
                             .coordinates_y = (double[]){ 0.0*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                         -0.5*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 10.2) touching edges \n\n");
#endif

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){160.0*YAC_RAD,
                                                         180.0*YAC_RAD,
                                                         180.0*YAC_RAD},
                             .coordinates_y = (double[]){ 65.0*YAC_RAD,
                                                          65.0*YAC_RAD,
                                                          90.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){180.0*YAC_RAD,
                                                         200.0*YAC_RAD,
                                                           0.0*YAC_RAD},
                             .coordinates_y = (double[]){ 70.0*YAC_RAD,
                                                          70.0*YAC_RAD,
                                                          90.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 11.1) test from test_interpolation_method_conserv.x \n\n");
#endif
  /* 11.) source cell test data

     +------+  0
     |      |
     |      |
     |      |
     +------+ -1
     0      1
  */
  /* target cell test data

     +------+  0.5
     |      |
     |      |
     |      |
     +------+ -0.5
   -0.5    0.5
  */

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){0.0*YAC_RAD,
                                                         1.0*YAC_RAD,
                                                         1.0*YAC_RAD,
                                                         0.0*YAC_RAD},
                             .coordinates_y = (double[]){-1.0*YAC_RAD,
                                                         -1.0*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-0.5*YAC_RAD,
                                                          0.5*YAC_RAD,
                                                          0.5*YAC_RAD,
                                                         -0.5*YAC_RAD},
                             .coordinates_y = (double[]){-0.5*YAC_RAD,
                                                         -0.5*YAC_RAD,
                                                          0.5*YAC_RAD,
                                                          0.5*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){0.0*YAC_RAD,
                                                        0.5*YAC_RAD,
                                                        0.5*YAC_RAD,
                                                        0.0*YAC_RAD},
                            .coordinates_y = (double[]){ 0.0*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                        -0.5*YAC_RAD,
                                                        -0.5*YAC_RAD,},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = latlon_edges,
                            .num_corners = 4}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 11.2) test from test_interpolation_method_conserv.x \n\n");
#endif
  /* 11.) source cell test data

    -1      0
     +------+  0
     |      |
     |      |
     |      |
     +------+ -1
  */
  /* target cell test data

     +------+  0.5
     |      |
     |      |
     |      |
     +------+ -0.5
   -0.5    0.5
  */

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                         -1.0*YAC_RAD,
                                                         -1.0*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_y = (double[]){-1.0*YAC_RAD,
                                                         -1.0*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                          0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-0.5*YAC_RAD,
                                                          0.5*YAC_RAD,
                                                          0.5*YAC_RAD,
                                                         -0.5*YAC_RAD},
                             .coordinates_y = (double[]){-0.5*YAC_RAD,
                                                         -0.5*YAC_RAD,
                                                          0.5*YAC_RAD,
                                                          0.5*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){ 0.0*YAC_RAD,
                                                        -0.5*YAC_RAD,
                                                        -0.5*YAC_RAD,
                                                         0.0*YAC_RAD},
                            .coordinates_y = (double[]){ 0.0*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                        -0.5*YAC_RAD,
                                                        -0.5*YAC_RAD,},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = latlon_edges,
                            .num_corners = 4}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
    printf ("\n 12.1) some mean tests \n\n");
#endif
    /* cell a

       +------+  60
       |      |
       |      |
       |      |
       +------+ 40
      -20     20
    */

    /* cell b

        --------  39
        \      /
         \    /
          \  /
           \/     22
     -20   0   20
    */

  {
    struct point intersection;

    if (!yac_intersect((struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  39.0*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  39.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       &intersection))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                          20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 39.0*YAC_RAD,
                                                          22.0*YAC_RAD,
                                                          39.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){ intersection.lon,
                                                        -intersection.lon},
                            .coordinates_y = (double[]){intersection.lat,
                                                        intersection.lat},
                            .coordinates_xyz = (double[2][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 2},
                           {.coordinates_x = (double[]){ intersection.lon,
                                                        -intersection.lon},
                            .coordinates_y = (double[]){intersection.lat,
                                                        intersection.lat},
                            .coordinates_xyz = (double[2][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE},
                            .num_corners = 2}},
      // number of reference cells
      2);
  }

#ifdef VERBOSE
  printf ("\n 12.2) some mean tests \n\n");
#endif
  /* cell a

     +------+  60
     |      |
     |      |
     |      |
     +------+ 40
    -20     20
  */

  /* cell b

  40  --....__  39
      \      /
       \    /
        \  /
         \/     22
   -20   0   20
  */

  {
    struct point intersection;

    if (!yac_intersect((struct edge){.points[0] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .points[1] = {.lon = -19.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  39.0*YAC_RAD},
                                     .points[1] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       &intersection))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                          20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          22.0*YAC_RAD,
                                                          39.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                        intersection.lon},
                            .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                        intersection.lat},
                            .coordinates_xyz = (double[2][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 2},
                           {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                        intersection.lon},
                            .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                        intersection.lat},
                            .coordinates_xyz = (double[2][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE},
                            .num_corners = 2}},
      // number of reference cells
      2);
  }

#ifdef VERBOSE
  printf ("\n 12.3) some mean tests \n\n");
#endif
  /* cell a

     +------+  60
     |      |
     |      |
     |      |
     +------+ 40
    -20     20
  */

  /* cell b

         /\     80
        /  \
       /    \
      /      \
      --------  59
   -19   0   19
  */

  {
    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon = 19.0*YAC_RAD,
                                                   .lat = 59.0*YAC_RAD},
                                     .points[1] = {.lon =  0.0*YAC_RAD,
                                                   .lat = 80.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  19.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .points[1] = {.lon = -19.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-19.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                          19.0*YAC_RAD},
                             .coordinates_y = (double[]){59.0*YAC_RAD,
                                                         80.0*YAC_RAD,
                                                         59.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){19.0*YAC_RAD,
                                                        intersections[0].lon,
                                                        fabs(intersections[1].lon),
                                                        -fabs(intersections[1].lon),
                                                        -intersections[0].lon,
                                                        -19.0*YAC_RAD},
                            .coordinates_y = (double[]){59.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        59.0*YAC_RAD},
                            .coordinates_xyz = (double[6][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 6},
                           {.coordinates_x = (double[]){19.0*YAC_RAD,
                                                        intersections[0].lon,
                                                        -intersections[0].lon,
                                                        -19.0*YAC_RAD,
                                                        -fabs(intersections[1].lon),
                                                        fabs(intersections[1].lon)},
                            .coordinates_y = (double[]){59.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        59.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD},
                            .coordinates_xyz = (double[6][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 6}},
      // number of reference cells
      2);
  }

#ifdef VERBOSE
  printf ("\n 12.4) some mean tests \n\n");
#endif
  /* cell a

     +------+  60
     |      |
     |      |
     |      |
     +------+ 40
    -20     20
  */

  /* cell b

            /| 80
          /  |
        /    |
      /      |
  60  --...__|
               59
   -20       21
  */

  {
    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .points[1] = {.lon =  21.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -19.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .points[1] = {.lon =  21.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 40.0*YAC_RAD},
                                     .points[1] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .edge_type = LON_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          21.0*YAC_RAD,
                                                          21.0*YAC_RAD},
                             .coordinates_y = (double[]){60.0*YAC_RAD,
                                                         80.0*YAC_RAD,
                                                         59.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){20.0*YAC_RAD,
                                                        intersections[0].lon,
                                                        intersections[1].lon},
                            .coordinates_y = (double[]){60.0*YAC_RAD,
                                                        intersections[0].lat,
                                                        intersections[1].lat},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LON_CIRCLE},
                            .num_corners = 3},
                           {.coordinates_x = (double[]){20.0*YAC_RAD,
                                                        intersections[0].lon,
                                                        -20.0*YAC_RAD,
                                                        intersections[1].lon},
                            .coordinates_y = (double[]){60.0*YAC_RAD,
                                                        intersections[0].lat,
                                                        60.0*YAC_RAD,
                                                        intersections[1].lat},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LON_CIRCLE},
                            .num_corners = 4}},
      // number of reference cells
      2);
  }

#ifdef VERBOSE
  printf ("\n 12.5) some mean tests \n\n");
#endif
  /* cell a

     +------+  60
     |      |
     |      |
     |      |
     +------+ 40
    -20     20
  */

  /* cell b

 60 --...__  59
    \      |
      \    |
        \  |
          \| 40

  -20     21

  */

  {
    struct point intersections[3];

    if (!yac_intersect((struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .points[1] = {.lon =  21.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -19.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .points[1] = {.lon =  21.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 40.0*YAC_RAD},
                                     .points[1] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .edge_type = LON_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .points[1] = {.lon =  21.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 40.0*YAC_RAD},
                                     .points[1] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .edge_type = LON_CIRCLE},
                       intersections+2))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          21.0*YAC_RAD,
                                                          21.0*YAC_RAD},
                             .coordinates_y = (double[]){60.0*YAC_RAD,
                                                         40.0*YAC_RAD,
                                                         59.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){intersections[0].lon,
                                                        intersections[1].lon,
                                                        intersections[2].lon,
                                                        -20.0*YAC_RAD},
                            .coordinates_y = (double[]){intersections[0].lat,
                                                        intersections[1].lat,
                                                        intersections[2].lat,
                                                        60.0*YAC_RAD},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LON_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE},
                            .num_corners = 4}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 12.6) some mean tests \n\n");
#endif
    /* cell a

       +------+  60
       |      |
       |      |
       |      |
       +------+ 40
       x      x+40
    */

    /* cell b

        --------  39
        \      /
         \    /
          \  /
           \/     20
      -20  0   20
    */

  {
    struct point intersection;

    if (!yac_intersect((struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  39.0*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  39.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .points[1] = {.lon =   0.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       &intersection))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){intersection.lon,
                                                         intersection.lon+40.0*YAC_RAD,
                                                         intersection.lon+40.0*YAC_RAD,
                                                         intersection.lon},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                           0.0*YAC_RAD},
                             .coordinates_y = (double[]){39.0*YAC_RAD,
                                                         39.0*YAC_RAD,
                                                         20.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){intersection.lon,
                                                        -intersection.lon},
                            .coordinates_y = (double[]){40.0*YAC_RAD,
                                                        40.0*YAC_RAD},
                            .coordinates_xyz = (double[2][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 2},
                           {.coordinates_x = (double[]){intersection.lon,
                                                        -intersection.lon},
                            .coordinates_y = (double[]){40.0*YAC_RAD,
                                                        40.0*YAC_RAD},
                            .coordinates_xyz = (double[2][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE},
                            .num_corners = 2}},
      // number of reference cells
      2);
  }

#ifdef VERBOSE
  printf ("\n 12.7) some mean tests \n\n");
#endif
    /* cell a

       +------+  60
       |      |
       |      |
       |      |
       +------+ 40
       x      x+40
    */

    /* cell b

      |\        50
      |  \
      |    \
      |      \
      --------  39
   -20   0   20
    */

  {
    struct point intersections[3];

    if (!yac_intersect((struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  39.0*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  39.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .points[1] = {.lon =   0.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  39.0*YAC_RAD},
                                     .points[1] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  50.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  39.0*YAC_RAD},
                                     .points[1] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  50.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = intersections[0].lon,
                                                   .lat = 40.0*YAC_RAD},
                                     .points[1] = {.lon = intersections[0].lon,
                                                   .lat = 60.0*YAC_RAD},
                                     .edge_type = LON_CIRCLE},
                       intersections+2))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){intersections[0].lon,
                                                         intersections[0].lon+40.0*YAC_RAD,
                                                         intersections[0].lon+40.0*YAC_RAD,
                                                         intersections[0].lon},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){39.0*YAC_RAD,
                                                         39.0*YAC_RAD,
                                                         50.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){ intersections[0].lon,
                                                        -intersections[0].lon,
                                                         intersections[1].lon,
                                                         intersections[2].lon},
                            .coordinates_y = (double[]){40.0*YAC_RAD,
                                                        40.0*YAC_RAD,
                                                        40.0*YAC_RAD,
                                                        intersections[2].lat},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LON_CIRCLE},
                            .num_corners = 4}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
    printf ("\n 12.8) some mean tests \n\n");
#endif
    /* cell a

       +------+  60
       |      |
       |      |
       |      |
       +------+ 40
      -20     20
    */

    /* cell b

        --------  40
        \      /
         \    /
          \  /
           \/     20
     -20   0   20
    */

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                          20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          40.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                         20.0*YAC_RAD},
                            .coordinates_y = (double[]){40.0*YAC_RAD,
                                                        40.0*YAC_RAD},
                            .coordinates_xyz = (double[2][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 2},
                           {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                         20.0*YAC_RAD},
                            .coordinates_y = (double[]){40.0*YAC_RAD,
                                                        40.0*YAC_RAD},
                            .coordinates_xyz = (double[2][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE},
                            .num_corners = 2}},
      // number of reference cells
      2);
  }

#ifdef VERBOSE
    printf ("\n 12.9) some mean tests \n\n");
#endif
    /* cell a

       +------+  60
       |      |
       |      |
       |      |
       +------+ 40
      -20     20
    */

    /* cell b

         /\     50
        /  \
       /    \
      /      \
      --------  40
   -20   0   20
    */

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                          20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          50.0*YAC_RAD,
                                                          40.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                         20.0*YAC_RAD,
                                                          0.0*YAC_RAD},
                            .coordinates_y = (double[]){40.0*YAC_RAD,
                                                        40.0*YAC_RAD,
                                                        50.0*YAC_RAD},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 3}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
    printf ("\n 12.10) some mean tests \n\n");
#endif
    /* cell a

       +------+  60
       |      |
       |      |
       |      |
       +------+ 40
      -20     20
    */

    /* cell b

        --------  60
        \      /
         \    /
          \  /
           \/     50
     -20   0   20
    */

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                          20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 60.0*YAC_RAD,
                                                          50.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                         20.0*YAC_RAD,
                                                          0.0*YAC_RAD},
                            .coordinates_y = (double[]){60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        50.0*YAC_RAD},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 3}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
    printf ("\n 12.11) some mean tests \n\n");
#endif
    /* cell a

       +------+  60
       |      |
       |      |
       |      |
       +------+ 40
      -20     20
    */

    /* cell b

         /\     70
        /  \
       /    \
      /      \
      --------  60
   -20   0   20
    */

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                          20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 60.0*YAC_RAD,
                                                          70.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                         20.0*YAC_RAD},
                            .coordinates_y = (double[]){60.0*YAC_RAD,
                                                        60.0*YAC_RAD},
                            .coordinates_xyz = (double[2][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 2},
                           {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                         20.0*YAC_RAD},
                            .coordinates_y = (double[]){60.0*YAC_RAD,
                                                        60.0*YAC_RAD},
                            .coordinates_xyz = (double[2][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE},
                            .num_corners = 2}},
      // number of reference cells
      2);
  }

#ifdef VERBOSE
    printf ("\n 12.12) some mean tests \n\n");
#endif
    /* cell a

       +------+ -40
       |      |
       |      |
       |      |
       +------+ -60
      -20     20
    */

    /* cell b

        -------- -40
        \      /
         \    /
          \  /
           \/    -50
     -20   0   20
    */

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){-40.0*YAC_RAD,
                                                         -40.0*YAC_RAD,
                                                         -60.0*YAC_RAD,
                                                         -60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                          20.0*YAC_RAD},
                             .coordinates_y = (double[]){-40.0*YAC_RAD,
                                                         -50.0*YAC_RAD,
                                                         -40.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                         20.0*YAC_RAD,
                                                          0.0*YAC_RAD},
                            .coordinates_y = (double[]){-40.0*YAC_RAD,
                                                        -40.0*YAC_RAD,
                                                        -50.0*YAC_RAD},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 3}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
    printf ("\n 12.13) some mean tests \n\n");
#endif
    /* cell a

       +------+  60
       |      |
       |      |
       |      |
       +------+  40
      -20     20
    */

    /* cell b

     59 ___..--- 60
        \      /
         \    /
          \  /
           \/    45
     -15   0   15
    */

  {
    struct point intersection;

    if (!yac_intersect((struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                               .lat =  60.0*YAC_RAD},
                                 .points[1] = {.lon =  14.9*YAC_RAD,
                                               .lat =  60.0*YAC_RAD},
                                 .edge_type = LAT_CIRCLE},
                   (struct edge){.points[0] = {.lon = -15.0*YAC_RAD,
                                               .lat =  59.0*YAC_RAD},
                                 .points[1] = {.lon =  15.0*YAC_RAD,
                                               .lat =  60.0*YAC_RAD},
                                 .edge_type = GREAT_CIRCLE},
                   &intersection))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){60.0*YAC_RAD,
                                                         60.0*YAC_RAD,
                                                         40.0*YAC_RAD,
                                                         40.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-15.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                          15.0*YAC_RAD},
                             .coordinates_y = (double[]){59.0*YAC_RAD,
                                                         45.0*YAC_RAD,
                                                         60.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){intersection.lon,
                                                         15.0*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                        -15.0*YAC_RAD},
                            .coordinates_y = (double[]){intersection.lat,
                                                        60.0*YAC_RAD,
                                                        45.0*YAC_RAD,
                                                        59.0*YAC_RAD},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 4}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
    printf ("\n 12.14) some mean tests \n\n");
#endif
    /* cell a

       +------+  60
       +------+  59
      -20     20
    */

    /* cell b

        --------  58.5
        \      /
         \    /
          \  /
           \/     40
     -20   0   20
    */

  {
    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .points[1] = {.lon =   0.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  58.5*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  58.5*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .points[1] = {.lon =   0.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  58.5*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  58.5*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){60.0*YAC_RAD,
                                                         60.0*YAC_RAD,
                                                         59.0*YAC_RAD,
                                                         59.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                          20.0*YAC_RAD},
                             .coordinates_y = (double[]){58.5*YAC_RAD,
                                                         40.0*YAC_RAD,
                                                         58.5*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){ intersections[0].lon,
                                                         intersections[1].lon,
                                                        -intersections[1].lon,
                                                        -intersections[0].lon},
                            .coordinates_y = (double[]){59.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        59.0*YAC_RAD},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE},
                            .num_corners = 4}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 12.15) some mean tests \n\n");
#endif
  /* cell a

     +------+  60
     |      |
     |      |
     |      |
     +------+ 40
    -20     20
  */

  /* cell b

         /\     80
        /  \
       /    \
      /      \
      --------  59
   -20   0   20
  */

  {
    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 59.0*YAC_RAD},
                                     .points[1] = {.lon =  0.0*YAC_RAD,
                                                   .lat = 80.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .points[1] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                 .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                          20.0*YAC_RAD},
                             .coordinates_y = (double[]){59.0*YAC_RAD,
                                                         80.0*YAC_RAD,
                                                         59.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){20.0*YAC_RAD,
                                                        intersections[0].lon,
                                                        fabs(intersections[1].lon),
                                                        -fabs(intersections[1].lon),
                                                        -intersections[0].lon,
                                                        -20.0*YAC_RAD},
                            .coordinates_y = (double[]){59.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        59.0*YAC_RAD},
                            .coordinates_xyz = (double[6][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 6},
                           {.coordinates_x = (double[]){20.0*YAC_RAD,
                                                        intersections[0].lon,
                                                        -intersections[0].lon,
                                                        -20.0*YAC_RAD,
                                                        -fabs(intersections[1].lon),
                                                        fabs(intersections[1].lon)},
                            .coordinates_y = (double[]){59.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        59.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD},
                            .coordinates_xyz = (double[6][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 6}},
      // number of reference cells
      2);
  }

#ifdef VERBOSE
  printf ("\n 12.16) some mean tests \n\n");
#endif
  /* cell a

     +------+  60
     |      |
     |      |
     |      |
     +------+ 40
    -20     20
  */

  /* cell b

        --------  59
        \      /
         \    /
          \  /
           \/     50
     -20   0   20
  */

  {
    struct point intersection;

    if (!yac_intersect((struct edge){.points[0] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 59.0*YAC_RAD},
                                     .points[1] = {.lon =-20.0*YAC_RAD,
                                                   .lat = 59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .points[1] = {.lon =   0.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       &intersection))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                          20.0*YAC_RAD},
                             .coordinates_y = (double[]){59.0*YAC_RAD,
                                                         50.0*YAC_RAD,
                                                         59.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){ intersection.lon,
                                                        -intersection.lon,
                                                         20.0*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                        -20.0*YAC_RAD},
                            .coordinates_y = (double[]){60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        59.0*YAC_RAD,
                                                        50.0*YAC_RAD,
                                                        59.0*YAC_RAD},
                            .coordinates_xyz = (double[5][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 5}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 12.17) some mean tests \n\n");
#endif
  /* cell a

     +------+  20
     |      |
     |      |
     |      |
     +------+ -20
    -20     20
  */

  /* cell b

     +------+  19
     |      |
     |      |
     |      |
     +------+ -19
    -19     19
  */

  {
    struct point intersection;

    if (!yac_intersect((struct edge){.points[0] = {.lon =-19.0*YAC_RAD,
                                                   .lat = 19.0*YAC_RAD},
                                     .points[1] = {.lon = 19.0*YAC_RAD,
                                                   .lat = 19.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  20.0*YAC_RAD},
                                     .points[1] = {.lon =   0.0*YAC_RAD,
                                                   .lat =  20.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       &intersection))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-19.0*YAC_RAD,
                                                         -19.0*YAC_RAD,
                                                          19.0*YAC_RAD,
                                                          19.0*YAC_RAD},
                             .coordinates_y = (double[]){ 19.0*YAC_RAD,
                                                         -19.0*YAC_RAD,
                                                         -19.0*YAC_RAD,
                                                          19.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){-19.0*YAC_RAD,
                                                         intersection.lon,
                                                        -intersection.lon,
                                                         19.0*YAC_RAD,
                                                         19.0*YAC_RAD,
                                                        -intersection.lon,
                                                         intersection.lon,
                                                        -19.0*YAC_RAD},
                            .coordinates_y = (double[]){ 19.0*YAC_RAD,
                                                         20.0*YAC_RAD,
                                                         20.0*YAC_RAD,
                                                         19.0*YAC_RAD,
                                                        -19.0*YAC_RAD,
                                                        -20.0*YAC_RAD,
                                                        -20.0*YAC_RAD,
                                                        -19.0*YAC_RAD},
                            .coordinates_xyz = (double[8][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 8}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
    printf ("\n 12.18) some mean tests \n\n");
#endif
    /* cell a

       +------+  60
       |      |
       |      |
       |      |
       +------+ 40
      -20     20
    */

    /* cell b

         /\     55
        /  \
       /    \
      /      \
      --------  45
   -40  -20   0
    */

  {
    struct point intersection;

    if (!yac_intersect((struct edge){.points[0] = {.lon = -40.0*YAC_RAD,
                                                   .lat =  45.0*YAC_RAD},
                                     .points[1] = {.lon =   0.0*YAC_RAD,
                                                   .lat =  45.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .points[1] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .edge_type = LON_CIRCLE},
                       &intersection))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-40.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 45.0*YAC_RAD,
                                                          45.0*YAC_RAD,
                                                          55.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                        -20.0*YAC_RAD,
                                                          0.0*YAC_RAD},
                            .coordinates_y = (double[]){intersection.lat,
                                                        55.0*YAC_RAD,
                                                        45.0*YAC_RAD},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LON_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 3}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
    printf ("\n 12.20) some mean tests \n\n");
#endif
    /* cell a

       +------+  60
       |      |
       |      |
       |      |
       +------+ 40
      -20     20
    */

    /* cell b

            /| 50
          /  |
        /    | 40
        \    |
          \  |
            \| 30
      -10   10
    */

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-10.0*YAC_RAD,
                                                          10.0*YAC_RAD,
                                                          10.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          30.0*YAC_RAD,
                                                          50.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){-10.0*YAC_RAD,
                                                         10.0*YAC_RAD,
                                                         10.0*YAC_RAD},
                            .coordinates_y = (double[]){40.0*YAC_RAD,
                                                        50.0*YAC_RAD,
                                                        40.0*YAC_RAD},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE},
                            .num_corners = 3}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 12.21) some mean tests \n\n");
#endif
  /* cell a

     +------+  60
     |      |
     |      |
     |      |
     +------+ 40
    -20     20
  */

  /* cell b

            /| 80
          /  |
        /    |
      /      |
  61  --...__|
               59
   -20       21
  */

  {
    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  61.0*YAC_RAD},
                                     .points[1] = {.lon =  21.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  61.0*YAC_RAD},
                                     .points[1] = {.lon =  21.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 40.0*YAC_RAD},
                                     .points[1] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .edge_type = LON_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          21.0*YAC_RAD,
                                                          21.0*YAC_RAD},
                             .coordinates_y = (double[]){61.0*YAC_RAD,
                                                         80.0*YAC_RAD,
                                                         59.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){20.0*YAC_RAD,
                                                        intersections[0].lon,
                                                        intersections[1].lon},
                            .coordinates_y = (double[]){60.0*YAC_RAD,
                                                        intersections[0].lat,
                                                        intersections[1].lat},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LON_CIRCLE},
                            .num_corners = 3}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
    printf ("\n 12.22) some mean tests \n\n");
#endif
    /* cell a

       +------+  60
       |      |
       |      |
       |      |
       +------+ 40
      -20     20
    */

    /* cell b

         /\     x (slightly above 40)
        /  \
       /    \
      /      \
      --------  30
   -20   0   20
    */

  {
    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon =  0.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .points[1] = {.lon =  0.0*YAC_RAD,
                                                   .lat = 30.0*YAC_RAD},
                                     .edge_type = LON_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  0.0*YAC_RAD,
                                                   .lat = intersections[0].lat},
                                     .points[1] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 30.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  40.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                           0.0*YAC_RAD},
                             .coordinates_y = (double[]){30.0*YAC_RAD,
                                                         30.0*YAC_RAD,
                                                         intersections[0].lat},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){0.0*YAC_RAD,
                                                         intersections[1].lon,
                                                        -intersections[1].lon},
                             .coordinates_y = (double[]){intersections[0].lat,
                                                         40.0*YAC_RAD,
                                                         40.0*YAC_RAD},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 3}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
    printf ("\n 12.23) some mean tests \n\n");
#endif
    /* cell a

       +------+  60
       |      |
       |      |
       |      |
       +------+ 40
      -20     20
    */

    /* cell b

         /\     70
        /  \
       /    \
      /      \
      --------  59
   -21   0   21
    */

  {
    struct point intersections[4];

    if (!yac_intersect((struct edge){.points[0] = {.lon = -21.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .points[1] = {.lon =  21.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 40.0*YAC_RAD},
                                     .points[1] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .edge_type = LON_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  0.0*YAC_RAD,
                                                   .lat = 70.0*YAC_RAD},
                                     .points[1] = {.lon = 21.0*YAC_RAD,
                                                   .lat = 59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 40.0*YAC_RAD},
                                     .points[1] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .edge_type = LON_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  0.0*YAC_RAD,
                                                   .lat = 70.0*YAC_RAD},
                                     .points[1] = {.lon = 21.0*YAC_RAD,
                                                   .lat = 59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = -20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .points[1] = {.lon =  20.0*YAC_RAD,
                                                   .lat =  60.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       intersections+2))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = -21.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .points[1] = {.lon =  21.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  0.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .points[1] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       intersections+3))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-21.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                          21.0*YAC_RAD},
                             .coordinates_y = (double[]){59.0*YAC_RAD,
                                                         70.0*YAC_RAD,
                                                         59.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){ 20.0*YAC_RAD,
                                                         20.0*YAC_RAD,
                                                         intersections[2].lon,
                                                         intersections[3].lon,
                                                        -intersections[3].lon,
                                                        -intersections[2].lon,
                                                        -20.0*YAC_RAD,
                                                        -20.0*YAC_RAD},
                            .coordinates_y = (double[]){intersections[0].lat,
                                                        intersections[1].lat,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        intersections[1].lat,
                                                        intersections[0].lat},
                            .coordinates_xyz = (double[8][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LON_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LON_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 8},
                           {.coordinates_x = (double[]){ 20.0*YAC_RAD,
                                                         20.0*YAC_RAD,
                                                         intersections[2].lon,
                                                         intersections[3].lon,
                                                        -intersections[3].lon,
                                                        -intersections[2].lon,
                                                        -20.0*YAC_RAD,
                                                        -20.0*YAC_RAD,
                                                        -intersections[3].lon,
                                                         intersections[3].lon,},
                            .coordinates_y = (double[]){intersections[0].lat,
                                                        intersections[1].lat,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        intersections[1].lat,
                                                        intersections[0].lat,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD},
                            .coordinates_xyz = (double[10][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LON_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LON_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 10},
                           {.coordinates_x = (double[]){ 20.0*YAC_RAD,
                                                         20.0*YAC_RAD,
                                                         intersections[2].lon,
                                                        -intersections[2].lon,
                                                        -20.0*YAC_RAD,
                                                        -20.0*YAC_RAD,
                                                        -intersections[3].lon,
                                                         intersections[3].lon,},
                            .coordinates_y = (double[]){intersections[0].lat,
                                                        intersections[1].lat,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        intersections[1].lat,
                                                        intersections[0].lat,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD},
                            .coordinates_xyz = (double[8][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LON_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LON_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 8}},
      // number of reference cells
      3);
  }

#ifdef VERBOSE
    printf ("\n 12.24) some mean tests \n\n");
#endif
    /* cell a

       +------+  60
       |      |
       |      |
       |      |
       +------+ 40
      -x      x      (x is choosen such that the upper corners of the
                      quadrangle touch the sloping edges of the triangle)
    */

    /* cell b

         /\     70
        /  \
       /    \
      /      \
      --------  59
   -21   0   21
    */

  {
    struct point intersections[3];

    if (!yac_intersect((struct edge){.points[0] = {.lon =  0.0*YAC_RAD,
                                                   .lat = 70.0*YAC_RAD},
                                     .points[1] = {.lon = 21.0*YAC_RAD,
                                                   .lat = 59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  0.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .points[1] = {.lon = 25.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = -21.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .points[1] = {.lon =  21.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  0.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .points[1] = {.lon = 25.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = -21.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .points[1] = {.lon =  21.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = intersections[0].lon,
                                                   .lat = 60.0*YAC_RAD},
                                     .points[1] = {.lon = intersections[0].lon,
                                                   .lat = 40.0*YAC_RAD},
                                     .edge_type = LON_CIRCLE},
                       intersections+2))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-intersections[0].lon,
                                                          intersections[0].lon,
                                                          intersections[0].lon,
                                                         -intersections[0].lon},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-21.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                          21.0*YAC_RAD},
                             .coordinates_y = (double[]){59.0*YAC_RAD,
                                                         70.0*YAC_RAD,
                                                         59.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){ intersections[2].lon,
                                                         intersections[0].lon,
                                                         intersections[1].lon,
                                                        -intersections[1].lon,
                                                        -intersections[0].lon,
                                                        -intersections[2].lon},
                            .coordinates_y = (double[]){intersections[2].lat,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        intersections[2].lat},
                            .coordinates_xyz = (double[9][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LON_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                LON_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 6},
                           {.coordinates_x = (double[]){ intersections[2].lon,
                                                         intersections[0].lon,
                                                         intersections[1].lon,
                                                        -intersections[1].lon,
                                                        -intersections[0].lon,
                                                        -intersections[2].lon,
                                                        -intersections[1].lon,
                                                         intersections[1].lon},
                            .coordinates_y = (double[]){intersections[2].lat,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        intersections[2].lat,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD},
                            .coordinates_xyz = (double[8][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LON_CIRCLE,
                                                                LAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                LON_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 8},
                           {.coordinates_x = (double[]){ intersections[2].lon,
                                                         intersections[0].lon,
                                                        -intersections[0].lon,
                                                        -intersections[2].lon,
                                                        -intersections[1].lon,
                                                         intersections[1].lon},
                            .coordinates_y = (double[]){intersections[2].lat,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD,
                                                        intersections[2].lat,
                                                        60.0*YAC_RAD,
                                                        60.0*YAC_RAD},
                            .coordinates_xyz = (double[6][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LON_CIRCLE,
                                                                LAT_CIRCLE,
                                                                LON_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 6}},
      // number of reference cells
      3);
  }

#ifdef VERBOSE
  printf ("\n 12.25) some mean tests \n\n");
#endif
  /* cell a

     +------+  60
     |      |
     |      |
     |      |
     +------+ 40
    -20     20
  */

  /* cell b

        --------  59
        \      /
         \    /
          \  /
           \/     50
     -21   0   21
  */

  {
    struct point intersections[3];

    if (!yac_intersect((struct edge){.points[0] = {.lon = 21.0*YAC_RAD,
                                                   .lat = 59.0*YAC_RAD},
                                     .points[1] = {.lon =-21.0*YAC_RAD,
                                                   .lat = 59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  0.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .points[1] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  21.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .points[1] = {.lon = -21.0*YAC_RAD,
                                                   .lat =  59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 40.0*YAC_RAD},
                                     .points[1] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .edge_type = LON_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    if (!yac_intersect((struct edge){.points[0] = {.lon =  0.0*YAC_RAD,
                                                   .lat = 50.0*YAC_RAD},
                                     .points[1] = {.lon = 21.0*YAC_RAD,
                                                   .lat = 59.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 40.0*YAC_RAD},
                                     .points[1] = {.lon = 20.0*YAC_RAD,
                                                   .lat = 60.0*YAC_RAD},
                                     .edge_type = LON_CIRCLE},
                       intersections+2))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                          20.0*YAC_RAD,
                                                         -20.0*YAC_RAD},
                             .coordinates_y = (double[]){ 40.0*YAC_RAD,
                                                          40.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                          60.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){-21.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                          21.0*YAC_RAD},
                             .coordinates_y = (double[]){59.0*YAC_RAD,
                                                         50.0*YAC_RAD,
                                                         59.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){ intersections[0].lon,
                                                         20.0*YAC_RAD,
                                                         20.0*YAC_RAD,
                                                          0.0*YAC_RAD,
                                                        -20.0*YAC_RAD,
                                                        -20.0*YAC_RAD,
                                                        -intersections[0].lon},
                            .coordinates_y = (double[]){60.0*YAC_RAD,
                                                        intersections[1].lat,
                                                        intersections[2].lat,
                                                        50.0*YAC_RAD,
                                                        intersections[2].lat,
                                                        intersections[1].lat,
                                                        60.0*YAC_RAD},
                            .coordinates_xyz = (double[7][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LON_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LON_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE},
                            .num_corners = 7}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 12.26) some mean tests \n\n");
#endif
  /* cell a (circle around a pole)

         .--.
        /    \
        \    /
         '--'
  */

  /* cell b (circle around a pole; bigger than cell a)

       .-""-.
      /      \  
     ;        ;
      \      /
       '-..-'
  */

  {

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){  0.0*YAC_RAD,
                                                          72.0*YAC_RAD,
                                                         144.0*YAC_RAD,
                                                         216.0*YAC_RAD,
                                                         288.0*YAC_RAD},
                             .coordinates_y = (double[]){ 85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD},
                             .coordinates_xyz = (double[5][3]){{-1}},
                             .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE},
                             .num_corners = 5},
                            {.coordinates_x = (double[]){  0.0*YAC_RAD,
                                                          72.0*YAC_RAD,
                                                         144.0*YAC_RAD,
                                                         216.0*YAC_RAD,
                                                         288.0*YAC_RAD},
                             .coordinates_y = (double[]){ 80.0*YAC_RAD,
                                                          80.0*YAC_RAD,
                                                          80.0*YAC_RAD,
                                                          80.0*YAC_RAD,
                                                          80.0*YAC_RAD},
                             .coordinates_xyz = (double[5][3]){{-1}},
                             .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE},
                             .num_corners = 5}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){  0.0*YAC_RAD,
                                                          72.0*YAC_RAD,
                                                         144.0*YAC_RAD,
                                                         216.0*YAC_RAD,
                                                         288.0*YAC_RAD},
                             .coordinates_y = (double[]){ 85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD},
                             .coordinates_xyz = (double[5][3]){{-1}},
                             .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE},
                             .num_corners = 5}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 12.27) some mean tests \n\n");
#endif
  /* cell a (circle around a pole)

       .-""-.
      /      \  
     ;        ;
      \      /
       '-..-'
  */

  /* cell b (circle around a pole; same size as cell a)

       .-""-.
      /      \  
     ;        ;
      \      /
       '-..-'
  */

  {

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){  0.0*YAC_RAD,
                                                          72.0*YAC_RAD,
                                                         144.0*YAC_RAD,
                                                         216.0*YAC_RAD,
                                                         288.0*YAC_RAD},
                             .coordinates_y = (double[]){ 85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD},
                             .coordinates_xyz = (double[5][3]){{-1}},
                             .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE},
                             .num_corners = 5},
                            {.coordinates_x = (double[]){  0.0*YAC_RAD,
                                                          72.0*YAC_RAD,
                                                         144.0*YAC_RAD,
                                                         216.0*YAC_RAD,
                                                         288.0*YAC_RAD},
                             .coordinates_y = (double[]){ 85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD},
                             .coordinates_xyz = (double[5][3]){{-1}},
                             .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE},
                             .num_corners = 5}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){  0.0*YAC_RAD,
                                                          72.0*YAC_RAD,
                                                         144.0*YAC_RAD,
                                                         216.0*YAC_RAD,
                                                         288.0*YAC_RAD},
                             .coordinates_y = (double[]){ 85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD},
                             .coordinates_xyz = (double[5][3]){{-1}},
                             .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE},
                             .num_corners = 5}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 12.28) some mean tests \n\n");
#endif
  /* cell a (circle around a pole)

       .-""-.
      /      \  
     ;        ;
      \      /
       '-..-'
  */

  /* cell b (circle around different pole than cell a)

       .-""-.
      /      \  
     ;        ;
      \      /
       '-..-'
  */

  {

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){  0.0*YAC_RAD,
                                                          72.0*YAC_RAD,
                                                         144.0*YAC_RAD,
                                                         216.0*YAC_RAD,
                                                         288.0*YAC_RAD},
                             .coordinates_y = (double[]){ 85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD},
                             .coordinates_xyz = (double[5][3]){{-1}},
                             .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE},
                             .num_corners = 5},
                            {.coordinates_x = (double[]){  0.0*YAC_RAD,
                                                          72.0*YAC_RAD,
                                                         144.0*YAC_RAD,
                                                         216.0*YAC_RAD,
                                                         288.0*YAC_RAD},
                             .coordinates_y = (double[]){-85.0*YAC_RAD,
                                                         -85.0*YAC_RAD,
                                                         -85.0*YAC_RAD,
                                                         -85.0*YAC_RAD,
                                                         -85.0*YAC_RAD},
                             .coordinates_xyz = (double[5][3]){{-1}},
                             .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE},
                             .num_corners = 5}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = NULL,
                             .coordinates_y = NULL,
                             .coordinates_xyz = NULL,
                             .edge_type = NULL,
                             .num_corners = 0}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 12.29) some mean tests \n\n");
#endif
  /* cell a (circle around a pole; made of lat circle edges)

       .-""-.
      /      \  
     ;        ;
      \      /
       '-..-'
  */

  /* cell b (circle around a pole; made of great circle edges;
             corners are on the lat circles)

        ____
       /    \
      /      \  
     *        *
      \      /
       \____/
  */

  {

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){  0.0*YAC_RAD,
                                                          72.0*YAC_RAD,
                                                         144.0*YAC_RAD,
                                                         216.0*YAC_RAD,
                                                         288.0*YAC_RAD},
                             .coordinates_y = (double[]){ 85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD},
                             .coordinates_xyz = (double[5][3]){{-1}},
                             .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE},
                             .num_corners = 5},
                            {.coordinates_x = (double[]){  0.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                         120.0*YAC_RAD,
                                                         180.0*YAC_RAD,
                                                         240.0*YAC_RAD,
                                                         300.0*YAC_RAD},
                             .coordinates_y = (double[]){ 85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD},
                             .coordinates_xyz = (double[6][3]){{-1}},
                             .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                 GREAT_CIRCLE,
                                                                 GREAT_CIRCLE,
                                                                 GREAT_CIRCLE,
                                                                 GREAT_CIRCLE,
                                                                 GREAT_CIRCLE},
                             .num_corners = 6}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){  0.0*YAC_RAD,
                                                         60.0*YAC_RAD,
                                                        120.0*YAC_RAD,
                                                        180.0*YAC_RAD,
                                                        240.0*YAC_RAD,
                                                        300.0*YAC_RAD},
                            .coordinates_y = (double[]){ 85.0*YAC_RAD,
                                                         85.0*YAC_RAD,
                                                         85.0*YAC_RAD,
                                                         85.0*YAC_RAD,
                                                         85.0*YAC_RAD,
                                                         85.0*YAC_RAD},
                            .coordinates_xyz = (double[6][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 6}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 12.30) some mean tests \n\n");
#endif
  /* cell a (circle around a pole; made of lat circle edges)

       .-""-.
      /      \  
     ;        ;
      \      /
       '-..-'
  */

  /* cell b (circle around a pole; made of great circle edges;
             all great circles edges intersect with the lat circles twice)

        ____
       /    \
      /      \  
     *        *
      \      /
       \____/
  */

  {

    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon =  0.0*YAC_RAD,
                                                   .lat = 85.5*YAC_RAD},
                                     .points[1] = {.lon = 30.0*YAC_RAD,
                                                   .lat = 85.5*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  0.0*YAC_RAD,
                                                   .lat = 85.0*YAC_RAD},
                                     .points[1] = {.lon = 60.0*YAC_RAD,
                                                   .lat = 85.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = 30.0*YAC_RAD,
                                                   .lat = 85.5*YAC_RAD},
                                     .points[1] = {.lon = 60.0*YAC_RAD,
                                                   .lat = 85.5*YAC_RAD},
                                     .edge_type = LAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  0.0*YAC_RAD,
                                                   .lat = 85.0*YAC_RAD},
                                     .points[1] = {.lon = 60.0*YAC_RAD,
                                                   .lat = 85.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){  0.0*YAC_RAD,
                                                          72.0*YAC_RAD,
                                                         144.0*YAC_RAD,
                                                         216.0*YAC_RAD,
                                                         288.0*YAC_RAD},
                             .coordinates_y = (double[]){ 85.5*YAC_RAD,
                                                          85.5*YAC_RAD,
                                                          85.5*YAC_RAD,
                                                          85.5*YAC_RAD,
                                                          85.5*YAC_RAD},
                             .coordinates_xyz = (double[5][3]){{-1}},
                             .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE,
                                                                 LAT_CIRCLE},
                             .num_corners = 5},
                            {.coordinates_x = (double[]){  0.0*YAC_RAD,
                                                          60.0*YAC_RAD,
                                                         120.0*YAC_RAD,
                                                         180.0*YAC_RAD,
                                                         240.0*YAC_RAD,
                                                         300.0*YAC_RAD},
                             .coordinates_y = (double[]){ 85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD,
                                                          85.0*YAC_RAD},
                             .coordinates_xyz = (double[6][3]){{-1}},
                             .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                 GREAT_CIRCLE,
                                                                 GREAT_CIRCLE,
                                                                 GREAT_CIRCLE,
                                                                 GREAT_CIRCLE,
                                                                 GREAT_CIRCLE},
                             .num_corners = 6}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x =
                             (double[]){intersections[0].lon+60.0*0.0*YAC_RAD,
                                        intersections[1].lon+60.0*0.0*YAC_RAD,
                                        intersections[0].lon+60.0*1.0*YAC_RAD,
                                        intersections[1].lon+60.0*1.0*YAC_RAD,
                                        intersections[0].lon+60.0*2.0*YAC_RAD,
                                        intersections[1].lon+60.0*2.0*YAC_RAD,
                                        intersections[0].lon+60.0*3.0*YAC_RAD,
                                        intersections[1].lon+60.0*3.0*YAC_RAD,
                                        intersections[0].lon+60.0*4.0*YAC_RAD,
                                        intersections[1].lon+60.0*4.0*YAC_RAD,
                                        intersections[0].lon+60.0*5.0*YAC_RAD,
                                        intersections[1].lon+60.0*5.0*YAC_RAD},
                            .coordinates_y = (double[]){ 85.5*YAC_RAD,
                                                         85.5*YAC_RAD,
                                                         85.5*YAC_RAD,
                                                         85.5*YAC_RAD,
                                                         85.5*YAC_RAD,
                                                         85.5*YAC_RAD,
                                                         85.5*YAC_RAD,
                                                         85.5*YAC_RAD,
                                                         85.5*YAC_RAD,
                                                         85.5*YAC_RAD,
                                                         85.5*YAC_RAD,
                                                         85.5*YAC_RAD},
                            .coordinates_xyz = (double[12][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE},
                            .num_corners = 12}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 13.) example taken from bug report by Uwe \n\n");
#endif

  /* 13.) source cell test data

     +------+ 5
     |      |
     |      |
     |      |
     +------+ 0
    355    360
  */

  /* target cell test data

     +------+   0
     |      |
     |      |
     |      |
     +------+ -10
    240    250
  */

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){355.0*YAC_RAD,
                                                         360.0*YAC_RAD,
                                                         360.0*YAC_RAD,
                                                         355.0*YAC_RAD},
                             .coordinates_y = (double[]){  0.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                           5.0*YAC_RAD,
                                                           5.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){240.0*YAC_RAD,
                                                         250.0*YAC_RAD,
                                                         250.0*YAC_RAD,
                                                         240.0*YAC_RAD},
                             .coordinates_y = (double[]){-10.0*YAC_RAD,
                                                         -10.0*YAC_RAD,
                                                           0.0*YAC_RAD,
                                                           0.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 14.1) example taken from icon toy \n\n");
#endif

  /* 14.1) source cell test data

     +------+ 2
     |      |
     |      |
     |      |
     +------+ 0
     0      2
  */

  /* target cell test data

     _______
    |      /  1
    |    /
    |  /
    |/       -1
    2      4
  */

  {
    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){ 2.0*YAC_RAD,
                                                          2.0*YAC_RAD,
                                                          4.0*YAC_RAD},
                             .coordinates_y = (double[]){-1.0*YAC_RAD,
                                                          1.0*YAC_RAD,
                                                          1.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){0.0*YAC_RAD,
                                                         2.0*YAC_RAD,
                                                         2.0*YAC_RAD,
                                                         0.0*YAC_RAD},
                             .coordinates_y = (double[]){0.0*YAC_RAD,
                                                         0.0*YAC_RAD,
                                                         2.0*YAC_RAD,
                                                         2.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 14.2) example taken from icon toy \n\n");
#endif

  {
    struct point intersection;

    if (!yac_intersect((struct edge){.points[0] = {.lon =  1.0*YAC_RAD,
                                                   .lat = 90.0*YAC_RAD},
                                     .points[1] = {.lon =  1.0*YAC_RAD,
                                                   .lat = 80.0*YAC_RAD},
                                     .edge_type = LON_CIRCLE},
                       (struct edge){.points[0] = {.lon =  0.0*YAC_RAD,
                                                   .lat = 85.0*YAC_RAD},
                                     .points[1] = {.lon =  5.0*YAC_RAD,
                                                   .lat = 85.0*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       &intersection))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){  0.0*YAC_RAD,
                                                         180.0*YAC_RAD,
                                                           5.0*YAC_RAD},
                             .coordinates_y = (double[]){85.0*YAC_RAD,
                                                         90.0*YAC_RAD,
                                                         85.0*YAC_RAD},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){ 1.0*YAC_RAD,
                                                         -5.0*YAC_RAD,
                                                         -5.0*YAC_RAD,
                                                          1.0*YAC_RAD},
                             .coordinates_y = (double[]){80.0*YAC_RAD,
                                                         80.0*YAC_RAD,
                                                         90.0*YAC_RAD,
                                                         90.0*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){1.0*YAC_RAD,
                                                        0.0*YAC_RAD,
                                                        0.0*YAC_RAD},
                            .coordinates_y = (double[]){intersection.lat,
                                                        85.0*YAC_RAD,
                                                        90.0*YAC_RAD},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LON_CIRCLE},
                            .num_corners = 3}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 14.3) example taken from icon toy \n\n");
#endif

  {

    struct point intersections[4];

    if (!yac_intersect((struct edge){.points[0] = {.lon =  0.46633015951723494341,
                                                   .lat = -0.88357293382212931387},
                                     .points[1] = {.lon =  0.49087385212340517437,
                                                   .lat = -0.88357293382212931387},
                                     .edge_type = LAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  0.47953965147452792817,
                                                   .lat = -0.90109638077934106626},
                                     .points[1] = {.lon =  0.46519399935013672209,
                                                   .lat = -0.88122392525212978054},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  0.46633015951723494341,
                                                   .lat = -0.88357293382212931387},
                                     .points[1] = {.lon =  0.46633015951723494341,
                                                   .lat = -0.85902924121595902740},
                                     .edge_type = LON_CIRCLE},
                       (struct edge){.points[0] = {.lon =  0.47953965147452792817,
                                                   .lat = -0.90109638077934106626},
                                     .points[1] = {.lon =  0.46519399935013672209,
                                                   .lat = -0.88122392525212978054},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  0.46633015951723494341,
                                                   .lat = -0.88357293382212931387},
                                     .points[1] = {.lon =  0.46633015951723494341,
                                                   .lat = -0.85902924121595902740},
                                     .edge_type = LON_CIRCLE},
                       (struct edge){.points[0] = {.lon =  0.50100250912415544846,
                                                   .lat = -0.88440597750634275531},
                                     .points[1] = {.lon =  0.46519399935013672209,
                                                   .lat = -0.88122392525212978054},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+2))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  0.49087385212340517437,
                                                   .lat = -0.88357293382212931387},
                                     .points[1] = {.lon =  0.49087385212340517437,
                                                   .lat = -0.85902924121595902740},
                                     .edge_type = LON_CIRCLE},
                       (struct edge){.points[0] = {.lon =  0.50100250912415544846,
                                                   .lat = -0.88440597750634275531},
                                     .points[1] = {.lon =  0.46519399935013672209,
                                                   .lat = -0.88122392525212978054},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+3))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){ 0.46633015951723494341,
                                                          0.49087385212340517437,
                                                          0.49087385212340517437,
                                                          0.46633015951723494341},
                             .coordinates_y = (double[]){-0.88357293382212931387,
                                                         -0.88357293382212931387,
                                                         -0.85902924121595902740,
                                                         -0.85902924121595902740},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){ 0.47953965147452792817,
                                                          0.50100250912415544846,
                                                          0.46519399935013672209},
                             .coordinates_y = (double[]){-0.90109638077934106626,
                                                         -0.88440597750634275531,
                                                         -0.88122392525212978054},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){0.49087385212340517437,
                                                        intersections[0].lon,
                                                        0.46633015951723494341,
                                                        0.46633015951723494341,
                                                        0.49087385212340517437},
                            .coordinates_y = (double[]){-0.88357293382212931387,
                                                        intersections[0].lat,
                                                        intersections[1].lat,
                                                        intersections[2].lat,
                                                        intersections[3].lat},
                            .coordinates_xyz = (double[5][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LON_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LON_CIRCLE},
                            .num_corners = 5}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 14.4) example provided by Uwe \n\n");
#endif

  {

    struct point intersections[4];

    if (!yac_intersect((struct edge){.points[0] = {.lon = 180.25*YAC_RAD,
                                                   .lat = -89.75*YAC_RAD},
                                     .points[1] = {.lon = 180.25*YAC_RAD,
                                                   .lat = -89.50*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 180.2*YAC_RAD,
                                                   .lat = -89.6*YAC_RAD},
                                     .points[1] = {.lon = 180.3*YAC_RAD,
                                                   .lat = -89.6*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = 180.25*YAC_RAD,
                                                   .lat = -89.75*YAC_RAD},
                                     .points[1] = {.lon = 180.25*YAC_RAD,
                                                   .lat = -89.50*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 180.3*YAC_RAD,
                                                   .lat = -89.5*YAC_RAD},
                                     .points[1] = {.lon = 180.2*YAC_RAD,
                                                   .lat = -89.5*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                     intersections+1))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = 180.25*YAC_RAD,
                                                   .lat = -89.5*YAC_RAD},
                                     .points[1] = {.lon = 180.00*YAC_RAD,
                                                   .lat = -89.5*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 180.3*YAC_RAD,
                                                   .lat = -89.5*YAC_RAD},
                                     .points[1] = {.lon = 180.2*YAC_RAD,
                                                   .lat = -89.5*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                         intersections+2))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = 180.25*YAC_RAD,
                                                   .lat = -89.5*YAC_RAD},
                                     .points[1] = {.lon = 180.00*YAC_RAD,
                                                   .lat = -89.5*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon = 180.2*YAC_RAD,
                                                   .lat = -89.5*YAC_RAD},
                                     .points[1] = {.lon = 180.2*YAC_RAD,
                                                   .lat = -89.6*YAC_RAD},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+3))
      return EXIT_FAILURE;

    test_clipping(
      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){180.2*YAC_RAD,
                                                         180.3*YAC_RAD,
                                                         180.3*YAC_RAD,
                                                         180.2*YAC_RAD},
                             .coordinates_y = (double[]){-89.6*YAC_RAD,
                                                         -89.6*YAC_RAD,
                                                         -89.5*YAC_RAD,
                                                         -89.5*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){180.00*YAC_RAD,
                                                         180.25*YAC_RAD,
                                                         180.25*YAC_RAD,
                                                         180.00*YAC_RAD},
                             .coordinates_y = (double[]){-89.75*YAC_RAD,
                                                         -89.75*YAC_RAD,
                                                         -89.50*YAC_RAD,
                                                         -89.50*YAC_RAD},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){180.2*YAC_RAD,
                                                        180.25*YAC_RAD,
                                                        180.25*YAC_RAD,
                                                        intersections[2].lon,
                                                        180.20*YAC_RAD},
                            .coordinates_y = (double[]){-89.6*YAC_RAD,
                                                        intersections[0].lat,
                                                        intersections[1].lat,
                                                        intersections[2].lat,
                                                        intersections[3].lat},
                            .coordinates_xyz = (double[5][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 5}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 14.5) example taken from icon toy \n\n");
#endif

  {
    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon =  1.20264093770234,
                                                   .lat = -0.76085447079128},
                                     .points[1] = {.lon =  1.22718463030851,
                                                   .lat = -0.76085447079128},
                                     .edge_type = LAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  1.20552692320133,
                                                   .lat = -0.76075226373764},
                                     .points[1] = {.lon =  1.30774719967050,
                                                   .lat = -0.76075226373764},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  1.20264093770234,
                                                   .lat = -0.76085447079128},
                                     .points[1] = {.lon =  1.22718463030851,
                                                   .lat = -0.76085447079128},
                                     .edge_type = LAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  1.20552692320133,
                                                   .lat = -0.76075226373764},
                                     .points[1] = {.lon =  1.25663706143592,
                                                   .lat = -0.82925269516665},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    test_clipping(
      // input cells

      (struct grid_cell[2]){{.coordinates_x = (double[]){1.20264093770234,
                                                         1.22718463030851,
                                                         1.22718463030851,
                                                         1.20264093770234},
                             .coordinates_y = (double[]){-0.76085447079128,
                                                         -0.76085447079128,
                                                         -0.73631077818511,
                                                         -0.73631077818511},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){1.20552692320133,
                                                         1.25663706143592,
                                                         1.30774719967050},
                             .coordinates_y = (double[]){-0.76075226373764,
                                                         -0.82925269516665,
                                                         -0.76075226373763},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){1.20552692320133,
                                                        intersections[0].lon,
                                                        intersections[1].lon},
                            .coordinates_y = (double[]){-0.76075226373764,
                                                        intersections[0].lat,
                                                        intersections[1].lat},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 3},
                           {.coordinates_x = (double[]){1.20552692320133,
                                                        intersections[0].lon,
                                                        1.22718463030851,
                                                        intersections[1].lon},
                            .coordinates_y = (double[]){-0.76075226373764,
                                                        intersections[0].lat,
                                                        -0.76085447079128,
                                                        intersections[1].lat},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 4}},
      // number of reference cells
      2);
  }

#ifdef VERBOSE
  printf ("\n 14.6) example in which all coordinates are 0 (occurred in a test"
          " of Uwe due to erroneous input data) \n\n");
#endif

  {
    test_clipping(
      // input cells

      (struct grid_cell[2]){{.coordinates_x = (double[]){0, 0, 0, 0},
                             .coordinates_y = (double[]){0, 0, 0, 0},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){0, 0, 0},
                             .coordinates_y = (double[]){0, 0, 0},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 14.7) example example take from ICON_toy \n\n");
#endif
  {
    test_clipping(
      // input cells

      (struct grid_cell[2]){{.coordinates_x = (double[]){ 3.1415926372157803,
                                                          0.9682712007396151,
                                                         -1.8849555862661402},
                             .coordinates_y = (double[]){-1.5418135425481947,
                                                         -1.5707963262077691,
                                                         -1.5418135428532602},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){1.2517283229146832,
                                                         1.2762720155208536,
                                                         1.2762720155208536,
                                                         1.2517283229146832},
                             .coordinates_y = (double[]){-1.5707963267948966,
                                                         -1.5707963267948966,
                                                         -1.5462526341887264,
                                                         -1.5462526341887264},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 14.8) example take from ICON_toy \n\n");
#endif
  {
    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon =  0.9682712007396151,
                                                   .lat = -1.5707963262077691},
                                     .points[1] = {.lon =  3.1415926372157803,
                                                   .lat = -1.5418135425481947},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  3.1170489609836229,
                                                   .lat = -1.5462526341887264},
                                     .points[1] = {.lon =  3.1415926535897931,
                                                   .lat = -1.5462526341887264},
                                     .edge_type = LAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = -1.8849555862661402,
                                                   .lat = -1.5418135428532602},
                                     .points[1] = {.lon =  3.1415926372157803,
                                                   .lat = -1.5418135425481947},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  3.1415926535897931,
                                                   .lat = -1.5462526341887264},
                                     .points[1] = {.lon =  3.1415926535897931,
                                                   .lat = -1.5217089415825560},
                                     .edge_type = LON_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    test_clipping(
      // input cells

      (struct grid_cell[2]){{.coordinates_x = (double[]){ 3.1415926372157803,
                                                          0.9682712007396151,
                                                         -1.8849555862661402},
                             .coordinates_y = (double[]){-1.5418135425481947,
                                                         -1.5707963262077691,
                                                         -1.5418135428532602},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){ 3.1170489609836229,
                                                          3.1415926535897931,
                                                          3.1415926535897931,
                                                          3.1170489609836229},
                             .coordinates_y = (double[]){-1.5462526341887264,
                                                         -1.5462526341887264,
                                                         -1.5217089415825560,
                                                         -1.5217089415825560},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){ 3.1415926535897931,
                                                        intersections[0].lon,
                                                         3.1415926372157803,
                                                        intersections[1].lon},
                            .coordinates_y = (double[]){-1.5462526341887264,
                                                        intersections[0].lat,
                                                        -1.5418135425481947,
                                                        intersections[1].lat},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LON_CIRCLE},
                            .num_corners = 4},
                           {.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0}},
      // number of reference cells
      2);
  }

#ifdef VERBOSE
  printf ("\n 14.9) example example take from ICON_toy \n\n");
#endif
  {
    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon = -1.2564300972590368,
                                                   .lat =  1.5418206553283160},
                                     .points[1] = {.lon =  1.2277260313871721,
                                                   .lat =  1.5707871564758291},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  0.0000000000000000,
                                                   .lat =  1.5462526341887264},
                                     .points[1] = {.lon =  0.0000000000000000,
                                                   .lat =  1.5707963267948966},
                                     .edge_type = LON_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = -1.2564300972590368,
                                                   .lat =  1.5418206553283160},
                                     .points[1] = {.lon =  1.2277260313871721,
                                                   .lat =  1.5707871564758291},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  0.0245436926061703,
                                                   .lat =  1.5462526341887264},
                                     .points[1] = {.lon =  0.0245436926061703,
                                                   .lat =  1.5707963267948966},
                                     .edge_type = LON_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    test_clipping(

      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-1.2564300972590368,
                                                          1.2277260313871721,
                                                         -2.5134348157836111},
                             .coordinates_y = (double[]){1.5418206553283160,
                                                         1.5707871564758291,
                                                         1.5418209335124486},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){0.0000000000000000,
                                                         0.0245436926061703,
                                                         0.0245436926061703,
                                                         0.0000000000000000},
                             .coordinates_y = (double[]){1.5462526341887264,
                                                         1.5462526341887264,
                                                         1.5707963267948966,
                                                         1.5707963267948966},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){0,
                                                        intersections[0].lon,
                                                        intersections[1].lon},
                            .coordinates_y = (double[]){M_PI_2,
                                                        intersections[0].lat,
                                                        intersections[1].lat},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LON_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LON_CIRCLE},
                            .num_corners = 3}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 14.10) example example take from ICON_toy \n\n");
#endif
  {
    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon = -2.1798479726998332,
                                                   .lat = -0.0311624518327736},
                                     .points[1] = {.lon = -2.1991148570983681,
                                                   .lat =  0.0000000043633749},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  4.0742529726242633,
                                                   .lat =  0.0000000000000000},
                                     .points[1] = {.lon =  4.0987966652304335,
                                                   .lat =  0.0000000000000000},
                                     .edge_type = LAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon = -2.2190472361409284,
                                                   .lat = -0.0338654325037920},
                                     .points[1] = {.lon = -2.1991148570983681,
                                                   .lat =  0.0000000043633749},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  4.0742529726242633,
                                                   .lat =  0.0000000000000000},
                                     .points[1] = {.lon =  4.0987966652304335,
                                                   .lat =  0.0000000000000000},
                                     .edge_type = LAT_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    test_clipping(

      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-2.1798479726998332,
                                                         -2.1991148570983681,
                                                         -2.2190472361409284},
                             .coordinates_y = (double[]){-0.0311624518327736,
                                                          0.0000000043633749,
                                                         -0.0338654325037920},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){4.0742529726242633,
                                                         4.0987966652304335,
                                                         4.0987966652304335,
                                                         4.0742529726242633},
                             .coordinates_y = (double[]){0.0000000000000000,
                                                         0.0000000000000000,
                                                         0.0245436926061703,
                                                         0.0245436926061703},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){-2.1991148570983681,
                                                        intersections[0].lon,
                                                        intersections[1].lon},
                            .coordinates_y = (double[]){0.0000000043633749,
                                                        0.0000000000000000,
                                                        0.0000000000000000},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 3}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 14.11) example example take from ICON_toy \n\n");
#endif
  {
    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon =  3.1415926521557882,
                                                   .lat = -0.9805860393425995},
                                     .points[1] = {.lon =  3.1415926520689506,
                                                   .lat = -1.0172219678978514},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  3.1415926535897931,
                                                   .lat = -1.0062913968529805},
                                     .points[1] = {.lon =  3.1170489609836229,
                                                   .lat = -1.0062913968529805},
                                     .edge_type = LAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  3.1415926521557882,
                                                   .lat = -0.9805860393425995},
                                     .points[1] = {.lon =  3.1415926520689506,
                                                   .lat = -1.0172219678978514},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  3.1415926535897931,
                                                   .lat = -0.9817477042468103},
                                     .points[1] = {.lon =  3.1170489609836229,
                                                   .lat = -0.9817477042468103},
                                     .edge_type = LAT_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    test_clipping(

      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){ 3.1415926521557882,
                                                          3.1415926520689506,
                                                         -3.0774466652769172},
                             .coordinates_y = (double[]){-0.9805860393425995,
                                                         -1.0172219678978514,
                                                         -0.9979427097227050},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){ 3.1170489609836229,
                                                          3.1415926535897931,
                                                          3.1415926535897931,
                                                          3.1170489609836229},
                             .coordinates_y = (double[]){-1.0062913968529805,
                                                         -1.0062913968529805,
                                                         -0.9817477042468103,
                                                         -0.9817477042468103},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){3.1415926535897931,
                                                        3.1415926535897931,
                                                        intersections[0].lon,
                                                        intersections[1].lon},
                            .coordinates_y = (double[]){-0.9817477042468103,
                                                        -1.0062913968529805,
                                                        -1.0062913968529805,
                                                        -0.9817477042468103},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LON_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LAT_CIRCLE},
                            .num_corners = 4},
                           {.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0}},
      // number of reference cells
      2);
  }

#ifdef VERBOSE
  printf ("\n 14.12) example example take from ICON_toy \n\n");
#endif
  {

    test_clipping(

      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){ 1.8849555814614660,
                                                          0.9682712007396151,
                                                          3.1415926372157803},
                             .coordinates_y = (double[]){-1.5418135420482968,
                                                         -1.5707963262077691,
                                                         -1.5418135425481947},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){ 1.0799224746714915,
                                                          1.1044661672776617,
                                                          1.1044661672776617,
                                                          1.0799224746714915},
                             .coordinates_y = (double[]){-1.5707963267948966,
                                                         -1.5707963267948966,
                                                         -1.5462526341887264,
                                                         -1.5462526341887264},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 14.13) example example take from ICON_toy \n\n");
#endif
  {

    test_clipping(

      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){ 3.1415926372157803,
                                                          0.9682712007396151,
                                                         -1.8849555862661402},
                             .coordinates_y = (double[]){-1.5418135425481947,
                                                         -1.5707963262077691,
                                                         -1.5418135428532602},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){1.5707963267948966,
                                                         1.5953400194010670,
                                                         1.5953400194010670,
                                                         1.5707963267948966},
                             .coordinates_y = (double[]){-1.5707963267948966,
                                                         -1.5707963267948966,
                                                         -1.5462526341887264,
                                                         -1.5462526341887264},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 14.14) example example take from ICON_toy \n\n");
#endif
  {
    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon =  1.25664579522418,
                                                   .lat =  1.54180416508220},
                                     .points[1] = {.lon =  0.00031736492867,
                                                   .lat =  1.54181036404629},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  0.90811662642830,
                                                   .lat =  1.54625263418873},
                                     .points[1] = {.lon =  0.93266031903447,
                                                   .lat =  1.54625263418873},
                                     .edge_type = LAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  1.25664579522418,
                                                   .lat =  1.54180416508220},
                                     .points[1] = {.lon =  0.00031736492867,
                                                   .lat =  1.54181036404629},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  0.93266031903447,
                                                   .lat =  1.52170894158256},
                                     .points[1] = {.lon =  0.93266031903447,
                                                   .lat =  1.54625263418873},
                                     .edge_type = LON_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    test_clipping(

      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){1.25664579522418,
                                                         1.22772603138717,
                                                         0.00031736492867},
                             .coordinates_y = (double[]){1.54180416508220,
                                                         1.57078715647583,
                                                         1.54181036404629},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){0.90811662642830,
                                                         0.93266031903447,
                                                         0.93266031903447,
                                                         0.90811662642830},
                             .coordinates_y = (double[]){1.52170894158256,
                                                         1.52170894158256,
                                                         1.54625263418873,
                                                         1.54625263418873},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){0.93266031903447,
                                                        intersections[0].lon,
                                                        intersections[1].lon},
                            .coordinates_y = (double[]){1.54625263418873,
                                                        intersections[0].lat,
                                                        intersections[1].lat},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LON_CIRCLE},
                            .num_corners = 3},
                           {.coordinates_x = (double[]){0.93266031903447,
                                                        0.90811662642830,
                                                        intersections[0].lon,
                                                        intersections[1].lon},
                            .coordinates_y = (double[]){1.54625263418873,
                                                        1.54625263418873,
                                                        intersections[0].lat,
                                                        intersections[1].lat},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                LAT_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LON_CIRCLE},
                            .num_corners = 4}},
      // number of reference cells
      2);
  }

#ifdef VERBOSE
  printf ("\n 14.15) example example take from ICON_toy \n\n");
#endif
  {
    test_clipping(

      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){4.9999997254688875,
                                                         4.9999997254688875,
                                                         5.0078539201557488,
                                                         5.0078539201557488},
                             .coordinates_y = (double[]){-1.2878426515089207,
                                                         -1.2946756570757916,
                                                         -1.2946797849754812,
                                                         -1.2878469125666649},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){5.0069132916587327,
                                                         5.0130492148102759,
                                                         5.0130492148102759,
                                                         5.0069132916587327},
                             .coordinates_y = (double[]){-1.2946797849754812,
                                                         -1.2946797849754812,
                                                         -1.2885438618239387,
                                                         -1.2885438618239387},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){5.0069132916587327,
                                                        5.0078539201557488,
                                                        5.0078539201557488,
                                                        5.0069132916587327},
                            .coordinates_y = (double[]){-1.2946797849754812,
                                                        -1.2946797849754812,
                                                        -1.2885438618239387,
                                                        -1.2885438618239387},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                            GREAT_CIRCLE,
                                                            LAT_CIRCLE,
                                                            LON_CIRCLE},
                            .num_corners = 4}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 15) example provided by Uwe \n\n");
#endif
  {

    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon =  2.5132741228718345,
                                                   .lat =  1.5494144670074232},
                                     .points[1] = {.lon =  3.7699111843077517,
                                                   .lat =  1.5494144670074232},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  3.2637657012293961,
                                                   .lat =  1.5533430342749532},
                                     .points[1] = {.lon =  3.2812189937493392,
                                                   .lat =  1.5533430342749532},
                                     .edge_type = LAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  2.5132741228718345,
                                                   .lat =  1.5494144670074232},
                                     .points[1] = {.lon =  3.7699111843077517,
                                                   .lat =  1.5494144670074232},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  3.2812189937493392,
                                                   .lat =  1.5533430342749532},
                                     .points[1] = {.lon =  3.2812189937493392,
                                                   .lat =  1.5358897417550099},
                                     .edge_type = LON_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    test_clipping(

      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){3.2637657012293961,
                                                         3.2812189937493392,
                                                         3.2812189937493392,
                                                         3.2637657012293961},
                             .coordinates_y = (double[]){1.5358897417550099,
                                                         1.5358897417550099,
                                                         1.5533430342749532,
                                                         1.5533430342749532},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){1.2566370614359175,
                                                         2.5132741228718345,
                                                         3.7699111843077517,
                                                         5.0265482457436690,
                                                         0.0000000000000000,
                                                         0.0000000000000000},
                             .coordinates_y = (double[]){1.5494144670074232,
                                                         1.5494144670074232,
                                                         1.5494144670074232,
                                                         1.5494144670074232,
                                                         1.5494144670074232,
                                                         1.5494144670074232},
                             .coordinates_xyz = (double[6][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 6}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){intersections[0].lon,
                                                        3.2812189937493392,
                                                        3.2812189937493392},
                            .coordinates_y = (double[]){1.5533430342749532,
                                                        1.5533430342749532,
                                                        intersections[1].lat},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                LON_CIRCLE,
                                                                GREAT_CIRCLE},
                            .num_corners = 3}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 16.1) example provided by Rene \n\n");
#endif
  {

    test_clipping(

      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){1.8849563731522099,
                                                         1.8465416851483734,
                                                         1.8661276404324745},
                             .coordinates_y = (double[]){-0.4124872765902973,
                                                         -0.4149531265147232,
                                                         -0.4401304121445256},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){1.8443992887189797,
                                                         1.8849555921538759,
                                                         1.9255118955887724},
                             .coordinates_y = (double[]){-0.4121554426487201,
                                                         -0.4636476090008059,
                                                         -0.4121554426487201},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){1.8849563731522099,
                                                        1.8465416851483734,
                                                        1.8661276404324745},
                            .coordinates_y = (double[]){-0.4124872765902973,
                                                        -0.4149531265147232,
                                                        -0.4401304121445256},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 3}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 16.2) example provided by Rene \n\n");
#endif
  {
    struct point intersection;

    if (!yac_intersect((struct edge){.points[0] = {.lon =  -1.9037843097448526,
                                                   .lat =  -0.4401294427665800},
                                     .points[1] = {.lon =  -1.8849555925510342,
                                                   .lat =  -0.4636476094897370},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  -1.8849555921538759,
                                                   .lat =  -0.4636476090008059},
                                     .points[1] = {.lon =  -1.9528559331227824,
                                                   .lat =  -0.4822665528104840},
                                     .edge_type = GREAT_CIRCLE},
                     &intersection))
      return EXIT_FAILURE;

    test_clipping(

      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-1.9037843097448526,
                                                         -1.8849555925510342,
                                                         -1.8661268754099334},
                             .coordinates_y = (double[]){-0.4401294427665800,
                                                         -0.4636476094897370,
                                                         -0.4401294427950335},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){-1.8849555921538759,
                                                         -1.8849555921538759,
                                                         -1.9528559331227824},
                             .coordinates_y = (double[]){-0.5268929705698896,
                                                         -0.4636476090008059,
                                                         -0.4822665528104840},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){-1.8849555921538759,
                                                        -1.8849555921538759,
                                                        intersection.lon},
                            .coordinates_y = (double[]){-0.4636476094897370,
                                                        -0.4636476090008059,
                                                        intersection.lat},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 3},
                           {.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0},
                            },
      // number of reference cells
      2);
  }

#ifdef VERBOSE
  printf ("\n 16.3) example provided by Rene \n\n");
#endif
  {

    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon =  -1.2566370614359172,
                                                   .lat =  -0.5535743588970451},
                                     .points[1] = {.lon =  -1.1715639552378114,
                                                   .lat =  -0.5519553785110131},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  -1.2566370573287293,
                                                   .lat =  -0.5535743562505872},
                                     .points[1] = {.lon =  -1.2346024154180213,
                                                   .lat =  -0.5882204740702072},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  -1.2566370614359172,
                                                   .lat =  -0.5535743588970451},
                                     .points[1] = {.lon =  -1.1715639552378114,
                                                   .lat =  -0.5519553785110131},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =  -1.2566370573287293,
                                                   .lat =  -0.5535743562505872},
                                     .points[1] = {.lon =  -1.2786717012447690,
                                                   .lat =  -0.5882204751704803},
                                     .edge_type = GREAT_CIRCLE},
                       intersections+1))
      return EXIT_FAILURE;

    test_clipping(

      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){-1.2346024154180213,
                                                         -1.2566370573287293,
                                                         -1.2786717012447690},
                             .coordinates_y = (double[]){-0.5882204740702072,
                                                         -0.5535743562505872,
                                                         -0.5882204751704803},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){-1.2156135107669752,
                                                         -1.2566370614359172,
                                                         -1.1715639552378114},
                             .coordinates_y = (double[]){-0.4835340719879115,
                                                         -0.5535743588970451,
                                                         -0.5519553785110131},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0},
                           {.coordinates_x = (double[]){-1.2566370573287293,
                                                        intersections[0].lon,
                                                        intersections[1].lon},
                            .coordinates_y = (double[]){-0.5535743562505872,
                                                        intersections[0].lat,
                                                        intersections[1].lat},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 3}},
      // number of reference cells
      2);
  }

#ifdef VERBOSE
  printf ("\n 17.1) handling of 'triangle' in which all corners on on a gc \n\n");
#endif
  {
    test_clipping(

      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){0, 1, 2},
                             .coordinates_y = (double[]){0, 0, 0},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){0, 1, 2},
                             .coordinates_y = (double[]){0, 0, 0},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0},
                           {.coordinates_x = (double[]){0, 1, 2},
                            .coordinates_y = (double[]){0, 0, 0},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 3}},
      // number of reference cells
      2);
  }

#ifdef VERBOSE
  printf ("\n 17.1) handling of 'triangle' in which all corners on on a gc \n\n");
#endif
  {
    test_clipping(

      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){0, 1, 1},
                             .coordinates_y = (double[]){0, 0, 1},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3},
                            {.coordinates_x = (double[]){0, 1, 2},
                             .coordinates_y = (double[]){0, 0, 0},
                             .coordinates_xyz = (double[3][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 3}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0},
                           {.coordinates_x = (double[]){0, 1, 2},
                            .coordinates_y = (double[]){0, 0, 0},
                            .coordinates_xyz = (double[3][3]){{-1}},
                            .edge_type = gc_edges,
                            .num_corners = 3}},
      // number of reference cells
      2);
  }

#ifdef VERBOSE
  printf ("\n 17.2)\n\n");
#endif
  {

    test_clipping(

      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){1.5462526341887264,
                                                         1.5707963267948966,
                                                         1.5707963267948966,
                                                         1.5462526341887264},
                             .coordinates_y = (double[]){-1.5707963267948966,
                                                         -1.5707963267948966,
                                                         -1.5462526341887264,
                                                         -1.5462526341887264},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){1.5707963267948966,
                                                         1.5707963267948966,
                                                         1.2491662739534428,
                                                         1.1072089649419674},
                             .coordinates_y = (double[]){-1.5462526341887277,
                                                         -1.5339807878856395,
                                                         -1.5319928455462422,
                                                         -1.5433578475477794},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = NULL,
                            .coordinates_y = NULL,
                            .coordinates_xyz = NULL,
                            .edge_type = NULL,
                            .num_corners = 0}},
      // number of reference cells
      1);
  }

#ifdef VERBOSE
  printf ("\n 17.3)\n\n");
#endif
  {

    struct point intersection;

    if (!yac_intersect((struct edge){.points[0] = {.lon =   0,
                                                   .lat =  -1.5339807878856395},
                                     .points[1] = {.lon =   0.32163005284145374,
                                                   .lat =  -1.5319928455462422},
                                     .edge_type = GREAT_CIRCLE},
                       (struct edge){.points[0] = {.lon =   0.024543692606170259,
                                                   .lat =  -1.5462526341887264},
                                     .points[1] = {.lon =   0.024543692606170259,
                                                   .lat =  -1.5217089415825564},
                                     .edge_type = LON_CIRCLE},
                       &intersection))
      return EXIT_FAILURE;

    test_clipping(

      // input cells
      (struct grid_cell[2]){{.coordinates_x = (double[]){0,
                                                         0.024543692606170259,
                                                         0.024543692606170259,
                                                         0},
                             .coordinates_y = (double[]){-1.5462526341887264,
                                                         -1.5462526341887264,
                                                         -1.5217089415825564,
                                                         -1.5217089415825564},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = latlon_edges,
                             .num_corners = 4},
                            {.coordinates_x = (double[]){0,
                                                         0.32163005284145374,
                                                         0.24480144176769322,
                                                         0},
                             .coordinates_y = (double[]){-1.5339807878856395,
                                                         -1.5319928455462422,
                                                         -1.5202029837469972,
                                                         -1.5217089415825564},
                             .coordinates_xyz = (double[4][3]){{-1}},
                             .edge_type = gc_edges,
                             .num_corners = 4}},
      // reference cells
      (struct grid_cell[]){{.coordinates_x = (double[]){0,
                                                        0.024543692606170259,
                                                        0.024543692606170259,
                                                        0},
                            .coordinates_y = (double[]){-1.5217089415825564,
                                                        -1.5217089415825564,
                                                        intersection.lat,
                                                        -1.5339807878856395},
                            .coordinates_xyz = (double[4][3]){{-1}},
                            .edge_type = (enum yac_edge_type[]){LAT_CIRCLE,
                                                                LON_CIRCLE,
                                                                GREAT_CIRCLE,
                                                                LON_CIRCLE},
                            .num_corners = 4}},
      // number of reference cells
      1);
  }

  yac_free_grid_cell(overlap_cell);
  yac_free_grid_cell(overlap_cell+1);

  return TEST_EXIT_CODE;
}

static int compare_cells(struct grid_cell a, struct grid_cell b) {

  double const tol = 1e-8;

  if ((yac_huiliers_area(a) < yac_angle_tol * yac_angle_tol) &&
      (yac_huiliers_area(b) < yac_angle_tol * yac_angle_tol)) return 0;

  if (a.num_corners != b.num_corners) return 1;

  if (a.num_corners == 0) return 0;

  for (int order = -1; order <= 1; order += 2) {
    for (int start = 0; start < (int)(a.num_corners); ++start) {

      int differences = 0;

      for (int i = 0; i < (int)(a.num_corners); ++i) {

        int j =
          ((int)(a.num_corners) + start + order * i) % (int)(a.num_corners);

        if (get_vector_angle(a.coordinates_xyz[i], b.coordinates_xyz[j]) > tol)
          ++differences;

        if (fabs(a.coordinates_y[i] - b.coordinates_y[j]) > tol)
          ++differences;
        else if ((fabs(fabs(a.coordinates_y[i]) - 90.0*YAC_RAD) > tol) &&
                 (get_angle(a.coordinates_x[i], b.coordinates_x[j]) > tol))
          ++differences;
        else if ((a.edge_type[i] == LAT_CIRCLE) !=
                 (b.edge_type[
                    (((int)(a.num_corners))+j-(order<0))%
                    ((int)(a.num_corners))] == LAT_CIRCLE))
          ++differences;
      }

      if (!differences) return 0;
    }
  }

  return 1;
}

static void test_clipping(struct grid_cell cells[2],
                          struct grid_cell * ref_cells,
                          unsigned num_ref_cells) {

  int order[2], start[2];
  double mem_dummy[2][2][128];
  double mem_dummy_[2][128][3];
  enum yac_edge_type edge_dummy[2][128];
  struct grid_cell test_cells[2] ={{.coordinates_x = mem_dummy[0][0],
                                    .coordinates_y = mem_dummy[0][1],
                                    .coordinates_xyz = mem_dummy_[0],
                                    .edge_type = edge_dummy[0],
                                    .num_corners = cells[0].num_corners},
                                   {.coordinates_x = mem_dummy[1][0],
                                    .coordinates_y = mem_dummy[1][1],
                                    .coordinates_xyz = mem_dummy_[1],
                                    .edge_type = edge_dummy[1],
                                    .num_corners = cells[1].num_corners}},
                   overlap_cell;

  for (int i = 0; i < (int)num_ref_cells; ++i)
    generate_XYZ(ref_cells[i]);

  yac_init_grid_cell(&overlap_cell);

  for (order[0] = -1; order[0] <= 1; order[0] += 2) {
    for (order[1] = -1; order[1] <= 1; order[1] += 2) {
      for (start[0] = 0; start[0] < (int)(cells[0].num_corners); ++start[0]) {
        for (start[1] = 0; start[1] < (int)(cells[1].num_corners); ++start[1]) {

          for (int k = 0; k <= 1; ++k) {
            for (int i = 0; i < ((int)(cells[k].num_corners)); ++i) {
              int j =
                (((int)(cells[k].num_corners))+start[k]+i*order[k])%
                ((int)(cells[k].num_corners));
              test_cells[k].coordinates_x[i] = cells[k].coordinates_x[j];
              test_cells[k].coordinates_y[i] = cells[k].coordinates_y[j];
              j =
                (((int)(cells[k].num_corners)) + j - (order[k] < 0))%
                ((int)(cells[k].num_corners));
              test_cells[k].edge_type[i] = cells[k].edge_type[j];
            }
            generate_XYZ(test_cells[k]);
          }

          for (int k = 0; k <= 1; ++k) {
            yac_cell_clipping(1, test_cells + k, test_cells[k^1], &overlap_cell);
            int match = 0;
            for (int i = 0; i < (int)num_ref_cells; ++i)
              match |= !compare_cells(overlap_cell, ref_cells[i]);
            if (!match)
              PUT_ERR("ERROR: wrong clipping cell\n");
          }
        }
      }
    }
  }
  yac_free_grid_cell(&overlap_cell);
}

